------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Strings.Unbounded;  use Ada.Strings.Unbounded;
with Unbounded_Strings;      use Unbounded_Strings;
use Unbounded_Strings.Unbounded_String_List_Package;
with Call_Framework;         use Call_Framework;
with GNAT.Command_Line;      use GNAT.Command_Line;
with GNAT.OS_Lib;            use GNAT.OS_Lib;
with Ada.IO_Exceptions;      use Ada.IO_Exceptions;
with GNAT.Current_Exception; use GNAT.Current_Exception;
with Systems;                use Systems;
with Task_Set;               use Task_Set;
with Processor_Set;          use Processor_Set;
with Buffer_Set;             use Buffer_Set;
with Resource_Set;           use Resource_Set;
with Address_Space_Set;      use Address_Space_Set;
with Xml_Parsers;            use Xml_Parsers;
with AADL_Parsers;           use AADL_Parsers;
with Text_IO;                use Text_IO;
with Version;                use Version;
with Project;                use Project;

procedure Aadl2xml is

   File_Name : constant Unbounded_String_Ptr := new Unbounded_String;

   procedure Usage is
   begin
      Put_Line(
"aadl2xml is a program which translate an AADL specification into an XML Cheddar's project file."
);

      New_Line;
      Put_Line(
"Check Cheddar home page for details :  http://beru.univ-brest.fr/~singhoff/cheddar "
);
      New_Line;
      New_Line;
      Put_Line ("Usage : aadl2xml [switch] aadl-file-to-translate");
      Put_Line ("   switch can be :");
      Put_Line ("            -u get this help");
      Put_Line ("            -v verbose mode ");
      Put_Line ("            -o file-name, write XML file into file-name ");
      New_Line;
   end Usage;

   -- Is the command should run in verbose mode ?
   --
   Verbose : Boolean := False;

   -- The XML to create
   --
   Output_File      : Boolean := False;
   Output_File_Name : Unbounded_String;

   -- The resulting system
   --
   Sys : System;

begin

   Copyright ("aadl2xml");

   -- Get arguments
   --
   loop
      case Getopt ("u v o:") is
         when ASCII.NUL =>
            exit;

         when 'o' =>
            Output_File      := True;
            Output_File_Name := To_Unbounded_String (Parameter);
         when 'v' =>
            Verbose := True;
         when 'u' =>
            Usage;
            OS_Exit (0);

         when others =>
            Usage;
            OS_Exit (0);
      end case;
   end loop;

   loop
      declare
         S : constant String := Get_Argument (Do_Expansion => True);

      begin
         exit when S'Length = 0;
         File_Name.all := File_Name.all & S;
      end;
   end loop;
   Add (Aadl_Project_File_Name, File_Name);

   Call_Framework.initialize;

   if (File_Name.all = "") then
      Usage;
      OS_Exit (0);
   end if;

   -- Transform  the file
   --
   if Verbose then
      Put ("Transforming the file ");
      Put (To_String (File_Name.all) & " ....");
   end if;
   New_Line;
   New_Line;
   New_Line;
   New_Line;

   Read_From_Aadl_File (Sys, Project_File_Dir_List, Aadl_Project_File_Name);

   if Output_File then
      Write_To_Xml_File (Sys, Output_File_Name);
   else
      Put_Xml (Sys);
   end if;

exception
   when Task_Set.Invalid_Parameter =>
      Put_Line
        ("aadl2xml.adb : invalid task argument ; " & Exception_Message);
   when Address_Space_Set.Invalid_Parameter =>
      Put_Line
        ("aadl2xml.adb : invalid addresss space argument ; " &
         Exception_Message);
   when Processor_Set.Invalid_Parameter =>
      Put_Line
        ("aadl2xml.adb : invalid processor argument ; " & Exception_Message);
   when Buffer_Set.Invalid_Parameter =>
      Put_Line
        ("aadl2xml.adb : invalid buffer argument ; " & Exception_Message);
   when Resource_Set.Invalid_Parameter =>
      Put_Line
        ("aadl2xml.adb : invalid resource argument ; " & Exception_Message);

   when GNAT.Command_Line.Invalid_Switch =>
      begin
         Put_Line ("aadl2xml.adb : invalid Switch ; " & Full_Switch);
         Usage;
      end;
   when GNAT.Command_Line.Invalid_Parameter =>
      begin
         Put_Line
           ("aadl2xml.adb : missing parameter for switch ; " & Full_Switch);
         Usage;
      end;

   when Xml_Read_Error =>
      Put_Line
        ("aadl2xml.adb : Can not read project file ; Xml Fatal Error ; " &
         Exception_Message);
   when AADL_Read_Error =>
      Put_Line
        ("aadl2xml.adb : Can not read project files ; AADL parsing Error ; " &
         Exception_Message);

   when Ada.IO_Exceptions.Name_Error =>
      Put_Line ("aadl2xml.adb : Can not open project files, Name_Error");
   when Ada.IO_Exceptions.Status_Error =>
      Put_Line ("aadl2xml.adb : Can not open project files, Status_Error");
   when Ada.IO_Exceptions.Mode_Error =>
      Put_Line ("aadl2xml.adb : Can not open project files, Mode_Error");
   when Ada.IO_Exceptions.Use_Error =>
      Put_Line ("aadl2xml.adb : Can not open project files, Use_Error");
   when Ada.IO_Exceptions.Device_Error =>
      Put_Line ("aadl2xml.adb : Can not open project files, Device_Error");
   when Ada.IO_Exceptions.End_Error =>
      Put_Line ("aadl2xml.adb : Can not open project files, End_Error");
   when Ada.IO_Exceptions.Data_Error =>
      Put_Line ("aadl2xml.adb : Can not open project files, Data_Error");
   when Ada.IO_Exceptions.Layout_Error =>
      Put_Line ("aadl2xml.adb : Can not open project files, Layout_Error");

   when others =>
      Put_Line ("This is an internal aadl2xml bug ... sorry");
      Put_Line ("Exception name : " & Exception_Name);
      Put_Line ("Exception message : " & Exception_Message);
      Put_Line ("Please, send a bug report to cheddar@listes.univ-brest.fr");
      Put_Line(
"Do not forget to join the AADL/XML Cheddar project files with your bug report"
);

end Aadl2xml;
