------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Gtk;                    use Gtk;
with Gtk.Main;
with Ada.IO_Exceptions;      use Ada.IO_Exceptions;
with GNAT.Current_Exception; use GNAT.Current_Exception;
with Ada.Strings.Unbounded;  use Ada.Strings.Unbounded;
with Version;                use Version;
with Framework_Config;       use Framework_Config;
with Project;                use Project;
with Systems;                use Systems;
with Task_Set;               use Task_Set;
with Processor_Set;          use Processor_Set;
with Buffer_Set;             use Buffer_Set;
with Resource_Set;           use Resource_Set;
with Address_Space_Set;      use Address_Space_Set;
with Xml_Parsers;            use Xml_Parsers;
with AADL_Parsers;           use AADL_Parsers;
with Root_Window_Pkg;        use Root_Window_Pkg;
with Graph_Window_Pkg;       use Graph_Window_Pkg;
with My_Canvas;              use My_Canvas;
with Gtk.Widget;             use Gtk.Widget;
with Gdk.Font;               use Gdk.Font;
with Gtk.Style;              use Gtk.Style;
with GNAT.Command_Line;      use GNAT.Command_Line;
with GNAT.OS_Lib;            use GNAT.OS_Lib;
with Text_IO;                use Text_IO;
with Unbounded_Strings;      use Unbounded_Strings;
use Unbounded_Strings.Unbounded_String_List_Package;
with Debug;                  use Debug;
with Call_Framework;         use Call_Framework;

procedure Cheddar is

   procedure Usage is
   begin
      Put_Line ("Cheddar is a free real time scheduling simulator.");
      Put_Line(
"This program provides services to automatically check temporal constraints of real time tasks."
);
      New_Line;
      Put_Line(
"Check Cheddar home page for details :  http://beru.univ-brest.fr/~singhoff/cheddar "
);
      New_Line;
      New_Line;
      Put_Line ("Usage : cheddar [switches] foo");
      Put_Line ("      where foo can be : ");
      Put_Line ("            - an unique XML file ");
      Put_Line ("            - one or several AADL files");
      New_Line;
      Put_Line ("   Switches can be :");

      Put_Line ("            -u get this help");

      Put_Line(
"            -l select a language : ""fr"" for franais; ""en"" for english ");
      Put ("               Default language is ");
      if (Current_Language = English) then
         Put ("English.");
      else
         Put ("Franais.");
      end if;
      New_Line;

      Put(
"            -a the file names given to cheddar contain AADL descriptions instead of an XML file"
);
      Put_Line(
" Only one XML file can be provided to Cheddar but several AADL files can be sent to Cheddar."
);

      Put_Line(
"            -I directory-name : gives an extra directory name where to look for XML or AADL files. By default, Cheddar only looks for project files into the Cheddar's current directory."
);
      Put_Line ("            -d activate Cheddar's debug mode ");
      Put_Line
        ("            -c put the current Cheddar's configuration at the screen"
);

   end Usage;

   My_Style          : Gtk_Style;
   Global_Font       : Gdk_Font;
   Aadl_File         : Boolean;
   A_Dir_Name        : Unbounded_String_Ptr;
   Project_File_Name : Unbounded_String_Ptr;
   Project_File_List : Unbounded_String_List;

begin

   Copyright ("cheddar");

   -- New project
   --
   Aadl_File := False;
   Initialize (Project_File_List);
   Do_New;

   -- Get arguments
   --
   loop
      case Getopt ("c d u l: a I:") is
         when ASCII.NUL =>
            exit;
         when 'c' =>
            Framework_Config.Put;

         when 'a' =>
            Aadl_File := True;

         when 'd' =>
            Cheddar_Debug := True;

         when 'I' =>
            A_Dir_Name     := new Unbounded_String;
            A_Dir_Name.all := To_Unbounded_String (Parameter);
            Put_Debug
              ("Project file directory : " & To_String (A_Dir_Name.all));
            Add (Project_File_Dir_List, A_Dir_Name);

         when 'u' =>
            Usage;
            OS_Exit (0);

         when 'l' =>
            if (Parameter /= "fr") and (Parameter /= "en") then
               Put_Line ("cheddar.adb : invalid language parameter");
               New_Line;
               Usage;
               OS_Exit (0);
            else
               if Parameter = "fr" then
                  Current_Language := Francais;
               else
                  Current_Language := English;
               end if;
            end if;

         when others =>
            Usage;
            OS_Exit (0);
      end case;
   end loop;

   loop
      Project_File_Name     := new Unbounded_String;
      Project_File_Name.all :=
         To_Unbounded_String (Get_Argument (Do_Expansion => True));

      exit when Project_File_Name.all = Empty_String;

      Put_Debug
        ("Project file to load : " & To_String (Project_File_Name.all));

      Add (Project_File_List, Project_File_Name);
   end loop;

   if (Get_Number_Of_Elements (Project_File_List) > 1) and
      (not Aadl_File)
   then
      Put_Line
        ("cheddar.adb : can not load more than one XML file at the same time ")
;
      OS_Exit (0);
   end if;

   Call_Framework.initialize;

   if (Get_Number_Of_Elements (Project_File_List) > 0) then
      Is_New   := False;
      Is_Saved := True;
      if (Aadl_File) then
         Duplicate (Project_File_List, Aadl_Project_File_Name);
         Systems.Read_From_Aadl_File
           (Sys,
            Project_File_Dir_List,
            Aadl_Project_File_Name);

      --
      -- The piece of code below allows us to call aadl2xml externally instead
      --of
      -- directly calling the package aadl_parser.ads
      --
      --         declare
      --            Ok                        : Boolean;
      --            Spawn_Args                : String_List (1 .. 2);
      --            Temp_File_Name            : aliased String       := "-o
      --cheddar_temp_file.xml";
      --            Aadl_File_Name_To_Perform : aliased String       :=
      --To_String
      --              (Aadl_Project_File_Name);
      --         begin
      --            Spawn_Args(1) := Temp_File_Name'Unchecked_Access;
      --            Spawn_Args(2):=Aadl_File_Name_To_Perform'Unchecked_Access;
      --            Spawn("aadl2xml", Spawn_Args, Ok);
      --            if not Ok then
      --               Put_Line("marche pas");
      --            end if;
      --         end;

      else
         Project_File_Name     := Get_Head (Project_File_List);
         Xml_Project_File_Name := Project_File_Name.all;
         Systems.Read_From_Xml_File
           (Sys,
            Project_File_Dir_List,
            Xml_Project_File_Name);
      end if;
   end if;

   -- Initialize the graphical editor
   --
   Gtk.Main.Set_Locale;
   Gtk.Main.Init;

   Load (Global_Font, "-*-*-*-*-*-*-*-*-*-*-*-*-iso8859-1");
   if Global_Font = Null_Font then
      Put_Line ("cheddar.adb : warning, can not load ISO 8859-1 font ");
   end if;

   Gtk_New (My_Style);
   Set_Font (My_Style, Global_Font);
   Set_Default_Style (My_Style);

   Gtk_New (Root_Window);
   Show_All (Root_Window);

   -- Graphical editor event loop
   --
   Gtk.Main.Main;

exception

   when Task_Set.Invalid_Parameter =>
      Put_Line ("cheddar.adb : invalid task argument ; " & Exception_Message);
   when Address_Space_Set.Invalid_Parameter =>
      Put_Line
        ("cheddar.adb : invalid addresss space argument ; " &
         Exception_Message);
   when Processor_Set.Invalid_Parameter =>
      Put_Line
        ("cheddar.adb : invalid processor argument ; " & Exception_Message);
   when Buffer_Set.Invalid_Parameter =>
      Put_Line
        ("cheddar.adb : invalid buffer argument ; " & Exception_Message);
   when Resource_Set.Invalid_Parameter =>
      Put_Line
        ("cheddar.adb : invalid resource argument ; " & Exception_Message);

   when GNAT.Command_Line.Invalid_Switch =>
      begin
         Put_Line ("cheddar.adb : invalid Switch ; " & Full_Switch);
         Usage;
      end;
   when GNAT.Command_Line.Invalid_Parameter =>
      begin
         Put_Line
           ("cheddar.adb : missing parameter for switch ; " & Full_Switch);
         Usage;
      end;

   when Xml_Read_Error =>
      Put_Line
        ("cheddar.adb : Can not read project file ; Xml Fatal Error ; " &
         Exception_Message);
   when AADL_Read_Error =>
      Put_Line
        ("cheddar.adb : Can not read project files ; AADL parsing Error ; " &
         Exception_Message);

   when Ada.IO_Exceptions.Name_Error =>
      Put_Line ("cheddar.adb : Can not open project files, Name_Error");
   when Ada.IO_Exceptions.Status_Error =>
      Put_Line ("cheddar.adb : Can not open project files, Status_Error");
   when Ada.IO_Exceptions.Mode_Error =>
      Put_Line ("cheddar.adb : Can not open project files, Mode_Error");
   when Ada.IO_Exceptions.Use_Error =>
      Put_Line ("cheddar.adb : Can not open project files, Use_Error");
   when Ada.IO_Exceptions.Device_Error =>
      Put_Line ("cheddar.adb : Can not open project files, Device_Error");
   when Ada.IO_Exceptions.End_Error =>
      Put_Line ("cheddar.adb : Can not open project files, End_Error");
   when Ada.IO_Exceptions.Data_Error =>
      Put_Line ("cheddar.adb : Can not open project files, Data_Error");
   when Ada.IO_Exceptions.Layout_Error =>
      Put_Line ("cheddar.adb : Can not open project files, Layout_Error");

   when others =>
      Put_Line ("This is an internal Cheddar bug ... sorry");
      Put_Line(
"WARNING           : your project was saved in the file cheddar_bug.xml");
      Put_Line ("Exception name    : " & Exception_Name);
      Put_Line ("Exception message : " & Exception_Message);

      if Graph_Window /= null then
         Get_Coordinates (Graph_Window.Buffer, Sys);
      end if;
      Systems.Write_To_Xml_File
        (Sys,
         To_Unbounded_String ("cheddar_bug.xml"));

      Put_Line ("Please, send a bug report to cheddar@listes.univ-brest.fr");
      Put_Line(
"Do not forget to join your AADL/XML/SC Cheddar project files with your bug report"
);

end Cheddar;
