------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Address_Spaces;
USE Address_Spaces;
WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Unbounded_Strings;
USE Unbounded_Strings;
WITH Framework_Config;
USE Framework_Config;
WITH Scheduler;
USE Scheduler;
WITH Task_Set;
USE Task_Set;
WITH Resource_Set;
USE Resource_Set;
WITH Sets;




PACKAGE Address_Space_Set IS





   ----------------------------------------------------------
   -- Definition of a set
   ----------------------------------------------------------


   PACKAGE Generic_Address_Space_Set IS NEW Sets (
      Max_Element => Framework_Config.Max_Address_Space,
      Element     => Address_Space_Ptr,
      Free        => Free,
      Copy        => Copy,
      Put         => Put);



   USE Generic_Address_Space_Set;

   TYPE Address_Spaces_Set IS NEW Generic_Address_Space_Set.Set WITH PRIVATE;


   SUBTYPE Address_Spaces_Range IS Generic_Address_Space_Set.
      Element_Range;
   SUBTYPE Address_Spaces_Iterator IS   Generic_Address_Space_Set.
      Iterator;




   ----------------------------------------------------------
   -- Exceptions defined in the set package
   ----------------------------------------------------------

   -- Raised when a given address space is not in a address space set
   --
   Address_Space_Not_Found : EXCEPTION;

   -- Raised when parameters provided to Add_Address_Space are wrong
   --
   Invalid_Parameter : EXCEPTION;



   ----------------------------------------------------------
   -- I/O operations
   ----------------------------------------------------------
   FUNCTION Export_Aadl_Declarations (
         My_Address_Spaces : IN     Address_Spaces_Set;
         Number_Of_Ht      : IN     Natural)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_Implementations (
         My_Address_Spaces : IN     Address_Spaces_Set;
         My_Tasks          : IN     Tasks_Set;
         My_Resources      : IN     Resources_Set)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_Properties (
         My_Address_Spaces : IN     Address_Spaces_Set;
         Number_Of_Ht      : IN     Natural)
     RETURN Unbounded_String;

   FUNCTION Export_Xml (
         My_Address_Spaces : IN     Address_Spaces_Set)
     RETURN Unbounded_String;




   ----------------------------------------------------------
   -- Procedure to proceed modification on a set
   ----------------------------------------------------------


   PROCEDURE Check_Address_Space (
         My_Address_Spaces : IN     Address_Spaces_Set;
         Name              : IN     Unbounded_String;
         Cpu_Name          : IN     Unbounded_String;
         Text_Memory_Size  : IN     Integer;
         Stack_Memory_Size : IN     Integer;
         Data_Memory_Size  : IN     Integer;
         Heap_Memory_Size  : IN     Integer;
         Is_Preemptive     : IN     Preemptive_Type    := Preemptive;
         Quantum           : IN     Integer            := 0;
         File_Name         : IN     Unbounded_String   := Empty_String;
         A_Scheduler       : IN     Scheduler_Type     := No_Scheduling_Protocol);


   PROCEDURE Add_Address_Space (
         My_Address_Spaces : IN OUT Address_Spaces_Set;
         Name              : IN     Unbounded_String;
         Cpu_Name          : IN     Unbounded_String;
         Text_Memory_Size  : IN     Integer;
         Stack_Memory_Size : IN     Integer;
         Data_Memory_Size  : IN     Integer;
         Heap_Memory_Size  : IN     Integer;
         Is_Preemptive     : IN     Preemptive_Type    := Preemptive;
         Quantum           : IN     Integer            := 0;
         File_Name         : IN     Unbounded_String   := Empty_String;
         A_Scheduler       : IN     Scheduler_Type     := No_Scheduling_Protocol);


   PROCEDURE Add_Address_Space (
         My_Address_Spaces : IN OUT Address_Spaces_Set;
         A_Address_Space   : IN OUT Address_Space_Ptr;
         Name              : IN     Unbounded_String;
         Cpu_Name          : IN     Unbounded_String;
         Text_Memory_Size  : IN     Integer;
         Stack_Memory_Size : IN     Integer;
         Data_Memory_Size  : IN     Integer;
         Heap_Memory_Size  : IN     Integer;
         Is_Preemptive     : IN     Preemptive_Type    := Preemptive;
         Quantum           : IN     Integer            := 0;
         File_Name         : IN     Unbounded_String   := Empty_String;
         A_Scheduler       : IN     Scheduler_Type     := No_Scheduling_Protocol);


   PROCEDURE Delete_Processor (
         My_Address_Spaces : IN OUT Address_Spaces_Set;
         A_Processor       : IN     Unbounded_String);



   ------------------------------------------------------
   -- Read information from a set
   ------------------------------------------------------


   FUNCTION Search_Address_Space (
         My_Address_Spaces : IN     Address_Spaces_Set;
         Name              : IN     Unbounded_String)
     RETURN Address_Space_Ptr;




PRIVATE

   TYPE Address_Spaces_Set IS NEW Generic_Address_Space_Set.Set WITH NULL RECORD;

END Address_Space_Set;









