------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Exceptions;    use Ada.Exceptions;
with Translate;         use Translate;
with Objects;           use Objects;
with Objects.extended;  use Objects.extended;
with Unbounded_Strings; use Unbounded_Strings;

package body Buffer_Set is

   procedure Add_Buffer
     (My_Buffers         : in out Buffers_Set;
      Name               : in Unbounded_String;
      Size               : in Integer;
      Cpu_Name           : in Unbounded_String;
      Address_Space_Name : in Unbounded_String;
      A_Qs               : in Queueing_Systems_Type;
      Roles              : in Buffer_Roles_Table)
   is

      Dummy : Buffer_Ptr;

   begin
      Add_Buffer
        (My_Buffers,
         Dummy,
         Name,
         Size,
         Cpu_Name,
         Address_Space_Name,
         A_Qs,
         Roles);
   end Add_Buffer;

   procedure Check_Buffer
     (My_Buffers         : in Buffers_Set;
      Name               : in Unbounded_String;
      Size               : in Integer;
      Cpu_Name           : in Unbounded_String;
      Address_Space_Name : in Unbounded_String)
   is

   begin

      if (Name = "") then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Buffer_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

      if not Is_A_Valid_Identifier (Name) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Buffer (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Buffer_Name (Current_Language) &
                Lb_Colon &
                Lb_Invalid_Identifier (Current_Language)));
      end if;

      if (Cpu_Name = "") then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Buffer (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Processor_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

      if (Address_Space_Name = "") then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Buffer (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Address_Space_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

      if (Size <= 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Buffer (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Size (Current_Language) &
                Lb_Must_Be (Current_Language) &
                ">0"));
      end if;

   end Check_Buffer;

   procedure Add_Buffer
     (My_Buffers         : in out Buffers_Set;
      A_Buffer           : in out Buffer_Ptr;
      Name               : in Unbounded_String;
      Size               : in Integer;
      Cpu_Name           : in Unbounded_String;
      Address_Space_Name : in Unbounded_String;
      A_Qs               : in Queueing_Systems_Type;
      Roles              : in Buffer_Roles_Table)
   is

      My_Iterator : Iterator;

   begin

      Check_Buffer (My_Buffers, Name, Size, Cpu_Name, Address_Space_Name);

      if (Get_Number_Of_Elements (My_Buffers) > 0) then

         Reset_Iterator (My_Buffers, My_Iterator);

         loop
            Current_Element (My_Buffers, A_Buffer, My_Iterator);
            if (Name = A_Buffer.Name) then
               Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_Buffer (Current_Language) &
                      " " &
                      Name &
                      " : " &
                      Lb_Buffer_Name (Current_Language) &
                      Lb_Already_Defined (Current_Language)));
            end if;

            exit when Is_Last_Element (My_Buffers, My_Iterator);

            Next_Element (My_Buffers, My_Iterator);
         end loop;
      end if;

      A_Buffer                      := new Buffer;
      A_Buffer.Name                 := Name;
      A_Buffer.Cpu_Name             := Cpu_Name;
      A_Buffer.Address_Space_Name   := Address_Space_Name;
      A_Buffer.Size                 := Size;
      A_Buffer.Roles                := Roles;
      A_Buffer.Queueing_System_Type := A_Qs;

      Add (My_Buffers, A_Buffer);

   end Add_Buffer;

   procedure Update_Buffer
     (My_Buffers         : in out Buffers_Set;
      Name               : in Unbounded_String;
      New_Name           : in Unbounded_String;
      Size               : in Integer;
      Cpu_Name           : in Unbounded_String;
      Address_Space_Name : in Unbounded_String;
      Roles              : in Buffer_Roles_Table)
   is
      The_Buffer : Buffer_Ptr;
   begin
      The_Buffer                    := Search_Buffer (My_Buffers, Name);
      The_Buffer.Name               := New_Name;
      The_Buffer.Address_Space_Name := Address_Space_Name;
      The_Buffer.Cpu_Name           := Cpu_Name;
      The_Buffer.Size               := Size;
      The_Buffer.Roles              := Roles;

   end Update_Buffer;

   function Get
     (My_Buffers  : in Buffers_Set;
      Buffer_Name : in Unbounded_String;
      Param_Name  : in Buffer_Parameters)
      return        Natural
   is
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Size) then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Buffers, My_Iterator);

      loop
         Current_Element (My_Buffers, A_Buffer, My_Iterator);

         if (A_Buffer.Name = Buffer_Name) then
            exit;
         end if;

         exit when Is_Last_Element (My_Buffers, My_Iterator);

         Next_Element (My_Buffers, My_Iterator);

      end loop;

      return A_Buffer.Size;

   end Get;

   function Get
     (My_Buffers  : in Buffers_Set;
      Buffer_Name : in Unbounded_String;
      Param_Name  : in Buffer_Parameters)
      return        Unbounded_String
   is
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Cpu_Name) and
         (Param_Name /= Address_Space_Name)
      then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Buffers, My_Iterator);

      loop
         Current_Element (My_Buffers, A_Buffer, My_Iterator);

         if (A_Buffer.Name = Buffer_Name) then
            exit;
         end if;

         exit when Is_Last_Element (My_Buffers, My_Iterator);

         Next_Element (My_Buffers, My_Iterator);

      end loop;

      if (Param_Name /= Cpu_Name) then
         return A_Buffer.Cpu_Name;
      else
         return A_Buffer.Address_Space_Name;
      end if;

   end Get;

   procedure Set
     (My_Buffers  : in out Buffers_Set;
      Buffer_Name : in Unbounded_String;
      Param_Name  : in Buffer_Parameters;
      Param_Value : in Natural)
   is
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Size) then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Buffers, My_Iterator);

      loop
         Current_Element (My_Buffers, A_Buffer, My_Iterator);

         if (A_Buffer.Name = Buffer_Name) then
            A_Buffer.Size := Param_Value;
            exit;
         end if;

         exit when Is_Last_Element (My_Buffers, My_Iterator);

         Next_Element (My_Buffers, My_Iterator);

      end loop;
   end Set;

   procedure Set
     (My_Buffers  : in out Buffers_Set;
      Buffer_Name : in Unbounded_String;
      Param_Name  : in Buffer_Parameters;
      Param_Value : in Unbounded_String)
   is
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Cpu_Name) and
         (Param_Name /= Address_Space_Name)
      then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Buffers, My_Iterator);

      loop
         Current_Element (My_Buffers, A_Buffer, My_Iterator);

         if (A_Buffer.Name = Buffer_Name) then
            if (Param_Name /= Cpu_Name) then
               A_Buffer.Cpu_Name := Param_Value;
            else
               A_Buffer.Address_Space_Name := Param_Value;
            end if;
            exit;
         end if;

         exit when Is_Last_Element (My_Buffers, My_Iterator);

         Next_Element (My_Buffers, My_Iterator);

      end loop;
   end Set;

   function Search_Buffer
     (My_Buffers : in Buffers_Set;
      Name       : in Unbounded_String)
      return       Buffer_Ptr
   is
      My_Iterator : Iterator;
      A_Buffer    : Buffer_Ptr;
      Result      : Buffer_Ptr;

      Found : Boolean := False;

   begin

      Reset_Iterator (My_Buffers, My_Iterator);

      loop
         Current_Element (My_Buffers, A_Buffer, My_Iterator);

         if (A_Buffer.Name = Name) then
            Found  := True;
            Result := A_Buffer;
         end if;

         exit when Is_Last_Element (My_Buffers, My_Iterator);

         Next_Element (My_Buffers, My_Iterator);

      end loop;

      if not Found then
         raise Buffer_Not_Found;
      end if;

      return Result;

   end Search_Buffer;

   procedure Delete_Task
     (My_Buffers : in out Buffers_Set;
      A_Task     : in Unbounded_String)
   is

      A_Buffer    : Buffer_Ptr;
      My_Iterator : Buffers_Iterator;
      Nb_Entries  : Buffer_Roles_Range;

   begin

      if (Get_Number_Of_Elements (My_Buffers) > 0) then

         Reset_Iterator (My_Buffers, My_Iterator);
         loop
            Current_Element (My_Buffers, A_Buffer, My_Iterator);

            -- Looking for a_task in roles
            --
            Nb_Entries := A_Buffer.Roles.Nb_Entries;
            for Index in  0 .. Nb_Entries - 1 loop
               if (A_Buffer.Roles.Entries (Index).Item = A_Task) then
                  A_Buffer.Roles.Entries (Index) :=
                    A_Buffer.Roles.Entries (A_Buffer.Roles.Nb_Entries - 1);

                  A_Buffer.Roles.Nb_Entries := A_Buffer.Roles.Nb_Entries - 1;
               end if;
            end loop;

            exit when Is_Last_Element (My_Buffers, My_Iterator);
            Next_Element (My_Buffers, My_Iterator);
         end loop;

      end if;

   end Delete_Task;

   procedure Delete_Address_Space
     (My_Buffers : in out Buffers_Set;
      A_Addr     : in Unbounded_String)
   is

      Tmp         : Buffers_Set;
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Buffers_Iterator;

   begin

      if (Get_Number_Of_Elements (My_Buffers) > 0) then

         Reset_Iterator (My_Buffers, My_Iterator);
         loop
            Current_Element (My_Buffers, A_Buffer, My_Iterator);

            if (A_Buffer.Address_Space_Name = A_Addr) then
               Add (Tmp, A_Buffer);
            end if;

            exit when Is_Last_Element (My_Buffers, My_Iterator);
            Next_Element (My_Buffers, My_Iterator);
         end loop;

         if not Is_Empty (Tmp) then
            Reset_Iterator (Tmp, My_Iterator);
            loop
               Current_Element (Tmp, A_Buffer, My_Iterator);

               Delete (My_Buffers, A_Buffer);

               exit when Is_Last_Element (Tmp, My_Iterator);
               Next_Element (Tmp, My_Iterator);
            end loop;
         end if;

      end if;
   end Delete_Address_Space;

   function Get_Number_Of_Buffer_From_Processor
     (My_Buffers     : in Buffers_Set;
      Processor_Name : in Unbounded_String)
      return           Buffers_Range
   is
      Number      : Buffers_Range := 0;
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   begin

      if Is_Empty (My_Buffers) then
         return 0;
      end if;

      Reset_Iterator (My_Buffers, My_Iterator);

      loop
         Current_Element (My_Buffers, A_Buffer, My_Iterator);

         if (A_Buffer.Cpu_Name = Processor_Name) then
            Number := Number + 1;
         end if;

         exit when Is_Last_Element (My_Buffers, My_Iterator);

         Next_Element (My_Buffers, My_Iterator);

      end loop;

      return Number;
   end Get_Number_Of_Buffer_From_Processor;

   procedure Delete_Processor
     (My_Buffers  : in out Buffers_Set;
      A_Processor : in Unbounded_String)
   is

      Tmp         : Buffers_Set;
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Buffers_Iterator;

   begin

      if (Get_Number_Of_Elements (My_Buffers) > 0) then

         Reset_Iterator (My_Buffers, My_Iterator);
         loop
            Current_Element (My_Buffers, A_Buffer, My_Iterator);

            if (A_Buffer.Cpu_Name = A_Processor) then
               Add (Tmp, A_Buffer);
            end if;

            exit when Is_Last_Element (My_Buffers, My_Iterator);
            Next_Element (My_Buffers, My_Iterator);
         end loop;

         if not Is_Empty (Tmp) then
            Reset_Iterator (Tmp, My_Iterator);
            loop
               Current_Element (Tmp, A_Buffer, My_Iterator);

               Delete (My_Buffers, A_Buffer);

               exit when Is_Last_Element (Tmp, My_Iterator);
               Next_Element (Tmp, My_Iterator);
            end loop;

            Reset (Tmp, False);
         end if;
      end if;

   end Delete_Processor;

   function Export_Xml
     (My_Buffers : in Buffers_Set)
      return       Unbounded_String
   is
      My_Iterator : Buffers_Iterator;
      A_Buffer    : Buffer_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Buffers) then

         Result := Result &
                   To_Unbounded_String ("<buffers>") &
                   Unbounded_Lf;

         Reset_Iterator (My_Buffers, My_Iterator);

         loop
            Current_Element (My_Buffers, A_Buffer, My_Iterator);

            Result := Result & To_Unbounded_String (ASCII.HT & "<buffer");
            if A_Buffer.X /= 0 then
               Result := Result &
                         To_Unbounded_String
                            (" x=""" & A_Buffer.X'Img & """ ");
            end if;
            if A_Buffer.Y /= 0 then
               Result := Result &
                         To_Unbounded_String
                            ("y=""" & A_Buffer.Y'Img & """ ");
            end if;
            Result := Result & To_Unbounded_String (">") & Unbounded_Lf;

            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<cpu_name>" &
                  To_String (A_Buffer.Cpu_Name) &
                  "</cpu_name>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<address_space_name>" &
                  To_String (A_Buffer.Address_Space_Name) &
                  "</address_space_name>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<name>" &
                  To_String (A_Buffer.Name) &
                  "</name>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<size>" &
                  A_Buffer.Size'Img &
                  "</size>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<qs>" &
                  A_Buffer.Queueing_System_Type'Img &
                  "</qs>") &
              Unbounded_Lf;

            Result := Result &
                      To_Unbounded_String
                         (ASCII.HT & ASCII.HT & "<buffer_used_by>") &
                      Unbounded_Lf;
            for I in  0 .. A_Buffer.Roles.Nb_Entries - 1 loop
               Result := Result &
                         To_Unbounded_String
                            (ASCII.HT & ASCII.HT & ASCII.HT & "<buffer_user");
               if A_Buffer.Roles.Entries (I).Data.The_Role = Consumer then
                  Result := Result &
                            To_Unbounded_String
                               (" buffer_role=""consumer"">");
               else
                  Result := Result & To_Unbounded_String (">");
               end if;
               Result := Result &
                         To_Unbounded_String
                            (" " &
                             To_String (A_Buffer.Roles.Entries (I).Item) &
                             " ");
               Result := Result &
                         To_Unbounded_String
                            (" " &
                             A_Buffer.Roles.Entries (I).Data.Size'Img &
                             " ");
               Result := Result &
                         To_Unbounded_String
                            (" " &
                             A_Buffer.Roles.Entries (I).Data.Time'Img &
                             " ");

               Result := Result &
                         To_Unbounded_String ("</buffer_user>") &
                         Unbounded_Lf;
            end loop;
            Result := Result &
                      To_Unbounded_String
                         (ASCII.HT & ASCII.HT & "</buffer_used_by>") &
                      Unbounded_Lf;

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "</buffer>") &
                      Unbounded_Lf;

            exit when Is_Last_Element (My_Buffers, My_Iterator);

            Next_Element (My_Buffers, My_Iterator);

         end loop;
         Result := Result &
                   To_Unbounded_String ("</buffers>") &
                   Unbounded_Lf;
      end if;

      return Result;

   end Export_Xml;

   function Export_Aadl_Implementations
     (My_Buffers : in Buffers_Set)
      return       Unbounded_String
   is
      Result : Unbounded_String := Empty_String;

   begin

      return Result;

   end Export_Aadl_Implementations;

end Buffer_Set;
