------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Framework_Config;
USE Framework_Config;
WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Queueing_System;
USE Queueing_System;
with Buffers;          use Buffers;
use Buffers.Buffer_Roles;
WITH Sets;


PACKAGE Buffer_Set IS



   ----------------------------------------------------------
   -- Definition of a set
   ----------------------------------------------------------



   PACKAGE Generic_Buffer_Set IS NEW Sets (
      Max_Element => Framework_Config.Max_Buffers,
      Element     => Buffer_Ptr,
      Free        => Free,
      Copy        => Copy,
      Put         => Put);


   USE Generic_Buffer_Set;

   TYPE Buffers_Set IS NEW Generic_Buffer_Set.Set WITH PRIVATE;


   SUBTYPE Buffers_Range IS Generic_Buffer_Set.Element_Range;
   SUBTYPE Buffers_Iterator IS   Generic_Buffer_Set.Iterator;






   ----------------------------------------------------------
   -- Exceptions defined in the set package
   ----------------------------------------------------------

   -- Raised when a given buffer is not in a network set
   --
   Buffer_Not_Found : EXCEPTION;

   -- Raised when parameters provided to Add_buffer are wrong
   --
   Invalid_Parameter : EXCEPTION;




   ----------------------------------------------------------
   -- I/O operations
   ----------------------------------------------------------
   FUNCTION Export_Aadl_Implementations (
         My_Buffers : IN     Buffers_Set)
     RETURN Unbounded_String;

   FUNCTION Export_Xml (
         My_Buffers : IN     Buffers_Set)
     RETURN Unbounded_String;




   ----------------------------------------------------------
   -- Procedure to proceed modification on a set
   ----------------------------------------------------------


   PROCEDURE Check_Buffer (
         My_Buffers         : IN     Buffers_Set;
         Name               : IN     Unbounded_String;
         Size               : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String);


   PROCEDURE Add_Buffer (
         My_Buffers         : IN OUT Buffers_Set;
         A_Buffer           : IN OUT Buffer_Ptr;
         Name               : IN     Unbounded_String;
         Size               : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         A_Qs               : IN     Queueing_Systems_Type;
         Roles              : IN     Buffer_Roles_Table);


   PROCEDURE Add_Buffer (
         My_Buffers         : IN OUT Buffers_Set;
         Name               : IN     Unbounded_String;
         Size               : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         A_Qs               : IN     Queueing_Systems_Type;
         Roles              : IN     Buffer_Roles_Table);


   PROCEDURE Update_Buffer (
         My_Buffers         : IN OUT Buffers_Set;
         Name               : IN     Unbounded_String;
         New_Name           : IN     Unbounded_String;
         Size               : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Roles              : IN     Buffer_Roles_Table);


   PROCEDURE Delete_Address_Space (
         My_Buffers : IN OUT Buffers_Set;
         A_Addr     : IN     Unbounded_String);


   PROCEDURE Delete_Task (
         My_Buffers : IN OUT Buffers_Set;
         A_Task     : IN     Unbounded_String);


   PROCEDURE Delete_Processor (
         My_Buffers  : IN OUT Buffers_Set;
         A_Processor : IN     Unbounded_String);



   ------------------------------------------------------
   -- Read information from a set
   ------------------------------------------------------



   FUNCTION Search_Buffer (
         My_Buffers : IN     Buffers_Set;
         Name       : IN     Unbounded_String)
     RETURN Buffer_Ptr;


   FUNCTION Get_Number_Of_Buffer_From_Processor (
         My_Buffers     : IN     Buffers_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Buffers_Range;





   ----------------------------------------------------------
   -- List of parameters the user can set of get directly
   -- throught the set
   ----------------------------------------------------------




   TYPE Buffer_Parameters IS
         (Size,
          Cpu_Name,
          Address_Space_Name,
          Roles);



   FUNCTION Get (
         My_Buffers  : IN     Buffers_Set;
         Buffer_Name : IN     Unbounded_String;
         Param_Name  : IN     Buffer_Parameters)
     RETURN Unbounded_String;


   PROCEDURE Set (
         My_Buffers  : IN OUT Buffers_Set;
         Buffer_Name : IN     Unbounded_String;
         Param_Name  : IN     Buffer_Parameters;
         Param_Value : IN     Unbounded_String);


   FUNCTION Get (
         My_Buffers  : IN     Buffers_Set;
         Buffer_Name : IN     Unbounded_String;
         Param_Name  : IN     Buffer_Parameters)
     RETURN Natural;

   PROCEDURE Set (
         My_Buffers  : IN OUT Buffers_Set;
         Buffer_Name : IN     Unbounded_String;
         Param_Name  : IN     Buffer_Parameters;
         Param_Value : IN     Natural);



PRIVATE

   TYPE Buffers_Set IS NEW Generic_Buffer_Set.Set WITH NULL RECORD;




END Buffer_Set;
