------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Framework_Config;        use Framework_Config;
with Xml_Tag;                 use Xml_Tag;
with Scheduler;               use Scheduler;
use Scheduler.Double_Tasks_Parameters;
with Translate;               use Translate;
with Tasks;                   use Tasks;
with Task_Set;                use Task_Set;
use Task_Set.Generic_Task_Set;
with Unbounded_Strings;       use Unbounded_Strings;
with Dependency_Services;     use Dependency_Services;
with Multiprocessor_Services; use Multiprocessor_Services;
use Multiprocessor_Services.Scheduling_Result_Per_Processor;

package body Call_Dependency_Framework is

   procedure Chetto_Parameter_Modifications
     (Sys                : in out System;
      Result             : in out Unbounded_String;
      A_Processor        : in Processor_Ptr;
      Compute_Deadlines  : in Boolean;
      Compute_Priorities : in Boolean;
      Update_Tasks_Set   : in Boolean;
      Output             : in Output_Format := String_Output)
   is

      Tmp       : Tasks_Set;
      A_Task    : Generic_Task_Ptr;
      Iterator1 : Tasks_Iterator;

   begin

      if Compute_Deadlines then

         Result :=
           Result &
           Unbounded_Lf &
           Lb_Minus &
           Lb_Chetto_Deadline (Current_Language) &
           To_Unbounded_String (" (") &
           Lb_See (Current_Language) &
           To_Unbounded_String ("[5], [6]).") &
           Unbounded_Lf &
           Unbounded_Lf;

      else
         Result :=
           Result &
           Unbounded_Lf &
           Lb_Minus &
           Lb_Chetto_Priority (Current_Language) &
           To_Unbounded_String (" (") &
           Lb_See (Current_Language) &
           To_Unbounded_String ("[5], [6]).") &
           Unbounded_Lf &
           Unbounded_Lf;
      end if;

      -- If we must update parameters, we do not work on temporary tasks_set
      -- Call logic and display new parameters
      --
      if Update_Tasks_Set then
         Set_Parameters_According_To_Chetto
           (Sys.Dependencies,
            Sys.Tasks,
            Compute_Deadlines,
            Compute_Priorities);

         -- Display priorities/deadlines
         --
         Reset_Iterator (Sys.Tasks, Iterator1);
         loop

            Current_Element (Sys.Tasks, A_Task, Iterator1);
            Result := Result &
                      To_Unbounded_String ("- ") &
                      A_Task.Name &
                      To_Unbounded_String (" => ");
            if Compute_Deadlines then
               Result := Result &
                         To_Unbounded_String (A_Task.Deadline'Img) &
                         Unbounded_Lf;
            else
               Result := Result &
                         To_Unbounded_String (A_Task.Priority'Img) &
                         Unbounded_Lf;
            end if;

            exit when Is_Last_Element (Sys.Tasks, Iterator1);

            Next_Element (Sys.Tasks, Iterator1);
         end loop;

         if Compute_Priorities then
            Result := Result & Lb_Priorities_Are_Updated (Current_Language);
         else
            Result := Result & Lb_Deadlines_Are_Updated (Current_Language);
         end if;
         Result := Result & Unbounded_Lf & Unbounded_Lf & Unbounded_Lf;

      else
         Duplicate (Sys.Tasks, Tmp);
         Set_Parameters_According_To_Chetto
           (Sys.Dependencies,
            Tmp,
            Compute_Deadlines,
            Compute_Priorities);

         -- Display priorities/deadlines
         --
         Reset_Iterator (Tmp, Iterator1);
         loop

            Current_Element (Tmp, A_Task, Iterator1);
            Result := Result &
                      To_Unbounded_String ("- ") &
                      A_Task.Name &
                      To_Unbounded_String (" => ");
            if Compute_Deadlines then
               Result := Result &
                         To_Unbounded_String (A_Task.Deadline'Img) &
                         Unbounded_Lf;
            else
               Result := Result &
                         To_Unbounded_String (A_Task.Priority'Img) &
                         Unbounded_Lf;
            end if;

            exit when Is_Last_Element (Tmp, Iterator1);

            Next_Element (Tmp, Iterator1);
         end loop;

         Free (Tmp);

      end if;

      Result := Result & Unbounded_Lf;

   exception
      when Dependency_Services.Dependencies_Period_Error =>
         Result := Result &
                   Lb_Chetto_Error2 (Current_Language) &
                   Unbounded_Lf;
      when Dependency_Services.Cyclic_Graph_Error =>
         Result := Result &
                   Lb_Chetto_Error1 (Current_Language) &
                   Unbounded_Lf;
      when Dependency_Services.Task_Model_Error =>
         Result := Result &
                   Lb_Chetto_Error3 (Current_Language) &
                   Unbounded_Lf;

   end Chetto_Parameter_Modifications;

   procedure Compute_End_To_End_Response_Time
     (Sys              : in out System;
      Result           : in out Unbounded_String;
      Update_Tasks_Set : in Boolean;
      One_Step         : in Boolean;
      Output           : in Output_Format := String_Output)
   is

      Rt          : Response_Time_Table;
      Referencies : Unbounded_String := Empty_String;
      Message     : Unbounded_String := Empty_String;

   begin

      Initialize (Rt);
      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      begin

         Compute_End_To_End_Response_Time
           (Sys.Dependencies,
            Sys.Processors,
            Sys.Tasks,
            One_Step,
            Update_Tasks_Set,
            Referencies,
            Rt);

         if One_Step then
            Message := Message &
                       Start_Line &
                       Lb_Minus &
                       Lb_Jitter_From_Response_Time_One_Step (Current_Language)
                       &
                       Start_Ref;
         else
            Message := Message &
                       Start_Line &
                       Lb_Minus &
                       Lb_Jitter_From_Response_Time_All_Steps (
              Current_Language) &
                       Start_Ref;
         end if;

         Message := Message &
                    Referencies &
                    End_Ref &
                    End_Line &
                    Unbounded_Lf;

         if Update_Tasks_Set then
            Message := Message &
                       Start_Line &
                       Lb_Jitters_Are_Updated (Current_Language) &
                       End_Line &
                       Unbounded_Lf;
         end if;

         Message := Message &
                    Start_Line &
                    Lb_Task_Response_Time (Current_Language) &
                    Unbounded_Lf;

         for I in  0 .. Rt.Nb_Entries - 1 loop
            if (Rt.Entries (I).Item /= null) then
               Message :=
                 Message &
                 Start_Task &
                 To_Unbounded_String ("    ") &
                 Rt.Entries (I).Item.Name &
                 To_Unbounded_String (" => ") &
                 Integer'Image (Integer (Rt.Entries (I).Data)) &
                 End_Task &
                 Lb_Comma &
                 Start_Jit &
                 Lb_Jitter (Current_Language) &
                 To_Unbounded_String ("=") &
                 Periodic_Task_Ptr (Rt.Entries (I).Item).Jitter'Img &
                 To_Unbounded_String (" ") &
                 End_Jit;
               if (Natural (Rt.Entries (I).Data) >
                   Rt.Entries (I).Item.Deadline)
               then
                  Message := Message &
                             Lb_Comma &
                             Lb_Check_Deadline (Current_Language) &
                             Rt.Entries (I).Item.Deadline'Img;
                  Message := Message & To_Unbounded_String (")");
               end if;

               Message := Message & Unbounded_Lf;
            end if;
         end loop;
         Message := Message & End_Line & Unbounded_Lf;
         Message := Message & End_End_To_End & Unbounded_Lf;

      exception
         when Task_Set.Task_Must_Have_Period_Equal_To_Deadline =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_4 (Current_Language) &
                       Unbounded_Lf;
         when Task_Set.Task_Must_Be_Periodic =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_5 (Current_Language) &
                       Unbounded_Lf;
         when Task_Set.Task_Model_Error =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_6 (Current_Language) &
                       Unbounded_Lf;
         when Scheduler.Processor_Utilization_Exceeded =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_7 (Current_Language) &
                       Unbounded_Lf;
         when Task_Set.Start_Time_Error =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_12 (Current_Language) &
                       Unbounded_Lf;
         when Task_Set.Offset_Error =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_13 (Current_Language) &
                       Unbounded_Lf;

      end;

      Result := Result & Message;

   end Compute_End_To_End_Response_Time;

end Call_Dependency_Framework;
