------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_IO;                   use Text_IO;
with Events;                    use Events;
with Xml_Tag;                   use Xml_Tag;
with Unbounded_Strings;         use Unbounded_Strings;
with Task_Set;                  use Task_Set;
use Task_Set.Generic_Task_Set;
with Buffer_Set;                use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Ada.Exceptions;            use Ada.Exceptions;
with Processors;                use Processors;
with Processor_Set;             use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Call_Scheduling_Framework; use Call_Scheduling_Framework;
with Call_Memory_Framework;     use Call_Memory_Framework;
use Call_Memory_Framework.Buffer_Result_Package;
with Call_Resource_Framework;   use Call_Resource_Framework;
with Call_Dependency_Framework; use Call_Dependency_Framework;
with Call_Random_Framework;     use Call_Random_Framework;
with Scheduler;                 use Scheduler;
with Multiprocessor_Services;   use Multiprocessor_Services;
use Multiprocessor_Services.Scheduling_Result_Per_Processor;
with Translate;                 use Translate;
with Translate.English_Labels;
with Translate.Francais_Labels;

package body Call_Framework is

   type Sub_Framework_Request is record
      Statement : Framework_Statement_Type;
      Param     : Framework_Parameters_Table;
      Target    : Processor_Ptr;
   end record;

   procedure initialize is
   begin
      Translate.Initialize;
      if Current_Language = English then
         Translate.English_Labels.Initialize;
      else
         Translate.Francais_Labels.Initialize;
      end if;

   end initialize;

   procedure Sequential_Framework_Request
     (Sys      : in out System;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table)
   is
   begin
      Sequential_Framework_Request
        (Sys,
         Request,
         Response,
         Total_Order,
         String_Output);
   end Sequential_Framework_Request;

   procedure Put (F : in Framework_Request) is
   begin
      Put (F.Statement'Img & "/" & To_String (F.Target));
   end Put;

   procedure Initialize (F : in out Framework_Request) is
   begin
      F.Target := Empty_String;
      Initialize (F.Param);
   end Initialize;

   procedure Initialize (F : in out Framework_Response) is
   begin
      F.Title := Empty_String;
      F.Text  := Empty_String;
   end Initialize;

   procedure Put (F : in Framework_Response) is
   begin
      Put_Line (To_String (F.Title));
      Put_Line (To_String (F.Text));
   end Put;

   procedure Call_A_Request
     (Sys        : in out System;
      Request    : in Sub_Framework_Request;
      A_Response : in out Unbounded_String;
      Output     : in Output_Format := String_Output)
   is

      -- Parameters to be used for calling the "computing time line" service
      --
      Event_To_Generate          : Time_Unit_Event_Type_Boolean_Table;
      Period                     : Natural := 0;
      Schedule_With_Offsets      : Boolean := True;
      Schedule_With_Precedencies : Boolean := True;
      Schedule_With_Resources    : Boolean := True;
      With_Task_Specific_Seed    : Boolean := True;
      Global_Seed_Value          : Integer := 0;
      Predictable_Global_Seed    : Boolean := True;

   begin

      case Request.Statement is

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_Time_Line =>

         Initialize (Event_To_Generate, True);

         -- Check parameters
         --
         for I in  0 .. Request.Param.Nb_Entries - 1 loop
            if (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("period")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("seed_value")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("schedule_with_offsets")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("schedule_with_precedencies")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("schedule_with_resources")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("predictable")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("start_of_task_capacity")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("end_of_task_capacity")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("write_to_buffer")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("read_from_buffer")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("running_task")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("task_activation")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("send_message")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("receive_message")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("allocate_resource")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("release_resource")) and
               (Request.Param.Entries (I).Name /=
                To_Unbounded_String ("wait_for_resource"))
            then
               Raise_Exception
                 (Unexpected_Parameter'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;

            if (Request.Param.Entries (I).ptype /= Integer_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("period"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Integer_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("seed_value"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("schedule_with_offsets"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("schedule_with_precedencies"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("schedule_with_resources"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("predictable"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("start_of_task_capacity"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("end_of_task_capacity"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("write_to_buffer"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("read_from_buffer"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("running_task"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("task_activation"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("send_message"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("receive_message"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("allocate_resource"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("release_resource"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;
            if (Request.Param.Entries (I).ptype /= Boolean_Parameter) and
               (Request.Param.Entries (I).Name =
                To_Unbounded_String ("wait_for_resource"))
            then
               Raise_Exception
                 (Parameter_Type_Error'Identity,
                  To_String (Request.Param.Entries (I).Name));
            end if;

            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("period"))
            then
               Period := Request.Param.Entries (I).Integer_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("seed_value"))
            then
               Global_Seed_Value       :=
                 Request.Param.Entries (I).Integer_value;
               With_Task_Specific_Seed := False;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("schedule_with_offsets"))
            then
               Schedule_With_Offsets :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("schedule_with_precedencies"))
            then
               Schedule_With_Precedencies :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("schedule_with_resources"))
            then
               Schedule_With_Resources :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("predictable"))
            then
               Predictable_Global_Seed :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("start_of_task_capacity"))
            then
               Event_To_Generate (Start_Of_Task_Capacity) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("end_of_task_capacity"))
            then
               Event_To_Generate (End_Of_Task_Capacity) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("write_to_buffer"))
            then
               Event_To_Generate (Write_To_Buffer) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("read_from_buffer"))
            then
               Event_To_Generate (Read_From_Buffer) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("running_task"))
            then
               Event_To_Generate (Running_Task) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("task_activation"))
            then
               Event_To_Generate (Task_Activation) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("send_message"))
            then
               Event_To_Generate (Send_Message) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("receive_message"))
            then
               Event_To_Generate (Receive_Message) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("allocate_resource"))
            then
               Event_To_Generate (Receive_Message) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("release_resource"))
            then
               Event_To_Generate (Release_Resource) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;
            if (Request.Param.Entries (I).Name =
                To_Unbounded_String ("wait_for_resource"))
            then
               Event_To_Generate (Wait_For_Resource) :=
                 Request.Param.Entries (I).Boolean_value;
            end if;

         end loop;

         -- Call framework
         --
         Compute_Simulation_Time_Line
           (Sys,
            A_Response,

         -- With period
         --
            Period,

         -- With offsets
         --
            Schedule_With_Offsets,

         -- With Task precedencies
         --
            Schedule_With_Precedencies,

         -- With Shared resources
         --
            Schedule_With_Resources,

         -- With task specific seed
         --
            With_Task_Specific_Seed,

         -- With a global seed value
         --
            Global_Seed_Value,

         -- Predictable seed of not ?
         --
            Predictable_Global_Seed,

         -- Which event we should generate
         --
            Event_To_Generate,

            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_Basics =>

         Compute_Simulation_Basics (Sys, A_Response, Request.Target, Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_Run_Event_Handler =>

         Run_User_Defined_Event_Handlers (Sys, A_Response, Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_Priority_Inversion =>

         Priority_Inversion_Detection
           (Sys,
            A_Response,
            Request.Target,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_General_Task =>

         Partionning_Tasks (Sys, A_Response, General_Task, Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_First_Fit =>

         Partionning_Tasks (Sys, A_Response, First_Fit, Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_Next_Fit =>

         Partionning_Tasks (Sys, A_Response, Next_Fit, Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_Best_Fit =>

         Partionning_Tasks (Sys, A_Response, Best_Fit, Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_Small_Task =>

         Partionning_Tasks (Sys, A_Response, Small_Task, Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_Deadlock =>

         Deadlock_Detection (Sys, A_Response, Request.Target, Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_Context_Switch_Number =>

         Compute_Simulation_Number_Of_Context_Switch
           (Sys,
            A_Response,
            Request.Target,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_Preemption_Number =>

         Compute_Simulation_Number_Of_Preemption
           (Sys,
            A_Response,
            Request.Target,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_Response_Time =>

         if (Get_Number_Of_Task_From_Processor
                (Sys.Tasks,
                 Request.Target.Name) /=
             0)
         then

            Compute_Feasibility_Response_Time
              (Sys,
               A_Response,
               Request.Target,
               Output);

         end if;

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_Cpu_Utilization =>

         if (Get_Number_Of_Task_From_Processor
                (Sys.Tasks,
                 Request.Target.Name) /=
             0)
         then

            Compute_Feasibility_Processor_Utilization_Factor
              (Sys,
               A_Response,
               Request.Target,
               Output);

         end if;

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_Response_Time =>

         -- Check parameters
         --
         if (Request.Param.Nb_Entries /= 3) then
            Raise_Exception
              (Invalid_Number_Of_Parameter'Identity,
               "3 expected parameters");
         end if;

         if (Request.Param.Entries (0).Name /=
             To_Unbounded_String ("worst_case"))
         then
            Raise_Exception (Unexpected_Parameter'Identity, "worst_case");
         end if;

         if (Request.Param.Entries (1).Name /=
             To_Unbounded_String ("best_case"))
         then
            Raise_Exception (Unexpected_Parameter'Identity, "best_case");
         end if;

         if (Request.Param.Entries (2).Name /=
             To_Unbounded_String ("average_case"))
         then
            Raise_Exception (Unexpected_Parameter'Identity, "average_case");
         end if;

         if (Request.Param.Entries (0).ptype /= Boolean_Parameter) then
            Raise_Exception (Parameter_Type_Error'Identity, "worst_case");
         end if;
         if (Request.Param.Entries (1).ptype /= Boolean_Parameter) then
            Raise_Exception (Parameter_Type_Error'Identity, "best_case");
         end if;
         if (Request.Param.Entries (2).ptype /= Boolean_Parameter) then
            Raise_Exception (Parameter_Type_Error'Identity, "average_case");
         end if;

         Compute_Simulation_Response_Time
           (Sys,
            A_Response,
            Request.Target,
            Request.Param.Entries (0).Boolean_value,
            Request.Param.Entries (1).Boolean_value,
            Request.Param.Entries (2).Boolean_value,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_Blocking_Time =>

         -- Check parameters
         --
         if (Request.Param.Nb_Entries /= 3) then
            Raise_Exception
              (Invalid_Number_Of_Parameter'Identity,
               "3 expected parameters");
         end if;

         if (Request.Param.Entries (0).Name /=
             To_Unbounded_String ("worst_case"))
         then
            Raise_Exception (Unexpected_Parameter'Identity, "worst_case");
         end if;

         if (Request.Param.Entries (1).Name /=
             To_Unbounded_String ("best_case"))
         then
            Raise_Exception (Unexpected_Parameter'Identity, "best_case");
         end if;

         if (Request.Param.Entries (2).Name /=
             To_Unbounded_String ("average_case"))
         then
            Raise_Exception (Unexpected_Parameter'Identity, "average_case");
         end if;

         if (Request.Param.Entries (0).ptype /= Boolean_Parameter) then
            Raise_Exception (Parameter_Type_Error'Identity, "worst_case");
         end if;
         if (Request.Param.Entries (1).ptype /= Boolean_Parameter) then
            Raise_Exception (Parameter_Type_Error'Identity, "best_case");
         end if;
         if (Request.Param.Entries (2).ptype /= Boolean_Parameter) then
            Raise_Exception (Parameter_Type_Error'Identity, "average_case");
         end if;

         Compute_Simulation_Blocking_Time
           (Sys,
            A_Response,
            Request.Target,
            Request.Param.Entries (0).Boolean_value,
            Request.Param.Entries (1).Boolean_value,
            Request.Param.Entries (2).Boolean_value,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Simulation_All_Response_Time =>

         null;

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_Compute_And_Set_Blocking_Time =>

         Compute_Bound_On_Blocking_Time
           (Sys,
            A_Response,
            Request.Target,
            True,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_Compute_Blocking_Time =>

         Compute_Bound_On_Blocking_Time
           (Sys,
            A_Response,
            Request.Target,
            False,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Feasibility_Basics =>

         if (Get_Number_Of_Task_From_Processor
                (Sys.Tasks,
                 Request.Target.Name) /=
             0)
         then

            Compute_Feasibility_Basics
              (Sys,
               A_Response,
               Request.Target,
               Output);

         end if;

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Set_Priorities_According_To_Deadline_Monotonic =>

         Set_Priorities
           (Sys,
            A_Response,
            Request.Target,
            Deadline_Monotonic_Protocol,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Scheduling_Set_Priorities_According_To_Rate_Monotonic =>

         Set_Priorities
           (Sys,
            A_Response,
            Request.Target,
            Rate_Monotonic_Protocol,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Buffer_Feasibility_Tests =>

         if (Get_Number_Of_Buffer_From_Processor
                (Sys.Buffers,
                 Request.Target.Name) /=
             0)
         then

            Compute_Buffer_Feasibility_Tests
              (Sys,
               A_Response,
               Request.Target,
               Output);
         end if;

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Global_Memory_Compute_Requirement_Analysis =>
         Compute_Memory_Requirement_Analysis (Sys, A_Response, False, Output);

      when Global_Memory_Set_Requirement_Analysis =>
         Compute_Memory_Requirement_Analysis (Sys, A_Response, True, Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Buffer_Scheduling_Simulation =>
         if (Get_Number_Of_Buffer_From_Processor
                (Sys.Buffers,
                 Request.Target.Name) /=
             0)
         then
            Compute_Buffer_Scheduling_Simulation
              (Sys,
               A_Response,
               Request.Target,
               Output);
         end if;

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Random_Response_Time_Density =>

         Compute_Response_Time_Density
           (Sys,
            A_Response,
            Request.Target,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Dependency_Compute_Chetto_Blazewicz_Priority =>

         Chetto_Parameter_Modifications
           (Sys,
            A_Response,
            Request.Target,
            False,
            True,
            False,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Dependency_Compute_Chetto_Blazewicz_Deadline =>

         Chetto_Parameter_Modifications
           (Sys,
            A_Response,
            Request.Target,
            True,
            False,
            False,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Dependency_Set_Chetto_Blazewicz_Deadline =>

         Chetto_Parameter_Modifications
           (Sys,
            A_Response,
            Request.Target,
            True,
            False,
            True,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Dependency_Set_Chetto_Blazewicz_Priority =>

         Chetto_Parameter_Modifications
           (Sys,
            A_Response,
            Request.Target,
            False,
            True,
            True,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Dependency_Compute_End_To_End_Response_Time_One_Step =>

         Compute_End_To_End_Response_Time
           (Sys,
            A_Response,
            False,
            True,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Dependency_Set_End_To_End_Response_Time_One_Step =>

         Compute_End_To_End_Response_Time
           (Sys,
            A_Response,
            True,
            True,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Dependency_Compute_End_To_End_Response_Time_All_Steps =>

         Compute_End_To_End_Response_Time
           (Sys,
            A_Response,
            False,
            False,
            Output);

      -----------------------------------------------------------
      -----------------------------------------------------------
      when Dependency_Set_End_To_End_Response_Time_All_Steps =>

         Compute_End_To_End_Response_Time
           (Sys,
            A_Response,
            True,
            False,
            Output);

      end case;
   end Call_A_Request;

   procedure Initialize_Before_Request (Request : in Framework_Request) is
   begin
      case Request.Statement is
      when Scheduling_Simulation_Time_Line =>
         Initialize (Sched.all);
      when Scheduling_Simulation_Basics =>
         null;
      when Scheduling_Simulation_Preemption_Number =>
         null;
      when Scheduling_Simulation_Context_Switch_Number =>
         null;
      when Scheduling_Simulation_Response_Time =>
         null;
      when Scheduling_Simulation_All_Response_Time =>
         null;
      when Scheduling_Simulation_Blocking_Time =>
         null;
      when Scheduling_Simulation_Run_Event_Handler =>
         null;
      when Scheduling_Feasibility_First_Fit =>
         null;
      when Scheduling_Feasibility_Next_Fit =>
         null;
      when Scheduling_Feasibility_Best_Fit =>
         null;
      when Scheduling_Feasibility_Small_Task =>
         null;
      when Scheduling_Feasibility_General_Task =>
         null;
      when Scheduling_Feasibility_Response_Time =>
         null;
      when Scheduling_Feasibility_Cpu_Utilization =>
         null;
      when Scheduling_Feasibility_Basics =>
         null;
      when Scheduling_Set_Priorities_According_To_Deadline_Monotonic =>
         null;
      when Scheduling_Set_Priorities_According_To_Rate_Monotonic =>
         null;
      when Scheduling_Feasibility_Compute_And_Set_Blocking_Time =>
         null;
      when Scheduling_Feasibility_Compute_Blocking_Time =>
         null;
      when Scheduling_Simulation_Priority_Inversion =>
         null;
      when Scheduling_Simulation_Deadlock =>
         null;

      when Global_Memory_Set_Requirement_Analysis =>
         null;
      when Global_Memory_Compute_Requirement_Analysis =>
         null;
      when Buffer_Feasibility_Tests =>
         null;
      when Buffer_Scheduling_Simulation =>
         Initialize (Buff);

      when Random_Response_Time_Density =>
         null;

      when Dependency_Set_Chetto_Blazewicz_Deadline =>
         null;
      when Dependency_Compute_End_To_End_Response_Time_One_Step =>
         null;
      when Dependency_Set_End_To_End_Response_Time_One_Step =>
         null;
      when Dependency_Compute_End_To_End_Response_Time_All_Steps =>
         null;
      when Dependency_Set_End_To_End_Response_Time_All_Steps =>
         null;
      when Dependency_Compute_Chetto_Blazewicz_Priority =>
         null;
      when Dependency_Compute_Chetto_Blazewicz_Deadline =>
         null;
      when Dependency_Set_Chetto_Blazewicz_Priority =>
         null;
      end case;
   end Initialize_Before_Request;

   function Build_A_Title
     (Request : in Framework_Request)
      return    Unbounded_String
   is
   begin

      case Request.Statement is

      when Scheduling_Simulation_Time_Line =>
         return Empty_String;

      when Scheduling_Simulation_Basics =>
         return Start_Sched_Sim &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Simulation (Current_Language);

      when Scheduling_Simulation_Preemption_Number =>
         return Start_Sched_Sim &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Simulation (Current_Language);

      when Scheduling_Simulation_Context_Switch_Number =>
         return Start_Sched_Sim &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Simulation (Current_Language);

      when Scheduling_Simulation_All_Response_Time =>
         return Start_Sched_Sim &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Simulation (Current_Language);

      when Scheduling_Simulation_Priority_Inversion =>
         return Start_Sched_Sim &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Simulation (Current_Language);

      when Scheduling_Simulation_Deadlock =>
         return Start_Sched_Sim &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Simulation (Current_Language);

      when Scheduling_Simulation_Response_Time =>
         return Start_Sched_Sim &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Simulation (Current_Language);

      when Scheduling_Simulation_Blocking_Time =>
         return Start_Sched_Sim &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Simulation (Current_Language);

      when Scheduling_Simulation_Run_Event_Handler =>
         return Start_Sched_Sim &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Simulation (Current_Language) &
                Lb_Comma &
                Lb_Event_Analyzer (Current_Language);

      when Scheduling_Feasibility_First_Fit =>
         return Start_Block &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language) &
                Lb_Comma &
                Lb_Partition (Current_Language) &
                Lb_Partition_First_Fit (Current_Language);

      when Scheduling_Feasibility_General_Task =>
         return Start_Block &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language) &
                Lb_Comma &
                Lb_Partition (Current_Language) &
                Lb_Partition_General_Task (Current_Language);

      when Scheduling_Feasibility_Best_Fit =>
         return Start_Block &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language) &
                Lb_Comma &
                Lb_Partition (Current_Language) &
                Lb_Partition_Best_Fit (Current_Language);

      when Scheduling_Feasibility_Next_Fit =>
         return Start_Block &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language) &
                Lb_Comma &
                Lb_Partition (Current_Language) &
                Lb_Partition_Next_Fit (Current_Language);

      when Scheduling_Feasibility_Small_Task =>
         return Start_Block &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language) &
                Lb_Comma &
                Lb_Partition (Current_Language) &
                Lb_Partition_Small_Task (Current_Language);

      when Scheduling_Feasibility_Compute_And_Set_Blocking_Time =>
         return Start_Block &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language);

      when Scheduling_Feasibility_Compute_Blocking_Time =>
         return Start_Block &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language);

      when Scheduling_Feasibility_Response_Time =>
         return Start_Sched_Feas &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language);

      when Scheduling_Feasibility_Basics =>
         return Start_Sched_Feas &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language);

      when Scheduling_Feasibility_Cpu_Utilization =>
         return Start_Sched_Feas &
                Unbounded_Lf &
                Start_Title &
                Lb_Scheduling_Feasibility (Current_Language);

      when Scheduling_Set_Priorities_According_To_Deadline_Monotonic =>
         return Start_Set_Prio &
                Unbounded_Lf &
                Start_Title &
                Unbounded_Lf &
                Lb_Set_Priorities (Current_Language);

      when Scheduling_Set_Priorities_According_To_Rate_Monotonic =>
         return Start_Set_Prio &
                Unbounded_Lf &
                Start_Title &
                Unbounded_Lf &
                Lb_Set_Priorities (Current_Language);

      when Global_Memory_Set_Requirement_Analysis =>
         return Start_Buf_Bound &
                Unbounded_Lf &
                Start_Title &
                Lb_Memory_Requirement_Analysis (Current_Language);
      when Global_Memory_Compute_Requirement_Analysis =>
         return Start_Buf_Bound &
                Unbounded_Lf &
                Start_Title &
                Lb_Memory_Requirement_Analysis (Current_Language);
      when Buffer_Feasibility_Tests =>
         return Start_Buf_Bound &
                Unbounded_Lf &
                Start_Title &
                Lb_Compute_Buffer (Current_Language);
      when Buffer_Scheduling_Simulation =>
         return Unbounded_Lf & Lb_Draw_Buffer (Current_Language);

      when Random_Response_Time_Density =>
         return Start_Response_Time_Density &
                Start_Title &
                Unbounded_Lf &
                Lb_Response_Time_Density (Current_Language);

      when Dependency_Compute_End_To_End_Response_Time_One_Step =>
         return Unbounded_Lf &
                Lb_Jitter_From_Response_Time (Current_Language);
      when Dependency_Set_End_To_End_Response_Time_One_Step =>
         return Unbounded_Lf &
                Lb_Jitter_From_Response_Time (Current_Language);
      when Dependency_Compute_End_To_End_Response_Time_All_Steps =>
         return Start_End_To_End &
                Unbounded_Lf &
                Start_Title &
                Lb_Jitter_From_Response_Time (Current_Language);
      when Dependency_Set_End_To_End_Response_Time_All_Steps =>
         return Start_End_To_End &
                Unbounded_Lf &
                Start_Title &
                Lb_Jitter_From_Response_Time (Current_Language);

      when Dependency_Compute_Chetto_Blazewicz_Priority =>
         return Lb_Chetto (Current_Language);
      when Dependency_Compute_Chetto_Blazewicz_Deadline =>
         return Lb_Chetto (Current_Language);
      when Dependency_Set_Chetto_Blazewicz_Priority =>
         return Lb_Chetto (Current_Language);
      when Dependency_Set_Chetto_Blazewicz_Deadline =>
         return Lb_Chetto (Current_Language);

      end case;

   end Build_A_Title;

   procedure Sequential_Framework_Request
     (Sys      : in out System_Ptr;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table;
      Order    : in Perform_Order;
      Output   : in Output_Format)
   is
   begin
      Sequential_Framework_Request
        (Sys.all,
         Request,
         Response,
         Order,
         Output);
   end Sequential_Framework_Request;

   procedure Sequential_Framework_Request
     (Sys      : in out System;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table;
      Order    : in Perform_Order;
      Output   : in Output_Format)
   is

      A_Response  : Framework_Response;
      A_Processor : Processor_Ptr;
      My_Iterator : Processors_Iterator;
      A_Request   : Sub_Framework_Request;

   begin

      for I in  0 .. Request.Nb_Entries - 1 loop

         Initialize_Before_Request (Request.Entries (I));

         -- Here, we should take care of two cases :
         -- case 1 : the called service should be run for
         --          each processor (that's the most of services)
         -- case 2 : the called service is global to all processor and
         --          should be called one time only
         --
         --
         if (Request.Entries (I).Statement =
             Scheduling_Simulation_Run_Event_Handler) or
            (Request.Entries (I).Statement =
             Global_Memory_Set_Requirement_Analysis) or
            (Request.Entries (I).Statement =
             Global_Memory_Compute_Requirement_Analysis) or
            (Request.Entries (I).Statement = Scheduling_Simulation_Time_Line)
            or
            (Request.Entries (I).Statement =
             Scheduling_Feasibility_Small_Task) or
            (Request.Entries (I).Statement = Scheduling_Feasibility_Best_Fit)
            or
            (Request.Entries (I).Statement = Scheduling_Feasibility_Next_Fit)
            or
            (Request.Entries (I).Statement =
             Scheduling_Feasibility_First_Fit) or
            (Request.Entries (I).Statement =
             Scheduling_Feasibility_General_Task) or
            (Request.Entries (I).Statement =
             Dependency_Compute_End_To_End_Response_Time_One_Step) or
            (Request.Entries (I).Statement =
             Dependency_Set_End_To_End_Response_Time_One_Step) or
            (Request.Entries (I).Statement =
             Dependency_Compute_End_To_End_Response_Time_All_Steps) or
            (Request.Entries (I).Statement =
             Dependency_Set_End_To_End_Response_Time_All_Steps) or
            (Request.Entries (I).Statement =
             Dependency_Set_Chetto_Blazewicz_Deadline) or
            (Request.Entries (I).Statement =
             Dependency_Compute_Chetto_Blazewicz_Priority) or
            (Request.Entries (I).Statement =
             Dependency_Compute_Chetto_Blazewicz_Deadline) or
            (Request.Entries (I).Statement =
             Dependency_Set_Chetto_Blazewicz_Priority)
         then

            -- Only one call for all processors
            --
            Initialize (A_Response);
            A_Request.Param     := Request.Entries (I).Param;
            A_Request.Statement := Request.Entries (I).Statement;

            Call_A_Request (Sys, A_Request, A_Response.Text, Output);

            if A_Response.Text /= Empty_String then

               A_Response.Title := Build_A_Title (Request.Entries (I));
               if A_Response.Title /= Empty_String then
                  A_Response.Title := Unbounded_Lf &
                                      Unbounded_Lf &
                                      A_Response.Title &
                                      Lb_Colon &
                                      Unbounded_Lf &
                                      End_Title;
               end if;

               Add (Response, A_Response);
            end if;

         else

            -- A call for each processor
            --
            Reset_Iterator (Sys.Processors, My_Iterator);

            loop
               Current_Element (Sys.Processors, A_Processor, My_Iterator);

               -- Do a request
               --
               if (A_Processor.Name = Request.Entries (I).Target) or
                  (Request.Entries (I).Target = Empty_String)
               then

                  Initialize (A_Response);
                  A_Request.Target    := A_Processor;
                  A_Request.Param     := Request.Entries (I).Param;
                  A_Request.Statement := Request.Entries (I).Statement;

                  Call_A_Request (Sys, A_Request, A_Response.Text, Output);

                  if A_Response.Text /= Empty_String then

                     A_Response.Title := Build_A_Title (Request.Entries (I));
                     if A_Response.Title /= Empty_String then
                        A_Response.Title :=
                          Unbounded_Lf &
                          A_Response.Title &
                          Lb_Comma &
                          Lb_Processor (Current_Language) &
                          Start_Proc &
                          A_Processor.Name &
                          End_Proc &
                          Lb_Colon &
                          Unbounded_Lf &
                          End_Title;
                     end if;

                     Add (Response, A_Response);
                  end if;

               end if;

               exit when Is_Last_Element (Sys.Processors, My_Iterator);

               Next_Element (Sys.Processors, My_Iterator);
            end loop;
         end if;

      end loop;
   end Sequential_Framework_Request;

   procedure Parallel_Framework_Request
     (Sys      : in out System;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table;
      Order    : in Perform_Order := Total_Order;
      Output   : in Output_Format := String_Output)
   is
   begin
      null;
   end Parallel_Framework_Request;

   procedure Distributed_Framework_Request
     (Sys      : in out System;
      Request  : in Framework_Request_Table;
      Response : in out Framework_Response_Table;
      Order    : in Perform_Order := Total_Order;
      Output   : in Output_Format := String_Output)
   is
   begin
      null;
   end Distributed_Framework_Request;

end Call_Framework;
