------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Ada.Strings.Unbounded;
USE  Ada.Strings.Unbounded;
WITH Parameters;
USE Parameters;
WITH Parameters.extended;
USE Parameters.extended;
USE Parameters.extended.Framework_Parameters_Table_Package;
WITH Framework_Config;
USE Framework_Config;
WITH Systems;
USE Systems;
WITH Tables;
WITH Convert_Strings;
WITH Convert_Unbounded_Strings;



PACKAGE Call_Framework  IS


   TYPE Framework_Statement_Type IS
         (-- Compute a set of basic simulation services
          --
          Scheduling_Simulation_Basics,

          -- Compute the scheduling time line
          --
          Scheduling_Simulation_Time_Line,

          -- Compute from a time line, the number
          -- of preemption
          --
          Scheduling_Simulation_Preemption_Number,

          -- Compute from a time line, the number
          -- of context switch
          --
          Scheduling_Simulation_Context_Switch_Number,

          -- Compute maximum bound on task response
          -- time from a simulation
          --
          Scheduling_Simulation_Response_Time,

          -- Compute from a simulation, for each task,
          -- response times for all its activation
          --
          Scheduling_Simulation_All_Response_Time,

          -- Compute from a simulation, for each task,
          -- its blocking time on shared resources
          --
          Scheduling_Simulation_Blocking_Time,


          -- Find in a simulation, if priority
          -- inversions exist
          --
          Scheduling_Simulation_Priority_Inversion,


          -- Find in a simulation, if
          -- deadlocks exist
          --
          Scheduling_Simulation_Deadlock,


          -- Run user's defined event handlers.
          -- This service allows users to dfine their
          -- own scheduling data inspectors
          --
          Scheduling_Simulation_Run_Event_Handler,



          -- Compute a set of basic feasibility tests services
          --
          Scheduling_Feasibility_Basics,

          -- Compute bound on task response time
          --
          Scheduling_Feasibility_Response_Time,

          -- Compute feasibility test based on cpu utilization factor
          --
          Scheduling_Feasibility_Cpu_Utilization,

          -- Set task priority according to Deadline Monotonic
          -- scheduler
          --
          Scheduling_Set_Priorities_According_To_Deadline_Monotonic,

          -- Set task priority according to Rate Monotonic
          -- scheduler
          --
          Scheduling_Set_Priorities_According_To_Rate_Monotonic,

          -- Compute shared resources blocking time
          --
          Scheduling_Feasibility_Compute_Blocking_Time,

          -- Compute shared resources blocking time
          -- and update the task set wxith the resulting blocking time
          --
          Scheduling_Feasibility_Compute_And_Set_Blocking_Time,


          -- Apply several multiprocessor partition rules
          --
          Scheduling_Feasibility_First_Fit,
          Scheduling_Feasibility_General_Task,
          Scheduling_Feasibility_Next_Fit,
          Scheduling_Feasibility_Best_Fit,
          Scheduling_Feasibility_Small_Task,



          -- Compute the global memory requirement
          --
          Global_Memory_Set_Requirement_Analysis,
          Global_Memory_Compute_Requirement_Analysis,


          -- Compute bound on buffer size
          --
          Buffer_Feasibility_Tests,

          -- Compute buffer utilization factor from a scheduling
          -- simulation
          --
          Buffer_Scheduling_Simulation,


          -- Compute a random law mean response time and its
          --  variance from a given scheduling
          --
          Random_Response_Time_Density,





          -- Compute end to end response time
          -- response time between several tasks which can
          -- be hostedon different processors
          --
          Dependency_Compute_End_To_End_Response_Time_One_Step,
          Dependency_Set_End_To_End_Response_Time_One_Step,
          Dependency_Compute_End_To_End_Response_Time_All_Steps,
          Dependency_Set_End_To_End_Response_Time_All_Steps,


          -- Assign deadlines or fixed priority
          -- to a set of tasks depending the precedency relatioship
          --
          Dependency_Compute_Chetto_Blazewicz_Priority,
          Dependency_Compute_Chetto_Blazewicz_Deadline,
          Dependency_Set_Chetto_Blazewicz_Priority,
          Dependency_Set_Chetto_Blazewicz_Deadline);



   PRAGMA Convention( C,  Framework_Statement_Type);



   PROCEDURE To_Framework_Statement IS
   NEW Convert_Strings(Framework_Statement_Type,
      Scheduling_Simulation_Basics);
   PROCEDURE To_Framework_Statement IS
   NEW Convert_Unbounded_Strings(Framework_Statement_Type,
      Scheduling_Simulation_Basics);





   -- Store a computing request.
   -- If "target" contains something, the computing should only
   -- done on one processor which is referecendd in "target"
   -- if "target" is a empty string, we compute for all processors
   --
   TYPE Framework_Request IS
      RECORD
         Statement : Framework_Statement_Type;
         Param     : Framework_Parameters_Table;
         Target    : Unbounded_String;
      END RECORD;

   TYPE Framework_Request_Ptr IS ACCESS Framework_Request;

   Invalid_Number_Of_Parameter : EXCEPTION;
   Unexpected_Parameter        : EXCEPTION;
   Parameter_Type_Error        : EXCEPTION;


   PROCEDURE Put (
         F : IN     Framework_Request);
   PROCEDURE Initialize (
         F : IN OUT Framework_Request);



   PACKAGE Framework_Request_Package IS NEW Tables(Framework_Request,
      Max_Requests,Put);
   USE Framework_Request_Package;

   SUBTYPE Framework_Request_Range IS Framework_Request_Package.
      Table_Range;
   SUBTYPE Framework_Request_Table IS Framework_Request_Package.
      Table;

   TYPE Framework_Request_Table_Ptr IS ACCESS Framework_Request_Table;


   TYPE Framework_Response IS
      RECORD
         Title : Unbounded_String;
         Text  : Unbounded_String;
      END RECORD;

   TYPE Framework_Response_Ptr IS ACCESS Framework_Response;

   PROCEDURE Put (
         F : IN     Framework_Response);
   PROCEDURE Initialize (
         F : IN OUT Framework_Response);


   PACKAGE Framework_Response_Package IS NEW Tables(Framework_Response,
      Max_Requests,Put);
   USE Framework_Response_Package;

   SUBTYPE Framework_Response_Range IS Framework_Response_Package.
      Table_Range;
   SUBTYPE Framework_Response_Table IS Framework_Response_Package.
      Table;

   TYPE Framework_Response_Table_Ptr IS ACCESS Framework_Response_Table;


   TYPE Output_Format IS
         (Xml_Output,
          String_Output,
          Raw_Output);
   PRAGMA Convention( C, Output_Format);


   TYPE Perform_Order IS
         (Total_Order,
          Causal_Order);
   PRAGMA Convention( C, Perform_Order);


-- Initialize the framework. Should be called before any
-- use of the framework
--
procedure initialize;


   PROCEDURE Sequential_Framework_Request (
         Sys      : IN OUT System;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table;
         Order    : IN     Perform_Order;
         Output   : IN     Output_Format);


   PROCEDURE Sequential_Framework_Request (
         Sys      : IN OUT System_Ptr;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table;
         Order    : IN     Perform_Order;
         Output   : IN     Output_Format);


   PROCEDURE Sequential_Framework_Request (
         Sys      : IN OUT System;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table);


   PROCEDURE Parallel_Framework_Request (
         Sys      : IN OUT System;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table;
         Order    : IN     Perform_Order            := Total_Order;
         Output   : IN     Output_Format            := String_Output);


   PROCEDURE Distributed_Framework_Request (
         Sys      : IN OUT System;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table;
         Order    : IN     Perform_Order            := Total_Order;
         Output   : IN     Output_Format            := String_Output);



END Call_Framework;


