------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Translate;                       use Translate;
with Xml_Tag;                         use Xml_Tag;
with Buffers.Extended;                use Buffers.Extended;
with Buffer_Set;                      use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Unbounded_Strings;               use Unbounded_Strings;
with Call_Scheduling_Framework;       use Call_Scheduling_Framework;
with Multiprocessor_Services;         use Multiprocessor_Services;
use Multiprocessor_Services.Scheduling_Result_Per_Processor;
with Tasks;                           use Tasks;
with Queueing_System;                 use Queueing_System;
with Queueing_System.theoretical;     use Queueing_System.theoretical;
with Queueing_System.theoretical.Mp1; use Queueing_System.theoretical.Mp1;
with Queueing_System.theoretical.Pp1; use Queueing_System.theoretical.Pp1;
with Queueing_System.theoretical.Mm1; use Queueing_System.theoretical.Mm1;
with Queueing_System.theoretical.Md1; use Queueing_System.theoretical.Md1;
with Queueing_System.theoretical.Mg1; use Queueing_System.theoretical.Mg1;
with Double_Util;                     use Double_Util;
with Ada.Numerics.Aux;                use Ada.Numerics.Aux;

package body Call_Memory_Framework is

   procedure Compute_Buffer_Scheduling_Simulation
     (Sys         : in System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      A_Buffer        : Buffer_Ptr;
      My_Iterator     : Buffers_Iterator;
      Index           : Scheduling_Table_Range := 0;
      A_Buffer_Result : Buffer_Size_Table_Ptr;
      A_Task          : Generic_Task_Ptr;

      Maximum_Size         : Natural;
      Maximum_Waiting_Time : Natural;
      Average_Size         : Double;
      Average_Waiting_Time : Double;

      Average_Consumer_Period : Double := 0.0;
      Sum_Period              : Double := 0.0;
      Nb_Consumer             : Double := 0.0;

   begin

      -- find the processor on which we want to compute buffer utilization
      --
      for I in  0 .. Call_Scheduling_Framework.Sched.Nb_Entries - 1 loop

         if Sched.Entries (I).Item.Name = A_Processor.Name then
            if Sched.Entries (I).Data.Error_Msg = Empty_String then
               Index := I;
               exit;
            end if;
         end if;
      end loop;

      -- Compute buffer utilization on all buffer of the
      -- selected processor
      --

      A_Buffer_Result := new Buffer_Size_Table;

      loop
         Current_Element (Sys.Buffers, A_Buffer, My_Iterator);

         if A_Processor.Name = A_Buffer.Cpu_Name then

            Initialize (A_Buffer_Result.all);

            Compute_Buffer_Size_From_Simulation
              (Sched.Entries (Index).Data.Result,
               A_Buffer,
               A_Buffer_Result);

            Add (Buff, A_Buffer, A_Buffer_Result.all);

            -- Compute average consumer period
            --
            Average_Consumer_Period := 0.0;
            Sum_Period              := 0.0;
            Nb_Consumer             := 0.0;
            for J in  0 .. A_Buffer.Roles.Nb_Entries - 1 loop
               if (A_Buffer.Roles.Entries (J).Data.The_Role = Consumer) then

                  A_Task :=
                     Search_Task (Sys.Tasks, A_Buffer.Roles.Entries (J).Item);
                  if (A_Task.Task_Type = Periodic_Type) then
                     Sum_Period  :=
                       +Double (Periodic_Task_Ptr (A_Task).Period);
                     Nb_Consumer := Nb_Consumer + 1.0;
                  end if;

               end if;
            end loop;

            Average_Consumer_Period := Sum_Period / Nb_Consumer;

            -- Compute results
            --
            Maximum_Size         :=
               Compute_Maximum_Buffer_Size_From_Simulation (A_Buffer_Result);
            Average_Size         :=
               Compute_Average_Buffer_Size_From_Simulation (A_Buffer_Result);
            Maximum_Waiting_Time :=
               Compute_Maximum_Waiting_Time_From_Simulation
                 (A_Buffer_Result,
                  Average_Consumer_Period);
            Average_Waiting_Time :=
               Compute_Average_Waiting_Time_From_Simulation
                 (A_Buffer_Result,
                  Average_Consumer_Period);

            -- Display results
            --
            Result :=
              Result &
              Unbounded_Lf &
              Unbounded_Lf &
              Start_Task &
              Lb_Buffer (Current_Language) &
              " " &
              A_Buffer.Name &
              To_Unbounded_String (" => ") &
              Unbounded_Lf;

            Result :=
              Result &
              To_Unbounded_String ("    ") &
              Lb_Maximum_Buffer_Size (Current_Language) &
              Maximum_Size'Img &
              Unbounded_Lf &
              To_Unbounded_String ("    ") &
              Lb_Maximum_Waiting_Time (Current_Language) &
              Maximum_Waiting_Time'Img &
              Unbounded_Lf;

            Result :=
              Result &
              To_Unbounded_String ("    ") &
              Lb_Average_Buffer_Size (Current_Language) &
              Format (Average_Size, 8) &
              Unbounded_Lf &
              To_Unbounded_String ("    ") &
              Lb_Average_Waiting_Time (Current_Language) &
              Format (Average_Waiting_Time, 8) &
              Unbounded_Lf;

         end if;

         exit when Is_Last_Element (Sys.Buffers, My_Iterator);
         Next_Element (Sys.Buffers, My_Iterator);

      end loop;

      Free (A_Buffer_Result);

   end Compute_Buffer_Scheduling_Simulation;

   procedure Periodic_Buffer_Control
     (My_Tasks : in Tasks_Set;
      A_Buffer : in Buffer_Ptr)
   is

      Iterator1 : Tasks_Iterator;
      Task1     : Generic_Task_Ptr;

   begin

      for I in  0 .. A_Buffer.Roles.Nb_Entries - 1 loop

         Reset_Iterator (My_Tasks, Iterator1);

         loop
            Current_Element (My_Tasks, Task1, Iterator1);

            if (Task1.Name = A_Buffer.Roles.Entries (I).Item) then
               if (Task1.Task_Type /= Periodic_Type) then
                  raise Task_Must_Be_Periodic;
               end if;
            end if;

            exit when Is_Last_Element (My_Tasks, Iterator1);
            Next_Element (My_Tasks, Iterator1);

         end loop;

      end loop;

   end Periodic_Buffer_Control;

   procedure Compute_Buffer_Feasibility_Tests
     (Sys         : in System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      A_Buffer : Buffer_Ptr;

      Waiting_Time : Double := 0.0;
      Occupation   : Double := 0.0;

      Reference, Ref1, Ref2 : Unbounded_String;
      My_Iterator           : Iterator;

      Nb_Arrival : Double := 0.0;
      Nb_Server  : Double := 0.0;

      My_Consumer_Task : Generic_Task_Ptr;

      Service_Rate : Double := 0.0;
      Arrival_Rate : Double := 0.0;

      A_Mm1_Queueing_System : Mm1_Queueing_System_theoretical;
      A_Md1_Queueing_System : Md1_Queueing_System_theoretical;
      A_Mg1_Queueing_System : Mg1_Queueing_System_theoretical;
      A_Mp1_Queueing_System : Mp1_Queueing_System_theoretical;
      A_Pp1_Queueing_System : Pp1_Queueing_System_theoretical;

      No_Performance_Criteria : Boolean := False;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      Result := Unbounded_Lf & Start_Line & Unbounded_Lf;

      Reset_Iterator (Sys.Buffers, My_Iterator);
      loop
         Current_Element (Sys.Buffers, A_Buffer, My_Iterator);
         if (A_Buffer.Cpu_Name = A_Processor.Name) then
            if (Natural (A_Buffer.Roles.Nb_Entries) >= 1) then
               begin

                  Result :=
                    Result &
                    Unbounded_Lf &
                    Unbounded_Lf &
                    Start_Task &
                    Lb_Buffer (Current_Language) &
                    " " &
                    A_Buffer.Name &
                    To_Unbounded_String (" => (") &
                    Qs_To_Display_String (A_Buffer.Queueing_System_Type) &
                    To_Unbounded_String (")") &
                    Unbounded_Lf;

                  Ref1                    := Empty_String;
                  Ref2                    := Empty_String;
                  Reference               := Empty_String;
                  Nb_Arrival              := 0.0;
                  Nb_Server               := 0.0;
                  No_Performance_Criteria := False;

                  -- Number of producer (arrival) and consumer (server)
                  --
                  for J in  0 .. A_Buffer.Roles.Nb_Entries - 1 loop
                     if (A_Buffer.Roles.Entries (J).Data.The_Role =
                         Producer)
                     then
                        Nb_Arrival := Nb_Arrival + 1.0;
                     end if;
                     if (A_Buffer.Roles.Entries (J).Data.The_Role =
                         Consumer)
                     then
                        My_Consumer_Task :=
                           Search_Task
                             (Sys.Tasks,
                              A_Buffer.Roles.Entries (J).Item);

                        Nb_Server := Nb_Server + 1.0;
                     end if;
                  end loop;

                  -- Control before buffer computation
                  --
                  Periodic_Buffer_Control (Sys.Tasks, A_Buffer);
                  Buffer_Flow_Control
                    (A_Buffer,
                     Sys.Tasks,
                     Service_Rate,
                     Arrival_Rate);

                  case A_Buffer.Queueing_System_Type is

                     when Qs_Mm1 =>

                        -- initalization
                        --
                        Reset (A_Mm1_Queueing_System);
                        Set_Qs_Arrival_Rate
                          (A_Mm1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate
                          (A_Mm1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival (A_Mm1_Queueing_System, Nb_Arrival);
                        Set_Qs_Nb_Server (A_Mm1_Queueing_System, Nb_Server);
                        Set_Qs_Harmonic
                          (A_Mm1_Queueing_System,
                           Is_Cons_Prod_Harmonic (A_Buffer, Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time
                          (A_Mm1_Queueing_System,
                           Double (My_Consumer_Task.Deadline));

                        if (Nb_Server <= 1.0) then

                           -- occupation computation
                           --
                           Queueing_System.theoretical.Mm1.
                             Qs_Average_Number_Customer
  (A_Mm1_Queueing_System,
   Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.theoretical.Mm1.
                             Qs_Average_Waiting_Time
  (A_Mm1_Queueing_System,
   Waiting_Time);

                           -- Update message to be displayed
                           --
                           Ref1 := To_Unbounded_String (" (") &
                                   Lb_See (Current_Language) &
                                   To_Unbounded_String ("[12], page 96, ") &
                                   Lb_Theorem (Current_Language) &
                                   To_Unbounded_String ("3.24). ");

                           Ref2 := To_Unbounded_String (" (") &
                                   Lb_See (Current_Language) &
                                   To_Unbounded_String ("[12], page 98, ") &
                                   Lb_Theorem (Current_Language) &
                                   To_Unbounded_String ("3.26). ");

                           Reference :=
                             To_Unbounded_String ("    ") &
                             Lb_Average_Buffer_Size (Current_Language) &
                             Format (Occupation, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref1 &
                             End_Ref &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Lb_Average_Waiting_Time (Current_Language) &
                             Format (Waiting_Time, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref2 &
                             End_Ref &
                             End_Task &
                             Unbounded_Lf &
                             End_Line &
                             End_Buf_Bound &
                             Unbounded_Lf;

                        else
                           No_Performance_Criteria := True;
                        end if;

                     when Qs_Md1 =>

                        -- initalization
                        --
                        Reset (A_Md1_Queueing_System);
                        Set_Qs_Arrival_Rate
                          (A_Md1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate
                          (A_Md1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival (A_Md1_Queueing_System, Nb_Arrival);
                        Set_Qs_Nb_Server (A_Md1_Queueing_System, Nb_Server);
                        Set_Qs_Harmonic
                          (A_Md1_Queueing_System,
                           Is_Cons_Prod_Harmonic (A_Buffer, Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time
                          (A_Md1_Queueing_System,
                           Double (My_Consumer_Task.Deadline));

                        if (Nb_Server <= 1.0) then

                           -- occupation computation
                           --
                           Queueing_System.theoretical.Md1.
                             Qs_Average_Number_Customer
  (A_Md1_Queueing_System,
   Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.theoretical.Md1.
                             Qs_Average_Waiting_Time
  (A_Md1_Queueing_System,
   Waiting_Time);

                           -- Update message to be displayed
                           --
                           Ref1 := To_Unbounded_String ("to be filled");

                           Ref2 := To_Unbounded_String ("to be filled");

                           Reference :=
                             To_Unbounded_String ("    ") &
                             Lb_Average_Buffer_Size (Current_Language) &
                             Format (Occupation, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref1 &
                             End_Ref &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Lb_Average_Waiting_Time (Current_Language) &
                             Format (Waiting_Time, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref2 &
                             End_Ref &
                             End_Task &
                             Unbounded_Lf &
                             End_Line &
                             End_Buf_Bound &
                             Unbounded_Lf;

                        else
                           No_Performance_Criteria := True;
                        end if;

                     when Qs_Mg1 =>

                        -- initalization
                        --
                        Reset (A_Mg1_Queueing_System);
                        Set_Qs_Arrival_Rate
                          (A_Mg1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate
                          (A_Mg1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival (A_Mg1_Queueing_System, Nb_Arrival);
                        Set_Qs_Nb_Server (A_Mg1_Queueing_System, Nb_Server);
                        Set_Qs_Harmonic
                          (A_Mg1_Queueing_System,
                           Is_Cons_Prod_Harmonic (A_Buffer, Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time
                          (A_Mg1_Queueing_System,
                           Double (My_Consumer_Task.Deadline));

                        if (Nb_Server <= 1.0) then

                           -- occupation computation
                           --
                           Queueing_System.theoretical.Mg1.
                             Qs_Average_Number_Customer
  (A_Mg1_Queueing_System,
   Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.theoretical.Mg1.
                             Qs_Average_Waiting_Time
  (A_Mg1_Queueing_System,
   Waiting_Time);

                           -- Update message to be displayed
                           --
                           Ref1 := To_Unbounded_String ("to be filled");

                           Ref2 := To_Unbounded_String ("to be filled");

                           Reference :=
                             To_Unbounded_String ("    ") &
                             Lb_Average_Buffer_Size (Current_Language) &
                             Format (Occupation, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref1 &
                             End_Ref &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Lb_Average_Waiting_Time (Current_Language) &
                             Format (Waiting_Time, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref2 &
                             End_Ref &
                             End_Task &
                             Unbounded_Lf &
                             End_Line &
                             End_Buf_Bound &
                             Unbounded_Lf;

                        else
                           No_Performance_Criteria := True;
                        end if;

                     when Qs_Mp1 =>

                        -- initalization
                        --
                        Reset (A_Mp1_Queueing_System);
                        Set_Qs_Arrival_Rate
                          (A_Mp1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate
                          (A_Mp1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival (A_Mp1_Queueing_System, Nb_Arrival);
                        Set_Qs_Nb_Server (A_Mp1_Queueing_System, Nb_Server);
                        Set_Qs_Harmonic
                          (A_Mp1_Queueing_System,
                           Is_Cons_Prod_Harmonic (A_Buffer, Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time
                          (A_Mp1_Queueing_System,
                           Double (My_Consumer_Task.Deadline));

                        if (Nb_Server <= 1.0) then

                           -- occupation computation
                           --
                           Queueing_System.theoretical.Mp1.
                             Qs_Average_Number_Customer
  (A_Mp1_Queueing_System,
   Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.theoretical.Mp1.
                             Qs_Average_Waiting_Time
  (A_Mp1_Queueing_System,
   Waiting_Time);

                           -- Update message to be displayed
                           --
                           Ref1 := To_Unbounded_String (" (") &
                                   Lb_See (Current_Language) &
                                   To_Unbounded_String ("[10], ") &
                                   Lb_Theorem (Current_Language) &
                                   To_Unbounded_String ("6). ");

                           Ref2 := To_Unbounded_String (" (") &
                                   Lb_See (Current_Language) &
                                   To_Unbounded_String ("[10], ") &
                                   Lb_Theorem (Current_Language) &
                                   To_Unbounded_String ("6). ");

                           Reference :=
                             To_Unbounded_String ("    ") &
                             Lb_Average_Buffer_Size (Current_Language) &
                             Format (Occupation, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref1 &
                             End_Ref &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Lb_Average_Waiting_Time (Current_Language) &
                             Format (Waiting_Time, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref2 &
                             End_Ref &
                             End_Task &
                             Unbounded_Lf &
                             End_Line &
                             End_Buf_Bound &
                             Unbounded_Lf;

                        else
                           No_Performance_Criteria := True;
                        end if;

                     when Qs_Pp1 =>

                        -- initalization
                        --
                        Reset (A_Pp1_Queueing_System);
                        Set_Qs_Arrival_Rate
                          (A_Pp1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate
                          (A_Pp1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival (A_Pp1_Queueing_System, Nb_Arrival);
                        Set_Qs_Nb_Server (A_Pp1_Queueing_System, Nb_Server);
                        Set_Qs_Harmonic
                          (A_Pp1_Queueing_System,
                           Is_Cons_Prod_Harmonic (A_Buffer, Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time
                          (A_Mp1_Queueing_System,
                           Double (My_Consumer_Task.Deadline));

                        if (Nb_Server <= 1.0) then

                           -- occupation computation
                           --
                           Queueing_System.theoretical.Pp1.
                             Qs_Maximum_Number_Customer
  (A_Pp1_Queueing_System,
   Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.theoretical.Pp1.
                             Qs_Maximum_Waiting_Time
  (A_Pp1_Queueing_System,
   Waiting_Time);

                           -- Update message to be displayed
                           --
                           Ref1 := To_Unbounded_String (" (") &
                                   Lb_See (Current_Language) &
                                   To_Unbounded_String ("[4,10], ") &
                                   Lb_Theorem (Current_Language) &
                                   To_Unbounded_String ("1 or 8). ");

                           Ref2 := To_Unbounded_String (" (") &
                                   Lb_See (Current_Language) &
                                   To_Unbounded_String ("[10], ") &
                                   Lb_Theorem (Current_Language) &
                                   To_Unbounded_String ("8). ");

                           Reference :=
                             To_Unbounded_String ("    ") &
                             Lb_Maximum_Buffer_Size (Current_Language) &
                             Format (Occupation, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref1 &
                             End_Ref &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Lb_Maximum_Waiting_Time (Current_Language) &
                             Format (Waiting_Time, 8) &
                             Unbounded_Lf &
                             To_Unbounded_String ("    ") &
                             Start_Ref &
                             Ref2 &
                             End_Ref &
                             End_Task &
                             Unbounded_Lf &
                             End_Line &
                             End_Buf_Bound &
                             Unbounded_Lf;

                        else
                           No_Performance_Criteria := True;
                        end if;

                     when others =>
                        No_Performance_Criteria := True;

                  end case;

                  if No_Performance_Criteria then
                     Result := Result &
                               To_Unbounded_String ("    ") &
                               Lb_Compute_Buffer_Error_4 (Current_Language) &
                               Unbounded_Lf;
                  else
                     Result := Result & Reference;
                  end if;

               exception
                  when Not_Implemented =>
                     Result := Result &
                               To_Unbounded_String ("    ") &
                               Lb_Compute_Buffer_Error_4 (Current_Language) &
                               Unbounded_Lf;
                  when Task_Set.Task_Model_Error =>
                     Result := Result &
                               To_Unbounded_String ("    ") &
                               Lb_Compute_Buffer_Error_3 (Current_Language) &
                               Unbounded_Lf;
                  when Task_Must_Be_Periodic =>
                     Result := Result &
                               To_Unbounded_String ("    ") &
                               Lb_Compute_Buffer_Error_2 (Current_Language) &
                               Unbounded_Lf;
                  when Flow_Constraint_Not_Respected =>
                     Result := Result &
                               To_Unbounded_String ("    ") &
                               Lb_Compute_Buffer_Error_1 (Current_Language) &
                               Unbounded_Lf;
               end;
            else
               Result := Result &
                         To_Unbounded_String ("    ") &
                         A_Buffer.Name &
                         To_Unbounded_String (" => NA");
            end if;
         end if;
         exit when Is_Last_Element (Sys.Buffers, My_Iterator);
         Next_Element (Sys.Buffers, My_Iterator);
      end loop;

   end Compute_Buffer_Feasibility_Tests;

   procedure Compute_Memory_Requirement_Analysis
     (Sys                       : in System;
      Result                    : in out Unbounded_String;
      Update_Address_Spaces_Set : in Boolean;
      Output                    : in Output_Format := String_Output)
   is
   begin
      Result := To_Unbounded_String ("to be written");
   end Compute_Memory_Requirement_Analysis;

end Call_Memory_Framework;
