------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Xml_Tag;                   use Xml_Tag;
with Tasks;                     use Tasks;
with Task_Set;                  use Task_Set;
use Task_Set.Generic_Task_Set;
with Resource_Set;              use Resource_Set;
use Resource_Set.Generic_Resource_Set;
with Processors.extended;       use Processors.extended;
with Processor_Set;             use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Translate;                 use Translate;
with Unbounded_Strings;         use Unbounded_Strings;
with Framework_Config;          use Framework_Config;
with Scheduler;                 use Scheduler;
with Scheduler.Fixed_Priority;  use Scheduler.Fixed_Priority;
with Scheduling_Sequences;      use Scheduling_Sequences;
use Scheduling_Sequences.Deadlock_List_Package;
use Scheduling_Sequences.Priority_Inversion_List_Package;
with Multiprocessor_Services;   use Multiprocessor_Services;
use Multiprocessor_Services.Scheduling_Result_Per_Processor;
with Call_Scheduling_Framework; use Call_Scheduling_Framework;
with Double_Util;               use Double_Util;
with Ada.Numerics.Aux;          use Ada.Numerics.Aux;

package body Call_Resource_Framework is

   procedure Compute_Bound_On_Blocking_Time
     (Sys                  : in out System;
      Result               : in out Unbounded_String;
      A_Processor          : in Processor_Ptr;
      Inject_Blocking_Time : in Boolean;
      Output               : in Output_Format := String_Output)
   is

      A_Task               : Generic_Task_Ptr;
      Has_Raised_Exception : Boolean := False;
      Msg                  : Unbounded_String;
      Blocking_Time        : Scheduler.Blocking_Time_Table;
      A_Scheduler          : Scheduler_Ptr;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      -- And now, compute blocking time
      --
      Result      :=
        Result &
        Start_Line &
        Lb_Minus &
        Lb_Bound_Blocking_Time (Current_Language) &
        Start_Ref &
        To_Unbounded_String (" (") &
        Lb_See (Current_Language) &
        To_Unbounded_String ("[3]) ") &
        End_Ref &
        Lb_Colon &
        Unbounded_Lf;
      Msg         := Empty_String;
      A_Scheduler := Get_Scheduler (extended_Processor (A_Processor.all));
      Compute_Blocking_Time
        (A_Scheduler.all,
         Sys.Tasks,
         Sys.Resources,
         A_Processor.Name,
         Msg,
         Blocking_Time);
      if Msg /= Empty_String then
         Has_Raised_Exception := True;
         Result               := Result & Msg;
      else
         Result := Result & Unbounded_Lf;
      end if;

      -- Display all task blocking time
      --

      for I in  0 .. Scheduler.Blocking_Time_Range (
                                             Get_Number_Of_Task_From_Processor
                                               (Sys.Tasks,
                                                A_Processor.Name) -
                                                    1)
      loop
         if (Blocking_Time.Entries (I).Item /= null) then
            Result :=
              Result &
              Start_Task &
              To_Unbounded_String ("    ") &
              Blocking_Time.Entries (I).Item.Name &
              To_Unbounded_String (" =>") &
              Integer'Image (Integer (Blocking_Time.Entries (I).Data)) &
              End_Task &
              Unbounded_Lf;
         end if;
      end loop;

      Result := Result &
                Unbounded_Lf &
                End_Line &
                Unbounded_Lf &
                Unbounded_Lf;

      -- Update blocking time
      --
      if Inject_Blocking_Time then

         if Has_Raised_Exception then

            Result :=
              Result &
              Unbounded_Lf &
              Start_Line &
              Lb_Blocking_Time_Inject_Failed (Current_Language) &
              End_Line &
              Unbounded_Lf &
              End_Block &
              Unbounded_Lf;

         else
            for I in  0 .. Scheduler.Blocking_Time_Range (
                                             Get_Number_Of_Task_From_Processor
                                               (Sys.Tasks,
                                                A_Processor.Name) -
                                                          1)
            loop
               if (Blocking_Time.Entries (I).Item /= null) then

                  A_Task               :=
                     Search_Task
                       (Sys.Tasks,
                        Blocking_Time.Entries (I).Item.Name);
                  A_Task.Blocking_Time := Blocking_Time.Entries (I).Data;
               end if;
            end loop;

            Result :=
              Result &
              Unbounded_Lf &
              Start_Line &
              Lb_Blocking_Time_Inject_Success (Current_Language) &
              End_Line &
              Unbounded_Lf &
              End_Block &
              Unbounded_Lf;
         end if;
      end if;

   exception
      when Invalid_Scheduler =>
         Result := Result &
                   Unbounded_Lf &
                   Lb_Compute_Blocking_Error2 (Current_Language) &
                   Unbounded_Lf;

   end Compute_Bound_On_Blocking_Time;

   procedure Compute_Simulation_Blocking_Time
     (Sys          : in out System;
      Result       : in out Unbounded_String;
      A_Processor  : in Processor_Ptr;
      Worst_Case   : in Boolean;
      Best_Case    : in Boolean;
      Average_Case : in Boolean;
      Output       : in Output_Format := String_Output)
   is

      A_Task      : Generic_Task_Ptr;
      My_Iterator : Tasks_Iterator;
      Min         : Natural := 0;
      Max         : Natural := 0;
      Average     : Double  := 0.0;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      for J in  0 .. Sched.Nb_Entries - 1 loop
         if Sched.Entries (J).Item.Name = A_Processor.Name then

            Result := Result &
                      Start_Line &
                      Lb_Minus &
                      Lb_Simulation_Blocking_Time (Current_Language) &
                      Lb_Colon &
                      Unbounded_Lf;

            Reset_Iterator (Sys.Tasks, My_Iterator);

            loop
               Current_Element (Sys.Tasks, A_Task, My_Iterator);

               if (A_Processor.Name = A_Task.Cpu_Name) then

                  Blocking_Time_From_Simulation
                    (A_Task,
                     Sys.Resources,
                     Sched.Entries (J).Data.Result,
                     Max,
                     Min,
                     Average);

                  Result := Result &
                            Start_Task &
                            To_Unbounded_String ("    ") &
                            A_Task.Name &
                            To_Unbounded_String (" =>");
                  if Worst_Case then
                     Result := Result &
                               Max'Img &
                               To_Unbounded_String ("/worst ");
                  end if;
                  if Best_Case then
                     Result := Result &
                               Min'Img &
                               To_Unbounded_String ("/best ");
                  end if;
                  if Average_Case then
                     Result := Result &
                               Format (Average) &
                               To_Unbounded_String ("/average ");
                  end if;
                  Result := Result & End_Task & Unbounded_Lf;

               end if;

               exit when Is_Last_Element (Sys.Tasks, My_Iterator);

               Next_Element (Sys.Tasks, My_Iterator);

            end loop;

         end if;
      end loop;

   end Compute_Simulation_Blocking_Time;

   procedure Priority_Inversion_Detection
     (Sys         : in out System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      Priority_Inversion : Priority_Inversion_List;
      Item               : Priority_Inversion_Item_Ptr;
      Tmp_Tasks          : Tasks_Set;
      My_Scheduler       : constant Scheduler_Ptr :=
         Get_Scheduler (extended_Processor (A_Processor.all));

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      Current_Processor_Name := A_Processor.Name;
      Select_And_Copy (Sys.Tasks, Tmp_Tasks, Select_Cpu'Access);

      -- Set priority according to the scheduler
      --
      if (Get_Name (My_Scheduler.all) = Deadline_Monotonic_Protocol) then
         Set_Priority_According_To_Dm
           (Fixed_Priority_Scheduler (My_Scheduler.all),
            Tmp_Tasks);
      else
         if (Get_Name (My_Scheduler.all) = Rate_Monotonic_Protocol) then
            Set_Priority_According_To_Rm
              (Fixed_Priority_Scheduler (My_Scheduler.all),
               Tmp_Tasks);
         end if;
      end if;

      for J in  0 .. Sched.Nb_Entries - 1 loop
         if Sched.Entries (J).Item.Name = A_Processor.Name then

            Priority_Inversion :=
               Priority_Inversion_From_Simulation
                 (Tmp_Tasks,
                  Sys.Resources,
                  A_Processor.Name,
                  Sched.Entries (J).Data.Result);

            if Is_Empty (Priority_Inversion) then
               Result := Result &
                         Lb_No_Priority_Inversion_Found (Current_Language) &
                         Unbounded_Lf;
            else
               Result := Result &
                         Start_Line &
                         Lb_Minus &
                         Lb_Simulation_Priority_Inversion (Current_Language) &
                         Lb_Colon &
                         Unbounded_Lf;

               while not Is_Empty (Priority_Inversion) loop
                  Item := Get_Head (Priority_Inversion);

                  Result :=
                    Result &
                    To_Unbounded_String ("   ") &
                    Item.Task_Name &
                    Lb_Has_Priority_Inversion (Current_Language) &
                    Item.Resource_Name &
                    Lb_From_The_Time (Current_Language) &
                    To_Unbounded_String (Item.Start_Time'Img) &
                    Lb_To_The_Time (Current_Language) &
                    To_Unbounded_String (Item.End_Time'Img) &
                    Lb_Dot &
                    Unbounded_Lf;

                  Delete (Priority_Inversion, Item);

               end loop;
            end if;
         end if;
      end loop;

      Free (Tmp_Tasks);
      Result := Result & Unbounded_Lf & Unbounded_Lf;

   end Priority_Inversion_Detection;

   procedure Deadlock_Detection
     (Sys         : in out System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      Deadlock : Deadlock_List;
      Item     : Deadlock_Item_Ptr;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      for J in  0 .. Sched.Nb_Entries - 1 loop
         if Sched.Entries (J).Item.Name = A_Processor.Name then

            Deadlock :=
               Deadlock_From_Simulation
                 (Sys.Tasks,
                  Sys.Resources,
                  A_Processor.Name,
                  Sched.Entries (J).Data.Result);

            if Is_Empty (Deadlock) then
               Result := Result &
                         Lb_No_Deadlock_Found (Current_Language) &
                         Unbounded_Lf;

            else
               Result := Result &
                         Start_Line &
                         Lb_Minus &
                         Lb_Simulation_Deadlock (Current_Language) &
                         Lb_Colon &
                         Unbounded_Lf;

               while not Is_Empty (Deadlock) loop
                  Item := Get_Head (Deadlock);

                  Result := Result &
                            To_Unbounded_String ("   ") &
                            Lb_Deadlock_At_Time (Current_Language) &
                            To_Unbounded_String (Item.Time'Img) &
                            Lb_Colon;
                  Result := Result & " " & Item.Task_Name;
                  Result := Result &
                            " " &
                            Item.Resource_Name &
                            Lb_Dot &
                            Unbounded_Lf;

                  Delete (Deadlock, Item);

               end loop;
            end if;
         end if;
      end loop;

   end Deadlock_Detection;

end Call_Resource_Framework;
