------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Xml_Tag;                     use Xml_Tag;
with Scheduler.Fixed_Priority;    use Scheduler.Fixed_Priority;
with Scheduler.Fixed_Priority.Dm; use Scheduler.Fixed_Priority.Dm;
with Scheduler.Fixed_Priority.Rm; use Scheduler.Fixed_Priority.Rm;
with Double_Util;                 use Double_Util;
with Translate;                   use Translate;
with Tasks;                       use Tasks;
with Event_Analyzers;             use Event_Analyzers;
with Event_Analyzer_Set;          use Event_Analyzer_Set;
use Event_Analyzer_Set.Generic_Event_Analyzer_Set;
with Ada.Numerics.Aux;            use Ada.Numerics.Aux;
with Unbounded_Strings;           use Unbounded_Strings;
with Scheduling_Sequences;        use Scheduling_Sequences;
with GNAT.Current_Exception;      use GNAT.Current_Exception;
with Expressions;                 use Expressions;
with Framework_Config;            use Framework_Config;
with Time_Units;                  use Time_Units;

package body Call_Scheduling_Framework is

   procedure Compute_Simulation_Time_Line
     (Sys                        : in out System;
      Result                     : in out Unbounded_String;
      Period                     : in Natural;
      Schedule_With_Offsets      : in Boolean;
      Schedule_With_Precedencies : in Boolean;
      Schedule_With_Resources    : in Boolean;
      With_Task_Specific_Seed    : in Boolean       := True;
      Global_Seed_Value          : in Integer       := 0;
      Predictable_Global_Seed    : in Boolean       := True;
      Event_To_Generate          : in Time_Unit_Event_Type_Boolean_Table;
      Output                     : in Output_Format := String_Output)
   is

   begin

      Sort (Sys.Tasks, Increasing_Name'Access);
      Scheduling_Base_Period := Period;

      Initialize (Sched.all);

      Build_Multiprocessor_Scheduling
        (Sys.Processors,
         Sys.Address_Spaces,
         Sys.Tasks,
         Sys.Resources,
         Sys.Buffers,
         Sys.Messages,
         Sys.Dependencies,
         Sched,
         Event_To_Generate,
         Period,
         Schedule_With_Offsets,
         Schedule_With_Precedencies,
         Schedule_With_Resources,
         With_Task_Specific_Seed,
         Global_Seed_Value,
         Predictable_Global_Seed);

   exception
      when Parametric_File_Error =>
         Result :=
           Result &
           Lb_Minus &
           Lb_Simulation_Error (Current_Language) &
           Lb_Parametric_File_Error (Current_Language) &
           Lb_Comma &
           Exception_Message &
           Unbounded_Lf;
      when Type_Error =>
         Result :=
           Result &
           Lb_Minus &
           Lb_Simulation_Error (Current_Language) &
           Lb_Type_Error (Current_Language) &
           Lb_Comma &
           Exception_Message &
           Unbounded_Lf;
      when Statement_Error =>
         Result :=
           Result &
           Lb_Minus &
           Lb_Simulation_Error (Current_Language) &
           Lb_Statement_Error (Current_Language) &
           Lb_Comma &
           Exception_Message &
           Unbounded_Lf;
      when Variable_Error =>
         Result :=
           Result &
           Lb_Minus &
           Lb_Simulation_Error (Current_Language) &
           Lb_Variable_Error (Current_Language) &
           Lb_Comma &
           Exception_Message &
           Unbounded_Lf;
      when Syntax_Error =>
         Result :=
           Result &
           Lb_Minus &
           Lb_Simulation_Error (Current_Language) &
           Lb_Syntax_Error (Current_Language) &
           Lb_Comma &
           Exception_Message &
           Unbounded_Lf;
      when A_Time_Unit.Indexed_Table_Full =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_22 (Current_Language) &
                   Unbounded_Lf;
         Initialize (Sched.all);

   end Compute_Simulation_Time_Line;

   procedure Compute_Simulation_Basics
     (Sys         : in System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      Check : Unbounded_String;

   begin

      Check := Empty_String;
      Compute_Simulation_Number_Of_Context_Switch
        (Sys,
         Check,
         A_Processor,
         Output);
      Result := Result & Check;

      Check := Empty_String;
      Compute_Simulation_Number_Of_Preemption
        (Sys,
         Check,
         A_Processor,
         Output);
      Result := Result & Check;

      Check := Empty_String;
      Compute_Simulation_Response_Time
        (Sys,
         Check,
         A_Processor,
         True,
         False,
         False,
         Output);
      Result := Result & Unbounded_Lf & Check & Unbounded_Lf;

   end Compute_Simulation_Basics;

   procedure Compute_Simulation_Number_Of_Context_Switch
     (Sys         : in System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      Nb : Natural := 0;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      for J in  0 .. Sched.Nb_Entries - 1 loop

         if Sched.Entries (J).Item.Name = A_Processor.Name then

            if Sched.Entries (J).Data.Error_Msg /= Empty_String then
               Result := Result &
                         Sched.Entries (J).Data.Error_Msg &
                         Unbounded_Lf;

            else
               Nb     :=
                  Number_Of_Switch_Context_From_Simulation
                    (Sched.Entries (J).Data.Result);
               Result :=
                 Result &
                 Start_Line &
                 Lb_Minus &
                 Lb_Number_Of_Context_Switch (Current_Language) &
                 Lb_Colon &
                 Start_Num_Of_Context_Switch &
                 Nb'Img &
                 End_Num_Of_Context_Switch &
                 End_Line &
                 Unbounded_Lf;

            end if;
         end if;
      end loop;
   end Compute_Simulation_Number_Of_Context_Switch;

   procedure Compute_Simulation_Number_Of_Preemption
     (Sys         : in System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      Nb : Natural := 0;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      for J in  0 .. Sched.Nb_Entries - 1 loop

         if Sched.Entries (J).Item.Name = A_Processor.Name then

            if Sched.Entries (J).Data.Error_Msg /= Empty_String then
               Result := Result &
                         Sched.Entries (J).Data.Error_Msg &
                         Unbounded_Lf;

            else
               Nb :=
                  Number_Of_Preemption_From_Simulation
                    (Sched.Entries (J).Data.Result,
                     Sys.Tasks);

               Result := Result & Sched.Entries (J).Data.Scheduling_Msg;
               Result :=
                 Result &
                 Start_Line &
                 Lb_Minus &
                 Lb_Number_Of_Preemption (Current_Language) &
                 Lb_Colon &
                 Start_Num_Of_Preemp &
                 Nb'Img &
                 End_Num_Of_Preemp &
                 End_Line &
                 Unbounded_Lf;
            end if;

         end if;
      end loop;

   end Compute_Simulation_Number_Of_Preemption;

   procedure Compute_Simulation_Response_Time
     (Sys          : in System;
      Result       : in out Unbounded_String;
      A_Processor  : in Processor_Ptr;
      Worst_Case   : Boolean;
      Best_Case    : Boolean;
      Average_Case : Boolean;
      Output       : in Output_Format := String_Output)
   is

      Check       : Unbounded_String := Empty_String;
      A_Task      : Generic_Task_Ptr;
      My_Iterator : Tasks_Iterator;
      Min         : Natural          := 0;
      Max         : Natural          := 0;
      Average     : Double           := 0.0;
      Missed      : Boolean          := False;

   begin

      for J in  0 .. Sched.Nb_Entries - 1 loop
         if Sched.Entries (J).Item.Name = A_Processor.Name then

            if Get_Number_Of_Task_From_Processor
                  (Sys.Tasks,
                   A_Processor.Name) >
               0
            then

               Result := Lb_Minus &
                         Lb_Task_Response_Time_From_Simulation (
                 Current_Language) &
                         Lb_Colon &
                         Unbounded_Lf;

               Reset_Iterator (Sys.Tasks, My_Iterator);

               loop
                  Current_Element (Sys.Tasks, A_Task, My_Iterator);

                  Check := Empty_String;
                  if (A_Processor.Name = A_Task.Cpu_Name) then
                     Response_Time_By_Simulation
                       (A_Task,
                        Sched.Entries (J).Data.Result,
                        Check,
                        Max,
                        Min,
                        Average);

                     Result := Result &
                               Start_Task &
                               To_Unbounded_String ("    ") &
                               A_Task.Name &
                               To_Unbounded_String (" =>");
                     if Worst_Case then
                        Result := Result &
                                  Max'Img &
                                  To_Unbounded_String ("/worst ");
                     end if;
                     if Best_Case then
                        Result := Result &
                                  Min'Img &
                                  To_Unbounded_String ("/best ");
                     end if;
                     if Average_Case then
                        Result := Result &
                                  Format (Average) &
                                  To_Unbounded_String ("/average ");
                     end if;
                     Result := Result & Check & Unbounded_Lf;

                     if Check /= Empty_String then
                        Missed := True;
                     end if;
                  end if;

                  exit when Is_Last_Element (Sys.Tasks, My_Iterator);

                  Next_Element (Sys.Tasks, My_Iterator);

               end loop;
            end if;
         end if;
      end loop;

      if Missed = False then
         Result := Result &
                   Lb_No_Deadline_Missed_In_The_Computed_Scheduling (
           Current_Language) &
                   Unbounded_Lf;
      else
         Result := Result &
                   Lb_Deadline_Will_Be_Missed_Task_Are_Not_Schedulable (
           Current_Language) &
                   Unbounded_Lf;

      end if;

   end Compute_Simulation_Response_Time;

   procedure Compute_Feasibility_Period
     (Sys         : in System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      Period_String  : Unbounded_String;
      Free_Ut_String : Unbounded_String;

      Period  : Double := 0.0;
      Util    : Double := 0.0;
      Free_Ut : Double := 0.0;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      begin
         Period := Scheduling_Period (Sys.Tasks, A_Processor.Name);

      exception
      -- A large task set with many different period values can lead to
      -- a CONSTRAINT_ERROR exception due to type overflow
      -- This "Scheduling Period" is computed with the help of a "big
      -- number" package. If you want to increase the accurrancy of this
      -- computation, increase Max_Big_Number_Size in config/config.ads
      --
         when others =>
            null;
      end;

      if Period > Double (Natural'Last) then
         Period_String := To_Unbounded_String (Double'Image (Period));
      else
         Period_String :=
            To_Unbounded_String (Natural'Image (Natural (Period)));
      end if;

      Result :=
        Result &
        Start_Line &
        Lb_Scheduling_Period (Current_Language) &
        Start_Bas_Per &
        Period_String &
        End_Bas_Per &
        Start_Ref &
        To_Unbounded_String (" (") &
        Lb_See (Current_Language) &
        To_Unbounded_String ("[18], page 5). ") &
        End_Ref &
        End_Line &
        Unbounded_Lf;

      Util := Processor_Utilization_Over_Period (Sys.Tasks, A_Processor.Name);

      Free_Ut := Period * (1.0 - Util);
      Free_Ut := Double'Max (0.0, Free_Ut);

      if Free_Ut > Double (Natural'Last) then
         Free_Ut_String := To_Unbounded_String (Double'Image (Free_Ut));
      else
         Free_Ut_String :=
            To_Unbounded_String (Natural'Image (Natural (Free_Ut)));
      end if;

      Result :=
        Result &
        Start_Line &
        To_Unbounded_String ("-") &
        Start_Unit &
        Free_Ut_String &
        End_Unit &
        Lb_Free_Unit_Time (Current_Language) &
        End_Line &
        Unbounded_Lf;

   exception
      when Task_Set.Task_Must_Be_Periodic =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_2 (Current_Language) &
                   Unbounded_Lf;
      when Task_Set.Task_Model_Error =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_3 (Current_Language) &
                   Unbounded_Lf;

   end Compute_Feasibility_Period;

   procedure Compute_Feasibility_Processor_Utilization_Factor
     (Sys         : in System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      Util   : Double;
      Utildi : Double;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      Result := Result &
                Lb_Feasibility_Number1 (Current_Language) &
                Unbounded_Lf &
                Unbounded_Lf;

      Compute_Feasibility_Period (Sys, Result, A_Processor, Output);

      begin
         Utildi :=
            Processor_Utilization_Over_Deadline (Sys.Tasks, A_Processor.Name);
         Result :=
           Result &
           Start_Line &
           Lb_Utilization_With_Deadline (Current_Language) &
           Start_Deadl_Fact &
           Format (Utildi) &
           End_Deadl_Fact &
           Start_Ref &
           To_Unbounded_String (" (") &
           Lb_See (Current_Language) &
           To_Unbounded_String ("[1], page 6). ") &
           End_Ref &
           End_Line &
           Unbounded_Lf;
      exception
         when Task_Set.Task_Must_Be_Periodic =>
            Result := Result &
                      Lb_Compute_Scheduling_Error_9 (Current_Language) &
                      Unbounded_Lf;
         when Task_Set.Task_Model_Error =>
            Result := Result &
                      Lb_Compute_Scheduling_Error_10 (Current_Language) &
                      Unbounded_Lf;
         when Deadline_Error =>
            Result := Result &
                      Lb_Compute_Scheduling_Error_21 (Current_Language) &
                      Unbounded_Lf;

      end;

      begin
         Util   :=
            Processor_Utilization_Over_Period (Sys.Tasks, A_Processor.Name);
         Result :=
           Result &
           Start_Line &
           Lb_Utilization_With_Period (Current_Language) &
           Start_Per_Fact &
           Format (Util) &
           End_Per_Fact &
           Start_Ref &
           To_Unbounded_String (" (") &
           Lb_See (Current_Language) &
           To_Unbounded_String ("[1], page 6). ") &
           End_Ref &
           Unbounded_Lf;
      exception
         when Task_Set.Task_Must_Be_Periodic =>
            Result := Result &
                      Lb_Compute_Scheduling_Error_9 (Current_Language) &
                      Unbounded_Lf;
         when Task_Set.Task_Model_Error =>
            Result := Result &
                      Lb_Compute_Scheduling_Error_10 (Current_Language) &
                      Unbounded_Lf;
      end;

      begin

         Result := Result & Start_Line;
         Utilization_Factor_Feasibility_Test
           (A_Processor.Scheduler.all,
            Sys.Tasks,
            A_Processor.Name,
            Result);

      exception
         when Task_Set.Task_Must_Be_Periodic =>
            Result := Result &
                      Lb_Compute_Scheduling_Error_9 (Current_Language) &
                      Unbounded_Lf;
         when Task_Set.Task_Model_Error =>
            Result := Result &
                      Lb_Compute_Scheduling_Error_10 (Current_Language) &
                      Unbounded_Lf;
         when Deadline_Error =>
            Result := Result &
                      Lb_Compute_Scheduling_Error_21 (Current_Language) &
                      Unbounded_Lf;

      end;

   end Compute_Feasibility_Processor_Utilization_Factor;

   procedure Compute_Feasibility_Response_Time
     (Sys         : in out System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is

      Rt     : Response_Time_Table;
      Msg    : Unbounded_String := Empty_String;
      Missed : Boolean          := False;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      Result := Result &
                Unbounded_Lf &
                Lb_Feasibility_Number2 (Current_Language) &
                Unbounded_Lf &
                Unbounded_Lf;

      Compute_Response_Time
        (A_Processor.Scheduler.all,
         Sys.Tasks,
         A_Processor.Name,
         Msg,
         Rt);

      Result :=
        Result &
        Start_Line &
        Lb_Minus &
        Lb_Bound_On_Task_Response_Time (Current_Language) &
        Lb_Colon &
        Start_Ref &
        Msg &
        End_Ref &
        Unbounded_Lf;

      for I in  0 .. Response_Time_Range (Get_Number_Of_Task_From_Processor
                                             (Sys.Tasks,
                                              A_Processor.Name) -
                                          1)
      loop
         if (Rt.Entries (I).Item /= null) then
            Result :=
              Result &
              Start_Task &
              To_Unbounded_String ("    ") &
              Rt.Entries (I).Item.Name &
              To_Unbounded_String (" =>") &
              Integer'Image (Integer (Rt.Entries (I).Data)) &
              End_Task;

            if (Natural (Rt.Entries (I).Data) >
                Rt.Entries (I).Item.Deadline)
            then
               Result := Result &
                         Lb_Comma &
                         Lb_Check_Deadline (Current_Language) &
                         Rt.Entries (I).Item.Deadline'Img &
                         To_Unbounded_String (")");
               Missed := True;
            end if;

            Result := Result & Unbounded_Lf;

         end if;
      end loop;

      if Missed = False then
         Result := Result &
                   Lb_No_Deadline_Will_Be_Missed_Task_Are_Schedulable (
           Current_Language);
      else
         Result := Result &
                   Lb_Deadline_Will_Be_Missed_Task_Are_Not_Schedulable (
           Current_Language);
      end if;
      Result := Result &
                End_Line &
                Unbounded_Lf &
                End_Sched_Feas &
                Unbounded_Lf;

   exception
      when Task_Set.Task_Must_Have_Period_Equal_To_Deadline =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_4 (Current_Language) &
                   Unbounded_Lf;
      when Task_Set.Task_Must_Be_Periodic =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_5 (Current_Language) &
                   Unbounded_Lf;
      when Task_Set.Task_Model_Error =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_6 (Current_Language) &
                   Unbounded_Lf;
      when Scheduler.Processor_Utilization_Exceeded =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_7 (Current_Language) &
                   Unbounded_Lf;
      when Task_Set.Start_Time_Error =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_12 (Current_Language) &
                   Unbounded_Lf;
      when Task_Set.Offset_Error =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_13 (Current_Language) &
                   Unbounded_Lf;
      when Invalid_Scheduler =>
         Result := Result &
                   Lb_Compute_Scheduling_Error_18 (Current_Language) &
                   Unbounded_Lf;
   end Compute_Feasibility_Response_Time;

   procedure Set_Priorities
     (Sys         : in out System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      A_Scheduler : in Scheduler_Type;
      Output      : in Output_Format := String_Output)
   is

      A_Rm_Scheduler   : Rm_Scheduler;
      A_Dm_Scheduler   : Dm_Scheduler;
      Ok               : Boolean := True;
      A_Task           : Generic_Task_Ptr;
      My_Task_Iterator : Tasks_Iterator;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      begin
         if (A_Scheduler = Rate_Monotonic_Protocol) then
            Set_Priority_According_To_Rm
              (A_Rm_Scheduler,
               Sys.Tasks,
               A_Processor.Name);
         else
            Set_Priority_According_To_Dm
              (A_Dm_Scheduler,
               Sys.Tasks,
               A_Processor.Name);
         end if;

      exception
         when Task_Set.Task_Must_Be_Periodic =>
            Ok := False;

      end;

      if Ok then
         if (A_Scheduler = Rate_Monotonic_Protocol) then
            Result := Result &
                      Start_Line &
                      Lb_Set_Priorities_Rm (Current_Language) &
                      End_Line;
         else
            Result := Result &
                      Start_Line &
                      Lb_Set_Priorities_Dm (Current_Language) &
                      End_Line;
         end if;

         Result := Result &
                   Unbounded_Lf &
                   Start_Line &
                   Lb_Priorities (Current_Language);

         Reset_Iterator (Sys.Tasks, My_Task_Iterator);
         loop
            Current_Element (Sys.Tasks, A_Task, My_Task_Iterator);

            if (A_Task.Cpu_Name = A_Processor.Name) then
               Result :=
                 Result &
                 Unbounded_Lf &
                 Start_Task &
                 To_Unbounded_String ("    ") &
                 A_Task.Name &
                 To_Unbounded_String (" =>") &
                 To_Unbounded_String (A_Task.Priority'Img) &
                 End_Task;
            end if;

            exit when Is_Last_Element (Sys.Tasks, My_Task_Iterator);
            Next_Element (Sys.Tasks, My_Task_Iterator);
         end loop;
         Result := Result &
                   Unbounded_Lf &
                   End_Line &
                   Unbounded_Lf &
                   End_Set_Prio;

      else
         Result := Result & Lb_Priorities_Error (Current_Language);
      end if;

   end Set_Priorities;

   procedure Run_User_Defined_Event_Handlers
     (Sys    : in System;
      Result : in out Unbounded_String;
      Output : in Output_Format := String_Output)
   is

      Msg               : Unbounded_String;
      My_Iterator       : Event_Analyzers_Iterator;
      An_Event_Analyzer : Event_Analyzer_Ptr;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      if (Get_Number_Of_Elements (Sys.Event_Analyzers) > 0) then

         Reset_Iterator (Sys.Event_Analyzers, My_Iterator);

         loop
            Current_Element
              (Sys.Event_Analyzers,
               An_Event_Analyzer,
               My_Iterator);

         -- Run an event analyzer
         ---
            begin
               Msg := Empty_String;

               Result := Result & Unbounded_Lf & Unbounded_Lf;

               Result :=
                 Result &
                 Lb_Minus &
                 Lb_Event_Analyzer_Name (Current_Language) &
                 Lb_Colon &
                 An_Event_Analyzer.Behavior_File_Name &
                 Unbounded_Lf &
                 Unbounded_Lf;

               Run_An_Event_Analyzer
                 (Sys.Tasks,
                  Sys.Processors,
                  Sys.Address_Spaces,
                  Sys.Resources,
                  Sys.Buffers,
                  Sys.Messages,
                  Sched,
                  An_Event_Analyzer,
                  Msg);

               Result := Result & Msg;

               Result := Result & Unbounded_Lf & Unbounded_Lf;

            exception
               when Parametric_File_Error =>
                  Result :=
                    Result &
                    Lb_Minus &
                    Lb_Simulation_Error (Current_Language) &
                    Lb_Parametric_File_Error (Current_Language) &
                    Lb_Comma &
                    Exception_Message &
                    Unbounded_Lf;
               when Type_Error =>
                  Result :=
                    Result &
                    Lb_Minus &
                    Lb_Simulation_Error (Current_Language) &
                    Lb_Type_Error (Current_Language) &
                    Lb_Comma &
                    Exception_Message &
                    Unbounded_Lf;
               when Statement_Error =>
                  Result :=
                    Result &
                    Lb_Minus &
                    Lb_Simulation_Error (Current_Language) &
                    Lb_Statement_Error (Current_Language) &
                    Lb_Comma &
                    Exception_Message &
                    Unbounded_Lf;
               when Variable_Error =>
                  Result :=
                    Result &
                    Lb_Minus &
                    Lb_Simulation_Error (Current_Language) &
                    Lb_Variable_Error (Current_Language) &
                    Lb_Comma &
                    Exception_Message &
                    Unbounded_Lf;
               when Syntax_Error =>
                  Result :=
                    Result &
                    Lb_Minus &
                    Lb_Simulation_Error (Current_Language) &
                    Lb_Syntax_Error (Current_Language) &
                    Lb_Comma &
                    Exception_Message &
                    Unbounded_Lf;
            end;

            exit when Is_Last_Element (Sys.Event_Analyzers, My_Iterator);

            Next_Element (Sys.Event_Analyzers, My_Iterator);

         end loop;

      end if;

   end Run_User_Defined_Event_Handlers;

   procedure Partionning_Tasks
     (Sys             : in out System;
      Result          : in out Unbounded_String;
      A_Parition_Rule : in Partioning_Type;
      Output          : in Output_Format := String_Output)
   is

      Partitionned_Task_Set : Tasks_Set;
      A_Task                : Generic_Task_Ptr;
      My_Task_Iterator      : Tasks_Iterator;
      Msg                   : Unbounded_String := Empty_String;

   begin

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      case A_Parition_Rule is

         when General_Task =>
            Partition_General_Task
              (Sys.Processors,
               Sys.Tasks,
               Msg,
               Partitionned_Task_Set);
         when Best_Fit =>
            Partition_Best_Fit
              (Sys.Processors,
               Sys.Tasks,
               Msg,
               Partitionned_Task_Set);
         when Next_Fit =>
            Partition_Next_Fit
              (Sys.Processors,
               Sys.Tasks,
               Msg,
               Partitionned_Task_Set);
         when First_Fit =>
            Partition_First_Fit
              (Sys.Processors,
               Sys.Tasks,
               Msg,
               Partitionned_Task_Set);
         when Small_Task =>
            Partition_Small_Task
              (Sys.Processors,
               Sys.Tasks,
               Msg,
               Partitionned_Task_Set);
      end case;

      Result :=
        Result &
        Start_Line &
        Lb_Minus &
        Lb_Partitioning_Result (Current_Language) &
        Start_Ref &
        Msg &
        End_Ref &
        Lb_Colon &
        Unbounded_Lf;

      Sys.Tasks := Partitionned_Task_Set;

      Reset_Iterator (Sys.Tasks, My_Task_Iterator);
      loop
         Current_Element (Sys.Tasks, A_Task, My_Task_Iterator);

         Result := Result &
                   Unbounded_Lf &
                   To_Unbounded_String ("    ") &
                   A_Task.Name &
                   To_Unbounded_String (" =>") &
                   A_Task.Cpu_Name;

         exit when Is_Last_Element (Sys.Tasks, My_Task_Iterator);
         Next_Element (Sys.Tasks, My_Task_Iterator);
      end loop;
      Result := Result & Unbounded_Lf & Unbounded_Lf;

   exception
      when Processors_Should_Have_The_Same_Scheduler =>
         Result := Result & Lb_Partition_Error1 (Current_Language);
      when No_Such_Processors =>
         Result := Result & Lb_Partition_Error2 (Current_Language);
      when Invalid_Scheduler_Multiprocessors =>
         Result := Result & Lb_Partition_Error3 (Current_Language);

   end Partionning_Tasks;

   procedure Compute_Feasibility_Basics
     (Sys         : in out System;
      Result      : in out Unbounded_String;
      A_Processor : in Processor_Ptr;
      Output      : in Output_Format := String_Output)
   is
   begin

      Compute_Feasibility_Processor_Utilization_Factor
        (Sys,
         Result,
         A_Processor,
         Output);

      Compute_Feasibility_Response_Time (Sys, Result, A_Processor, Output);

   end Compute_Feasibility_Basics;

begin

   Sched := new Scheduling_Table;
   Initialize (Sched.all);

end Call_Scheduling_Framework;
