------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Tasks;             use Tasks;
with Unbounded_Strings; use Unbounded_Strings;
with Ada.Numerics.Aux;  use Ada.Numerics.Aux;
with Framework_Config;  use Framework_Config;
with Processors;        use Processors;

package body Dependency_Services is

   procedure Set_Parameters_According_To_Chetto
     (My_Dependencies    : in Tasks_Dependencies_Ptr;
      My_Tasks           : in out Tasks_Set;
      Compute_Deadlines  : in Boolean;
      Compute_Priorities : in Boolean)
   is

      Leaf, Step                         : Tasks_Set;
      Previous                           : Tasks_Set;
      Real_Left, Real_Right, Left, Right : Generic_Task_Ptr;
      Ite1, Ite2                         : Tasks_Iterator;

   begin

      if Is_Cyclic (My_Dependencies) then
         raise Cyclic_Graph_Error;
      end if;

      Dependencies_Periods_Control (My_Tasks, My_Dependencies, Empty_String);

      -- Assume the graph is acyclic.
      -- Extract leaves of the graph and update deadline
      -- of previous node
      --
      Step := Get_Leaf_Tasks (My_Dependencies);

      while not Is_Empty (Step) loop

         Reset (Leaf);
         Duplicate (Step, Leaf);
         Reset (Step);

         Reset_Iterator (Leaf, Ite1);
         for I in  0 .. Get_Number_Of_Elements (Leaf) - 1 loop

            -- The task to be updated
            --
            Current_Element (Leaf, Right, Ite1);

            -- Take previous tasks
            --
            if Has_Predecessor (My_Dependencies, Right) then

               Reset (Previous);
               Previous := Get_Predecessors_List (My_Dependencies, Right);

               Reset_Iterator (Previous, Ite2);

               for J in  0 .. Get_Number_Of_Elements (Previous) - 1 loop

                  Current_Element (Previous, Left, Ite2);
                  Add (Step, Left);

                  Real_Left  := Search_Task (My_Tasks, Left.Name);
                  Real_Right := Search_Task (My_Tasks, Right.Name);

                  if Compute_Priorities then
                     if Right.Priority >= 1 then
                        Real_Left.Priority :=
                           Priority_Range'Max
                             (Real_Left.Priority,
                              Real_Right.Priority + 1);
                     end if;
                  end if;

                  if Compute_Deadlines then
                     Real_Left.Deadline :=
                        Natural'Min
                          (Real_Left.Deadline,
                           Real_Right.Deadline - Real_Right.Capacity);
                  end if;

                  Next_Element (Previous, Ite2);
               end loop;
            end if;

            Next_Element (Leaf, Ite1);
         end loop;

      end loop;

   end Set_Parameters_According_To_Chetto;

   function Get_Response_Time
     (Rtn    : in Scheduler.Response_Time_Table;
      A_Task : Generic_Task_Ptr)
      return   Double
   is
   begin
      for I in  Scheduler.Response_Time_Range loop
         if (Rtn.Entries (I).Item.Name = A_Task.Name) then
            return Rtn.Entries (I).Data;
         end if;
      end loop;

      raise Task_Set.Task_Not_Found;
   end Get_Response_Time;

   procedure Inject_Response_Time_Into_Jitter
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Tasks        : in out Tasks_Set;
      Rtn             : in Scheduler.Response_Time_Table)
   is

      Max_Jitter       : Double := 0.0;
      Ite1, Ite2       : Tasks_Iterator;
      Previous, A_Task : Generic_Task_Ptr;
      Pred             : Tasks_Set;

   begin

      -- Scan each task. If a task has a predecessor, set its jitter
      -- to the response time of its predecessors
      --
      Reset_Iterator (My_Tasks, Ite1);
      loop
         Current_Element (My_Tasks, A_Task, Ite1);

         -- Find maximum response time of predecessors
         --
         if Has_Predecessor (My_Dependencies, A_Task) then
            Reset (Pred);
            Pred := Get_Predecessors_List (My_Dependencies, A_Task);

            if not Is_Empty (Pred) then
               Reset_Iterator (Pred, Ite2);
               Max_Jitter := 0.0;
               for I in  0 .. Get_Number_Of_Elements (Pred) - 1 loop

                  Current_Element (Pred, Previous, Ite2);

                  Max_Jitter :=
                     Double'Max
                       (Max_Jitter,
                        Get_Response_Time (Rtn, Previous));

                  Next_Element (Pred, Ite2);
               end loop;

               -- set jitter with max_jitter
               --
               case A_Task.Task_Type is
                  when Periodic_Type =>
                     Periodic_Task_Ptr (A_Task).Jitter :=
                        Natural'Max
                          (Periodic_Task_Ptr (A_Task).Jitter,
                           Natural (Max_Jitter));
                  when Poisson_Type =>
                     Poisson_Task_Ptr (A_Task).Jitter :=
                        Natural'Max
                          (Poisson_Task_Ptr (A_Task).Jitter,
                           Natural (Max_Jitter));
                  when others =>
                     raise Constraint_Error;
               end case;
            end if;
         end if;

         exit when Is_Last_Element (My_Tasks, Ite1);
         Next_Element (My_Tasks, Ite1);
      end loop;

   end Inject_Response_Time_Into_Jitter;

   procedure Compute_All_Response_Times
     (My_Processors : in Processors_Set;
      My_Tasks      : in out Tasks_Set;
      Msg           : in out Unbounded_String;
      Rtn           : in out Scheduler.Response_Time_Table)
   is

      I           : Natural;
      K, J        : Scheduler.Response_Time_Range;
      A_Processor : Processor_Ptr;
      My_Iterator : Processors_Iterator;

      type Rt is
        array (0 .. Max_Processors - 1) of Scheduler.Response_Time_Table;
      Rt_By_Processor : Rt;

   begin
      -- Compute tasks
      --
      I := 0;
      Reset_Iterator (My_Processors, My_Iterator);

      loop
         Current_Element (My_Processors, A_Processor, My_Iterator);

         Msg := Empty_String;
         if Get_Number_Of_Task_From_Processor (My_Tasks, A_Processor.Name) >
            0
         then
            Compute_Response_Time
              (A_Processor.Scheduler.all,
               My_Tasks,
               A_Processor.Name,
               Msg,
               Rt_By_Processor (I));
         end if;

         exit when Is_Last_Element (My_Processors, My_Iterator);

         Next_Element (My_Processors, My_Iterator);
         I := I + 1;
      end loop;

      I := 0;
      K := 0;

      Reset_Iterator (My_Processors, My_Iterator);

      loop
         J := 0;
         Current_Element (My_Processors, A_Processor, My_Iterator);

         for L in  0 .. (Get_Number_Of_Task_From_Processor
                            (My_Tasks,
                             A_Processor.Name) -
                         1)
         loop
            Rtn.Entries (K) := Rt_By_Processor (I).Entries (J);
            Rtn.Nb_Entries  := Rtn.Nb_Entries + 1;
            K               := K + 1;
            J               := J + 1;
         end loop;

         exit when Is_Last_Element (My_Processors, My_Iterator);

         Next_Element (My_Processors, My_Iterator);
         I := I + 1;

      end loop;

   end Compute_All_Response_Times;

   procedure Compute_End_To_End_Response_Time
     (My_Dependencies  : in Tasks_Dependencies_Ptr;
      My_Processors    : in Processors_Set;
      My_Tasks         : in out Tasks_Set;
      One_Step         : in Boolean;
      Update_Tasks_Set : in Boolean;
      Msg              : in out Unbounded_String;
      Rt               : in out Response_Time_Table)
   is

      Rtn, Rtn1 : Response_Time_Table;
      Tmp_Tasks : Tasks_Set;

   begin

      Initialize (Rtn);
      Initialize (Rtn1);

      if not Update_Tasks_Set then
         Duplicate (My_Tasks, Tmp_Tasks);
      end if;

      Compute_All_Response_Times (My_Processors, My_Tasks, Msg, Rtn);

      loop

      -- Stop when end to end response time do not change
      -- from an iteration to another
      --
         exit when Is_Equal (Rtn, Rtn1);

         -- Update jitter ....
         --
         if not Update_Tasks_Set then
            Inject_Response_Time_Into_Jitter
              (My_Dependencies,
               Tmp_Tasks,
               Rtn);
         else
            Inject_Response_Time_Into_Jitter (My_Dependencies, My_Tasks, Rtn);
         end if;

         if One_Step then
            exit;
         end if;

         Rtn1 := Rtn;
         Initialize (Rtn);

         if not Update_Tasks_Set then
            Compute_All_Response_Times (My_Processors, Tmp_Tasks, Msg, Rtn);
         else
            Compute_All_Response_Times (My_Processors, My_Tasks, Msg, Rtn);
         end if;

         -- Stop when all deadlines are missed
         --
         exit when All_Deadlines_Missed (Rtn);

      end loop;

      Rt := Rtn;

   end Compute_End_To_End_Response_Time;

   function All_Deadlines_Missed
     (Rtn  : in Response_Time_Table)
      return Boolean
   is
   begin

      for I in  0 .. Rtn.Nb_Entries - 1 loop
         if Rtn.Entries (I).Data <
            Double (Rtn.Entries (I).Item.Deadline)
         then
            return False;
         end if;
      end loop;

      return True;
   end All_Deadlines_Missed;

   function Is_Equal
     (Rtn  : in Response_Time_Table;
      Rtn1 : in Response_Time_Table)
      return Boolean
   is
   begin

      if Rtn.Nb_Entries /= Rtn1.Nb_Entries then
         return False;
      end if;

      for I in  0 .. Rtn.Nb_Entries - 1 loop
         if Rtn.Entries (I).Data /= Rtn1.Entries (I).Data then
            return False;
         end if;
      end loop;

      return True;
   end Is_Equal;

   -- To schedule with precedencies
   -- all task of a dag must have the same period
   --
   procedure Dependencies_Periods_Control
     (My_Tasks       : in Tasks_Set;
      My_Deps        : in Tasks_Dependencies_Ptr;
      Processor_Name : in Unbounded_String)
   is

      Next                          : Tasks_Set;
      Real_Task1, Real_Right, Right : Generic_Task_Ptr;
      Ite1, Ite2                    : Tasks_Iterator;
      Task1                         : Generic_Task_Ptr;

   begin

      Reset_Iterator (My_Tasks, Ite1);

      loop
         Current_Element (My_Tasks, Task1, Ite1);

         if (Task1.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then

            -- Check period of successors
            --
            if Has_Successor (My_Deps, Task1) then

               Reset (Next);
               Next := Get_Successors_List (My_Deps, Task1);

               Reset_Iterator (Next, Ite2);
               for J in  0 .. Get_Number_Of_Elements (Next) - 1 loop

                  Current_Element (Next, Right, Ite2);

                  Real_Task1 := Search_Task (My_Tasks, Task1.Name);
                  Real_Right := Search_Task (My_Tasks, Right.Name);

                  if (Real_Right.Task_Type /= Real_Task1.Task_Type) then
                     raise Dependencies_Period_Error;
                  end if;

                  if (Real_Right.Task_Type = Parametric_Type) then
                     raise Dependencies_Period_Error;
                  end if;

                  if (Real_Right.Task_Type = Periodic_Type) then
                     if (Periodic_Task_Ptr (Real_Right).Period /=
                         Periodic_Task_Ptr (Real_Task1).Period)
                     then
                        raise Dependencies_Period_Error;
                     end if;
                  end if;

                  if (Real_Right.Task_Type = Poisson_Type) then
                     if (Poisson_Task_Ptr (Real_Right).Period /=
                         Poisson_Task_Ptr (Real_Task1).Period)
                     then
                        raise Dependencies_Period_Error;
                     end if;
                  end if;

                  Next_Element (Next, Ite2);
               end loop;
            end if;

            -- Check period of predecessors
            --
            if Has_Predecessor (My_Deps, Task1) then

               Reset (Next);
               Next := Get_Predecessors_List (My_Deps, Task1);

               Reset_Iterator (Next, Ite2);
               for J in  0 .. Get_Number_Of_Elements (Next) - 1 loop

                  Current_Element (Next, Right, Ite2);

                  Real_Task1 := Search_Task (My_Tasks, Task1.Name);
                  Real_Right := Search_Task (My_Tasks, Right.Name);

                  if (Real_Right.Task_Type /= Real_Task1.Task_Type) then
                     raise Dependencies_Period_Error;
                  end if;

                  if (Real_Right.Task_Type = Parametric_Type) then
                     raise Dependencies_Period_Error;
                  end if;

                  if (Real_Right.Task_Type = Periodic_Type) then
                     if (Periodic_Task_Ptr (Real_Right).Period /=
                         Periodic_Task_Ptr (Real_Task1).Period)
                     then
                        raise Dependencies_Period_Error;
                     end if;
                  end if;

                  if (Real_Right.Task_Type = Poisson_Type) then
                     if (Poisson_Task_Ptr (Real_Right).Period /=
                         Poisson_Task_Ptr (Real_Task1).Period)
                     then
                        raise Dependencies_Period_Error;
                     end if;
                  end if;

                  Next_Element (Next, Ite2);
               end loop;
            end if;

         end if;

         exit when Is_Last_Element (My_Tasks, Ite1);
         Next_Element (My_Tasks, Ite1);

      end loop;
   end Dependencies_Periods_Control;

end Dependency_Services;
