------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Exceptions;    use Ada.Exceptions;
with Objects;           use Objects;
with Objects.extended;  use Objects.extended;
with Translate;         use Translate;
with Unbounded_Strings; use Unbounded_Strings;

package body Event_Analyzer_Set is

   procedure Check_Event_Analyzer
     (My_Event_Analyzers : in Event_Analyzers_Set;
      File_Name          : in Unbounded_String)
   is

   begin

      if File_Name = "" then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Event_Analyzer_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

      if not Is_A_Valid_Identifier (File_Name) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Event_Analyzer_Name (Current_Language) &
                Lb_Colon &
                Lb_Invalid_Identifier (Current_Language)));
      end if;

   end Check_Event_Analyzer;

   procedure Add_Event_Analyzer
     (My_Event_Analyzers : in out Event_Analyzers_Set;
      File_Name          : in Unbounded_String)
   is

      An_Event_Analyzer : Event_Analyzer_Ptr;
      My_Iterator       : Iterator;

   begin

      Check_Event_Analyzer (My_Event_Analyzers, File_Name);

      if (Get_Number_Of_Elements (My_Event_Analyzers) > 0) then

         Reset_Iterator (My_Event_Analyzers, My_Iterator);

         loop
            Current_Element
              (My_Event_Analyzers,
               An_Event_Analyzer,
               My_Iterator);
            if (File_Name = An_Event_Analyzer.Behavior_File_Name) then
               Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_Event_Analyzer (Current_Language) &
                      " " &
                      File_Name &
                      " : " &
                      Lb_Event_Analyzer_Name (Current_Language) &
                      Lb_Already_Defined (Current_Language)));
            end if;

            exit when Is_Last_Element (My_Event_Analyzers, My_Iterator);

            Next_Element (My_Event_Analyzers, My_Iterator);

         end loop;

      end if;

      An_Event_Analyzer                    := new Event_Analyzer;
      An_Event_Analyzer.Behavior_File_Name := File_Name;

      Add (My_Event_Analyzers, An_Event_Analyzer);

   exception
      when Full_Set =>
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Can_Not_Define_More_Event_Analyzers (Current_Language)));

   end Add_Event_Analyzer;

   function Search_Event_Analyzer
     (My_Event_Analyzers : in Event_Analyzers_Set;
      Name               : in Unbounded_String)
      return               Event_Analyzer_Ptr
   is
      My_Iterator       : Iterator;
      An_Event_Analyzer : Event_Analyzer_Ptr;
      Result            : Event_Analyzer_Ptr;

      Found : Boolean := False;

   begin

      Reset_Iterator (My_Event_Analyzers, My_Iterator);

      loop
         Current_Element (My_Event_Analyzers, An_Event_Analyzer, My_Iterator);

         if (An_Event_Analyzer.Behavior_File_Name = Name) then
            Found  := True;
            Result := An_Event_Analyzer;
         end if;

         exit when Is_Last_Element (My_Event_Analyzers, My_Iterator);

         Next_Element (My_Event_Analyzers, My_Iterator);

      end loop;

      if not Found then
         raise Event_Analyzer_Not_Found;
      end if;

      return Result;
   end Search_Event_Analyzer;

   function Export_Xml
     (My_Event_Analyzers : in Event_Analyzers_Set)
      return               Unbounded_String
   is
      A_Event_Analyzer : Event_Analyzer_Ptr;
      My_Iterator      : Event_Analyzers_Iterator;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Event_Analyzers) then

         Result := Result &
                   To_Unbounded_String ("<event_analyzers>") &
                   Unbounded_Lf;
         Reset_Iterator (My_Event_Analyzers, My_Iterator);

         loop
            Current_Element
              (My_Event_Analyzers,
               A_Event_Analyzer,
               My_Iterator);

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "<event_analyzer");
            Result :=
              Result &
              To_Unbounded_String
                 (" parametric_file_name=""" &
                  To_String (A_Event_Analyzer.Behavior_File_Name) &
                  """" &
                  "> </event_analyzer>") &
              Unbounded_Lf;

            exit when Is_Last_Element (My_Event_Analyzers, My_Iterator);

            Next_Element (My_Event_Analyzers, My_Iterator);

         end loop;
         Result := Result &
                   To_Unbounded_String ("</event_analyzers>") &
                   Unbounded_Lf;
      end if;

      return Result;

   end Export_Xml;

   function Export_Aadl_Properties
     (My_Event_Analyzers : in Event_Analyzers_Set;
      Number_Of_Ht       : in Natural)
      return               Unbounded_String
   is
      An_Event_Analyzer : Event_Analyzer_Ptr;
      My_Iterator       : Event_Analyzers_Iterator;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Event_Analyzers) then

         Reset_Iterator (My_Event_Analyzers, My_Iterator);

         for I in  1 .. Number_Of_Ht loop
            Result := Result & ASCII.HT;
         end loop;
         Result := Result & To_Unbounded_String ("Source_Text => ") & """";

         loop
            Current_Element
              (My_Event_Analyzers,
               An_Event_Analyzer,
               My_Iterator);

            Result := Result &
                      To_String (An_Event_Analyzer.Behavior_File_Name);

            exit when Is_Last_Element (My_Event_Analyzers, My_Iterator);

            Next_Element (My_Event_Analyzers, My_Iterator);

            Result := Result & ",";

         end loop;

         Result := Result & """" & ";" & Unbounded_Lf;

      end if;

      return Result;

   end Export_Aadl_Properties;

end Event_Analyzer_Set;
