------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Time_Units;        use Time_Units;
use Time_Units.A_Time_Unit;
with Double_Util;       use Double_Util;
with Unbounded_Strings; use Unbounded_Strings;

package body Message_Set is

   procedure Add_Message
     (My_Messages        : in out Messages_Set;
      Name               : in Unbounded_String;
      Size               : in Natural;
      Period             : in Natural;
      Deadline           : in Natural;
      Jitter             : in Natural;
      Param              : in User_Defined_Parameters_Table :=
        No_User_Defined_Parameter;
      Response_Time      : in Natural;
      Communication_Time : in Natural)
   is

      Dummy : Message_Ptr;
   begin
      Add_Message
        (My_Messages,
         Dummy,
         Name,
         Size,
         Period,
         Deadline,
         Jitter,
         Param,
         Response_Time,
         Communication_Time);
   end Add_Message;

   procedure Check_Message
     (My_Messages        : in Messages_Set;
      Name               : in Unbounded_String;
      Size               : in Natural;
      Period             : in Natural;
      Deadline           : in Natural;
      Jitter             : in Natural;
      Param              : in User_Defined_Parameters_Table :=
        No_User_Defined_Parameter;
      Response_Time      : in Natural;
      Communication_Time : in Natural)
   is

      My_Iterator : Iterator;
      A_Message   : Message_Ptr;

   begin

      if (Name = "") then
         raise Invalid_Parameter;
      end if;

      if (Size = 0) then
         raise Invalid_Parameter;
      end if;

      if (Jitter > Deadline) then
         raise Invalid_Parameter;
      end if;

      if (Get_Number_Of_Elements (My_Messages) > 0) then

         Reset_Iterator (My_Messages, My_Iterator);

         loop
            Current_Element (My_Messages, A_Message, My_Iterator);
            if (Name = A_Message.Name) then
               raise Message_Already_Defined;
            end if;

            exit when Is_Last_Element (My_Messages, My_Iterator);

            Next_Element (My_Messages, My_Iterator);
         end loop;
      end if;

   end Check_Message;

   procedure Add_Message
     (My_Messages        : in out Messages_Set;
      A_Message          : in out Message_Ptr;
      Name               : in Unbounded_String;
      Size               : in Natural;
      Period             : in Natural;
      Deadline           : in Natural;
      Jitter             : in Natural;
      Param              : in User_Defined_Parameters_Table :=
        No_User_Defined_Parameter;
      Response_Time      : in Natural;
      Communication_Time : in Natural)
   is

      New_Periodic_Message  : Periodic_Message_Ptr;
      New_Aperiodic_Message : Aperiodic_Message_Ptr;

   begin

      Check_Message
        (My_Messages,
         Name,
         Size,
         Period,
         Deadline,
         Jitter,
         Param,
         Response_Time,
         Communication_Time);

      if (Period /= 0) then
         New_Periodic_Message        := new Periodic_Message;
         New_Periodic_Message.Jitter := Jitter;
         New_Periodic_Message.Period := Period;
         A_Message                   := Message_Ptr (New_Periodic_Message);

      else
         New_Aperiodic_Message := new Aperiodic_Message;
         A_Message             := Message_Ptr (New_Aperiodic_Message);
      end if;

      A_Message.Name               := Name;
      A_Message.Size               := Size;
      A_Message.Response_Time      := Response_Time;
      A_Message.Communication_Time := Communication_Time;
      A_Message.Deadline           := Deadline;
      A_Message.Param              := Param;

      Add (My_Messages, A_Message);

   end Add_Message;

   procedure Update_Message
     (My_Messages        : in out Messages_Set;
      Name               : in Unbounded_String;
      New_Name           : in Unbounded_String;
      Size               : in Natural;
      Period             : in Natural;
      Deadline           : in Natural;
      Jitter             : in Natural;
      Param              : in User_Defined_Parameters_Table :=
        No_User_Defined_Parameter;
      Response_Time      : in Natural;
      Communication_Time : in Natural)
   is

      A_Message          : Message_Ptr;
      A_Periodic_Message : Periodic_Message_Ptr;
   begin
      A_Message                    := Search_Message (My_Messages, Name);
      A_Message.Name               := New_Name;
      A_Message.Size               := Size;
      A_Message.Deadline           := Deadline;
      A_Message.Param              := Param;
      A_Message.Response_Time      := Response_Time;
      A_Message.Communication_Time := Communication_Time;
      if (A_Message.Message_Type = Periodic_Type) then
         A_Periodic_Message        := Periodic_Message_Ptr (A_Message);
         A_Periodic_Message.Period := Period;
         A_Periodic_Message.Jitter := Jitter;
      end if;

   end Update_Message;

   function Have_Deadlines_Equal_Than_Periods
     (My_Messages : in Messages_Set)
      return        Boolean
   is
      Equal_Than  : Boolean := True;
      A_Message   : Message_Ptr;
      My_Iterator : Iterator;

   begin

      Reset_Iterator (My_Messages, My_Iterator);

      loop
         Current_Element (My_Messages, A_Message, My_Iterator);

         if (A_Message.Message_Type /= Periodic_Type) then
            return False;
         end if;

         if (Periodic_Message_Ptr (A_Message).Deadline =
             Periodic_Message_Ptr (A_Message).Period)
         then
            Equal_Than := False;
         end if;

         exit when Is_Last_Element (My_Messages, My_Iterator);

         Next_Element (My_Messages, My_Iterator);

      end loop;

      return Equal_Than;

   end Have_Deadlines_Equal_Than_Periods;

   procedure Get_Global_Message_Type
     (My_Messages     : in Messages_Set;
      Has_Global_Type : out Boolean;
      Global_Type     : out Messages_Type)
   is

      A_Message   : Message_Ptr;
      My_Iterator : Iterator;

   begin

      Has_Global_Type := True;
      Reset_Iterator (My_Messages, My_Iterator);

      -- Find the first message of the processor
      --
      Current_Element (My_Messages, A_Message, My_Iterator);

      Global_Type := A_Message.Message_Type;
      if Is_Last_Element (My_Messages, My_Iterator) then
         return;
      else
         Next_Element (My_Messages, My_Iterator);
      end if;

      loop
         Current_Element (My_Messages, A_Message, My_Iterator);

         if (Global_Type /= A_Message.Message_Type) then
            Has_Global_Type := False;
            return;
         end if;

         exit when Is_Last_Element (My_Messages, My_Iterator);

         Next_Element (My_Messages, My_Iterator);

      end loop;

   end Get_Global_Message_Type;

   function Search_Message
     (My_Messages : in Messages_Set;
      Name        : in Unbounded_String)
      return        Message_Ptr
   is
      My_Iterator : Iterator;
      A_Message   : Message_Ptr;
      Result      : Message_Ptr;

      Found : Boolean := False;

   begin

      Reset_Iterator (My_Messages, My_Iterator);

      loop
         Current_Element (My_Messages, A_Message, My_Iterator);

         if (A_Message.Name = Name) then
            Found  := True;
            Result := A_Message;
         end if;

         exit when Is_Last_Element (My_Messages, My_Iterator);

         Next_Element (My_Messages, My_Iterator);

      end loop;

      if not Found then
         raise Message_Not_Found;
      end if;

      return Result;
   end Search_Message;

   function Increasing_Period
     (Op1  : in Message_Ptr;
      Op2  : in Message_Ptr)
      return Boolean
   is
   begin
      return (Periodic_Message_Ptr (Op1).Period <=
              Periodic_Message_Ptr (Op2).Period);
   end Increasing_Period;

   function Decreasing_Period
     (Op1  : in Message_Ptr;
      Op2  : in Message_Ptr)
      return Boolean
   is
   begin
      return (Periodic_Message_Ptr (Op1).Period >=
              Periodic_Message_Ptr (Op2).Period);
   end Decreasing_Period;

   function Increasing_Deadline
     (Op1  : in Message_Ptr;
      Op2  : in Message_Ptr)
      return Boolean
   is
   begin
      return (Periodic_Message_Ptr (Op1).Deadline <=
              Periodic_Message_Ptr (Op2).Deadline);
   end Increasing_Deadline;

   function Decreasing_Deadline
     (Op1  : in Message_Ptr;
      Op2  : in Message_Ptr)
      return Boolean
   is
   begin
      return (Periodic_Message_Ptr (Op1).Deadline >=
              Periodic_Message_Ptr (Op2).Deadline);
   end Decreasing_Deadline;

   function Get
     (My_Messages  : in Messages_Set;
      Message_Name : in Unbounded_String;
      Param_Name   : in Message_Parameters)
      return         Natural
   is
      A_Message   : Message_Ptr;
      My_Iterator : Iterator;

   begin

      if ((Param_Name /= Size) and
          (Param_Name /= Period) and
          (Param_Name /= Deadline) and
          (Param_Name /= Response_Time) and
          (Param_Name /= Jitter))
      then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Messages, My_Iterator);

      loop
         Current_Element (My_Messages, A_Message, My_Iterator);

         if (A_Message.Name = Message_Name) then
            exit;
         end if;

         exit when Is_Last_Element (My_Messages, My_Iterator);

         Next_Element (My_Messages, My_Iterator);

      end loop;

      if (Param_Name = Size) then
         return A_Message.Size;
      else
         if (Param_Name = Deadline) then
            return Periodic_Message_Ptr (A_Message).Deadline;
         else
            if (Param_Name = Response_Time) then
               return A_Message.Response_Time;
            else
               if (Param_Name = Communication_Time) then
                  return A_Message.Communication_Time;
               else
                  if (Param_Name = Period) then
                     return Periodic_Message_Ptr (A_Message).Period;
                  else
                     return Periodic_Message_Ptr (A_Message).Jitter;
                  end if;
               end if;
            end if;
         end if;
      end if;
   end Get;

   procedure Set
     (My_Messages  : in out Messages_Set;
      Message_Name : in Unbounded_String;
      Param_Name   : in Message_Parameters;
      Param_Value  : in Natural)
   is
      A_Message   : Message_Ptr;
      My_Iterator : Iterator;

   begin

      if ((Param_Name /= Size) and
          (Param_Name /= Period) and
          (Param_Name /= Deadline) and
          (Param_Name /= Jitter)) and
         (Param_Name /= Response_Time) and
         (Param_Name /= Communication_Time)
      then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Messages, My_Iterator);

      loop
         Current_Element (My_Messages, A_Message, My_Iterator);

         if (A_Message.Name = Message_Name) then
            if (Param_Name = Size) then
               A_Message.Size := Param_Value;
            else
               if (Param_Name = Response_Time) then
                  A_Message.Response_Time := Param_Value;
               else
                  if (Param_Name = Communication_Time) then
                     A_Message.Communication_Time := Param_Value;
                     exit;
                  end if;
               end if;
            end if;
         end if;
         if (Param_Name = Deadline) then
            Periodic_Message_Ptr (A_Message).Deadline := Param_Value;
         end if;
         if (Param_Name = Period) then
            Periodic_Message_Ptr (A_Message).Period := Param_Value;
            exit;
         end if;
         if (Param_Name = Jitter) then
            Periodic_Message_Ptr (A_Message).Jitter := Param_Value;
            exit;
         end if;
         exit when Is_Last_Element (My_Messages, My_Iterator);

         Next_Element (My_Messages, My_Iterator);

      end loop;
   end Set;

   function Export_Xml
     (My_Messages : in Messages_Set)
      return        Unbounded_String
   is
      My_Iterator : Messages_Iterator;
      A_Message   : Message_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Messages) then

         Result := Result &
                   To_Unbounded_String ("<messages>") &
                   Unbounded_Lf;
         Reset_Iterator (My_Messages, My_Iterator);

         loop
            Current_Element (My_Messages, A_Message, My_Iterator);

            Result := Result & To_Unbounded_String (ASCII.HT & "<message");
            if A_Message.X /= 0 then
               Result := Result &
                         To_Unbounded_String
                            (" x=""" & A_Message.X'Img & """ ");
            end if;
            if A_Message.Y /= 0 then
               Result := Result &
                         To_Unbounded_String
                            ("y=""" & A_Message.Y'Img & """ ");
            end if;
            Result := Result & To_Unbounded_String (">") & Unbounded_Lf;

            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<name>" &
                  To_String (A_Message.Name) &
                  "</name>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<size>" &
                  A_Message.Size'Img &
                  "</size>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<deadline>" &
                  A_Message.Deadline'Img &
                  "</deadline>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<response_time>" &
                  A_Message.Response_Time'Img &
                  "</response_time>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<communication_time>" &
                  A_Message.Communication_Time'Img &
                  "</communication_time>") &
              Unbounded_Lf;

            if (A_Message.Message_Type = Periodic_Type) then
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "<period>" &
                     Periodic_Message_Ptr (A_Message).Period'Img &
                     "</period>") &
                 Unbounded_Lf;
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "<jitter>" &
                     Periodic_Message_Ptr (A_Message).Jitter'Img &
                     "</jitter>") &
                 Unbounded_Lf;
            end if;

            if A_Message.Param.Nb_Entries > 0 then
               Result := Result &
                         To_Unbounded_String
                            (ASCII.HT & ASCII.HT & "<parameters>") &
                         Unbounded_Lf;
               for I in  0 .. A_Message.Param.Nb_Entries - 1 loop

                  Result := Result &
                            To_Unbounded_String
                               (ASCII.HT &
                                ASCII.HT &
                                ASCII.HT &
                                "<parameter");
                  if A_Message.Param.Entries (I).ptype =
                     Boolean_Parameter
                  then
                     Result := Result &
                               To_Unbounded_String
                                  (" parameter_type=""boolean""> ");
                     Result := Result &
                               To_Unbounded_String
                                  (" " &
                                   A_Message.Param.Entries (I).Boolean_value'
                       Img &
                                   " ");
                  else
                     if A_Message.Param.Entries (I).ptype =
                        Integer_Parameter
                     then
                        Result := Result &
                                  To_Unbounded_String
                                     (" parameter_type=""integer""> ");
                        Result := Result &
                                  To_Unbounded_String
                                     (" " &
                                      A_Message.Param.Entries (I).
                          Integer_value'Img &
                                      " ");
                     else
                        if A_Message.Param.Entries (I).ptype =
                           Double_Parameter
                        then
                           Result := Result &
                                     To_Unbounded_String
                                        (" parameter_type=""double""> ");
                           Result :=
                             Result &
                             To_Unbounded_String
                                (" " &
                                 To_String
                                    (Format
                                        (A_Message.Param.Entries (I).
                             double_value)) &
                                 " ");
                        else
                           Result := Result &
                                     To_Unbounded_String
                                        (" parameter_type=""string""> ");
                           Result :=
                             Result &
                             To_Unbounded_String
                                (" " &
                                 To_String
                                    (A_Message.Param.Entries (I).string_value)
                                 &
                                 " ");
                        end if;
                     end if;
                  end if;
                  Result := Result &
                            To_Unbounded_String
                               (To_String
                                   (" " & A_Message.Param.Entries (I).Name) &
                                " ");
                  Result := Result &
                            To_Unbounded_String ("</parameter>") &
                            Unbounded_Lf;
               end loop;
               Result := Result &
                         To_Unbounded_String
                            (ASCII.HT & ASCII.HT & "</parameters>") &
                         Unbounded_Lf;
            end if;

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "</message>") &
                      Unbounded_Lf;

            exit when Is_Last_Element (My_Messages, My_Iterator);

            Next_Element (My_Messages, My_Iterator);

         end loop;
         Result := Result &
                   To_Unbounded_String ("</messages>") &
                   Unbounded_Lf;
      end if;

      return Result;

   end Export_Xml;

   function Export_Aadl_User_Defined_Properties
     (My_Messages : in Messages_Set)
      return        Unbounded_String
   is
      Result : Unbounded_String := Empty_String;

   begin
      return Result;
   end Export_Aadl_User_Defined_Properties;

   function Export_Aadl_Implementations
     (My_Messages : in Messages_Set)
      return        Unbounded_String
   is
      Result : Unbounded_String := Empty_String;

   begin
      return Result;
   end Export_Aadl_Implementations;

end Message_Set;
