------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Messages;
USE Messages;
WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Framework_Config;
USE Framework_Config;
WITH Parameters;
USE Parameters;
USE Parameters.User_Defined_Parameters_Table_Package;
WITH Parameters.extended;
USE Parameters.extended;
WITH Offsets;
USE Offsets;
USE Offsets.Offsets_Table_package;
WITH Sets;

PACKAGE Message_Set IS





   ----------------------------------------------------------
   -- Definition of a set
   ----------------------------------------------------------



   PACKAGE Generic_Message_Set IS NEW Sets (
      Max_Element => Framework_Config.Max_Messages,
      Element     => Message_Ptr,
      Free        => Free,
      Copy        => Copy,
      Put         => Put);



   USE Generic_Message_Set;

   TYPE Messages_Set IS NEW Generic_Message_Set.Set WITH PRIVATE;


   SUBTYPE Messages_Range IS Generic_Message_Set.Element_Range;
   SUBTYPE Messages_Iterator IS   Generic_Message_Set.Iterator;



   ----------------------------------------------------------
   -- Exceptions defined in the set package
   ----------------------------------------------------------

   -- Raised when a given message is not in a message set
   --
   Message_Not_Found : EXCEPTION;

   -- Raised when 2 messages with the same name are added
   -- to a message set
   --
   Message_Already_Defined : EXCEPTION;

   -- Raised when parameters provided to Add_message are wrong
   --
   Invalid_Parameter : EXCEPTION;

   -- Raised if a message has a wrong type
   --
   Message_Model_Error : EXCEPTION;



   ----------------------------------------------------------
   -- I/O operations
   ----------------------------------------------------------
   FUNCTION Export_Xml (
         My_Messages : IN     Messages_Set)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_User_Defined_Properties (
         My_Messages : IN     Messages_Set)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_Implementations (
         My_Messages : IN     Messages_Set)
     RETURN Unbounded_String;




   ----------------------------------------------------------
   -- Procedure to proceed modification on a set
   ----------------------------------------------------------



   PROCEDURE Check_Message (
         My_Messages        : IN     Messages_Set;
         Name               : IN     Unbounded_String;
         Size               : IN     Natural;
         Period             : IN     Natural;
         Deadline           : IN     Natural;
         Jitter             : IN     Natural;
         Param              : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Response_Time      : IN     Natural;
         Communication_Time : IN     Natural);


   PROCEDURE Add_Message (
         My_Messages        : IN OUT Messages_Set;
         Name               : IN     Unbounded_String;
         Size               : IN     Natural;
         Period             : IN     Natural;
         Deadline           : IN     Natural;
         Jitter             : IN     Natural;
         Param              : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Response_Time      : IN     Natural;
         Communication_Time : IN     Natural);


   PROCEDURE Add_Message (
         My_Messages        : IN OUT Messages_Set;
         A_Message          : IN OUT Message_Ptr;
         Name               : IN     Unbounded_String;
         Size               : IN     Natural;
         Period             : IN     Natural;
         Deadline           : IN     Natural;
         Jitter             : IN     Natural;
         Param              : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Response_Time      : IN     Natural;
         Communication_Time : IN     Natural);


   PROCEDURE Update_Message (
         My_Messages        : IN OUT Messages_Set;
         Name               : IN     Unbounded_String;
         New_Name           : IN     Unbounded_String;
         Size               : IN     Natural;
         Period             : IN     Natural;
         Deadline           : IN     Natural;
         Jitter             : IN     Natural;
         Param              : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Response_Time      : IN     Natural;
         Communication_Time : IN     Natural);


   ------------------------------------------------------
   -- Read information from a set
   ------------------------------------------------------



   PROCEDURE Get_Global_Message_Type (
         My_Messages     : IN     Messages_Set;
         Has_Global_Type :    OUT Boolean;
         Global_Type     :    OUT Messages_Type);


   FUNCTION Search_Message (
         My_Messages : IN     Messages_Set;
         Name        : IN     Unbounded_String)
     RETURN Message_Ptr;





   ----------------------------------------------------------
   -- List of parameters the user can set of get directly
   -- throught the set
   ----------------------------------------------------------



   TYPE Message_Parameters IS
         (Size,
          Period,
          Deadline,
          Jitter,
          Response_Time,
          Communication_Time,
          Send_Time);


   FUNCTION Get (
         My_Messages  : IN     Messages_Set;
         Message_Name : IN     Unbounded_String;
         Param_Name   : IN     Message_Parameters)
     RETURN Natural;


   PROCEDURE Set (
         My_Messages  : IN OUT Messages_Set;
         Message_Name : IN     Unbounded_String;
         Param_Name   : IN     Message_Parameters;
         Param_Value  : IN     Natural);


   --------------------------------------------------------
   -- Functions to proceed sort on a set
   --------------------------------------------------------




   FUNCTION Increasing_Deadline (
         Op1 : IN     Message_Ptr;
         Op2 : IN     Message_Ptr)
     RETURN Boolean;

   FUNCTION Decreasing_Deadline (
         Op1 : IN     Message_Ptr;
         Op2 : IN     Message_Ptr)
     RETURN Boolean;

   FUNCTION Increasing_Period (
         Op1 : IN     Message_Ptr;
         Op2 : IN     Message_Ptr)
     RETURN Boolean;

   FUNCTION Decreasing_Period (
         Op1 : IN     Message_Ptr;
         Op2 : IN     Message_Ptr)
     RETURN Boolean;


   ----------------------------------------------------
   -- Control sub-programs
   ----------------------------------------------------



   FUNCTION Have_Deadlines_Equal_Than_Periods (
         My_Messages : IN     Messages_Set)
     RETURN Boolean;





PRIVATE

   TYPE Messages_Set IS NEW Generic_Message_Set.Set WITH NULL RECORD;




END Message_Set;
