------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Numerics;                      use Ada.Numerics;
with Ada.Numerics.Elementary_Functions; use Ada.Numerics.Elementary_Functions;
with scheduler_io;                      use scheduler_io;
with Statements;                        use Statements;
use Statements.Sets_Type_Package;
with Expressions;                       use Expressions;
use Expressions.Variables_Type_Package;
with Processors.extended;               use Processors.extended;
with Tasks;                             use Tasks;
with Ada.Strings.Unbounded.Text_IO;     use Ada.Strings.Unbounded.Text_IO;
with Unbounded_Strings;                 use Unbounded_Strings;
with Scheduler;                         use Scheduler;
with Dependency_Services;               use Dependency_Services;
with Translate;                         use Translate;
with Io_Tools;                          use Io_Tools;
with Ada.IO_Exceptions;                 use Ada.IO_Exceptions;
with Ada.Exceptions;                    use Ada.Exceptions;
with Xml_Parsers;                       use Xml_Parsers;
with Input_Sources.File;                use Input_Sources.File;
with Sax.Readers;                       use Sax.Readers;
with Text_IO;                           use Text_IO;
with Parser;                            use Parser;

package body Multiprocessor_Services is

   procedure Put (R : in Scheduling_Result) is
   begin
      Put (To_Unbounded_String ("Error Message : ") & R.Error_Msg);
      New_Line;
      Put (To_Unbounded_String ("Scheduling Message : ") & R.Scheduling_Msg);
      New_Line;

      Put (R.Result.all);
   end Put;

   procedure Initialize (R : out Scheduling_Result) is
   begin
      R.Scheduling_Msg := Empty_String;
      R.Error_Msg      := Empty_String;
      R.Result         := new Scheduling_Sequence;
      Initialize (R.Result.all);
   end Initialize;

   procedure Build_Multiprocessor_Scheduling
     (My_Processors           : in Processors_Set;
      My_Address_Spaces       : in Address_Spaces_Set;
      My_Tasks                : in Tasks_Set;
      My_Resources            : in Resources_Set;
      My_Buffers              : in Buffers_Set;
      My_Messages             : in Messages_Set;
      My_Dependencies         : in Tasks_Dependencies_Ptr;
      Result                  : in out Scheduling_Table_Ptr;
      Event_To_Generate       : in Time_Unit_Event_Type_Boolean_Table;
      Last_Time               : in Natural;
      With_Offsets            : in Boolean := True;
      With_Precedencies       : in Boolean := True;
      With_Resources          : in Boolean := True;
      With_Task_Specific_Seed : in Boolean := True;
      Global_Seed_Value       : in Integer := 0;
      Predictable_Global_Seed : in Boolean := True)
   is

      Tmp_Buffers   : array (Scheduling_Table_Range) of Buffers_Set;
      Tmp_Tasks     : array (Scheduling_Table_Range) of Tasks_Set;
      Tmp_Resources : array (Scheduling_Table_Range) of Resources_Set;

      Si : Scheduling_Information;

      Iterator3 : Processors_Iterator;

      A_Processor : Processor_Ptr;

      Last_Current_Time : Natural := 0;

      Current_Time : array (Scheduling_Table_Range) of Natural :=
        (others => 0);
      Elected      : Tasks_Range                               :=
         Tasks_Range'First;

      No_Task : Boolean := True;

   begin

      Si.Number_Of_Tasks          := 0;
      Si.Number_Of_Resources      := 0;
      Si.Number_Of_Processors     :=
        Integer (Get_Number_Of_Elements (My_Processors));
      Si.Number_Of_Address_Spaces :=
        Integer (Get_Number_Of_Elements (My_Address_Spaces));
      Si.Simulation_Length        := Last_Time;

      -- Before scheduling, check if the tasks can be
      -- schedule with the selected scheduler
      --
      Reset_Iterator (My_Processors, Iterator3);

      loop
         Current_Element (My_Processors, A_Processor, Iterator3);

         begin

            Result.Entries (Result.Nb_Entries).Item                :=
              A_Processor;
            Result.Entries (Result.Nb_Entries).Data.Scheduling_Msg :=
              Empty_String;
            Result.Entries (Result.Nb_Entries).Data.Error_Msg      :=
              Empty_String;

            if (Get_Number_Of_Task_From_Processor
                   (My_Tasks,
                    A_Processor.Name) /=
                0)
            then

               Check_Before_Scheduling
                 (A_Processor.Scheduler.all,
                  My_Tasks,
                  A_Processor.Name);

               if With_Precedencies then
                  Dependencies_Periods_Control
                    (My_Tasks,
                     My_Dependencies,
                     A_Processor.Name);
               end if;

            else
               Result.Entries (Result.Nb_Entries).Data.Error_Msg := Lb_Minus &
                                                                    Lb_Define_Tasks_Before
(Current_Language);

            end if;

         exception
            when Dependency_Services.Dependencies_Period_Error =>
               Result.Entries (Result.Nb_Entries).Data.Error_Msg :=
                 Lb_Precedencies_Period_Error (Current_Language) &
                 Unbounded_Lf;
            when Task_Set.Task_Must_Be_Periodic =>
               Result.Entries (Result.Nb_Entries).Data.Error_Msg :=
                 Lb_Compute_Scheduling_Error_15 (Current_Language) &
                 Unbounded_Lf;
            when Task_Set.Task_Must_Have_Period_Equal_To_Deadline =>
               Result.Entries (Result.Nb_Entries).Data.Error_Msg :=
                 Lb_Compute_Scheduling_Error_14 (Current_Language) &
                 Unbounded_Lf;
            when Task_Set.Task_Model_Error =>
               Result.Entries (Result.Nb_Entries).Data.Error_Msg :=
                 Lb_Compute_Scheduling_Error_1 (Current_Language) &
                 Unbounded_Lf;
            when Task_Set.Priority_Error =>
               Result.Entries (Result.Nb_Entries).Data.Error_Msg :=
                 Lb_Compute_Scheduling_Error_11 (Current_Language) &
                 Unbounded_Lf;

         end;

         Result.Nb_Entries := Result.Nb_Entries + 1;

         exit when Is_Last_Element (My_Processors, Iterator3);
         Next_Element (My_Processors, Iterator3);

      end loop;

      -- First : gather information
      --
      for J in  0 .. Result.Nb_Entries - 1 loop

         -- Duplicate data before scheduling (if case of data are changed)
         --
         if Result.Entries (J).Data.Error_Msg = Empty_String then

            Current_Processor_Name := Result.Entries (J).Item.Name;

            Select_And_Copy (My_Tasks, Tmp_Tasks (J), Select_Cpu'Access);
            Select_And_Copy
              (My_Resources,
               Tmp_Resources (J),
               Select_Cpu'Access);
            Select_And_Copy (My_Buffers, Tmp_Buffers (J), Select_Cpu'Access);

            Scheduler_Initialization
              (Result.Entries (J).Item.Scheduler.all,
               Si,
               Result.Entries (J).Item.Name,
               Tmp_Tasks (J),
               Tmp_Resources (J),
               Tmp_Buffers (J),
               Result.Entries (J).Data.Result,
               With_Offsets,
               With_Precedencies,
               With_Resources,
               With_Task_Specific_Seed,
               Global_Seed_Value,
               Predictable_Global_Seed,
               Last_Time,
               Event_To_Generate);

            Specific_Scheduler_Initialization
              (Result.Entries (J).Item.Scheduler.all,
               Si,
               Result.Entries (J).Item.Name,
               Tmp_Tasks (J),
               Tmp_Resources (J),
               Tmp_Buffers (J),
               My_Messages,
               Result.Entries (J).Data.Scheduling_Msg);

         end if;

      end loop;

      -- Second : compute scheduling
      --

      while (Last_Current_Time < Last_Time) loop

         Last_Current_Time := Natural'Last;

         for J in  0 .. Result.Nb_Entries - 1 loop

            if (Result.Entries (J).Data.Error_Msg = Empty_String) then

               if (Current_Time (J) < Last_Time) then

                  Do_Election
                    (Result.Entries (J).Item.Scheduler.all,
                     Si,
                     Result.Entries (J).Data.Result,
                     Result.Entries (J).Data.Scheduling_Msg,
                     Current_Time (J),
                     Result.Entries (J).Item.Name,
                     My_Dependencies,
                     With_Offsets,
                     With_Precedencies,
                     With_Resources,
                     Event_To_Generate,
                     Elected,
                     No_Task);

                  -- No task to schedule
                  --
                  if No_Task then
                     Current_Time (J) := Current_Time (J) + 1;

                  else
                     Run_Task
                       (Result.Entries (J).Item.Scheduler.all,
                        Result.Entries (J).Item.Name,
                        Si,
                        My_Dependencies,
                        Elected,
                        Result.Entries (J).Data.Result,
                        Current_Time (J),
                        Last_Time,
                        With_Offsets,
                        With_Precedencies,
                        With_Resources,
                        Event_To_Generate);

                  end if;

                  Last_Current_Time :=
                     Natural'Min (Last_Current_Time, Current_Time (J));

               end if;
            end if;

         end loop;

      end loop;

   end Build_Multiprocessor_Scheduling;

   procedure Put (Sched : in Scheduling_Table_Ptr) is

      Global_Sched :
        array (Scheduling_Table_Range) of Scheduling_Sequence_Ptr;

      Global_Sched_Index : array (Scheduling_Table_Range) of Time_Unit_Range
         := (others => 0);

      Global_Sched_Number        : Scheduling_Table_Range := 0;
      Min_Time                   : Natural                := 0;
      Current_Min_Time_Processor : Scheduling_Table_Range;

   begin

      New_Line;
      New_Line;
      Put_Line ("Event Table after scheduling : ");
      New_Line;
      Put_Line ("Event type/Task, buffer, message or resource name/Date");
      New_Line;
      New_Line;

      -- how many processors scheduling should we run analysis ?
      --
      for I in  0 .. Sched.Nb_Entries - 1 loop
         if (Sched.Entries (I).Data.Error_Msg = Empty_String) then
            Global_Sched (Global_Sched_Number) :=
              Sched.Entries (I).Data.Result;
            Global_Sched_Number                := Global_Sched_Number + 1;
         end if;
      end loop;

      -- Now, scan each time unit
      --
      while Min_Time /= Natural'Last loop

         -- find the smallest time unit
         --
         Min_Time := Natural'Last;
         for I in  0 .. Global_Sched_Number - 1 loop
            if Global_Sched_Index (I) < Global_Sched (I).Nb_Entries then
               if Global_Sched (I).Entries (Global_Sched_Index (I)).Data <
                  Min_Time
               then
                  Min_Time                   :=
                    Global_Sched (I).Entries (Global_Sched_Index (I)).Data;
                  Current_Min_Time_Processor := I;
               end if;
            end if;
         end loop;

         if Min_Time /= Natural'Last then

            Put
              (Global_Sched (Current_Min_Time_Processor).Entries (
              Global_Sched_Index (Current_Min_Time_Processor)));

            Global_Sched_Index (Current_Min_Time_Processor) :=
              Global_Sched_Index (Current_Min_Time_Processor) + 1;

         end if;

      end loop;

      New_Line;
      New_Line;

   end Put;

   procedure Run_An_Event_Analyzer
     (My_Tasks          : in Tasks_Set;
      My_Processors     : in Processors_Set;
      My_Address_Spaces : in Address_Spaces_Set;
      My_Resources      : in Resources_Set;
      My_Buffers        : in Buffers_Set;
      My_Messages       : in Messages_Set;
      Sched             : in Scheduling_Table_Ptr;
      An_Event_Analyzer : in Event_Analyzer_Ptr;
      Result            : in out Unbounded_String)
   is

      Current : Statement_Ptr;

      Var, Var_Time, Var_Type, Var_Processor, Var_Message, Var_Resource, 
Var_Task, Var_Buffer : Variables_Range;

      Global_Sched :
        array (Scheduling_Table_Range) of Scheduling_Sequence_Ptr;

      Global_Sched_Index    :
        array (Scheduling_Table_Range) of Time_Unit_Range := (others => 0);
      Global_Processor_Name :
        array (Scheduling_Table_Range) of Unbounded_String;

      Global_Sched_Number        : Scheduling_Table_Range := 0;
      Min_Time                   : Natural                := 0;
      Current_Min_Time_Processor : Scheduling_Table_Range;

   begin

      begin
         An_Event_Analyzer.String_Behavior :=
            Read_Sequential_File (An_Event_Analyzer.Behavior_File_Name);

      exception
         when Ada.IO_Exceptions.Name_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   An_Event_Analyzer.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Name_Error");
         when Ada.IO_Exceptions.Status_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   An_Event_Analyzer.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Status_Error");
         when Ada.IO_Exceptions.Mode_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   An_Event_Analyzer.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Mode_Error");
         when Ada.IO_Exceptions.Use_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   An_Event_Analyzer.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Use_Error");
         when Ada.IO_Exceptions.Device_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   An_Event_Analyzer.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Device_Error");
         when Ada.IO_Exceptions.End_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   An_Event_Analyzer.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", End_Error");
         when Ada.IO_Exceptions.Data_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   An_Event_Analyzer.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Data_Error");
         when Ada.IO_Exceptions.Layout_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   An_Event_Analyzer.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Layout_Error");
         when others =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   An_Event_Analyzer.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)));
      end;

      -- We should check the scheduler behavior syntax ...
      --
      Open_Input (An_Event_Analyzer.String_Behavior);
      Parser.Init_Parser;
      Create_Parametric_Variables (Parser.Variables_Table, My_Tasks);
      Parser.Yyparse;

      -- Store syntax tree and compiling information
      --
      An_Event_Analyzer.Root_Statement_Pointer :=
        Parser.Root_Statement_Pointer;

      An_Event_Analyzer.Variables_Table := Parser.Variables_Table;

      An_Event_Analyzer.Sets_Table := Parser.Sets_Table;

      -- Initialize parametric variables which are constant
      --
      Initialize_Parametric_Variables
        (An_Event_Analyzer.Variables_Table,
         My_Messages,
         My_Buffers,
         My_Resources,
         My_Tasks);

      -- "nb_processors"
      --
      Var :=
         Find_Variable
           (An_Event_Analyzer.Variables_Table,
            To_Unbounded_String ("nb_processors"));
      An_Event_Analyzer.Variables_Table.Entries (Var).Simulation.Integer_Value
         := Integer (Get_Number_Of_Elements (My_Processors));

      -- "nb_address_spaces"
      --
      Var :=
         Find_Variable
           (An_Event_Analyzer.Variables_Table,
            To_Unbounded_String ("nb_address_spaces"));
      An_Event_Analyzer.Variables_Table.Entries (Var).Simulation.Integer_Value
         := Integer (Get_Number_Of_Elements (My_Address_Spaces));

      -- Now : run "start" statements
      --
      Current := An_Event_Analyzer.Root_Statement_Pointer (Statements.Start);
      Dispatch (Current, Start, An_Event_Analyzer.Variables_Table, Result);

      -- Now : run "event analyzer" statements
      --
      Current :=
        An_Event_Analyzer.Root_Statement_Pointer (
        Statements.Gather_Event_Analyzer);

      -- Find variables before running the simulation
      --
      Var_Time      :=
         Find_Variable
           (An_Event_Analyzer.Variables_Table,
            To_Unbounded_String ("events.time"));
      Var_Type      :=
         Find_Variable
           (An_Event_Analyzer.Variables_Table,
            To_Unbounded_String ("events.type"));
      Var_Processor :=
         Find_Variable
           (An_Event_Analyzer.Variables_Table,
            To_Unbounded_String ("events.processor_name"));
      Var_Message   :=
         Find_Variable
           (An_Event_Analyzer.Variables_Table,
            To_Unbounded_String ("events.message_name"));
      Var_Resource  :=
         Find_Variable
           (An_Event_Analyzer.Variables_Table,
            To_Unbounded_String ("events.resource_name"));
      Var_Task      :=
         Find_Variable
           (An_Event_Analyzer.Variables_Table,
            To_Unbounded_String ("events.task_name"));
      Var_Buffer    :=
         Find_Variable
           (An_Event_Analyzer.Variables_Table,
            To_Unbounded_String ("events.buffer_name"));

      -- how many processors scheduling should we run analysis ?
      --
      for I in  0 .. Sched.Nb_Entries - 1 loop
         if (Sched.Entries (I).Data.Error_Msg = Empty_String) then
            Global_Sched (Global_Sched_Number)          :=
              Sched.Entries (I).Data.Result;
            Global_Processor_Name (Global_Sched_Number) :=
              Sched.Entries (I).Item.Name;
            Global_Sched_Number                         :=
              Global_Sched_Number + 1;
         end if;
      end loop;

      -- Now, scan each time unit
      --
      while Min_Time /= Natural'Last loop

         -- find the smallest time unit
         --
         Min_Time := Natural'Last;
         for I in  0 .. Global_Sched_Number - 1 loop
            if Global_Sched_Index (I) < Global_Sched (I).Nb_Entries then
               if Global_Sched (I).Entries (Global_Sched_Index (I)).Data <
                  Min_Time
               then
                  Min_Time                   :=
                    Global_Sched (I).Entries (Global_Sched_Index (I)).Data;
                  Current_Min_Time_Processor := I;
               end if;
            end if;
         end loop;

         if Min_Time /= Natural'Last then

            -- "events.time"
            --
            An_Event_Analyzer.Variables_Table.Entries (Var_Time).Simulation.
              Integer_Value :=
              Integer (
              Global_Sched (Current_Min_Time_Processor).Entries (
              Global_Sched_Index (Current_Min_Time_Processor)).Data);

            -- "events.type"
            --
            An_Event_Analyzer.Variables_Table.Entries (Var_Type).Simulation.
              String_Value :=
               To_Lower
                 (
              Global_Sched (Current_Min_Time_Processor).Entries (
              Global_Sched_Index (Current_Min_Time_Processor)).Item.Event_Type'
              Img);

            -- "events.processor_name"
            --
            An_Event_Analyzer.Variables_Table.Entries (Var_Processor).
              Simulation.String_Value :=
              Global_Processor_Name (Current_Min_Time_Processor);

            -- "events.buffer_name"
            --
            An_Event_Analyzer.Variables_Table.Entries (Var_Buffer).Simulation.
              String_Value := Empty_String;

            -- "events.resource_name"
            --
            An_Event_Analyzer.Variables_Table.Entries (Var_Resource).
              Simulation.String_Value := Empty_String;

            -- "events.message_name"
            --
            An_Event_Analyzer.Variables_Table.Entries (Var_Message).Simulation.
              String_Value := Empty_String;

            case 
              Global_Sched (Current_Min_Time_Processor).Entries (
              Global_Sched_Index (Current_Min_Time_Processor)).Item.Event_Type
is

            when Start_Of_Task_Capacity =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Start_Task.Name;

            when End_Of_Task_Capacity =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 End_Task.Name;

            when Write_To_Buffer =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Write_Task.Name;
               An_Event_Analyzer.Variables_Table.Entries (Var_Buffer).
                 Simulation.String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Write_Buffer.Name;

            when Read_From_Buffer =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Read_Task.Name;
               An_Event_Analyzer.Variables_Table.Entries (Var_Buffer).
                 Simulation.String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Read_Buffer.Name;

            when Running_Task =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Running_Task.Name;

            when Task_Activation =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Activation_Task.Name;

            when Allocate_Resource =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Allocate_Task.Name;
               An_Event_Analyzer.Variables_Table.Entries (Var_Resource).
                 Simulation.String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Allocate_Resource.Name;

            when Release_Resource =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Release_Task.Name;
               An_Event_Analyzer.Variables_Table.Entries (Var_Resource).
                 Simulation.String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Release_Resource.Name;

            when Wait_For_Resource =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Wait_Task.Name;
               An_Event_Analyzer.Variables_Table.Entries (Var_Resource).
                 Simulation.String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Wait_Resource.Name;

            when Send_Message =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Send_Task.Name;
               An_Event_Analyzer.Variables_Table.Entries (Var_Message).
                 Simulation.String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Send_Message.Name;

            when Receive_Message =>
               An_Event_Analyzer.Variables_Table.Entries (Var_Task).Simulation.
                 String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Receive_Task.Name;
               An_Event_Analyzer.Variables_Table.Entries (Var_Message).
                 Simulation.String_Value :=
                 Global_Sched (Current_Min_Time_Processor).Entries (
                 Global_Sched_Index (Current_Min_Time_Processor)).Item.
                 Receive_Message.Name;

            end case;

            -- Do analysis on the current time : run "event_analyzer"
            --section/statements
            --
            Dispatch
              (Current,
               Statements.Gather_Event_Analyzer,
               An_Event_Analyzer.Variables_Table,
               Result);

            Global_Sched_Index (Current_Min_Time_Processor) :=
              Global_Sched_Index (Current_Min_Time_Processor) + 1;

         end if;

      end loop;

      -- The end of the line : run "display_event" section/statements
      --
      Current :=
        An_Event_Analyzer.Root_Statement_Pointer (
        Statements.Display_Event_Analyzer);
      Dispatch
        (Current,
         Display_Event_Analyzer,
         An_Event_Analyzer.Variables_Table,
         Result);

   end Run_An_Event_Analyzer;

   procedure Free (Sched : in out Scheduling_Table_Ptr) is

      procedure Free_Pointer is new Unchecked_Deallocation (
         Scheduling_Table,
         Scheduling_Table_Ptr);

   begin
      for I in  Scheduling_Table_Range loop
         if Sched.Entries (I).Data.Result /= null then
            Free (Sched.Entries (I).Data.Result);
         end if;
      end loop;
      Free_Pointer (Sched);
   end Free;

   procedure Read_From_Xml_File
     (Sched     : in out Scheduling_Table_Ptr;
      Sys       : in System;
      File_Name : in String)
   is
   begin
      Read_From_Xml_File (Sched, Sys, To_Unbounded_String (File_Name));
   end Read_From_Xml_File;

   procedure Read_From_Xml_File
     (Sched     : in out Scheduling_Table_Ptr;
      Sys       : in System;
      File_Name : in Unbounded_String)
   is

      Read       : File_Input;
      My_Reader  : Xml_Event_Table_Parser;
      Name_Start : Natural;
      S          : constant String := To_String (File_Name);

   begin

      --  Base file name should be used as the public Id
      --
      Name_Start := S'Last;
      while Name_Start >= S'First and then S (Name_Start) /= '/' loop
         Name_Start := Name_Start - 1;
      end loop;
      Set_Public_Id (Read, S (Name_Start + 1 .. S'Last));
      Set_System_Id (Read, S);
      Open (S, Read);

      --  xmlns:* attributes will be reported in Start_Element
      --
      Set_Feature (My_Reader, Namespace_Prefixes_Feature, False);
      Set_Feature (My_Reader, Validation_Feature, True);

      Set_Scheduled_System (My_Reader, Sys);
      Parse (My_Reader, Read);
      Free (Sched);
      Sched := Get_Parsed_Event_Table (My_Reader);
      Close (Read);

   exception
      when XML_Fatal_Error =>
         Close (Read);
         raise Xml_Read_Error;

   end Read_From_Xml_File;

   procedure Write_To_Xml_File
     (Sched     : in Scheduling_Table_Ptr;
      Sys       : in System;
      File_Name : in String)
   is
   begin
      Write_To_Xml_File (Sched, Sys, To_Unbounded_String (File_Name));
   end Write_To_Xml_File;

   procedure Write_To_Xml_File
     (Sched     : in Scheduling_Table_Ptr;
      Sys       : in System;
      File_Name : in Unbounded_String)
   is

      Into         : File_Type;
      An_Event     : Time_Unit_Event_Ptr;
      My_Scheduler : Scheduler_Ptr;
      A_Processor  : Processor_Ptr;

   begin

      -- Open file and Write DTD/XML Header
      --
      Create (Into, Mode => Out_File, Name => To_String (File_Name));

      Put_Line (Into, "<?xml version=""1.0"" standalone=""yes""?>  ");
      Put_Line
        (Into,
         "<?xml-stylesheet type=""text/xsl"" href=""event_table.xsl""?>");
      Put_Line (Into, "<!DOCTYPE event_table [  ");
      for E in  Time_Unit_Event_Type'Range loop
         Put_Line (Into, "<!ELEMENT " & To_Lower (E'Img) & " (#PCDATA) >  ");
      end loop;
      Put_Line (Into, "<!ELEMENT name (#PCDATA) >  ");
      Put_Line (Into, "<!ELEMENT processor (#PCDATA) >  ");

      Put (Into, "<!ELEMENT processor (name");
      for E in  Time_Unit_Event_Type'Range loop
         Put (Into, "|(" & To_Lower (E'Img) & ")?");
      end loop;
      Put (Into, " ) >  ");

      Put_Line (Into, "] >  ");
      New_Line (Into, 2);

      Put_Line (Into, "<event_table>");

      -- Export event table for each processor
      --
      for I in  0 .. Sched.Nb_Entries - 1 loop
         if (Sched.Entries (I).Data.Error_Msg = Empty_String) then

            Put_Line (Into, "   <processor> ");
            Put_Line
              (Into,
               "     <name>" &
               To_String (Sched.Entries (I).Item.Name) &
               " </name>");

            A_Processor  :=
               Search_Processor (Sys.Processors, Sched.Entries (I).Item.Name);
            My_Scheduler :=
               Get_Scheduler (extended_Processor (A_Processor.all));

            for J in  0 .. Sched.Entries (I).Data.Result.Nb_Entries - 1 loop

               An_Event := Sched.Entries (I).Data.Result.Entries (J).Item;
               Put
                 (Into,
                  "      <" & To_Lower (An_Event.Event_Type'Img) & "> ");
               Put
                 (Into,
                  " " &
                  Sched.Entries (I).Data.Result.Entries (J).Data'Img &
                  " ");

               case An_Event.Event_Type is
               when Start_Of_Task_Capacity =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Start_Of_Task_Capacity
                            (My_Scheduler.all,
                             An_Event)));

               when End_Of_Task_Capacity =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_End_Of_Task_Capacity
                            (My_Scheduler.all,
                             An_Event)));

               when Write_To_Buffer =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Write_To_Buffer
                            (My_Scheduler.all,
                             An_Event)));

               when Read_From_Buffer =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Read_From_Buffer
                            (My_Scheduler.all,
                             An_Event)));

               when Running_Task =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Running_Task
                            (My_Scheduler.all,
                             An_Event)));

               when Task_Activation =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Task_Activation
                            (My_Scheduler.all,
                             An_Event)));

               when Send_Message =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Send_Message
                            (My_Scheduler.all,
                             An_Event)));

               when Receive_Message =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Receive_Message
                            (My_Scheduler.all,
                             An_Event)));

               when Allocate_Resource =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Allocate_Resource
                            (My_Scheduler.all,
                             An_Event)));

               when Release_Resource =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Release_Resource
                            (My_Scheduler.all,
                             An_Event)));

               when Wait_For_Resource =>
                  Put
                    (Into,
                     To_String
                        (Export_Xml_Event_Wait_For_Resource
                            (My_Scheduler.all,
                             An_Event)));

               end case;

               Put (Into, "</" & To_Lower (An_Event.Event_Type'Img) & "> ");
               New_Line (Into);

            end loop;
            Put_Line (Into, "   </processor>");
            New_Line (Into);

         end if;
      end loop;

      Put_Line (Into, "</event_table>");
      New_Line (Into);
      New_Line (Into);

      Close (Into);

   end Write_To_Xml_File;

   procedure Partition_Small_Task
     (Src_Processors   : in Processors_Set;
      Src_Tasks        : in Tasks_Set;
      Msg              : in out Unbounded_String;
      Result_Tasks     : in out Tasks_Set;
      Processor_Result : out Processors_Iterator)
   is

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;

      Processor_Util, Beta : Float   := 0.0;
      S_Cur                : Float   := 1.0;
      First_Run            : Boolean := True;

   begin

      Validate_Multiprocessor_Tasks (Src_Processors, Src_Tasks);

      Duplicate (Src_Tasks, Result_Tasks);
      Sort (Result_Tasks, Increasing_Si'Access);

      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);
      Reset_Iterator (Src_Processors, My_Processor_Iterator);
      Current_Element (Src_Processors, A_Processor, My_Processor_Iterator);

      loop
         if First_Run = False then
            if Is_Last_Element (Src_Processors, My_Processor_Iterator) then
               Duplicate (Src_Tasks, Result_Tasks);
               raise No_Such_Processors;
            end if;
            Next_Element (Src_Processors, My_Processor_Iterator);
            Current_Element
              (Src_Processors,
               A_Processor,
               My_Processor_Iterator);
         end if;
         First_Run := False;
         if Float (A_Task_Result.Capacity) / Float (A_Task_Result.Period) >
            1.0
         then
            Duplicate (Src_Tasks, Result_Tasks);
            raise No_Such_Processors;
         end if;
         A_Task_Result.Cpu_Name := A_Processor.Name;
         Processor_Util         := Float (A_Task_Result.Capacity) /
                                   Float (A_Task_Result.Period);
         S_Cur                  :=
           Float (Log (Float (A_Task_Result.Period), 2.0)) -
           Float'Floor (Float (Log (Float (A_Task_Result.Period), 2.0)));
         Beta                   := 0.0;
         exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);

         loop
            if Is_Last_Element (Result_Tasks, My_Task_Iterator_Result) then
               First_Run := True;
               exit;
            end if;
            Next_Element (Result_Tasks, My_Task_Iterator_Result);
            Current_Element
              (Result_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);
            Beta           :=
              (Float (Log (Float (A_Task_Result.Period), 2.0)) -
               Float'Floor
                  (Float (Log (Float (A_Task_Result.Period), 2.0)))) -
              S_Cur;
            Processor_Util := Processor_Util +
                              Float (A_Task_Result.Capacity) /
                              Float (A_Task_Result.Period);
            if (Processor_Util <=
                Float'Max
                   (Float (Log (2.0, e)),
                    1.0 - Beta * Float (Log (2.0, e))))
            then
               A_Task_Result.Cpu_Name := A_Processor.Name;
            else
               exit;
            end if;
         end loop;

      end loop;
      Processor_Result := My_Processor_Iterator;
      Sort (Result_Tasks, Increasing_Name'Access);

   end Partition_Small_Task;

   procedure Partition_Small_Task
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set)
   is
      Tmp : Processors_Iterator;
   begin

      Msg := To_Unbounded_String (" (") &
             Lb_See (Current_Language) &
             To_Unbounded_String ("[9], [10]) ");

      Partition_Small_Task
        (Src_Processors,
         Src_Tasks,
         Msg,
         Result_Tasks,
         Tmp);
   end Partition_Small_Task;

   procedure Partition_Small_Task
     (Src_Processors   : in Processors_Set;
      Result_Tasks     : in out Tasks_Set;
      Processor_Result : out Processors_Iterator)
   is
      Src_Tasks : Tasks_Set;
      Msg       : Unbounded_String;
   begin
      Duplicate (Result_Tasks, Src_Tasks);
      Partition_Small_Task
        (Src_Processors,
         Src_Tasks,
         Msg,
         Result_Tasks,
         Processor_Result);
   end Partition_Small_Task;

   procedure Rm_General_Tasks
     (Src_Processors : in Processors_Set;
      Result_Tasks   : in out Tasks_Set)
   is

      A_Task_Result           : Periodic_Task_Ptr;
      A_Task_Tmp              : Periodic_Task_Ptr;
      A_Task_Iterator         : Tasks_Iterator;
      My_Task_Iterator_Result : Tasks_Iterator;
      Tmp_Iterator            : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;
      Src_Tasks               : Tasks_Set;

      Num_Of_Tasks                        : Integer := 0;
      Task_Utilization, Total_Utilization : Float   := 0.0;
      Goto_First_Loop                     : Boolean := False;

   begin
      Duplicate (Result_Tasks, Src_Tasks);
      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);
      Reset_Iterator (Src_Processors, My_Processor_Iterator);
      Current_Element (Src_Processors, A_Processor, My_Processor_Iterator);

      Sort (Result_Tasks, Increasing_Period'Access);

      -- Zero processor locations
      --
      loop
         A_Task_Result.Cpu_Name := To_Unbounded_String ("");
         exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
         Next_Element (Result_Tasks, My_Task_Iterator_Result);
         Current_Element
           (Result_Tasks,
            Generic_Task_Ptr (A_Task_Result),
            My_Task_Iterator_Result);
      end loop;

      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);
      loop

         loop
            if Goto_First_Loop then
               Goto_First_Loop := False;
               exit;
            end if;
            -- Fetch information for condition-IP
            --
            Tmp_Iterator := My_Task_Iterator_Result;
            Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
            Current_Element
              (Result_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);
            Num_Of_Tasks      := 0;
            Total_Utilization := 0.0;
            loop
               if A_Task_Result.Cpu_Name = A_Processor.Name then
                  Num_Of_Tasks      := Num_Of_Tasks + 1;
                  Total_Utilization := Total_Utilization +
                                       Float (A_Task_Result.Capacity) /
                                       Float (A_Task_Result.Period);
               end if;
               exit when Is_Last_Element
                            (Result_Tasks,
                             My_Task_Iterator_Result);
               Next_Element (Result_Tasks, My_Task_Iterator_Result);
               Current_Element
                 (Result_Tasks,
                  Generic_Task_Ptr (A_Task_Result),
                  My_Task_Iterator_Result);
            end loop;
            My_Task_Iterator_Result := Tmp_Iterator;
            Current_Element
              (Result_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);
            Task_Utilization := Float (A_Task_Result.Capacity) /
                                Float (A_Task_Result.Period);

            if Task_Utilization <=
               2.0 *
               (1.0 /
                ((1.0 + Total_Utilization / Float (Num_Of_Tasks))) **
                Num_Of_Tasks) -
               1.0
            then
               if Total_Utilization = 0.0 then
                  A_Task_Result.Cpu_Name := A_Processor.Name;
                  exit;
               end if;
               Reset_Iterator (Result_Tasks, A_Task_Iterator);
               Current_Element
                 (Result_Tasks,
                  Generic_Task_Ptr (A_Task_Tmp),
                  A_Task_Iterator);

               loop
                  if A_Task_Tmp.Cpu_Name = A_Processor.Name and
                     A_Task_Tmp /= A_Task_Result
                  then

                     if A_Task_Result.Period < A_Task_Tmp.Period and
                        (Integer (Float'Floor
                                     (Float (A_Task_Tmp.Period) /
                                      Float (A_Task_Result.Period))) *
                         (A_Task_Tmp.Period - A_Task_Result.Capacity) >=
                         A_Task_Tmp.Capacity or
                         A_Task_Tmp.Period >=
                         Integer (Float'Ceiling
                                     (Float (A_Task_Tmp.Period) /
                                      Float (A_Task_Result.Period))) *
                         A_Task_Result.Capacity +
                         A_Task_Tmp.Capacity)
                     then
                        A_Task_Result.Cpu_Name := A_Processor.Name;
                        Reset_Iterator
                          (Src_Processors,
                           My_Processor_Iterator);
                        Goto_First_Loop := True;
                        exit;
                     else
                        if A_Task_Result.Period >= A_Task_Tmp.Period and
                           (((Integer (Float'Floor
                                          (Float (A_Task_Result.Period) /
                                           Float (A_Task_Tmp.Period))) *
                              (A_Task_Result.Period - A_Task_Tmp.Capacity)) >=
                             A_Task_Result.Capacity) or
                            (A_Task_Result.Period >=
                             (Integer (Float'Ceiling
                                          (Float (A_Task_Result.Period) /
                                           Float (A_Task_Tmp.Period))) *
                              A_Task_Tmp.Capacity +
                              A_Task_Result.Capacity)))
                        then
                           A_Task_Result.Cpu_Name := A_Processor.Name;
                           Reset_Iterator
                             (Src_Processors,
                              My_Processor_Iterator);
                           Goto_First_Loop := True;
                           exit;
                        else
                           if Is_Last_Element
                                (Src_Processors,
                                 My_Processor_Iterator)
                           then
                              Duplicate (Src_Tasks, Result_Tasks);
                              raise No_Such_Processors;
                           end if;
                           Next_Element
                             (Src_Processors,
                              My_Processor_Iterator);
                           Current_Element
                             (Src_Processors,
                              A_Processor,
                              My_Processor_Iterator);
                           exit;
                        end if;
                     end if;
                  end if;
                  exit when Is_Last_Element (Result_Tasks, A_Task_Iterator);
                  Next_Element (Result_Tasks, A_Task_Iterator);
                  Current_Element
                    (Result_Tasks,
                     Generic_Task_Ptr (A_Task_Tmp),
                     A_Task_Iterator);
               end loop;

            else
               if Is_Last_Element
                    (Src_Processors,
                     My_Processor_Iterator)
               then
                  Duplicate (Src_Tasks, Result_Tasks);
                  raise No_Such_Processors;
               end if;
               Next_Element (Src_Processors, My_Processor_Iterator);
               Current_Element
                 (Src_Processors,
                  A_Processor,
                  My_Processor_Iterator);
            end if;

         end loop;

         exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
         Next_Element (Result_Tasks, My_Task_Iterator_Result);
      end loop;

   end Rm_General_Tasks;

   procedure Partition_General_Task
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set)
   is

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      My_Processor_Iterator   : Processors_Iterator;
      A_Processor             : Processor_Ptr;

      Rmst_Tasks : Tasks_Set;
      Rmgt_Tasks : Tasks_Set;
      Tmp_Cpus   : Processors_Set;

      Rmst_Index, Rmgt, Num_Of_Tasks : Integer := 0;

   begin

      Validate_Multiprocessor_Tasks (Src_Processors, Src_Tasks);

      Msg := To_Unbounded_String (" (") &
             Lb_See (Current_Language) &
             To_Unbounded_String ("[9], [10]) ");

      Duplicate (Src_Tasks, Result_Tasks);
      Sort (Result_Tasks, Increasing_Utilization'Access);

      -- Scan for max Alfa to make division between RMST and RMGT tasks
      --
      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);
      loop
         if Float (A_Task_Result.Capacity) / Float (A_Task_Result.Period) >
            1.0 / 3.0
         then
            exit;
         else
            Rmgt := Rmgt + 1;
         end if;
         exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
         Next_Element (Result_Tasks, My_Task_Iterator_Result);
         Current_Element
           (Result_Tasks,
            Generic_Task_Ptr (A_Task_Result),
            My_Task_Iterator_Result);
      end loop;

      -- Make bin-packing for RMST and RMGT
      --
      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);
      loop
         exit when Rmst_Index = Rmgt;
         Add (Rmst_Tasks, Generic_Task_Ptr (A_Task_Result));
         Rmst_Index := Rmst_Index + 1;
         Next_Element (Result_Tasks, My_Task_Iterator_Result);
         Current_Element
           (Result_Tasks,
            Generic_Task_Ptr (A_Task_Result),
            My_Task_Iterator_Result);
      end loop;

      Num_Of_Tasks := Integer (Get_Number_Of_Elements (Result_Tasks));
      if Rmgt /= Num_Of_Tasks then
         loop
            Current_Element
              (Result_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);
            Add (Rmgt_Tasks, Generic_Task_Ptr (A_Task_Result));
            exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
            Next_Element (Result_Tasks, My_Task_Iterator_Result);
         end loop;
      end if;

      -- Call RMST and RMGT functions respectively
      --
      if (Is_Empty (Rmst_Tasks) = False) then
         Partition_Small_Task
           (Src_Processors,
            Rmst_Tasks,
            My_Processor_Iterator);
      end if;

      if False = Is_Last_Element (Src_Processors, My_Processor_Iterator) and
         Is_Empty (Rmgt_Tasks) = False
      then
         -- Make processor set for RMGT
         --
         Next_Element (Src_Processors, My_Processor_Iterator);
         Current_Element (Src_Processors, A_Processor, My_Processor_Iterator);
         loop
            Add (Tmp_Cpus, A_Processor);
            exit when Is_Last_Element (Src_Processors, My_Processor_Iterator);
            Next_Element (Src_Processors, My_Processor_Iterator);
            Current_Element
              (Src_Processors,
               A_Processor,
               My_Processor_Iterator);
         end loop;
         Rm_General_Tasks (Tmp_Cpus, Rmgt_Tasks);

      else
         if Is_Last_Element (Src_Processors, My_Processor_Iterator) and
            Is_Empty (Rmgt_Tasks) = False
         then
            Duplicate (Src_Tasks, Result_Tasks);
            raise No_Such_Processors;
         end if;
      end if;

      -- Put results of two functions to returning task set
      --
      if not Is_Empty (Rmst_Tasks) then
         Duplicate (Rmst_Tasks, Result_Tasks);
      end if;

      if not Is_Empty (Rmgt_Tasks) then
         Reset_Iterator (Rmgt_Tasks, My_Task_Iterator_Result);
         Current_Element
           (Rmgt_Tasks,
            Generic_Task_Ptr (A_Task_Result),
            My_Task_Iterator_Result);
         loop
            Add (Result_Tasks, Generic_Task_Ptr (A_Task_Result));
            exit when Is_Last_Element (Rmgt_Tasks, My_Task_Iterator_Result);
            Next_Element (Rmgt_Tasks, My_Task_Iterator_Result);
            Current_Element
              (Rmgt_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);
         end loop;
      end if;
      Sort (Result_Tasks, Increasing_Name'Access);

   end Partition_General_Task;

   procedure Partition_Next_Fit
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set)
   is

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;

      M                 : Integer := 2;
      Total_Utilization : Float   := 0.0;
      Condition_Ip      : Boolean := False;

   begin

      Validate_Multiprocessor_Tasks (Src_Processors, Src_Tasks);

      Msg := To_Unbounded_String (" (") &
             Lb_See (Current_Language) &
             To_Unbounded_String ("[9], [10]) ");

      Duplicate (Src_Tasks, Result_Tasks);
      Sort (Result_Tasks, Increasing_Period'Access);

      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);
      Reset_Iterator (Src_Processors, My_Processor_Iterator);
      Current_Element (Src_Processors, A_Processor, My_Processor_Iterator);

      loop
         if Float (A_Task_Result.Capacity) / Float (A_Task_Result.Period) <=
            1.0
         then
            A_Task_Result.Cpu_Name := A_Processor.Name;
         else
            Duplicate (Src_Tasks, Result_Tasks);
            raise No_Such_Processors;
         end if;
         exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
         loop
            exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
            Total_Utilization := Total_Utilization +
                                 Float (A_Task_Result.Capacity) /
                                 Float (A_Task_Result.Period);
            Next_Element (Result_Tasks, My_Task_Iterator_Result);
            Current_Element
              (Result_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);

            if Float (A_Task_Result.Capacity) /
               Float (A_Task_Result.Period) <=
               2.0 *
               (1.0 /
                (1.0 + Total_Utilization / Float ((M - 1))) ** (M - 1)) -
               1.0 and
               Total_Utilization <=
               Float (M - 1) * (2.0 ** (1.0 / Float (M - 1)) - 1.0)
            then
               Condition_Ip := True;
            else
               Condition_Ip := False;
            end if;

            if Condition_Ip then
               A_Task_Result.Cpu_Name := A_Processor.Name;
               M                      := M + 1;
            else
               if Is_Last_Element
                    (Src_Processors,
                     My_Processor_Iterator)
               then
                  Duplicate (Src_Tasks, Result_Tasks);
                  raise No_Such_Processors;
               end if;
               Next_Element (Src_Processors, My_Processor_Iterator);
               Current_Element
                 (Src_Processors,
                  A_Processor,
                  My_Processor_Iterator);
               Total_Utilization := 0.0;
               M                 := 2;
               exit;
            end if;
         end loop;

      end loop;

      Sort (Result_Tasks, Increasing_Name'Access);

   end Partition_Next_Fit;

   procedure Partition_First_Fit
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set)
   is

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      Temp_Task_Iterator      : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;

      Num_Of_Tasks                        : Integer := 0;
      Task_Utilization, Total_Utilization : Float   := 0.0;
   begin
      Validate_Multiprocessor_Tasks (Src_Processors, Src_Tasks);

      Msg := To_Unbounded_String (" (") &
             Lb_See (Current_Language) &
             To_Unbounded_String ("[9], [10]) ");

      Duplicate (Src_Tasks, Result_Tasks);
      Sort (Result_Tasks, Increasing_Period'Access);

      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Temp_Task_Iterator := My_Task_Iterator_Result;
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);
      Reset_Iterator (Src_Processors, My_Processor_Iterator);
      Current_Element (Src_Processors, A_Processor, My_Processor_Iterator);

      -- Zero processor locations
      --
      loop
         A_Task_Result.Cpu_Name := To_Unbounded_String ("");
         exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
         Next_Element (Result_Tasks, My_Task_Iterator_Result);
         Current_Element
           (Result_Tasks,
            Generic_Task_Ptr (A_Task_Result),
            My_Task_Iterator_Result);
      end loop;

      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);
      loop
         loop
         -- Fetch information for condition-IP
         --
            Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
            Current_Element
              (Result_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);
            Num_Of_Tasks      := 0;
            Total_Utilization := 0.0;
            loop
               if A_Task_Result.Cpu_Name = A_Processor.Name then
                  Num_Of_Tasks      := Num_Of_Tasks + 1;
                  Total_Utilization := Total_Utilization +
                                       Float (A_Task_Result.Capacity) /
                                       Float (A_Task_Result.Period);
               end if;
               exit when Is_Last_Element
                            (Result_Tasks,
                             My_Task_Iterator_Result);
               Next_Element (Result_Tasks, My_Task_Iterator_Result);
               Current_Element
                 (Result_Tasks,
                  Generic_Task_Ptr (A_Task_Result),
                  My_Task_Iterator_Result);
            end loop;
            My_Task_Iterator_Result := Temp_Task_Iterator;
            Current_Element
              (Result_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);

            Task_Utilization := Float (A_Task_Result.Capacity) /
                                Float (A_Task_Result.Period);
            if (Task_Utilization <=
                2.0 *
                (1.0 /
                 (1.0 + Total_Utilization / Float (Num_Of_Tasks)) **
                 Num_Of_Tasks) -
                1.0)
            then
               A_Task_Result.Cpu_Name := A_Processor.Name;
               Reset_Iterator (Src_Processors, My_Processor_Iterator);
               Current_Element
                 (Src_Processors,
                  A_Processor,
                  My_Processor_Iterator);
               exit;
            else
               if Is_Last_Element
                    (Src_Processors,
                     My_Processor_Iterator)
               then
                  Duplicate (Src_Tasks, Result_Tasks);
                  raise No_Such_Processors;
               end if;
               Next_Element (Src_Processors, My_Processor_Iterator);
               Current_Element
                 (Src_Processors,
                  A_Processor,
                  My_Processor_Iterator);
            end if;
         end loop;

         exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
         Next_Element (Result_Tasks, My_Task_Iterator_Result);
         Temp_Task_Iterator := My_Task_Iterator_Result;
      end loop;

      Sort (Result_Tasks, Increasing_Name'Access);

   end Partition_First_Fit;

   procedure Partition_Best_Fit
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set;
      Msg            : in out Unbounded_String;
      Result_Tasks   : in out Tasks_Set)
   is

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;
      Temp_Task_Iterator      : Tasks_Iterator;
      Best_Cpu                : Processors_Iterator;

      M                            : Integer := 1;
      Best_Fit_Value               : Float   := 2.0;
      Best_Fit_Tmp_Value, Tot_Util : Float;
      First_Task                   : Boolean := True;
      Assignable                   : Boolean;

   begin

      Validate_Multiprocessor_Tasks (Src_Processors, Src_Tasks);

      Msg := To_Unbounded_String (" (") &
             Lb_See (Current_Language) &
             To_Unbounded_String ("[9], [10]) ");

      Duplicate (Src_Tasks, Result_Tasks);
      Sort (Result_Tasks, Increasing_Period'Access);

      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);

      -- Zero processor locations
      --
      loop
         A_Task_Result.Cpu_Name := To_Unbounded_String ("");
         exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
         Next_Element (Result_Tasks, My_Task_Iterator_Result);
         Current_Element
           (Result_Tasks,
            Generic_Task_Ptr (A_Task_Result),
            My_Task_Iterator_Result);
      end loop;
      Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
      Current_Element
        (Result_Tasks,
         Generic_Task_Ptr (A_Task_Result),
         My_Task_Iterator_Result);

      loop
         Reset_Iterator (Src_Processors, My_Processor_Iterator);
         Current_Element (Src_Processors, A_Processor, My_Processor_Iterator);

         -- Loop scans best processor for task to be placed
         --
         loop
            Temp_Task_Iterator := My_Task_Iterator_Result;

            -- get CPU data
            --
            Reset_Iterator (Result_Tasks, My_Task_Iterator_Result);
            Current_Element
              (Result_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);
            M        := 1;
            Tot_Util := 0.0;
            loop
               if A_Task_Result.Cpu_Name = A_Processor.Name then
                  M        := M + 1;
                  Tot_Util := Tot_Util +
                              Float (A_Task_Result.Capacity) /
                              Float (A_Task_Result.Period);
               end if;
               exit when Is_Last_Element
                            (Result_Tasks,
                             My_Task_Iterator_Result);
               Next_Element (Result_Tasks, My_Task_Iterator_Result);
               Current_Element
                 (Result_Tasks,
                  Generic_Task_Ptr (A_Task_Result),
                  My_Task_Iterator_Result);
            end loop;
            My_Task_Iterator_Result := Temp_Task_Iterator;
            Current_Element
              (Result_Tasks,
               Generic_Task_Ptr (A_Task_Result),
               My_Task_Iterator_Result);

            -- test for condition IP
            --
            if First_Task then
               if Float (A_Task_Result.Capacity) /
                  Float (A_Task_Result.Period) <=
                  1.0
               then
                  Reset_Iterator (Src_Processors, Best_Cpu);
                  My_Processor_Iterator := Best_Cpu;
                  First_Task            := False;
                  Assignable            := True;
                  exit;
               else
                  Assignable := False;
                  exit;
               end if;
            else
               if Float (A_Task_Result.Capacity) /
                  Float (A_Task_Result.Period) <=
                  2.0 *
                  (1.0 / (1.0 + Tot_Util / Float ((M - 1))) ** (M - 1)) -
                  1.0
               then
                  -- Tricky Condition IP testing because mathematics function
                  -- doesn't seem to understand handle
                  -- zero number very well in following check (M := 0
                  -- happens when processor is empty)
                  --
                  if M > 1 then
                     if Tot_Util <=
                        Float (M - 1) *
                        (2.0 ** Float (1.0 / Float (M - 1)) - 1.0)
                     then
                        Best_Fit_Tmp_Value :=
                          2.0 *
                          (1.0 /
                           ((1.0 + Tot_Util / Float (M - 1)) ** (M - 1))) -
                          1.0;
                        if Best_Fit_Tmp_Value < Best_Fit_Value then
                           Best_Fit_Value := Best_Fit_Tmp_Value;
                           Best_Cpu       := My_Processor_Iterator;
                           Assignable     := True;
                        end if;
                     end if;
                  else
                     Best_Fit_Tmp_Value :=
                       2.0 *
                       (1.0 /
                        (1.0 + Tot_Util / Float (M - 1) ** (M - 1))) -
                       1.0;
                     if Best_Fit_Tmp_Value < Best_Fit_Value then
                        Best_Fit_Value := Best_Fit_Tmp_Value;
                        Best_Cpu       := My_Processor_Iterator;
                        Assignable     := True;
                     end if;
                  end if;
                  -- End of tricky testing
                  --
               end if;
            end if;

            exit when Is_Last_Element (Src_Processors, My_Processor_Iterator);
            Next_Element (Src_Processors, My_Processor_Iterator);
            Current_Element
              (Src_Processors,
               A_Processor,
               My_Processor_Iterator);
         end loop;

         if Assignable then
            -- save task to best CPU
            --
            Current_Element (Src_Processors, A_Processor, Best_Cpu);
            A_Task_Result.Cpu_Name := A_Processor.Name;
            Assignable             := False;
            Best_Fit_Value         := 2.0;
         else
            Duplicate (Src_Tasks, Result_Tasks);
            raise No_Such_Processors;
         end if;

         exit when Is_Last_Element (Result_Tasks, My_Task_Iterator_Result);
         Next_Element (Result_Tasks, My_Task_Iterator_Result);
         Current_Element
           (Result_Tasks,
            Generic_Task_Ptr (A_Task_Result),
            My_Task_Iterator_Result);
      end loop;

      Sort (Result_Tasks, Increasing_Name'Access);
   end Partition_Best_Fit;

   procedure Validate_Multiprocessor_Tasks
     (Src_Processors : in Processors_Set;
      Src_Tasks      : in Tasks_Set)
   is

      A_Processor           : Processor_Ptr;
      My_Processor_Iterator : Processors_Iterator;

   begin
      -- check if tasks are periodic
      --
      Reset_Iterator (Src_Processors, My_Processor_Iterator);
      loop
         Current_Element (Src_Processors, A_Processor, My_Processor_Iterator);
         Periodic_Control (Src_Tasks, A_Processor.Name);
         exit when True =
                   Is_Last_Element (Src_Processors, My_Processor_Iterator);
         Next_Element (Src_Processors, My_Processor_Iterator);
      end loop;

      -- check if processors are rate-monotonic and non pre-emptive
      --
      Reset_Iterator (Src_Processors, My_Processor_Iterator);
      loop
         Current_Element (Src_Processors, A_Processor, My_Processor_Iterator);
         if Get_Name (A_Processor.Scheduler.all) /=
            Rate_Monotonic_Protocol
         then
            raise Invalid_Scheduler_Multiprocessors;
         end if;
         if Get_Preemptive (A_Processor.Scheduler.all) =
            Not_Preemptive
         then
            raise Invalid_Scheduler_Multiprocessors;
         end if;
         exit when Is_Last_Element (Src_Processors, My_Processor_Iterator);
         Next_Element (Src_Processors, My_Processor_Iterator);
      end loop;

   end Validate_Multiprocessor_Tasks;

end Multiprocessor_Services;
