------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------




WITH Events;
USE Events;
WITH Systems;
USE Systems;
WITH Processors;
USE Processors;
WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;
WITH Message_Set;
USE Message_Set;
USE Message_Set.Generic_Message_Set;
WITH Event_Analyzers;
USE Event_Analyzers;
WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Framework_Config;
USE Framework_Config;
WITH Scheduling_Sequences;
USE Scheduling_Sequences;
WITH Task_Dependencies;
USE Task_Dependencies;
WITH Indexed_Tables;
WITH Unchecked_Deallocation;
with Processor_Set;       use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Address_Space_Set; use Address_Space_Set;
use Address_Space_Set.Generic_Address_Space_Set;
with Time_Units; use Time_Units;
use Time_Units.A_Time_Unit;
with Task_Set; use Task_Set;
use Task_Set.Generic_Task_Set;
with Resources;    use Resources;
use Resources.Task_Lists;
with Resource_Set; use Resource_Set;
use Resource_Set.Generic_Resource_Set;



PACKAGE Multiprocessor_Services IS



   --------------------------------------------------
   -- The types below define data layout to
   -- store information computed at simulation time
   --------------------------------------------------


   TYPE Scheduling_Result IS
      RECORD
         Scheduling_Msg : Unbounded_String;
         Has_Error      : Boolean;
         Error_Msg      : Unbounded_String;
         Result         : Scheduling_Sequence_Ptr;
      END RECORD;

   TYPE Scheduling_Result_Ptr IS ACCESS ALL Scheduling_Result;


   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Scheduling_Result, Scheduling_Result_Ptr);

   PROCEDURE Put (
         R : IN     Scheduling_Result);

   PROCEDURE Initialize (
         R :   OUT Scheduling_Result);



   PACKAGE Scheduling_Result_Per_Processor IS NEW Indexed_Tables(
      Scheduling_Result, Processor_Ptr, Framework_Config.Max_Processors, 0,
      Put,
      Initialize, Processors.Put_Name);
   USE Scheduling_Result_Per_Processor;

   SUBTYPE Scheduling_Table IS Scheduling_Result_Per_Processor.
      Indexed_Table;
   SUBTYPE Scheduling_Table_Range IS Scheduling_Result_Per_Processor.
      Indexed_Table_Range;

   TYPE Scheduling_Table_Ptr IS ACCESS ALL Scheduling_Table;

   PROCEDURE Free (
         Sched : IN OUT Scheduling_Table_Ptr);


   ---------------------------------------
   -- Put the event table to the screen
   ---------------------------------------

   PROCEDURE Put (
         Sched : IN     Scheduling_Table_Ptr);


   ------------------------------------------------------
   -- Write/read the event table into/from a XML file
   ------------------------------------------------------

   PROCEDURE Write_To_Xml_File (
         Sched     : IN     Scheduling_Table_Ptr;
         Sys       : IN     System;
         File_Name : IN     String);

   PROCEDURE Write_To_Xml_File (
         Sched     : IN     Scheduling_Table_Ptr;
         Sys       : IN     System;
         File_Name : IN     Unbounded_String);

   PROCEDURE Read_From_Xml_File (
         Sched     : IN OUT Scheduling_Table_Ptr;
         Sys       : IN     System;
         File_Name : IN     String);

   PROCEDURE Read_From_Xml_File (
         Sched     : IN OUT Scheduling_Table_Ptr;
         Sys       : IN     System;
         File_Name : IN     Unbounded_String);



   -----------------------------------------------------------------
   --  This function is the major entry of the scheduling simulator
   -----------------------------------------------------------------



   PROCEDURE Build_Multiprocessor_Scheduling (
         My_Processors           : IN     Processors_Set;
         My_Address_Spaces       : IN     Address_Spaces_Set;
         My_Tasks                : IN     Tasks_Set;
         My_Resources            : IN     Resources_Set;
         My_Buffers              : IN     Buffers_Set;
         My_Messages             : IN     Messages_Set;
         My_Dependencies         : IN     Tasks_Dependencies_Ptr;
         Result                  : IN OUT Scheduling_Table_Ptr;
         Event_To_Generate       : IN     Time_Unit_Event_Type_Boolean_Table;
         Last_Time               : IN     Natural;
         With_Offsets            : IN     Boolean                            := True;
         With_Precedencies       : IN     Boolean                            := True;
         With_Resources          : IN     Boolean                            := True;
         With_Task_Specific_Seed : IN     Boolean                            := True;
         Global_Seed_Value       : IN     Integer                            := 0;
         Predictable_Global_Seed : IN     Boolean                            := True);


   --------------------------------------------------------------
   -- User's defined event anayser
   -- From a scheduling table, this function do an
   -- user's defined analysis (computed with "event analyzers"
   --------------------------------------------------------------


   PROCEDURE Run_An_Event_Analyzer (
         My_Tasks          : IN     Tasks_Set;
         My_Processors     : IN     Processors_Set;
         My_Address_Spaces : IN     Address_Spaces_Set;
         My_Resources      : IN     Resources_Set;
         My_Buffers        : IN     Buffers_Set;
         My_Messages       : IN     Messages_Set;
         Sched             : IN     Scheduling_Table_Ptr;
         An_Event_Analyzer : IN     Event_Analyzer_Ptr;
         Result            : IN OUT Unbounded_String);



   --------------------------------------------------------------
   -- Partitionning tools  : these subprograms allow users to
   -- split tasks on a set of processors
   --------------------------------------------------------------


   -- All processors should have the same scheduler
   --
   Processors_Should_Have_The_Same_Scheduler : EXCEPTION;

   -- We need more processors to do partition of this task set
   --
   No_Such_Processors : EXCEPTION;

   -- can we applied scheduling partition algorithms with this scheduler ?
   --
   Invalid_Scheduler_Multiprocessors : EXCEPTION;



   PROCEDURE Partition_General_Task (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set);


   PROCEDURE Partition_Small_Task (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set);

   PROCEDURE Partition_Small_Task (
         Src_Processors   : IN     Processors_Set;
         Src_Tasks        : IN     Tasks_Set;
         Msg              : IN OUT Unbounded_String;
         Result_Tasks     : IN OUT Tasks_Set;
         Processor_Result :    OUT Processors_Iterator);

   PROCEDURE Partition_Small_Task (
         Src_Processors   : IN     Processors_Set;
         Result_Tasks     : IN OUT Tasks_Set;
         Processor_Result :    OUT Processors_Iterator);



   PROCEDURE Rm_General_Tasks (
         Src_Processors : IN     Processors_Set;
         Result_Tasks   : IN OUT Tasks_Set);



   PROCEDURE Partition_Next_Fit (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set);



   PROCEDURE Partition_First_Fit (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set);



   PROCEDURE Partition_Best_Fit (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set);



   PROCEDURE Validate_Multiprocessor_Tasks (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set);








END Multiprocessor_Services;


