------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------






WITH Ada.Sequential_Io;
WITH Io_Tools;
USE Io_Tools;


PACKAGE BODY Network.Predefined IS



   PROCEDURE Initialize (
         A_Network : IN OUT Predefined_Network) IS
   BEGIN
      Reset(A_Network);
   END Initialize;

   PROCEDURE Finalize (
         A_Network : IN OUT Predefined_Network) IS
   BEGIN
      NULL;
   END Finalize;

   PROCEDURE Adjust (
         A_Network : IN OUT Predefined_Network) IS
   BEGIN
      NULL;
   END Adjust;



   TYPE Descriptor IS
      RECORD
         Reception_Latency_Bound : Natural;
         Emission_Latency_Bound  : Natural;
         Media_Length            : Natural;
         Name                    : Unbounded_String;
         Number_Of_Messages      : Response_Range;
         Rt                      : Response_Table;
      END RECORD;


   PROCEDURE Dispatched_Read_From_File (
         My_Network : IN OUT Predefined_Network;
         File_Name  : IN     Unbounded_String) IS

      Desc : Descriptor;

      PACKAGE Uni_File IS NEW Ada.Sequential_Io( Descriptor );
      USE Uni_File;

      A_Uni_File : Uni_File.File_Type;
   BEGIN
      Open(A_Uni_File,
         Mode => In_File,
         Name => To_String(File_Name));

      IF NOT Is_Open(A_Uni_File)
            THEN
         RAISE Can_Not_Open_File_For_Reading;
      END IF;

      Read(A_Uni_File,Desc);

      My_Network.Reception_Latency_Bound:=Desc.Reception_Latency_Bound;
      My_Network.Emission_Latency_Bound:=Desc.Emission_Latency_Bound;
      My_Network.Media_Length:=Desc.Media_Length;
      My_Network.Name:=Desc.Name;
      My_Network.Rt:=Desc.Rt;
      My_Network.Number_Of_Messages:=Desc.Number_Of_Messages;

      Close(A_Uni_File);

   END Dispatched_Read_From_File;


   PROCEDURE Dispatched_Write_To_File (
         My_Network : IN     Predefined_Network;
         File_Name  : IN     Unbounded_String) IS

      Desc : Descriptor;

      PACKAGE Uni_File IS NEW Ada.Sequential_Io( Descriptor );
      USE Uni_File;

      A_Uni_File : Uni_File.File_Type;
   BEGIN

      Create(A_Uni_File,
         Mode => Out_File,
         Name => To_String(File_Name));

      IF NOT Is_Open(A_Uni_File)
            THEN
         RAISE Can_Not_Open_File_For_Writing;
      END IF;

      Desc.Reception_Latency_Bound:=My_Network.Reception_Latency_Bound;
      Desc.Emission_Latency_Bound:=My_Network.Emission_Latency_Bound;
      Desc.Media_Length:=My_Network.Media_Length;
      Desc.Name:=My_Network.Name;
      Desc.Rt:=My_Network.Rt;
      Desc.Number_Of_Messages:=My_Network.Number_Of_Messages;

      Write(A_Uni_File,Desc);

      Close(A_Uni_File);

   END Dispatched_Write_To_File;


   PROCEDURE Dispatched_Reset (
         My_Network : IN OUT Predefined_Network) IS
   BEGIN
      My_Network.Number_Of_Messages:=0;
      My_Network.Architecture:=Unknown;
      My_Network.Name:=To_Unbounded_String(
         "Network with predefined response time");
      Initialize(My_Network.Rt);
   END Dispatched_Reset;


   PROCEDURE Dispatched_Put (
         My_Network : IN     Predefined_Network) IS
   BEGIN
      Put(My_Network.Rt);
   END Dispatched_Put;


   PROCEDURE Dispatched_Transmission_Delay (
         My_Network    : IN     Predefined_Network;
         My_Messages   : IN     Messages_Set;
         Response_Time :    OUT Response_Table) IS

      Msg_Jitter : Double;
   BEGIN
      Response_Time:=My_Network.Rt;

      -- Add message jitter to predefined message response time
      --
      FOR I IN 0..(My_Network.Number_Of_Messages-1) LOOP
         Msg_Jitter:=Double(Message_Set.Get(My_Messages, My_Network.Rt.
               Entries(I).
               Item.Name, Jitter));
         Response_Time.Entries(I).Data:=Msg_Jitter+Response_Time.Entries(
            I).Data;
      END LOOP;


   END Dispatched_Transmission_Delay;



   PROCEDURE Register_Message (
         A_Network     : IN OUT Predefined_Network;
         Message_Name  : IN     Unbounded_String;
         Response_Time : IN     Double) IS
   BEGIN
      FOR I IN 0..(A_Network.Number_Of_Messages-1) LOOP
         IF (A_Network.Rt.Entries(I).Item.Name = Message_Name)
               THEN
            A_Network.Rt.Entries(I).Data:=Response_Time;
            RETURN;
         END IF;
      END LOOP;

      A_Network.Rt.Entries(A_Network.Number_Of_Messages).Data:=
         Response_Time;
      A_Network.Rt.Entries(A_Network.Number_Of_Messages).Item.Name:=
         Message_Name;
      A_Network.Number_Of_Messages:=A_Network.Number_Of_Messages+1;

   END Register_Message;


   PROCEDURE Register_Messages (
         A_Network     : IN OUT Predefined_Network;
         Number        : IN     Response_Range;
         Response_Time : IN     Response_Table) IS
   BEGIN
      FOR I IN 0..(Number-1) LOOP
         Register_Message(A_Network,Response_Time.Entries(I).Item.Name,
            Response_Time.Entries(I).Data);
      END LOOP;
   END Register_Messages;




END Network.Predefined;
