------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Unbounded_Strings; use Unbounded_Strings;

package body Network_Set is

   procedure Check_Network
     (My_Networks  : in Networks_Set;
      Name         : in Unbounded_String;
      Network_Type : in Unbounded_String)
   is

      My_Iterator : Iterator;
      A_Network   : Network_Ptr;

   begin

      if (Name = "") then
         raise Invalid_Parameter;
      end if;

      if (Get_Number_Of_Elements (My_Networks) > 0) then
         Reset_Iterator (My_Networks, My_Iterator);

         loop
            Current_Element (My_Networks, A_Network, My_Iterator);

            if (Name = A_Network.Name) then
               raise Network_Already_Defined;
            end if;

            exit when Is_Last_Element (My_Networks, My_Iterator);

            Next_Element (My_Networks, My_Iterator);

         end loop;
      end if;

      if (Network_Type /= "BOUNDED_DELAY") and
         (Network_Type /= "PARAMETRIC_DELAY")
      then
         raise Invalid_Parameter;
      end if;

   end Check_Network;

   procedure Add_Network
     (My_Networks  : in out Networks_Set;
      Name         : in Unbounded_String;
      Network_Type : in Unbounded_String)
   is

      A_Network : Network_Ptr;

   begin

      Check_Network (My_Networks, Name, Network_Type);

      A_Network := new Network;

      if (Network_Type = "BOUNDED_DELAY") then
         A_Network.Network_Type := Bounded_Delay;
      else
         A_Network.Network_Type := Parametric_Delay;
      end if;

      A_Network.Name := Name;

      Add (My_Networks, A_Network);

   end Add_Network;

   procedure Update_Network
     (My_Networks  : in out Networks_Set;
      Name         : in Unbounded_String;
      Network_Type : in Unbounded_String)
   is

      A_Network : Network_Ptr;

   begin

      A_Network := Search_Network (My_Networks, Name);
      Delete (My_Networks, A_Network);
      Add_Network (My_Networks, Name, Network_Type);

   end Update_Network;

   function Search_Network
     (My_Networks : in Networks_Set;
      Name        : in Unbounded_String)
      return        Network_Ptr
   is
      My_Iterator : Iterator;
      A_Network   : Network_Ptr;
      Result      : Network_Ptr;

      Found : Boolean := False;

   begin

      Reset_Iterator (My_Networks, My_Iterator);

      loop
         Current_Element (My_Networks, A_Network, My_Iterator);

         if (A_Network.Name = Name) then
            Found  := True;
            Result := A_Network;
         end if;

         exit when Is_Last_Element (My_Networks, My_Iterator);

         Next_Element (My_Networks, My_Iterator);

      end loop;

      if not Found then
         raise Network_Not_Found;
      end if;

      return Result;

   end Search_Network;

   function Export_Xml
     (My_Networks : in Networks_Set)
      return        Unbounded_String
   is
      My_Iterator : Network_Iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Networks) then

         Result := Result &
                   To_Unbounded_String ("<networks>") &
                   Unbounded_Lf;

         Reset_Iterator (My_Networks, My_Iterator);

         loop
            Current_Element (My_Networks, A_Network, My_Iterator);

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "<network>") &
                      Unbounded_Lf;

            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<name>" &
                  To_String (A_Network.Name) &
                  "</name>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<type>" &
                  A_Network.Network_Type'Img &
                  "</type>") &
              Unbounded_Lf;

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "</network>") &
                      Unbounded_Lf;

            exit when Is_Last_Element (My_Networks, My_Iterator);

            Next_Element (My_Networks, My_Iterator);

         end loop;
         Result := Result &
                   To_Unbounded_String ("</networks>") &
                   Unbounded_Lf;
      end if;

      return Result;

   end Export_Xml;

   function Export_Aadl_Implementations
     (My_Networks : in Networks_Set)
      return        Unbounded_String
   is
      My_Iterator : Iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Networks) then

         Reset_Iterator (My_Networks, My_Iterator);

         loop
            Current_Element (My_Networks, A_Network, My_Iterator);
            exit when Is_Last_Element (My_Networks, My_Iterator);

            Result := Result &
                      To_Unbounded_String
                         ("bus " & To_String (A_Network.Name));
            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Network.Name) & ";");

            Result := Result &
                      To_Unbounded_String
                         ("bus implementation " &
                          To_String (A_Network.Name) &
                          ".Impl");
            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Network.Name) & ".Impl");

            Next_Element (My_Networks, My_Iterator);

         end loop;
      end if;

      return Result;

   end Export_Aadl_Implementations;

   function Export_Aadl_Declarations
     (My_Networks  : in Networks_Set;
      Number_Of_Ht : in Natural)
      return         Unbounded_String
   is
      My_Iterator : Iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Networks) then

         Reset_Iterator (My_Networks, My_Iterator);

         loop
            Current_Element (My_Networks, A_Network, My_Iterator);
            exit when Is_Last_Element (My_Networks, My_Iterator);

            for I in  1 .. Number_Of_Ht loop
               Result := Result & ASCII.HT;
            end loop;
            Result :=
              Result &
              To_Unbounded_String
                 ("instancied_" &
                  To_String (A_Network.Name) &
                  " : bus " &
                  To_String (A_Network.Name) &
                  ".Impl;");

            Next_Element (My_Networks, My_Iterator);

         end loop;
      end if;

      return Result;

   end Export_Aadl_Declarations;

end Network_Set;
