------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Message_Set;           use Message_Set;
use Message_Set.Generic_Message_Set;
with Buffer_Set;            use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Task_Set;              use Task_Set;
use Task_Set.Generic_Task_Set;
with Resource_Set;          use Resource_Set;
use Resource_Set.Generic_Resource_Set;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Unbounded_Strings;     use Unbounded_Strings;
with Framework_Config;      use Framework_Config;
with Simulations;           use Simulations;
with Tables;
with Ada.Finalization;

package Expressions is

   -- The data type does not match the expected type
   --
   Type_Error : exception;

   -- Statement exception (sush as unexpected statement)
   --
   Statement_Error : exception;

   -- Exception raised when an error occurs
   -- on variable  operations
   --
   Variable_Error : exception;

   -- Syntax error detected
   --
   Syntax_Error : exception;

   -- This exception is raised when a parametric file can not be read
   --
   Parametric_File_Error : exception;

   ---------------------------------------------------------
   -- Valid operator used to build expressions
   --
   type Operator is (
      Plus,
      Minus,
      Div,
      Mult,
      Exp,
      Modulo,
      Equal,
      Not_Equal,
      Equal_Less,
      Equal_Greater,
      Inf,
      Sup,
      Logic_And,
      Logic_Or,
      Logic_Not,
      Min_Operator,
      Max_Operator,
      Min_To_Index,
      Max_To_Index,
      Lcm,
      Abs_Operator,
      Get_Resource_Index,
      Get_Task_Index,
      Get_Message_Index,
      Get_Buffer_Index,
      Concat);

   procedure Put (S : in Operator);

   -----------------------------------------------------------

   -- A predefinition of the variable type
   --
   type Variable_Expr;
   type Variable_Expr_Ptr is access all Variable_Expr'Class;

   ---------------------------------------------------------
   -- Table to store the list of variables of
   -- a parametric scheduler. For each variable,
   -- the current simulation values are also stored
   --

   type Variable_Record is record
      Variable   : Variable_Expr_Ptr;
      Simulation : Simulation_Value_Ptr;
   end record;
   type Variable_Record_Ptr is access Variable_Record;

   procedure Put (V : in Variable_Record_Ptr);

   package Variables_Type_Package is new Tables (
      Variable_Record_Ptr,
      Max_Variables,
      Put);
   use Variables_Type_Package;

   subtype Variables_Range is Variables_Type_Package.Table_Range;
   subtype Variables_Table_Type is Variables_Type_Package.Table;

   -- Looking for a variable in the variable table
   --
   function Find_Variable
     (Var_Table : in Variables_Table_Type;
      S         : in Unbounded_String;
      Line      : in Natural := 0)
      return      Variables_Range;

   function Find_Variable
     (Var_Table : in Variables_Table_Type;
      S         : in String;
      Line      : in Natural := 0)
      return      Variables_Range;

   -- Check is a given variable is already defined in a
   -- variable table
   --
   procedure Check_Variable_Declaration
     (Var_Table : in Variables_Table_Type;
      S         : in Unbounded_String;
      Line      : in Natural := 0);

   -- Load into the parametric simulator
   -- data from task, processor, message and buffer elements
   --
   -- The number of object (tasks, resources, buffers) is initialized
   -- accordig to the correspondig processor (parameter processor_name)
   -- if processor_name is an empty string, all tasks/buffers/resources
   -- of all procesors are considered
   --
   procedure Initialize_Parametric_Variables
     (Var_Table      : in out Variables_Table_Type;
      My_Messages    : in Messages_Set;
      My_Buffers     : in Buffers_Set;
      My_Resources   : in Resources_Set;
      My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String := Empty_String);

   -- Add the parametric variable in a variable table
   --
   procedure Create_Parametric_Variables
     (Variables_Table : in out Variables_Table_Type;
      My_Tasks        : in Tasks_Set);

   ---------------------------------------------------------

   type Expressions_Type is (
      Constant_Type,
      Variable_Type,
      Array_Variable_Type,
      Binary_Type,
      Unary_Type,
      Random_Type);

   type Generic_Expression is abstract new Ada.Finalization.Controlled with
      record
         Expression_Type : Expressions_Type;
      end record;
   type Expression_Ptr is access all Generic_Expression'Class;

   procedure Put (S : in Generic_Expression) is abstract;

   procedure Put (S : in Expression_Ptr) is abstract;

   function To_Unbounded_String
     (C    : Generic_Expression)
      return Unbounded_String
   is abstract;

   function Get_Type
     (C    : Generic_Expression)
      return Simulation_Type
   is abstract;

   function Value_Of
     (C    : Generic_Expression;
      V    : Variables_Table_Type)
      return Simulation_Value_Ptr
   is abstract;

   ------------------------------------------

   type Constant_Expr is new Generic_Expression with record
      Value : Simulation_Value_Ptr;
   end record;

   type Constant_Expr_Ptr is access all Constant_Expr'Class;

   procedure Put (S : in Constant_Expr);
   procedure Put (S : in Constant_Expr_Ptr);

   function To_Unbounded_String (C : Constant_Expr) return Unbounded_String;

   function Get_Type (C : Constant_Expr) return Simulation_Type;

   procedure Initialize (S : in out Constant_Expr);

   function Value_Of
     (C    : Constant_Expr;
      V    : Variables_Table_Type)
      return Simulation_Value_Ptr;

   ------------------------------------------

   type Variable_Expr is new Generic_Expression with record

   -- Name of the variable
   --
      Name : Unbounded_String;

      -- Type of the variable (integer, boolean, ...)
      --
      Variable_Type : Simulation_Type;
   end record;

   procedure Put (S : in Variable_Expr);
   procedure Put (S : in Variable_Expr_Ptr);

   function To_Unbounded_String (C : Variable_Expr) return Unbounded_String;

   function Get_Type (C : Variable_Expr) return Simulation_Type;

   procedure Initialize (S : in out Variable_Expr);

   function Value_Of
     (C    : Variable_Expr;
      V    : Variables_Table_Type)
      return Simulation_Value_Ptr;

   ------------------------------------------

   type Array_Variable_Expr is new Variable_Expr with record
      Array_Index : Expression_Ptr;
   end record;

   type Array_Variable_Expr_Ptr is access all Array_Variable_Expr'Class;

   procedure Put (S : in Array_Variable_Expr);
   procedure Put (S : in Array_Variable_Expr_Ptr);

   function To_Unbounded_String
     (C    : Array_Variable_Expr)
      return Unbounded_String;

   procedure Initialize (S : in out Array_Variable_Expr);

   function Value_Of
     (C    : Array_Variable_Expr;
      V    : Variables_Table_Type)
      return Simulation_Value_Ptr;

   ------------------------------------------

   type Binary_Expr is new Generic_Expression with record
      Ope    : Operator;
      Rvalue : Expression_Ptr;
      Lvalue : Expression_Ptr;
   end record;
   type Binary_Expr_Ptr is access all Binary_Expr'Class;

   procedure Put (S : in Binary_Expr);
   procedure Put (S : in Binary_Expr_Ptr);

   function To_Unbounded_String (C : Binary_Expr) return Unbounded_String;

   function Get_Type (C : Binary_Expr) return Simulation_Type;

   procedure Initialize (S : in out Binary_Expr);

   function Value_Of
     (C    : Binary_Expr;
      V    : Variables_Table_Type)
      return Simulation_Value_Ptr;

   ------------------------------------------

   type Unary_Expr is new Generic_Expression with record
      Ope   : Operator;
      Value : Expression_Ptr;
   end record;
   type Unary_Expr_Ptr is access all Unary_Expr'Class;

   procedure Put (S : in Unary_Expr);
   procedure Put (S : in Unary_Expr_Ptr);

   function To_Unbounded_String (C : Unary_Expr) return Unbounded_String;

   function Get_Type (C : Unary_Expr) return Simulation_Type;

   procedure Initialize (S : in out Unary_Expr);

   function Value_Of
     (C    : Unary_Expr;
      V    : Variables_Table_Type)
      return Simulation_Value_Ptr;

end Expressions;
