------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_IO;                       use Text_IO;
with Ada.Strings.Unbounded.Text_IO; use Ada.Strings.Unbounded.Text_IO;
with Unbounded_Strings;             use Unbounded_Strings;
with Ada.Exceptions;                use Ada.Exceptions;
with GNAT.Current_Exception;        use GNAT.Current_Exception;
with Translate;                     use Translate;
with Time_Units;                    use Time_Units;
with Ada.Numerics.Float_Random;     use Ada.Numerics.Float_Random;

package body Statements is

   function Find_Set
     (Var  : in Sets_Table_Type;
      S    : in Unbounded_String)
      return Sets_Range
   is
   begin
      for I in  0 .. Var.Nb_Entries - 1 loop
         if Var.Entries (I).Set_Id = S then
            return I;
         end if;
      end loop;

      Raise_Exception
        (Expressions.Syntax_Error'Identity,
         To_String
            (S & Lb_Comma & Lb_Undeclared_Identifier (Current_Language)));

   end Find_Set;

   procedure Check_Set_Declaration
     (Var : in Sets_Table_Type;
      S   : in Unbounded_String)
   is

      Dummy : constant Sets_Range := Find_Set (Var, S);
   begin
      null;
   end Check_Set_Declaration;

   --------------------------------------------------------------

   procedure Initialize (S : in out Put_Statement) is
   begin
      S.Statement_Type             := Put_Statement_Type;
      S.Expression_To_Be_Displayed := null;
      S.Put_To                     := null;
      S.Put_From                   := null;
   end Initialize;

   procedure Initialize (S : in out Nop_Statement) is
   begin
      S.Statement_Type := Nop_Statement_Type;
   end Initialize;

   procedure Initialize (S : in out If_Statement) is
   begin
      S.Statement_Type := If_Statement_Type;
   end Initialize;

   procedure Initialize (S : in out Random_Initialize_Statement) is
   begin
      S.Statement_Type := Random_Initialize_Statement_Type;
   end Initialize;

   procedure Initialize (S : in out Affectation_Statement) is
   begin
      S.Statement_Type := Affectation_Statement_Type;
   end Initialize;

   procedure Initialize (S : in out For_Statement) is
   begin
      S.Statement_Type := For_Statement_Type;
   end Initialize;

   procedure Initialize (S : in out While_Statement) is
   begin
      S.Statement_Type := While_Statement_Type;
   end Initialize;

   procedure Initialize (S : in out Return_Statement) is
   begin
      S.Statement_Type := Return_Statement_Type;
   end Initialize;

   procedure Initialize (S : in out Set_Statement) is
   begin
      S.Statement_Type := Set_Statement_Type;
   end Initialize;

   procedure Initialize (S : in out Subprogram_Call_Statement) is
   begin
      S.Statement_Type := Subprogram_Call_Statement_Type;
   end Initialize;

   procedure Initialize (S : in out Subprogram_Statement) is
   begin
      S.Statement_Type := Subprogram_Statement_Type;
   end Initialize;

   --------------------------------------------------------------

   procedure Recursive_Put (S : in Statement_Ptr) is

   begin

      Put (S.all);

      if (S.Statement_Type = While_Statement_Type) then
         if (While_Statement_Ptr (S).Included_Statement /= null) then
            Put ("Included statement : ");
            New_Line;
            Recursive_Put (While_Statement_Ptr (S).Included_Statement);
         end if;
      end if;
      if (S.Statement_Type = For_Statement_Type) then
         if (For_Statement_Ptr (S).Included_Statement /= null) then
            Put ("Included statement : ");
            New_Line;
            Recursive_Put (For_Statement_Ptr (S).Included_Statement);
         end if;
      end if;

      if (S.Statement_Type = If_Statement_Type) then
         if (If_Statement_Ptr (S).Then_Statement /= null) then
            Put ("Then statement : ");
            New_Line;
            New_Line;
            Recursive_Put (If_Statement_Ptr (S).Then_Statement);
         end if;
         if (If_Statement_Ptr (S).Else_Statement /= null) then
            Put ("Else statement : ");
            New_Line;
            New_Line;
            Recursive_Put (If_Statement_Ptr (S).Else_Statement);
         end if;

      else
         if (S.Next_Statement /= null) then
            Put ("Next statement  : ");
            New_Line;
            New_Line;
            Recursive_Put (S.Next_Statement);
         end if;

      end if;

   end Recursive_Put;

   procedure Put (S : in Generic_Statement) is

   begin
      Put_Line ("Type : " & S.Statement_Type'Img);
      Put_Line ("Line number : " & S.Line_Number'Img);
   end Put;

   procedure Put (S : in Subprogram_Statement) is

   begin
      Put_Line ("Subprogram statement :");
      Put (Generic_Statement (S));
      Put ("Is a function : ");
      Put (S.Is_A_Function'Img);
      New_Line;
      Put ("Subprogram name  : ");
      Put (S.Subprogram_Name);
      New_Line;
   end Put;

   procedure Put (S : in Subprogram_Call_Statement) is

   begin
      Put_Line ("Subprogram call statement :");
      Put (Generic_Statement (S));
      Put ("Is a function : ");
      Put (S.Is_A_Function'Img);
      New_Line;
   end Put;

   procedure Put (S : in Nop_Statement) is

   begin
      Put_Line ("Nop statement :");
      Put (Generic_Statement (S));
      New_Line;
   end Put;

   procedure Put (S : in Put_Statement) is

   begin
      Put_Line ("Put statement :");
      Put (Generic_Statement (S));
      Put ("Expression  : ");
      Put (S.Expression_To_Be_Displayed.all);
      New_Line;
      if (S.Put_From /= null) and (S.Put_To /= null) then
         Put_Line ("From/to : ");
         Put (S.Put_From.all);
         Put (" / ");
         Put (S.Put_To.all);
         New_Line;
      end if;
   end Put;

   procedure Put (S : in If_Statement) is

   begin
      Put_Line ("If statement :");
      Put (Generic_Statement (S));
      Put ("Bool expression : ");
      Put (S.Bool_Expression.all);
      New_Line;
      New_Line;
   end Put;

   procedure Put (S : in For_Statement) is

   begin
      Put_Line ("For statement :");
      Put (Generic_Statement (S));
      Put ("For index : ");
      Put (S.For_Index.all);
      New_Line;
      New_Line;
   end Put;

   procedure Put (S : in While_Statement) is

   begin
      Put_Line ("While statement :");
      Put (Generic_Statement (S));
      Put ("Condition : ");
      Put (S.Condition.all);
      New_Line;
      New_Line;
   end Put;

   procedure Put (S : in Affectation_Statement) is

   begin
      Put_Line ("Affectation statement :");
      Put (Generic_Statement (S));
      Put ("Variable name/index : ");
      New_Line;
      Put ("Lvalue : ");
      Put (S.Lvalue.all);
      New_Line;
      Put ("Rvalue : ");
      Put (S.Rvalue.all);
      New_Line;
      New_Line;
   end Put;

   procedure Put (S : in Random_Initialize_Statement) is

   begin
      Put_Line ("Random Initialize statement :");
      Put (Generic_Statement (S));
      Put ("Variable name/index : ");
      New_Line;
      Put ("Lvalue : ");
      Put (To_String (S.Lvalue));
      New_Line;
      Put ("Law : ");
      Put (S.Law);
      New_Line;
      Put ("Parameter1 : ");
      Put (S.Parameter1.all);
      New_Line;
      Put ("Parameter2 : ");
      Put (S.Parameter2.all);
      New_Line;
      New_Line;
   end Put;

   procedure Put (S : in Return_Statement) is

   begin
      Put_Line ("Return statement :");
      Put (Generic_Statement (S));
      Put ("Return value: ");
      Put (S.Return_Value.all);
      New_Line;
      New_Line;
   end Put;

   procedure Put (S : in Set_Statement) is

   begin
      Put_Line ("Set statement :");
      Put (Generic_Statement (S));
      Put ("Set value: ");
      Put (S.Set_Value.all);
      New_Line;
      New_Line;
   end Put;

   procedure Put (S : in Set_Statement_Ptr) is
   begin
      Put (S.all);
   end Put;

   --------------------------------------------------------------

   function Copy (S : in Statement_Ptr) return Statement_Ptr is
      New_If_Statement                : If_Statement_Ptr;
      New_For_Statement               : For_Statement_Ptr;
      New_Return_Statement            : Return_Statement_Ptr;
      New_Affectation_Statement       : Affectation_Statement_Ptr;
      New_Put_Statement               : Put_Statement_Ptr;
      New_Nop_Statement               : Nop_Statement_Ptr;
      New_While_Statement             : While_Statement_Ptr;
      New_Set_Statement               : Set_Statement_Ptr;
      New_Subprogram_Statement        : Subprogram_Statement_Ptr;
      New_Subprogram_Call_Statement   : Subprogram_Call_Statement_Ptr;
      New_Random_Initialize_Statement : Random_Initialize_Statement_Ptr;

      Ptr : Statement_Ptr;

   begin

      case S.Statement_Type is
      when Affectation_Statement_Type =>
         New_Affectation_Statement     := new Affectation_Statement;
         New_Affectation_Statement.all := Affectation_Statement_Ptr (S).all;
         Ptr                           :=
           Statement_Ptr (New_Affectation_Statement);
      when For_Statement_Type =>
         New_For_Statement     := new For_Statement;
         New_For_Statement.all := For_Statement_Ptr (S).all;
         Ptr                   := Statement_Ptr (New_For_Statement);
      when While_Statement_Type =>
         New_While_Statement     := new While_Statement;
         New_While_Statement.all := While_Statement_Ptr (S).all;
         Ptr                     := Statement_Ptr (New_While_Statement);
      when If_Statement_Type =>
         New_If_Statement     := new If_Statement;
         New_If_Statement.all := If_Statement_Ptr (S).all;
         Ptr                  := Statement_Ptr (New_If_Statement);
      when Return_Statement_Type =>
         New_Return_Statement     := new Return_Statement;
         New_Return_Statement.all := Return_Statement_Ptr (S).all;
         Ptr                      := Statement_Ptr (New_Return_Statement);
      when Put_Statement_Type =>
         New_Put_Statement     := new Put_Statement;
         New_Put_Statement.all := Put_Statement_Ptr (S).all;
         Ptr                   := Statement_Ptr (New_Put_Statement);
      when Nop_Statement_Type =>
         New_Nop_Statement     := new Nop_Statement;
         New_Nop_Statement.all := Nop_Statement_Ptr (S).all;
         Ptr                   := Statement_Ptr (New_Nop_Statement);
      when Set_Statement_Type =>
         New_Set_Statement     := new Set_Statement;
         New_Set_Statement.all := Set_Statement_Ptr (S).all;
         Ptr                   := Statement_Ptr (New_Set_Statement);
      when Random_Initialize_Statement_Type =>
         New_Random_Initialize_Statement     :=
           new Random_Initialize_Statement;
         New_Random_Initialize_Statement.all :=
           Random_Initialize_Statement_Ptr (S).all;
         Ptr                                 :=
           Statement_Ptr (New_Random_Initialize_Statement);
      when Subprogram_Statement_Type =>
         New_Subprogram_Statement     := new Subprogram_Statement;
         New_Subprogram_Statement.all := Subprogram_Statement_Ptr (S).all;
         Ptr                          :=
           Statement_Ptr (New_Subprogram_Statement);
      when Subprogram_Call_Statement_Type =>
         New_Subprogram_Call_Statement     := new Subprogram_Call_Statement;
         New_Subprogram_Call_Statement.all :=
           Subprogram_Call_Statement_Ptr (S).all;
         Ptr                               :=
           Statement_Ptr (New_Subprogram_Call_Statement);
      end case;

      return Ptr;
   end Copy;

   --------------------------------------------------------------

   procedure If_Dispatch
     (Current         : in If_Statement_Ptr;
      Current_Section : in Sections;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr)
   is

      Simu_Ptr : Simulation_Value_Ptr;

   begin

      Simu_Ptr := Value_Of (Current.Bool_Expression.all, Variables_Table);

      if Simu_Ptr.Ptype /= Simulation_Boolean then
         Raise_Exception
           (Type_Error'Identity,
            "If statement, line " &
            Next.Line_Number'Img &
            To_String
               (Lb_Comma & Lb_Uncompatible_Type_Error (Current_Language)));
      end if;

      if Simu_Ptr.Boolean_Value then
         Dispatch
           (Current.Then_Statement,
            Current_Section,
            Variables_Table,
            Msg);
      else
         Dispatch
           (Current.Else_Statement,
            Current_Section,
            Variables_Table,
            Msg);
      end if;

      Free (Simu_Ptr);

      Next := Current.Next_Statement;

   exception
      when Constraint_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Storage_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Program_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);

   end If_Dispatch;

   procedure Random_Initialize_Dispatch
     (Current         : in Random_Initialize_Statement_Ptr;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr)
   is

      Var        : Variables_Range;
      Val1, Val2 : Simulation_Value_Ptr;
      Seed       : Generator;

   begin

      Var :=
         Find_Variable (Variables_Table, Current.Lvalue, Current.Line_Number);

      if Current.Law = Uniform_Law then
         Val1 := Value_Of (Current.Parameter1.all, Variables_Table);
         Val2 := Value_Of (Current.Parameter2.all, Variables_Table);
      else
         Val1 := Value_Of (Current.Parameter1.all, Variables_Table);
         Val2 := new Simulation_Value (Simulation_Integer);
      end if;

      if Val1.Ptype /= Simulation_Integer then
         Raise_Exception
           (Type_Error'Identity,
            "Random  statement, line " &
            Next.Line_Number'Img &
            To_String
               (Lb_Comma & Lb_Uncompatible_Type_Error (Current_Language)));
      end if;
      if Val2.Ptype /= Simulation_Integer then
         Raise_Exception
           (Type_Error'Identity,
            "Random  statement, line " &
            Next.Line_Number'Img &
            To_String
               (Lb_Comma & Lb_Uncompatible_Type_Error (Current_Language)));
      end if;

      -- Store random law parameters  in simulation data
      -- and initialize seed
      --
      if (Variables_Table.Entries (Var).Simulation.Ptype =
          Simulation_Array_Random)
      then
         for I in  Simulations_Range loop
            Reset (Seed);
            Save
              (Seed,
               Variables_Table.Entries (Var).Simulation.Random_Table_Value (I))
;
         end loop;
         Variables_Table.Entries (Var).Simulation.Random_Table_Law_Parameters.
           Law := Current.Law;
         Variables_Table.Entries (Var).Simulation.Random_Table_Law_Parameters.
           Parameter1 := Val1.Integer_Value;
         Variables_Table.Entries (Var).Simulation.Random_Table_Law_Parameters.
           Parameter2 := Val2.Integer_Value;

      else
         if (Variables_Table.Entries (Var).Simulation.Ptype =
             Simulation_Time_Unit_Array_Random)
         then
            for I in  Time_Unit_Range loop
               Reset (Seed);
               Save
                 (Seed,
                  Variables_Table.Entries (Var).Simulation.
                 Random_Time_Unit_Table_Value (I));
            end loop;
            Variables_Table.Entries (Var).Simulation.
              Random_Time_Unit_Table_Law_Parameters.Law := Current.Law;
            Variables_Table.Entries (Var).Simulation.
              Random_Time_Unit_Table_Law_Parameters.Parameter1 :=
              Val1.Integer_Value;
            Variables_Table.Entries (Var).Simulation.
              Random_Time_Unit_Table_Law_Parameters.Parameter2 :=
              Val2.Integer_Value;

         else
            Reset (Seed);
            Save
              (Seed,
               Variables_Table.Entries (Var).Simulation.Random_Value);
            Variables_Table.Entries (Var).Simulation.Random_Law_Parameters.Law
               := Current.Law;
            Variables_Table.Entries (Var).Simulation.Random_Law_Parameters.
              Parameter1 := Val1.Integer_Value;
            Variables_Table.Entries (Var).Simulation.Random_Law_Parameters.
              Parameter2 := Val2.Integer_Value;
         end if;
      end if;

      Free (Val1);
      Free (Val2);

      Next := Current.Next_Statement;

   exception
      when Constraint_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Storage_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Program_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);

   end Random_Initialize_Dispatch;

   procedure Affectation_Dispatch
     (Current         : in Affectation_Statement_Ptr;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr)
   is

      Left_Var      : Variables_Range;
      Fully_Indexed : Boolean := False;
      Val, Index    : Simulation_Value_Ptr;

   begin

      if Current.Lvalue.Expression_Type = Array_Variable_Type then
         if Array_Variable_Expr_Ptr (Current.Lvalue).Array_Index /=
            null
         then
            Fully_Indexed := True;
         end if;
      end if;

      -- Affectation from a scalar to another scalar
      --
      if not Fully_Indexed then

         Left_Var :=
            Find_Variable
              (Variables_Table,
               Variable_Expr_Ptr (Current.Lvalue).Name,
               Current.Line_Number);
         Val      := Value_Of (Current.Rvalue.all, Variables_Table);

         -- We can not initialize the left part when its type is
         -- scalar and when the right part type is vectorial
         -- The only exception is when the operator is a min_to_index
         -- or a max_to_index operator
         --

         if ((Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Integer) or
             (Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Boolean) or
             (Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Random) or
             (Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_String) or
             (Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Double)) and
            ((Val.Ptype = Simulation_Array_Boolean) or
             (Val.Ptype = Simulation_Array_Integer) or
             (Val.Ptype = Simulation_Array_Double) or
             (Val.Ptype = Simulation_Array_String) or
             (Val.Ptype = Simulation_Array_Random))
         then
            Raise_Exception
              (Type_Error'Identity,
               "Affectation statement, line " &
               Current.Line_Number'Img &
               To_String (Lb_Comma & Lb_Variable_Size (Current_Language)));
         end if;

         if ((Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Array_Integer) or
             (Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Array_Boolean) or
             (Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Array_Random) or
             (Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Array_String) or
             (Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Array_Double)) and
            ((Val.Ptype = Simulation_Time_Unit_Array_Double) or
             (Val.Ptype = Simulation_Time_Unit_Array_Random) or
             (Val.Ptype = Simulation_Time_Unit_Array_Double) or
             (Val.Ptype = Simulation_Time_Unit_Array_String) or
             (Val.Ptype = Simulation_Time_Unit_Array_Integer))
         then
            Raise_Exception
              (Type_Error'Identity,
               "Affectation statement, line " &
               Current.Line_Number'Img &
               To_String (Lb_Comma & Lb_Variable_Size (Current_Language)));
         end if;

         -- Start to compute expression evaluation
         --

         if (Variables_Table.Entries (Left_Var).Simulation.Ptype =
             Simulation_Array_Integer) and
            (Val.Ptype = Simulation_Integer)
         then
            for I in  Simulations_Range loop
               Variables_Table.Entries (Left_Var).Simulation.
                 Integer_Table_Value (I) := Val.Integer_Value;
            end loop;

         else
            if (Variables_Table.Entries (Left_Var).Simulation.Ptype =
                Simulation_Time_Unit_Array_Integer) and
               (Val.Ptype = Simulation_Integer)
            then
               for I in  Time_Unit_Range loop
                  Variables_Table.Entries (Left_Var).Simulation.
                    Integer_Time_Unit_Table_Value (I) := Val.Integer_Value;
               end loop;

            else
               if (Variables_Table.Entries (Left_Var).Simulation.Ptype =
                   Simulation_Time_Unit_Array_Double) and
                  (Val.Ptype = Simulation_Double)
               then
                  for I in  Time_Unit_Range loop
                     Variables_Table.Entries (Left_Var).Simulation.
                       Double_Time_Unit_Table_Value (I) := Val.Double_Value;
                  end loop;

               else
                  if (Variables_Table.Entries (Left_Var).Simulation.Ptype =
                      Simulation_Time_Unit_Array_Boolean) and
                     (Val.Ptype = Simulation_Boolean)
                  then
                     for I in  Time_Unit_Range loop
                        Variables_Table.Entries (Left_Var).Simulation.
                          Boolean_Time_Unit_Table_Value (I) :=
                          Val.Boolean_Value;
                     end loop;

                  else
                     if (Variables_Table.Entries (Left_Var).Simulation.Ptype =
                         Simulation_Time_Unit_Array_String) and
                        (Val.Ptype = Simulation_String)
                     then
                        for I in  Time_Unit_Range loop
                           Variables_Table.Entries (Left_Var).Simulation.
                             String_Time_Unit_Table_Value (I) :=
                             Val.String_Value;
                        end loop;

                     else
                        if (Variables_Table.Entries (Left_Var).Simulation.
                             Ptype =
                            Simulation_Array_Boolean) and
                           (Val.Ptype = Simulation_Boolean)
                        then
                           for I in  Simulations_Range loop
                              Variables_Table.Entries (Left_Var).Simulation.
                                Boolean_Table_Value (I) := Val.Boolean_Value;
                           end loop;

                        else
                           if (Variables_Table.Entries (Left_Var).Simulation.
                                Ptype =
                               Simulation_Array_Double) and
                              (Val.Ptype = Simulation_Double)
                           then
                              for I in  Simulations_Range loop
                                 Variables_Table.Entries (Left_Var).Simulation.
                                   Double_Table_Value (I) := Val.Double_Value;
                              end loop;

                           else
                              if (Variables_Table.Entries (Left_Var).
                                   Simulation.Ptype =
                                  Simulation_Array_String) and
                                 (Val.Ptype = Simulation_String)
                              then
                                 for I in  Simulations_Range loop
                                    Variables_Table.Entries (Left_Var).
                                      Simulation.String_Table_Value (I) :=
                                      Val.String_Value;
                                 end loop;

                              else
                                 Variables_Table.Entries (Left_Var).Simulation
                                    :=
                                    Value_Of
                                      (Current.Rvalue.all,
                                       Variables_Table);
                              end if;
                           end if;
                        end if;
                     end if;
                  end if;
               end if;
            end if;
         end if;

         Free (Val);

      else

         -- Indexed affectation : the left side AND the right side are indexed
         --

         Index :=
            Value_Of
              (Array_Variable_Expr_Ptr (Current.Lvalue).Array_Index.all,
               Variables_Table);

         if Index.Ptype /= Simulation_Integer then
            Raise_Exception
              (Type_Error'Identity,
               "Affectation statement, line " &
               Current.Line_Number'Img &
               To_String (Lb_Comma & Lb_Index_Error (Current_Language)));
         end if;

         Left_Var :=
            Find_Variable
              (Variables_Table,
               Variable_Expr_Ptr (Current.Lvalue).Name,
               Current.Line_Number);

         Val := Value_Of (Current.Rvalue.all, Variables_Table);

         if (Val.Ptype /= Simulation_Integer) and
            (Val.Ptype /= Simulation_Boolean) and
            (Val.Ptype /= Simulation_Double) and
            (Val.Ptype /= Simulation_String)
         then
            Raise_Exception
              (Type_Error'Identity,
               "Affectation statement, line " &
               Current.Line_Number'Img &
               ", rvalue " &
               To_String (Lb_Uncompatible_Type_Error (Current_Language)));
         end if;

         if ((Val.Ptype = Simulation_Integer) and
             (Variables_Table.Entries (Left_Var).Simulation.Ptype =
              Simulation_Array_Integer))
         then
            Variables_Table.Entries (Left_Var).Simulation.Integer_Table_Value (
              Simulations_Range (Index.Integer_Value)) := Val.Integer_Value;

         else
            if ((Val.Ptype = Simulation_Double) and
                (Variables_Table.Entries (Left_Var).Simulation.Ptype =
                 Simulation_Array_Double))
            then
               Variables_Table.Entries (Left_Var).Simulation.
                 Double_Table_Value (Simulations_Range (Index.Integer_Value))
                  := Val.Double_Value;
            else
               if ((Val.Ptype = Simulation_Boolean) and
                   (Variables_Table.Entries (Left_Var).Simulation.Ptype =
                    Simulation_Array_Boolean))
               then

                  Variables_Table.Entries (Left_Var).Simulation.
                    Boolean_Table_Value (Simulations_Range (
                    Index.Integer_Value)) := Val.Boolean_Value;

               else
                  if ((Val.Ptype = Simulation_String) and
                      (Variables_Table.Entries (Left_Var).Simulation.Ptype =
                       Simulation_Array_String))
                  then

                     Variables_Table.Entries (Left_Var).Simulation.
                       String_Table_Value (Simulations_Range (
                       Index.Integer_Value)) := Val.String_Value;

                  else
                     if ((Val.Ptype = Simulation_Integer) and
                         (Variables_Table.Entries (Left_Var).Simulation.Ptype =
                          Simulation_Time_Unit_Array_Integer))
                     then
                        Variables_Table.Entries (Left_Var).Simulation.
                          Integer_Time_Unit_Table_Value (Time_Unit_Range (
                          Index.Integer_Value)) := Val.Integer_Value;

                     else
                        if ((Val.Ptype = Simulation_Double) and
                            (Variables_Table.Entries (Left_Var).Simulation.
                             Ptype =
                             Simulation_Time_Unit_Array_Double))
                        then
                           Variables_Table.Entries (Left_Var).Simulation.
                             Double_Time_Unit_Table_Value (Time_Unit_Range (
                             Index.Integer_Value)) := Val.Double_Value;
                        else
                           if ((Val.Ptype = Simulation_Boolean) and
                               (Variables_Table.Entries (Left_Var).Simulation.
                                Ptype =
                                Simulation_Time_Unit_Array_Boolean))
                           then

                              Variables_Table.Entries (Left_Var).Simulation.
                                Boolean_Time_Unit_Table_Value (Time_Unit_Range(
                                Index.Integer_Value)) := Val.Boolean_Value;

                           else
                              if ((Val.Ptype = Simulation_String) and
                                  (Variables_Table.Entries (Left_Var).
                                   Simulation.Ptype =
                                   Simulation_Time_Unit_Array_String))
                              then

                                 Variables_Table.Entries (Left_Var).Simulation.
                                   String_Time_Unit_Table_Value (
                                   Time_Unit_Range (Index.Integer_Value)) :=
                                   Val.String_Value;
                              else
                                 Raise_Exception
                                   (Type_Error'Identity,
                                    "Affectation statement, line " &
                                    Current.Line_Number'Img &
                                    To_String
                                       (Lb_Comma &
                                        Lb_Uncompatible_Type_Error (
                                   Current_Language)));
                              end if;
                           end if;
                        end if;
                     end if;
                  end if;
               end if;
            end if;
         end if;

         Free (Index);
         Free (Val);

      end if;

      Next := Current.Next_Statement;

   exception
      when Constraint_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Storage_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Program_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);

   end Affectation_Dispatch;

   procedure Put_Dispatch
     (Current         : in Put_Statement_Ptr;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr)
   is

      Expr_Simu_Ptr : Simulation_Value_Ptr;
      From_Simu_Ptr : Simulation_Value_Ptr;
      To_Simu_Ptr   : Simulation_Value_Ptr;

   begin

      if (Current.Put_From /= null) then
         From_Simu_Ptr := Value_Of (Current.Put_From.all, Variables_Table);
      end if;
      if (Current.Put_To /= null) then
         To_Simu_Ptr := Value_Of (Current.Put_To.all, Variables_Table);
      end if;
      Expr_Simu_Ptr :=
         Value_Of (Current.Expression_To_Be_Displayed.all, Variables_Table);

      if (Current.Put_To /= null) then
         if (To_Simu_Ptr.Ptype /= Simulation_Integer) or
            (From_Simu_Ptr.Ptype /= Simulation_Integer)
         then
            Raise_Exception
              (Type_Error'Identity,
               "Put parameter should be an integer, line " &
               Current.Line_Number'Img &
               To_String
                  (Lb_Comma & Lb_Uncompatible_Type_Error (Current_Language)));
         end if;
      end if;

      Msg := Msg &
             To_Unbounded_String ("- Line " & Current.Line_Number'Img & ", ");

      if (Current.Put_To /= null) and (Current.Put_From /= null) then
         Msg :=
           Msg &
           To_Unbounded_String (Current.Expression_To_Be_Displayed.all) &
           "=" &
           To_Unbounded_String
              (Expr_Simu_Ptr,
               From_Simu_Ptr.Integer_Value,
               To_Simu_Ptr.Integer_Value) &
           Unbounded_Lf;

         Free (From_Simu_Ptr);
         Free (To_Simu_Ptr);

      else
         if ((Expr_Simu_Ptr.Ptype = Simulation_Time_Unit_Array_Boolean) or
             (Expr_Simu_Ptr.Ptype = Simulation_Time_Unit_Array_Integer) or
             (Expr_Simu_Ptr.Ptype = Simulation_Time_Unit_Array_Double) or
             (Expr_Simu_Ptr.Ptype = Simulation_Time_Unit_Array_String))
         then
            Msg :=
              Msg &
              To_Unbounded_String (Current.Expression_To_Be_Displayed.all) &
              "=" &
              To_Unbounded_String
                 (Expr_Simu_Ptr,
                  Integer (Time_Unit_Range'First),
                  Integer (Time_Unit_Range'Last)) &
              Unbounded_Lf;
         else
            if ((Expr_Simu_Ptr.Ptype = Simulation_Array_Boolean) or
                (Expr_Simu_Ptr.Ptype = Simulation_Array_Integer) or
                (Expr_Simu_Ptr.Ptype = Simulation_Array_String) or
                (Expr_Simu_Ptr.Ptype = Simulation_Array_Double))
            then
               Msg :=
                 Msg &
                 To_Unbounded_String (Current.Expression_To_Be_Displayed.all) &
                 "=" &
                 To_Unbounded_String
                    (Expr_Simu_Ptr,
                     Integer (Simulations_Range'First),
                     Integer (Simulations_Range'Last)) &
                 Unbounded_Lf;
            else
               Msg := Msg &
                      To_Unbounded_String
                         (Current.Expression_To_Be_Displayed.all) &
                      "=" &
                      To_Unbounded_String (Expr_Simu_Ptr) &
                      Unbounded_Lf;
            end if;
         end if;
      end if;

      Free (Expr_Simu_Ptr);

      Next := Current.Next_Statement;

   exception
      when Constraint_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Storage_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Program_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);

   end Put_Dispatch;

   procedure While_Dispatch
     (Current         : in While_Statement_Ptr;
      Current_Section : in Sections;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr)
   is

      Simu_Ptr : Simulation_Value_Ptr;

   begin

      -- Let's run included statements :
      -- do a loop iteration
      --
      loop

         Simu_Ptr := Value_Of (Current.Condition.all, Variables_Table);
         if Simu_Ptr.Ptype /= Simulation_Boolean then
            Raise_Exception
              (Type_Error'Identity,
               "while statement, line " &
               Current.Line_Number'Img &
               ", " &
               To_String (Lb_Uncompatible_Type_Error (Current_Language)));
         end if;

         exit when Simu_Ptr.Boolean_Value = False;

         Dispatch
           (Current.Included_Statement,
            Current_Section,
            Variables_Table,
            Msg);

      end loop;

      Free (Simu_Ptr);

      Next := Statement_Ptr (Current.Next_Statement);

   exception
      when Constraint_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Storage_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Program_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);

   end While_Dispatch;

   procedure For_Dispatch
     (Current         : in For_Statement_Ptr;
      Current_Section : in Sections;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr)
   is

      Var            : Variables_Range;
      Var_Loop_Index : Variables_Range;

   begin

      -- Test is the expression included in the for
      -- statement is a variable ... raise a exception
      -- in the other cases
      --
      if Current.For_Index.Expression_Type = Variable_Type then
         if Current.For_Index.Variable_Type /= Simulation_Integer then
            Raise_Exception
              (Type_Error'Identity,
               "for statement, line " &
               Current.Line_Number'Img &
               ", for expression have to be an integer variable");
         end if;
      else
         Raise_Exception
           (Type_Error'Identity,
            "for statement, line " &
            Current.Line_Number'Img &
            ", for expression have to be an integer variable");
      end if;

      -- Let's run included statements :
      -- do a loop iteration
      --

      case Current.For_Type is
         when Task_Table_Type =>
            Var_Loop_Index := Find_Variable (Variables_Table, "nb_tasks");
         when Processor_Table_Type =>
            Var_Loop_Index :=
               Find_Variable (Variables_Table, "nb_processors");
         when Buffer_Table_Type =>
            Var_Loop_Index := Find_Variable (Variables_Table, "nb_buffers");
         when Time_Unit_Table_Type =>
            Var_Loop_Index :=
               Find_Variable (Variables_Table, "nb_time_units");
         when Resource_Table_Type =>
            Var_Loop_Index := Find_Variable (Variables_Table, "nb_resources");
         when Message_Table_Type =>
            Var_Loop_Index := Find_Variable (Variables_Table, "nb_messages");
      end case;

      for I in  0 .. Natural (Variables_Table.Entries (Var_Loop_Index).
           Simulation.Integer_Value -
                              1)
      loop

         Var                                                    :=
            Find_Variable
              (Variables_Table,
               Variable_Expr_Ptr (Current.For_Index).Name,
               Current.Line_Number);
         Variables_Table.Entries (Var).Simulation.Integer_Value :=
           Integer (I);

         Dispatch
           (Current.Included_Statement,
            Current_Section,
            Variables_Table,
            Msg);

      end loop;

      Next := Statement_Ptr (Current.Next_Statement);

   exception
      when Constraint_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Storage_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);
      when Program_Error =>
         Raise_Exception
           (Statement_Error'Identity,
            "line " &
            Current.Line_Number'Img &
            " ; Exception raised :" &
            Exception_Name &
            ":" &
            Exception_Message);

   end For_Dispatch;

   procedure Dispatch
     (Current         : in Statement_Ptr;
      Current_Section : in Sections;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String)
   is

      Next       : Statement_Ptr;
      My_Current : Statement_Ptr := Current;

   begin

      if My_Current /= null then
         loop

         -- Run the associated statement
         --
            case My_Current.Statement_Type is

            when Put_Statement_Type =>
               Put_Dispatch
                 (Put_Statement_Ptr (My_Current),
                  Variables_Table,
                  Msg,
                  Next);

            when Nop_Statement_Type =>
               Next := My_Current.Next_Statement;

            when If_Statement_Type =>
               If_Dispatch
                 (If_Statement_Ptr (My_Current),
                  Current_Section,
                  Variables_Table,
                  Msg,
                  Next);

            when Random_Initialize_Statement_Type =>
               Random_Initialize_Dispatch
                 (Random_Initialize_Statement_Ptr (My_Current),
                  Variables_Table,
                  Msg,
                  Next);

            when Affectation_Statement_Type =>
               Affectation_Dispatch
                 (Affectation_Statement_Ptr (My_Current),
                  Variables_Table,
                  Msg,
                  Next);

            when For_Statement_Type =>
               For_Dispatch
                 (For_Statement_Ptr (My_Current),
                  Current_Section,
                  Variables_Table,
                  Msg,
                  Next);

            when Return_Statement_Type =>
               -- Check if a return statement is called
               --
               if (Return_Statement_Ptr (My_Current).Return_Value /= null) and
                  (Current_Section /= Election)
               then
                  Raise_Exception
                    (Type_Error'Identity,
                     Current_Section'Img &
                     " section, line " &
                     Current.Line_Number'Img &
                     ", invalid return value from return statement");
               end if;
               exit;

            when While_Statement_Type =>
               While_Dispatch
                 (While_Statement_Ptr (My_Current),
                  Current_Section,
                  Variables_Table,
                  Msg,
                  Next);

            when Set_Statement_Type =>
               null;

            when Subprogram_Statement_Type =>
               null;

            when Subprogram_Call_Statement_Type =>
               null;

            end case;

            -- Last statement
            --
            exit when Next = null;

            My_Current := Next;

         end loop;
      end if;

   end Dispatch;

end Statements;
