------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Expressions;           use Expressions;
with Ada.Finalization;
with Tables;
with Framework_Config;      use Framework_Config;
with Simulations;           use Simulations;

package Statements is

   ------------------------------------------------------------------------
   -- Definition of all kind of statement we can find in parametric code
   ------------------------------------------------------------------------

   type Table_Types is (
      Task_Table_Type,
      Message_Table_Type,
      Processor_Table_Type,
      Buffer_Table_Type,
      Time_Unit_Table_Type,
      Resource_Table_Type);

   type Generic_Statement;
   type Statement_Ptr is access all Generic_Statement'Class;

   type Statements_Type is (
      Put_Statement_Type,
      Nop_Statement_Type,
      If_Statement_Type,
      Affectation_Statement_Type,
      For_Statement_Type,
      Return_Statement_Type,
      While_Statement_Type,
      Set_Statement_Type,
      Subprogram_Statement_Type,
      Subprogram_Call_Statement_Type,
      Random_Initialize_Statement_Type);

   type Generic_Statement is abstract new Ada.Finalization.Controlled with
      record
         Statement_Type : Statements_Type;
         Line_Number    : Natural       := 1;
         Next_Statement : Statement_Ptr := null;
      end record;

   procedure Put (S : in Generic_Statement);
   procedure Recursive_Put (S : in Statement_Ptr);

   function Copy (S : in Statement_Ptr) return Statement_Ptr;

   ------------------------------------------

   type Nop_Statement is new Generic_Statement with null record;

   type Nop_Statement_Ptr is access all Nop_Statement'Class;

   procedure Initialize (S : in out Nop_Statement);

   procedure Put (S : in Nop_Statement);

   ------------------------------------------

   type Put_Statement is new Generic_Statement with record
      Expression_To_Be_Displayed : Expression_Ptr;
      Put_From                   : Expression_Ptr;
      Put_To                     : Expression_Ptr;
   end record;

   type Put_Statement_Ptr is access all Put_Statement'Class;

   procedure Initialize (S : in out Put_Statement);

   procedure Put (S : in Put_Statement);

   ------------------------------------------

   type If_Statement is new Generic_Statement with record
      Bool_Expression : Expression_Ptr;
      Else_Statement  : Statement_Ptr;
      Then_Statement  : Statement_Ptr;
   end record;

   type If_Statement_Ptr is access all If_Statement'Class;

   procedure Initialize (S : in out If_Statement);

   procedure Put (S : in If_Statement);

   ------------------------------------------

   type Affectation_Statement is new Generic_Statement with record
      Lvalue : Expression_Ptr;
      Rvalue : Expression_Ptr;
   end record;

   type Affectation_Statement_Ptr is access all Affectation_Statement'Class;

   procedure Initialize (S : in out Affectation_Statement);

   procedure Put (S : in Affectation_Statement);

   ------------------------------------------

   type For_Statement is new Generic_Statement with record
      For_Type           : Table_Types;
      Included_Statement : Statement_Ptr;
      For_Index          : Variable_Expr_Ptr;
   end record;

   type For_Statement_Ptr is access all For_Statement'Class;

   procedure Initialize (S : in out For_Statement);

   procedure Put (S : in For_Statement);

   ------------------------------------------

   type Return_Statement is new Generic_Statement with record
      Return_Value : Expression_Ptr;
   end record;

   type Return_Statement_Ptr is access all Return_Statement'Class;

   procedure Initialize (S : in out Return_Statement);

   procedure Put (S : in Return_Statement);

   ------------------------------------------

   type While_Statement is new Generic_Statement with record
      Included_Statement : Statement_Ptr;
      Condition          : Expression_Ptr;
   end record;

   type While_Statement_Ptr is access all While_Statement'Class;

   procedure Initialize (S : in out While_Statement);

   procedure Put (S : in While_Statement);

   ------------------------------------------

   type Random_Initialize_Statement is new Generic_Statement with record
      Lvalue     : Unbounded_String;
      Law        : Law_Type;
      Parameter1 : Expression_Ptr;
      Parameter2 : Expression_Ptr;
   end record;

   type Random_Initialize_Statement_Ptr is access all 
     Random_Initialize_Statement'Class;

   procedure Initialize (S : in out Random_Initialize_Statement);

   procedure Put (S : in Random_Initialize_Statement);

   ------------------------------------------

   type Set_Statement is new Generic_Statement with record
      Set_Value : Expression_Ptr;
      Set_Id    : Unbounded_String;
   end record;

   type Set_Statement_Ptr is access all Set_Statement'Class;

   procedure Initialize (S : in out Set_Statement);

   procedure Put (S : in Set_Statement);

   procedure Put (S : in Set_Statement_Ptr);

   ------------------------------------------

   type Subprogram_Statement is new Generic_Statement with record
      Included_Statement : Statement_Ptr;
      Is_A_Function      : Boolean;
      Subprogram_Name    : Unbounded_String;
   end record;

   type Subprogram_Statement_Ptr is access all Subprogram_Statement'Class;

   procedure Initialize (S : in out Subprogram_Statement);

   procedure Put (S : in Subprogram_Statement);

   ------------------------------------------

   type Subprogram_Call_Statement is new Generic_Statement with record
      Is_A_Function     : Boolean;
      Called_Subprogram : Statement_Ptr;
      Return_Value      : Expression_Ptr;
   end record;

   type Subprogram_Call_Statement_Ptr is access all Subprogram_Call_Statement'
     Class;

   procedure Initialize (S : in out Subprogram_Call_Statement);

   procedure Put (S : in Subprogram_Call_Statement);

   ---------------------------------------------------------
   -- Table to store the list of sets statements
   ---------------------------------------------------------

   package Sets_Type_Package is new Tables (
      Set_Statement_Ptr,
      Max_Statements,
      Put);
   use Sets_Type_Package;

   subtype Sets_Range is Sets_Type_Package.Table_Range;
   subtype Sets_Table_Type is Sets_Type_Package.Table;

   -- Looking for a set statement in the table
   --
   function Find_Set
     (Var  : in Sets_Table_Type;
      S    : in Unbounded_String)
      return Sets_Range;

   procedure Check_Set_Declaration
     (Var : in Sets_Table_Type;
      S   : in Unbounded_String);

   ------------------------------------------------------------------------
   -- Parametric codes are sorted in different sections. Each section is
   --associated
   -- to a parametric functionnality
   ------------------------------------------------------------------------

   type Sections is (
      Start,
      Priority,
      Election,
      Activation,
      Check_Resource,
      Allocate_Resource,
      Release_Resource,
      Gather_Event_Analyzer,
      Display_Event_Analyzer,
      Subprogram);

   type Section_Pointers is array (Sections'Range) of Statement_Ptr;

   ---------------------------------------------------------------------
   --- Functions and procedure in this part of the package run
   --  parametric scheduler statements
   ----------------------------------------------------------------------

   procedure Dispatch
     (Current         : in Statement_Ptr;
      Current_Section : in Sections;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String);

   procedure If_Dispatch
     (Current         : in If_Statement_Ptr;
      Current_Section : in Sections;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr);

   procedure Affectation_Dispatch
     (Current         : in Affectation_Statement_Ptr;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr);

   procedure Random_Initialize_Dispatch
     (Current         : in Random_Initialize_Statement_Ptr;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr);

   procedure Put_Dispatch
     (Current         : in Put_Statement_Ptr;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr);

   procedure While_Dispatch
     (Current         : in While_Statement_Ptr;
      Current_Section : in Sections;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr);

   procedure For_Dispatch
     (Current         : in For_Statement_Ptr;
      Current_Section : in Sections;
      Variables_Table : in out Variables_Table_Type;
      Msg             : in out Unbounded_String;
      Next            : in out Statement_Ptr);

end Statements;
