------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Exceptions;                                use Ada.Exceptions;
with Scheduler.user_defined;                        use Scheduler.user_defined;
with Scheduler.user_defined.pipeline;               use 
  Scheduler.user_defined.pipeline;
with Scheduler.user_defined.automata;               use 
  Scheduler.user_defined.automata;
with Scheduler.user_defined.compiled;               use 
  Scheduler.user_defined.compiled;
with Scheduler.Fixed_Priority.Dm;                   use 
  Scheduler.Fixed_Priority.Dm;
with Scheduler.Fixed_Priority.Rm;                   use 
  Scheduler.Fixed_Priority.Rm;
with Scheduler.Fixed_Priority.Hpf;                  use 
  Scheduler.Fixed_Priority.Hpf;
with Scheduler.Dynamic_Priority.Edf;                use 
  Scheduler.Dynamic_Priority.Edf;
with Scheduler.Dynamic_Priority.Llf;                use 
  Scheduler.Dynamic_Priority.Llf;
with Scheduler.Dynamic_Priority.Muf.Deadline_Based;
use Scheduler.Dynamic_Priority.Muf.Deadline_Based;
with Scheduler.Dynamic_Priority.Muf.Laxity_Based;
use Scheduler.Dynamic_Priority.Muf.Laxity_Based;
with Scheduler.Dynamic_Priority.D_Over;             use 
  Scheduler.Dynamic_Priority.D_Over;
with Scheduler.Time_Sharing_Based_On_Wait_Time;
use Scheduler.Time_Sharing_Based_On_Wait_Time;
with Scheduler.Time_Sharing_Based_On_Cpu_Usage;
use Scheduler.Time_Sharing_Based_On_Cpu_Usage;
with Scheduler.Round_Robin;                         use Scheduler.Round_Robin;
with Scheduler.Hierarchical;                        use Scheduler.Hierarchical;
with Scheduler.Hierarchical.Round_Robin;
use Scheduler.Hierarchical.Round_Robin;
with Processors.extended;                           use Processors.extended;
with Unbounded_Strings;                             use Unbounded_Strings;
with Translate;                                     use Translate;
with Objects;                                       use Objects;
with Objects.extended;                              use Objects.extended;

package body Processor_Set is

   procedure Check_Processor
     (My_Processors : in Processors_Set;
      Name          : in Unbounded_String;
      Is_Preemptive : in Preemptive_Type;
      Quantum       : in Integer;
      File_Name     : in Unbounded_String;
      A_Scheduler   : in Scheduler_Type;
      A_Network     : in Unbounded_String)
   is

   begin

      if (Name = "") then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Processor_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

      if not Is_A_Valid_Identifier (Name) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Processor (Current_Language) &
                Name &
                " : " &
                Lb_Processor_Name (Current_Language) &
                Lb_Colon &
                Lb_Invalid_Identifier (Current_Language)));
      end if;

      if ((A_Scheduler = pipeline_User_Defined_Protocol) or
          (A_Scheduler = pipeline_User_Defined_Protocol) or
          (A_Scheduler = pipeline_User_Defined_Protocol)) and
         (File_Name = "")
      then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Processor (Current_Language) &
                Name &
                " : " &
                Lb_File_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

      if (File_Name /= "") and
         (A_Scheduler /= pipeline_User_Defined_Protocol) and
         (A_Scheduler /= Automata_User_Defined_Protocol) and
         (A_Scheduler /= Compiled_User_Defined_Protocol)
      then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Processor (Current_Language) &
                Name &
                " : " &
                Lb_File_Name_Control (Current_Language)));
      end if;

      if (Quantum /= 0) and
         (A_Scheduler /= Posix_1003_Highest_Priority_First_Protocol) and
         (A_Scheduler /= Round_Robin_Protocol) and
         (A_Scheduler /= Hierarchical_Round_Robin_Protocol)
      then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Processor (Current_Language) &
                Name &
                " : " &
                Lb_Quantum_Control (Current_Language)));
      end if;

      if (Quantum < 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Processor (Current_Language) &
                Name &
                " : " &
                "Quantum" &
                Lb_Must_Be (Current_Language) &
                To_Unbounded_String (" >= 0")));
      end if;

      if (A_Scheduler = No_Scheduling_Protocol) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Processor (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Invalid_Scheduler (Current_Language)));
      end if;

   end Check_Processor;

   procedure Add_Processor
     (My_Processors : in out Processors_Set;
      A_Processor   : in out Processor_Ptr;
      Name          : in Unbounded_String;
      Is_Preemptive : in Preemptive_Type;
      Quantum       : in Integer;
      File_Name     : in Unbounded_String;
      A_Scheduler   : in Scheduler_Type;
      A_Network     : in Unbounded_String)
   is

      My_Iterator : Iterator;

   begin

      Check_Processor
        (My_Processors,
         Name,
         Is_Preemptive,
         Quantum,
         File_Name,
         A_Scheduler,
         A_Network);

      if (Get_Number_Of_Elements (My_Processors) > 0) then

         Reset_Iterator (My_Processors, My_Iterator);

         loop
            Current_Element (My_Processors, A_Processor, My_Iterator);
            if (Name = A_Processor.Name) then
               Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_Processor (Current_Language) &
                      " " &
                      Name &
                      " : " &
                      Lb_Processor_Name (Current_Language) &
                      Lb_Already_Defined (Current_Language)));
            end if;

            exit when Is_Last_Element (My_Processors, My_Iterator);

            Next_Element (My_Processors, My_Iterator);
         end loop;
      end if;

      A_Processor      := new extended_Processor;
      A_Processor.Name := To_Unbounded_String (To_String (Name));
      case A_Scheduler is

      when Rate_Monotonic_Protocol =>
         A_Processor.Scheduler := new Rm_Scheduler;
      when Deadline_Monotonic_Protocol =>
         A_Processor.Scheduler := new Dm_Scheduler;
      when Posix_1003_Highest_Priority_First_Protocol =>
         A_Processor.Scheduler := new Hpf_Scheduler;
      when Earliest_Deadline_First_Protocol =>
         A_Processor.Scheduler := new Edf_Scheduler;
      when Least_Laxity_First_Protocol =>
         A_Processor.Scheduler := new Llf_Scheduler;
      when Round_Robin_Protocol =>
         A_Processor.Scheduler := new Round_Robin_Scheduler;
      when Maximum_Urgency_First_Based_On_Deadline_Protocol =>
         A_Processor.Scheduler := new Deadline_Muf_Scheduler;
      when Maximum_Urgency_First_Based_On_Laxity_Protocol =>
         A_Processor.Scheduler := new Laxity_Muf_Scheduler;
      when Time_Sharing_Based_On_Wait_Time_Protocol =>
         A_Processor.Scheduler :=
           new Time_Sharing_Based_On_Wait_Time_Scheduler;
      when Time_Sharing_Based_On_Cpu_Usage_Protocol =>
         A_Processor.Scheduler :=
           new Time_Sharing_Based_On_Cpu_Usage_Scheduler;
      when D_Over_Protocol =>
         A_Processor.Scheduler := new D_Over_Scheduler;
      when pipeline_User_Defined_Protocol =>
         A_Processor.Scheduler := new pipeline_user_defined_Scheduler;
      when Automata_User_Defined_Protocol =>
         A_Processor.Scheduler := new automata_user_defined_Scheduler;
      when Compiled_User_Defined_Protocol =>
         A_Processor.Scheduler := new compiled_user_defined_Scheduler;

      when Hierarchical_Cyclic_Protocol =>
         A_Processor.Scheduler := new Rm_Scheduler;
      when Hierarchical_Round_Robin_Protocol =>
         A_Processor.Scheduler := new Hierarchical_Round_Robin_Scheduler;
      when Hierarchical_Fixed_Priority_Protocol =>
         A_Processor.Scheduler := new Rm_Scheduler;

      when User_Defined_Protocol =>
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Processor (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Invalid_Scheduler (Current_Language)));
      when No_Scheduling_Protocol =>
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Processor (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Invalid_Scheduler (Current_Language)));
      end case;

      A_Processor.Network := A_Network;

      if (A_Scheduler = Automata_User_Defined_Protocol) then
         Set_Behavior_File_Name
           (automata_user_defined_Scheduler (A_Processor.Scheduler.all),
            File_Name);
      end if;
      if (A_Scheduler = Compiled_User_Defined_Protocol) then
         Set_Behavior_File_Name
           (compiled_user_defined_Scheduler (A_Processor.Scheduler.all),
            File_Name);
      end if;
      if (A_Scheduler = pipeline_User_Defined_Protocol) then
         Set_Behavior_File_Name
           (pipeline_user_defined_Scheduler (A_Processor.Scheduler.all),
            File_Name);
      end if;

      Set_Quantum (A_Processor.Scheduler.all, Quantum);
      Set_Preemptive (A_Processor.Scheduler.all, Is_Preemptive);

      Add (My_Processors, A_Processor);

   exception
      when Full_Set =>
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String (Lb_Can_Not_Define_More_Processors (Current_Language)));

   end Add_Processor;

   procedure Add_Processor
     (My_Processors : in out Processors_Set;
      Name          : in Unbounded_String;
      Is_Preemptive : in Preemptive_Type;
      Quantum       : in Integer;
      File_Name     : in Unbounded_String;
      A_Scheduler   : in Scheduler_Type;
      A_Network     : in Unbounded_String)
   is

      Dummy : Processor_Ptr;

   begin
      Add_Processor
        (My_Processors,
         Dummy,
         Name,
         Is_Preemptive,
         Quantum,
         File_Name,
         A_Scheduler,
         A_Network);
   end Add_Processor;

   function Get
     (My_Processors  : in Processors_Set;
      Processor_Name : in Unbounded_String;
      Param_Name     : in Processor_Parameters)
      return           Scheduler_Ptr
   is
      A_Processor : Processor_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Scheduler) then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Processors, My_Iterator);

      loop
         Current_Element (My_Processors, A_Processor, My_Iterator);

         if (A_Processor.Name = Processor_Name) then
            exit;
         end if;

         exit when Is_Last_Element (My_Processors, My_Iterator);

         Next_Element (My_Processors, My_Iterator);

      end loop;

      return A_Processor.Scheduler;

   end Get;

   procedure Set
     (My_Processors  : in out Processors_Set;
      Processor_Name : in Unbounded_String;
      Param_Name     : in Processor_Parameters;
      Param_Value    : in Scheduler_Ptr)
   is

      A_Processor : Processor_Ptr;
      My_Iterator : Iterator;
   begin

      if (Param_Name /= Scheduler) then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Processors, My_Iterator);

      loop
         Current_Element (My_Processors, A_Processor, My_Iterator);

         if (A_Processor.Name = Processor_Name) then
            A_Processor.Scheduler := Param_Value;
            exit;
         end if;

         exit when Is_Last_Element (My_Processors, My_Iterator);

         Next_Element (My_Processors, My_Iterator);

      end loop;
   end Set;

   function Search_Processor
     (My_Processors : in Processors_Set;
      Name          : in Unbounded_String)
      return          Processor_Ptr
   is
      My_Iterator : Iterator;
      A_Processor : Processor_Ptr;
      Result      : Processor_Ptr;

      Found : Boolean := False;

   begin

      Reset_Iterator (My_Processors, My_Iterator);

      loop
         Current_Element (My_Processors, A_Processor, My_Iterator);

         if (A_Processor.Name = Name) then
            Found  := True;
            Result := A_Processor;
         end if;

         exit when Is_Last_Element (My_Processors, My_Iterator);

         Next_Element (My_Processors, My_Iterator);

      end loop;

      if not Found then
         Raise_Exception
           (Processor_Not_Found'Identity,
            To_String (Lb_Processor_Name (Current_Language) & "=" & Name));
      end if;

      return Result;

   end Search_Processor;

   function Export_Xml
     (My_Processors : in Processors_Set)
      return          Unbounded_String
   is
      A_Processor : Processor_Ptr;
      My_Iterator : Processors_Iterator;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Processors) then

         Result := Result &
                   To_Unbounded_String ("<processors>") &
                   Unbounded_Lf;
         Reset_Iterator (My_Processors, My_Iterator);

         loop
            Current_Element (My_Processors, A_Processor, My_Iterator);

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "<processor>") &
                      Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<name>" &
                  To_String (A_Processor.Name) &
                  "</name>") &
              Unbounded_Lf;

            Result := Result & Export_Xml (A_Processor.Scheduler.all);

            if A_Processor.Network /= Empty_String then
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "<network_link> " &
                     To_String (A_Processor.Network) &
                     " </network_link>") &
                 Unbounded_Lf;
            end if;
            Result := Result &
                      To_Unbounded_String (ASCII.HT & "</processor>") &
                      Unbounded_Lf;

            exit when Is_Last_Element (My_Processors, My_Iterator);

            Next_Element (My_Processors, My_Iterator);

         end loop;
         Result := Result &
                   To_Unbounded_String ("</processors>") &
                   Unbounded_Lf;
      end if;

      return Result;

   end Export_Xml;

   function Export_Aadl_Implementations
     (My_Processors : in Processors_Set)
      return          Unbounded_String
   is
      My_Iterator : Iterator;
      A_Processor : Processor_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Processors) then

         Reset_Iterator (My_Processors, My_Iterator);

         loop
            Current_Element (My_Processors, A_Processor, My_Iterator);

            Result := Result &
                      To_Unbounded_String
                         ("processor " & To_String (A_Processor.Name)) &
                      Unbounded_Lf;

            if A_Processor.Network /=
               To_Unbounded_String ("No_Network")
            then
               Result := Result &
                         To_Unbounded_String (ASCII.HT & "features") &
                         Unbounded_Lf;
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "bus_connected : requires bus access " &
                     To_String (A_Processor.Network) &
                     ";") &
                 Unbounded_Lf;
            end if;
            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Processor.Name) & ";") &
                      Unbounded_Lf &
                      Unbounded_Lf;

            Result := Result &
                      To_Unbounded_String
                         ("processor implementation " &
                          To_String (A_Processor.Name) &
                          ".Impl") &
                      Unbounded_Lf;
            Result := Result &
                      To_Unbounded_String (ASCII.HT & "properties ") &
                      Unbounded_Lf;

            Result := Result &
                      Export_Aadl_Properties (A_Processor.Scheduler.all, 2);

            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Processor.Name) & ".Impl;") &
                      Unbounded_Lf &
                      Unbounded_Lf;

            exit when Is_Last_Element (My_Processors, My_Iterator);

            Next_Element (My_Processors, My_Iterator);

         end loop;

      end if;

      return Result;

   end Export_Aadl_Implementations;

   function Export_Aadl_Declarations
     (My_Processors : in Processors_Set;
      Number_Of_Ht  : in Natural)
      return          Unbounded_String
   is
      My_Iterator : Iterator;
      A_Processor : Processor_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Processors) then

         Reset_Iterator (My_Processors, My_Iterator);

         loop
            Current_Element (My_Processors, A_Processor, My_Iterator);

            for I in  1 .. Number_Of_Ht loop
               Result := Result & ASCII.HT;
            end loop;
            Result :=
              Result &
              To_Unbounded_String
                 ("instancied_" &
                  To_String (A_Processor.Name) &
                  " : processor " &
                  To_String (A_Processor.Name) &
                  ".Impl;") &
              Unbounded_Lf;

            exit when Is_Last_Element (My_Processors, My_Iterator);

            Next_Element (My_Processors, My_Iterator);

         end loop;

      end if;

      return Result;

   end Export_Aadl_Declarations;

end Processor_Set;
