------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developped by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Numerics.Aux;          use Ada.Numerics.Aux;
with Indexed_Tables;
with Natural_Util;
with Double_Util;               use Double_Util;
with Text_IO;                   use Text_IO;
with Ada.Strings.Unbounded;     use Ada.Strings.Unbounded;
with Convert_Strings;
with Convert_Unbounded_Strings;
with Ada.Finalization;

package Queueing_System is

   Not_Implemented : exception;

   -- different type of queueing sytem
   -- (there are not all implemented)
   --
   type Queueing_Systems_Type is (
      Qs_Pp1,
      Qs_Mm1,
      Qs_Md1,
      Qs_Mp1,
      Qs_Mg1,
      Qs_Mms,
      Qs_Mds,
      Qs_Mps,
      Qs_Mgs,
      Qs_Mm1n,
      Qs_Md1n,
      Qs_Mp1n,
      Qs_Mg1n,
      Qs_Mmsn,
      Qs_Mdsn,
      Qs_Mpsn,
      Qs_Mgsn);

   procedure To_Qs is new Convert_Strings (Queueing_Systems_Type, Qs_Mm1);
   procedure To_Qs is new Convert_Unbounded_Strings (
      Queueing_Systems_Type,
      Qs_Mm1);

   package qs_io is new Text_IO.Enumeration_IO (Queueing_Systems_Type);
   use qs_io;

   function Qs_To_Display_String
     (A_Qs : in Queueing_Systems_Type)
      return Unbounded_String;
   procedure Display_String_To_Qs
     (From : in Unbounded_String;
      To   : out Queueing_Systems_Type;
      Ok   : out Boolean);

   Arrival_Rate_Table_Size       : constant Natural := 100;
   Service_Rate_Table_Size       : constant Natural := 100;
   Resp_Time_Table_Size          : constant Natural := 1000;
   Resp_Time_Consumer_Table_Size : constant Natural := 1000;

   package A_Arrival_Rate is new Indexed_Tables (
      Double,
      Natural,
      Arrival_Rate_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   use A_Arrival_Rate;

   package A_Service_Rate is new Indexed_Tables (
      Double,
      Natural,
      Service_Rate_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   use A_Service_Rate;

   package A_Resp_Time is new Indexed_Tables (
      Double,
      Natural,
      Resp_Time_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   use A_Resp_Time;

   package A_Resp_Time_Consumer is new Indexed_Tables (
      Double,
      Natural,
      Resp_Time_Consumer_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   use A_Resp_Time_Consumer;

   subtype Arrival_Rate_Table is A_Arrival_Rate.Indexed_Table;
   subtype Arrival_Rate_Range is A_Arrival_Rate.Indexed_Table_Range;

   subtype Service_Rate_Table is A_Service_Rate.Indexed_Table;
   subtype Service_Rate_Range is A_Service_Rate.Indexed_Table_Range;

   subtype Resp_Time_Consumer_Table is A_Resp_Time_Consumer.Indexed_Table;
   subtype Resp_Time_Consumer_Table_Ptr is
     A_Resp_Time_Consumer.Indexed_Table_Ptr;
   subtype Resp_Time_Consumer_Range is
     A_Resp_Time_Consumer.Indexed_Table_Range;

   subtype Resp_Time_Table is A_Resp_Time.Indexed_Table;
   subtype Resp_Time_Table_Ptr is A_Resp_Time.Indexed_Table_Ptr;
   subtype Resp_Time_Range is A_Resp_Time.Indexed_Table_Range;

   type Generic_Queueing_System is abstract new Ada.Finalization.Controlled
   with
      record
         Queueing_System_Type : Queueing_Systems_Type;
         Arrival_Rate         : Arrival_Rate_Table;
         Service_Rate         : Service_Rate_Table;
         Nb_Of_Messages       : Double;

         Var_Service : Double; -- variance service time for mg1
         Avg_Service : Double; -- average service time for mg1

         Waiting_Time : Double;
         Utilisation  : Double;
         Max_State    : Natural;

         Nb_Arrival : Double;  -- nb of producer or arrival flow
         Nb_Server  : Double;  -- nb of server or consumer
         Harmonic   : Boolean; -- for PP1 qs

         -- mp1 criteria computation
         --
         Consumer_Resp_Time     : Resp_Time_Consumer_Table_Ptr;
         Consumer_Resp_Time_Cpt : Resp_Time_Consumer_Range;
      end record;

   type Queueing_System_Ptr is access all Generic_Queueing_System'Class;

   procedure Initialize
     (A_Queueing_System : in out Generic_Queueing_System);

   -- Initialize a queueing system with a constant consumer response time
   --
   procedure Set_Constant_Consumer_Response_Time
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Response_Time     : in Double);

   -- reset queueing system values
   --
   procedure Reset (A_Queueing_System : in out Generic_Queueing_System);

   -- display all queueing system caracteristics
   --
   procedure Put (A_Queueing_System : in Generic_Queueing_System);

   -- get or set queueing system caracteristics
   --
   procedure Set_Qs_Arrival_Rate
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Value             : in Double);

   procedure Set_Qs_Arrival_Rate
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Place             : in Arrival_Rate_Range;
      Value             : in Double);

   function Get_Qs_Arrival_Rate
     (A_Queueing_System : in Generic_Queueing_System'Class;
      Place             : in Arrival_Rate_Range)
      return              Double;

   procedure Set_Qs_Service_Rate
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Value             : in Double);

   procedure Set_Qs_Service_Rate
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Place             : in Service_Rate_Range;
      Value             : in Double);

   function Get_Qs_Service_Rate
     (A_Queueing_System : in Generic_Queueing_System'Class;
      Place             : in Service_Rate_Range)
      return              Double;

   procedure Set_Qs_Utilisation
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Value             : Double);

   function Get_Qs_Utilisation
     (A_Queueing_System : in Generic_Queueing_System'Class)
      return              Double;

   procedure Set_Qs_Type
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Qs_Type           : in Queueing_Systems_Type);

   function Get_Qs_Type
     (A_Queueing_System : in Generic_Queueing_System'Class)
      return              Queueing_Systems_Type;

   procedure Set_Qs_Nb_Of_Messages
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Value             : Double);

   function Get_Qs_Nb_Of_Messages
     (A_Queueing_System : in Generic_Queueing_System'Class)
      return              Double;

   procedure Set_Qs_Waiting_Time
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Value             : Double);

   function Get_Qs_Waiting_Time
     (A_Queueing_System : in Generic_Queueing_System'Class)
      return              Double;

   procedure Set_Qs_Max_State
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Value             : Natural);

   function Get_Qs_Max_State
     (A_Queueing_System : in Generic_Queueing_System'Class)
      return              Natural;

   procedure Set_Qs_Nb_Arrival
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Value             : Double);

   function Get_Qs_Nb_Arrival
     (A_Queueing_System : in Generic_Queueing_System'Class)
      return              Double;

   procedure Set_Qs_Nb_Server
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Value             : Double);

   function Get_Qs_Nb_Server
     (A_Queueing_System : in Generic_Queueing_System'Class)
      return              Double;

   procedure Set_Qs_Harmonic
     (A_Queueing_System : in out Generic_Queueing_System'Class;
      Value             : Boolean);

   function Get_Qs_Harmonic
     (A_Queueing_System : in Generic_Queueing_System'Class)
      return              Boolean;

   -- return the name of the queueing system
   --
   function Get_Name
     (A_Queueing_System : in Generic_Queueing_System'Class)
      return              Unbounded_String;

   -- set queueing system caracteristics
   -- needed, for instance, if you want theoretical results after a simulation
   --
   procedure Set_Qs_Basic_Parameters
     (A_Queueing_System        : in out Generic_Queueing_System'Class;
      A_Source_Queueing_System : in Generic_Queueing_System);

end Queueing_System;
