------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Framework_Config;
USE Framework_Config;
WITH Resources;
USE Resources;
USE Resources.Task_Lists;
WITH Sets;


PACKAGE Resource_Set IS




   ----------------------------------------------------------
   -- Definition of a shared resource set
   ----------------------------------------------------------



   PACKAGE Generic_Resource_Set IS NEW Sets (
      Max_Element => Framework_Config.Max_Resources,
      Element     => Generic_Resource_Ptr,
      Free        => Free,
      Copy        => Copy,
      Put         => Put);

   USE Generic_Resource_Set;

   TYPE Resources_Set IS NEW Generic_Resource_Set.Set WITH PRIVATE;


   SUBTYPE Resources_Range IS Generic_Resource_Set.Element_Range;
   SUBTYPE Resources_Iterator IS Generic_Resource_Set.Iterator;



   ----------------------------------------------------------
   -- Exceptions defined in the resource set package
   ----------------------------------------------------------

   -- Try to get/set a unkwon parameter
   --
   Invalid_Parameter : EXCEPTION;

   -- The the resource name does ot exist in the set
   --
   Resource_Not_Found : EXCEPTION;

   -- all the resource  accessed by a task should have
   -- the same protocol type
   --
   Can_Not_Used_Different_Protocol : EXCEPTION;



   ----------------------------------------------------------
   -- I/O operations
   ----------------------------------------------------------
   FUNCTION Export_Xml (
         My_Resources : IN     Resources_Set)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_Implementations (
         My_Resources : IN     Resources_Set)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_Declarations (
         My_Resources       : IN     Resources_Set;
         Address_Space_Name : IN     Unbounded_String;
         Number_Of_Ht       : IN     Natural)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_Connections (
         My_Resources : IN     Resources_Set;
         A_Task_Name  : IN     Unbounded_String;
         Number_Of_Ht : IN     Natural)
     RETURN Unbounded_String;




   ----------------------------------------------------------
   -- Procedure to perform modification on a set
   ----------------------------------------------------------



   PROCEDURE Task_Begin_End (
         Task_Name      : IN     Unbounded_String;
         Resource_Name  : IN     Unbounded_String;
         A_Resource_Set : IN     Resources_Set;
         Starts_Ends    : IN OUT Start_End_Table;
         Elements       : IN OUT Task_Lists_Range);


   PROCEDURE Update_Resource (
         My_Resources       : IN OUT Resources_Set;
         Name               : IN     Unbounded_String;
         State              : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Protocol           : IN     Resources_Type;
         Affected_Tasks     : IN     Task_Lists_table);


   PROCEDURE Check_Resource (
         My_Resources       : IN     Resources_Set;
         Name               : IN     Unbounded_String;
         State              : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Protocol           : IN     Resources_Type);



   PROCEDURE Add_Resource (
         My_Resources       : IN OUT Resources_Set;
         Name               : IN     Unbounded_String;
         State              : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Protocol           : IN     Resources_Type;
         Affected_Tasks     : IN     Task_Lists_table);


   PROCEDURE Add_Resource (
         My_Resources       : IN OUT Resources_Set;
         A_Resource         : IN OUT Generic_Resource_Ptr;
         Name               : IN     Unbounded_String;
         State              : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Protocol           : IN     Resources_Type;
         Affected_Tasks     : IN     Task_Lists_table);


   PROCEDURE Delete_Task (
         My_Resources : IN OUT Resources_Set;
         A_Task       : IN     Unbounded_String);


   PROCEDURE Delete_Address_Space (
         My_Resources : IN OUT Resources_Set;
         A_Addr       : IN     Unbounded_String);


   PROCEDURE Delete_Processor (
         My_Resources : IN OUT Resources_Set;
         A_Processor  : IN     Unbounded_String);




   ------------------------------------------------------
   -- Read information from a set
   ------------------------------------------------------


   -- Get the number of tasks hosted by the
   -- processor "Processor_Name"
   --
   FUNCTION Get_Number_Of_Resource_From_Processor (
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Resources_Range;


   -- Get the task pointer of the task named "Name"
   --
   FUNCTION Search_Resource (
         My_Resources : IN     Resources_Set;
         Name         : IN     Unbounded_String)
     RETURN Generic_Resource_Ptr;



   --------------------------------------------------------
   -- Functions to perform sort on a set
   --------------------------------------------------------



   PROCEDURE Same_Protocol_Control (
         My_Resources : IN     Resources_Set;
         A_Task       : IN     Unbounded_String);





PRIVATE

   TYPE Resources_Set IS NEW Generic_Resource_Set.Set WITH NULL RECORD;

END Resource_Set;




