------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Xml_Tag;           use Xml_Tag;
with Double_Util;       use Double_Util;
with Translate;         use Translate;
with Ada.Numerics.Aux;  use Ada.Numerics.Aux;
with Unbounded_Strings; use Unbounded_Strings;

package body Scheduler.Dynamic_Priority.Llf is

   procedure Initialize (A_Scheduler : in out Llf_Scheduler) is
   begin
      Reset (A_Scheduler);
      A_Scheduler.Name := Least_Laxity_First_Protocol;
   end Initialize;

   function Copy (A_Scheduler : in Llf_Scheduler) return Scheduler_Ptr is
      Ptr : Llf_Scheduler_Ptr;

   begin

      Ptr := new Llf_Scheduler;

      Ptr.Name               := A_Scheduler.Name;
      Ptr.Is_Preemptive      := A_Scheduler.Is_Preemptive;
      Ptr.Quantum            := A_Scheduler.Quantum;
      Ptr.Previously_Elected := A_Scheduler.Previously_Elected;

      return Scheduler_Ptr (Ptr);

   end Copy;

   procedure Do_Election
     (My_Scheduler      : in out Llf_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Msg               : in out Unbounded_String;
      Current_Time      : in Natural;
      Processor_Name    : in Unbounded_String;
      My_Dependencies   : in Tasks_Dependencies_Ptr;
      With_Offsets      : in Boolean;
      With_Precedencies : in Boolean;
      With_Resources    : in Boolean;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table;
      Elected           : in out Tasks_Range;
      No_Task           : in out Boolean)
   is

      Smallest_Laxity : Natural     := Natural'Last;
      Laxity          : Natural;
      I               : Tasks_Range := 0;
   begin

      loop
         if (Si.Tcbs (I).Tsk.Cpu_Name = Processor_Name) then

            Laxity :=
              Dynamic_Priority_Tcb_Ptr (Si.Tcbs (I)).Dynamic_Deadline -
              Si.Tcbs (I).Rest_Of_Capacity;

            if (Si.Tcbs (I).Wake_Up_Time <= Current_Time) and
               (Laxity < Smallest_Laxity) and
               (Si.Tcbs (I).Rest_Of_Capacity /= 0)
            then
               if (With_Offsets = False) or
                  Check_Offset (Si.Tcbs (I), Current_Time)
               then
                  if (With_Precedencies = False) or
                     Check_Precedencies
                        (Si,
                         My_Dependencies,
                         Current_Time,
                         Si.Tcbs (I))
                  then

                     Smallest_Laxity := Laxity;
                     Elected         := I;
                  end if;
               end if;
            end if;
         end if;

         I := I + 1;
         exit when Si.Tcbs (I) = null;
      end loop;

      if Smallest_Laxity = Natural'Last then
         No_Task := True;
      else
         No_Task := False;

         -- last time unit : compute dynamic deadline for periodic task
         --
         if (Si.Tcbs (Elected).Rest_Of_Capacity = 1) and
            (Si.Tcbs (Elected).Tsk.Task_Type = Periodic_Type)
         then
            Dynamic_Priority_Tcb_Ptr (Si.Tcbs (Elected)).Dynamic_Deadline  :=
              Dynamic_Priority_Tcb_Ptr (Si.Tcbs (Elected)).Dynamic_Deadline +
              Periodic_Task_Ptr (Si.Tcbs (Elected).Tsk).Period;
         end if;

      end if;

   end Do_Election;

   procedure Utilization_Factor_Feasibility_Test
     (My_Scheduler   : in Llf_Scheduler;
      My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String;
      Result         : in out Unbounded_String)
   is

      Bound  : Double           := 0.0;
      Utildi : Double           := 0.0;
      Msg    : Unbounded_String := Empty_String;

   begin

      case Get_Preemptive (My_Scheduler) is

         when Preemptive =>

            Utildi :=
               Processor_Utilization_Over_Deadline (My_Tasks, Processor_Name);

            Bound_On_Processor_Utilization
              (My_Scheduler,
               My_Tasks,
               Processor_Name,
               Bound,
               Msg);

            Result := Result & Lb_Utilization_Bound1 (Current_Language);
            Result := Result & Lb_Llf (Current_Language);

            if Utildi <= Bound then
               Result := Result & Lb_Sched_Explanation1 (Current_Language);
               Result := Result &
                         Start_Fact_Bound &
                         Format (Utildi) &
                         End_Fact_Bound;
               Result := Result & Lb_Sched_Explanation12 (Current_Language);
               Result := Result &
                         Start_Fact_Bound &
                         Format (Bound) &
                         End_Fact_Bound;
            else
               Result := Result & Lb_Sched_Explanation2 (Current_Language);
               Result := Result &
                         Start_Fact_Bound &
                         Format (Utildi) &
                         End_Fact_Bound;
               Result := Result & Lb_Sched_Explanation22 (Current_Language);
               Result := Result &
                         Start_Fact_Bound &
                         Format (Bound) &
                         End_Fact_Bound;
            end if;
            Result :=
              Result &
              Start_Ref &
              To_Unbounded_String (" (") &
              Lb_See (Current_Language) &
              To_Unbounded_String ("[7]).") &
              End_Ref &
              Unbounded_Lf;

         when Not_Preemptive =>

            Result := Result &
                      Lb_Pb_Sched_Unknown (Current_Language) &
                      Unbounded_Lf;

      end case;

   end Utilization_Factor_Feasibility_Test;

end Scheduler.Dynamic_Priority.Llf;
