------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

package body Scheduler.Dynamic_Priority.Muf.Deadline_Based is

   procedure Initialize (A_Scheduler : in out Deadline_Muf_Scheduler) is
   begin
      Reset (A_Scheduler);
      A_Scheduler.Name := Maximum_Urgency_First_Based_On_Deadline_Protocol;
   end Initialize;

   function Copy
     (A_Scheduler : in Deadline_Muf_Scheduler)
      return        Scheduler_Ptr
   is
      Ptr : Deadline_Muf_Scheduler_Ptr;

   begin

      Ptr := new Deadline_Muf_Scheduler;

      Ptr.Name               := A_Scheduler.Name;
      Ptr.Is_Preemptive      := A_Scheduler.Is_Preemptive;
      Ptr.Quantum            := A_Scheduler.Quantum;
      Ptr.Previously_Elected := A_Scheduler.Previously_Elected;

      return Scheduler_Ptr (Ptr);

   end Copy;

   procedure Do_Election
     (My_Scheduler      : in out Deadline_Muf_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Msg               : in out Unbounded_String;
      Current_Time      : in Natural;
      Processor_Name    : in Unbounded_String;
      My_Dependencies   : in Tasks_Dependencies_Ptr;
      With_Offsets      : in Boolean;
      With_Precedencies : in Boolean;
      With_Resources    : in Boolean;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table;
      Elected           : in out Tasks_Range;
      No_Task           : in out Boolean)
   is

      Smallest_Deadline : Natural     := Natural'Last;
      I                 : Tasks_Range := 0;
      Almost_One_Task   : Boolean     := False;

   begin

      -- Find the first ready task and assume we'll run this one
      --
      loop
         if (Si.Tcbs (I).Tsk.Cpu_Name = Processor_Name) then

            -- Is the task ready to run ?
            --
            if (Si.Tcbs (I).Wake_Up_Time <= Current_Time) and
               (Si.Tcbs (I).Rest_Of_Capacity /= 0)
            then
               if (With_Offsets = False) or
                  Check_Offset (Si.Tcbs (I), Current_Time)
               then
                  if (With_Precedencies = False) or
                     Check_Precedencies
                        (Si,
                         My_Dependencies,
                         Current_Time,
                         Si.Tcbs (I))
                  then

                     Almost_One_Task   := True;
                     Smallest_Deadline :=
                       Dynamic_Priority_Tcb_Ptr (Si.Tcbs (I)).Dynamic_Deadline;
                     Elected           := I;

                  end if;
               end if;
            end if;
         end if;

         I := I + 1;
         exit when Si.Tcbs (I) = null;
      end loop;

      if not Almost_One_Task then
         No_Task := True;
      else
         No_Task := False;

         -- Let see now if some more urgent/critical task exists
         --
         I := 0;
         loop
            if (Si.Tcbs (I).Tsk.Cpu_Name = Processor_Name) then

               -- Is the task ready to run ?
               --
               if (Si.Tcbs (I).Wake_Up_Time <= Current_Time) and
                  (Si.Tcbs (I).Rest_Of_Capacity /= 0)
               then
                  if (With_Offsets = False) or
                     Check_Offset (Si.Tcbs (I), Current_Time)
                  then
                     if (With_Precedencies = False) or
                        Check_Precedencies
                           (Si,
                            My_Dependencies,
                            Current_Time,
                            Si.Tcbs (I))
                     then

                        -- Select according to the deadline, criticality and
                        --priority
                        --

                        -- First case : we find a more critical task
                        --
                        if (Si.Tcbs (I).Tsk.Criticality >
                            Si.Tcbs (Elected).Tsk.Criticality)
                        then
                           Smallest_Deadline :=
                             Dynamic_Priority_Tcb_Ptr (Si.Tcbs (I)).
                             Dynamic_Deadline;
                           Elected           := I;
                        else

                           -- Second case : we find a more urgent task
                           --
                           if (Si.Tcbs (I).Tsk.Criticality =
                               Si.Tcbs (Elected).Tsk.Criticality) and
                              (Dynamic_Priority_Tcb_Ptr (Si.Tcbs (I)).
                                Dynamic_Deadline <
                               Smallest_Deadline)
                           then
                              Smallest_Deadline :=
                                Dynamic_Priority_Tcb_Ptr (Si.Tcbs (I)).
                                Dynamic_Deadline;
                              Elected           := I;
                           else

                              -- Last case : we find a task with a highiest
                              --priority
                              --
                              if (Si.Tcbs (I).Tsk.Criticality =
                                  Si.Tcbs (Elected).Tsk.Criticality) and
                                 (Dynamic_Priority_Tcb_Ptr (Si.Tcbs (I)).
                                   Dynamic_Deadline =
                                  Smallest_Deadline) and
                                 (Si.Tcbs (I).Tsk.Priority >
                                  Si.Tcbs (Elected).Tsk.Priority)
                              then
                                 Smallest_Deadline :=
                                   Dynamic_Priority_Tcb_Ptr (Si.Tcbs (I)).
                                   Dynamic_Deadline;
                                 Elected           := I;
                              end if;
                           end if;
                        end if;

                     end if;
                  end if;
               end if;
            end if;

            I := I + 1;
            exit when Si.Tcbs (I) = null;
         end loop;

         -- last time unit : compute dynamic deadline for periodic task
         --
         if (Si.Tcbs (Elected).Rest_Of_Capacity = 1) and
            (Si.Tcbs (Elected).Tsk.Task_Type = Periodic_Type)
         then
            Dynamic_Priority_Tcb_Ptr (Si.Tcbs (Elected)).Dynamic_Deadline  :=
              Dynamic_Priority_Tcb_Ptr (Si.Tcbs (Elected)).Dynamic_Deadline +
              Periodic_Task_Ptr (Si.Tcbs (Elected).Tsk).Period;
         end if;

      end if;

   end Do_Election;

end Scheduler.Dynamic_Priority.Muf.Deadline_Based;
