------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Translate;         use Translate;
with Unbounded_Strings; use Unbounded_Strings;

package body Scheduler.Dynamic_Priority is

   function Build_Tcb
     (My_Scheduler : in Dynamic_Priority_Scheduler;
      A_Task       : Generic_Task_Ptr)
      return         Tcb_Ptr
   is
      A_Tcb : Dynamic_Priority_Tcb_Ptr;
   begin
      A_Tcb := new Dynamic_Priority_Tcb;
      Initialize (Tcb (A_Tcb.all), A_Task);
      Initialize (A_Tcb.all);
      return Tcb_Ptr (A_Tcb);
   end Build_Tcb;

   procedure Initialize (A_Tcb : in out Dynamic_Priority_Tcb) is
   begin
      A_Tcb.Dynamic_Deadline := A_Tcb.Tsk.Deadline + A_Tcb.Wake_Up_Time;
   end Initialize;

   procedure Check_Before_Scheduling
     (My_Scheduler   : in Dynamic_Priority_Scheduler;
      My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

   begin
      null;
   end Check_Before_Scheduling;

   procedure Compute_Blocking_Time
     (My_Scheduler   : in Dynamic_Priority_Scheduler;
      My_Tasks       : in Tasks_Set;
      My_Resources   : in Resources_Set;
      Processor_Name : in Unbounded_String;
      Msg            : in out Unbounded_String;
      Blocking_Time  : out Blocking_Time_Table)
   is
   begin
      raise Invalid_Scheduler;
   end Compute_Blocking_Time;

   procedure Bound_On_Processor_Utilization
     (My_Scheduler   : in Dynamic_Priority_Scheduler;
      My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String;
      Result         : in out Double;
      Msg            : in out Unbounded_String)
   is
   begin
      Result := 1.0;
      Msg    := To_Unbounded_String (" (") &
                Lb_See (Current_Language) &
                To_Unbounded_String ("[1], page 8, ") &
                Lb_Theorem (Current_Language) &
                To_Unbounded_String ("2). ") &
                Unbounded_Lf;

   end Bound_On_Processor_Utilization;

   function Compute_Wi_A_T
     (My_Tasks : in Tasks_Set;
      Taski    : in Generic_Task_Ptr;
      A        : in Integer;
      T        : in Double)
      return     Double
   is
      Result   : Double := 0.0;
      Min      : Double := 0.0;
      Iterator : Tasks_Iterator;
      Taskj    : Generic_Task_Ptr;
   begin

      Reset_Iterator (My_Tasks, Iterator);
      loop
         Current_Element (My_Tasks, Taskj, Iterator);

         if ((Taskj.Name /= Taski.Name) and
             (Taskj.Deadline <= A + Taski.Deadline))
         then
            Min := 1.0 +
                   Double'Floor
                      ((Double (A) +
                        Double (Taski.Deadline) -
                        Double (Taskj.Deadline)) /
                       Double (Periodic_Task_Ptr (Taskj).Period));

            if (Double'Ceiling
                   (T / Double (Periodic_Task_Ptr (Taskj).Period)) <
                Min)
            then
               Min :=
                  Double'Ceiling
                    (T / Double (Periodic_Task_Ptr (Taskj).Period));
            end if;
            Result := Result + Min * Double (Taskj.Capacity);
         end if;

         -- exit loop when there is no more task
         --
         exit when Is_Last_Element (My_Tasks, Iterator);
         Next_Element (My_Tasks, Iterator);
      end loop;

      return Result;
   end Compute_Wi_A_T;

   function Compute_Li_A
     (My_Scheduler : in Dynamic_Priority_Scheduler;
      My_Tasks     : in Tasks_Set;
      Taski        : in Generic_Task_Ptr;
      A            : in Integer)
      return         Double
   is
      Li_A     : Double  := 0.1;
      Li_A_1   : Double  := 0.0;
      Iterator : Tasks_Iterator;
      Taskj    : Generic_Task_Ptr;
      Max_Cj   : Natural := 0;
      Tmp      : Double  := 0.0;
      Min      : Double  := 0.0;
      Min1     : Double  := 0.0;
      Min2     : Double  := 0.0;
   begin

      if (My_Scheduler.Is_Preemptive = Preemptive) then
         -- preemptive case
         --
         while Li_A_1 /= Li_A loop

            Li_A := Li_A_1;

            Li_A_1 := Compute_Wi_A_T (My_Tasks, Taski, A, Li_A);

            case Taski.Task_Type is
               when Periodic_Type =>
                  Li_A_1 := Li_A_1 +
                            (1.0 +
                             Double'Floor
                                (Double (A) /
                                 Double (Periodic_Task_Ptr (Taski).Period))) *
                            Double (Taski.Capacity);
               when others =>
                  raise Constraint_Error;
            end case;
         end loop;

      else
         -- non preemptive case
         --
         while Li_A_1 /= Li_A loop

            Li_A := Li_A_1;
            Tmp  := 0.0;
            Reset_Iterator (My_Tasks, Iterator);
            loop
               Current_Element (My_Tasks, Taskj, Iterator);

               if ((A + Taski.Deadline) < Taskj.Deadline) then
                  if (Taskj.Capacity > Max_Cj) then
                     Max_Cj := Taskj.Capacity;
                  end if;
               end if;

               if ((Taskj.Name /= Taski.Name) and
                   (Taskj.Deadline <= A + Taski.Deadline))
               then
                  Min1 := 1.0 +
                          Double'Floor
                             (Li_A /
                              Double (Periodic_Task_Ptr (Taskj).Period));

                  Min2 := 1.0 +
                          Double'Floor
                             (Double (A) +
                              Double (Taski.Deadline) -
                              Double (Taskj.Deadline)) /
                          Double (Periodic_Task_Ptr (Taskj).Period);

                  Min := Min1;
                  if (Min2 < Min) then
                     Min := Min2;
                  end if;

                  Tmp := Tmp +
                         Min * Double (Taskj.Capacity) +
                         Double'Floor
                            (Double (A) /
                             Double (Periodic_Task_Ptr (Taskj).Period)) *
                         Double (Taski.Capacity);
               end if;

               -- exit loop when there is no more task
               --
               exit when Is_Last_Element (My_Tasks, Iterator);
               Next_Element (My_Tasks, Iterator);
            end loop;

            Li_A_1 := Double (Max_Cj) + Tmp;
         end loop;

      end if;

      return Li_A;
   end Compute_Li_A;

   -- This function may present an infinite loop for non-preemptive case
   -- It may come from the upper bound for a: (exit when a >= Li(a))
   --
   procedure Compute_Response_Time
     (My_Scheduler   : in Dynamic_Priority_Scheduler;
      My_Tasks       : in out Tasks_Set;
      Processor_Name : in Unbounded_String;
      Msg            : in out Unbounded_String;
      Response_Time  : out Response_Time_Table)
   is

      Tmp       : Tasks_Set;
      Iterator1 : Tasks_Iterator;
      Taski     : Generic_Task_Ptr;
      Iterator2 : Tasks_Iterator;
      Taskj     : Generic_Task_Ptr;
      I         : Response_Time_Range := 0;

      A : Integer := 0;

      B_Sup : Double  := 0.0;
      Li_A  : Double  := 0.0;
      K     : Natural := 0;

   begin
      Double_Tasks_Parameters.Initialize (Response_Time);

      -- check if tasks are periodics
      --
      Periodic_Control (My_Tasks, Processor_Name);

      -- check start time
      --
      Start_Time_Control (My_Tasks, Processor_Name);

      -- check offset
      --
      Offset_Control (My_Tasks, Processor_Name);

      -- Check processor utilization first
      --
      if (Processor_Utilization_Over_Period (My_Tasks, Processor_Name) >
          1.0)
      then
         raise Processor_Utilization_Exceeded;
      end if;

      -- Select task to proceed
      --
      Current_Processor_Name := Processor_Name;
      Select_And_Copy (My_Tasks, Tmp, Select_Cpu'Access);
      Reset_Iterator (My_Tasks, Iterator1);

      loop
      -- selection of the current task
      --
         Current_Element (Tmp, Taski, Iterator1);

         -- initialize response time for current task
         --
         Response_Time.Entries (I).Data := 0.0;
         Response_Time.Entries (I).Item := Taski;
         Response_Time.Nb_Entries       := Response_Time.Nb_Entries + 1;

         if (My_Scheduler.Is_Preemptive = Preemptive) then
            Response_Time.Entries (I).Data := Double (Taski.Capacity);
         end if;

         -- compute S "on the fly"
         --
         Reset_Iterator (My_Tasks, Iterator2);

         loop
            Current_Element (Tmp, Taskj, Iterator2);

            K := 0;
            loop
               case Taskj.Task_Type is
                  when Periodic_Type =>

                     A := (K * Periodic_Task_Ptr (Taskj).Period) +
                          Taskj.Deadline -
                          Taski.Deadline;

                  when others =>
                     raise Constraint_Error;
               end case;

               Li_A := Compute_Li_A (My_Scheduler, Tmp, Taski, A);

               if ((Li_A - Double (A)) >
                   Response_Time.Entries (I).Data)
               then
                  Response_Time.Entries (I).Data := Li_A - Double (A);

                  -- PROBLEM
                  -- with non preemptive case response_time has no upper bound
                  --in some cases
                  -- Put(Natural(Response_Time.Entries(i).Data)'Img) ;
               end if;

               if (My_Scheduler.Is_Preemptive = Preemptive) then
                  B_Sup := Li_A;
               else
                  B_Sup := Li_A;
               end if;

               exit when (Double (A) >= B_Sup);

               -- PROBLEM : the lines below try yo fix the problem described
               --10 lines
               -- before (see morin2003.doc)
               -- The "exit when K" statement is not a part of the EDF
               --response time
               -- computation : it's just a way to be sure that the response
               -- time computation never loop for ever .... This Exit statement
               -- HAVE TO BE REMOVED BEFORE EDF response time computation will
               --be
               -- strongly tested by users
               --
               --exit when Response_Time.Entries(I).Data > Double(
               --  Taski.Deadline);
               exit when K > 10000;

               K := K + 1;
            end loop;

            exit when Is_Last_Element (Tmp, Iterator2);
            -- go to the next task
            --
            Next_Element (Tmp, Iterator2);
         end loop;

         exit when Is_Last_Element (Tmp, Iterator1);

         -- go to the next task
         --
         Next_Element (Tmp, Iterator1);

         I := I + 1;
      end loop;

   end Compute_Response_Time;

   procedure Specific_Scheduler_Initialization
     (My_Scheduler   : in out Dynamic_Priority_Scheduler;
      Si             : in out Scheduling_Information;
      Processor_Name : in Unbounded_String;
      My_Tasks       : in out Tasks_Set;
      My_Resources   : in out Resources_Set;
      My_Buffers     : in out Buffers_Set;
      My_Messages    : in Messages_Set;
      Msg            : in out Unbounded_String)
   is

   begin
      null;
   end Specific_Scheduler_Initialization;

end Scheduler.Dynamic_Priority;
