------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Unbounded_Strings;
USE Unbounded_Strings;



PACKAGE Scheduler.Fixed_Priority IS


   TYPE Fixed_Priority_Scheduler IS ABSTRACT NEW Generic_Scheduler WITH PRIVATE;


   TYPE Fixed_Priority_Tcb IS NEW Tcb WITH
      RECORD
         Current_Priority : Priority_Range;
         Is_Ready         : Boolean;
      END RECORD;

   TYPE Fixed_Priority_Tcb_Ptr IS ACCESS ALL Fixed_Priority_Tcb'Class;

   PROCEDURE Initialize (
         A_Tcb : IN OUT Fixed_Priority_Tcb);

   ----------------------------------------

   TYPE Fixed_Priority_Resource IS NEW Shared_Resource WITH
      RECORD
         Priority_Ceiling : Priority_Range;
      END RECORD;

   TYPE Fixed_Priority_Resource_Ptr IS ACCESS ALL Fixed_Priority_Resource'Class;

   ----------------------------------------

   FUNCTION Build_Resource (
         My_Scheduler : IN     Fixed_Priority_Scheduler;
         A_Resource   :        Generic_Resource_Ptr)
     RETURN Shared_Resource_Ptr;


   FUNCTION Build_Tcb (
         My_Scheduler : IN     Fixed_Priority_Scheduler;
         A_Task       :        Generic_Task_Ptr)
     RETURN Tcb_Ptr;


   PROCEDURE Do_Election (
         My_Scheduler      : IN OUT Fixed_Priority_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Msg               : IN OUT Unbounded_String;
         Current_Time      : IN     Natural;
         Processor_Name    : IN     Unbounded_String;
         My_Dependencies   : IN     Tasks_Dependencies_Ptr;
         With_Offsets      :        Boolean;
         With_Precedencies :        Boolean;
         With_Resources    :        Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table;
         Elected           : IN OUT Tasks_Range;
         No_Task           : IN OUT Boolean);


   PROCEDURE Compute_Response_Time (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Response_Time  :    OUT Response_Time_Table);


   PROCEDURE Compute_Blocking_Time (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Blocking_Time  :    OUT Blocking_Time_Table);


   PROCEDURE Bound_On_Processor_Utilization (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Double;
         Msg            : IN OUT Unbounded_String);


   PROCEDURE Set_Priority_According_To_Dm (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String         := Empty_String);


   PROCEDURE Set_Priority_According_To_Rm (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String         := Empty_String);


   PROCEDURE Check_Resource (
         My_Scheduler      : IN OUT Fixed_Priority_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             : IN     Tcb_Ptr;
         Is_Ready          :    OUT Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);


   PROCEDURE Allocate_Resource (
         My_Scheduler      : IN OUT Fixed_Priority_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);


   PROCEDURE Release_Resource (
         My_Scheduler      : IN OUT Fixed_Priority_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);


   PROCEDURE Specific_Scheduler_Initialization (
         My_Scheduler   : IN OUT Fixed_Priority_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String;
         My_Tasks       : IN OUT Tasks_Set;
         My_Resources   : IN OUT Resources_Set;
         My_Buffers     : IN OUT Buffers_Set;
         My_Messages    : IN     Messages_Set;
         Msg            : IN OUT Unbounded_String);


   PROCEDURE Compute_Ceiling_Of_Resources (
         My_Scheduler   : IN OUT Fixed_Priority_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String;
         My_Tasks       : IN OUT Tasks_Set;
         My_Resources   : IN OUT Resources_Set);


   PROCEDURE Dispatched_Change_Current_Priority (
         My_Scheduler : IN OUT Fixed_Priority_Scheduler;
         A_Tcb        : IN     Fixed_Priority_Tcb_Ptr;
         New_Priority :        Priority_Range);


   PROCEDURE Change_Current_Priority (
         My_Scheduler : IN OUT Fixed_Priority_Scheduler'Class;
         A_Tcb        : IN     Fixed_Priority_Tcb_Ptr;
         New_Priority :        Priority_Range);


   FUNCTION Produce_Running_Task_Event (
         My_Scheduler : IN     Fixed_Priority_Scheduler;
         A_Task       : IN     Tcb_Ptr)
     RETURN Time_Unit_Event_Ptr;


   FUNCTION Export_Xml_Event_Running_Task (
         My_Scheduler : IN     Fixed_Priority_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;



PRIVATE

   TYPE Fixed_Priority_Scheduler IS ABSTRACT NEW Generic_Scheduler WITH
      RECORD
         Used_Resource : Boolean := False;
      END RECORD;


END Scheduler.Fixed_Priority ;


