------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Io_Tools;                      use Io_Tools;
with Ada.IO_Exceptions;             use Ada.IO_Exceptions;
with Ada.Strings.Unbounded.Text_IO; use Ada.Strings.Unbounded.Text_IO;
with Ada.Exceptions;                use Ada.Exceptions;
with Translate;                     use Translate;
with scheduler_io;                  use scheduler_io;
with Simulations;                   use Simulations;
with Parser;                        use Parser;
with Debug;                         use Debug;
with Parameters;                    use Parameters;
use Parameters.User_Defined_Parameters_Table_Package;

package body Scheduler.user_defined.automata is

   function Export_Aadl_Properties
     (My_Scheduler : in automata_user_defined_Scheduler;
      Number_Of_Ht : in Natural)
      return         Unbounded_String
   is
      Result : Unbounded_String :=
         Export_Aadl_Properties
           (Generic_Scheduler (My_Scheduler),
            Number_Of_Ht);

   begin

      for I in  1 .. Number_Of_Ht loop
         Result := Result & ASCII.HT;
      end loop;

      Result := Result &
                To_Unbounded_String
                   ("Source_Text => """ &
                    To_String (Get_Behavior_File_Name (My_Scheduler)) &
                    """;") &
                Unbounded_Lf;

      return Result;

   end Export_Aadl_Properties;

   function Export_Xml
     (My_Scheduler : in automata_user_defined_Scheduler)
      return         Unbounded_String
   is
      Result        : Unbounded_String         := Empty_String;
      File          : Unbounded_String         := Empty_String;
      Quantum       : constant Natural         := Get_Quantum (My_Scheduler);
      Is_Preemptive : constant Preemptive_Type :=
         Get_Preemptive (My_Scheduler);
      Name          : Unbounded_String         := Get_Name (My_Scheduler);

   begin

      Result := Result &
                To_Unbounded_String (ASCII.HT & ASCII.HT & "<scheduler");

      if Quantum /= 0 then
         Result := Result &
                   To_Unbounded_String (" quantum=""" & Quantum'Img & """");
      end if;

      if Is_Preemptive /= Preemptive then
         Result := Result &
                   To_Unbounded_String
                      (" is_preemptive=""" & Is_Preemptive'Img & """");
      end if;

      File := Get_Behavior_File_Name (My_Scheduler);

      Result := Result &
                To_Unbounded_String
                   (" parametric_filen_name=""" & To_String (File) & """");

      Result := Result & To_Unbounded_String (">");
      Result := Result &
                To_Unbounded_String (" ") &
                Name &
                To_Unbounded_String (" ");
      Result := Result &
                To_Unbounded_String ("</scheduler>") &
                Unbounded_Lf;

      return Result;

   end Export_Xml;

   procedure Initialize
     (A_Scheduler : in out automata_user_defined_Scheduler)
   is
   begin
      Reset (A_Scheduler);
      A_Scheduler.Name            := Automata_User_Defined_Protocol;
      A_Scheduler.String_Behavior := new Unbounded_String;
   end Initialize;

   function Copy
     (A_Scheduler : in automata_user_defined_Scheduler)
      return        Scheduler_Ptr
   is
      Ptr : automata_user_defined_Scheduler_Ptr;

   begin

      Ptr := new automata_user_defined_Scheduler;

      Ptr.Name               := A_Scheduler.Name;
      Ptr.Is_Preemptive      := A_Scheduler.Is_Preemptive;
      Ptr.Quantum            := A_Scheduler.Quantum;
      Ptr.Previously_Elected := A_Scheduler.Previously_Elected;

      Ptr.Behavior_File_Name     := A_Scheduler.Behavior_File_Name;
      Ptr.String_Behavior        := new Unbounded_String;
      Ptr.String_Behavior.all    := A_Scheduler.String_Behavior.all;
      Ptr.Root_Statement_Pointer := A_Scheduler.Root_Statement_Pointer;
      Ptr.Variables_Table        := A_Scheduler.Variables_Table;

      return Scheduler_Ptr (Ptr);

   end Copy;

   function Get_String_Behavior
     (My_Scheduler : in automata_user_defined_Scheduler)
      return         Unbounded_String
   is
   begin
      return My_Scheduler.String_Behavior.all;
   end Get_String_Behavior;

   function Get_String_Behavior
     (My_Scheduler : in automata_user_defined_Scheduler)
      return         Unbounded_String_Ptr
   is
   begin
      return My_Scheduler.String_Behavior;
   end Get_String_Behavior;

   procedure Put (My_Scheduler : in automata_user_defined_Scheduler) is
   begin
      Put (Generic_Scheduler (My_Scheduler));
      Put_Line ("Behavior file  name : " & My_Scheduler.Behavior_File_Name);
      Put_Line ("Behavior file  : " & My_Scheduler.String_Behavior.all);
      Put (My_Scheduler.Root_Statement_Pointer);
      Put (My_Scheduler.Variables_Table);
   end Put;

   procedure Set_Behavior_File_Name
     (My_Scheduler : in out automata_user_defined_Scheduler;
      To_Set       : in Unbounded_String)
   is
   begin
      My_Scheduler.Behavior_File_Name := To_Set;
   end Set_Behavior_File_Name;

   function Get_Behavior_File_Name
     (My_Scheduler : in automata_user_defined_Scheduler)
      return         Unbounded_String
   is
   begin
      return My_Scheduler.Behavior_File_Name;
   end Get_Behavior_File_Name;

   procedure Set_String_Behavior
     (My_Scheduler : in out automata_user_defined_Scheduler;
      To_Set       : in Unbounded_String)
   is
   begin
      My_Scheduler.String_Behavior     := new Unbounded_String;
      My_Scheduler.String_Behavior.all := To_Set;
   end Set_String_Behavior;

   procedure Do_Election
     (My_Scheduler      : in out automata_user_defined_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Msg               : in out Unbounded_String;
      Current_Time      : in Natural;
      Processor_Name    : in Unbounded_String;
      My_Dependencies   : in Tasks_Dependencies_Ptr;
      With_Offsets      : in Boolean;
      With_Precedencies : in Boolean;
      With_Resources    : in Boolean;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table;
      Elected           : in out Tasks_Range;
      No_Task           : in out Boolean)
   is

      Return_Val             : Expression_Ptr;
      Current                : Statement_Ptr;
      Election_Value         : Simulation_Value_Ptr;
      Current_Selected_Value : Simulation_Value_Ptr;
      I                      : Tasks_Range := 0;
      Var                    : Variables_Range;
      Var_Task_Name          : Variables_Range;

   begin

      -- By default, no task are elected : it's a way
      -- to safer schedule a user_defined scheduler without
      -- a correctly designed code (election part of it)
      --
      No_Task := True;

      -- Look for the variable which stores task names
      --
      Var_Task_Name :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("tasks.name"));

      -- We set the variables changed by the simulation each unit of time
      --
      Var :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("tasks.activation_number"));
      for I in  0 .. Si.Number_Of_Tasks - 1 loop
         if Processor_Name = Si.Tcbs (I).Tsk.Cpu_Name then
            for J in  0 .. Si.Number_Of_Tasks - 1 loop
               if 
                    My_Scheduler.Variables_Table.Entries (Var_Task_Name).
                    Simulation.String_Table_Value (Simulations_Range (J)) =
                  Si.Tcbs (I).Tsk.Name
               then
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (J)) :=
                    Integer (Si.Tcbs (I).Activation);
               end if;
            end loop;
         end if;
      end loop;

      for I in  0 .. Si.Number_Of_Tasks - 1 loop
         if 
              My_Scheduler.Variables_Table.Entries (Var_Task_Name).Simulation.
              String_Table_Value (Simulations_Range (I)) =
            Si.Tcbs (My_Scheduler.Previously_Elected).Tsk.Name
         then
            Var :=
               Find_Variable
                 (My_Scheduler.Variables_Table,
                  To_Unbounded_String ("tasks.rest_of_capacity"));
            My_Scheduler.Variables_Table.Entries (Var).Simulation.
              Integer_Table_Value (Simulations_Range (I)) :=
              Integer (
              Si.Tcbs (My_Scheduler.Previously_Elected).Rest_Of_Capacity);

            Var :=
               Find_Variable
                 (My_Scheduler.Variables_Table,
                  To_Unbounded_String ("tasks.used_cpu"));
            My_Scheduler.Variables_Table.Entries (Var).Simulation.
              Integer_Table_Value (Simulations_Range (I)) :=
              Integer (Si.Tcbs (My_Scheduler.Previously_Elected).Used_Cpu);
         end if;
      end loop;

      Var                                                                 :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("previously_elected"));
      My_Scheduler.Variables_Table.Entries (Var).Simulation.Integer_Value :=
        Integer (My_Scheduler.Previously_Elected);

      Var                                                                 :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("simulation_time"));
      My_Scheduler.Variables_Table.Entries (Var).Simulation.Integer_Value :=
        Integer (Current_Time);

      -- Check if tasks are ready or not
      --
      Var :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("tasks.ready"));

      for I in  0 .. Si.Number_Of_Tasks - 1 loop
         if (Si.Tcbs (I).Tsk.Cpu_Name = Processor_Name) then
            for J in  0 .. Si.Number_Of_Tasks - 1 loop
               if 
                    My_Scheduler.Variables_Table.Entries (Var_Task_Name).
                    Simulation.String_Table_Value (Simulations_Range (J)) =
                  Si.Tcbs (I).Tsk.Name
               then
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Boolean_Table_Value (Simulations_Range (J)) := False;
               end if;
            end loop;
         end if;
      end loop;

      for I in  0 .. Si.Number_Of_Tasks - 1 loop
         if (Si.Tcbs (I).Tsk.Cpu_Name = Processor_Name) then
            if (Si.Tcbs (I).Wake_Up_Time <= Current_Time) and
               (Si.Tcbs (I).Rest_Of_Capacity /= 0)
            then
               if (With_Offsets = False) or
                  Check_Offset (Si.Tcbs (I), Current_Time)
               then
                  if (With_Precedencies = False) or
                     Check_Precedencies
                        (Si,
                         My_Dependencies,
                         Current_Time,
                         Si.Tcbs (I))
                  then
                     for J in  0 .. Si.Number_Of_Tasks - 1 loop
                        if 
                             My_Scheduler.Variables_Table.Entries (
                             Var_Task_Name).Simulation.String_Table_Value (
                             Simulations_Range (J)) =
                           Si.Tcbs (I).Tsk.Name
                        then
                           My_Scheduler.Variables_Table.Entries (Var).
                             Simulation.Boolean_Table_Value (Simulations_Range(
                             J)) := True;
                        end if;
                     end loop;
                  end if;
               end if;
            end if;
         end if;
      end loop;

      -- Now, do "priority" section
      --
      Current := My_Scheduler.Root_Statement_Pointer (Priority);
      Dispatch (Current, Priority, My_Scheduler.Variables_Table, Msg);

      -- After priority section, we
      -- save some task run time information
      --
      Save_Read_Write_user_defined_Variables
        (My_Scheduler,
         Si,
         Processor_Name);

      -- Do "election" section
      --

      -- A election is done in 3 steps :
      --        1) we check and retrieve return information
      --        2) tasks queues should be updated
      --        3) we find the running task
      --
      Current := My_Scheduler.Root_Statement_Pointer (Election);

      -- We should have, at least a election_section !
      --
      if Current = null then
         Raise_Exception
           (Statement_Error'Identity,
            "An election section is mandatory");
      end if;

      -- Only one type of statement is allowed on "election" section :
      -- a return statement
      --
      if Current.all not  in Return_Statement'Class then
         Raise_Exception
           (Statement_Error'Identity,
            "Election section, line " &
            Current.Line_Number'Img &
            ", return statement expected");
      end if;

      -- Do return statement
      --

      -- The return statement should be :
      --      - either a min/max_to_index expression composed of
      --        a array variable only. Boolean/double/string array variable
      --are forbidden
      --      - or an integer constant
      --      - or an integer variable
      --
      Return_Val := Return_Statement_Ptr (Current).Return_Value;

      ------------------------------------------------------------
      -- The return statement is a constant or a variable ?
      ------------------------------------------------------------

      if (Return_Val.Expression_Type = Variable_Type) or
         (Return_Val.Expression_Type = Constant_Type)
      then

         if (Return_Val.Expression_Type = Constant_Type) then
            Put_Debug ("Do a constant election");
            if Constant_Expr_Ptr (Return_Val).Value.Ptype /=
               Simulation_Integer
            then
               Raise_Exception
                 (Statement_Error'Identity,
                  "Election section, line " &
                  Current.Line_Number'Img & 
", invalid constant expression type referenced in a return statement");
            end if;
            if (Constant_Expr_Ptr (Return_Val).Value.Integer_Value <
                Integer (Tasks_Range'First)) or
               (Constant_Expr_Ptr (Return_Val).Value.Integer_Value >
                Integer (Tasks_Range'Last))
            then
               Raise_Exception
                 (Statement_Error'Identity,
                  "Election section, line " &
                  Current.Line_Number'Img & 
", invalid range constant type referenced in a return statement");
            end if;

            I :=
              Tasks_Range (Constant_Expr_Ptr (Return_Val).Value.Integer_Value);
         end if;

         if (Return_Val.Expression_Type = Variable_Type) then
            Put_Debug ("Do a variable election");
            Var :=
               Find_Variable
                 (My_Scheduler.Variables_Table,
                  Variable_Expr_Ptr (Return_Val).Name);
            if My_Scheduler.Variables_Table.Entries (Var).Simulation.Ptype /=
               Simulation_Integer
            then
               Raise_Exception
                 (Statement_Error'Identity,
                  "Election section, line " &
                  Current.Line_Number'Img & 
", invalid variable expression type referenced in a return statement");
            end if;
            if (My_Scheduler.Variables_Table.Entries (Var).Simulation.
                 Integer_Value <
                Integer (Simulations_Range'First)) or
               (My_Scheduler.Variables_Table.Entries (Var).Simulation.
                 Integer_Value >
                Integer (Tasks_Range'Last))
            then
               Raise_Exception
                 (Statement_Error'Identity,
                  "Election section, line " &
                  Current.Line_Number'Img & 
", invalid range variable type referenced in a return statement");
            end if;
            I :=
              Tasks_Range (
              My_Scheduler.Variables_Table.Entries (Var).Simulation.
              Integer_Value);
         end if;

         -- Convert user_defined variable index towards tcb index
         --
         for J in  0 .. Si.Number_Of_Tasks - 1 loop
            if 
                 My_Scheduler.Variables_Table.Entries (Var_Task_Name).
                 Simulation.String_Table_Value (Simulations_Range (I)) =
               Si.Tcbs (J).Tsk.Name
            then

               -- Check that the task is ready ... and elect it !
               --
               if (Si.Tcbs (J).Tsk.Cpu_Name = Processor_Name) then
                  if (Si.Tcbs (J).Wake_Up_Time <= Current_Time) then
                     if (Si.Tcbs (J).Rest_Of_Capacity /= 0) then
                        if (With_Offsets = False) or
                           Check_Offset (Si.Tcbs (J), Current_Time)
                        then
                           if (With_Precedencies = False) or
                              Check_Precedencies
                                 (Si,
                                  My_Dependencies,
                                  Current_Time,
                                  Si.Tcbs (J))
                           then
                              Elected := J;
                              No_Task := False;
                           end if;
                        end if;
                     end if;
                  end if;
               end if;

            end if;
         end loop;

      -------------------------------------------------------------------------
      -------
      -- Now, let see the case when the return expression is a max/min_to_index
      -------------------------------------------------------------------------
      -------

      else
         Put_Debug ("Do a min/max_to_index election");

         if (Return_Val.Expression_Type = Unary_Type) then
            if (Unary_Expr_Ptr (Return_Val).Ope = Min_To_Index) or
               (Unary_Expr_Ptr (Return_Val).Ope = Max_To_Index)
            then

               if (Unary_Expr_Ptr (Return_Val).Ope = Min_To_Index) then
                  if (Unary_Expr_Ptr (Return_Val).Value.Expression_Type /=
                      Array_Variable_Type)
                  then
                     Raise_Exception
                       (Statement_Error'Identity,
                        "Election section, line " &
                        Current.Line_Number'Img & 
", invalid min_to_index expression in return statement");
                  end if;

                  if (Array_Variable_Expr_Ptr (
                       Unary_Expr_Ptr (Return_Val).Value).Variable_Type /=
                      Simulation_Array_Integer) and
                     (Array_Variable_Expr_Ptr (
                       Unary_Expr_Ptr (Return_Val).Value).Variable_Type /=
                      Simulation_Array_Double)
                  then
                     Raise_Exception
                       (Statement_Error'Identity,
                        "Election section, line " &
                        Current.Line_Number'Img & 
", min_to_index expression in return statement should be an integer array variable"
);
                  end if;
               end if;
            end if;

            if (Return_Val.Expression_Type = Unary_Type) then
               if (Unary_Expr_Ptr (Return_Val).Ope = Max_To_Index) then
                  if (Unary_Expr_Ptr (Return_Val).Value.Expression_Type /=
                      Array_Variable_Type)
                  then
                     Raise_Exception
                       (Statement_Error'Identity,
                        "Election section, line " &
                        Current.Line_Number'Img & 
", invalid max_to_index expression in return statement");
                  end if;

                  if (Array_Variable_Expr_Ptr (
                       Unary_Expr_Ptr (Return_Val).Value).Variable_Type /=
                      Simulation_Array_Integer)
                  then
                     Raise_Exception
                       (Statement_Error'Identity,
                        "Election section, line " &
                        Current.Line_Number'Img & 
", max_to_index expression in return statement should be an integer array variable"
);
                  end if;
               end if;
            end if;

            -----------------------------------------------------
            -- All the code is now done : we do the last
            -- operation : election !!!
            -----------------------------------------------------

            Election_Value :=
               Value_Of
                 (Array_Variable_Expr (Unary_Expr_Ptr (Return_Val).Value.all),
                  My_Scheduler.Variables_Table);

            if (Election_Value.Ptype = Simulation_Array_Double) then
               Current_Selected_Value :=
                 new Simulation_Value (Simulation_Double);
               if Unary_Expr_Ptr (Return_Val).Ope = Min_To_Index then
                  Current_Selected_Value.Double_Value := Double'Last;
               else
                  Current_Selected_Value.Double_Value := Double'First;
               end if;
            else
               Current_Selected_Value :=
                 new Simulation_Value (Simulation_Integer);
               if Unary_Expr_Ptr (Return_Val).Ope = Min_To_Index then
                  Current_Selected_Value.Integer_Value := Integer'Last;
               else
                  Current_Selected_Value.Integer_Value := Integer'First;
               end if;
            end if;

            -- Now, loop on the array variable to find the appropriate task to
            --be
            -- elected
            --

            I := 0;
            loop
               if (Si.Tcbs (I).Tsk.Cpu_Name = Processor_Name) then
                  if (Si.Tcbs (I).Wake_Up_Time <= Current_Time) and
                     (Si.Tcbs (I).Rest_Of_Capacity /= 0)
                  then
                     if (With_Offsets = False) or
                        Check_Offset (Si.Tcbs (I), Current_Time)
                     then
                        if (With_Precedencies = False) or
                           Check_Precedencies
                              (Si,
                               My_Dependencies,
                               Current_Time,
                               Si.Tcbs (I))
                        then

                           if (Election_Value.Ptype =
                               Simulation_Array_Integer)
                           then

                              if ((Unary_Expr_Ptr (Return_Val).Ope =
                                   Min_To_Index) and
                                  (Election_Value.Integer_Table_Value (
                                   Simulations_Range (I)) <
                                   Current_Selected_Value.Integer_Value)) or
                                 ((Unary_Expr_Ptr (Return_Val).Ope =
                                   Max_To_Index) and
                                  (Election_Value.Integer_Table_Value (
                                   Simulations_Range (I)) >
                                   Current_Selected_Value.Integer_Value))
                              then
                                 Elected                              := I;
                                 Current_Selected_Value.Integer_Value :=
                                   Election_Value.Integer_Table_Value (
                                   Simulations_Range (I));
                              end if;
                           else
                              if ((Unary_Expr_Ptr (Return_Val).Ope =
                                   Min_To_Index) and
                                  (Election_Value.Double_Table_Value (
                                   Simulations_Range (I)) <
                                   Current_Selected_Value.Double_Value)) or
                                 ((Unary_Expr_Ptr (Return_Val).Ope =
                                   Max_To_Index) and
                                  (Election_Value.Double_Table_Value (
                                   Simulations_Range (I)) >
                                   Current_Selected_Value.Double_Value))
                              then
                                 Elected                             := I;
                                 Current_Selected_Value.Double_Value :=
                                   Election_Value.Double_Table_Value (
                                   Simulations_Range (I));
                              end if;

                           end if;
                        end if;
                     end if;
                  end if;
               end if;

               I := I + 1;
               exit when Si.Tcbs (I) = null;
            end loop;

            if (Election_Value.Ptype = Simulation_Array_Integer) then

               if ((Unary_Expr_Ptr (Return_Val).Ope = Min_To_Index) and
                   (Current_Selected_Value.Integer_Value /= Integer'Last)) or
                  ((Unary_Expr_Ptr (Return_Val).Ope = Max_To_Index) and
                   (Current_Selected_Value.Integer_Value /= Integer'First))
               then
                  No_Task := False;
               end if;
            else
               if ((Unary_Expr_Ptr (Return_Val).Ope = Min_To_Index) and
                   (Current_Selected_Value.Double_Value /= Double'Last)) or
                  ((Unary_Expr_Ptr (Return_Val).Ope = Max_To_Index) and
                   (Current_Selected_Value.Double_Value /= Double'First))
               then
                  No_Task := False;
               end if;
            end if;

            Free (Election_Value);

         -- End of return statement
         --

         ------------------------------------------------------------
         -- Others expression in a return statement is a mistake
         ------------------------------------------------------------
         else
            Raise_Exception
              (Statement_Error'Identity,
               "Election section, line " &
               Current.Line_Number'Img &
               ", invalid expression referenced in a return statement");

         end if;
      end if;

   end Do_Election;

   -- Do "start" section statements
   --
   procedure Specific_Scheduler_Initialization
     (My_Scheduler   : in out automata_user_defined_Scheduler;
      Si             : in out Scheduling_Information;
      Processor_Name : in Unbounded_String;
      My_Tasks       : in out Tasks_Set;
      My_Resources   : in out Resources_Set;
      My_Buffers     : in out Buffers_Set;
      My_Messages    : in Messages_Set;
      Msg            : in out Unbounded_String)
   is

      Current : Statement_Ptr;
      Var     : Variables_Range;

   begin

   -- Read all user_defined files before simulation
   --
      begin
         My_Scheduler.String_Behavior     := new Unbounded_String;
         My_Scheduler.String_Behavior.all :=
            Read_Sequential_File (My_Scheduler.Behavior_File_Name);

      exception
         when Ada.IO_Exceptions.Name_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   My_Scheduler.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Name_Error");
         when Ada.IO_Exceptions.Status_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   My_Scheduler.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Status_Error");
         when Ada.IO_Exceptions.Mode_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   My_Scheduler.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Mode_Error");
         when Ada.IO_Exceptions.Use_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   My_Scheduler.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Use_Error");
         when Ada.IO_Exceptions.Device_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   My_Scheduler.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Device_Error");
         when Ada.IO_Exceptions.End_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   My_Scheduler.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", End_Error");
         when Ada.IO_Exceptions.Data_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   My_Scheduler.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Data_Error");
         when Ada.IO_Exceptions.Layout_Error =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   My_Scheduler.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)) &
               ", Layout_Error");
         when others =>
            Raise_Exception
              (Parametric_File_Error'Identity,
               To_String
                  (Lb_File (Current_Language) &
                   Lb_Colon &
                   My_Scheduler.Behavior_File_Name &
                   Lb_Comma &
                   Lb_Can_Not_Open_File (Current_Language)));
      end;

      -- We should check the scheduler behavior syntax ...
      --
      Open_Input (My_Scheduler.String_Behavior.all);
      Parser.Init_Parser;
      Create_Parametric_Variables (Parser.Variables_Table, My_Tasks);
      Parser.Yyparse;

      -- Store syntax tree and compiling information
      --
      My_Scheduler.Root_Statement_Pointer := Parser.Root_Statement_Pointer;

      My_Scheduler.Variables_Table := Parser.Variables_Table;

      My_Scheduler.Sets_Table := Parser.Sets_Table;

      Current := My_Scheduler.Root_Statement_Pointer (Start);

      -- Initialize user_defined variables which are constant
      --
      Initialize_Parametric_Variables
        (My_Scheduler.Variables_Table,
         My_Messages,
         My_Buffers,
         My_Resources,
         My_Tasks,
         Processor_Name);

      -- Initialize scheduler-user_defined.adb specific user_defined variables
      --

      -- "nb_processors"
      --
      Var                                                                 :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("nb_processors"));
      My_Scheduler.Variables_Table.Entries (Var).Simulation.Integer_Value :=
        Si.Number_Of_Processors;

      -- "nb_address_spaces"
      --
      Var                                                                 :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("nb_address_spaces"));
      My_Scheduler.Variables_Table.Entries (Var).Simulation.Integer_Value :=
        Si.Number_Of_Address_Spaces;

      -- "simulation_length"
      --
      Var                                                                 :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("simulation_length"));
      My_Scheduler.Variables_Table.Entries (Var).Simulation.Integer_Value :=
        Si.Simulation_Length;

      -- Load user_defined variable which can be changed by the programmer
      -- before executing sections
      --
      Load_Read_Write_user_defined_Variables
        (My_Scheduler,
         Si,
         Processor_Name);

      -- Now : run "start" statements
      --
      Dispatch (Current, Start, My_Scheduler.Variables_Table, Msg);

      -- After start section, we
      -- save some task run time information
      --
      Save_Read_Write_user_defined_Variables
        (My_Scheduler,
         Si,
         Processor_Name);

   end Specific_Scheduler_Initialization;

   procedure Compute_Activation_Time
     (My_Scheduler : in automata_user_defined_Scheduler;
      Si           : in out Scheduling_Information;
      Elected      : in Tasks_Range;
      Value        : in out Natural)
   is

      Set_Id             : Sets_Range;
      user_defined_Delay : Simulation_Value_Ptr;

   begin

      Set_Id :=
         Find_Set
           (My_Scheduler.Sets_Table,
            Parametric_Task_Ptr (Si.Tcbs (Elected).Tsk).Activation_Rule);

      user_defined_Delay :=
         Value_Of
           (My_Scheduler.Sets_Table.Entries (Set_Id).Set_Value.all,
            My_Scheduler.Variables_Table);

      if (user_defined_Delay.Ptype /= Simulation_Integer) and
         (user_defined_Delay.Ptype /= Simulation_Array_Integer)
      then
         Raise_Exception
           (Type_Error'Identity,
            "Task activation rule, activation rule " &
            To_String
               (Parametric_Task_Ptr (Si.Tcbs (Elected).Tsk).Activation_Rule) &
            To_String
               (Lb_Comma & Lb_Uncompatible_Type_Error (Current_Language)));
      end if;

      -- If it's a vector, return the corresponding index
      -- else, just return the computed integer
      --
      if user_defined_Delay.Ptype = Simulation_Integer then
         Value := user_defined_Delay.Integer_Value;
      else
         Value :=
           user_defined_Delay.Integer_Table_Value (Simulations_Range (Elected))
;
      end if;

   end Compute_Activation_Time;

   procedure Load_Read_Write_user_defined_Variables
     (My_Scheduler   : in out automata_user_defined_Scheduler;
      Si             : in Scheduling_Information;
      Processor_Name : in Unbounded_String)
   is

      Var           : Variables_Range;
      Var_Task_Name : Variables_Range;

   begin

      Var_Task_Name :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("tasks.name"));

      for J in  0 .. Si.Number_Of_Tasks - 1 loop

         if Processor_Name = Si.Tcbs (J).Tsk.Cpu_Name then

            for Z in  0 .. Si.Number_Of_Tasks - 1 loop
               if 
                    My_Scheduler.Variables_Table.Entries (Var_Task_Name).
                    Simulation.String_Table_Value (Simulations_Range (Z)) =
                  Si.Tcbs (J).Tsk.Name
               then

                  -- "tasks.priority"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.priority"));
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z)) :=
                    Integer (Si.Tcbs (J).Tsk.Priority);

                  -- "tasks.deadline"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.deadline"));
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z)) :=
                    Integer (Si.Tcbs (J).Tsk.Deadline);

                  -- "tasks.start_time"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.start_time"));
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z)) :=
                    Integer (Si.Tcbs (J).Tsk.Start_Time);

                  -- "tasks.processor_name"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.processor_name"));
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    String_Table_Value (Simulations_Range (Z)) :=
                    Si.Tcbs (J).Tsk.Cpu_Name;

                  -- "tasks.name"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.name"));
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    String_Table_Value (Simulations_Range (Z)) :=
                    Si.Tcbs (J).Tsk.Name;

                  -- "tasks.capacity"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.capacity"));
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z)) :=
                    Integer (Si.Tcbs (J).Tsk.Capacity);

                  -- "tasks.rest_of_capacity"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.rest_of_capacity"));
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z)) :=
                    Integer (Si.Tcbs (J).Rest_Of_Capacity);

                  -- "tasks.blocking_time"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.blocking_time"));
                  My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z)) :=
                    Integer (Si.Tcbs (J).Tsk.Blocking_Time);

                  -- "tasks.period"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.period"));
                  case Si.Tcbs (J).Tsk.Task_Type is
                     when Aperiodic_Type =>
                        My_Scheduler.Variables_Table.Entries (Var).Simulation.
                          Integer_Table_Value (Simulations_Range (Z)) := 0;
                     when others =>
                        My_Scheduler.Variables_Table.Entries (Var).Simulation.
                          Integer_Table_Value (Simulations_Range (Z)) :=
                          Integer (Periodic_Task_Ptr (Si.Tcbs (J).Tsk).Period);
                  end case;

                  -- "tasks.jitter"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.jitter"));
                  case Si.Tcbs (J).Tsk.Task_Type is
                     when Aperiodic_Type =>
                        My_Scheduler.Variables_Table.Entries (Var).Simulation.
                          Integer_Table_Value (Simulations_Range (Z)) := 0;
                     when others =>
                        My_Scheduler.Variables_Table.Entries (Var).Simulation.
                          Integer_Table_Value (Simulations_Range (Z)) :=
                          Integer (Periodic_Task_Ptr (Si.Tcbs (J).Tsk).Jitter);
                  end case;

                  -- user defined task's parameter :
                  --

                  for I in  0 .. Si.Tcbs (J).Tsk.Param.Nb_Entries - 1 loop

                     -- Find the user defined variable
                     -- and set its value
                     --
                     for K in  0 .. Variables_Table.Nb_Entries - 1 loop
                        if My_Scheduler.Variables_Table.Entries (K).Variable.
                             Name =
                           To_Unbounded_String ("tasks.") &
                           Si.Tcbs (J).Tsk.Param.Entries (I).Name
                        then

                           -- Now, set the value
                           --
                           if Si.Tcbs (J).Tsk.Param.Entries (I).ptype =
                              Boolean_Parameter
                           then
                              My_Scheduler.Variables_Table.Entries (K).
                                Simulation.Boolean_Table_Value (
                                Simulations_Range (Z)) :=
                                Boolean (
                                Si.Tcbs (J).Tsk.Param.Entries (I).
                                Boolean_value);
                           end if;
                           if Si.Tcbs (J).Tsk.Param.Entries (I).ptype =
                              Integer_Parameter
                           then
                              My_Scheduler.Variables_Table.Entries (K).
                                Simulation.Integer_Table_Value (
                                Simulations_Range (Z)) :=
                                Integer (
                                Si.Tcbs (J).Tsk.Param.Entries (I).
                                Integer_value);
                           end if;
                           if Si.Tcbs (J).Tsk.Param.Entries (I).ptype =
                              Double_Parameter
                           then
                              My_Scheduler.Variables_Table.Entries (K).
                                Simulation.Double_Table_Value (
                                Simulations_Range (Z)) :=
                                Double (
                                Si.Tcbs (J).Tsk.Param.Entries (I).double_value)
;
                           end if;
                           if Si.Tcbs (J).Tsk.Param.Entries (I).ptype =
                              String_Parameter
                           then
                              My_Scheduler.Variables_Table.Entries (K).
                                Simulation.String_Table_Value (
                                Simulations_Range (Z)) :=
                                Si.Tcbs (J).Tsk.Param.Entries (I).string_value;
                           end if;
                        end if;
                     end loop;
                  end loop;

               end if;
            end loop;

         end if;
      end loop;

   end Load_Read_Write_user_defined_Variables;

   procedure Save_Read_Write_user_defined_Variables
     (My_Scheduler   : in automata_user_defined_Scheduler;
      Si             : in out Scheduling_Information;
      Processor_Name : in Unbounded_String)
   is

      Var           : Variables_Range;
      Var_Task_Name : Variables_Range;

   begin

      Var_Task_Name :=
         Find_Variable
           (My_Scheduler.Variables_Table,
            To_Unbounded_String ("tasks.name"));

      for J in  0 .. Si.Number_Of_Tasks - 1 loop

         if Processor_Name = Si.Tcbs (J).Tsk.Cpu_Name then

            for Z in  0 .. Si.Number_Of_Tasks - 1 loop
               if 
                    My_Scheduler.Variables_Table.Entries (Var_Task_Name).
                    Simulation.String_Table_Value (Simulations_Range (Z)) =
                  Si.Tcbs (J).Tsk.Name
               then

               -- "tasks.priority"
               --
                  begin
                     Var                      :=
                        Find_Variable
                          (Variables_Table,
                           To_Unbounded_String ("tasks.priority"));
                     Si.Tcbs (J).Tsk.Priority :=
                       Priority_Range (
                       My_Scheduler.Variables_Table.Entries (Var).Simulation.
                       Integer_Table_Value (Simulations_Range (Z)));
                  exception
                     when Constraint_Error =>
                        Raise_Exception
                          (Type_Error'Identity,
                           "Task priority range check error  " &
                           To_String
                              (Parametric_Task_Ptr (Si.Tcbs (J).Tsk).Name) &
                           " " &
                           Integer'Image
                              (
                          My_Scheduler.Variables_Table.Entries (Var).
                          Simulation.Integer_Table_Value (Simulations_Range (Z)
)));
                  end;

                  -- "tasks.deadline"
                  --
                  Var                      :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.deadline"));
                  Si.Tcbs (J).Tsk.Deadline :=
                    My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z));

                  -- "tasks.start_time"
                  --
                  Var                        :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.start_time"));
                  Si.Tcbs (J).Tsk.Start_Time :=
                    My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z));

                  -- "tasks.capacity"
                  --
                  Var                      :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.capacity"));
                  Si.Tcbs (J).Tsk.Capacity :=
                    My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z));

                  -- "tasks.rest_of_capacity"
                  --
                  Var                          :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.rest_of_capacity"));
                  Si.Tcbs (J).Rest_Of_Capacity :=
                    My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z));

                  -- "tasks.blocking_time"
                  --
                  Var                           :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.blocking_time"));
                  Si.Tcbs (J).Tsk.Blocking_Time :=
                    My_Scheduler.Variables_Table.Entries (Var).Simulation.
                    Integer_Table_Value (Simulations_Range (Z));

                  -- "tasks.period"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.period"));
                  case Si.Tcbs (J).Tsk.Task_Type is
                     when Aperiodic_Type =>
                        null;
                     when others =>
                        Periodic_Task_Ptr (Si.Tcbs (J).Tsk).Period :=
                          My_Scheduler.Variables_Table.Entries (Var).
                          Simulation.Integer_Table_Value (Simulations_Range (Z)
);
                  end case;

                  -- "tasks.jitter"
                  --
                  Var :=
                     Find_Variable
                       (Variables_Table,
                        To_Unbounded_String ("tasks.jitter"));
                  case Si.Tcbs (J).Tsk.Task_Type is
                     when Aperiodic_Type =>
                        null;
                     when others =>
                        Periodic_Task_Ptr (Si.Tcbs (J).Tsk).Jitter :=
                          My_Scheduler.Variables_Table.Entries (Var).
                          Simulation.Integer_Table_Value (Simulations_Range (Z)
);
                  end case;

                  -- user defined task's parameter :
                  --
                  for I in  0 .. Si.Tcbs (J).Tsk.Param.Nb_Entries - 1 loop

                     -- Find the user defined variable
                     -- and set its value
                     --
                     for K in  0 .. Variables_Table.Nb_Entries - 1 loop
                        if My_Scheduler.Variables_Table.Entries (K).Variable.
                             Name =
                           To_Unbounded_String ("tasks.") &
                           Si.Tcbs (J).Tsk.Param.Entries (I).Name
                        then

                           -- Now, set the value
                           --
                           if Si.Tcbs (J).Tsk.Param.Entries (I).ptype =
                              Boolean_Parameter
                           then
                              Si.Tcbs (J).Tsk.Param.Entries (I).Boolean_value
                                 :=
                                My_Scheduler.Variables_Table.Entries (K).
                                Simulation.Boolean_Table_Value (
                                Simulations_Range (Z));
                           end if;
                           if Si.Tcbs (J).Tsk.Param.Entries (I).ptype =
                              Integer_Parameter
                           then
                              Si.Tcbs (J).Tsk.Param.Entries (I).Integer_value
                                 :=
                                My_Scheduler.Variables_Table.Entries (K).
                                Simulation.Integer_Table_Value (
                                Simulations_Range (Z));
                           end if;
                           if Si.Tcbs (J).Tsk.Param.Entries (I).ptype =
                              Double_Parameter
                           then
                              Si.Tcbs (J).Tsk.Param.Entries (I).double_value
                                 :=
                                My_Scheduler.Variables_Table.Entries (K).
                                Simulation.Double_Table_Value (
                                Simulations_Range (Z));
                           end if;
                           if Si.Tcbs (J).Tsk.Param.Entries (I).ptype =
                              String_Parameter
                           then
                              Si.Tcbs (J).Tsk.Param.Entries (I).string_value
                                 :=
                                My_Scheduler.Variables_Table.Entries (K).
                                Simulation.String_Table_Value (
                                Simulations_Range (Z));
                           end if;
                        end if;
                     end loop;
                  end loop;

               end if;
            end loop;

         end if;
      end loop;

   end Save_Read_Write_user_defined_Variables;

end Scheduler.user_defined.automata;
