------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_IO;           use Text_IO;
with Offsets;           use Offsets;
use Offsets.Offsets_Table_package;
with Qs_Tools;          use Qs_Tools;
with Ada.Exceptions;    use Ada.Exceptions;
with Expressions;       use Expressions;
with Dependencies;      use Dependencies;
with Unbounded_Strings; use Unbounded_Strings;

package body Scheduler is

   procedure put (my_scheduler : in Scheduler_Ptr) is
   begin
      put (My_Scheduler.all);
   end put;

   function Export_Aadl_Properties
     (My_Scheduler : in Generic_Scheduler;
      Number_Of_Ht : in Natural)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;

   begin

      for I in  1 .. Number_Of_Ht loop
         Result := Result & ASCII.HT;
      end loop;

      Result := Result &
                To_Unbounded_String
                   ("Scheduling_Protocol => " &
                    To_String (Get_Name (My_Scheduler)) &
                    ";") &
                Unbounded_Lf;

      for I in  1 .. Number_Of_Ht loop
         Result := Result & ASCII.HT;
      end loop;

      Result := Result &
                To_Unbounded_String
                   ("Cheddar_Properties::Scheduler_Quantum => " &
                    Get_Quantum (My_Scheduler) &
                    " ms ;") &
                Unbounded_Lf;

      for I in  1 .. Number_Of_Ht loop
         Result := Result & ASCII.HT;
      end loop;

      if Get_Preemptive (My_Scheduler) = Preemptive then
         Result := Result &
                   To_Unbounded_String
                      ("Cheddar_Properties::Preemptive_Scheduler => True;") &
                   Unbounded_Lf;
      else
         Result := Result &
                   To_Unbounded_String
                      ("Cheddar_Properties::Preemptive_Scheduler => False;") &
                   Unbounded_Lf;
      end if;

      return Result;

   end Export_Aadl_Properties;

   function Export_Xml
     (My_Scheduler : in Generic_Scheduler)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;

      Quantum       : constant Natural         := Get_Quantum (My_Scheduler);
      Is_Preemptive : constant Preemptive_Type :=
         Get_Preemptive (My_Scheduler);
      Name          : Unbounded_String         := Get_Name (My_Scheduler);

   begin

      Result := Result &
                To_Unbounded_String (ASCII.HT & ASCII.HT & "<scheduler");

      if Quantum /= 0 then
         Result := Result &
                   To_Unbounded_String (" quantum=""" & Quantum'Img & """");
      end if;

      if Is_Preemptive /= Preemptive then
         Result := Result &
                   To_Unbounded_String
                      (" is_preemptive=""" & Is_Preemptive'Img & """");
      end if;

      Result := Result & To_Unbounded_String (">");
      Result := Result &
                To_Unbounded_String (" ") &
                Name &
                To_Unbounded_String (" ");
      Result := Result &
                To_Unbounded_String ("</scheduler>") &
                Unbounded_Lf;

      return Result;

   end Export_Xml;

   procedure Set_Preemptive
     (My_Scheduler : in out Generic_Scheduler'Class;
      Preempt      : in Preemptive_Type)
   is
   begin
      My_Scheduler.Is_Preemptive := Preempt;
   end Set_Preemptive;

   function Get_Preemptive
     (My_Scheduler : in Generic_Scheduler'Class)
      return         Preemptive_Type
   is
   begin
      return My_Scheduler.Is_Preemptive;
   end Get_Preemptive;

   function Get_Preemptive
     (My_Scheduler : in Generic_Scheduler'Class)
      return         String
   is
   begin
      return My_Scheduler.Is_Preemptive'Img;
   end Get_Preemptive;

   procedure Put (My_Scheduler : in Generic_Scheduler) is
   begin
      Put_Line ("Scheduler name : " & My_Scheduler.Name'Img);
      Put_Line ("Is preemptive  : " & My_Scheduler.Is_Preemptive'Img);
      Put_Line ("Quantum        : " & My_Scheduler.Quantum'Img);
   end Put;

   procedure Reset (A_Scheduler : in out Generic_Scheduler'Class) is
   begin
      A_Scheduler.Is_Preemptive := Preemptive;
      A_Scheduler.Quantum       := 0;
   end Reset;

   function Get_Name
     (My_Scheduler : in Generic_Scheduler'Class)
      return         Unbounded_String
   is
   begin
      return To_Unbounded_String (My_Scheduler.Name'Img);
   end Get_Name;

   function Get_Name
     (My_Scheduler : in Generic_Scheduler'Class)
      return         Scheduler_Type
   is
   begin
      return My_Scheduler.Name;
   end Get_Name;

   function Build_Tcb
     (My_Scheduler : in Generic_Scheduler;
      A_Task       : Generic_Task_Ptr)
      return         Tcb_Ptr
   is
      A_Tcb : Tcb_Ptr;
   begin
      A_Tcb := new Tcb;
      Initialize (A_Tcb.all, A_Task);
      return A_Tcb;
   end Build_Tcb;

   procedure Initialize (A_Tcb : in out Tcb; A_Task : Generic_Task_Ptr) is

      Seed : Generator;

   begin
      A_Tcb.Tsk := A_Task;

      A_Tcb.Activation       := 1;
      A_Tcb.End_Time         := 0;
      A_Tcb.Used_Cpu         := 0;
      A_Tcb.Wake_Up_Time     := A_Tcb.Tsk.Start_Time;
      A_Tcb.Rest_Of_Capacity := A_Tcb.Tsk.Capacity;

      if (A_Tcb.Tsk.Task_Type = Poisson_Type) or
         (A_Tcb.Tsk.Task_Type = Parametric_Type)
      then
         if (Poisson_Task_Ptr (A_Tcb.Tsk).Predictable) then
            Reset (Seed, Poisson_Task_Ptr (A_Tcb.Tsk).Seed);
         else
            Reset (Seed);
         end if;

         Save (Seed, A_Tcb.Task_Seed);
      end if;
   end Initialize;

   procedure Scheduler_Initialization
     (My_Scheduler            : in out Generic_Scheduler'Class;
      Si                      : in out Scheduling_Information;
      Processor_Name          : in Unbounded_String;
      My_Tasks                : in out Tasks_Set;
      My_Resources            : in out Resources_Set;
      My_Buffers              : in out Buffers_Set;
      Result                  : in out Scheduling_Sequence_Ptr;
      With_Offsets            : Boolean;
      With_Precedencies       : Boolean;
      With_Resources          : Boolean;
      With_Task_Specific_Seed : Boolean := True;
      Global_Seed_Value       : Integer := 0;
      Predictable_Global_Seed : Boolean := True;
      Given_Last_Time         : in Natural;
      Event_To_Generate       : in Time_Unit_Event_Type_Boolean_Table)
   is

      Iterator1     : Tasks_Iterator;
      Iterator2     : Resources_Iterator;
      A_Buffer_Size : Buffer_Scheduling_Information_Ptr;
      Iterator3     : Buffers_Iterator;
      A_Buffer      : Buffer_Ptr;
      A_Task        : Generic_Task_Ptr;
      A_Resource    : Generic_Resource_Ptr;
      Seed          : Generator;
      A_Item        : Time_Unit_Event_Ptr;

   begin

      My_Scheduler.Previously_Elected := Tasks_Range'First;

      if (Predictable_Global_Seed) then
         Reset (Seed, Global_Seed_Value);
      else
         Reset (Seed);
      end if;

      Si.With_Specific_Task_Seed := With_Task_Specific_Seed;
      Save (Seed, Si.Global_Seed);

      -- Build Task Control Blocks
      --
      Reset_Iterator (My_Tasks, Iterator1);

      loop
         Current_Element (My_Tasks, A_Task, Iterator1);

         Si.Tcbs (Si.Number_Of_Tasks) := Build_Tcb (My_Scheduler, A_Task);

         Si.Number_Of_Tasks := Si.Number_Of_Tasks + 1;

         exit when Is_Last_Element (My_Tasks, Iterator1);
         Next_Element (My_Tasks, Iterator1);
      end loop;

      -- Build of buffer list with their current size
      -- (in order to generate Write/Read buffer events
      --
      if (With_Precedencies) then
         if (Get_Number_Of_Elements (My_Buffers) > 0) then

            Reset_Iterator (My_Buffers, Iterator3);
            loop

               Current_Element (My_Buffers, A_Buffer, Iterator3);

               A_Buffer_Size                :=
                 new Buffer_Scheduling_Information;
               A_Buffer_Size.Written_Buffer := A_Buffer;
               A_Buffer_Size.Current_Size   := 0;
               Add (Si.Written_Buffers, A_Buffer_Size);

               exit when Is_Last_Element (My_Buffers, Iterator3);
               Next_Element (My_Buffers, Iterator3);

            end loop;
         end if;
      end if;

      if (With_Resources) then
         if (Get_Number_Of_Elements (My_Resources) > 0) then

            Reset_Iterator (My_Resources, Iterator2);
            loop

               Current_Element (My_Resources, A_Resource, Iterator2);

               Si.Shared_Resources (Si.Number_Of_Resources) :=
                  Build_Resource (My_Scheduler, A_Resource);

               Si.Number_Of_Resources := Si.Number_Of_Resources + 1;

               exit when Is_Last_Element (My_Resources, Iterator2);
               Next_Element (My_Resources, Iterator2);

            end loop;
         end if;
      end if;

      -- Initialize event table (store the computed scheduling)
      --
      Initialize (Result.all);

      -- First activation for all tasks
      --
      for I in  0 .. Si.Number_Of_Tasks - 1 loop

         if Si.Tcbs (I).Tsk.Cpu_Name = Processor_Name then

            if Event_To_Generate (Task_Activation) then
               A_Item                 :=
                 new Time_Unit_Event (Task_Activation);
               A_Item.Activation_Task := Si.Tcbs (I).Tsk;

               Add (Result.all, A_Item, Si.Tcbs (I).Tsk.Start_Time);
            end if;

         end if;

      end loop;

   end Scheduler_Initialization;

   procedure Build_Scheduling_Sequence
     (My_Scheduler            : in out Generic_Scheduler'Class;
      My_Tasks                : in Tasks_Set;
      My_Resources            : in Resources_Set;
      My_Buffers              : in Buffers_Set;
      My_Messages             : in Messages_Set;
      My_Dependencies         : in Tasks_Dependencies_Ptr;
      Processor_Name          : in Unbounded_String;
      Result                  : out Scheduling_Sequence_Ptr;
      Msg                     : in out Unbounded_String;
      Event_To_Generate       : in Time_Unit_Event_Type_Boolean_Table;
      Given_Last_Time         : Natural := 0;
      With_Offsets            : Boolean := True;
      With_Precedencies       : Boolean := True;
      With_Resources          : Boolean := True;
      With_Task_Specific_Seed : Boolean := True;
      Global_Seed_Value       : Integer := 0;
      Predictable_Global_Seed : Boolean := True)
   is

      Tmp_Tasks     : Tasks_Set;
      Tmp_Resources : Resources_Set;
      Tmp_Buffers   : Buffers_Set;
      Tmp_Messages  : Messages_Set;

      Last_Time    : Natural;
      Current_Time : Natural := 0;

      -- index for scheduling table
      --
      Elected : Tasks_Range := Tasks_Range'First;
      No_Task : Boolean     := True;

      Si : Scheduling_Information;

   begin

      Si.Number_Of_Tasks     := 0;
      Si.Number_Of_Resources := 0;

      -- Before scheduling, check if the tasks can be
      -- scheduled with the selected scheduler
      --

      Check_Before_Scheduling (My_Scheduler, My_Tasks, Processor_Name);

      -- First : gather information
      --
      Current_Processor_Name := Processor_Name;

      Select_And_Copy (My_Tasks, Tmp_Tasks, Select_Cpu'Access);
      Select_And_Copy (My_Resources, Tmp_Resources, Select_Cpu'Access);
      Select_And_Copy (My_Buffers, Tmp_Buffers, Select_Cpu'Access);
      Duplicate (My_Messages, Tmp_Messages);

      Scheduler_Initialization
        (My_Scheduler,
         Si,
         Processor_Name,
         Tmp_Tasks,
         Tmp_Resources,
         Tmp_Buffers,
         Result,
         With_Offsets,
         With_Precedencies,
         With_Resources,
         With_Task_Specific_Seed,
         Global_Seed_Value,
         Predictable_Global_Seed,
         Given_Last_Time,
         Event_To_Generate);

      Specific_Scheduler_Initialization
        (My_Scheduler,
         Si,
         Processor_Name,
         Tmp_Tasks,
         Tmp_Resources,
         Tmp_Buffers,
         Tmp_Messages,
         Msg);

      -- Compute last scheduled time unit
      --

      if (Given_Last_Time = 0) then
         Last_Time :=
            Natural'Min
              (Max_Scheduling_Period,
               Scheduling_Period (Tmp_Tasks, Processor_Name));
      else
         Last_Time := Given_Last_Time;
      end if;
      Si.Simulation_Length := Last_Time;

      -- Second : compute scheduling
      --
      while (Current_Time < Last_Time) loop

         Do_Election
           (My_Scheduler,
            Si,
            Result,
            Msg,
            Current_Time,
            Processor_Name,
            My_Dependencies,
            With_Offsets,
            With_Precedencies,
            With_Resources,
            Event_To_Generate,
            Elected,
            No_Task);

         -- No task to schedule
         --
         if No_Task then
            Current_Time := Current_Time + 1;

         else
            Run_Task
              (My_Scheduler,
               Processor_Name,
               Si,
               My_Dependencies,
               Elected,
               Result,
               Current_Time,
               Last_Time,
               With_Offsets,
               With_Precedencies,
               With_Resources,
               Event_To_Generate);

         end if;

      end loop;

   end Build_Scheduling_Sequence;

   procedure Run_Task
     (My_Scheduler      : in out Generic_Scheduler'Class;
      Processor_Name    : in Unbounded_String;
      Si                : in out Scheduling_Information;
      My_Dependencies   : in Tasks_Dependencies_Ptr;
      Elected           : in Tasks_Range;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in out Natural;
      Last_Time         : in Natural;
      With_Offsets      : Boolean := True;
      With_Precedencies : Boolean := True;
      With_Resources    : Boolean := True;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table)
   is

      -- For Poisson Process task
      --
      Seed : Generator;
      Temp : Natural;

      -- For Parametric task
      --
      Parametric_Delay : Natural := 0;

      A_Item : Time_Unit_Event_Ptr;

   begin

      ------------------------------------------------------
      -- Allocate/release shared resources
      ------------------------------------------------------

      if With_Resources then
         Allocate_Resource
           (My_Scheduler,
            Si,
            Result,
            Current_Time,
            Si.Tcbs (Elected),
            Event_To_Generate);

         Release_Resource
           (My_Scheduler,
            Si,
            Result,
            Current_Time,
            Si.Tcbs (Elected),
            Event_To_Generate);
      end if;

      ------------------------------------------------------
      -- Send/Receive Messages
      -- write/read on buffers
      ------------------------------------------------------

      if With_Precedencies then

         Send_Message
           (My_Scheduler,
            Si,
            Result,
            Current_Time,
            Si.Tcbs (Elected),
            My_Dependencies,
            Event_To_Generate);

         Receive_Message
           (My_Scheduler,
            Si,
            Result,
            Current_Time,
            Si.Tcbs (Elected),
            My_Dependencies,
            Event_To_Generate);

         Buffer_Write
           (My_Scheduler,
            Si,
            Result,
            Current_Time,
            Si.Tcbs (Elected),
            My_Dependencies,
            Event_To_Generate);

         Buffer_Read
           (My_Scheduler,
            Si,
            Result,
            Current_Time,
            Si.Tcbs (Elected),
            My_Dependencies,
            Event_To_Generate);

      end if;

      ------------------------------------------------------
      -- Computed scheduling
      ------------------------------------------------------

      -- Preemptive case
      --

      if (My_Scheduler.Is_Preemptive = Preemptive) then

         if Si.Tcbs (Elected).Rest_Of_Capacity =
            Si.Tcbs (Elected).Tsk.Capacity
         then
            if Event_To_Generate (Start_Of_Task_Capacity) then
               A_Item            :=
                 new Time_Unit_Event (Start_Of_Task_Capacity);
               A_Item.Start_Task := Si.Tcbs (Elected).Tsk;
               Add (Result.all, A_Item, Current_Time);
            end if;
         end if;

         if Event_To_Generate (Running_Task) then
            A_Item :=
               Produce_Running_Task_Event (My_Scheduler, Si.Tcbs (Elected));
            Add (Result.all, A_Item, Current_Time);
         end if;

         Si.Tcbs (Elected).Rest_Of_Capacity :=
           Si.Tcbs (Elected).Rest_Of_Capacity - 1;
         Si.Tcbs (Elected).Used_Cpu         := Si.Tcbs (Elected).Used_Cpu +
                                               1;

         Current_Time := Current_Time + 1;

         if Si.Tcbs (Elected).Rest_Of_Capacity = 0 then
            if Event_To_Generate (End_Of_Task_Capacity) then
               A_Item          := new Time_Unit_Event (End_Of_Task_Capacity);
               A_Item.End_Task := Si.Tcbs (Elected).Tsk;
               Add (Result.all, A_Item, Current_Time);
            end if;
         end if;

      else

         -- Non preemptive case
         --

         if Event_To_Generate (Start_Of_Task_Capacity) then
            A_Item            := new Time_Unit_Event (Start_Of_Task_Capacity);
            A_Item.Start_Task := Si.Tcbs (Elected).Tsk;
            Add (Result.all, A_Item, Current_Time);
         end if;

         for I in  0 .. (Si.Tcbs (Elected).Tsk.Capacity - 1) loop
            if Event_To_Generate (Running_Task) then
               A_Item :=
                  Produce_Running_Task_Event
                    (My_Scheduler,
                     Si.Tcbs (Elected));
               Add (Result.all, A_Item, Current_Time);
            end if;
            Current_Time               := Current_Time + 1;
            Si.Tcbs (Elected).Used_Cpu := Si.Tcbs (Elected).Used_Cpu + 1;
         end loop;

         if Event_To_Generate (End_Of_Task_Capacity) then
            A_Item          := new Time_Unit_Event (End_Of_Task_Capacity);
            A_Item.End_Task := Si.Tcbs (Elected).Tsk;
            Add (Result.all, A_Item, Current_Time);
         end if;

         Si.Tcbs (Elected).Rest_Of_Capacity := 0;

      end if;

      ------------------------------------------------------
      -- Compute the next task activations
      ------------------------------------------------------

      if (Si.Tcbs (Elected).Rest_Of_Capacity = 0) then

         Si.Tcbs (Elected).End_Time   := Current_Time;
         Si.Tcbs (Elected).Activation := Si.Tcbs (Elected).Activation + 1;

         case Si.Tcbs (Elected).Tsk.Task_Type is

            when Periodic_Type =>
               Si.Tcbs (Elected).Rest_Of_Capacity :=
                 Si.Tcbs (Elected).Tsk.Capacity;
               Si.Tcbs (Elected).Wake_Up_Time     :=
                 Si.Tcbs (Elected).Wake_Up_Time +
                 Periodic_Task_Ptr (Si.Tcbs (Elected).Tsk).Period;

            when Poisson_Type =>
               Si.Tcbs (Elected).Rest_Of_Capacity :=
                 Si.Tcbs (Elected).Tsk.Capacity;

               if Si.With_Specific_Task_Seed then
                  Reset (Seed, Si.Tcbs (Elected).Task_Seed);
               else
                  Reset (Seed, Si.Global_Seed);
               end if;

               Temp :=
                 Natural (Get_Exponential_Time
                             (Double (
                 Poisson_Task_Ptr (Si.Tcbs (Elected).Tsk).Period),
                              Seed));

               Si.Tcbs (Elected).Wake_Up_Time :=
                 Si.Tcbs (Elected).Wake_Up_Time + Temp;

               if Si.With_Specific_Task_Seed then
                  Save (Seed, Si.Tcbs (Elected).Task_Seed);
               else
                  Save (Seed, Si.Global_Seed);
               end if;

            when Sporadic_Type =>
               Si.Tcbs (Elected).Rest_Of_Capacity :=
                 Si.Tcbs (Elected).Tsk.Capacity;

               if Si.With_Specific_Task_Seed then
                  Reset (Seed, Si.Tcbs (Elected).Task_Seed);
               else
                  Reset (Seed, Si.Global_Seed);
               end if;

               Temp :=
                 Natural (Get_Exponential_Time
                             (Double (
                 Poisson_Task_Ptr (Si.Tcbs (Elected).Tsk).Period),
                              Seed));

               Si.Tcbs (Elected).Wake_Up_Time :=
                 Si.Tcbs (Elected).Wake_Up_Time +
                 Natural'Max
                    (Temp,
                     Periodic_Task_Ptr (Si.Tcbs (Elected).Tsk).Period);

               if Si.With_Specific_Task_Seed then
                  Save (Seed, Si.Tcbs (Elected).Task_Seed);
               else
                  Save (Seed, Si.Global_Seed);
               end if;

            when Aperiodic_Type =>
               null;

            when Parametric_Type =>
               Si.Tcbs (Elected).Rest_Of_Capacity :=
                 Si.Tcbs (Elected).Tsk.Capacity;

               -- The scheduler has to be a parametric one !!!
               --
               if (My_Scheduler.Name /= pipeline_User_Defined_Protocol) and
                (My_Scheduler.Name /= compiled_User_Defined_Protocol) and
                (My_Scheduler.Name /= automata_User_Defined_Protocol) then
                  Raise_Exception
                    (Statement_Error'Identity, 
"User-defined task only permitted with user-defined schedulers");
               end if;

               Compute_Activation_Time
                 (My_Scheduler,
                  Si,
                  Elected,
                  Parametric_Delay);

               Si.Tcbs (Elected).Wake_Up_Time :=
                 Si.Tcbs (Elected).Wake_Up_Time + Parametric_Delay;

         end case;

         -- Add to the scheduling table the next activation time
         --
         if Si.Tcbs (Elected).Tsk.Task_Type /= Aperiodic_Type and
            Si.Tcbs (Elected).Wake_Up_Time <= Last_Time
         then

            if Event_To_Generate (Task_Activation) then
               A_Item                 :=
                 new Time_Unit_Event (Task_Activation);
               A_Item.Activation_Task := Si.Tcbs (Elected).Tsk;
               Add (Result.all, A_Item, Si.Tcbs (Elected).Wake_Up_Time);
            end if;

         end if;
      end if;

      My_Scheduler.Previously_Elected := Elected;

   end Run_Task;

   -- Return true is the task "task_id" can be schedule now
   -- according its offsets
   --
   function Check_Offset
     (A_Tcb   : Tcb_Ptr;
      At_Time : Natural)
      return    Boolean
   is
      Target_Time : Integer := 0;
   begin

      for K in  0 .. A_Tcb.Tsk.Offset.Nb_Entries - 1 loop
         if A_Tcb.Tsk.Offset.Entries (K).Activation = A_Tcb.Activation then
            Target_Time := A_Tcb.Wake_Up_Time +
                           A_Tcb.Tsk.Offset.Entries (K).Offset_Value;
            if At_Time < Target_Time then
               return False;
            end if;
         end if;
      end loop;

      return True;
   end Check_Offset;

   function Select_Cpu (Op1 : in Generic_Task_Ptr) return Boolean is
   begin
      return (Op1.Cpu_Name = Current_Processor_Name);
   end Select_Cpu;

   function Select_Cpu (Op1 : in Generic_Resource_Ptr) return Boolean is
   begin
      return (Op1.cpu_name = Current_Processor_Name);
   end Select_Cpu;

   function Select_Cpu (Op1 : in Buffer_Ptr) return Boolean is
   begin
      return (Op1.Cpu_Name = Current_Processor_Name);
   end Select_Cpu;

   procedure Send_Message
     (My_Scheduler      : in out Generic_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : Tcb_Ptr;
      Deps              : Tasks_Dependencies_Ptr;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table)
   is

      A_Item                           : Time_Unit_Event_Ptr;
      My_Iterator                      : Tasks_Dependencies_Iterator;
      A_Half_Dep_Ptr                   : Dependency_Ptr;
      A_Message                        : Message_Ptr;
      A_Message_Scheduling_Information : Message_Scheduling_Information_Ptr;

   begin

      -- is the time to send message ?
      --
      if (A_Tcb.Rest_Of_Capacity = 1) then

         -- Check if the task has a message to send
         --
         if not Is_Empty (Deps.all) then
            Reset_Iterator (Deps.all, My_Iterator);
            loop
               Current_Element (Deps.all, A_Half_Dep_Ptr, My_Iterator);

               -- Is the task concerned ?
               -- Is it a communication dependency ?
               -- Is it a sending operation ?
               --
               if (A_Half_Dep_Ptr.Depend_Type =
                   Communication_Dependency)
               then
                  if (A_Half_Dep_Ptr.Communication_Dependent_Task.Name =
                      A_Tcb.Tsk.Name) and
                     (A_Half_Dep_Ptr.Communication_Orientation =
                      From_Task_To_Object)
                  then

                     A_Message :=
                       A_Half_Dep_Ptr.Communication_Dependency_Object;

                     A_Message_Scheduling_Information                :=
                       new Message_Scheduling_Information;
                     A_Message_Scheduling_Information.Send_Time      :=
                       Current_Time + 1;
                     A_Message_Scheduling_Information.Sended_Message :=
                       A_Message;
                     Add
                       (Si.Sended_Messages,
                        A_Message_Scheduling_Information);

                     if Event_To_Generate (Send_Message) then
                        A_Item              :=
                          new Time_Unit_Event (Send_Message);
                        A_Item.Send_Message := A_Message;
                        A_Item.Send_Task    := A_Tcb.Tsk;
                        Add
                          (Result.all,
                           A_Item,
                           A_Message_Scheduling_Information.Send_Time);
                     end if;

                  end if;
               end if;

               exit when Is_Last_Element (Deps.all, My_Iterator);

               Next_Element (Deps.all, My_Iterator);

            end loop;

         end if;
      end if;
   end Send_Message;

   procedure Receive_Message
     (My_Scheduler      : in out Generic_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : Tcb_Ptr;
      Deps              : Tasks_Dependencies_Ptr;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table)
   is

      A_Item         : Time_Unit_Event_Ptr;
      My_Iterator    : Tasks_Dependencies_Iterator;
      A_Half_Dep_Ptr : Dependency_Ptr;
      A_Message      : Message_Ptr;
      List_Ite       : Message_Scheduling_Information_Iterator;
      Msg            : Message_Scheduling_Information_Ptr;

   begin

      -- Is The time to receive a message ?
      --
      if (A_Tcb.Rest_Of_Capacity = A_Tcb.Tsk.Capacity) then

         -- Check the sended_messages to see if this
         -- messages was already sent
         --
         if not Is_Empty (Si.Sended_Messages) then

            Reset_Head_Iterator (Si.Sended_Messages, List_Ite);
            loop
               Current_Element (Si.Sended_Messages, Msg, List_Ite);

               Reset_Iterator (Deps.all, My_Iterator);
               loop
                  Current_Element (Deps.all, A_Half_Dep_Ptr, My_Iterator);

                  -- Is the task concerned ?
                  -- Is it a communication dependency ?
                  -- Is it a receiving operation ?
                  --
                  if (A_Half_Dep_Ptr.Depend_Type =
                      Communication_Dependency)
                  then
                     if (A_Half_Dep_Ptr.Communication_Dependent_Task.Name =
                         A_Tcb.Tsk.Name) and
                        (A_Half_Dep_Ptr.Communication_Orientation =
                         From_Object_To_Task)
                     then

                        A_Message :=
                          A_Half_Dep_Ptr.Communication_Dependency_Object;

                        -- Check the sended_messages to see if this
                        -- messages was already sent
                        --
                        -- Is it the right message ?
                        -- Is the message arrived at the receiving processor ?
                        --
                        if (Msg.Sended_Message.Name = A_Message.Name) then

                           if (Current_Time >=
                               (Msg.Send_Time + A_Message.Response_Time - 1))
                           then

                              if Event_To_Generate (Receive_Message) then
                                 A_Item                 :=
                                   new Time_Unit_Event (Receive_Message);
                                 A_Item.Receive_Message := A_Message;
                                 A_Item.Receive_Task    := A_Tcb.Tsk;
                                 Add (Result.all, A_Item, Current_Time);
                              end if;

                              -- Delete received message
                              --
                              Delete (Si.Sended_Messages, Msg);

                              -- Currently, a task reads only one message per
                              --activation
                              --
                              return;

                           end if;
                        end if;

                     end if;
                  end if;

                  exit when Is_Last_Element (Deps.all, My_Iterator);

                  Next_Element (Deps.all, My_Iterator);

               end loop;

               if Is_Tail_Element (Si.Sended_Messages, List_Ite) then
                  exit;
               end if;
               Next_Element (Si.Sended_Messages, List_Ite);
            end loop;

         end if;
      end if;

   end Receive_Message;

   procedure Buffer_Read
     (My_Scheduler      : in out Generic_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : Tcb_Ptr;
      Deps              : Tasks_Dependencies_Ptr;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table)
   is

      A_Item         : Time_Unit_Event_Ptr;
      My_Iterator    : Tasks_Dependencies_Iterator;
      A_Half_Dep_Ptr : Dependency_Ptr;
      A_Buffer       : Buffer_Ptr;
      A_Buffer_Size  : Buffer_Scheduling_Information_Ptr;
      List_Ite       : Buffer_Scheduling_Information_Iterator;

   begin

      -- Check if the task has a buffer to write
      --
      if (Has_Buffer_To_Read (Deps, A_Tcb.Tsk)) then

         -- for all tasks which have a precedency with the current task
         --
         loop
            Current_Element (Deps.all, A_Half_Dep_Ptr, My_Iterator);

            if (A_Half_Dep_Ptr.Depend_Type = Buffer_Dependency) then
               if (A_Half_Dep_Ptr.Buffer_Dependent_Task.Name =
                   A_Tcb.Tsk.Name) and
                  (A_Half_Dep_Ptr.Buffer_Orientation = From_Object_To_Task)
               then

                  A_Buffer := A_Half_Dep_Ptr.Buffer_Dependency_Object;

                  -- Must we generate Read event ?
                  --
                  if Event_To_Generate (Read_From_Buffer) then
                     for I in  0 .. A_Buffer.Roles.Nb_Entries - 1 loop

                        -- Is the right task ?
                        --
                        if (A_Buffer.Roles.Entries (I).Item =
                            A_Tcb.Tsk.Name)
                        then

                           -- Is it time to read ?
                           --
                           if (A_Buffer.Roles.Entries (I).Data.Time =
                               (A_Tcb.Tsk.Capacity +
                                1 -
                                A_Tcb.Rest_Of_Capacity))
                           then

                              -- Does the task is consumer ?
                              --
                              if (A_Buffer.Roles.Entries (I).Data.The_Role =
                                  Consumer)
                              then

                                 -- Can we find some information to read ?
                                 --
                                 if not Is_Empty (Si.Written_Buffers) then

                                    Reset_Head_Iterator
                                      (Si.Written_Buffers,
                                       List_Ite);
                                    loop
                                       Current_Element
                                         (Si.Written_Buffers,
                                          A_Buffer_Size,
                                          List_Ite);

                                       if (A_Buffer_Size.Written_Buffer.Name =
                                           A_Buffer.Name) and
                                          (A_Buffer_Size.Current_Size >=
                                           A_Buffer.Roles.Entries (I).Data.
                                            Size)
                                       then
                                          A_Item :=
                                            new Time_Unit_Event
                                            (Read_From_Buffer);

                                          A_Item.Read_Buffer := A_Buffer;
                                          A_Item.Read_Task   := A_Tcb.Tsk;
                                          A_Item.Read_Size   :=
                                            A_Buffer.Roles.Entries (I).Data.
                                            Size;
                                          Add
                                            (Result.all,
                                             A_Item,
                                             Current_Time);
                                          exit;
                                       end if;

                                       if Is_Tail_Element
                                            (Si.Written_Buffers,
                                             List_Ite)
                                       then
                                          exit;
                                       end if;
                                       Next_Element
                                         (Si.Written_Buffers,
                                          List_Ite);
                                    end loop;
                                 end if;

                              end if;
                           end if;
                        end if;
                     end loop;
                  end if;

               end if;
            end if;

            exit when Is_Last_Element (Deps.all, My_Iterator);

            Next_Element (Deps.all, My_Iterator);

         end loop;

      end if;

   end Buffer_Read;

   procedure Buffer_Write
     (My_Scheduler      : in out Generic_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : Tcb_Ptr;
      Deps              : Tasks_Dependencies_Ptr;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table)
   is

      A_Item         : Time_Unit_Event_Ptr;
      My_Iterator    : Tasks_Dependencies_Iterator;
      A_Half_Dep_Ptr : Dependency_Ptr;
      A_Buffer       : Buffer_Ptr;
      A_Buffer_Size  : Buffer_Scheduling_Information_Ptr;
      List_Ite       : Buffer_Scheduling_Information_Iterator;

   begin

      -- Check if the task has a buffer to write
      --
      if (Has_Buffer_To_Write (Deps, A_Tcb.Tsk)) then

         -- for all tasks which have a precedency with the current task
         --
         loop
            Current_Element (Deps.all, A_Half_Dep_Ptr, My_Iterator);

            if (A_Half_Dep_Ptr.Depend_Type = Buffer_Dependency) then

               if (A_Half_Dep_Ptr.Buffer_Dependent_Task.Name =
                   A_Tcb.Tsk.Name) and
                  (A_Half_Dep_Ptr.Buffer_Orientation = From_Task_To_Object)
               then

                  A_Buffer := A_Half_Dep_Ptr.Buffer_Dependency_Object;

                  -- Must we generate Write event ?
                  --
                  if Event_To_Generate (Write_To_Buffer) then
                     for I in  0 .. A_Buffer.Roles.Nb_Entries - 1 loop

                        -- Is the right task ?
                        --
                        if (A_Buffer.Roles.Entries (I).Item =
                            A_Tcb.Tsk.Name)
                        then
                           -- Does the task is producer ?
                           --
                           if (A_Buffer.Roles.Entries (I).Data.The_Role =
                               Producer)
                           then
                              -- Is it time to write ?
                              --
                              if (A_Buffer.Roles.Entries (I).Data.Time =
                                  (A_Tcb.Tsk.Capacity +
                                   1 -
                                   A_Tcb.Rest_Of_Capacity))
                              then
                                 A_Item :=
                                   new Time_Unit_Event (Write_To_Buffer);

                                 A_Item.Write_Buffer := A_Buffer;
                                 A_Item.Write_Task   := A_Tcb.Tsk;
                                 A_Item.Write_Size   :=
                                   A_Buffer.Roles.Entries (I).Data.Size;
                                 Add (Result.all, A_Item, Current_Time);

                                 -- Add the size write event to the buffer list
                                 --
                                 if not Is_Empty (Si.Written_Buffers) then

                                    Reset_Head_Iterator
                                      (Si.Written_Buffers,
                                       List_Ite);
                                    loop
                                       Current_Element
                                         (Si.Written_Buffers,
                                          A_Buffer_Size,
                                          List_Ite);

                                       if A_Buffer_Size.Written_Buffer.Name =
                                          A_Buffer.Name
                                       then
                                          A_Buffer_Size.Current_Size :=
                                            A_Buffer_Size.Current_Size +
                                            A_Item.Write_Size;
                                          exit;
                                       end if;

                                       if Is_Tail_Element
                                            (Si.Written_Buffers,
                                             List_Ite)
                                       then
                                          exit;
                                       end if;
                                       Next_Element
                                         (Si.Written_Buffers,
                                          List_Ite);
                                    end loop;
                                 end if;

                              end if;
                           end if;
                        end if;
                     end loop;
                  end if;

               end if;
            end if;

            exit when Is_Last_Element (Deps.all, My_Iterator);

            Next_Element (Deps.all, My_Iterator);

         end loop;

      end if;

   end Buffer_Write;

   procedure Release_Resource
     (My_Scheduler      : in out Generic_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : Tcb_Ptr;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table)
   is

      Current_Capacity : Natural := 0;
      A_Item           : Time_Unit_Event_Ptr;

   begin

      if (Si.Number_Of_Resources > 0) then
         Current_Capacity := A_Tcb.Tsk.Capacity - A_Tcb.Rest_Of_Capacity + 1;

         for Index1 in  0 .. Si.Number_Of_Resources - 1 loop

            -- For each resource, check if the task hold resources
            -- and then release them
            --
            for Index2 in  0 .. Si.Shared_Resources (Index1).Shared.task_tab.
                 Nb_Entries -
                                1
            loop

               if (Current_Capacity =
                   Si.Shared_Resources (Index1).Shared.task_tab.Entries (
                    Index2).Data.Task_End) and
                  (Si.Shared_Resources (Index1).Shared.task_tab.Entries (
                    Index2).Item =
                   A_Tcb.Tsk.Name)
               then

                  Si.Shared_Resources (Index1).Shared.state :=
                    Si.Shared_Resources (Index1).Shared.state + 1;
                  for I in  0 .. Si.Shared_Resources (Index1).Nb_Allocated -
                                 1
                  loop
                     if Si.Shared_Resources (Index1).Allocated_By (I) =
                        A_Tcb.Tsk.Name
                     then

                        Si.Shared_Resources (Index1).Allocated_By (I) :=
                          Si.Shared_Resources (Index1).Allocated_By (
                          Si.Shared_Resources (Index1).Nb_Allocated -
                                                                     1);
                        Si.Shared_Resources (Index1).Nb_Allocated     :=
                          Si.Shared_Resources (Index1).Nb_Allocated - 1;

                        if Event_To_Generate (Release_Resource) then
                           A_Item                  :=
                             new Time_Unit_Event (Release_Resource);
                           A_Item.Release_Resource :=
                             Si.Shared_Resources (Index1).Shared;
                           A_Item.Release_Task     := A_Tcb.Tsk;
                           Add (Result.all, A_Item, Current_Time);
                        end if;

                        exit;
                     end if;
                  end loop;

               end if;

            end loop;

         end loop;
      end if;
   end Release_Resource;

   procedure Allocate_Resource
     (My_Scheduler      : in out Generic_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : Tcb_Ptr;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table)
   is

      Current_Capacity : Natural := 0;
      A_Item           : Time_Unit_Event_Ptr;

   begin

      if (Si.Number_Of_Resources > 0) then
         Current_Capacity := A_Tcb.Tsk.Capacity - A_Tcb.Rest_Of_Capacity + 1;

         for Index1 in  0 .. Si.Number_Of_Resources - 1 loop

            -- For each resource, check if the task request the resource
            -- and if the resource is free
            -- otherwise, block the task
            --
            for Index2 in  0 .. Si.Shared_Resources (Index1).Shared.task_tab.
                 Nb_Entries -
                                1
            loop

               if (Current_Capacity =
                   Si.Shared_Resources (Index1).Shared.task_tab.Entries (
                    Index2).Data.Task_Begin) and
                  (Si.Shared_Resources (Index1).Shared.task_tab.Entries (
                    Index2).Item =
                   A_Tcb.Tsk.Name)
               then

                  Si.Shared_Resources (Index1).Shared.state :=
                    Si.Shared_Resources (Index1).Shared.state - 1;

                  Si.Shared_Resources (Index1).Allocated_By (
                    Si.Shared_Resources (Index1).Nb_Allocated) :=
                    A_Tcb.Tsk.Name;
                  Si.Shared_Resources (Index1).Nb_Allocated :=
                    Si.Shared_Resources (Index1).Nb_Allocated + 1;

                  if Event_To_Generate (Allocate_Resource) then
                     A_Item                   :=
                       new Time_Unit_Event (Allocate_Resource);
                     A_Item.Allocate_Resource :=
                       Si.Shared_Resources (Index1).Shared;
                     A_Item.Allocate_Task     := A_Tcb.Tsk;
                     Add (Result.all, A_Item, Current_Time);
                  end if;

               end if;

            end loop;

         end loop;
      end if;
   end Allocate_Resource;

   procedure Check_Resource
     (My_Scheduler      : in out Generic_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : in Tcb_Ptr;
      Is_Ready          : out Boolean;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table)
   is

      Current_Capacity : Natural := 0;
      A_Item           : Time_Unit_Event_Ptr;

   begin

      Is_Ready := True;

      if (Si.Number_Of_Resources > 0) then

         Current_Capacity := A_Tcb.Tsk.Capacity - A_Tcb.Rest_Of_Capacity + 1;

         for Index1 in  0 .. Si.Number_Of_Resources - 1 loop

            -- For each resource, check if the task request the resource
            -- and if the resource is free
            --
            for Index2 in  0 .. Si.Shared_Resources (Index1).Shared.task_tab.
                 Nb_Entries -
                                1
            loop

               if (Current_Capacity =
                   Si.Shared_Resources (Index1).Shared.task_tab.Entries (
                    Index2).Data.Task_Begin) and
                  (Si.Shared_Resources (Index1).Shared.task_tab.Entries (
                    Index2).Item =
                   A_Tcb.Tsk.Name)
               then

                  if (Si.Shared_Resources (Index1).Shared.state <= 0) then
                     Is_Ready := False;
                  end if;

                  if not Is_Ready then

                     if Event_To_Generate (Wait_For_Resource) then
                        A_Item               :=
                          new Time_Unit_Event (Wait_For_Resource);
                        A_Item.Wait_Resource :=
                          Si.Shared_Resources (Index1).Shared;
                        A_Item.Wait_Task     := A_Tcb.Tsk;
                        Add (Result.all, A_Item, Current_Time);
                     end if;

                  end if;

               end if;

            end loop;

         end loop;
      end if;
   end Check_Resource;

   procedure Set_Quantum
     (My_Scheduler : in out Generic_Scheduler'Class;
      Q            : in Natural)
   is
   begin
      My_Scheduler.Quantum := Q;
   end Set_Quantum;

   function Get_Quantum
     (My_Scheduler : in Generic_Scheduler'Class)
      return         Natural
   is
   begin
      return My_Scheduler.Quantum;
   end Get_Quantum;

   function Get_Quantum
     (My_Scheduler : in Generic_Scheduler'Class)
      return         Unbounded_String
   is
   begin
      return To_Unbounded_String (My_Scheduler.Quantum'Img);
   end Get_Quantum;

   function Get_Quantum
     (My_Scheduler : in Generic_Scheduler'Class)
      return         String
   is
   begin
      return My_Scheduler.Quantum'Img;
   end Get_Quantum;

   function Build_Resource
     (My_Scheduler : in Generic_Scheduler;
      A_Resource   : Generic_Resource_Ptr)
      return         Shared_Resource_Ptr
   is
      New_A_Resource : Shared_Resource_Ptr;

   begin

      New_A_Resource        := new Shared_Resource;
      New_A_Resource.Shared := A_Resource;

      return New_A_Resource;

   end Build_Resource;

   procedure Initialize (S : in out Scheduling_Information) is
   begin
      S.Number_Of_Tasks     := 0;
      S.Number_Of_Resources := 0;
   end Initialize;

   -- If this function return "True", it means that
   -- the task "A_Tcb" can be selected because of all its
   -- predecessors have ended their activation
   --
   function Check_Precedencies
     (Si           : in Scheduling_Information;
      Deps         : Tasks_Dependencies_Ptr;
      Current_Time : in Natural;
      A_Tcb        : Tcb_Ptr)
      return         Boolean
   is
      Has_Message_To_Receive : Boolean := False;
      Wait_Message           : Boolean := False;

      Previous : Tasks_Set;
      Left     : Generic_Task_Ptr;
      Ite1     : Tasks_Iterator;

      A_Half_Dep_Ptr : Dependency_Ptr;
      A_Message      : Message_Ptr;

      My_Iterator : Tasks_Dependencies_Iterator;
      List_Ite    : Message_Scheduling_Information_Iterator;
      Msg         : Message_Scheduling_Information_Ptr;

   begin

      -------------------------------------------------
      -- Check if the task has a task precedency and
      -- if they are met
      -------------------------------------------------

      if Has_Predecessor (Deps, A_Tcb.Tsk) then

         Reset (Previous);
         Previous := Get_Predecessors_List (Deps, A_Tcb.Tsk);

         Reset_Iterator (Previous, Ite1);

         for J in  0 .. Get_Number_Of_Elements (Previous) - 1 loop

            Current_Element (Previous, Left, Ite1);

            for I in  0 .. Si.Number_Of_Tasks - 1 loop
               if Si.Tcbs (I).Tsk.Name = Left.Name then

                  if (Si.Tcbs (I).Activation <= A_Tcb.Activation) then
                     return False;
                  end if;

                  -- Same activation ... be sure the previous task
                  -- is ended
                  --
                  if (Si.Tcbs (I).Activation = A_Tcb.Activation + 1) and
                     (Si.Tcbs (I).End_Time > Current_Time)
                  then
                     return False;
                  end if;

               end if;
            end loop;

            Next_Element (Previous, Ite1);
         end loop;
      end if;

      -------------------------------------------------
      -- Check communication dependencies first
      -- A "receiver" has to read at most one message
      -- to be activated
      -------------------------------------------------

      if not Is_Empty (Deps.all) then

         Reset_Iterator (Deps.all, My_Iterator);
         loop
            Current_Element (Deps.all, A_Half_Dep_Ptr, My_Iterator);

            -- Is the task concerned ?
            -- Is it a communication dependency ?
            -- Is it a receiving operation ?
            --
            if (A_Half_Dep_Ptr.Depend_Type = Communication_Dependency) then
               if (A_Half_Dep_Ptr.Communication_Dependent_Task.Name =
                   A_Tcb.Tsk.Name) and
                  (A_Half_Dep_Ptr.Communication_Orientation =
                   From_Object_To_Task)
               then

                  A_Message := A_Half_Dep_Ptr.Communication_Dependency_Object;

                  -- Is The task waiting for a message ?
                  --
                  if (A_Tcb.Rest_Of_Capacity = A_Tcb.Tsk.Capacity) then
                     Wait_Message := True;
                  end if;

                  -- Check the sended_messages to see if this
                  -- messages is just arrived
                  --
                  if not Is_Empty (Si.Sended_Messages) then
                     Reset_Head_Iterator (Si.Sended_Messages, List_Ite);
                     loop
                        Current_Element (Si.Sended_Messages, Msg, List_Ite);

                        -- Is it the right message ?
                        -- Is the message arrived at the receiving processor ?
                        --
                        if Wait_Message then
                           if (Msg.Sended_Message.Name = A_Message.Name) then
                              if (Current_Time >=
                                  (Msg.Send_Time + A_Message.Response_Time))
                              then
                                 Has_Message_To_Receive := True;
                              end if;
                           end if;
                        end if;

                        if Is_Tail_Element
                             (Si.Sended_Messages,
                              List_Ite)
                        then
                           exit;
                        end if;

                        Next_Element (Si.Sended_Messages, List_Ite);
                     end loop;
                  end if;
               end if;
            end if;

            exit when Is_Last_Element (Deps.all, My_Iterator);

            Next_Element (Deps.all, My_Iterator);

         end loop;
      end if;

      -- If the task wait for a message which is not
      -- arrived, do not wake up it
      --
      if Wait_Message and (not Has_Message_To_Receive) then
         return False;
      end if;

      -- All dependency constraints are met
      --
      return True;

   end Check_Precedencies;

   procedure Compute_Activation_Time
     (My_Scheduler : in Generic_Scheduler;
      Si           : in out Scheduling_Information;
      Elected      : in Tasks_Range;
      Value        : in out Natural)
   is

   begin
      Value := 0;
   end Compute_Activation_Time;

   procedure Put (M : in Message_Scheduling_Information_Ptr) is
   begin
      Put ("Message " & To_String (M.Sended_Message.Name));
      Put_Line (" ; send_time = " & M.Send_Time'Img);
   end Put;

   procedure Put (M : in Buffer_Scheduling_Information_Ptr) is
   begin
      Put ("Buffer " & To_String (M.Written_Buffer.Name));
      Put_Line (" ; current_size = " & M.Current_Size'Img);
   end Put;

   function Produce_Running_Task_Event
     (My_Scheduler : in Generic_Scheduler;
      A_Task       : in Tcb_Ptr)
      return         Time_Unit_Event_Ptr
   is
      A_Item : Time_Unit_Event_Ptr;

   begin
      A_Item              := new Time_Unit_Event (Running_Task);
      A_Item.Running_Task := A_Task.Tsk;
      return A_Item;
   end Produce_Running_Task_Event;

   function Export_Xml_Event_Write_To_Buffer
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " &
                An_Event.Write_Buffer.Name &
                " " &
                An_Event.Write_Task.Name &
                " " &
                An_Event.Write_Size'Img;
      return Result;
   end Export_Xml_Event_Write_To_Buffer;

   function Export_Xml_Event_Read_From_Buffer
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " &
                An_Event.Read_Buffer.Name &
                " " &
                An_Event.Read_Task.Name &
                " " &
                An_Event.Read_Size'Img;
      return Result;
   end Export_Xml_Event_Read_From_Buffer;

   function Export_Xml_Event_Running_Task
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " & An_Event.Running_Task.Name;
      return Result;
   end Export_Xml_Event_Running_Task;

   function Export_Xml_Event_Task_Activation
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " & An_Event.Activation_Task.Name;
      return Result;
   end Export_Xml_Event_Task_Activation;

   function Export_Xml_Event_Start_Of_Task_Capacity
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " & An_Event.Start_Task.Name;
      return Result;
   end Export_Xml_Event_Start_Of_Task_Capacity;

   function Export_Xml_Event_End_Of_Task_Capacity
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " & An_Event.End_Task.Name;
      return Result;
   end Export_Xml_Event_End_Of_Task_Capacity;

   function Export_Xml_Event_Send_Message
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " &
                An_Event.Send_Message.Name &
                " " &
                An_Event.Send_Task.Name;
      return Result;
   end Export_Xml_Event_Send_Message;

   function Export_Xml_Event_Receive_Message
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " &
                An_Event.Receive_Message.Name &
                " " &
                An_Event.Receive_Task.Name;
      return Result;
   end Export_Xml_Event_Receive_Message;

   function Export_Xml_Event_Allocate_Resource
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " &
                An_Event.Allocate_Resource.Name &
                " " &
                An_Event.Allocate_Task.Name;
      return Result;
   end Export_Xml_Event_Allocate_Resource;

   function Export_Xml_Event_Release_Resource
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " &
                An_Event.Release_Resource.Name &
                " " &
                An_Event.Release_Task.Name;
      return Result;
   end Export_Xml_Event_Release_Resource;

   function Export_Xml_Event_Wait_For_Resource
     (My_Scheduler : in Generic_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin
      Result := " " &
                An_Event.Wait_Resource.Name &
                " " &
                An_Event.Wait_Task.Name;
      return Result;
   end Export_Xml_Event_Wait_For_Resource;

end Scheduler;
