------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Events;            use Events;
with Translate;         use Translate;
with Unbounded_Strings; use Unbounded_Strings;
with Text_IO;           use Text_IO;
with Buffer_Set;        use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Resources;         use Resources;
with Natural_Util;      use Natural_Util;
use Natural_Util.Natural_List;

package body Scheduling_Sequences is

   procedure Initialize (B : out Buffer_Size_Item) is
   begin
      B.Time := 0;
      B.Size := 0;
   end Initialize;

   procedure Put (B : in Buffer_Size_Item) is
   begin
      Put (B.Time'Img & "/" & B.Size'Img);
   end Put;

   function Compute_Average_Buffer_Size_From_Simulation
     (Buff : in Buffer_Size_Table_Ptr)
      return Double
   is
      Sum : Double := 0.0;

   begin

      for I in  0 .. Buff.Nb_Entries - 1 loop
         Sum := Sum + Double (Buff.Entries (I).Size);
      end loop;

      return Sum / Double (Buff.Nb_Entries);

   end Compute_Average_Buffer_Size_From_Simulation;

   function Compute_Maximum_Buffer_Size_From_Simulation
     (Buff : in Buffer_Size_Table_Ptr)
      return Natural
   is
      Max : Natural := 0;
   begin

      for I in  0 .. Buff.Nb_Entries - 1 loop
         if Buff.Entries (I).Size > Max then
            Max := Buff.Entries (I).Size;
         end if;
      end loop;

      return Max;

   end Compute_Maximum_Buffer_Size_From_Simulation;

   procedure Compute_Buffer_Size_From_Simulation
     (Sched   : in Scheduling_Sequence_Ptr;
      My_Buff : in Buffer_Ptr;
      Buff    : in out Buffer_Size_Table_Ptr)
   is

      Current_Size : Natural := 0;
      A_Buff_Size  : Buffer_Size_Item;

   begin

      for I in  0 .. Sched.Nb_Entries - 1 loop

         if (Sched.Entries (I).Item.Event_Type = Read_From_Buffer) then
            if (Sched.Entries (I).Item.Read_Buffer.Name = My_Buff.Name) then
               A_Buff_Size.Time := Sched.Entries (I).Data;
               Current_Size     := Current_Size -
                                   Sched.Entries (I).Item.Read_Size;
               A_Buff_Size.Size := Current_Size;
               Add (Buff.all, A_Buff_Size);
            end if;
         end if;

         if (Sched.Entries (I).Item.Event_Type = Write_To_Buffer) then
            if (Sched.Entries (I).Item.Write_Buffer.Name = My_Buff.Name) then
               A_Buff_Size.Time := Sched.Entries (I).Data;
               Current_Size     := Current_Size +
                                   Sched.Entries (I).Item.Write_Size;
               A_Buff_Size.Size := Current_Size;
               Add (Buff.all, A_Buff_Size);
            end if;
         end if;

      end loop;
   end Compute_Buffer_Size_From_Simulation;

   function Compute_Maximum_Waiting_Time_From_Simulation
     (Buff                    : in Buffer_Size_Table_Ptr;
      Average_Consumer_Period : in Double)
      return                    Natural
   is
      Max : Natural := 0;
   begin

      for I in  0 .. Buff.Nb_Entries - 1 loop
         if Buff.Entries (I).Size > Max then
            Max := Buff.Entries (I).Size;
         end if;
      end loop;

      return (Max * Natural (Average_Consumer_Period));

   end Compute_Maximum_Waiting_Time_From_Simulation;

   function Compute_Average_Waiting_Time_From_Simulation
     (Buff                    : in Buffer_Size_Table_Ptr;
      Average_Consumer_Period : in Double)
      return                    Double
   is
      Sum : Double := 0.0;

   begin

      for I in  0 .. Buff.Nb_Entries - 1 loop
         Sum := Sum + Double (Buff.Entries (I).Size);
      end loop;

      return ((Sum / Double (Buff.Nb_Entries)) * Average_Consumer_Period);

   end Compute_Average_Waiting_Time_From_Simulation;

   procedure Cyclic_Response_Time_By_Simulation
     (My_Task : in Generic_Task_Ptr;
      Sched   : in Scheduling_Sequence_Ptr;
      Msg     : in out Unbounded_String;
      Max     : in out Natural;
      Min     : in out Natural;
      Average : in out Double)
   is

      Task_Occurence_Number        : Task_Occurence_Range := 0;
      Awoken_Task_Occurence_Number : Task_Occurence_Range := 0;
      Running_Awoken_Task_Index    : Task_Occurence_Range := 0;
      Next_My_Task_Occurence_Awake : Task_Occurence_Range := 0;

      A_Running_Task : Generic_Task_Ptr;
      Effective_End  : Natural := 0;

      Current_Date            : Natural := 0;
      Number_Of_Response_Time : Natural;
      My_Task_Occurence_Table : Task_Occurence_Table_Ptr;

   begin

      -- initialize all occurences
      -- dates are taken from scheduling table
      --
      My_Task_Occurence_Table := new Task_Occurence_Table;
      Min                     := Natural'Last;
      Max                     := Natural'First;
      Average                 := 0.0;
      Number_Of_Response_Time := 0;
      Msg                     := Empty_String;

      for J in  0 .. Sched.Nb_Entries - 1 loop

         if Sched.Entries (J).Item.Event_Type = Task_Activation then
            if Sched.Entries (J).Item.Activation_Task.Name =
               My_Task.Name
            then
               Initialize_Cyclic
                 (My_Task_Occurence_Table.Entries (Task_Occurence_Number),
                  My_Task,
                  Sched.Entries (J).Data);
               Task_Occurence_Number              := Task_Occurence_Number +
                                                     1;
               My_Task_Occurence_Table.Nb_Entries :=
                 My_Task_Occurence_Table.Nb_Entries + 1;
            end if;
         end if;
      end loop;

      for T in  0 .. Sched.Nb_Entries - 1 loop

         -- the current date
         --
         Current_Date := Sched.Entries (T).Data;

         -- the current task running is the running occurence
         --
         case Sched.Entries (T).Item.Event_Type is

            when Running_Task =>
               A_Running_Task := Sched.Entries (T).Item.Running_Task;

               -- New occurence of My_task
               --
               if (Current_Date =
                   My_Task_Occurence_Table.Entries (
                    Next_My_Task_Occurence_Awake).Arrival_Date)
               then

                  -- one more task is awoken
                  --
                  Awoken_Task_Occurence_Number :=
                    Awoken_Task_Occurence_Number + 1;

                  -- activate the occurence
                  --
                  My_Task_Occurence_Table.Entries (
                    Next_My_Task_Occurence_Awake).Is_Active := True;

                  -- if there is not active occurence, run the new occurence
                  --
                  if (Awoken_Task_Occurence_Number = 0) then
                     Running_Awoken_Task_Index :=
                       Next_My_Task_Occurence_Awake;
                  end if;

                  -- the next occurence to awake
                  --
                  Next_My_Task_Occurence_Awake :=
                    Next_My_Task_Occurence_Awake + 1;

               end if;

               -- is there any activated occurences
               --
               if (Awoken_Task_Occurence_Number > 0) then

                  -- increment response time for all active occurences
                  --
                  for I_Occurence in  0 .. My_Task_Occurence_Table.Nb_Entries -
                                           1
                  loop
                     if (
                          My_Task_Occurence_Table.Entries (I_Occurence).
                          Is_Active)
                     then
                        My_Task_Occurence_Table.Entries (I_Occurence).
                          Response_Time :=
                          My_Task_Occurence_Table.Entries (I_Occurence).
                          Response_Time +
                          1;
                     end if;
                  end loop;

                  if (A_Running_Task.Name = My_Task.Name) then

                     My_Task_Occurence_Table.Entries (
                       Running_Awoken_Task_Index).Residual_Capacity :=
                       My_Task_Occurence_Table.Entries (
                       Running_Awoken_Task_Index).Residual_Capacity -
                       1;

                     -- is The Current Occurence finished
                     --
                     if (My_Task_Occurence_Table.Entries (
                          Running_Awoken_Task_Index).Residual_Capacity =
                         0)
                     then
                        Effective_End := Current_Date + 1;

                        -- Check deadline missed
                        --
                        if (My_Task_Occurence_Table.Entries (
                             Running_Awoken_Task_Index).Response_Time >
                            My_Task.Deadline)
                        then
                           Msg :=
                             Msg &
                             Lb_Comma &
                             Lb_Check_Deadline (Current_Language) &
                             My_Task_Occurence_Table.Entries (
                             Running_Awoken_Task_Index).Absolute_Deadline'Img &
                             Lb_Completion_Time (Current_Language) &
                             Effective_End'Img &
                             To_Unbounded_String (")");

                        end if;

                        -- Update worst/best/average case response time
                        --
                        if (My_Task_Occurence_Table.Entries (
                             Running_Awoken_Task_Index).Response_Time <
                            Min)
                        then
                           Min :=
                             My_Task_Occurence_Table.Entries (
                             Running_Awoken_Task_Index).Response_Time;
                        end if;
                        if (My_Task_Occurence_Table.Entries (
                             Running_Awoken_Task_Index).Response_Time >
                            Max)
                        then
                           Max :=
                             My_Task_Occurence_Table.Entries (
                             Running_Awoken_Task_Index).Response_Time;
                        end if;
                        Average                 := Average +
                                                   Double (
                          My_Task_Occurence_Table.Entries (
                          Running_Awoken_Task_Index).Response_Time);
                        Number_Of_Response_Time := Number_Of_Response_Time +
                                                   1;

                        -- desactivate the current occurence
                        --
                        My_Task_Occurence_Table.Entries (
                          Running_Awoken_Task_Index).Is_Active := False;

                        -- give hand to the next awoken occurence
                        --
                        Running_Awoken_Task_Index :=
                          Running_Awoken_Task_Index + 1;

                        -- one less Awoken Occurence
                        --
                        Awoken_Task_Occurence_Number :=
                          Awoken_Task_Occurence_Number - 1;

                     end if;

                  end if;

               end if;

            when Task_Activation =>
               A_Running_Task := Sched.Entries (T).Item.Activation_Task;

            when others =>
               null;

         end case;

      end loop;

      if Number_Of_Response_Time > 0 then
         Average := Average / Double (Number_Of_Response_Time);
      end if;
      if Min = Natural'Last then
         Min := 0;
      end if;

   end Cyclic_Response_Time_By_Simulation;

   procedure Acyclic_Response_Time_By_Simulation
     (My_Task : in Aperiodic_Task_Ptr;
      Sched   : in Scheduling_Sequence_Ptr;
      Msg     : in out Unbounded_String;
      Max     : in out Natural;
      Min     : in out Natural;
      Average : in out Double)
   is

      My_Task_Occurence : Task_Occurence;
      A_Running_Task    : Generic_Task_Ptr;
      Effective_End     : Natural := 0;

      Current_Date : Natural := 0;

   begin

      Initialize_Acyclic (My_Task_Occurence, My_Task);
      Msg := To_Unbounded_String ("");

      for T in  0 .. Sched.Nb_Entries - 1 loop
         -- the current date
         --
         Current_Date := Sched.Entries (T).Data;

         -- the current task running is the running occurence
         --
         case Sched.Entries (T).Item.Event_Type is
            when Running_Task =>
               A_Running_Task := Sched.Entries (T).Item.Running_Task;

               -- if the occurence of My_task starts
               --
               if (Current_Date = My_Task_Occurence.Arrival_Date) then
                  My_Task_Occurence.Is_Active := True;
               end if;

               if (A_Running_Task.Name = My_Task.Name) then

                  My_Task_Occurence.Residual_Capacity :=
                    My_Task_Occurence.Residual_Capacity - 1;

                  -- is The Current Occurence finished
                  --
                  if (My_Task_Occurence.Residual_Capacity = 0) then

                     My_Task_Occurence.Is_Active := False;

                     Effective_End := Current_Date + 1;

                     -- Check deadline missed
                     --
                     if (Effective_End >
                         My_Task_Occurence.Absolute_Deadline)
                     then
                        Msg :=
                          Msg &
                          Lb_Comma &
                          Lb_Check_Deadline (Current_Language) &
                          My_Task_Occurence.Absolute_Deadline'Img &
                          Lb_Completion_Time (Current_Language) &
                          Effective_End'Img &
                          To_Unbounded_String (")");
                     end if;

                  end if;

               end if;

               -- increment response time if the occurence is Active
               --
               if (My_Task_Occurence.Is_Active) then
                  My_Task_Occurence.Response_Time :=
                    My_Task_Occurence.Response_Time + 1;
               end if;
            -- if the current task running is the occurence of My_Task
            --
            when Task_Activation =>
               A_Running_Task := Sched.Entries (T).Item.Activation_Task;
            when others =>
               null;
         end case;

      end loop;

      if Effective_End > 0 then
         Average := Double (Effective_End - My_Task_Occurence.Arrival_Date);
         Min     := Effective_End - My_Task_Occurence.Arrival_Date;
         Max     := Effective_End - My_Task_Occurence.Arrival_Date;
      else
         Average := 0.0;
         Min     := 0;
         Max     := 0;
      end if;

   end Acyclic_Response_Time_By_Simulation;

   procedure Compute_Response_Time_Distribution
     (My_Task : in Generic_Task_Ptr;
      Sched   : in Scheduling_Sequence_Ptr;
      Result  : in out Density_Table)
   is

      Awoken_Task_Occurence_Number : Task_Occurence_Range := 0;
      Task_Occurence_Number        : Task_Occurence_Range := 0;
      Running_Awoken_Task_Index    : Task_Occurence_Range := 0;
      Next_My_Task_Occurence_Awake : Task_Occurence_Range := 0;
      A_Running_Task               : Generic_Task_Ptr;
      Current_Date                 : Natural              := 0;
      Found                        : Boolean              := False;
      A_Density                    : Density_Item;
      Number_Of_Response_Time      : Double;

      My_Task_Occurence_Table : Task_Occurence_Table_Ptr;

   begin

      My_Task_Occurence_Table := new Task_Occurence_Table;

      -- initialize all occurences
      if My_Task.Task_Type = Aperiodic_Type then

         Initialize_Acyclic
           (My_Task_Occurence_Table.Entries (Task_Occurence_Number),
            Aperiodic_Task_Ptr (My_Task));
         Task_Occurence_Number := Task_Occurence_Number + 1;

      else

         -- initialize all occurences
         -- dates are taken from scheduling table
         --
         for J in  0 .. Sched.Nb_Entries - 1 loop

            if Sched.Entries (J).Item.Event_Type = Task_Activation then
               if Sched.Entries (J).Item.Activation_Task.Name =
                  My_Task.Name
               then
                  Initialize_Cyclic
                    (My_Task_Occurence_Table.Entries (Task_Occurence_Number),
                     My_Task,
                     Sched.Entries (J).Data);
                  Task_Occurence_Number              := Task_Occurence_Number +
                                                        1;
                  My_Task_Occurence_Table.Nb_Entries :=
                    My_Task_Occurence_Table.Nb_Entries + 1;
               end if;
            end if;
         end loop;

      end if;

      for T in  0 .. Sched.Nb_Entries - 1 loop

         if Sched.Entries (T).Item.Event_Type = Running_Task then

            -- the current date
            Current_Date := Sched.Entries (T).Data;

            -- the current task running is the running occurence
            A_Running_Task := Sched.Entries (T).Item.Running_Task;
            -- New occurence of My_task
            if (Current_Date =
                My_Task_Occurence_Table.Entries (Next_My_Task_Occurence_Awake).
                 Arrival_Date)
            then

               -- one more task is awoken
               Awoken_Task_Occurence_Number := Awoken_Task_Occurence_Number +
                                               1;

               -- activate the occurence
               My_Task_Occurence_Table.Entries (Next_My_Task_Occurence_Awake).
                 Is_Active := True;

               -- if there is not active occurence, run the new occurence
               if (Awoken_Task_Occurence_Number = 0) then
                  Running_Awoken_Task_Index := Next_My_Task_Occurence_Awake;
               end if;

               -- the next occurence to awake
               Next_My_Task_Occurence_Awake := Next_My_Task_Occurence_Awake +
                                               1;

            end if;

            -- is there any activated occurences
            if (Awoken_Task_Occurence_Number > 0) then

               -- increment response time for all active occurences
               --
               for I_Occurence in  0 .. My_Task_Occurence_Table.Nb_Entries -
                                        1
               loop
                  if (My_Task_Occurence_Table.Entries (I_Occurence).Is_Active)
                  then
                     My_Task_Occurence_Table.Entries (I_Occurence).
                       Response_Time :=
                       My_Task_Occurence_Table.Entries (I_Occurence).
                       Response_Time +
                       1;
                  end if;
               end loop;

               if (A_Running_Task.Name = My_Task.Name) then

                  My_Task_Occurence_Table.Entries (Running_Awoken_Task_Index).
                    Residual_Capacity :=
                    My_Task_Occurence_Table.Entries (Running_Awoken_Task_Index)
.Residual_Capacity -
                    1;

                  -- is The Current Occurence finished
                  if (My_Task_Occurence_Table.Entries (
                       Running_Awoken_Task_Index).Residual_Capacity =
                      0)
                  then

                     -- We have a response time. We look for the
                     --   density table its response time is already
                     -- known. If so, increment the frequency, if
                     -- not, add it to the table
                     --
                     Found := False;
                     for K in  0 .. Result.Nb_Entries - 1 loop
                        if Result.Entries (K).Response_Time =
                           My_Task_Occurence_Table.Entries (
                             Running_Awoken_Task_Index).Response_Time
                        then
                           Result.Entries (K).Probability :=
                             Result.Entries (K).Probability + 1.0;
                           Found                          := True;
                        end if;
                     end loop;
                     if not Found then
                        A_Density.Response_Time :=
                          My_Task_Occurence_Table.Entries (
                          Running_Awoken_Task_Index).Response_Time;
                        A_Density.Probability   := 1.0;
                        Add (Result, A_Density);
                     end if;

                     -- desactivate the current occurence
                     --
                     My_Task_Occurence_Table.Entries (
                       Running_Awoken_Task_Index).Is_Active := False;

                     -- give hand to the next awoken occurence
                     --
                     Running_Awoken_Task_Index := Running_Awoken_Task_Index +
                                                  1;

                     -- one less Awoken Occurence
                     --
                     Awoken_Task_Occurence_Number :=
                       Awoken_Task_Occurence_Number - 1;

                  end if;

               end if;

            end if;

         end if;

      end loop;

      -- Now, compute density : find number of activation done and
      -- divide all entries of the table by this number
      --
      Number_Of_Response_Time := 0.0;
      for K in  0 .. Result.Nb_Entries - 1 loop
         Number_Of_Response_Time := Number_Of_Response_Time +
                                    Result.Entries (K).Probability;
      end loop;
      for K in  0 .. Result.Nb_Entries - 1 loop
         Result.Entries (K).Probability := Result.Entries (K).Probability /
                                           Number_Of_Response_Time;
      end loop;

   end Compute_Response_Time_Distribution;

   procedure Initialize (D : out Density_Item) is
   begin
      D.Response_Time := 0;
      D.Probability   := 0.0;
   end Initialize;

   procedure Put (D : in Density_Item) is
   begin
      Put (D.Response_Time'Img & " " & D.Probability'Img);
   end Put;

   procedure Compute_Response_Time_Distribution
     (My_Tasks       : in Tasks_Set;
      Sched          : in Scheduling_Sequence_Ptr;
      Processor_Name : Unbounded_String;
      Result         : in out Densities_Table)
   is

      A_Task          : Generic_Task_Ptr;
      My_Iterator     : Tasks_Iterator;
      Working_Density : Density_Table;

   begin

      Initialize (Result);

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (Processor_Name = A_Task.Cpu_Name) then
            Initialize (Working_Density);
            Compute_Response_Time_Distribution
              (A_Task,
               Sched,
               Working_Density);
            Add (Result, A_Task, Working_Density);

         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

   end Compute_Response_Time_Distribution;

   procedure Response_Time_By_Simulation
     (My_Task : Generic_Task_Ptr;
      Sched   : in Scheduling_Sequence_Ptr;
      Msg     : in out Unbounded_String;
      Max     : in out Natural;
      Min     : in out Natural;
      Average : in out Double)
   is

   begin

      case My_Task.Task_Type is
         when Aperiodic_Type =>
            Acyclic_Response_Time_By_Simulation
              (Aperiodic_Task_Ptr (My_Task),
               Sched,
               Msg,
               Max,
               Min,
               Average);
         when others =>
            Cyclic_Response_Time_By_Simulation
              (My_Task,
               Sched,
               Msg,
               Max,
               Min,
               Average);
      end case;
   end Response_Time_By_Simulation;

   procedure All_Response_Time_By_Simulation
     (My_Task : in Generic_Task_Ptr;
      Sched   : in Scheduling_Sequence_Ptr;
      Result  : in out Task_Occurence_Table_Ptr)
   is

      Result_Number                : Task_Occurence_Range := 0;
      Task_Occurence_Number        : Task_Occurence_Range := 0;
      Awoken_Task_Occurence_Number : Task_Occurence_Range := 0;
      Running_Awoken_Task_Index    : Task_Occurence_Range := 0;
      Next_My_Task_Occurence_Awake : Task_Occurence_Range := 0;
      A_Running_Task               : Generic_Task_Ptr;
      Current_Date                 : Natural              := 0;
      My_Task_Occurence_Table      : Task_Occurence_Table_Ptr;

   begin

      -- initialize all occurences
      -- dates are taken from scheduling table
      --
      My_Task_Occurence_Table := new Task_Occurence_Table;
      for J in  0 .. Sched.Nb_Entries - 1 loop

         if Sched.Entries (J).Item.Event_Type = Task_Activation then
            if Sched.Entries (J).Item.Activation_Task.Name =
               My_Task.Name
            then
               Initialize_Cyclic
                 (My_Task_Occurence_Table.Entries (Task_Occurence_Number),
                  My_Task,
                  Sched.Entries (J).Data);
               Task_Occurence_Number := Task_Occurence_Number + 1;
            end if;
         end if;
      end loop;

      for T in  0 .. Sched.Nb_Entries - 1 loop

         -- the current date
         Current_Date := Sched.Entries (T).Data;

         -- the current task running is the running occurence
         case Sched.Entries (T).Item.Event_Type is

            when Running_Task =>
               A_Running_Task := Sched.Entries (T).Item.Running_Task;
               --               Put();

               -- New occurence of My_task
               if (Current_Date =
                   My_Task_Occurence_Table.Entries (
                    Next_My_Task_Occurence_Awake).Arrival_Date)
               then

                  -- one more task is awoken
                  Awoken_Task_Occurence_Number :=
                    Awoken_Task_Occurence_Number + 1;

                  -- activate the occurence
                  My_Task_Occurence_Table.Entries (
                    Next_My_Task_Occurence_Awake).Is_Active := True;

                  -- if there is not active occurence, run the new occurence
                  if (Awoken_Task_Occurence_Number = 0) then
                     Running_Awoken_Task_Index :=
                       Next_My_Task_Occurence_Awake;
                  end if;

                  -- the next occurence to awake
                  Next_My_Task_Occurence_Awake :=
                    Next_My_Task_Occurence_Awake + 1;

               end if;

               -- is there any activated occurences
               if (Awoken_Task_Occurence_Number > 0) then

                  -- increment response time for all active occurences
                  -- for I_Occurence in Running_Awoken_Task_Index ..
                  -- (Running_Awoken_Task_Index +
                  -- Awoken_Task_Occurence_Number - 1) loop
                  for I_Occurence in  Task_Occurence_Range loop
                     if (
                          My_Task_Occurence_Table.Entries (I_Occurence).
                          Is_Active)
                     then
                        My_Task_Occurence_Table.Entries (I_Occurence).
                          Response_Time :=
                          My_Task_Occurence_Table.Entries (I_Occurence).
                          Response_Time +
                          1;
                     end if;
                  end loop;

                  if (A_Running_Task.Name = My_Task.Name) then

                     My_Task_Occurence_Table.Entries (
                       Running_Awoken_Task_Index).Residual_Capacity :=
                       My_Task_Occurence_Table.Entries (
                       Running_Awoken_Task_Index).Residual_Capacity -
                       1;

                     -- is The Current Occurence finished
                     if (My_Task_Occurence_Table.Entries (
                          Running_Awoken_Task_Index).Residual_Capacity =
                         0)
                     then

                        -- store the response time
                        Result.Nb_Entries                            :=
                          Result.Nb_Entries + 1;
                        Result.Entries (Result_Number).Response_Time :=
                          My_Task_Occurence_Table.Entries (
                          Running_Awoken_Task_Index).Response_Time;
                        Result_Number                                :=
                          Result_Number + 1;

                        -- desactivate the current occurence
                        My_Task_Occurence_Table.Entries (
                          Running_Awoken_Task_Index).Is_Active := False;

                        -- give hand to the next awoken occurence
                        Running_Awoken_Task_Index :=
                          Running_Awoken_Task_Index + 1;

                        -- one less Awoken Occurence
                        --
                        Awoken_Task_Occurence_Number :=
                          Awoken_Task_Occurence_Number - 1;

                     end if;

                  end if;

               end if;

            when Task_Activation =>
               A_Running_Task := Sched.Entries (T).Item.Activation_Task;

            when others =>
               null;
         end case;

      end loop;

      Free (My_Task_Occurence_Table);

   end All_Response_Time_By_Simulation;

   procedure Blocking_Time_From_Simulation
     (My_Task      : in Generic_Task_Ptr;
      My_Resources : in Resources_Set;
      Sched        : in Scheduling_Sequence_Ptr;
      Max          : in out Natural;
      Min          : in out Natural;
      Average      : in out Double)
   is

      A_Resource   : Generic_Resource_Ptr;
      Resource_Ite : Resources_Iterator;

      Running_Task : Generic_Task_Ptr;

      Event_List : Time_Unit_List.List;
      An_Event   : Time_Unit_Event_Ptr;
      Wait_Event : Time_Unit_Event_Ptr;
      Event_Ite  : Time_Unit_List.Iterator;

      Date_List : Natural_List.List;
      A_Date    : Natural_Ptr;
      Wait_Date : Natural_Ptr;
      Date_Ite  : Natural_List.Iterator;

      Nb_Blocking_Time : Natural;

   begin

      Max              := 0;
      Min              := Natural'Last;
      Average          := 0.0;
      Nb_Blocking_Time := 0;

      for T in  0 .. Sched.Nb_Entries - 1 loop

         -- Store events on shared resources
         --
         case Sched.Entries (T).Item.Event_Type is
            when Wait_For_Resource =>
               Running_Task := Sched.Entries (T).Item.Wait_Task;
               if Running_Task.Name = My_Task.Name then
                  Add (Event_List, Sched.Entries (T).Item);

                  -- the current date
                  --
                  A_Date     := new Natural;
                  A_Date.all := Sched.Entries (T).Data;
                  Add (Date_List, A_Date);
               end if;

            when Allocate_Resource =>
               Running_Task := Sched.Entries (T).Item.Allocate_Task;
               if Running_Task.Name = My_Task.Name then
                  Add (Event_List, Sched.Entries (T).Item);

                  -- the current date
                  --
                  A_Date     := new Natural;
                  A_Date.all := Sched.Entries (T).Data;
                  Add (Date_List, A_Date);
               end if;
            when others =>
               null;
         end case;
      end loop;

      -- compute maximum, minimum and average blocking time
      --
      if not Is_Empty (Event_List) then

         Reset_Iterator (My_Resources, Resource_Ite);
         loop
            Current_Element (My_Resources, A_Resource, Resource_Ite);

            -- for each resource allocation, find the corresponding
            -- wait resource event
            --

            Reset_Tail_Iterator (Event_List, Event_Ite);
            Reset_Tail_Iterator (Date_List, Date_Ite);

            loop
               Current_Element (Event_List, An_Event, Event_Ite);

               Current_Element (Date_List, A_Date, Date_Ite);

               -- first, find a wait event for the current resource
               --
               if (An_Event.Event_Type = Wait_For_Resource) and
                  (Wait_Event = null)
               then
                  if (An_Event.Wait_Resource.Name = A_Resource.Name) then
                     Wait_Event := An_Event;
                     Wait_Date  := A_Date;
                  end if;
               end if;
               if (An_Event.Event_Type = Allocate_Resource) and
                  (Wait_Event /= null)
               then
                  Max              :=
                     Natural'Max (Max, (A_Date.all - Wait_Date.all));
                  Min              :=
                     Natural'Min (Min, (A_Date.all - Wait_Date.all));
                  Average          := Average +
                                      Double (A_Date.all - Wait_Date.all);
                  Nb_Blocking_Time := Nb_Blocking_Time + 1;
                  Wait_Event       := null;
               end if;

               exit when Is_Head_Element (Event_List, Event_Ite);

               Previous_Element (Event_List, Event_Ite);
               Previous_Element (Date_List, Date_Ite);
            end loop;

            exit when Is_Last_Element (My_Resources, Resource_Ite);

            Next_Element (My_Resources, Resource_Ite);
         end loop;
      end if;

      if Min = Natural'Last then
         Min := 0;
      end if;
      if Nb_Blocking_Time /= 0 then
         Average := Average / Double (Nb_Blocking_Time);
      end if;

   end Blocking_Time_From_Simulation;

   procedure Initialize (D : in out Deadlock_Item) is
   begin
      D.Time          := 0;
      D.Task_Name     := Empty_String;
      D.Resource_Name := Empty_String;
   end Initialize;

   procedure Put (D : in Deadlock_Item) is
   begin
      Put ("Time : " & D.Time'Img & " ; ");
      if D.Task_Name /= Empty_String then
         Put (To_String (D.Task_Name & " "));
      end if;
      Put (" ; ");
      if D.Resource_Name /= Empty_String then
         Put (To_String (D.Resource_Name & " "));
      end if;
      New_Line;
   end Put;

   procedure Put (D : in Deadlock_Item_Ptr) is
   begin
      Put (D.all);
   end Put;

   procedure Initialize (D : in out Priority_Inversion_Item) is
   begin
      D.Start_Time    := 0;
      D.End_Time      := 0;
      D.Task_Name     := Empty_String;
      D.Resource_Name := Empty_String;
   end Initialize;

   procedure Put (D : in Priority_Inversion_Item) is
   begin
      Put (" Start/end time  = " & D.Start_Time'Img & D.End_Time'Img);
      Put
        (" ; Task/Resource name = " &
         To_String (D.Task_Name) &
         To_String (D.Resource_Name));
      New_Line;
   end Put;

   procedure Put (D : in Priority_Inversion_Item_Ptr) is
   begin
      Put (D.all);
   end Put;

   function Deadlock_From_Simulation
     (My_Tasks       : in Tasks_Set;
      My_Resources   : in Resources_Set;
      Processor_Name : in Unbounded_String;
      Sched          : in Scheduling_Sequence_Ptr)
      return           Deadlock_List
   is
      Result : Deadlock_List;

      -- Variables required to scan the resource and task
      --
      Item          : Deadlock_Item_Ptr;
      Item_Task     : Generic_Task_Ptr;
      Item_Resource : Generic_Resource_Ptr;

      -- To store number of Tasks
      --
      Number_Of_Tasks : constant Tasks_Range :=
         Get_Number_Of_Elements (My_Tasks);

      -- To store number of resource
      --
      Number_Of_Resources : constant Resources_Range :=
         Get_Number_Of_Elements (My_Resources);

      -- Stamp to represent the tree graphic for the allocated resources and
      --the awaited resources
      --
      Stamp_For_Allocate :
        array (Tasks_Range'Range, Resources_Range' Range) of Integer;

      Stamp_For_Wait :
        array (Tasks_Range'Range, Resources_Range' Range) of Integer;

      --Tables of names of Tasks
      --
      Table_Name_Tasks : array (Tasks_Range'Range) of Unbounded_String;

      -- Table of names of resources
      --
      Table_Name_Resources :
        array (Resources_Range'Range) of Unbounded_String;

      -- Tables to store the indices of the tasks and the resources
      --
      Table_Index_Task     :
        array (Natural range 0 .. Framework_Config.Max_Resources - 1)
               of Tasks_Range;
      Table_Index_Resource :
        array (Natural range 0 .. Framework_Config.Max_Resources - 1)
               of Resources_Range;

      -- Index
      --
      Current_Time   : Natural         := 0;
      Index_Task     : Tasks_Range     := 200;
      Index_Resource : Resources_Range := 30;

      --  booleen to say if there is deadlock
      --
      Ok : Boolean;

      -- procedure being used to find the index of a task
      --
      procedure Index_Of_Tasks
        (Name  : in out Unbounded_String;
         Index : in out Tasks_Range)
      is
      begin

         for I in  0 .. Number_Of_Tasks - 1 loop
            if (Table_Name_Tasks (I) = Name) then
               Index := I;
            end if;
         end loop;

      end Index_Of_Tasks;

      -- procedure being used to find the index of a resource
      --

      procedure Index_Of_Resources
        (Name  : in out Unbounded_String;
         Index : in out Resources_Range)
      is
      begin

         for I in  0 .. Number_Of_Resources - 1 loop
            if (Table_Name_Resources (I) = Name) then
               Index := I;
            end if;
         end loop;

      end Index_Of_Resources;

      --  procedure being used to find if a resource is held by a task
      --

      procedure Seek_Of_Task
        (R     : in out Resources_Range;
         Index : in out Tasks_Range)
      is
      begin

         for I in  0 .. Number_Of_Tasks - 1 loop
            if (Stamp_For_Allocate (I, R) = 1) then
               Index := I;
            end if;
         end loop;

      end Seek_Of_Task;

      --  procedure being used to find if a task awaits a resource
      --

      procedure Seek_Of_Resource
        (T     : in out Tasks_Range;
         Index : in out Resources_Range)
      is
      begin

         for I in  0 .. Number_Of_Resources - 1 loop
            if (Stamp_For_Wait (T, I) = 1) then
               Index := I;
            end if;
         end loop;

      end Seek_Of_Resource;

      procedure Perform_Analysis_Of_Resources is

         Item2    : Deadlock_Item_Ptr;
         List_Ite : Deadlock_Iterator;
         Found    : Boolean := False;

      begin
         -- Initiazation of stamps
         --

         for I in  0 .. Number_Of_Tasks - 1 loop
            for J in  0 .. Number_Of_Resources - 1 loop
               Stamp_For_Allocate (I, J) := 0;
               Stamp_For_Wait (I, J)     := 0;
            end loop;
         end loop;

         for I in  0 .. Sched.Entries (Sched.Nb_Entries - 1).Data loop
            for J in  0 .. Sched.Nb_Entries - 1 loop

               -- construction of the tree
               --
               if (Sched.Entries (J).Data = I) then
                  if (Sched.Entries (J).Item.Event_Type =
                      Allocate_Resource)
                  then
                     Item_Task     := Sched.Entries (J).Item.Allocate_Task;
                     Item_Resource :=
                       Sched.Entries (J).Item.Allocate_Resource;
                     Index_Of_Tasks (Item_Task.Name, Index_Task);
                     Index_Of_Resources (Item_Resource.Name, Index_Resource);
                     Stamp_For_Allocate (Index_Task, Index_Resource) := 1;
                     Stamp_For_Wait (Index_Task, Index_Resource)     := 0;
                  end if;

                  if (Sched.Entries (J).Item.Event_Type =
                      Wait_For_Resource)
                  then
                     Item_Task     := Sched.Entries (J).Item.Wait_Task;
                     Item_Resource := Sched.Entries (J).Item.Wait_Resource;
                     Index_Of_Tasks (Item_Task.Name, Index_Task);
                     Index_Of_Resources (Item_Resource.Name, Index_Resource);
                     Stamp_For_Wait (Index_Task, Index_Resource) := 1;
                  end if;

                  if (Sched.Entries (J).Item.Event_Type =
                      Release_Resource)
                  then
                     Item_Task     := Sched.Entries (J).Item.Release_Task;
                     Item_Resource := Sched.Entries (J).Item.Release_Resource;
                     Index_Of_Tasks (Item_Task.Name, Index_Task);
                     Index_Of_Resources (Item_Resource.Name, Index_Resource);
                     Stamp_For_Allocate (Index_Task, Index_Resource) := 0;
                  end if;
               end if;

            end loop;

            --analyse of stamps
            --

            for K in  0 .. Number_Of_Resources - 1 loop
               for L in  0 .. Framework_Config.Max_Resources - 1 loop
                  Table_Index_Task (L)     := 0;
                  Table_Index_Resource (L) := 0;
               end loop;
               Current_Time   := 0;
               Index_Resource := K;
               loop
                  Index_Task := 30;
                  Ok         := False;
                  Seek_Of_Task (Index_Resource, Index_Task);
                  if (Index_Task /= 30) then
                     Table_Index_Task (Current_Time)     := Index_Task;
                     Table_Index_Resource (Current_Time) := Index_Resource;
                     Index_Resource                      := 30;
                     Seek_Of_Resource (Index_Task, Index_Resource);
                     if (Index_Resource /= 30) then
                        for P in  0 .. Current_Time loop
                           if (Table_Index_Resource (P) =
                               Index_Resource)
                           then
                              Ok := True;
                           end if;
                        end loop;
                        if Ok then
                           for P in  0 .. Current_Time loop
                              Item               := new Deadlock_Item;
                              Item.Time          := I;
                              Item.Task_Name     :=
                                Table_Name_Tasks (Table_Index_Task (P));
                              Item.Resource_Name :=
                                Table_Name_Resources (Table_Index_Resource (P))
;
                              Found              := False;

                              -- to check if item is already present in the
                              --list
                              --

                              if not Is_Empty (Result) then
                                 Reset_Head_Iterator (Result, List_Ite);
                                 loop
                                    Current_Element (Result, Item2, List_Ite);

                                    if ((Item2.Task_Name = Item.Task_Name) and
                                        (Item2.Resource_Name =
                                         Item.Resource_Name))
                                    then
                                       Found := True;
                                       exit;
                                    end if;

                                    if Is_Tail_Element
                                         (Result,
                                          List_Ite)
                                    then
                                       exit;
                                    end if;
                                    Next_Element (Result, List_Ite);
                                 end loop;
                              end if;

                              -- addition of item in the list
                              --

                              if not Found then
                                 Add (Result, Item);
                              end if;

                           end loop;

                           exit;
                        end if;
                     else
                        exit;
                     end if;
                  else
                     exit;
                  end if;
                  Current_Time := Current_Time + 1;
               end loop;
            end loop;

         end loop;

      end Perform_Analysis_Of_Resources;

   begin
      -- Initialization of tables of tasks names
      --
      for I in  0 .. Number_Of_Tasks - 1 loop

         Get_Element_Number (My_Tasks, Item_Task, I);
         Table_Name_Tasks (I) := Item_Task.Name;

      end loop;

      -- Initialization of tables of resources of names
      --
      for I in  0 .. Number_Of_Resources - 1 loop

         Get_Element_Number (My_Resources, Item_Resource, I);
         Table_Name_Resources (I) := Item_Resource.Name;
      end loop;

      Perform_Analysis_Of_Resources;
      return Result;

   end Deadlock_From_Simulation;

   function Priority_Inversion_From_Simulation
     (My_Tasks       : in Tasks_Set;
      My_Resources   : in Resources_Set;
      Processor_Name : in Unbounded_String;
      Sched          : in Scheduling_Sequence_Ptr)
      return           Priority_Inversion_List
   is
      Result : Priority_Inversion_List;

      -- Variables required to scan the resource set
      --
      Item        : Priority_Inversion_Item_Ptr;
      A_Resource  : Generic_Resource_Ptr;
      My_Iterator : Resources_Iterator;

      -- Store the event of the resource allocation
      --
      Allocation : Time_Unit_Event_Ptr;

      -- Store all Wait_For_Resource events
      --
      Wait_Events : array (Time_Unit_Range) of Time_Unit_Event_Ptr;

      -- Simulation time of the stored event
      --
      Wait_Event_Time : array (Time_Unit_Range) of Natural;

      -- Index on  Wait_Events and Indexes  tables
      --
      Index : Time_Unit_Range := 0;

      -- Index on the scheduling table
      --
      Current_Time : Time_Unit_Range := 0;

      procedure Perform_Analysis_Of_One_Resource is

      begin

         Current_Time := 0;

         -- First, find an Allocation Event
         --
         loop
            if (Current_Time >= Sched.Nb_Entries - 1) then
               return;
            end if;

            if (Sched.Entries (Current_Time).Item.Event_Type =
                Allocate_Resource)
            then
               if (Sched.Entries (Current_Time).Item.Allocate_Resource.Name =
                   A_Resource.Name)
               then
                  Allocation := Sched.Entries (Current_Time).Item;
                  exit;
               end if;
            end if;

            Current_Time := Current_Time + 1;
         end loop;

         loop

            Current_Time := Current_Time + 1;

            -- Search all Wait_For_Resource events upto the next allocation
            --
            if (Current_Time >= Sched.Nb_Entries - 1) then
               return;
            end if;

            if (Sched.Entries (Current_Time).Item.Event_Type =
                Allocate_Resource)
            then
               if (Sched.Entries (Current_Time).Item.Allocate_Resource.Name =
                   A_Resource.Name)
               then

                  -- The next Allocate event is found, stop collecting data
                  -- and do analysis
                  --

                  -- Check that between the two "Allocate" events when have
                  --middle priority
                  -- tasks and that the priority of the first allocating task
                  --is less than
                  -- 2nd one.
                  --
                  if (Allocation.Allocate_Task.Priority <
                      Sched.Entries (Current_Time).Item.Allocate_Task.Priority)
                  then

                     declare

                     -- Start time of the priority inversion
                     --
                        Start_Time : Natural :=
                          Sched.Entries (Current_Time).Data;

                     begin
                        -- look for the start time of the priority inversion
                        --
                        for Z in reverse  0 .. Index - 1 loop
                           if Wait_Events (Z).Wait_Task.Name =
                              Sched.Entries (Current_Time).Item.Allocate_Task.
                                Name
                           then
                              Start_Time := Wait_Event_Time (Z);
                           end if;
                        end loop;

                        -- Store the detected priority inversion
                        --
                        if Start_Time <
                           Sched.Entries (Current_Time).Data
                        then
                           Item := new Priority_Inversion_Item;

                           Item.Start_Time    := Start_Time;
                           Item.End_Time      :=
                             Sched.Entries (Current_Time).Data;
                           Item.Task_Name     :=
                             Sched.Entries (Current_Time).Item.Allocate_Task.
                             Name;
                           Item.Resource_Name :=
                             Sched.Entries (Current_Time).Item.
                             Allocate_Resource.Name;

                           Add (Result, Item);
                        end if;
                     end;

                  end if;

                  -- Ready for the next priority inversion analysis
                  --
                  Index      := 0;
                  Allocation := Sched.Entries (Current_Time).Item;

               end if;
            end if;

            if (Sched.Entries (Current_Time).Item.Event_Type =
                Wait_For_Resource)
            then
               if (Sched.Entries (Current_Time).Item.Wait_Resource.Name =
                   A_Resource.Name)
               then

                  Wait_Events (Index)     :=
                    Sched.Entries (Current_Time).Item;
                  Wait_Event_Time (Index) :=
                    Sched.Entries (Current_Time).Data;
                  Index                   := Index + 1;
               end if;
            end if;
         end loop;

      end Perform_Analysis_Of_One_Resource;

   begin

      Reset_Iterator (My_Resources, My_Iterator);

      Initialize (Result);

      loop
         Current_Element (My_Resources, A_Resource, My_Iterator);

         if (Processor_Name = A_Resource.cpu_name) then
            Perform_Analysis_Of_One_Resource;
         end if;

         exit when Is_Last_Element (My_Resources, My_Iterator);

         Next_Element (My_Resources, My_Iterator);
      end loop;

      return Result;

   end Priority_Inversion_From_Simulation;

   function Number_Of_Preemption_From_Simulation
     (Sched    : in Scheduling_Sequence_Ptr;
      My_Tasks : in Tasks_Set)
      return     Natural
   is
      Nb         : Natural := 0;
      A_Task     : Generic_Task_Ptr;
      Usage_Time : Natural;

   begin

      for I in  0 .. Sched.Nb_Entries - 1 loop
         if (Sched.Entries (I).Item.Event_Type = Running_Task) then

            for K in reverse  0 .. I - 1 loop
               if (Sched.Entries (K).Item.Event_Type = Running_Task) then
                  if ((Sched.Entries (K).Item.Running_Task.Name /=
                       Sched.Entries (I).Item.Running_Task.Name) and
                      (Sched.Entries (K).Data + 1 = Sched.Entries (I).Data))
                  then

                     -- We know that the previous unit of time, another
                     -- task was running. Check now that is was a preeemption
                     --

                     -- look for the previously elected task and its state
                     --
                     A_Task :=
                        Search_Task
                          (My_Tasks,
                           Sched.Entries (I).Item.Running_Task.Name);

                     -- Compute the running time of the task until its wake up
                     --time

                     --
                     Usage_Time := 0;
                     for Z in reverse  0 .. K - 1 loop
                        if (Sched.Entries (Z).Item.Event_Type =
                            Running_Task)
                        then
                           if A_Task.Name =
                              Sched.Entries (Z).Item.Running_Task.Name
                           then
                              Usage_Time := Usage_Time + 1;
                           end if;
                        end if;
                        if (Sched.Entries (Z).Item.Event_Type =
                            Task_Activation)
                        then
                           if A_Task.Name =
                              Sched.Entries (Z).Item.Activation_Task.Name
                           then
                              if (Usage_Time /= A_Task.Capacity) and
                                 (Usage_Time /= 0)
                              then
                                 Nb := Nb + 1;
                              end if;
                              exit;
                           end if;
                        end if;
                     end loop;
                     exit;
                  end if;
               end if;
            end loop;

         end if;
      end loop;

      return Nb;
   end Number_Of_Preemption_From_Simulation;

   function Number_Of_Switch_Context_From_Simulation
     (Sched : in Scheduling_Sequence_Ptr)
      return  Natural
   is
      Nb : Natural := 0;

   begin

      for I in  0 .. Sched.Nb_Entries - 1 loop
         if (Sched.Entries (I).Item.Event_Type = Running_Task) then

            for K in reverse  0 .. I - 1 loop
               if (Sched.Entries (K).Item.Event_Type = Running_Task) then

                  if (Sched.Entries (K).Item.Running_Task.Name /=
                      Sched.Entries (I).Item.Running_Task.Name)
                  then
                     Nb := Nb + 1;
                  end if;

                  exit;
               end if;
            end loop;
         end if;
      end loop;

      return Nb;
   end Number_Of_Switch_Context_From_Simulation;

   procedure Consumer_Response_Time
     (Sched              : in Scheduling_Sequence_Ptr;
      Consumer_Resp_Time : in out Resp_Time_Consumer_Table_Ptr;
      My_Consumer_Task   : in Generic_Task_Ptr)
   is

      My_Consumer_Task_Occurence_Table : Task_Occurence_Table_Ptr;
      Index                            : Resp_Time_Consumer_Range := 0;

   begin

      -- initialize
      --
      My_Consumer_Task_Occurence_Table := new Task_Occurence_Table;
      All_Response_Time_By_Simulation
        (My_Consumer_Task,
         Sched,
         My_Consumer_Task_Occurence_Table);

      -- affect response time value
      --
      for I_Occurence in  Task_Occurence_Range loop
         if (My_Consumer_Task_Occurence_Table.Entries (I_Occurence).
              Response_Time /=
             0)
         then

            Consumer_Resp_Time.Entries (Index).Data :=
              Double (
              My_Consumer_Task_Occurence_Table.Entries (I_Occurence).
              Response_Time);
            Index                                   := Index + 1;

         end if;
      end loop;
      Consumer_Resp_Time.Nb_Entries := Index;

      Free (My_Consumer_Task_Occurence_Table);
   end Consumer_Response_Time;

end Scheduling_Sequences;
