------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Task_Set; use Task_Set;
use Task_Set.Generic_Task_Set;
with Resource_Set; use Resource_Set;
use Resource_Set.Generic_Resource_Set;
WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;
WITH Framework_Config;
USE Framework_Config;
WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;
use Time_Units.Time_Unit_List;
WITH Double_Util;
USE Double_Util;
WITH Tasks;
USE Tasks;
WITH Tasks.Extended;
USE Tasks.Extended;
WITH Buffers;
USE Buffers;
use Buffers.Buffer_Roles;
WITH Event_Analyzer_Set;
USE Event_Analyzer_Set;
USE Event_Analyzer_Set.Generic_Event_Analyzer_Set;
with Queueing_System; use Queueing_System;
use Queueing_System.A_Resp_Time_Consumer;
WITH Tables;
WITH Indexed_Tables;
WITH Lists;
WITH Unchecked_Deallocation;


PACKAGE Scheduling_Sequences IS



   --------------------------------------------------------
   --------------------------------------------------------
   --
   -- Definition of scheduling sequence type and its services
   --
   --------------------------------------------------------
   --------------------------------------------------------


   SUBTYPE Scheduling_Sequence IS A_Time_Unit.Indexed_Table;
   TYPE Scheduling_Sequence_Ptr IS ACCESS Scheduling_Sequence;


   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Scheduling_Sequence, Scheduling_Sequence_Ptr);



   --------------------------------------------------------
   --------------------------------------------------------
   -- Services to do analysis of scheduling sequences
   --------------------------------------------------------
   --------------------------------------------------------



   --------------------------------------------------------
   -- Processor analysis
   --------------------------------------------------------


   -- Compute number of task preemption
   --
   FUNCTION Number_Of_Preemption_From_Simulation (
         Sched    : IN     Scheduling_Sequence_Ptr;
         My_Tasks : IN     Tasks_Set)
     RETURN Natural;


   -- Compute number of switch context
   --
   FUNCTION Number_Of_Switch_Context_From_Simulation (
         Sched : IN     Scheduling_Sequence_Ptr)
     RETURN Natural;




   --------------------------------------------------------
   -- Buffer analysis (buffer utilization factor)
   --------------------------------------------------------


   TYPE Buffer_Size_Item IS
      RECORD
         Time : Natural;
         Size : Natural;
      END RECORD;

   PROCEDURE Initialize (
         B :    OUT Buffer_Size_Item);
   PROCEDURE Put (
         B : IN     Buffer_Size_Item);




   PACKAGE Buffer_Size_Package IS NEW Tables(Buffer_Size_Item,
      Framework_Config.Max_Scheduling_Period,  Put);
   USE Buffer_Size_Package;

   SUBTYPE Buffer_Size_Table IS Buffer_Size_Package.Table;
   SUBTYPE Buffer_Size_Range IS Buffer_Size_Package.Table_Range;
   TYPE Buffer_Size_Table_Ptr IS ACCESS Buffer_Size_Table;


   PACKAGE Waiting_Time_Package IS NEW Tables(Double,
      Framework_Config.Max_Scheduling_Period,  Put);
   USE Waiting_Time_Package;

   SUBTYPE Waiting_Time_Table IS Waiting_Time_Package.Table;
   SUBTYPE Waiting_Time_Range IS Waiting_Time_Package.Table_Range;
   TYPE Waiting_Time_Table_Ptr IS ACCESS Waiting_Time_Table;


   -- Compute from a scheduling sequence buffer size for each
   -- time unit
   --
   PROCEDURE Compute_Buffer_Size_From_Simulation (
         Sched   : IN     Scheduling_Sequence_Ptr;
         My_Buff : IN     Buffer_Ptr;
         Buff    : IN OUT Buffer_Size_Table_Ptr);


   -- Compute, from a buffer_Size_Table, maximum and average
   -- performance criterion
   --
   FUNCTION Compute_Average_Buffer_Size_From_Simulation (
         Buff : IN     Buffer_Size_Table_Ptr)
     RETURN Double;

   FUNCTION Compute_Maximum_Buffer_Size_From_Simulation (
         Buff : IN     Buffer_Size_Table_Ptr)
     RETURN Natural;

   FUNCTION Compute_Maximum_Waiting_Time_From_Simulation (
         Buff                    : IN     Buffer_Size_Table_Ptr;
         Average_Consumer_Period : IN     Double)
     RETURN Natural;

   FUNCTION Compute_Average_Waiting_Time_From_Simulation (
         Buff                    : IN     Buffer_Size_Table_Ptr;
         Average_Consumer_Period : IN     Double)
     RETURN Double;





   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Buffer_Size_Table, Buffer_Size_Table_Ptr);

   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Waiting_Time_Table, Waiting_Time_Table_Ptr);

   No_Performance_Criteria : EXCEPTION;

   -- copy consumer response time from simulation of a scheduling
   -- in cheddar to a qs (example : MP1 criteria computation)
   --
   PROCEDURE Consumer_Response_Time (
         Sched              : IN     Scheduling_Sequence_Ptr;
         Consumer_Resp_Time : IN OUT Resp_Time_Consumer_Table_Ptr;
         My_Consumer_Task   : IN     Generic_Task_Ptr);

   PACKAGE Generic_Index_Lst IS NEW Lists (
      Put     => Put,
      Element => Double_Ptr,
      Free    => Double_Util.Free);

   USE Generic_Index_Lst;



   --------------------------------------------------------
   -- Task analysis (eg. response time from simulation)
   --------------------------------------------------------


   -- Compute bound on response time from simulation
   --
   PROCEDURE Response_Time_By_Simulation (
         My_Task :        Generic_Task_Ptr;
         Sched   : IN     Scheduling_Sequence_Ptr;
         Msg     : IN OUT Unbounded_String;
         Max     : IN OUT Natural;
         Min     : IN OUT Natural;
         Average : IN OUT Double);


   -- Compute probability distribution of response time
   --
   TYPE Density_Item IS
      RECORD
         Response_Time : Natural;
         Probability   : Double;
      END RECORD;


   PROCEDURE Initialize (
         D :    OUT Density_Item);
   PROCEDURE Put (
         D : IN     Density_Item);


   PACKAGE Density_Table_Package IS NEW Tables(
      Density_Item, Framework_Config.Max_Density_Size, Put);
   USE Density_Table_Package;
   SUBTYPE Density_Table IS Density_Table_Package.Table;
   SUBTYPE Density_Range IS Density_Table_Package.Table_Range;


   PACKAGE Densities_Table_Package IS NEW Indexed_Tables(Density_Table,
      Generic_Task_Ptr, Framework_Config.Max_Tasks, 0, Put, Initialize,
      Put_Name);
   USE  Densities_Table_Package;
   SUBTYPE Densities_Table IS Densities_Table_Package.Indexed_Table;
   SUBTYPE Densities_Range IS
      Densities_Table_Package.Indexed_Table_Range;


   PROCEDURE Compute_Response_Time_Distribution (
         My_Tasks       : IN     Tasks_Set;
         Sched          : IN     Scheduling_Sequence_Ptr;
         Processor_Name :        Unbounded_String;
         Result         : IN OUT Densities_Table);



   -- Compute the response time for each task activation
   --
   PACKAGE Task_Occurence_Table_Package IS NEW Tables(
      Task_Occurence, Framework_Config.Max_Scheduling_Period, Put);
   USE Task_Occurence_Table_Package;

   SUBTYPE Task_Occurence_Table IS Task_Occurence_Table_Package.Table;
   SUBTYPE Task_Occurence_Range IS Task_Occurence_Table_Package.
      Table_Range;
   TYPE Task_Occurence_Table_Ptr IS ACCESS Task_Occurence_Table;



   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Task_Occurence_Table, Task_Occurence_Table_Ptr);


   PROCEDURE All_Response_Time_By_Simulation (
         My_Task : IN     Generic_Task_Ptr;
         Sched   : IN     Scheduling_Sequence_Ptr;
         Result  : IN OUT Task_Occurence_Table_Ptr);





   -----------------------------------------------------------------------
   -- Shared resource  analysis (eg. blocking time from simulation)
   -----------------------------------------------------------------------







   -- Compute blocking time from scheduling simulaion
   --
   PACKAGE Double_Tasks_Parameters IS NEW Indexed_Tables(Double,
      Generic_Task_Ptr,
      Framework_Config.Max_Tasks, 0, Double_Util.Put,
      Double_Util.Initialize,
      Put_Name);

   SUBTYPE Blocking_Time_Table IS Double_Tasks_Parameters.Indexed_Table;
   SUBTYPE Blocking_Time_Range IS Double_Tasks_Parameters.
      Indexed_Table_Range;


   PROCEDURE Blocking_Time_From_Simulation (
         My_Task      : IN     Generic_Task_Ptr;
         My_Resources : IN     Resources_Set;
         Sched        : IN     Scheduling_Sequence_Ptr;
         Max          : IN OUT Natural;
         Min          : IN OUT Natural;
         Average      : IN OUT Double);



   -- Check if deadlocks on shared resources exist
   --
   TYPE Deadlock_Item IS
      RECORD
         Time          : Natural;
         Task_Name     : Unbounded_String;
         Resource_Name : Unbounded_String;
      END RECORD;

   TYPE Deadlock_Item_Ptr IS ACCESS ALL Deadlock_Item;
   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Deadlock_Item, Deadlock_Item_Ptr );

   PROCEDURE Initialize (
         D : IN OUT Deadlock_Item);
   PROCEDURE Put (
         D : IN     Deadlock_Item);
   PROCEDURE Put (
         D : IN     Deadlock_Item_Ptr);


   PACKAGE Deadlock_List_Package IS NEW Lists(
      Deadlock_Item_Ptr, Put, Free);
   USE Deadlock_List_Package;
   SUBTYPE Deadlock_List IS Deadlock_List_Package.List;
   SUBTYPE Deadlock_Iterator IS
      Deadlock_List_Package.Iterator;



   FUNCTION Deadlock_From_Simulation (
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Sched          : IN     Scheduling_Sequence_Ptr)
     RETURN Deadlock_List;




   -- Check if priority inversions exist in a scheduling simulation
   --

   TYPE Priority_Inversion_Item IS
      RECORD
         Start_Time    : Natural;
         End_Time      : Natural;
         Task_Name     : Unbounded_String;
         Resource_Name : Unbounded_String;
      END RECORD;

   TYPE Priority_Inversion_Item_Ptr IS ACCESS ALL Priority_Inversion_Item;
   PROCEDURE Free IS
   NEW Unchecked_Deallocation (Priority_Inversion_Item,
      Priority_Inversion_Item_Ptr);


   PROCEDURE Initialize (
         D : IN OUT Priority_Inversion_Item);
   PROCEDURE Put (
         D : IN     Priority_Inversion_Item);
   PROCEDURE Put (
         D : IN     Priority_Inversion_Item_Ptr);


   PACKAGE Priority_Inversion_List_Package IS NEW Lists(
      Priority_Inversion_Item_Ptr, Put, Free);
   USE Priority_Inversion_List_Package;
   SUBTYPE Priority_Inversion_List IS
      Priority_Inversion_List_Package.List;
   SUBTYPE Priority_Inversion_Iterator IS
      Priority_Inversion_List_Package.Iterator;


   FUNCTION Priority_Inversion_From_Simulation (
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Sched          : IN     Scheduling_Sequence_Ptr)
     RETURN Priority_Inversion_List;




END Scheduling_Sequences;
