------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_IO;            use Text_IO;
with Scheduler;          use Scheduler;
with AADL_Parsers;       use AADL_Parsers;
with Xml_Parsers;        use Xml_Parsers;
with Input_Sources.File; use Input_Sources.File;
with Sax.Readers;        use Sax.Readers;
with Offsets;            use Offsets;
with Integer_Util;       use Integer_Util;

package body Systems is

   procedure Delete_Task (A_Sys : in out System; A_Task : Generic_Task_Ptr) is
   begin
      Delete_Task (A_Sys.Resources, A_Task.Name);
      Delete_Task (A_Sys.Buffers, A_Task.Name);
      Delete_All_Task_Dependencies (A_Sys.Dependencies, A_Task);
      Delete (A_Sys.Tasks, A_Task);
   end Delete_Task;

   procedure Delete_Buffer (A_Sys : in out System; A_Buffer : Buffer_Ptr) is
   begin
      Delete_All_Task_Dependencies (A_Sys.Dependencies, A_Buffer);
      Delete (A_Sys.Buffers, A_Buffer);
   end Delete_Buffer;

   procedure Delete_Message
     (A_Sys     : in out System;
      A_Message : Message_Ptr)
   is
   begin
      Delete_All_Task_Dependencies (A_Sys.Dependencies, A_Message);
      Delete (A_Sys.Messages, A_Message);
   end Delete_Message;

   procedure Delete_Processor
     (A_Sys       : in out System;
      A_Processor : Processor_Ptr)
   is
   begin

      Delete_Processor_All_Task_Dependencies
        (A_Sys.Dependencies,
         A_Processor.Name);
      Delete_Processor (A_Sys.Resources, A_Processor.Name);
      Delete_Processor (A_Sys.Buffers, A_Processor.Name);
      Delete_Processor (A_Sys.Tasks, A_Processor.Name);
      Delete_Processor (A_Sys.Address_Spaces, A_Processor.Name);
      Delete (A_Sys.Processors, A_Processor);
   end Delete_Processor;

   procedure Delete_Address_Space
     (A_Sys            : in out System;
      An_Address_Space : Address_Space_Ptr)
   is
   begin
      Delete_Address_Space_All_Task_Dependencies
        (A_Sys.Dependencies,
         An_Address_Space.Name);
      Delete_Address_Space (A_Sys.Resources, An_Address_Space.Name);
      Delete_Address_Space (A_Sys.Buffers, An_Address_Space.Name);
      Delete_Address_Space (A_Sys.Tasks, An_Address_Space.Name);
      Delete (A_Sys.Address_Spaces, An_Address_Space);
   end Delete_Address_Space;

   procedure Duplicate (Src : in System; Dest : in out System) is
   begin

      Duplicate (Src.Tasks, Dest.Tasks);
      Duplicate (Src.Resources, Dest.Resources);
      Duplicate (Src.Messages, Dest.Messages);
      Duplicate (Src.Dependencies, Dest.Dependencies);
      Duplicate (Src.Processors, Dest.Processors);
      Duplicate (Src.Buffers, Dest.Buffers);
      Duplicate (Src.Networks, Dest.Networks);
      Duplicate (Src.Event_Analyzers, Dest.Event_Analyzers);
      Duplicate (Src.Address_Spaces, Dest.Address_Spaces);

   end Duplicate;

   procedure Put_Xml (A_System : in System) is
   begin

      Put (To_String (Export_Xml_Cheddar_Dtd (A_System)));
      Put (To_String (Export_Xml (A_System)));

   end Put_Xml;

   procedure Put_Aadl (A_System : in System) is
   begin
      New_Line;

      -- Export AADL standard properties and Cheddar's properties
      --
      --   Put_Line(To_String(Export_Aadl_Standard_Properties(A_System)));
      --   Put_Line(To_String(Export_Aadl_Project_Properties(A_System)));
      --   Put_Line(To_String(Export_Aadl_Cheddar_Properties(A_System)));

      -- Export the system itself
      --
      Put_Line (To_String (Export_Aadl_Implementations (A_System)));

      New_Line;

   end Put_Aadl;

   procedure Put (A_System : in System) is
   begin

      Put_Line ("***** LIST OF PROCESSORS *****");
      Put (A_System.Processors);
      New_Line;
      New_Line;
      New_Line;
      Put_Line ("***** LIST OF TASKS *****");
      Put (A_System.Tasks);
      New_Line;
      New_Line;
      New_Line;
      Put_Line ("***** LIST OF RESSOURCES *****");
      Put (A_System.Resources);
      New_Line;
      New_Line;
      New_Line;
      Put_Line ("***** LIST OF BUFFERS *****");
      Put (A_System.Buffers);
      New_Line;
      New_Line;
      New_Line;
      Put_Line ("***** LIST OF MESSAGES *****");
      Put (A_System.Messages);
      New_Line;
      New_Line;
      New_Line;
      Put_Line ("***** LIST OF DEPENCENCIES *****");
      Put (A_System.Dependencies);
      New_Line;
      New_Line;
      New_Line;
      Put_Line ("***** LIST OF NETWORKS *****");
      Put (A_System.Networks);
      New_Line;
      New_Line;
      New_Line;
      Put_Line ("***** LIST OF EVENTS ANALYZERS *****");
      Put (A_System.Event_Analyzers);
      New_Line;
      New_Line;
      New_Line;
      Put_Line ("***** LIST OF ADDRESS SPACES *****");
      Put (A_System.Address_Spaces);
      New_Line;
      New_Line;
      New_Line;

   end Put;

   procedure Put (A_System : in System_Ptr) is
   begin
      Put (A_System.all);
   end Put;

   ------------------------------------------------

   procedure Initialize (A_System : in out System) is
   begin

      Reset (A_System.Networks);
      Reset (A_System.Tasks);
      Reset (A_System.Messages);
      Reset (A_System.Resources);
      if A_System.Dependencies /= null then
         Free (A_System.Dependencies);
      end if;
      A_System.Dependencies := new Tasks_Dependencies;
      Reset (A_System.Dependencies);
      Reset (A_System.Buffers);
      Reset (A_System.Processors);
      Reset (A_System.Event_Analyzers);
      Reset (A_System.Address_Spaces);
   end Initialize;

   procedure Initialize (A_System : in System_Ptr) is
   begin
      Initialize (A_System.all);
   end Initialize;

   ------------------------------------------------

   procedure Read_From_Xml_File
     (A_System  : in out System;
      Dir_List  : in Unbounded_String_List;
      File_Name : in String)
   is
   begin
      Read_From_Xml_File
        (A_System,
         Dir_List,
         To_Unbounded_String (File_Name));
   end Read_From_Xml_File;

   procedure Write_To_Xml_File
     (A_System  : in System;
      File_Name : in String)
   is
   begin
      Write_To_Xml_File (A_System, To_Unbounded_String (File_Name));
   end Write_To_Xml_File;

   procedure Read_From_Xml_File
     (A_System  : in out System;
      Dir_List  : in Unbounded_String_List;
      File_Name : in Unbounded_String)
   is

      Read       : File_Input;
      My_Reader  : Xml_Project_Parser;
      Name_Start : Natural;
      S          : constant String := To_String (File_Name);

   begin

      --  Base file name should be used as the public Id
      --
      Name_Start := S'Last;
      while Name_Start >= S'First and then S (Name_Start) /= '/' loop
         Name_Start := Name_Start - 1;
      end loop;
      Set_Public_Id (Read, S (Name_Start + 1 .. S'Last));
      Set_System_Id (Read, S);
      Open (S, Read);

      --  xmlns:* attributes will be reported in Start_Element
      --
      Set_Feature (My_Reader, Namespace_Prefixes_Feature, False);
      Set_Feature (My_Reader, Validation_Feature, True);

      Parse (My_Reader, Read);
      A_System := Get_Parsed_System (My_Reader);
      Close (Read);

   exception
      when XML_Fatal_Error =>
         Close (Read);
         raise Xml_Read_Error;

   end Read_From_Xml_File;

   procedure Write_To_Xml_File
     (A_System  : in System;
      File_Name : in Unbounded_String)
   is

      Into : File_Type;

   begin

      -- Open file and Write DTD/XML Header
      --
      Create (Into, Mode => Out_File, Name => To_String (File_Name));

      Put (Into, To_String (Export_Xml_Cheddar_Dtd (A_System)));
      Put (Into, To_String (Export_Xml (A_System)));

      Close (Into);

   end Write_To_Xml_File;

   procedure Read_From_Aadl_File
     (A_System          : in out System;
      Dir_List          : in Unbounded_String_List;
      Project_File_List : in Unbounded_String_List)
   is

      My_Reader : AADL_Project_Parser;

   begin

      Parse (My_Reader, Dir_List, Project_File_List);
      A_System := Get_Parsed_System (My_Reader);

   end Read_From_Aadl_File;

   procedure Write_To_Aadl_File
     (A_System  : in System;
      File_Name : in Unbounded_String)
   is

      Into : File_Type;

   begin

      -- Open file and Write aadl specification
      --
      Create (Into, Mode => Out_File, Name => To_String (File_Name));

      New_Line (Into);

      -- Export AADL standard customized properties
      --
      -- Put_Line(Into, To_String(Export_Aadl_Standard_Properties(A_System)));
      -- Put_Line(Into, To_String(Export_Aadl_Project_Properties(A_System)));
      -- Put_Line(Into, To_String(Export_Aadl_Cheddar_Properties(A_System)));

      -- Export the system itself
      --
      Put_Line (Into, To_String (Export_Aadl_Implementations (A_System)));

      New_Line (Into);

      Close (Into);

   end Write_To_Aadl_File;

   procedure Write_Aadl_Standard_Properties_Set_To_File
     (A_System : in System)
   is

      Into : File_Type;

   begin

      -- Open file and Write aadl specification
      --
      Create (Into, Mode => Out_File, Name => "AADL_Properties.aadl");

      New_Line (Into);
      Put_Line
        (Into,
         To_String (Export_Aadl_Standard_Properties (A_System)));
      New_Line (Into);

      Close (Into);
   end Write_Aadl_Standard_Properties_Set_To_File;

   procedure Write_Cheddar_Property_Sets_To_File (A_System : in System) is

      Into : File_Type;

   begin

      -- Open file and Write aadl specification
      --
      Create (Into, Mode => Out_File, Name => "AADL_Project.aadl");
      New_Line (Into);
      Put_Line (Into, To_String (Export_Aadl_Project_Properties (A_System)));
      New_Line (Into);
      Close (Into);

      Create (Into, Mode => Out_File, Name => "Cheddar_Properties.aadl");
      New_Line (Into);
      Put_Line (Into, To_String (Export_Aadl_Cheddar_Properties (A_System)));
      New_Line (Into);
      Close (Into);

      Create
        (Into,
         Mode => Out_File,
         Name => "User_Defined_Cheddar_Properties.aadl");
      New_Line (Into);
      Put_Line
        (Into,
         To_String (Export_Aadl_User_Defined_Cheddar_Properties (A_System)));
      New_Line (Into);
      Close (Into);

   end Write_Cheddar_Property_Sets_To_File;

   ------------------------------------------------

   function Export_Aadl_User_Defined_Cheddar_Properties
     (A_System : in System)
      return     Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
      Tmp    : Unbounded_String := Empty_String;

   begin

      -- Export user's defined properties
      --
      Tmp := Export_Aadl_User_Defined_Properties (A_System.Tasks);
      Tmp := Tmp & Export_Aadl_User_Defined_Properties (A_System.Messages);

      Result := Result &
                To_Unbounded_String
                   ("property set User_Defined_Cheddar_Properties is") &
                Unbounded_Lf &
                Unbounded_Lf;

      if Tmp = Empty_String then
         Result := Result &
                   "Dummy_User_Defined_Cheddar_Properties : aadlboolean" &
                   Unbounded_Lf &
                   "applies to (thread, thread group);";
      else
         Result := Result & Tmp;
      end if;

      Result := Result &
                Unbounded_Lf &
                To_Unbounded_String ("end User_Defined_Cheddar_Properties;") &
                Unbounded_Lf &
                Unbounded_Lf;

      return Result;

   end Export_Aadl_User_Defined_Cheddar_Properties;

   function Export_Aadl_Standard_Properties
     (A_System : in System)
      return     Unbounded_String
   is
      Result : Unbounded_String := Empty_String;

   begin

      Result := Result & ASCII.lf & To_Unbounded_String(" --****************************************************** ");

      Result := Result & ASCII.lf & To_Unbounded_String(" --  AADL Standard AADL_V1.0 ");

      Result := Result & ASCII.lf & To_Unbounded_String(" --  Appendix A (normative) ");

      Result := Result & ASCII.lf & To_Unbounded_String(" --  Predeclared Property Sets ");

      Result := Result & ASCII.lf & To_Unbounded_String(" --  03Nov04 ");

      Result := Result & ASCII.lf & To_Unbounded_String(" --  Update to reflect current standard on 28Mar06 ");

      Result := Result & ASCII.lf & To_Unbounded_String(" --****************************************************** ");
 
      Result := Result & ASCII.lf & To_Unbounded_String(" property set AADL_Properties is ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Activate_Deadline: Time applies to (thread); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Activate_Execution_Time: Time_Range applies to (thread); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Activate_Entrypoint: aadlstring applies to (thread); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Active_Thread_Handling_Protocol: inherit Supported_Active_Thread_Handling_Protocols => value(Default_Active_Thread_Handling_Protocol) applies to (thread, thread group, process, system); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Active_Thread_Queue_Handling_Protocol: inherit enumeration (flush, hold) => flush applies to (thread, thread group, process, system); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Actual_Connection_Binding: inherit list of reference (bus, processor, device) applies to (port connections, thread, thread group, process, system); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Actual_Latency: Time applies to (flow); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Actual_Memory_Binding: inherit reference (memory) applies to (thread, thread group, process, system, processor, data port, event data port, subprogram); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Actual_Processor_Binding: inherit reference (processor) applies to (thread, thread group, process, system); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Actual_Subprogram_Call: reference (server subprogram) applies to (subprogram); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Actual_Subprogram_Call_Binding: inherit list of reference (bus, processor, memory, device) applies to (subprogram); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Actual_Throughput: Data_Volume applies to (flow); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Aggregate_Data_Port: aadlboolean => false applies to (port group); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Access_Protocol: list of enumeration (Memory_Access, Device_Access) applies to (bus); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Connection_Binding: inherit list of reference (bus, processor, device) applies to (port connections, thread group, process, system); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Connection_Binding_Class: inherit list of classifier (processor, bus, device) applies to (port connections, thread, thread group, process, system); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Connection_Protocol: list of enumeration (Data_Connection, Event_Connection, Event_Data_Connection, Data_Access_Connection, Server_Subprogram_Call) applies to (bus, device); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Dispatch_Protocol: list of Supported_Dispatch_Protocols applies to (processor); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Memory_Binding: inherit list of reference (memory, system, processor) applies to (thread, thread group, process, system, device, data port, event data port, subprogram, processor); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Memory_Binding_Class: inherit list of classifier (memory, system, processor) applies to (thread, thread group, process, system, device, data port, event data port, subprogram, processor); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Message_Size: Size_Range applies to (bus); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Period: list of Time_Range applies to (processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Processor_Binding: inherit list of reference (processor, system) applies to (thread, thread group, process, system, device); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Processor_Binding_Class: inherit list of classifier (processor, system) applies to (thread, thread group, process, system, device); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Subprogram_Call: list of reference (server subprogram) applies to (subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Allowed_Subprogram_Call_Binding: inherit list of reference (bus, processor, device) applies to (subprogram, thread, thread group, process, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Assign_Time: Time applies to (processor, bus); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Assign_Byte_Time: Time applies to (processor, bus); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Assign_Fixed_Time: Time applies to (processor, bus); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Available_Memory_Binding: inherit list of reference (memory, system) applies to (system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Available_Processor_Binding: inherit list of reference (processor, system) applies to (system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Base_Address: access aadlinteger 0 .. value(Max_Base_Address) applies to (memory); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Client_Subprogram_Execution_Time: Time applies to (subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Clock_Jitter: Time applies to (processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Clock_Period: Time applies to (processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Clock_Period_Range: Time_Range applies to (processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Compute_Deadline: Time applies to (thread, device, subprogram, event port, event data port); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Compute_Entrypoint: aadlstring applies to (thread, subprogram, event port, event data port); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Compute_Execution_Time: Time_Range applies to (thread, device, subprogram, event port, event data port); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Concurrency_Control_Protocol: Supported_Concurrency_Control_Protocols => NoneSpecified applies to (data); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Connection_Protocol: Supported_Connection_Protocols applies to (connections); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Data_Volume: type aadlinteger 0 bitsps .. value(Max_Aadlinteger) units (bitsps, Bps    => bitsps * 8, Kbps   => Bps    * 1000, Mbps   => Kbps   * 1000, Gbps   => Mbps   * 1000 ); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Deactivate_Deadline: Time applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Deactivate_Execution_Time: Time_Range applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Deactivate_Entrypoint: aadlstring applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Deadline: inherit Time => value(Period) applies to (thread, thread group, process, system, device); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Dequeue_Protocol: enumeration (OneItem, AllItems) => OneItem applies to (event port, event data port); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Device_Dispatch_Protocol: Supported_Dispatch_Protocols => Aperiodic applies to (device); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Device_Register_Address: aadlinteger applies to (port, port group); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Dispatch_Protocol: Supported_Dispatch_Protocols applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Expected_Latency: Time applies to (flow); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Expected_Throughput: Data_Volume applies to (flow); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Finalize_Deadline: Time applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Finalize_Execution_Time: Time_Range applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Finalize_Entrypoint: aadlstring applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Hardware_Description_Source_Text: inherit list of aadlstring applies to (memory, bus, device, processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Hardware_Source_Language: Supported_Hardware_Source_Languages applies to (memory, bus, device, processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Initialize_Deadline: Time applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Initialize_Execution_Time: Time_Range applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Initialize_Entrypoint: aadlstring applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Latency: Time applies to (flow, connections); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Load_Deadline: Time applies to (process, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Load_Time: Time_Range applies to (process, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Memory_Protocol: enumeration (read_only, write_only, read_write) => read_write applies to (memory); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Not_Collocated: list of reference (data, thread, process, system, connections) applies to (data, thread, process, system, connections); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Overflow_Handling_Protocol: enumeration (DropOldest, DropNewest, Error) => DropOldest applies to (event port, event data port, subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Period: inherit Time applies to (thread, thread group, process, system, device); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Process_Swap_Execution_Time: Time_Range applies to (processor); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Propagation_Delay: Time_Range applies to (bus); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Provided_Access : access enumeration (read_only, write_only, read_write, by_method) => read_write applies to (data, bus); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Queue_Processing_Protocol: Supported_Queue_Processing_Protocols => FIFO applies to (event port, event data port, subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Queue_Size: aadlinteger 0 .. value(Max_Queue_Size) => 0 applies to (event port, event data port, subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Read_Time: list of Time_Range applies to (memory); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Recover_Deadline: Time applies to (thread, server subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Recover_Execution_Time: Time_Range applies to (thread, server subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Recover_Entrypoint: aadlstring applies to (thread); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Required_Access : access enumeration (read_only, write_only, read_write, by_method) => read_write applies to (data, bus); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Required_Connection : aadlboolean => true applies to (port); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Runtime_Protection : inherit aadlboolean => true applies to (process, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Scheduling_Protocol: list of Supported_Scheduling_Protocols applies to (processor); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Server_Subprogram_Call_Binding: inherit list of reference (thread, processor) applies to (subprogram, thread, thread group, process, system); "); 

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Size: type aadlinteger 0 B .. value (Max_Memory_Size) units Size_Units; ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Size_Range: type range of Size; ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Source_Code_Size: Size applies to (data, thread, thread group, process, system, subprogram, processor, device); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Source_Data_Size: Size applies to (data, subprogram, thread, thread group, process, system, processor, device); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Source_Heap_Size: Size applies to (thread, subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Source_Language: inherit Supported_Source_Languages applies to (subprogram, data, thread, thread group, process, bus, device, processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Source_Name: aadlstring applies to (data, port, subprogram, parameter); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Source_Stack_Size: Size applies to (thread, subprogram, processor, device); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Source_Text: inherit list of aadlstring applies to (data, port, subprogram, thread, thread group, process, system, memory, bus, device, processor, parameter, port group); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Startup_Deadline: inherit Time applies to (processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Subprogram_Execution_Time: Time_Range applies to (subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Supported_Source_Language: list of Supported_Source_Languages applies to (processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Synchronized_Component: inherit aadlboolean => true applies to (thread, thread group, process, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Thread_Limit: aadlinteger 0 .. value(Max_Thread_Limit) => value(Max_Thread_Limit) applies to (processor); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Thread_Swap_Execution_Time: Time_Range applies to (processor, system); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Throughput: Data_Volume applies to (flow, connections); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Time: type aadlinteger 0 ps .. value(Max_Time) units Time_Units; ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Time_Range: type range of Time; ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Transmission_Time: list of Time_Range applies to (bus); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Type_Source_Name: aadlstring applies to (data, port, subprogram); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Urgency: aadlinteger 0 .. value(Max_Urgency) applies to (port); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Word_Count: aadlinteger 0 .. value(Max_Word_Count) applies to (memory); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Word_Size: Size => 8 bits applies to (memory); ");
  
      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Word_Space: aadlinteger 1 .. value(Max_Word_Space) => 1 applies to (memory); ");

      Result := Result  & ASCII.lf & ASCII.HT & To_Unbounded_String(" Write_Time: list of Time_Range applies to (memory); ");

      Result := Result  & ASCII.lf & To_Unbounded_String(" end AADL_Properties;"); 

      Result := Result &
                Unbounded_Lf &
                Unbounded_Lf &
                Unbounded_Lf;

      return Result;

   end Export_Aadl_Standard_Properties;

   function Export_Aadl_Project_Properties
     (A_System : in System)
      return     Unbounded_String
   is
      Result : Unbounded_String := Empty_String;

   begin
      Result := Result &
                To_Unbounded_String ("property set AADL_Project is") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Default_Active_Thread_Handling_Protocol : constant Supported_Active_Thread_Handling_Protocols => abort;"
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    "Max_Aadlinteger : constant aadlinteger => 2#1#e32;") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Max_Memory_Size : constant aadlinteger Size_Units => 2#1#e32 B;") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    "Max_Queue_Size : constant aadlinteger => 512;") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    "Max_Base_Address : constant aadlinteger => 512;") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Max_Urgency : constant aadlinteger => 12;") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    "Max_Time : constant aadlinteger Time_Units => 1000 Hr;") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    "Max_Thread_Limit : constant aadlinteger => 32;") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    "Max_Word_Count : constant aadlinteger => 64000;") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    "Max_Word_Space : constant aadlinteger => 64;") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Supported_Active_Thread_Handling_Protocols : type enumeration (abort, complete_one_flush_queue, complete_one_transfer_queue, complete_one_preserve_queue, complete_all);"
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Supported_Connection_Protocols : type enumeration (UDP, TCP);") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Supported_Dispatch_Protocols : type enumeration (Periodic, Background, Poisson_Process, Sporadic, User_Defined);"
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Supported_Concurrency_Control_Protocols : type enumeration (No_Protocol, Priority_Ceiling_Protocol, Immediate_Priority_Ceiling_Protocol, Priority_Inheritance_Protocol);"
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Supported_Hardware_Source_Languages : type enumeration (VHDL);") &
                Unbounded_Lf;

      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    "Supported_Scheduling_Protocols : type enumeration (");
      for I in  Scheduler_Type'First .. Scheduler_Type'Last loop
         Result := Result & To_Unbounded_String (Scheduler_Type'Image (I));
         if I /= Scheduler_Type'Last then
            Result := Result & To_Unbounded_String (", ");
         else
            Result := Result & To_Unbounded_String (");") & Unbounded_Lf;
         end if;
      end loop;

      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Supported_Queue_Processing_Protocols : type enumeration (FIFO);") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Supported_Source_Languages : type enumeration (Ada95, C);") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Time_Units : type units (ps, Ns => ps * 1000, Us => Ns * 1000, Ms => Us * 1000, Sec => Ms * 1000, Min => Sec * 60, Hr => Min * 60);"
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"Size_Units : type units (Bits, B => Bits * 8, KB => B * 1000, MB => KB * 1000, GB => MB * 1000);"
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("end AADL_Project;") &
                Unbounded_Lf;

      return Result;

   end Export_Aadl_Project_Properties;

   function Export_Aadl_Cheddar_Properties
     (A_System : in System)
      return     Unbounded_String
   is
      Result : Unbounded_String;

   begin

      -- Export cheddar's specific properties
      --
      Result := Result &
                Unbounded_Lf &
                To_Unbounded_String ("property set Cheddar_Properties is") &
                Unbounded_Lf;

      -- Thread and thread group properties
      --
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Dispatch_Seed_is_Predictable : aadlboolean") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    ASCII.HT &
                    "applies to (thread, thread group);") &
                Unbounded_Lf;

      Result := Result &
                ASCII.HT &
                To_Unbounded_String(
"Source_Text : inherit list of aadlstring  applies to (data, thread, thread group, process, system, memory, bus, device, processor, port, subprogram, parameter, port group);"
) &
                Unbounded_Lf;

      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Dispatch_Seed_Value : aadlinteger") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    ASCII.HT &
                    "applies to (thread, thread group);") &
                Unbounded_Lf;

      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Dispatch_Absolute_Time :  inherit Time ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    ASCII.HT &
                    "applies to (thread, thread group);") &
                Unbounded_Lf;

      Result := Result &
                To_Unbounded_String (ASCII.HT & "Criticality : aadlinteger") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    ASCII.HT &
                    "applies to (thread, thread group);") &
                Unbounded_Lf;

      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Bound_On_Data_Blocking_Time :  inherit Time ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    ASCII.HT &
                    "applies to (thread, thread group);") &
                Unbounded_Lf;

      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Dispatch_Jitter :  inherit Time ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    ASCII.HT &
                    "applies to (thread, thread group);") &
                Unbounded_Lf;

      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Fixed_Priority : aadlinteger 0..255") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    ASCII.HT &
                    "applies to (thread, thread group);") &
                Unbounded_Lf;

      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & 
"POSIX_Scheduling_Policy : enumeration (SCHED_FIFO, SCHED_RR, SCHED_OTHERS)") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT &
                    ASCII.HT &
                    "applies to (thread, thread group);") &
                Unbounded_Lf;
      for I in  Offsets_Range'Range loop
         Result :=
           Result &
           To_Unbounded_String
              (ASCII.HT &
               "Dispatch_Offset_Value_" &
               To_String (Integer_Util.Format (Integer (I))) &
               " : aadlinteger") &
           Unbounded_Lf;
         Result := Result &
                   To_Unbounded_String
                      (ASCII.HT &
                       ASCII.HT &
                       "applies to (thread, thread group);") &
                   Unbounded_Lf;
         Result :=
           Result &
           To_Unbounded_String
              (ASCII.HT &
               "Dispatch_Offset_Time_" &
               To_String (Integer_Util.Format (Integer (I))) &
               " :  inherit Time ") &
           Unbounded_Lf;
         Result := Result &
                   To_Unbounded_String
                      (ASCII.HT &
                       ASCII.HT &
                       "applies to (thread, thread group);") &
                   Unbounded_Lf;
      end loop;

      -- System properties
      --
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Task_Precedencies : list of aadlstring") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & ASCII.HT & "applies to (system);") &
                Unbounded_Lf;

      -- Processor properties
      --
      Result := Result &
                Unbounded_Lf &
                To_Unbounded_String
                   (ASCII.HT & "Scheduler_Quantum :  inherit Time ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & ASCII.HT & "applies to (processor);") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Preemptive_Scheduler : aadlboolean") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & ASCII.HT & "applies to (processor);") &
                Unbounded_Lf &
                Unbounded_Lf;

      -- Data properties
      --
      Result := Result &
                Unbounded_Lf &
                To_Unbounded_String
                   (ASCII.HT & "Data_Concurrency_State : aadlinteger") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & ASCII.HT & "applies to (data);") &
                Unbounded_Lf &
                Unbounded_Lf;

      -- Process properties
      --
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Critical_Section : list of aadlstring") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & ASCII.HT & "applies to (process, thread, data);") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Source_Global_Heap_Size : Size ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & ASCII.HT & "applies to (process);") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Source_Global_Stack_Size : Size ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & ASCII.HT & "applies to (process);") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Source_Global_Text_Size : Size ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & ASCII.HT & "applies to (process);") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & "Source_Global_Data_Size : Size ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   (ASCII.HT & ASCII.HT & "applies to (process);") &
                Unbounded_Lf &
                Unbounded_Lf;

      Result := Result &
                To_Unbounded_String ("end Cheddar_Properties;") &
                Unbounded_Lf &
                Unbounded_Lf;

      return Result;

   end Export_Aadl_Cheddar_Properties;

   function Export_Aadl_Implementations
     (A_System : in System)
      return     Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
      Tmp    : Unbounded_String := Empty_String;

   begin

      -- Export all sub-components of the system
      --
      Result := Result & Export_Aadl_Implementations (A_System.Resources);
      Result := Result &
                Export_Aadl_Implementations
                   (A_System.Tasks,
                    A_System.Resources);
      Result := Result &
                Export_Aadl_Implementations
                   (A_System.Address_Spaces,
                    A_System.Tasks,
                    A_System.Resources);
      Result := Result & Export_Aadl_Implementations (A_System.Buffers);
      Result := Result & Export_Aadl_Implementations (A_System.Messages);
      Result := Result & Export_Aadl_Implementations (A_System.Networks);
      Result := Result & Export_Aadl_Implementations (A_System.Processors);

      -- Export the system composition
      --
      Result := Result &
                Unbounded_Lf &
                To_Unbounded_String ("system Cheddar") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("end Cheddar;") &
                Unbounded_Lf;
      Result := Result & Unbounded_Lf;

      Result := Result &
                To_Unbounded_String ("system implementation Cheddar.Impl") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String (ASCII.HT & "subcomponents") &
                Unbounded_Lf;

      -- List all processors, processes and networks
      --
      Result := Result & Export_Aadl_Declarations (A_System.Networks, 2);
      Result := Result & Export_Aadl_Declarations (A_System.Processors, 2);
      Result := Result &
                Export_Aadl_Declarations (A_System.Address_Spaces, 2);

      -- Properties related to the global system : processor and bus
      --connections, event
      -- analyzers, task precedencies
      --
      Tmp := Export_Aadl_Properties (A_System.Address_Spaces, 2);
      Tmp := Tmp & Export_Aadl_Properties (A_System.Event_Analyzers, 2);
      Tmp := Tmp & Export_Aadl_Properties (A_System.Dependencies, 2);
      if Tmp /= Empty_String then
         Result := Result &
                   To_Unbounded_String (ASCII.HT & "properties") &
                   Unbounded_Lf &
                   Tmp;
      end if;

      Result := Result &
                To_Unbounded_String ("end Cheddar.Impl;") &
                Unbounded_Lf &
                Unbounded_Lf;

      return Result;

   end Export_Aadl_Implementations;

   function Export_Aadl_Declarations
     (A_System     : in System;
      Number_Of_Ht : in Natural)
      return         Unbounded_String
   is
      Result : Unbounded_String := Empty_String;

   begin
      return Result;
   end Export_Aadl_Declarations;

   function Export_Xml_Cheddar_Dtd
     (A_System : in System)
      return     Unbounded_String
   is
      Result : Unbounded_String := Empty_String;
   begin

      Result := Result &
                To_Unbounded_String
                   ("<?xml version=""1.0"" standalone=""yes""?>  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String(
"<?xml-stylesheet type=""text/xsl"" href=""cheddar_project.xsl""?>") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!DOCTYPE cheddar [  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT name (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT scheduler (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT parameter (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT network_link (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT offset (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT period (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT capacity (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT deadline (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT blocking_time (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT policy (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT priority (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT criticality (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT cpu_name (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT address_space_name (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT jitter (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT seed (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT predictable_seed (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT activation_rule (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT state (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT protocol (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT size (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT qs (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT time (#PCDATA) > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT buffer_user (#PCDATA) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT resource_user (#PCDATA) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT event_Analyzer (#PCDATA) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT type (#PCDATA) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT text_memory_size (#PCDATA) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT heap_memory_size (#PCDATA) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT stack_memory_size (#PCDATA) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT data_memory_size (#PCDATA) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String(
"<!ELEMENT cheddar (processors, (address_spaces)?, (tasks)?, ((event_analyzers)?|(networks)?|(buffers)?|(resources)?|(messages)?),(dependencies)?  )  >   "
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT processors (processor)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT processor (name|scheduler|network_link) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT address_spaces (address_space)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String(
"<!ELEMENT address_space (name|text_memory_size|data_memory_size|stack_memory_size|heap_memory_size) >  "
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT networks (network)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT network (name|type) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT tasks (task)+>  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String(
"<!ELEMENT task (name|cpu_name|address_space_name|capacity|start_time|(stack_memory_size)?|(text_memory_size)?|(period)?|"
);
      Result := Result &
                To_Unbounded_String(
"(deadline)?|(parameters)?|(offsets)?|(jitter)?|(policy)?|(criticality)?|(priority)?|(predictable_seed)?|(blocking_time)?|(seed)?"
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("|(activation_rule)?) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT offsets (offset)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT parameters (parameter)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT messages (message)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String(
"<!ELEMENT message (name|size|(period)?|(deadline)?|(jitter)?) >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT buffers (buffer)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String(
"<!ELEMENT buffer (cpu_name|address_space_name|qs|name|size|(buffer_used_by)?) >  "
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT buffer_used_by (buffer_user)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT resources (resource)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String(
"<!ELEMENT resource (cpu_name|address_space_name|name|protocol|(state)?|(resource_used_by)?) >  "
) &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT resource_used_by (resource_user)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT dependencies (dependency)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("<!ELEMENT event_analyzers (event_analyzer)+ >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ELEMENT dependency (#PCDATA)  >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ATTLIST scheduler ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("        quantum CDATA  ""0""  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("        is_preemptive CDATA  ""PREEMPTIVE""  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("        parametric_file_name CDATA """" >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ATTLIST event_analyzer  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("        parametric_file_name CDATA """" >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ATTLIST task  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("        task_type CDATA  ""APERIODIC_TYPE""  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("        x CDATA ""0""  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("        y CDATA ""0"" > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ATTLIST buffer  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("        x CDATA ""0"" ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("        y CDATA ""0"" > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ATTLIST message  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("        x CDATA ""0"" ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("        y CDATA ""0"" > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ATTLIST buffer_user ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("        buffer_role CDATA  ""producer"" >  ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ATTLIST dependency ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("        from_type CDATA ""task"" ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("        to_type CDATA ""task"" > ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String ("<!ATTLIST parameter ") &
                Unbounded_Lf;
      Result := Result &
                To_Unbounded_String
                   ("        parameter_type CDATA  ""integer"" >  ") &
                Unbounded_Lf;
      Result := Result & To_Unbounded_String ("] >  ") & Unbounded_Lf;
      Result := Result & Unbounded_Lf & Unbounded_Lf & Unbounded_Lf;

      return Result;

   end Export_Xml_Cheddar_Dtd;

   function Export_Xml (A_System : in System) return Unbounded_String is
      Result : Unbounded_String := Empty_String;

   begin

      Result := Result & To_Unbounded_String ("<cheddar>") & Unbounded_Lf;

      -- Call each export part of the system
      --
      Result := Result & Export_Xml (A_System.Processors);
      Result := Result & Export_Xml (A_System.Address_Spaces);
      Result := Result & Export_Xml (A_System.Tasks);
      Result := Result & Export_Xml (A_System.Networks);
      Result := Result & Export_Xml (A_System.Buffers);
      Result := Result & Export_Xml (A_System.Messages);
      Result := Result & Export_Xml (A_System.Resources);
      Result := Result & Export_Xml (A_System.Dependencies);
      Result := Result & Export_Xml (A_System.Event_Analyzers);

      Result := Result &
                To_Unbounded_String ("</cheddar>") &
                Unbounded_Lf &
                Unbounded_Lf;
      Result := Result & Unbounded_Lf & Unbounded_Lf;

      return Result;

   end Export_Xml;

end Systems;
