------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------




WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Messages;
USE Messages;
WITH Message_Set;
USE Message_Set;
USE Message_Set.Generic_Message_Set;
WITH Buffers;
USE Buffers;
WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;
WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Task_Dependencies;
USE Task_Dependencies;
USE Task_Dependencies.Half_Dep_Set;
WITH Processors;
USE Processors;
WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;
WITH Address_Spaces;
USE Address_Spaces;
WITH Address_Space_Set;
USE Address_Space_Set;
USE Address_Space_Set.Generic_Address_Space_Set;
WITH Resource_Set;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;
WITH Network_Set;
USE Network_Set;
USE Network_Set.Generic_Network_Set;
WITH Event_Analyzer_Set;
USE Event_Analyzer_Set;
USE Event_Analyzer_Set.Generic_Event_Analyzer_Set;
WITH Unbounded_Strings;
USE Unbounded_Strings;
USE Unbounded_Strings.Unbounded_String_List_Package;



PACKAGE Systems IS




   TYPE System IS TAGGED
      RECORD
         Tasks           : Tasks_Set;
         Resources       : Resources_Set;
         Messages        : Messages_Set;
         Dependencies    : Tasks_Dependencies_Ptr;
         Processors      : Processors_Set;
         Buffers         : Buffers_Set;
         Networks        : Networks_Set;
         Event_Analyzers : Event_Analyzers_Set;
         Address_Spaces  : Address_Spaces_Set;
      END RECORD;

   TYPE System_Ptr IS ACCESS System;




   PROCEDURE Initialize (
         A_System : IN OUT System);
   PROCEDURE Initialize (
         A_System : IN     System_Ptr);


   PROCEDURE Duplicate (
         Src  : IN     System;
         Dest : IN OUT System);



   -- Delete sub_programs : sometimes, deleting a
   -- cheddar object requires to delete other cheddar
   -- objects. These sub-programs implement such multiple
   -- delete operation
   --
   PROCEDURE Delete_Processor (
         A_Sys       : IN OUT System;
         A_Processor :        Processor_Ptr);
   PROCEDURE Delete_Buffer (
         A_Sys    : IN OUT System;
         A_Buffer :        Buffer_Ptr);
   PROCEDURE Delete_Message (
         A_Sys     : IN OUT System;
         A_Message :        Message_Ptr);
   PROCEDURE Delete_Task (
         A_Sys  : IN OUT System;
         A_Task :        Generic_Task_Ptr);
   PROCEDURE Delete_Address_Space (
         A_Sys            : IN OUT System;
         An_Address_Space :        Address_Space_Ptr);


   -- Display a system to the screen in different languages
   --
   PROCEDURE Put_Xml (
         A_System : IN     System);
   PROCEDURE Put_Aadl (
         A_System : IN     System);
   PROCEDURE Put (
         A_System : IN     System_Ptr);



   -- I/O sub-programs : read/write a system in different languages
   --

   PROCEDURE Read_From_Aadl_File (
         A_System          : IN OUT System;
         Dir_List          : IN     Unbounded_String_List;
         Project_File_List : IN     Unbounded_String_List);
   PROCEDURE Read_From_Xml_File (
         A_System  : IN OUT System;
         Dir_List  : IN     Unbounded_String_List;
         File_Name : IN     Unbounded_String);
   PROCEDURE Read_From_Xml_File (
         A_System  : IN OUT System;
         Dir_List  : IN     Unbounded_String_List;
         File_Name : IN     String);


   PROCEDURE Write_To_Xml_File (
         A_System  : IN     System;
         File_Name : IN     Unbounded_String);
   PROCEDURE Write_To_Aadl_File (
         A_System  : IN     System;
         File_Name : IN     Unbounded_String);
   PROCEDURE Write_To_Xml_File (
         A_System  : IN     System;
         File_Name : IN     String);
   PROCEDURE Write_Aadl_Standard_Properties_Set_To_File (
         A_System : IN     System);
   PROCEDURE Write_Cheddar_Property_Sets_To_File (
         A_System : IN     System);



   FUNCTION Export_Xml (
         A_System : IN     System)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Cheddar_Dtd (
         A_System : IN     System)
     RETURN Unbounded_String;



   FUNCTION Export_Aadl_Project_Properties (
         A_System : IN     System)
     RETURN Unbounded_String;
   FUNCTION Export_Aadl_Cheddar_Properties (
         A_System : IN     System)
     RETURN Unbounded_String;
   FUNCTION Export_Aadl_User_Defined_Cheddar_Properties (
         A_System : IN     System)
     RETURN Unbounded_String;
   FUNCTION Export_Aadl_Standard_Properties (
         A_System : IN     System)
     RETURN Unbounded_String;
   FUNCTION Export_Aadl_Implementations (
         A_System : IN     System)
     RETURN Unbounded_String;
   FUNCTION Export_Aadl_Declarations (
         A_System     : IN     System;
         Number_Of_Ht : IN     Natural)
     RETURN Unbounded_String;





END Systems;








