------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Aadl_Config;       use Aadl_Config;
with Unbounded_Strings; use Unbounded_Strings;
with Text_IO;           use Text_IO;

package body Task_Dependencies is

   -- Add a task in the Dependent task list
   -- only if it now already added
   --
   procedure Add_Dependent_Tasks
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
   is

      Ite1   : Tasks_Iterator;
      A_Task : Generic_Task_Ptr;

   begin

      if not Is_Empty (My_Dependencies.Dependent_Tasks) then
         Reset_Iterator (My_Dependencies.Dependent_Tasks, Ite1);
         loop
            Current_Element (My_Dependencies.Dependent_Tasks, A_Task, Ite1);

            if A_Task.Name = My_Task.Name then
               return;
            end if;

            exit when Is_Last_Element (My_Dependencies.Dependent_Tasks, Ite1);
            Next_Element (My_Dependencies.Dependent_Tasks, Ite1);
         end loop;
      end if;

      Add (My_Dependencies.Dependent_Tasks, My_Task);

   end Add_Dependent_Tasks;

   -------------------------------------------------------------------------

   procedure Free (My_Dependencies : in out Tasks_Dependencies_Ptr) is
   begin
      Free (My_Dependencies.Dependent_Tasks);
      Free (Half_Dep_Set.Set (My_Dependencies.all));
   end Free;

   function Export_Xml
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Unbounded_String
   is
      My_Iterator : Tasks_Dependencies_Iterator;
      A_Half_Dep  : Dependency_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator);

         Result := Result &
                   To_Unbounded_String ("<dependencies>") &
                   Unbounded_Lf;
         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator);

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "<dependency");

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if A_Half_Dep.Buffer_Orientation = From_Task_To_Object then
                  Result :=
                    Result &
                    To_Unbounded_String
                       (" to_type=""buffer""> " &
                        To_String (A_Half_Dep.Buffer_Dependent_Task.Name) &
                        " " &
                        To_String (A_Half_Dep.Buffer_Dependency_Object.Name) &
                        " ");
               else
                  Result :=
                    Result &
                    To_Unbounded_String
                       (" from_type=""buffer""> " &
                        To_String (A_Half_Dep.Buffer_Dependency_Object.Name) &
                        " " &
                        To_String (A_Half_Dep.Buffer_Dependent_Task.Name) &
                        " ");
               end if;
            else
               if A_Half_Dep.Depend_Type = Communication_Dependency then
                  if A_Half_Dep.Communication_Orientation =
                     From_Task_To_Object
                  then
                     Result :=
                       Result &
                       To_Unbounded_String
                          (" to_type=""message""> " &
                           To_String
                              (A_Half_Dep.Communication_Dependent_Task.Name) &
                           " " &
                           To_String
                              (A_Half_Dep.Communication_Dependency_Object.Name)
                           &
                           " ");
                  else
                     Result :=
                       Result &
                       To_Unbounded_String
                          (" from_type=""message""> " &
                           To_String
                              (A_Half_Dep.Communication_Dependency_Object.Name)
                           &
                           " " &
                           To_String
                              (A_Half_Dep.Communication_Dependent_Task.Name) &
                           " ");
                  end if;
               else
                  Result :=
                    Result &
                    To_Unbounded_String
                       ("> " &
                        To_String (A_Half_Dep.Precedency_From_Task.Name) &
                        " " &
                        To_String (A_Half_Dep.Precedency_To_Task.Name) &
                        " ");
               end if;
            end if;

            Result := Result &
                      To_Unbounded_String ("</dependency>") &
                      Unbounded_Lf;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator);

            Next_Element (My_Dependencies.all, My_Iterator);

         end loop;
         Result := Result &
                   To_Unbounded_String ("</dependencies>") &
                   Unbounded_Lf;

      end if;

      return Result;

   end Export_Xml;

   function Export_Aadl_Properties
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      Number_Of_Ht    : in Natural)
      return            Unbounded_String
   is
      My_Iterator     : Tasks_Dependencies_Iterator;
      A_Half_Dep      : Dependency_Ptr;
      Head_Is_Written : Boolean := False;

      Result : Unbounded_String := Empty_String;

   begin

      if Aadl_Export_Precedencies_To_Properties then

         if not Is_Empty (My_Dependencies.all) then

            Reset_Iterator (My_Dependencies.all, My_Iterator);
            loop
               Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator);

               if A_Half_Dep.Depend_Type = Precedence_Dependency then

                  if not Head_Is_Written then
                     for I in  1 .. Number_Of_Ht loop
                        Result := Result & ASCII.HT;
                     end loop;
                     Result          := Result &
                                        To_Unbounded_String(
"Cheddar_Properties::Task_Precedencies => (") &
                                        Unbounded_Lf;
                     Head_Is_Written := True;
                  else
                     Result := Result & "," & Unbounded_Lf;
                  end if;

                  for I in  1 .. Number_Of_Ht + 1 loop
                     Result := Result & ASCII.HT;
                  end loop;
                  Result :=
                    Result &
                    """" &
                    "instancied_" &
                    A_Half_Dep.Precedency_From_Task.Address_Space_Name &
                    ".instancied_" &
                    A_Half_Dep.Precedency_From_Task.Name &
                    """" &
                    ", " &
                    """" &
                    "instancied_" &
                    A_Half_Dep.Precedency_From_Task.Address_Space_Name &
                    ".instancied_" &
                    A_Half_Dep.Precedency_To_Task.Name &
                    """";
               end if;

               exit when Is_Last_Element (My_Dependencies.all, My_Iterator);

               Next_Element (My_Dependencies.all, My_Iterator);

            end loop;

            if Head_Is_Written then
               Result := Result & To_Unbounded_String (");") & Unbounded_Lf;
            end if;
         end if;

      end if;

      return Result;

   end Export_Aadl_Properties;

   procedure Duplicate
     (Src  : in Tasks_Dependencies_Ptr;
      Dest : in out Tasks_Dependencies_Ptr)
   is

   begin
      Duplicate (Half_Dep_Set.Set (Src.all), Half_Dep_Set.Set (Dest.all));
      Duplicate (Src.Dependent_Tasks, Dest.Dependent_Tasks);
   end Duplicate;

   procedure Reset
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      Free_Object     : in Boolean := False)
   is

   begin
      Reset (Half_Dep_Set.Set (My_Dependencies.all), Free_Object);
      Reset (My_Dependencies.Dependent_Tasks, Free_Object);
   end Reset;

   procedure Put (My_Dependencies : in Tasks_Dependencies_Ptr) is

   begin
      Put ("Task list/index : ");
      New_Line;

      Put ("Dependencies : ");
      New_Line;
      Put (My_Dependencies.all);
   end Put;

   ------------------------------------------------------------------------

   ---------------------------------------
   --  Services below allow to add
   --  dependencies.
   ---------------------------------------

   -- Add ALL dependencies taken from the definition
   -- of a given buffer
   --
   procedure Add_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      My_Tasks        : in Tasks_Set;
      A_Buffer        : in Buffer_Ptr)
   is

      Task1 : Generic_Task_Ptr;

   begin

      Delete_All_Task_Dependencies (My_Dependencies, A_Buffer);

      for I in  0 .. A_Buffer.Roles.Nb_Entries - 1 loop
         Task1 := Search_Task (My_Tasks, A_Buffer.Roles.Entries (I).Item);
         if A_Buffer.Roles.Entries (I).Data.The_Role = Producer then
            Add_One_Task_Dependency
              (My_Dependencies,
               Task1,
               A_Buffer,
               From_Task_To_Object);
         else
            Add_One_Task_Dependency
              (My_Dependencies,
               Task1,
               A_Buffer,
               From_Object_To_Task);
         end if;
      end loop;

   end Add_All_Task_Dependencies;

   procedure Add_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr;
      A_Dep           : in Message_Ptr;
      A_Type          : in Orientation_Dependency_Type)
   is

      A_Dep_Ptr : Dependency_Ptr;
   begin
      A_Dep_Ptr                                 :=
        new Dependency (Communication_Dependency);
      A_Dep_Ptr.Communication_Dependent_Task    := A_Task;
      A_Dep_Ptr.Communication_Orientation       := A_Type;
      A_Dep_Ptr.Communication_Dependency_Object := A_Dep;

      Add (My_Dependencies.all, A_Dep_Ptr);

      Add_Dependent_Tasks (My_Dependencies, A_Task);
   end Add_One_Task_Dependency;

   procedure Add_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr;
      A_Dep           : in Buffer_Ptr;
      A_Type          : in Orientation_Dependency_Type)
   is

      A_Dep_Ptr : Dependency_Ptr;
   begin
      A_Dep_Ptr                          :=
        new Dependency (Buffer_Dependency);
      A_Dep_Ptr.Buffer_Dependent_Task    := A_Task;
      A_Dep_Ptr.Buffer_Orientation       := A_Type;
      A_Dep_Ptr.Buffer_Dependency_Object := A_Dep;

      Add (My_Dependencies.all, A_Dep_Ptr);

      Add_Dependent_Tasks (My_Dependencies, A_Task);
   end Add_One_Task_Dependency;

   procedure Add_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      From_Task       : in Generic_Task_Ptr;
      To_Task         : in Generic_Task_Ptr)
   is

      A_Dep_Ptr : Dependency_Ptr;
   begin
      A_Dep_Ptr                      :=
        new Dependency (Precedence_Dependency);
      A_Dep_Ptr.Precedency_From_Task := From_Task;
      A_Dep_Ptr.Precedency_To_Task   := To_Task;

      Add (My_Dependencies.all, A_Dep_Ptr);

      Add_Dependent_Tasks (My_Dependencies, From_Task);
      Add_Dependent_Tasks (My_Dependencies, To_Task);
   end Add_One_Task_Dependency;

   ------------------------------------------------------------------

   function Has_Buffer_To_Read
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Boolean
   is
      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if (A_Half_Dep.Buffer_Dependent_Task.Name = My_Task.Name) and
                  (A_Half_Dep.Buffer_Orientation = From_Object_To_Task)
               then
                  return True;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      return False;

   end Has_Buffer_To_Read;

   function Has_Buffer_To_Write
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Boolean
   is
      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if (A_Half_Dep.Buffer_Dependent_Task.Name = My_Task.Name) and
                  (A_Half_Dep.Buffer_Orientation = From_Task_To_Object)
               then
                  return True;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      return False;

   end Has_Buffer_To_Write;

   function Get_A_Successor
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Generic_Task_Ptr
   is
      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_From_Task.Name = My_Task.Name) then
                  return Copy (A_Half_Dep.Precedency_To_Task);
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      raise Dependency_Not_Found;
      return null;

   end Get_A_Successor;

   function Get_A_Predecessor
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Generic_Task_Ptr
   is
      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_To_Task.Name = My_Task.Name) then
                  return Copy (A_Half_Dep.Precedency_From_Task);
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      raise Dependency_Not_Found;
      return null;

   end Get_A_Predecessor;

   function Has_Predecessor
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Boolean
   is
      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_To_Task.Name = My_Task.Name) then
                  return True;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      return False;
   end Has_Predecessor;

   function Has_Successor
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Boolean
   is
      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_From_Task.Name = My_Task.Name) then
                  return True;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      return False;
   end Has_Successor;

   function Get_Successors_List
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Tasks_Set
   is
      Result   : Tasks_Set;
      New_Task : Generic_Task_Ptr;

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_From_Task.Name = My_Task.Name) then
                  New_Task := Copy (A_Half_Dep.Precedency_To_Task);
                  Add (Result, New_Task);
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      return Result;
   end Get_Successors_List;

   function Get_Predecessors_List
     (My_Dependencies : in Tasks_Dependencies_Ptr;
      My_Task         : in Generic_Task_Ptr)
      return            Tasks_Set
   is
      Result   : Tasks_Set;
      New_Task : Generic_Task_Ptr;

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_To_Task.Name = My_Task.Name) then
                  New_Task := Copy (A_Half_Dep.Precedency_From_Task);
                  Add (Result, New_Task);
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      return Result;
   end Get_Predecessors_List;

   function Get_A_Root_Task
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Generic_Task_Ptr
   is
      Is_Root : Boolean;

      My_Iterator2 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;
      My_Iterator1 : Tasks_Iterator;
      A_Task       : Generic_Task_Ptr;

   begin

      if not Is_Empty (My_Dependencies.Dependent_Tasks) then

         Reset_Iterator (My_Dependencies.Dependent_Tasks, My_Iterator1);

         loop
            Current_Element
              (My_Dependencies.Dependent_Tasks,
               A_Task,
               My_Iterator1);

            Reset_Iterator (My_Dependencies.all, My_Iterator2);

            Is_Root := True;
            loop
               Current_Element
                 (My_Dependencies.all,
                  A_Half_Dep,
                  My_Iterator2);

               if A_Half_Dep.Depend_Type = Buffer_Dependency then
                  if (A_Half_Dep.Buffer_Orientation = From_Object_To_Task) and
                     (A_Task.Name = A_Half_Dep.Buffer_Dependent_Task.Name)
                  then
                     Is_Root := False;
                  end if;
               end if;

               if A_Half_Dep.Depend_Type = Communication_Dependency then
                  if (A_Half_Dep.Communication_Orientation =
                      From_Object_To_Task) and
                     (A_Task.Name =
                      A_Half_Dep.Communication_Dependent_Task.Name)
                  then
                     Is_Root := False;
                  end if;
               end if;

               if A_Half_Dep.Depend_Type = Precedence_Dependency then
                  if (A_Half_Dep.Precedency_To_Task.Name = A_Task.Name) then
                     Is_Root := False;
                  end if;
               end if;

               exit when Is_Last_Element (My_Dependencies.all, My_Iterator2);

               Next_Element (My_Dependencies.all, My_Iterator2);
            end loop;

            if Is_Root then
               return A_Task;
            end if;

            exit when Is_Last_Element
                         (My_Dependencies.Dependent_Tasks,
                          My_Iterator1);

            Next_Element (My_Dependencies.Dependent_Tasks, My_Iterator1);

         end loop;
      end if;

      raise Dependency_Not_Found;
      return null;
   end Get_A_Root_Task;

   function Get_A_Leaf_Task
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Generic_Task_Ptr
   is
      Is_Leaf : Boolean;

      My_Iterator2 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;
      My_Iterator1 : Tasks_Iterator;
      A_Task       : Generic_Task_Ptr;

   begin

      if not Is_Empty (My_Dependencies.Dependent_Tasks) then

         Reset_Iterator (My_Dependencies.Dependent_Tasks, My_Iterator1);

         loop
            Current_Element
              (My_Dependencies.Dependent_Tasks,
               A_Task,
               My_Iterator1);

            Reset_Iterator (My_Dependencies.all, My_Iterator2);

            Is_Leaf := True;
            loop
               Current_Element
                 (My_Dependencies.all,
                  A_Half_Dep,
                  My_Iterator2);

               if A_Half_Dep.Depend_Type = Buffer_Dependency then
                  if (A_Half_Dep.Buffer_Orientation = From_Task_To_Object) and
                     (A_Task.Name = A_Half_Dep.Buffer_Dependent_Task.Name)
                  then
                     Is_Leaf := False;
                  end if;
               end if;

               if A_Half_Dep.Depend_Type = Communication_Dependency then
                  if (A_Half_Dep.Communication_Orientation =
                      From_Task_To_Object) and
                     (A_Task.Name =
                      A_Half_Dep.Communication_Dependent_Task.Name)
                  then
                     Is_Leaf := False;
                  end if;
               end if;

               if A_Half_Dep.Depend_Type = Precedence_Dependency then
                  if (A_Half_Dep.Precedency_From_Task.Name =
                      A_Task.Name)
                  then
                     Is_Leaf := False;
                  end if;
               end if;

               exit when Is_Last_Element (My_Dependencies.all, My_Iterator2);

               Next_Element (My_Dependencies.all, My_Iterator2);
            end loop;

            if Is_Leaf then
               return A_Task;
            end if;

            exit when Is_Last_Element
                         (My_Dependencies.Dependent_Tasks,
                          My_Iterator1);

            Next_Element (My_Dependencies.Dependent_Tasks, My_Iterator1);

         end loop;
      end if;

      raise Dependency_Not_Found;
      return null;
   end Get_A_Leaf_Task;

   function Get_Leaf_Tasks
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Tasks_Set
   is
      Result   : Tasks_Set;
      Is_Leaf  : Boolean;
      New_Task : Generic_Task_Ptr;

      My_Iterator2 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;
      My_Iterator1 : Tasks_Iterator;
      A_Task       : Generic_Task_Ptr;

   begin

      if not Is_Empty (My_Dependencies.Dependent_Tasks) then

         Reset_Iterator (My_Dependencies.Dependent_Tasks, My_Iterator1);

         loop
            Current_Element
              (My_Dependencies.Dependent_Tasks,
               A_Task,
               My_Iterator1);

            Reset_Iterator (My_Dependencies.all, My_Iterator2);
            Is_Leaf := True;
            loop
               Current_Element
                 (My_Dependencies.all,
                  A_Half_Dep,
                  My_Iterator2);

               if A_Half_Dep.Depend_Type = Buffer_Dependency then
                  if (A_Half_Dep.Buffer_Orientation = From_Task_To_Object) and
                     (A_Task.Name = A_Half_Dep.Buffer_Dependent_Task.Name)
                  then
                     Is_Leaf := False;
                  end if;
               end if;

               if A_Half_Dep.Depend_Type = Communication_Dependency then
                  if (A_Half_Dep.Communication_Orientation =
                      From_Task_To_Object) and
                     (A_Task.Name =
                      A_Half_Dep.Communication_Dependent_Task.Name)
                  then
                     Is_Leaf := False;
                  end if;
               end if;

               if A_Half_Dep.Depend_Type = Precedence_Dependency then
                  if (A_Half_Dep.Precedency_From_Task.Name =
                      A_Task.Name)
                  then
                     Is_Leaf := False;
                  end if;
               end if;

               exit when Is_Last_Element (My_Dependencies.all, My_Iterator2);

               Next_Element (My_Dependencies.all, My_Iterator2);
            end loop;

            if Is_Leaf then
               New_Task := Copy (A_Task);
               Add (Result, New_Task);
            end if;

            exit when Is_Last_Element
                         (My_Dependencies.Dependent_Tasks,
                          My_Iterator1);

            Next_Element (My_Dependencies.Dependent_Tasks, My_Iterator1);

         end loop;
      end if;
      return Result;
   end Get_Leaf_Tasks;

   function Get_Root_Tasks
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Tasks_Set
   is
      Result   : Tasks_Set;
      Is_Root  : Boolean;
      New_Task : Generic_Task_Ptr;

      My_Iterator2 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;
      My_Iterator1 : Tasks_Iterator;
      A_Task       : Generic_Task_Ptr;

   begin

      if not Is_Empty (My_Dependencies.Dependent_Tasks) then

         Reset_Iterator (My_Dependencies.Dependent_Tasks, My_Iterator1);

         loop
            Current_Element
              (My_Dependencies.Dependent_Tasks,
               A_Task,
               My_Iterator1);

            Reset_Iterator (My_Dependencies.all, My_Iterator2);

            Is_Root := True;
            loop
               Current_Element
                 (My_Dependencies.all,
                  A_Half_Dep,
                  My_Iterator2);

               if A_Half_Dep.Depend_Type = Buffer_Dependency then
                  if (A_Half_Dep.Buffer_Orientation = From_Object_To_Task) and
                     (A_Task.Name = A_Half_Dep.Buffer_Dependent_Task.Name)
                  then
                     Is_Root := False;
                  end if;
               end if;

               if A_Half_Dep.Depend_Type = Communication_Dependency then
                  if (A_Half_Dep.Communication_Orientation =
                      From_Object_To_Task) and
                     (A_Task.Name =
                      A_Half_Dep.Communication_Dependent_Task.Name)
                  then
                     Is_Root := False;
                  end if;
               end if;

               if A_Half_Dep.Depend_Type = Precedence_Dependency then
                  if (A_Half_Dep.Precedency_To_Task.Name = A_Task.Name) then
                     Is_Root := False;
                  end if;
               end if;

               exit when Is_Last_Element (My_Dependencies.all, My_Iterator2);

               Next_Element (My_Dependencies.all, My_Iterator2);
            end loop;

            if Is_Root then
               New_Task := Copy (A_Task);
               Add (Result, New_Task);
            end if;

            exit when Is_Last_Element
                         (My_Dependencies.Dependent_Tasks,
                          My_Iterator1);

            Next_Element (My_Dependencies.Dependent_Tasks, My_Iterator1);

         end loop;
      end if;
      return Result;

   end Get_Root_Tasks;

   -------------------------------------------------------------------

   -- Delete ALL dependencies  of an
   -- address space (parameter "An_Address_Space")
   --
   procedure Delete_Address_Space_All_Task_Dependencies
     (My_Dependencies  : in out Tasks_Dependencies_Ptr;
      An_Address_Space : in Unbounded_String)
   is

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr := new Tasks_Dependencies;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_From_Task.Address_Space_Name =
                   An_Address_Space) or
                  (A_Half_Dep.Precedency_To_Task.Address_Space_Name =
                   An_Address_Space)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if (A_Half_Dep.Buffer_Orientation = From_Object_To_Task) and
                  (An_Address_Space =
                   A_Half_Dep.Buffer_Dependent_Task.Address_Space_Name)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;

            if A_Half_Dep.Depend_Type = Communication_Dependency then
               if (A_Half_Dep.Communication_Orientation =
                   From_Object_To_Task) and
                  (An_Address_Space =
                   A_Half_Dep.Communication_Dependent_Task.Address_Space_Name)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      if not Is_Empty (To_Be_Deleted.all) then
         Delete (My_Dependencies.all, To_Be_Deleted.all);
         Free (Half_Dep_Set.Set_Ptr (To_Be_Deleted), False);
      end if;

   end Delete_Address_Space_All_Task_Dependencies;

   -- Delete ALL dependencies  of a
   -- processor (parameter "A_Processor")
   --
   procedure Delete_Processor_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Processor     : in Unbounded_String)
   is

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr := new Tasks_Dependencies;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_From_Task.Cpu_Name = A_Processor) or
                  (A_Half_Dep.Precedency_To_Task.Cpu_Name = A_Processor)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if (A_Half_Dep.Buffer_Orientation = From_Object_To_Task) and
                  (A_Processor = A_Half_Dep.Buffer_Dependent_Task.Cpu_Name)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;

            if A_Half_Dep.Depend_Type = Communication_Dependency then
               if (A_Half_Dep.Communication_Orientation =
                   From_Object_To_Task) and
                  (A_Processor =
                   A_Half_Dep.Communication_Dependent_Task.Cpu_Name)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      if not Is_Empty (To_Be_Deleted.all) then
         Delete (My_Dependencies.all, To_Be_Deleted.all);
         Free (Half_Dep_Set.Set_Ptr (To_Be_Deleted), False);
      end if;

   end Delete_Processor_All_Task_Dependencies;

   -- Delete ALL dependencies :
   --  1) of a task (if My_Dep_Name is a Task_ptr)
   --  2) of a message  (if My_Dep_Name is a Message_ptr)
   --  3) of a buffer  (if My_Dep_Name is a Buffer_ptr)
   --
   procedure Delete_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr)
   is

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr := new Tasks_Dependencies;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if (A_Half_Dep.Buffer_Dependent_Task.Name = A_Task.Name) then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;
            if A_Half_Dep.Depend_Type = Communication_Dependency then
               if (A_Half_Dep.Communication_Dependent_Task.Name =
                   A_Task.Name)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;
            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_From_Task.Name = A_Task.Name) or
                  (A_Half_Dep.Precedency_To_Task.Name = A_Task.Name)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      if not Is_Empty (To_Be_Deleted.all) then
         Delete (My_Dependencies.all, To_Be_Deleted.all);
         Free (Half_Dep_Set.Set_Ptr (To_Be_Deleted), False);
      end if;

   end Delete_All_Task_Dependencies;

   procedure Delete_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Buffer        : in Buffer_Ptr)
   is

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr := new Tasks_Dependencies;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if (A_Half_Dep.Buffer_Dependency_Object.Name =
                   A_Buffer.Name)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      if not Is_Empty (To_Be_Deleted.all) then
         Delete (My_Dependencies.all, To_Be_Deleted.all);
         Free (Half_Dep_Set.Set_Ptr (To_Be_Deleted), False);
      end if;

   end Delete_All_Task_Dependencies;

   procedure Delete_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Message       : in Message_Ptr)
   is

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr := new Tasks_Dependencies;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Communication_Dependency then
               if (A_Half_Dep.Communication_Dependency_Object.Name =
                   A_Message.Name)
               then
                  Add (To_Be_Deleted.all, A_Half_Dep);
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      if not Is_Empty (To_Be_Deleted.all) then
         Delete (My_Dependencies.all, To_Be_Deleted.all);
         Free (Half_Dep_Set.Set_Ptr (To_Be_Deleted), False);
      end if;

   end Delete_All_Task_Dependencies;

   -- Remove ONE dependency between two tasks
   -- (remove a precedency relationship, not with buffer of message)
   --
   procedure Delete_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      From_Task       : in Generic_Task_Ptr;
      To_Task         : in Generic_Task_Ptr)
   is

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_From_Task.Name = From_Task.Name) and
                  (A_Half_Dep.Precedency_To_Task.Name = To_Task.Name)
               then
                  Delete (My_Dependencies.all, A_Half_Dep);
                  return;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      raise Dependency_Not_Found;
   end Delete_One_Task_Dependency;

   -- Remove ONE dependency between a task and
   --   a message
   --
   procedure Delete_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr;
      A_Buffer        : in Buffer_Ptr;
      A_Type          : in Orientation_Dependency_Type)
   is

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if (A_Half_Dep.Buffer_Dependent_Task.Name = A_Task.Name) and
                  (A_Half_Dep.Buffer_Dependency_Object.Name = A_Buffer.Name)
                  and
                  (A_Half_Dep.Buffer_Orientation = A_Type)
               then
                  Delete (My_Dependencies.all, A_Half_Dep);
                  return;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      raise Dependency_Not_Found;
   end Delete_One_Task_Dependency;

   -- Remove ONE dependency between a task and
   --   a message
   --
   procedure Delete_One_Task_Dependency
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr;
      A_Message       : in Message_Ptr;
      A_Type          : in Orientation_Dependency_Type)
   is

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Communication_Dependency then
               if (A_Half_Dep.Communication_Dependent_Task.Name =
                   A_Task.Name) and
                  (A_Half_Dep.Communication_Dependency_Object.Name =
                   A_Message.Name) and
                  (A_Half_Dep.Communication_Orientation = A_Type)
               then
                  Delete (My_Dependencies.all, A_Half_Dep);
                  return;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

      raise Dependency_Not_Found;
   end Delete_One_Task_Dependency;

   -------------------------------------------------------------------

   procedure Update_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Task          : in Generic_Task_Ptr)
   is

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if (A_Half_Dep.Buffer_Dependent_Task.Name = A_Task.Name) then
                  A_Half_Dep.Buffer_Dependent_Task := A_Task;
               end if;
            end if;
            if A_Half_Dep.Depend_Type = Communication_Dependency then
               if (A_Half_Dep.Communication_Dependent_Task.Name =
                   A_Task.Name)
               then
                  A_Half_Dep.Communication_Dependent_Task := A_Task;
               end if;
            end if;
            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_From_Task.Name = A_Task.Name) then
                  A_Half_Dep.Precedency_From_Task := A_Task;
               end if;
            end if;
            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if (A_Half_Dep.Precedency_To_Task.Name = A_Task.Name) then
                  A_Half_Dep.Precedency_To_Task := A_Task;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

   end Update_All_Task_Dependencies;

   procedure Update_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Buffer        : in Buffer_Ptr)
   is

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if (A_Half_Dep.Buffer_Dependency_Object.Name =
                   A_Buffer.Name)
               then
                  A_Half_Dep.Buffer_Dependency_Object := A_Buffer;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

   end Update_All_Task_Dependencies;

   procedure Update_All_Task_Dependencies
     (My_Dependencies : in out Tasks_Dependencies_Ptr;
      A_Message       : in Message_Ptr)
   is

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   begin

      if not Is_Empty (My_Dependencies.all) then

         Reset_Iterator (My_Dependencies.all, My_Iterator1);

         loop
            Current_Element (My_Dependencies.all, A_Half_Dep, My_Iterator1);

            if A_Half_Dep.Depend_Type = Communication_Dependency then
               if (A_Half_Dep.Communication_Dependency_Object.Name =
                   A_Message.Name)
               then
                  A_Half_Dep.Communication_Dependency_Object := A_Message;
               end if;
            end if;

            exit when Is_Last_Element (My_Dependencies.all, My_Iterator1);

            Next_Element (My_Dependencies.all, My_Iterator1);
         end loop;
      end if;

   end Update_All_Task_Dependencies;

   -------------------------------------------------------------------

   -- Check is a task dependencies is acyclic or not
   --
   function Is_Cyclic
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Boolean
   is
   begin
      return not Is_Acyclic (My_Dependencies);
   end Is_Cyclic;

   function Is_Acyclic
     (My_Dependencies : in Tasks_Dependencies_Ptr)
      return            Boolean
   is
      type Tasks_Table_Range is new Integer range
         0 .. Framework_Config.Max_Tasks_Dependencies;
      type Tasks_Table is array (Tasks_Table_Range'Range) of Generic_Task_Ptr;

      My_Iterator1 : Tasks_Iterator;
      A_Task       : Generic_Task_Ptr;

      -- Stay true until we find a cycle
      --
      Result : Boolean := True;

      --  Check backward arrows (cycle detection)
      --
      function Has_Backward_Arrow
        (Current, Next_Of_Current : Generic_Task_Ptr;
         Number_Of_Arrows         : Tasks_Table_Range;
         From, To                 : Tasks_Table)
         return                     Boolean
      is
      begin
         for K in  0 .. Number_Of_Arrows - 1 loop
            if (Current.Name = From (K).Name) and
               (Next_Of_Current.Name = To (K).Name)
            then
               return True;
            end if;
         end loop;
         return False;
      end Has_Backward_Arrow;

      -- Each task of the set has to be inspected
      --
      procedure Scan_A_Task
        (Current          : Generic_Task_Ptr;
         Number_Of_Arrows : in Tasks_Table_Range;
         From, To         : in Tasks_Table)
      is

         Next            : Tasks_Set;
         Ite             : Tasks_Iterator;
         Next_Of_Current : Generic_Task_Ptr;

         Tmp_To : Tasks_Table := To;

         Tmp_From   : Tasks_Table                := From;
         Tmp_Number : constant Tasks_Table_Range := Number_Of_Arrows + 1;

      begin

         if Has_Successor (My_Dependencies, Current) then

            Reset (Next, False);
            Next := Get_Successors_List (My_Dependencies, Current);

            -- Run for each task a cyclicity test
            --
            Reset_Iterator (Next, Ite);
            for J in  0 .. Get_Number_Of_Elements (Next) - 1 loop
               Current_Element (Next, Next_Of_Current, Ite);

               -- Memorize that we had check this arrow
               --
               if Has_Backward_Arrow
                    (Current,
                     Next_Of_Current,
                     Number_Of_Arrows,
                     From,
                     To)
               then
                  Result := False;
               else
                  Tmp_From (Tmp_Number - 1) := Current;
                  Tmp_To (Tmp_Number - 1)   := Next_Of_Current;
                  Scan_A_Task (Next_Of_Current, Tmp_Number, Tmp_From, Tmp_To);
               end if;
               Next_Element (Next, Ite);
            end loop;
         end if;

      end Scan_A_Task;

      -- Used to store backward arrows
      --
      From, To : Tasks_Table;
      M        : Tasks_Table_Range := 0;

   begin

      -- Run for Each task A Cyclicity Test
      --
      if not Is_Empty (My_Dependencies.Dependent_Tasks) then

         Reset_Iterator (My_Dependencies.Dependent_Tasks, My_Iterator1);

         loop
            Current_Element
              (My_Dependencies.Dependent_Tasks,
               A_Task,
               My_Iterator1);

            if Result then
               M := 0;
               Scan_A_Task (A_Task, M, From, To);
            end if;

            exit when Is_Last_Element
                         (My_Dependencies.Dependent_Tasks,
                          My_Iterator1);

            Next_Element (My_Dependencies.Dependent_Tasks, My_Iterator1);

         end loop;
      end if;

      return Result;

   end Is_Acyclic;

end Task_Dependencies;
