------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Framework_Config;
USE Framework_Config;
WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Buffers;
USE Buffers;
use Buffers.Buffer_Roles;
WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;
WITH Messages;
USE Messages;
WITH Message_Set;
USE Message_Set;
USE Message_Set.Generic_Message_Set;
WITH Dependencies;
USE Dependencies;
WITH Sets;




PACKAGE Task_Dependencies IS



   -- Raised when a dependency accessor do not find the
   -- required dependency
   --
   Dependency_Not_Found : EXCEPTION;





   --------------------------------------------
   -- A set of half dependencies
   --------------------------------------------
   PACKAGE Half_Dep_Set IS
      NEW Sets (
      Max_Element => Framework_Config.Max_Tasks_Dependencies,
      Element     => Dependency_Ptr,
      Free        => Free,
      Copy        => Copy,
      Put         => Put);
   USE Half_Dep_Set;



   TYPE Tasks_Dependencies IS NEW Half_Dep_Set.Set WITH PRIVATE;

   TYPE Tasks_Dependencies_Ptr IS ACCESS ALL Tasks_Dependencies'Class;

   SUBTYPE Tasks_Dependencies_Range IS Half_Dep_Set.Element_Range;
   SUBTYPE Tasks_Dependencies_Iterator IS Half_Dep_Set.Iterator;



   ----------------------------------------------------------
   --  Basic services
   ----------------------------------------------------------



   PROCEDURE Free (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr);


   FUNCTION Export_Xml (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Unbounded_String;


   FUNCTION Export_Aadl_Properties (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         Number_Of_Ht    : IN     Natural)
     RETURN Unbounded_String;


   PROCEDURE Reset (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         Free_Object     : IN     Boolean                := False);


   PROCEDURE Duplicate (
         Src  : IN     Tasks_Dependencies_Ptr;
         Dest : IN OUT Tasks_Dependencies_Ptr);


   PROCEDURE Put (
         My_Dependencies : IN     Tasks_Dependencies_Ptr);




   -----------------------------------------------------------------------
   --  Services below allow us to test
   --  existing dependencies.
   -----------------------------------------------------------------------


   -- Sub programs to test precedency constraints : dependency between two tasks
   --

   FUNCTION Get_A_Successor (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Generic_Task_Ptr)
     RETURN Generic_Task_Ptr;

   FUNCTION Get_A_Predecessor (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Generic_Task_Ptr)
     RETURN Generic_Task_Ptr;

   FUNCTION Get_Successors_List (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Generic_Task_Ptr)
     RETURN Tasks_Set;

   FUNCTION Get_Predecessors_List (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Generic_Task_Ptr)
     RETURN Tasks_Set;

   FUNCTION Has_Predecessor (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Generic_Task_Ptr)
     RETURN Boolean;

   FUNCTION Has_Successor (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Generic_Task_Ptr)
     RETURN Boolean;


   -- Get a task without predecessor
   --
   FUNCTION Get_A_Root_Task (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Generic_Task_Ptr;


   -- Get a task without successor
   --
   FUNCTION Get_A_Leaf_Task (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Generic_Task_Ptr;


   -- Get all tasks without predecessor
   --
   FUNCTION Get_Root_Tasks (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Tasks_Set;


   -- Get all tasks without successor
   --
   FUNCTION Get_Leaf_Tasks (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Tasks_Set;


   -- Check is a task dependencies is acyclic or not
   --
   FUNCTION Is_Acyclic (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Boolean;

   FUNCTION Is_Cyclic (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Boolean;


   ----------------------------------------------------------------------
   --  CURRENTLY, IN THIS PACKAGE, THERE IS NO SUB PROGRAM TO TEST
   --    DEPENDENCIES WITH MESSAGES AND BUFFERS
   --
   --  IN THE FUTURE, SOME SUB-PROGRAMS AS BELOW SHOULD BE ADDED
   --   (ex : Has_Predecessor,  Get_A_Successor  Get_A_Predecessor
   --       Get_Successors_List  Get_Predecessors_List Has_Predecessor
   --       Has_Successor  Get_A_Root_Task  Get_A_Leaf_Task
   --       Get_Root_Tasks Get_Leaf_Tasks
   --
   --        but with "Message_Ptr" and "Buffer_Ptr" instead of "Task_Ptr"
   --        return values to test such properties
   ----------------------------------------------------------------------


   -- Sub programs to test buffer constraints : dependency between a task
   -- and a buffer
   --


   FUNCTION Has_Buffer_To_Read (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Generic_Task_Ptr)
     RETURN Boolean;

   FUNCTION Has_Buffer_To_Write (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Generic_Task_Ptr)
     RETURN Boolean;







   ----------------------------------------------------------------------
   --  Services below allow you to delete
   --  dependencies.
   ----------------------------------------------------------------------



   -- Delete ALL dependencies  of a
   -- processor (parameter "A_Processor")
   --
   PROCEDURE Delete_Processor_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Processor     : IN     Unbounded_String);


   -- Delete ALL dependencies  of an
   -- address space (parameter "An_Address_Space")
   --
   PROCEDURE Delete_Address_Space_All_Task_Dependencies (
         My_Dependencies  : IN OUT Tasks_Dependencies_Ptr;
         An_Address_Space : IN     Unbounded_String);


   -- Delete ALL dependencies :
   --  1) of a task (if My_Dep_Name is a Task_ptr)
   --  2) of a message  (if My_Dep_Name is a Message_ptr)
   --  3) of a buffer  (if My_Dep_Name is a Buffer_ptr)
   --
   PROCEDURE Delete_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Generic_Task_Ptr);
   PROCEDURE Delete_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Buffer        : IN     Buffer_Ptr);
   PROCEDURE Delete_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Message       : IN     Message_Ptr);



   -- Remove ONE dependncy between two tasks
   -- (remove a precedency relationship, not with bufer of message)
   --
   PROCEDURE Delete_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         From_Task       : IN     Generic_Task_Ptr;
         To_Task         : IN     Generic_Task_Ptr);



   -- Remove ONE dependency between a task and
   --   a message or a buffer
   --
   PROCEDURE Delete_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Generic_Task_Ptr;
         A_Buffer        : IN     Buffer_Ptr;
         A_Type          : IN     Orientation_Dependency_Type);
   PROCEDURE Delete_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Generic_Task_Ptr;
         A_Message       : IN     Message_Ptr;
         A_Type          : IN     Orientation_Dependency_Type);


   ---------------------------------------
   --  Services below allow to update
   --  dependencies.
   ---------------------------------------



   -- Update ALL dependencies :
   --  1) of a task (if My_Dep_Name is a Task_ptr)
   --  2) of a message  (if My_Dep_Name is a Message_ptr)
   --  3) of a buffer  (if My_Dep_Name is a Buffer_ptr)
   --
   PROCEDURE Update_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Generic_Task_Ptr);
   PROCEDURE Update_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Buffer        : IN     Buffer_Ptr);
   PROCEDURE Update_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Message       : IN     Message_Ptr);

   ---------------------------------------
   --  Services below allow to add
   --  dependencies.
   ---------------------------------------




   -- Add ALL dependencies taken from the definition
   -- of a given buffer
   --
   PROCEDURE Add_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         My_Tasks        : IN     Tasks_Set;
         A_Buffer        : IN     Buffer_Ptr);




   -- Add a dependency between a task and
   --   a message or a buffer
   --
   PROCEDURE Add_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Generic_Task_Ptr;
         A_Dep           : IN     Message_Ptr;
         A_Type          : IN     Orientation_Dependency_Type);

   PROCEDURE Add_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Generic_Task_Ptr;
         A_Dep           : IN     Buffer_Ptr;
         A_Type          : IN     Orientation_Dependency_Type);



   -- Add a precedency constraint between two tasks
   -- (not a dependency with message or buffer)
   --
   PROCEDURE Add_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         From_Task       : IN     Generic_Task_Ptr;
         To_Task         : IN     Generic_Task_Ptr);




PRIVATE

   TYPE Tasks_Dependencies IS NEW Half_Dep_Set.Set WITH
      RECORD
         Dependent_Tasks : Tasks_Set;
      END RECORD;


END Task_Dependencies;




