------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Exceptions;                    use Ada.Exceptions;
with Resources;                         use Resources;
use Resources.Task_Lists;
with Time_Units;                        use Time_Units;
use Time_Units.A_Time_Unit;
with Natural_Util;                      use Natural_Util;
with Double_Util;                       use Double_Util;
with Integer_Util;                      use Integer_Util;
with Ada.Numerics.Elementary_Functions; use Ada.Numerics.Elementary_Functions;
with Objects;                           use Objects;
with Objects.extended;                  use Objects.extended;
with Translate;                         use Translate;
with Big_Number;

package body Task_Set is

   procedure Check_Task_Critical_Section
     (My_Tasks          : in Tasks_Set;
      Resource_Name     : in Unbounded_String;
      Resource_Cpu_Name : in Unbounded_String;
      Task_Name         : in Unbounded_String;
      Share_Begin       : in Integer;
      Share_End         : in Integer)
   is

      A_Task : Generic_Task_Ptr;

   begin

      A_Task := Search_Task (My_Tasks, Task_Name);

      --
      -- Check if begin and end time are consistent
      --
      if (Share_Begin <= 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Resource_Name &
                " ; " &
                Task_Name &
                " ; " &
                Lb_Task_Begin (Current_Language) &
                Lb_Must_Be (Current_Language) &
                To_Unbounded_String (" >0 ")));
      end if;

      if (Share_End <= 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Resource_Name &
                " ; " &
                Task_Name &
                " ; " &
                Lb_Task_End (Current_Language) &
                Lb_Must_Be (Current_Language) &
                To_Unbounded_String (" >0 ")));
      end if;

      if Share_Begin > Share_End then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Resource_Name &
                " ; " &
                Task_Name &
                " ; " &
                Lb_Task_Begin (Current_Language) &
                Lb_Must_Be (Current_Language) &
                To_Unbounded_String (" <= ") &
                Lb_Task_End (Current_Language)));
      end if;

      -- Check that the critical section is consistent    according to
      -- the task definition
      --
      if (Natural (Share_End) > A_Task.Capacity) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Resource_Name &
                " ; " &
                Task_Name &
                " ; " &
                Lb_Task_End (Current_Language) &
                Lb_Must_Be (Current_Language) &
                To_Unbounded_String (" <= ") &
                Lb_Capacity (Current_Language)));
      end if;

      if (Natural (Share_Begin) > A_Task.Capacity) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Resource_Name &
                " ; " &
                Task_Name &
                " ; " &
                Lb_Task_Begin (Current_Language) &
                Lb_Must_Be (Current_Language) &
                To_Unbounded_String (" <= ") &
                Lb_Capacity (Current_Language)));
      end if;

      -- Check that the task is on the same processor that the
      -- resource
      --
      if A_Task.Cpu_Name /= Resource_Cpu_Name then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Resource_Name &
                " ; " &
                Task_Name &
                " ; " &
                Lb_Have_To_Be_On_The_Same_Processor (Current_Language)));
      end if;

   end Check_Task_Critical_Section;

   procedure Check_Buffer_Roles
     (My_Tasks    : in Tasks_Set;
      Buffer_Name : in Unbounded_String;
      Task_Name   : in Unbounded_String;
      Time        : in Integer;
      Size        : in Integer)
   is

      A_Task : Generic_Task_Ptr;

   begin

      if (Size <= 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Buffer_Name &
                " ; " &
                Task_Name &
                " ; " &
                Lb_Size (Current_Language) &
                Lb_Must_Be (Current_Language) &
                To_Unbounded_String (" >0 ")));
      end if;

      if (Time <= 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Buffer_Name &
                " ; " &
                Task_Name &
                " ; " &
                Lb_Time (Current_Language) &
                Lb_Must_Be (Current_Language) &
                To_Unbounded_String (" >0 ")));
      end if;

      -- Check that capacity is <= time
      --
      A_Task := Search_Task (My_Tasks, Task_Name);
      if A_Task.Capacity < Time then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Buffer_Name &
                " ; " &
                Task_Name &
                " ; " &
                Lb_Time (Current_Language) &
                Lb_Must_Be (Current_Language) &
                To_Unbounded_String (" <= ") &
                Lb_Capacity (Current_Language)));
      end if;

   end Check_Buffer_Roles;

   procedure Add_Task
     (My_Tasks             : in out Tasks_Set;
      Name                 : in Unbounded_String;
      Cpu_Name             : in Unbounded_String;
      Address_Space_Name   : in Unbounded_String;
      Task_Type            : in Tasks_Type;
      Start_Time           : in Integer;
      Capacity             : in Integer;
      Period               : in Integer;
      Deadline             : in Integer;
      Jitter               : in Integer;
      Blocking_Time        : in Integer;
      Priority             : in Integer;
      Criticality          : in Integer;
      Policy               : in Policies;
      Offset               : Offsets_Table;
      Stack_Memory_Size    : in Integer                       := 0;
      Text_Memory_Size     : in Integer                       := 0;
      Param                : in User_Defined_Parameters_Table :=
        No_User_Defined_Parameter;
      Parametric_Rule_Name : in Unbounded_String              := Empty_String;
      Seed_Value           : in Integer                       := 0;
      Predictable          : in Boolean                       := True)
   is

      Dummy : Generic_Task_Ptr;

   begin
      Add_Task
        (My_Tasks,
         Dummy,
         Name,
         Cpu_Name,
         Address_Space_Name,
         Task_Type,
         Start_Time,
         Capacity,
         Period,
         Deadline,
         Jitter,
         Blocking_Time,
         Priority,
         Criticality,
         Policy,
         Offset,
         Stack_Memory_Size,
         Text_Memory_Size,
         Param,
         Parametric_Rule_Name,
         Seed_Value,
         Predictable);
   end Add_Task;

   procedure Check_Task
     (My_Tasks             : in Tasks_Set;
      Name                 : in Unbounded_String;
      Cpu_Name             : in Unbounded_String;
      Address_Space_Name   : in Unbounded_String;
      Task_Type            : in Tasks_Type;
      Start_Time           : in Integer;
      Capacity             : in Integer;
      Period               : in Integer;
      Deadline             : in Integer;
      Jitter               : in Integer;
      Blocking_Time        : in Integer;
      Priority             : in Integer;
      Criticality          : in Integer;
      Policy               : in Policies;
      Offset               : Offsets_Table;
      Stack_Memory_Size    : in Integer                       := 0;
      Text_Memory_Size     : in Integer                       := 0;
      Param                : in User_Defined_Parameters_Table :=
        No_User_Defined_Parameter;
      Parametric_Rule_Name : in Unbounded_String              := Empty_String;
      Seed_Value           : in Integer                       := 0;
      Predictable          : in Boolean                       := True)
   is

   begin

      if Name = "" then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

      if not Is_A_Valid_Identifier (Name) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Task_Name (Current_Language) &
                Lb_Colon &
                Lb_Invalid_Identifier (Current_Language)));
      end if;

      if not Is_A_Valid_Identifier (Parametric_Rule_Name) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Activation_Rule (Current_Language) &
                Lb_Colon &
                Lb_Invalid_Identifier (Current_Language)));
      end if;

      if (Cpu_Name = "") then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Processor_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

      if (Address_Space_Name = "") then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Address_Space_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

      case Task_Type is
         when Periodic_Type =>
            if Period <= 0 then
               Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_Task (Current_Language) &
                      " " &
                      Name &
                      " : " &
                      Lb_Periodic (Current_Language) &
                      Lb_Require_Period (Current_Language)));
            end if;

         when Aperiodic_Type =>
            if Period /= 0 then
               Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_Task (Current_Language) &
                      " " &
                      Name &
                      " : " &
                      Lb_Aperiodic (Current_Language) &
                      Lb_No_Period (Current_Language)));
            end if;

         when Poisson_Type =>
            if Period <= 0 then
               Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_Task (Current_Language) &
                      " " &
                      Name &
                      " : " &
                      Lb_Poisson_Process_Task (Current_Language) &
                      Lb_Require_Period (Current_Language)));
            end if;

         when Sporadic_Type =>
            if Period <= 0 then
               Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_Task (Current_Language) &
                      " " &
                      Name &
                      " : " &
                      Lb_Sporadic_Task (Current_Language) &
                      Lb_Require_Period (Current_Language)));
            end if;

         when Parametric_Type =>
            if Parametric_Rule_Name = Empty_String then
               Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_Task (Current_Language) &
                      " " &
                      Name &
                      " : " &
                      Lb_Parametric_Task (Current_Language) &
                      Lb_Require_Activation_Rule (Current_Language)));
            end if;

      end case;

      if (Capacity <= 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Capacity (Current_Language) &
                Lb_Must_Be (Current_Language) &
                ">0"));
      end if;

      if (Jitter < 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Jitter (Current_Language) &
                Lb_Must_Be (Current_Language) &
                "=>0"));
      end if;

      if (Criticality < 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Criticality (Current_Language) &
                Lb_Must_Be (Current_Language) &
                "=>0"));
      end if;

      if (Period < 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Period (Current_Language) &
                Lb_Must_Be (Current_Language) &
                "=>0"));
      end if;

      if (Deadline < 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Deadline (Current_Language) &
                Lb_Must_Be (Current_Language) &
                ">=0"));
      end if;

      if (Deadline < Jitter) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Deadline (Current_Language) &
                Lb_Must_Be (Current_Language) &
                Lb_Superior (Current_Language) &
                Lb_Jitter (Current_Language)));
      end if;

      if (Start_Time < 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Start_Time (Current_Language) &
                Lb_Must_Be (Current_Language) &
                ">=0"));
      end if;

      if (Blocking_Time < 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Blocking_Time (Current_Language) &
                Lb_Must_Be (Current_Language) &
                ">=0"));
      end if;

      if (Text_Memory_Size < 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Text_Memory_Size (Current_Language) &
                Lb_Must_Be (Current_Language) &
                ">=0"));
      end if;

      if (Stack_Memory_Size < 0) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Stack_Memory_Size (Current_Language) &
                Lb_Must_Be (Current_Language) &
                ">=0"));
      end if;

      if (Priority < Integer (Priority_Range'First)) or
         (Priority > Integer (Priority_Range'Last))
      then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Invalid_Priority (Current_Language)));
      end if;

      if (Priority /= 0) and (Policy = Sched_Others) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Policy_Control (Current_Language)));
      end if;

      if (Priority = 0) and (Policy /= Sched_Others) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Policy_Control (Current_Language)));
      end if;

      if (To_String (Parametric_Rule_Name) /= "" and
          Task_Type /= Parametric_Type)
      then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_Task (Current_Language) &
                " " &
                Name &
                " : " &
                Lb_Rule_For_Parametric_Only (Current_Language)));
      end if;

      -- Offset integrity checks
      --

      for I in  0 .. Offset.Nb_Entries - 1 loop

         if Offset.Entries (I).Offset_Value < 0 then
            Raise_Exception
              (Invalid_Parameter'Identity,
               To_String
                  (Lb_Task (Current_Language) &
                   " " &
                   Name &
                   " : " &
                   To_Unbounded_String ("Offset/val.") &
                   Lb_Must_Be (Current_Language) &
                   To_Unbounded_String (" >= 0")));
         end if;

         if Offset.Entries (I).Activation < 0 then
            Raise_Exception
              (Invalid_Parameter'Identity,
               To_String
                  (Lb_Task (Current_Language) &
                   " " &
                   Name &
                   " : " &
                   To_Unbounded_String ("Offset/activation") &
                   Lb_Must_Be (Current_Language) &
                   To_Unbounded_String (" >= 0")));

         end if;

      end loop;

   end Check_Task;

   procedure Add_Task
     (My_Tasks             : in out Tasks_Set;
      A_Task               : in out Generic_Task_Ptr;
      Name                 : in Unbounded_String;
      Cpu_Name             : in Unbounded_String;
      Address_Space_Name   : in Unbounded_String;
      Task_Type            : in Tasks_Type;
      Start_Time           : in Integer;
      Capacity             : in Integer;
      Period               : in Integer;
      Deadline             : in Integer;
      Jitter               : in Integer;
      Blocking_Time        : in Integer;
      Priority             : in Integer;
      Criticality          : in Integer;
      Policy               : in Policies;
      Offset               : Offsets_Table;
      Stack_Memory_Size    : in Integer                       := 0;
      Text_Memory_Size     : in Integer                       := 0;
      Param                : in User_Defined_Parameters_Table :=
        No_User_Defined_Parameter;
      Parametric_Rule_Name : in Unbounded_String              := Empty_String;
      Seed_Value           : in Integer                       := 0;
      Predictable          : in Boolean                       := True)
   is

      New_Periodic_Task   : Periodic_Task_Ptr;
      New_Aperiodic_Task  : Aperiodic_Task_Ptr;
      New_Poisson_Task    : Poisson_Task_Ptr;
      New_Parametric_Task : Parametric_Task_Ptr;
      New_Sporadic_Task   : Sporadic_Task_Ptr;

      My_Iterator : Iterator;

   begin

      Check_Task
        (My_Tasks,
         Name,
         Cpu_Name,
         Address_Space_Name,
         Task_Type,
         Start_Time,
         Capacity,
         Period,
         Deadline,
         Jitter,
         Blocking_Time,
         Priority,
         Criticality,
         Policy,
         Offset,
         Stack_Memory_Size,
         Text_Memory_Size,
         Param,
         Parametric_Rule_Name,
         Seed_Value,
         Predictable);

      if (Get_Number_Of_Elements (My_Tasks) > 0) then

         Reset_Iterator (My_Tasks, My_Iterator);

         loop
            Current_Element (My_Tasks, A_Task, My_Iterator);
            if (Name = A_Task.Name) then
               Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_Task (Current_Language) &
                      " " &
                      Name &
                      " : " &
                      Lb_Task_Name (Current_Language) &
                      Lb_Already_Defined (Current_Language)));
            end if;
            exit when Is_Last_Element (My_Tasks, My_Iterator);
            Next_Element (My_Tasks, My_Iterator);
         end loop;
      end if;

      case Task_Type is
         when Periodic_Type =>
            New_Periodic_Task        := new Periodic_Task;
            New_Periodic_Task.Jitter := Natural (Jitter);
            New_Periodic_Task.Period := Natural (Period);
            A_Task                   := Generic_Task_Ptr (New_Periodic_Task);
         when Aperiodic_Type =>
            New_Aperiodic_Task := new Aperiodic_Task;
            A_Task             := Generic_Task_Ptr (New_Aperiodic_Task);
         when Poisson_Type =>
            New_Poisson_Task             := new Poisson_Task;
            New_Poisson_Task.Jitter      := Natural (Jitter);
            New_Poisson_Task.Period      := Natural (Period);
            New_Poisson_Task.Seed        := Seed_Value;
            New_Poisson_Task.Predictable := Predictable;
            A_Task                       :=
              Generic_Task_Ptr (New_Poisson_Task);
         when Sporadic_Type =>
            New_Sporadic_Task             := new Sporadic_Task;
            New_Sporadic_Task.Jitter      := Natural (Jitter);
            New_Sporadic_Task.Period      := Natural (Period);
            New_Sporadic_Task.Seed        := Seed_Value;
            New_Sporadic_Task.Predictable := Predictable;
            A_Task                        :=
              Generic_Task_Ptr (New_Sporadic_Task);
         when Parametric_Type =>
            New_Parametric_Task                 := new Parametric_Task;
            New_Parametric_Task.Jitter          := Natural (Jitter);
            New_Parametric_Task.Period          := Natural (Period);
            New_Parametric_Task.Seed            := Seed_Value;
            New_Parametric_Task.Predictable     := Predictable;
            New_Parametric_Task.Activation_Rule := Parametric_Rule_Name;
            A_Task                              :=
              Generic_Task_Ptr (New_Parametric_Task);
      end case;

      A_Task.Offset             := Offset;
      A_Task.Deadline           := Natural (Deadline);
      A_Task.Name               := Name;
      A_Task.Capacity           := Natural (Capacity);
      A_Task.Start_Time         := Natural (Start_Time);
      A_Task.Cpu_Name           := Cpu_Name;
      A_Task.Address_Space_Name := Address_Space_Name;
      A_Task.Priority           := Priority_Range (Priority);
      A_Task.Criticality        := Natural (Criticality);
      A_Task.Policy             := Policy;
      A_Task.Task_Type          := Task_Type;
      A_Task.Blocking_Time      := Natural (Blocking_Time);
      A_Task.Stack_Memory_Size  := Natural (Stack_Memory_Size);
      A_Task.Text_Memory_Size   := Natural (Text_Memory_Size);
      A_Task.Param              := Param;

      Add (My_Tasks, A_Task);

   exception
      when Full_Set =>
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String (Lb_Can_Not_Define_More_Tasks (Current_Language)));

   end Add_Task;

   function Task_Is_Present
     (My_Tasks : in Tasks_Set;
      Name     : in Unbounded_String)
      return     Boolean
   is
      My_Iterator : Iterator;
      A_Task      : Generic_Task_Ptr;

      Found : Boolean := False;

   begin

      if Is_Empty (My_Tasks) then
         return False;
      else
         Reset_Iterator (My_Tasks, My_Iterator);

         loop
            Current_Element (My_Tasks, A_Task, My_Iterator);

            if (A_Task.Name = Name) then
               Found := True;
            end if;

            exit when Is_Last_Element (My_Tasks, My_Iterator);

            Next_Element (My_Tasks, My_Iterator);

         end loop;

         return Found;
      end if;

   end Task_Is_Present;

   function Search_Task
     (My_Tasks : in Tasks_Set;
      Name     : in Unbounded_String)
      return     Generic_Task_Ptr
   is
      My_Iterator : Iterator;
      A_Task      : Generic_Task_Ptr;
      Result      : Generic_Task_Ptr;

      Found : Boolean := False;

   begin

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Name = Name) then
            Found  := True;
            Result := A_Task;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

      if not Found then
         Raise_Exception
           (Task_Not_Found'Identity,
            To_String (Lb_Task_Name (Current_Language) & "=" & Name));
      end if;

      return Result;
   end Search_Task;

   function Increasing_Si
     (Op1  : in Generic_Task_Ptr;
      Op2  : in Generic_Task_Ptr)
      return Boolean
   is
      Si1, Si2 : Float;
   begin
      Si1 := Log (Float (Periodic_Task_Ptr (Op1).Period), 2.0) -
             Float'Floor (Log (Float (Periodic_Task_Ptr (Op1).Period), 2.0));
      Si2 := Float (Log (Float (Periodic_Task_Ptr (Op2).Period), 2.0)) -
             Float'Floor
                (Float (Log (Float (Periodic_Task_Ptr (Op2).Period), 2.0)));
      return (Si1 <= Si2);
   end Increasing_Si;

   function Increasing_Utilization
     (Op1  : in Generic_Task_Ptr;
      Op2  : in Generic_Task_Ptr)
      return Boolean
   is
      U1, U2 : Float;
   begin
      U1 := Float (Op1.Capacity) / Float (Periodic_Task_Ptr (Op1).Period);
      U2 := Float (Op2.Capacity) / Float (Periodic_Task_Ptr (Op2).Period);
      return (U1 <= U2);
   end Increasing_Utilization;

   function Increasing_Priority
     (Op1  : in Generic_Task_Ptr;
      Op2  : in Generic_Task_Ptr)
      return Boolean
   is
   begin
      return (Op1.Priority <= Op2.Priority);
   end Increasing_Priority;

   function Decreasing_Priority
     (Op1  : in Generic_Task_Ptr;
      Op2  : in Generic_Task_Ptr)
      return Boolean
   is
   begin
      return (Op1.Priority >= Op2.Priority);
   end Decreasing_Priority;

   function Increasing_Period
     (Op1  : in Generic_Task_Ptr;
      Op2  : in Generic_Task_Ptr)
      return Boolean
   is
   begin
      case Op1.Task_Type is
         when Periodic_Type =>
            return (Periodic_Task_Ptr (Op1).Period <=
                    Periodic_Task_Ptr (Op2).Period);
         when Poisson_Type =>
            return (Poisson_Task_Ptr (Op1).Period <=
                    Poisson_Task_Ptr (Op2).Period);
         when Sporadic_Type =>
            return (Sporadic_Task_Ptr (Op1).Period <=
                    Sporadic_Task_Ptr (Op2).Period);
         when others =>
            return True;
      end case;

   end Increasing_Period;

   function Decreasing_Period
     (Op1  : in Generic_Task_Ptr;
      Op2  : in Generic_Task_Ptr)
      return Boolean
   is
   begin

      case Op1.Task_Type is
         when Periodic_Type =>
            return (Periodic_Task_Ptr (Op1).Period >=
                    Periodic_Task_Ptr (Op2).Period);
         when Poisson_Type =>
            return (Poisson_Task_Ptr (Op1).Period >=
                    Poisson_Task_Ptr (Op2).Period);
         when Sporadic_Type =>
            return (Sporadic_Task_Ptr (Op1).Period >=
                    Sporadic_Task_Ptr (Op2).Period);
         when others =>
            return True;
      end case;

   end Decreasing_Period;

   function Increasing_Deadline
     (Op1  : in Generic_Task_Ptr;
      Op2  : in Generic_Task_Ptr)
      return Boolean
   is
   begin
      return (Op1.Deadline <= Op2.Deadline);
   end Increasing_Deadline;

   function Decreasing_Deadline
     (Op1  : in Generic_Task_Ptr;
      Op2  : in Generic_Task_Ptr)
      return Boolean
   is
   begin
      return (Op1.Deadline >= Op2.Deadline);
   end Decreasing_Deadline;

   function Increasing_Name
     (Op1  : in Generic_Task_Ptr;
      Op2  : in Generic_Task_Ptr)
      return Boolean
   is
   begin
      return (Op1.Name <= Op2.Name);
   end Increasing_Name;

   function Get
     (My_Tasks   : in Tasks_Set;
      Task_Name  : in Unbounded_String;
      Param_Name : in Task_Parameters)
      return       Unbounded_String
   is
      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Cpu_Name) and
         (Param_Name /= Address_Space_Name)
      then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Name = Task_Name) then
            exit;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

      if (Param_Name = Cpu_Name) then
         return A_Task.Cpu_Name;
      else
         return A_Task.Address_Space_Name;
      end if;

   end Get;

   function Get
     (My_Tasks   : in Tasks_Set;
      Task_Name  : in Unbounded_String;
      Param_Name : in Task_Parameters)
      return       Boolean
   is
      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Predictable) then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Name = Task_Name) then
            exit;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

      if (A_Task.Task_Type /= Poisson_Type) and
         (A_Task.Task_Type /= Parametric_Type)
      then
         raise Invalid_Parameter;
      end if;

      return Poisson_Task_Ptr (A_Task).Predictable;

   end Get;

   function Get
     (My_Tasks   : in Tasks_Set;
      Task_Name  : in Unbounded_String;
      Param_Name : in Task_Parameters)
      return       Priority_Range
   is
      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Priority) then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Name = Task_Name) then
            exit;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

      return A_Task.Priority;

   end Get;

   function Get
     (My_Tasks   : in Tasks_Set;
      Task_Name  : in Unbounded_String;
      Param_Name : in Task_Parameters)
      return       Natural
   is
      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      if ((Param_Name /= Start_Time) and
          (Param_Name /= Blocking_Time) and
          (Param_Name /= Capacity) and
          (Param_Name /= Period) and
          (Param_Name /= Deadline) and
          (Param_Name /= Jitter))
      then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Name = Task_Name) then
            exit;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

      if (Param_Name = Start_Time) then
         return A_Task.Start_Time;
      else
         if (Param_Name = Deadline) then
            return A_Task.Deadline;
         else
            if (Param_Name = Period) then
               return Periodic_Task_Ptr (A_Task).Period;
            else
               if (Param_Name = Jitter) then
                  return Periodic_Task_Ptr (A_Task).Jitter;
               else
                  if (Param_Name = Capacity) then
                     return A_Task.Capacity;
                  else
                     return A_Task.Blocking_Time;
                  end if;
               end if;
            end if;
         end if;
      end if;

   end Get;

   procedure Set
     (My_Tasks    : in out Tasks_Set;
      Task_Name   : in Unbounded_String;
      Param_Name  : in Task_Parameters;
      Param_Value : in Unbounded_String)
   is

      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Cpu_Name) and
         (Param_Name /= Address_Space_Name)
      then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Name = Task_Name) then
            if (Param_Name = Cpu_Name) then
               A_Task.Cpu_Name := Param_Value;
            else
               A_Task.Address_Space_Name := Param_Value;
            end if;
            exit;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

   end Set;

   procedure Set
     (My_Tasks    : in out Tasks_Set;
      Task_Name   : in Unbounded_String;
      Param_Name  : in Task_Parameters;
      Param_Value : in Priority_Range)
   is

      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Priority) then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Name = Task_Name) then
            A_Task.Priority := Param_Value;
            exit;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;
   end Set;

   procedure Set
     (My_Tasks    : in out Tasks_Set;
      Task_Name   : in Unbounded_String;
      Param_Name  : in Task_Parameters;
      Param_Value : in Boolean)
   is

      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      if (Param_Name /= Predictable) then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Name = Task_Name) then
            if (A_Task.Task_Type /= Poisson_Type) and
               (A_Task.Task_Type /= Parametric_Type)
            then
               raise Invalid_Parameter;
            end if;
            Poisson_Task_Ptr (A_Task).Predictable := Param_Value;
            exit;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;
   end Set;

   procedure Set
     (My_Tasks    : in out Tasks_Set;
      Task_Name   : in Unbounded_String;
      Param_Name  : in Task_Parameters;
      Param_Value : in Natural)
   is
      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      if ((Param_Name /= Start_Time) and
          (Param_Name /= Blocking_Time) and
          (Param_Name /= Capacity) and
          (Param_Name /= Period) and
          (Param_Name /= Deadline) and
          (Param_Name /= Jitter))
      then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Name = Task_Name) then
            if (Param_Name = Start_Time) then
               A_Task.Start_Time := Param_Value;
               exit;
            end if;
            if (Param_Name = Deadline) then
               A_Task.Deadline := Param_Value;
            end if;
            if (Param_Name = Period and
                A_Task.Task_Type = Periodic_Type)
            then
               Periodic_Task_Ptr (A_Task).Period := Param_Value;
               exit;
            end if;
            if (Param_Name = Jitter and
                A_Task.Task_Type = Periodic_Type)
            then
               Periodic_Task_Ptr (A_Task).Jitter := Param_Value;
               exit;
            end if;
            if (Param_Name = Period and A_Task.Task_Type = Poisson_Type) then
               Poisson_Task_Ptr (A_Task).Period := Param_Value;
               exit;
            end if;
            if (Param_Name = Jitter and A_Task.Task_Type = Poisson_Type) then
               Poisson_Task_Ptr (A_Task).Jitter := Param_Value;
               exit;
            end if;
            if (Param_Name = Capacity) then
               A_Task.Capacity := Param_Value;
               exit;
            end if;
            if (Param_Name = Blocking_Time) then
               A_Task.Blocking_Time := Param_Value;
               exit;
            end if;
         end if;
         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;
   end Set;

   function Get_Number_Of_Task_From_Processor
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
      return           Tasks_Range
   is
      Number      : Tasks_Range := 0;
      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      if Is_Empty (My_Tasks) then
         return 0;
      end if;

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Cpu_Name = Processor_Name) then
            Number := Number + 1;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

      return Number;
   end Get_Number_Of_Task_From_Processor;

   procedure Periodic_Control
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

      Iterator1 : Tasks_Iterator;
      Task1     : Generic_Task_Ptr;

   begin

      Reset_Iterator (My_Tasks, Iterator1);

      loop
         Current_Element (My_Tasks, Task1, Iterator1);

         if (Task1.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then
            if (Task1.Task_Type /= Periodic_Type) then
               raise Task_Must_Be_Periodic;
            end if;
         end if;

         exit when Is_Last_Element (My_Tasks, Iterator1);
         Next_Element (My_Tasks, Iterator1);

      end loop;

   end Periodic_Control;

   procedure Periodic_Or_Sporadic_Control
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

      Iterator1 : Tasks_Iterator;
      Task1     : Generic_Task_Ptr;

   begin

      Reset_Iterator (My_Tasks, Iterator1);

      loop
         Current_Element (My_Tasks, Task1, Iterator1);

         if (Task1.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then
            if (Task1.Task_Type /= Periodic_Type) and
               (Task1.Task_Type /= Sporadic_Type)
            then
               raise Task_Must_Be_Periodic;
            end if;
         end if;

         exit when Is_Last_Element (My_Tasks, Iterator1);
         Next_Element (My_Tasks, Iterator1);

      end loop;

   end Periodic_Or_Sporadic_Control;

   procedure Periodic_Or_Poisson_Process_Control
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

      Iterator1 : Tasks_Iterator;
      Task1     : Generic_Task_Ptr;

   begin

      Reset_Iterator (My_Tasks, Iterator1);

      loop
         Current_Element (My_Tasks, Task1, Iterator1);

         if (Task1.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then
            if (Task1.Task_Type /= Periodic_Type) and
               (Task1.Task_Type /= Poisson_Type)
            then
               raise Task_Must_Be_Periodic_Or_Poisson_Process;
            end if;
         end if;

         exit when Is_Last_Element (My_Tasks, Iterator1);
         Next_Element (My_Tasks, Iterator1);

      end loop;

   end Periodic_Or_Poisson_Process_Control;

   procedure Priority_Control
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

      Iterator1 : Tasks_Iterator;
      Task1     : Generic_Task_Ptr;
      Iterator2 : Tasks_Iterator;
      Task2     : Generic_Task_Ptr;

   begin
      -- No duplicate priority
      --

      Reset_Iterator (My_Tasks, Iterator1);

      loop
         Current_Element (My_Tasks, Task1, Iterator1);

         if (Task1.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then
            Reset_Iterator (My_Tasks, Iterator2);

            loop
               Current_Element (My_Tasks, Task2, Iterator2);

               if (Task2.Cpu_Name = Task1.Cpu_Name) and
                  (Task1.Name /= Task2.Name) and
                  (Task1.Priority = Task2.Priority)
               then
                  raise Priority_Error;
               end if;

               exit when Is_Last_Element (My_Tasks, Iterator2);
               Next_Element (My_Tasks, Iterator2);

            end loop;

         end if;

         exit when Is_Last_Element (My_Tasks, Iterator1);
         Next_Element (My_Tasks, Iterator1);

      end loop;
   end Priority_Control;

   -- Check start time
   --
   procedure Start_Time_Control
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

      Iterator1 : Tasks_Iterator;
      Task1     : Generic_Task_Ptr;

   begin

      Reset_Iterator (My_Tasks, Iterator1);

      loop
         Current_Element (My_Tasks, Task1, Iterator1);

         if ((Task1.Cpu_Name = Processor_Name) or
             (Processor_Name = Empty_String)) and
            (Task1.Start_Time /= 0)
         then
            raise Start_Time_Error;
         end if;

         exit when Is_Last_Element (My_Tasks, Iterator1);
         Next_Element (My_Tasks, Iterator1);

      end loop;
   end Start_Time_Control;

   -- Check offset
   --
   procedure Offset_Control
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

      Iterator1 : Tasks_Iterator;
      Task1     : Generic_Task_Ptr;

   begin

      Reset_Iterator (My_Tasks, Iterator1);

      loop
         Current_Element (My_Tasks, Task1, Iterator1);

         if (Task1.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then
            for I in  0 .. Task1.Offset.Nb_Entries - 1 loop
               if (Task1.Offset.Entries (I).Activation /= 0) then
                  raise Offset_Error;
               end if;
            end loop;
         end if;

         exit when Is_Last_Element (My_Tasks, Iterator1);
         Next_Element (My_Tasks, Iterator1);

      end loop;
   end Offset_Control;

   procedure Have_Deadlines_Equal_Than_Periods_Control
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

      A_Task      : Generic_Task_Ptr;
      My_Iterator : Iterator;

   begin

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (not ((A_Task.Task_Type = Periodic_Type) or
                  (A_Task.Task_Type = Poisson_Type)) and
             (A_Task.Cpu_Name = Processor_Name))
         then
            raise Task_Must_Be_Periodic;
         end if;

         if (A_Task.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then
            if A_Task.Task_Type = Periodic_Type then
               if (A_Task.Deadline /= Periodic_Task_Ptr (A_Task).Period) then
                  raise Task_Must_Have_Period_Equal_To_Deadline;
               end if;
            end if;

            if A_Task.Task_Type = Poisson_Type then  -- check this
               if (A_Task.Deadline /= Poisson_Task_Ptr (A_Task).Period) then
                  raise Task_Must_Have_Period_Equal_To_Deadline;
               end if;
            end if;
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

   end Have_Deadlines_Equal_Than_Periods_Control;

   procedure Deadline_Control
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

      Iterator1 : Tasks_Iterator;
      Task1     : Generic_Task_Ptr;

   begin

      Reset_Iterator (My_Tasks, Iterator1);

      loop
         Current_Element (My_Tasks, Task1, Iterator1);

         if (Task1.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then
            if Task1.Deadline = 0 then
               raise Deadline_Error;
            end if;
         end if;

         exit when Is_Last_Element (My_Tasks, Iterator1);
         Next_Element (My_Tasks, Iterator1);

      end loop;

   end Deadline_Control;

   procedure Have_Same_Period_Control
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

      Iterator1    : Tasks_Iterator;
      Task1        : Generic_Task_Ptr;
      Period_Value : Natural := 0;
      First        : Boolean := True;

   begin

      Reset_Iterator (My_Tasks, Iterator1);

      loop
         Current_Element (My_Tasks, Task1, Iterator1);

         if not ((Task1.Task_Type = Periodic_Type) or
                 (Task1.Task_Type = Aperiodic_Type))
         then
            raise Task_Model_Error;
         end if;

         if (Task1.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then
            if First then
               First := False;
               if Task1.Task_Type = Periodic_Type then
                  Period_Value := Periodic_Task_Ptr (Task1).Period;
               end if;
            else
               if Task1.Task_Type = Aperiodic_Type and
                  Period_Value /= 0
               then
                  raise Task_Must_Have_The_Same_Period_Value;
               end if;
               if Task1.Task_Type = Periodic_Type and
                  Period_Value /= Periodic_Task_Ptr (Task1).Period
               then
                  raise Task_Must_Have_The_Same_Period_Value;
               end if;
            end if;
         end if;

         exit when Is_Last_Element (My_Tasks, Iterator1);
         Next_Element (My_Tasks, Iterator1);

      end loop;

   end Have_Same_Period_Control;

   function Deadline_Inferior_To_Period
     (My_Tasks : in Tasks_Set)
      return     Boolean
   is
      My_Iterator : Iterator;
      A_Task      : Generic_Task_Ptr;
      Result      : Boolean := True;

   begin
      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if A_Task.Task_Type = Periodic_Type then
            if (Periodic_Task_Ptr (A_Task).Period < A_Task.Deadline) then
               Result := False;
            end if;
         end if;

         if A_Task.Task_Type = Poisson_Type then
            if (Poisson_Task_Ptr (A_Task).Period < A_Task.Deadline) then
               Result := False;
            end if;
         end if;

         -- Next task
         --
         exit when Is_Last_Element (My_Tasks, My_Iterator);
         Next_Element (My_Tasks, My_Iterator);
      end loop;

      return Result;

   end Deadline_Inferior_To_Period;

   function Is_Harmonic
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
      return           Boolean
   is
      My_Iterator1, My_Iterator2 : Iterator;
      A_Task1, A_Task2           : Generic_Task_Ptr;
      Result                     : Boolean := True;
      Tmp_Mod                    : Natural;

   begin
      -- First loop
      --
      Reset_Iterator (My_Tasks, My_Iterator1);

      loop
         Current_Element (My_Tasks, A_Task1, My_Iterator1);

         if (A_Task1.Cpu_Name = Processor_Name) or
            (Processor_Name = Empty_String)
         then

            -- Second loop
            --
            Reset_Iterator (My_Tasks, My_Iterator2);

            loop
               Current_Element (My_Tasks, A_Task2, My_Iterator2);

               if (A_Task2.Cpu_Name = Processor_Name) or
                  (Processor_Name = Empty_String)
               then

                  if A_Task1.Task_Type = Periodic_Type then
                     if (Periodic_Task_Ptr (A_Task1).Period >
                         Periodic_Task_Ptr (A_Task2).Period)
                     then
                        Tmp_Mod := Periodic_Task_Ptr (A_Task1).Period mod
                                   Periodic_Task_Ptr (A_Task2).Period;
                     else
                        Tmp_Mod := Periodic_Task_Ptr (A_Task2).Period mod
                                   Periodic_Task_Ptr (A_Task1).Period;
                     end if;

                     if (A_Task1.Name /= A_Task2.Name and Tmp_Mod /= 0) then
                        Result := False;
                     end if;
                  end if;

                  if A_Task1.Task_Type = Poisson_Type then
                     if (Poisson_Task_Ptr (A_Task1).Period >
                         Poisson_Task_Ptr (A_Task2).Period)
                     then
                        Tmp_Mod := Poisson_Task_Ptr (A_Task1).Period mod
                                   Poisson_Task_Ptr (A_Task2).Period;
                     else
                        Tmp_Mod := Poisson_Task_Ptr (A_Task2).Period mod
                                   Poisson_Task_Ptr (A_Task1).Period;
                     end if;

                     if (A_Task1.Name /= A_Task2.Name and Tmp_Mod /= 0) then
                        Result := False;
                     end if;
                  end if;

               end if;

               -- next task
               --
               exit when Is_Last_Element (My_Tasks, My_Iterator2);
               Next_Element (My_Tasks, My_Iterator2);

            end loop;

         end if;

         -- next task
         --
         exit when Is_Last_Element (My_Tasks, My_Iterator1);
         Next_Element (My_Tasks, My_Iterator1);
      end loop;

      return Result;
   end Is_Harmonic;

   procedure Delete_Address_Space
     (My_Tasks : in out Tasks_Set;
      A_Addr   : in Unbounded_String)
   is

      Tmp         : Tasks_Set;
      A_Task      : Generic_Task_Ptr;
      My_Iterator : Tasks_Iterator;

   begin

      if (Get_Number_Of_Elements (My_Tasks) > 0) then

         Reset_Iterator (My_Tasks, My_Iterator);
         loop
            Current_Element (My_Tasks, A_Task, My_Iterator);

            if (A_Task.Address_Space_Name = A_Addr) then
               Add (Tmp, A_Task);
            end if;

            exit when Is_Last_Element (My_Tasks, My_Iterator);
            Next_Element (My_Tasks, My_Iterator);
         end loop;

         if not Is_Empty (Tmp) then
            Reset_Iterator (Tmp, My_Iterator);
            loop
               Current_Element (Tmp, A_Task, My_Iterator);

               Delete (My_Tasks, A_Task);

               exit when Is_Last_Element (Tmp, My_Iterator);
               Next_Element (Tmp, My_Iterator);
            end loop;

            Reset (Tmp, False);
         end if;

      end if;

   end Delete_Address_Space;

   function Scheduling_Period
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
      return           Double
   is
      Period     : Double  := 1.0;
      Start_Time : Integer := 0;

      A_Task      : Generic_Task_Ptr;
      My_Iterator : Tasks_Iterator;

      pragma Warnings (Off);

      type My_Index is mod 2 ** 16;
      for My_Index'Size use 16;

      package My_Big is new Big_Number (
         My_Index,
         My_Index (Framework_Config.Max_Big_Number_Size));
      use My_Big;
      use My_Big.Conversion;
      use My_Big.Generic_Conversion;
      use My_Big.String_Conversion;
      use My_Big.Unsigned_Number;

      Res, Current : Big_Unsigned;

      function Set is new Int_Number2Big_Unsigned (Integer);

      pragma Warnings (On);

   begin

      -- Check if Tasks Are Periodic
      --
      Periodic_Control (My_Tasks, Processor_Name);

      Res := Set (1);

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Cpu_Name = Processor_Name and
             A_Task.Task_Type = Periodic_Type)
         then
            Start_Time := Natural'Max (Start_Time, A_Task.Start_Time);
            Current    := Set (Periodic_Task_Ptr (A_Task).Period);
            Res        := LCM (Res, Current);
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);
      end loop;

      if Start_Time /= 0 then
         Current := Set (2);
         Res     := Res * Current;
         Current := Set (Start_Time);
         Res     := Res + Current;
      end if;

      Period := Double (Big_Unsigned2Long_Long_Float (Res));

      return Period;

   end Scheduling_Period;

   function Scheduling_Period
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
      return           Natural
   is
      Period     : Natural := 1;
      Start_Time : Natural := 0;

      A_Task      : Generic_Task_Ptr;
      My_Iterator : Tasks_Iterator;

   begin

      -- Check if tasks are periodic
      --
      Periodic_Control (My_Tasks, Processor_Name);

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Cpu_Name = Processor_Name and
             A_Task.Task_Type = Periodic_Type)
         then
            Start_Time := Natural'Max (Start_Time, A_Task.Start_Time);
            Period     :=
               Natural_Util.Lcm (Period, Periodic_Task_Ptr (A_Task).Period);
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);
      end loop;

      if Start_Time /= 0 then
         Period := (Period * 2) + Start_Time;
      end if;

      return Period;

   end Scheduling_Period;

   function Processor_Utilization_Over_Deadline
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
      return           Double
   is
      Utilization : Double := 0.0;

      My_Iterator : Tasks_Iterator;
      A_Task      : Generic_Task_Ptr;

   begin

      -- Check if tasks are periodic
      --
      Periodic_Control (My_Tasks, Processor_Name);

      -- Check if deadline a greater than 0
      --
      Deadline_Control (My_Tasks, Processor_Name);

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Cpu_Name = Processor_Name) then
            Utilization := Utilization +
                           (Double (A_Task.Capacity) /
                            Double (A_Task.Deadline));
         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

      return Utilization;

   end Processor_Utilization_Over_Deadline;

   function Processor_Utilization_Over_Period
     (My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
      return           Double
   is
      Utilization : Double := 0.0;

      My_Iterator : Tasks_Iterator;
      A_Task      : Generic_Task_Ptr;

   begin

      -- Check if tasks are periodics
      --
      Periodic_Control (My_Tasks, Processor_Name);

      Reset_Iterator (My_Tasks, My_Iterator);

      loop
         Current_Element (My_Tasks, A_Task, My_Iterator);

         if (A_Task.Cpu_Name = Processor_Name) then
            if A_Task.Task_Type /= Aperiodic_Type then
               Utilization := Utilization +
                              (Double (A_Task.Capacity) /
                               Double (Periodic_Task_Ptr (A_Task).Period));
            end if;

         end if;

         exit when Is_Last_Element (My_Tasks, My_Iterator);

         Next_Element (My_Tasks, My_Iterator);

      end loop;

      return Utilization;

   end Processor_Utilization_Over_Period;

   procedure Delete_Processor
     (My_Tasks    : in out Tasks_Set;
      A_Processor : in Unbounded_String)
   is

      Tmp         : Tasks_Set;
      A_Task      : Generic_Task_Ptr;
      My_Iterator : Tasks_Iterator;

   begin

      if (Get_Number_Of_Elements (My_Tasks) > 0) then

         Reset_Iterator (My_Tasks, My_Iterator);
         loop
            Current_Element (My_Tasks, A_Task, My_Iterator);

            if (A_Task.Cpu_Name = A_Processor) then
               Add (Tmp, A_Task);
            end if;

            exit when Is_Last_Element (My_Tasks, My_Iterator);
            Next_Element (My_Tasks, My_Iterator);
         end loop;

         if not Is_Empty (Tmp) then
            Reset_Iterator (Tmp, My_Iterator);
            loop
               Current_Element (Tmp, A_Task, My_Iterator);

               Delete (My_Tasks, A_Task);

               exit when Is_Last_Element (Tmp, My_Iterator);
               Next_Element (Tmp, My_Iterator);
            end loop;

            Reset (Tmp, False);
         end if;
      end if;

   end Delete_Processor;

   function Export_Xml (My_Tasks : in Tasks_Set) return Unbounded_String is
      My_Iterator : Tasks_Iterator;
      A_Task      : Generic_Task_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Tasks) then

         Result := Result & To_Unbounded_String ("<tasks>") & Unbounded_Lf;

         Reset_Iterator (My_Tasks, My_Iterator);

         loop
            Current_Element (My_Tasks, A_Task, My_Iterator);

            Result := Result & To_Unbounded_String (ASCII.HT & "<task");
            if A_Task.X /= 0 then
               Result := Result &
                         To_Unbounded_String (" x=""" & A_Task.X'Img & """ ");
            end if;
            if A_Task.Y /= 0 then
               Result := Result &
                         To_Unbounded_String ("y=""" & A_Task.Y'Img & """ ");
            end if;

            case A_Task.Task_Type is
               when Periodic_Type =>
                  Result := Result &
                            To_Unbounded_String
                               (" task_type=""PERIODIC_TYPE"" ");
               when Poisson_Type =>
                  Result := Result &
                            To_Unbounded_String
                               (" task_type=""POISSON_TYPE"" ");
               when Sporadic_Type =>
                  Result := Result &
                            To_Unbounded_String
                               (" task_type=""SPORADIC_TYPE"" ");
               when Parametric_Type =>
                  Result := Result &
                            To_Unbounded_String
                               (" task_type=""PARAMETRIC_TYPE"" ");
               when others =>
                  null;
            end case;
            Result := Result & To_Unbounded_String (">") & Unbounded_Lf;

            if A_Task.Param.Nb_Entries > 0 then
               Result := Result &
                         To_Unbounded_String
                            (ASCII.HT & ASCII.HT & "<parameters>") &
                         Unbounded_Lf;
               for I in  0 .. A_Task.Param.Nb_Entries - 1 loop

                  Result := Result &
                            To_Unbounded_String
                               (ASCII.HT &
                                ASCII.HT &
                                ASCII.HT &
                                "<parameter");
                  if A_Task.Param.Entries (I).ptype = Boolean_Parameter then
                     Result := Result &
                               To_Unbounded_String
                                  (" parameter_type=""boolean""> ");
                     Result := Result &
                               To_Unbounded_String
                                  (" " &
                                   A_Task.Param.Entries (I).Boolean_value'Img &
                                   " ");
                  else
                     if A_Task.Param.Entries (I).ptype =
                        Integer_Parameter
                     then
                        Result := Result &
                                  To_Unbounded_String
                                     (" parameter_type=""integer""> ");
                        Result := Result &
                                  To_Unbounded_String
                                     (" " &
                                      A_Task.Param.Entries (I).Integer_value'
                          Img &
                                      " ");
                     else
                        if A_Task.Param.Entries (I).ptype =
                           Double_Parameter
                        then
                           Result := Result &
                                     To_Unbounded_String
                                        (" parameter_type=""double""> ");
                           Result :=
                             Result &
                             To_Unbounded_String
                                (" " &
                                 To_String
                                    (Format
                                        (A_Task.Param.Entries (I).double_value)
) &
                                 " ");
                        else
                           Result := Result &
                                     To_Unbounded_String
                                        (" parameter_type=""string""> ");
                           Result :=
                             Result &
                             To_Unbounded_String
                                (" " &
                                 To_String
                                    (A_Task.Param.Entries (I).string_value) &
                                 " ");
                        end if;
                     end if;
                  end if;
                  Result := Result &
                            To_Unbounded_String
                               (To_String
                                   (" " & A_Task.Param.Entries (I).Name) &
                                " ");
                  Result := Result &
                            To_Unbounded_String ("</parameter>") &
                            Unbounded_Lf;
               end loop;
               Result := Result &
                         To_Unbounded_String
                            (ASCII.HT & ASCII.HT & "</parameters>") &
                         Unbounded_Lf;
            end if;

            if A_Task.Offset.Nb_Entries > 0 then
               Result := Result &
                         To_Unbounded_String
                            (ASCII.HT & ASCII.HT & "<offsets>") &
                         Unbounded_Lf;
               for I in  0 .. A_Task.Offset.Nb_Entries - 1 loop
                  Result := Result &
                            To_Unbounded_String
                               (ASCII.HT & ASCII.HT & ASCII.HT & "<offset>");
                  Result := Result &
                            To_Unbounded_String
                               (" " &
                                A_Task.Offset.Entries (I).Activation'Img &
                                " ");
                  Result := Result &
                            To_Unbounded_String
                               (" " &
                                A_Task.Offset.Entries (I).Offset_Value'Img &
                                " ");
                  Result := Result &
                            To_Unbounded_String ("</offset>") &
                            Unbounded_Lf;
               end loop;
               Result := Result &
                         To_Unbounded_String
                            (ASCII.HT & ASCII.HT & "</offsets>") &
                         Unbounded_Lf;
            end if;

            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<cpu_name>" &
                  To_String (A_Task.Cpu_Name) &
                  "</cpu_name>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<address_space_name>" &
                  To_String (A_Task.Address_Space_Name) &
                  "</address_space_name>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<name>" &
                  To_String (A_Task.Name) &
                  "</name>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<capacity>" &
                  A_Task.Capacity'Img &
                  "</capacity>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<start_time>" &
                  A_Task.Start_Time'Img &
                  "</start_time>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<policy>" &
                  A_Task.Policy'Img &
                  "</policy>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<deadline>" &
                  A_Task.Deadline'Img &
                  "</deadline>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<criticality>" &
                  A_Task.Criticality'Img &
                  "</criticality>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<blocking_time>" &
                  A_Task.Blocking_Time'Img &
                  "</blocking_time>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<priority>" &
                  A_Task.Priority'Img &
                  "</priority>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<text_memory_size>" &
                  A_Task.Text_Memory_Size'Img &
                  "</text_memory_size>") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "<stack_memory_size>" &
                  A_Task.Stack_Memory_Size'Img &
                  "</stack_memory_size>") &
              Unbounded_Lf;

            if (A_Task.Task_Type /= Aperiodic_Type) then
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "<period>" &
                     Periodic_Task_Ptr (A_Task).Period'Img &
                     "</period>") &
                 Unbounded_Lf;
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "<jitter>" &
                     Periodic_Task_Ptr (A_Task).Jitter'Img &
                     "</jitter>") &
                 Unbounded_Lf;

            end if;

            if (A_Task.Task_Type = Poisson_Type) or
               (A_Task.Task_Type = Sporadic_Type) or
               (A_Task.Task_Type = Parametric_Type)
            then
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "<seed>" &
                     Poisson_Task_Ptr (A_Task).Seed'Img &
                     "</seed>") &
                 Unbounded_Lf;
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "<predictable_seed>" &
                     Poisson_Task_Ptr (A_Task).Predictable'Img &
                     "</predictable_seed>") &
                 Unbounded_Lf;
            end if;

            if (A_Task.Task_Type = Parametric_Type) then
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "<activation_rule>" &
                     To_String (Parametric_Task_Ptr (A_Task).Activation_Rule) &
                     "</activation_rule>") &
                 Unbounded_Lf;
            end if;

            Result := Result &
                      To_Unbounded_String (ASCII.HT & "</task>") &
                      Unbounded_Lf;

            exit when Is_Last_Element (My_Tasks, My_Iterator);

            Next_Element (My_Tasks, My_Iterator);

         end loop;
         Result := Result & To_Unbounded_String ("</tasks>") & Unbounded_Lf;
      end if;

      return Result;

   end Export_Xml;

   function Export_Aadl_Implementations
     (My_Tasks     : in Tasks_Set;
      My_Resources : in Resources_Set)
      return         Unbounded_String
   is
      My_Iterator          : Tasks_Iterator;
      A_Task               : Generic_Task_Ptr;
      My_Resource_Iterator : Resources_Iterator;
      A_Resource           : Generic_Resource_Ptr;

      Print_Feature_Header : Boolean;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Tasks) then

         Reset_Iterator (My_Tasks, My_Iterator);

         loop
            Current_Element (My_Tasks, A_Task, My_Iterator);

            Result               := Result &
                                    To_Unbounded_String
                                       ("thread " & To_String (A_Task.Name)) &
                                    Unbounded_Lf;
            Print_Feature_Header := True;

            -- Look for shared resource access
            --
            if not Is_Empty (My_Resources) then

               Reset_Iterator (My_Resources, My_Resource_Iterator);

               loop
                  Current_Element
                    (My_Resources,
                     A_Resource,
                     My_Resource_Iterator);

                  if A_Resource.task_tab.Nb_Entries /= 0 then
                     for I in  0 .. A_Resource.task_tab.Nb_Entries - 1 loop
                        if A_Resource.task_tab.Entries (I).Item =
                           A_Task.Name
                        then
                           if Print_Feature_Header then
                              Result               := Result &
                                                      To_Unbounded_String
                                                         (ASCII.HT &
                                                          "features") &
                                                      Unbounded_Lf;
                              Print_Feature_Header := False;
                           end if;
                           Result :=
                             Result &
                             To_Unbounded_String
                                (ASCII.HT &
                                 ASCII.HT &
                                 To_String (A_Resource.Name) &
                                 "_features" &
                                 " : requires data access " &
                                 To_String (A_Resource.Name) &
                                 ".Impl;") &
                             Unbounded_Lf;
                           exit;
                        end if;
                     end loop;
                  end if;

                  exit when Is_Last_Element
                               (My_Resources,
                                My_Resource_Iterator);

                  Next_Element (My_Resources, My_Resource_Iterator);

               end loop;
            end if;

            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Task.Name) & ";") &
                      Unbounded_Lf &
                      Unbounded_Lf;

            Result := Result &
                      To_Unbounded_String
                         ("thread implementation " &
                          To_String (A_Task.Name) &
                          ".Impl") &
                      Unbounded_Lf;
            Result := Result &
                      To_Unbounded_String (ASCII.HT & "properties") &
                      Unbounded_Lf;

            case A_Task.Task_Type is
               when Aperiodic_Type =>
                  Result := Result &
                            To_Unbounded_String
                               (ASCII.HT &
                                ASCII.HT &
                                "Dispatch_Protocol => Background;") &
                            Unbounded_Lf;
               when Periodic_Type =>
                  Result := Result &
                            To_Unbounded_String
                               (ASCII.HT &
                                ASCII.HT &
                                "Dispatch_Protocol => Periodic;") &
                            Unbounded_Lf;
               when Poisson_Type =>
                  Result := Result &
                            To_Unbounded_String
                               (ASCII.HT &
                                ASCII.HT &
                                "Dispatch_Protocol => Poisson_Process;") &
                            Unbounded_Lf;
               when Sporadic_Type =>
                  Result := Result &
                            To_Unbounded_String
                               (ASCII.HT &
                                ASCII.HT &
                                "Dispatch_Protocol => Sporadic;") &
                            Unbounded_Lf;
               when Parametric_Type =>
                  Result := Result &
                            To_Unbounded_String
                               (ASCII.HT &
                                ASCII.HT &
                                "Dispatch_Protocol => User_Defined;") &
                            Unbounded_Lf;
                  Result :=
                    Result &
                    To_Unbounded_String
                       (ASCII.HT &
                        ASCII.HT &
                        "Cheddar_Properties::Source_Text => " &
                        """" &
                        To_String
                           (Parametric_Task_Ptr (A_Task).Activation_Rule) &
                        """" &
                        ";") &
                    Unbounded_Lf;
            end case;

            if A_Task.Param.Nb_Entries > 0 then

               for I in  0 .. A_Task.Param.Nb_Entries - 1 loop

                  Result :=
                    Result &
                    To_Unbounded_String
                       (ASCII.HT &
                        ASCII.HT &
                        "User_Defined_Cheddar_Properties::" &
                        To_String (A_Task.Param.Entries (I).Name) &
                        " => ");
                  if A_Task.Param.Entries (I).ptype = Boolean_Parameter then
                     Result := Result &
                               To_Unbounded_String
                                  (A_Task.Param.Entries (I).Boolean_value'Img);
                  else
                     if A_Task.Param.Entries (I).ptype =
                        Integer_Parameter
                     then
                        Result := Result &
                                  To_Unbounded_String
                                     (A_Task.Param.Entries (I).Integer_value'
                          Img);
                     else
                        if A_Task.Param.Entries (I).ptype =
                           Double_Parameter
                        then
                           Result := Result &
                                     To_Unbounded_String
                                        (To_String
                                            (Format
                                                (A_Task.Param.Entries (I).
                             double_value)));
                        else
                           Result :=
                             Result &
                             To_Unbounded_String
                                ("""" &
                                 To_String
                                    (A_Task.Param.Entries (I).string_value) &
                                 """");
                        end if;
                     end if;
                  end if;
                  Result := Result &
                            To_Unbounded_String (";") &
                            Unbounded_Lf;
               end loop;
            end if;

            if A_Task.Offset.Nb_Entries > 0 then
               for I in  0 .. A_Task.Offset.Nb_Entries - 1 loop
                  Result :=
                    Result &
                    To_Unbounded_String
                       (ASCII.HT &
                        ASCII.HT &
                        "Cheddar_Properties::Dispatch_Offset_Value_" &
                        To_String (Integer_Util.Format (Integer (I))) &
                        " => " &
                        A_Task.Offset.Entries (I).Offset_Value'Img &
                        ";");
                  Result :=
                    Result &
                    To_Unbounded_String
                       (ASCII.HT &
                        ASCII.HT &
                        "Cheddar_Properties::Dispatch_Offset_Time_" &
                        To_String (Integer_Util.Format (Integer (I))) &
                        " => " &
                        A_Task.Offset.Entries (I).Activation'Img &
                        " ms ;") &
                    Unbounded_Lf;
               end loop;
            end if;

            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "Source_Code_Size => " &
                  A_Task.Text_Memory_Size'Img &
                  " kb ;") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "Source_Stack_Size => " &
                  A_Task.Stack_Memory_Size'Img &
                  " kb ;") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "Compute_Execution_Time => " &
                  A_Task.Capacity'Img &
                  " ms .. " &
                  A_Task.Capacity'Img &
                  " ms;") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "Cheddar_Properties::Dispatch_Absolute_Time => " &
                  A_Task.Start_Time'Img &
                  " ms ;") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "Cheddar_Properties::POSIX_Scheduling_Policy => " &
                  A_Task.Policy'Img &
                  ";") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "Cheddar_Properties::Fixed_Priority => " &
                  A_Task.Priority'Img &
                  ";") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "Cheddar_Properties::Bound_On_Data_Blocking_Time => " &
                  A_Task.Blocking_Time'Img &
                  " ms ;") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "Deadline  => " &
                  A_Task.Deadline'Img &
                  " ms ;") &
              Unbounded_Lf;
            Result :=
              Result &
              To_Unbounded_String
                 (ASCII.HT &
                  ASCII.HT &
                  "Cheddar_Properties::Criticality  => " &
                  A_Task.Criticality'Img &
                  ";") &
              Unbounded_Lf;

            if (A_Task.Task_Type /= Aperiodic_Type) then
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "Period => " &
                     Periodic_Task_Ptr (A_Task).Period'Img &
                     " ms ;") &
                 Unbounded_Lf;
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "Cheddar_Properties::Dispatch_Jitter => " &
                     Periodic_Task_Ptr (A_Task).Jitter'Img &
                     " ms ;") &
                 Unbounded_Lf;
            end if;

            if (A_Task.Task_Type = Poisson_Type) or
               (A_Task.Task_Type = Sporadic_Type) or
               (A_Task.Task_Type = Parametric_Type)
            then
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "Cheddar_Properties::Dispatch_Seed_Value => " &
                     Poisson_Task_Ptr (A_Task).Seed'Img &
                     ";") &
                 Unbounded_Lf;
               Result :=
                 Result &
                 To_Unbounded_String
                    (ASCII.HT &
                     ASCII.HT &
                     "Cheddar_Properties::Dispatch_Seed_is_Predictable => " &
                     Poisson_Task_Ptr (A_Task).Predictable'Img &
                     ";") &
                 Unbounded_Lf;
            end if;

            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Task.Name) & ".Impl;") &
                      Unbounded_Lf &
                      Unbounded_Lf;

            exit when Is_Last_Element (My_Tasks, My_Iterator);

            Next_Element (My_Tasks, My_Iterator);

         end loop;

      end if;

      return Result;

   end Export_Aadl_Implementations;

   function Export_Aadl_User_Defined_Properties
     (My_Tasks : in Tasks_Set)
      return     Unbounded_String
   is
      My_Iterator : Tasks_Iterator;
      A_Task      : Generic_Task_Ptr;

      Find            : Boolean := False;
      Properties_List : Unbounded_String_List;
      List_Ite        : Unbounded_String_Iterator;
      Str             : Unbounded_String_Ptr;
      use Unbounded_Strings.Unbounded_String_List_Package;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Tasks) then

         Reset_Iterator (My_Tasks, My_Iterator);

         loop
            Current_Element (My_Tasks, A_Task, My_Iterator);

            if A_Task.Param.Nb_Entries > 0 then
               for I in  0 .. A_Task.Param.Nb_Entries - 1 loop

                  -- Check that the property does not exist
                  --
                  Find := False;
                  if not Is_Empty (Properties_List) then

                     Reset_Head_Iterator (Properties_List, List_Ite);
                     loop
                        Current_Element (Properties_List, Str, List_Ite);

                        if Str.all = A_Task.Param.Entries (I).Name then
                           Find := True;
                           exit;
                        end if;
                        if Is_Tail_Element (Properties_List, List_Ite) then
                           exit;
                        end if;
                        Next_Element (Properties_List, List_Ite);
                     end loop;

                  end if;

                  if not Find then
                     Str     := new Unbounded_String;
                     Str.all := A_Task.Param.Entries (I).Name;
                     Add (Properties_List, Str);

                     Result := Result &
                               To_Unbounded_String
                                  (ASCII.HT &
                                   To_String (A_Task.Param.Entries (I).Name) &
                                   " : ");
                     if A_Task.Param.Entries (I).ptype =
                        Boolean_Parameter
                     then
                        Result := Result &
                                  To_Unbounded_String ("aadlboolean") &
                                  Unbounded_Lf;
                     else
                        if A_Task.Param.Entries (I).ptype =
                           Integer_Parameter
                        then
                           Result := Result &
                                     To_Unbounded_String ("aadlinteger") &
                                     Unbounded_Lf;
                        else
                           if A_Task.Param.Entries (I).ptype =
                              Double_Parameter
                           then
                              Result := Result &
                                        To_Unbounded_String ("aadlreal") &
                                        Unbounded_Lf;
                           else
                              Result := Result &
                                        To_Unbounded_String ("aadlstring") &
                                        Unbounded_Lf;
                           end if;
                        end if;
                     end if;

                     Result := Result &
                               To_Unbounded_String
                                  (ASCII.HT &
                                   ASCII.HT &
                                   "applies to (thread, thread group);") &
                               Unbounded_Lf;
                  end if;

               end loop;

            end if;

            exit when Is_Last_Element (My_Tasks, My_Iterator);

            Next_Element (My_Tasks, My_Iterator);

         end loop;

      end if;

      return Result;

   end Export_Aadl_User_Defined_Properties;

   function Export_Aadl_Declarations
     (My_Tasks           : in Tasks_Set;
      Address_Space_Name : in Unbounded_String;
      Number_Of_Ht       : in Natural)
      return               Unbounded_String
   is
      My_Iterator : Tasks_Iterator;
      A_Task      : Generic_Task_Ptr;

      Result : Unbounded_String := Empty_String;

   begin

      if not Is_Empty (My_Tasks) then

         Reset_Iterator (My_Tasks, My_Iterator);

         loop
            Current_Element (My_Tasks, A_Task, My_Iterator);

            if A_Task.Address_Space_Name = Address_Space_Name then
               for I in  1 .. Number_Of_Ht loop
                  Result := Result & ASCII.HT;
               end loop;
               Result :=
                 Result &
                 To_Unbounded_String
                    ("instancied_" &
                     To_String (A_Task.Name) &
                     " : thread " &
                     To_String (A_Task.Name) &
                     ".Impl;") &
                 Unbounded_Lf;
            end if;

            exit when Is_Last_Element (My_Tasks, My_Iterator);

            Next_Element (My_Tasks, My_Iterator);

         end loop;

      end if;

      return Result;

   end Export_Aadl_Declarations;

end Task_Set;
