------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Unbounded_Strings;
USE Unbounded_Strings;
WITH Framework_Config;
USE Framework_Config;
with resource_set;
use resource_set;
WITH Tasks;
USE Tasks;
WITH Parameters;
USE Parameters;
WITH Parameters.extended;
USE Parameters.extended;
USE Parameters.User_Defined_Parameters_Table_Package;
WITH Offsets;
USE Offsets;
USE Offsets.Offsets_Table_package;
WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;
WITH Sets;


PACKAGE Task_Set IS



   ----------------------------------------------------------
   -- Definition of a set
   ----------------------------------------------------------


   PACKAGE Generic_Task_Set IS NEW Sets (
      Max_Element => framework_Config.Max_Tasks,
      Element     => Generic_Task_Ptr,
      Copy        => Copy,
      Free        => Free,
      Put         => Put);


   USE Generic_Task_Set;

   TYPE Tasks_Set IS NEW Generic_Task_Set.Set WITH PRIVATE;


   SUBTYPE Tasks_Range IS Generic_Task_Set.Element_Range;
   SUBTYPE Tasks_Iterator IS   Generic_Task_Set.Iterator;



   ----------------------------------------------------------
   -- Exceptions defined in the set package
   ----------------------------------------------------------


   -- Raised if periodic tasks are expected
   --
   Task_Must_Be_Periodic : EXCEPTION;

   -- Raised if tasks are not periodic and poisson process
   --
   Task_Must_Be_Periodic_Or_Poisson_Process : EXCEPTION;

   -- Raised if a task has a wrong type
   --
   Task_Model_Error : EXCEPTION;

   -- Raised when a given task is not in a task set
   --
   Task_Not_Found : EXCEPTION;

   -- Raised when parameters provided to Add_Task are wrong
   --
   Invalid_Parameter : EXCEPTION;

   -- Raised when a treatement required that the processor
   -- utilization is less than 1 for a given task set
   --
   Processor_Utilization_Exceeded : EXCEPTION;

   -- Tasks must have deadline equals to period
   --
   Task_Must_Have_Period_Equal_To_Deadline : EXCEPTION;

   -- All tasks must have different priority values
   --
   Priority_Error : EXCEPTION;

   -- Start times have to be null
   --
   Start_Time_Error : EXCEPTION;

   -- Offsets have to be null
   --
   Offset_Error : EXCEPTION;

   -- Tasks must have the same period value
   --
   Task_Must_Have_The_Same_Period_Value : EXCEPTION;

   -- Tasks must have deadline > 0
   --
   Deadline_Error : EXCEPTION;



   ----------------------------------------------------------
   -- I/O operations
   ----------------------------------------------------------
   FUNCTION Export_Xml (
         My_Tasks : IN     Tasks_Set)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_Implementations (
         My_Tasks     : IN     Tasks_Set;
         My_Resources : IN     Resources_Set)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_Declarations (
         My_Tasks           : IN     Tasks_Set;
         Address_Space_Name : IN     Unbounded_String;
         Number_Of_Ht       : IN     Natural)
     RETURN Unbounded_String;

   FUNCTION Export_Aadl_User_Defined_Properties (
         My_Tasks : IN     Tasks_Set)
     RETURN Unbounded_String;



   ----------------------------------------------------------
   -- Procedure to proceed modification on a set
   ----------------------------------------------------------


   PROCEDURE Check_Task_Critical_Section (
         My_Tasks          : IN     Tasks_Set;
         Resource_Name     : IN     Unbounded_String;
         Resource_Cpu_Name : IN     Unbounded_String;
         Task_Name         : IN     Unbounded_String;
         Share_Begin       : IN     Integer;
         Share_End         : IN     Integer);


   PROCEDURE Check_Buffer_Roles (
         My_Tasks    : IN     Tasks_Set;
         Buffer_Name : IN     Unbounded_String;
         Task_Name   : IN     Unbounded_String;
         Time        : IN     Integer;
         Size        : IN     Integer);




   PROCEDURE Check_Task (
         My_Tasks             : IN     Tasks_Set;
         Name                 : IN     Unbounded_String;
         Cpu_Name             : IN     Unbounded_String;
         Address_Space_Name   : IN     Unbounded_String;
         Task_Type            : IN     Tasks_Type;
         Start_Time           : IN     Integer;
         Capacity             : IN     Integer;
         Period               : IN     Integer;
         Deadline             : IN     Integer;
         Jitter               : IN     Integer;
         Blocking_Time        : IN     Integer;
         Priority             : IN     Integer;
         Criticality          : IN     Integer;
         Policy               : IN     Policies;
         Offset               : IN     Offsets_Table;
         Stack_Memory_Size    : IN     Integer                       := 0;
         Text_Memory_Size     : IN     Integer                       := 0;
         Param                : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Parametric_Rule_Name : IN     Unbounded_String              := Empty_String;
         Seed_Value           : IN     Integer                       := 0;
         Predictable          : IN     Boolean                       := True);



   PROCEDURE Add_Task (
         My_Tasks             : IN OUT Tasks_Set;
         Name                 : IN     Unbounded_String;
         Cpu_Name             : IN     Unbounded_String;
         Address_Space_Name   : IN     Unbounded_String;
         Task_Type            : IN     Tasks_Type;
         Start_Time           : IN     Integer;
         Capacity             : IN     Integer;
         Period               : IN     Integer;
         Deadline             : IN     Integer;
         Jitter               : IN     Integer;
         Blocking_Time        : IN     Integer;
         Priority             : IN     Integer;
         Criticality          : IN     Integer;
         Policy               : IN     Policies;
         Offset               : IN     Offsets_Table;
         Stack_Memory_Size    : IN     Integer                       := 0;
         Text_Memory_Size     : IN     Integer                       := 0;
         Param                : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Parametric_Rule_Name : IN     Unbounded_String              := Empty_String;
         Seed_Value           : IN     Integer                       := 0;
         Predictable          : IN     Boolean                       := True);


   PROCEDURE Add_Task (
         My_Tasks             : IN OUT Tasks_Set;
         A_Task               : IN OUT Generic_Task_Ptr;
         Name                 : IN     Unbounded_String;
         Cpu_Name             : IN     Unbounded_String;
         Address_Space_Name   : IN     Unbounded_String;
         Task_Type            : IN     Tasks_Type;
         Start_Time           : IN     Integer;
         Capacity             : IN     Integer;
         Period               : IN     Integer;
         Deadline             : IN     Integer;
         Jitter               : IN     Integer;
         Blocking_Time        : IN     Integer;
         Priority             : IN     Integer;
         Criticality          : IN     Integer;
         Policy               : IN     Policies;
         Offset               : IN     Offsets_Table;
         Stack_Memory_Size    : IN     Integer                       := 0;
         Text_Memory_Size     : IN     Integer                       := 0;
         Param                : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Parametric_Rule_Name : IN     Unbounded_String              := Empty_String;
         Seed_Value           : IN     Integer                       := 0;
         Predictable          : IN     Boolean                       := True);


   -- Retrieve task hosted by the address space/processor
   --
   PROCEDURE Delete_Address_Space (
         My_Tasks : IN OUT Tasks_Set;
         A_Addr   : IN     Unbounded_String);

   PROCEDURE Delete_Processor (
         My_Tasks    : IN OUT Tasks_Set;
         A_Processor : IN     Unbounded_String);


   ------------------------------------------------------
   -- Read information from a set
   ------------------------------------------------------


   -- Get the number of tasks hosted by the
   -- processor "Processor_Name"
   --
   FUNCTION Get_Number_Of_Task_From_Processor (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Tasks_Range;


   -- Get the task pointer of the task named "Name"
   --
   FUNCTION Search_Task (
         My_Tasks : IN     Tasks_Set;
         Name     : IN     Unbounded_String)
     RETURN Generic_Task_Ptr;


   -- Check if a given task is stored in the task set
   --
   FUNCTION Task_Is_Present (
         My_Tasks : IN     Tasks_Set;
         Name     : IN     Unbounded_String)
     RETURN Boolean;





   ----------------------------------------------------------
   -- List of parameters the user can set of get directly
   -- throught the set
   ----------------------------------------------------------


   TYPE Task_Parameters IS
         (Cpu_Name,
          Address_Space_Name,
          Capacity,
          Start_Time,
          Priority,
          Period,
          Blocking_Time,
          Deadline,
          Jitter,
          Seed,
          Predictable);


   -- Set the value of a parameter of a given task
   --
   PROCEDURE Set (
         My_Tasks    : IN OUT Tasks_Set;
         Task_Name   : IN     Unbounded_String;
         Param_Name  : IN     Task_Parameters;
         Param_Value : IN     Unbounded_String);


   PROCEDURE Set (
         My_Tasks    : IN OUT Tasks_Set;
         Task_Name   : IN     Unbounded_String;
         Param_Name  : IN     Task_Parameters;
         Param_Value : IN     Priority_Range);


   PROCEDURE Set (
         My_Tasks    : IN OUT Tasks_Set;
         Task_Name   : IN     Unbounded_String;
         Param_Name  : IN     Task_Parameters;
         Param_Value : IN     Natural);


   PROCEDURE Set (
         My_Tasks    : IN OUT Tasks_Set;
         Task_Name   : IN     Unbounded_String;
         Param_Name  : IN     Task_Parameters;
         Param_Value : IN     Boolean);


   -- Get the value of a parameter of a given task
   --
   FUNCTION Get (
         My_Tasks   : IN     Tasks_Set;
         Task_Name  : IN     Unbounded_String;
         Param_Name : IN     Task_Parameters)
     RETURN Unbounded_String;


   FUNCTION Get (
         My_Tasks   : IN     Tasks_Set;
         Task_Name  : IN     Unbounded_String;
         Param_Name : IN     Task_Parameters)
     RETURN Priority_Range;


   FUNCTION Get (
         My_Tasks   : IN     Tasks_Set;
         Task_Name  : IN     Unbounded_String;
         Param_Name : IN     Task_Parameters)
     RETURN Natural;


   FUNCTION Get (
         My_Tasks   : IN     Tasks_Set;
         Task_Name  : IN     Unbounded_String;
         Param_Name : IN     Task_Parameters)
     RETURN Boolean;



   --------------------------------------------------------
   -- Functions to proceed sort on a set
   --------------------------------------------------------



   FUNCTION Increasing_Si (
         Op1 : IN     Generic_Task_Ptr;
         Op2 : IN     Generic_Task_Ptr)
     RETURN Boolean;

   FUNCTION Increasing_Utilization (
         Op1 : IN     Generic_Task_Ptr;
         Op2 : IN     Generic_Task_Ptr)
     RETURN Boolean;

   FUNCTION Increasing_Name (
         Op1 : IN     Generic_Task_Ptr;
         Op2 : IN     Generic_Task_Ptr)
     RETURN Boolean;


   FUNCTION Increasing_Deadline (
         Op1 : IN     Generic_Task_Ptr;
         Op2 : IN     Generic_Task_Ptr)
     RETURN Boolean;


   FUNCTION Decreasing_Deadline (
         Op1 : IN     Generic_Task_Ptr;
         Op2 : IN     Generic_Task_Ptr)
     RETURN Boolean;


   FUNCTION Increasing_Period (
         Op1 : IN     Generic_Task_Ptr;
         Op2 : IN     Generic_Task_Ptr)
     RETURN Boolean;


   FUNCTION Decreasing_Period (
         Op1 : IN     Generic_Task_Ptr;
         Op2 : IN     Generic_Task_Ptr)
     RETURN Boolean;


   FUNCTION Increasing_Priority (
         Op1 : IN     Generic_Task_Ptr;
         Op2 : IN     Generic_Task_Ptr)
     RETURN Boolean;


   FUNCTION Decreasing_Priority (
         Op1 : IN     Generic_Task_Ptr;
         Op2 : IN     Generic_Task_Ptr)
     RETURN Boolean;


   ----------------------------------------------------
   -- Control sub-programs
   ----------------------------------------------------


   -- These functions test and return TRUE if all the
   -- tasks of a task set have a common property
   --
   FUNCTION Deadline_Inferior_To_Period (
         My_Tasks : IN     Tasks_Set)
     RETURN Boolean;


   -- These function tests if all the
   -- tasks of a task set are harmonic
   --
   -- If "Processor_Name" is a null string, the control
   -- is done on all the task of the set. In the other
   -- case, the control is done on tasks located on the
   -- processor "Processor_Name"
   --

   FUNCTION Is_Harmonic (
         My_Tasks : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Boolean;


   -- These subprogram test  if all the
   -- tasks of a task set have a common property
   --
   -- If "Processor_Name" is a null string, the control
   -- is done on all the task of the set. In the other
   -- case, the control is done on tasks located on the
   -- processor "Processor_Name"
   --

   PROCEDURE Have_Deadlines_Equal_Than_Periods_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String);


   PROCEDURE Have_Same_Period_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String);


   PROCEDURE Periodic_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String);


   PROCEDURE Periodic_Or_Sporadic_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String);


   PROCEDURE Periodic_Or_Poisson_Process_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String);


   PROCEDURE Priority_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String);


   PROCEDURE Start_Time_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String);


   PROCEDURE Offset_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String);


   PROCEDURE Deadline_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String);


   ----------------------------------------------------
   -- Compute some performance properties on
   -- a task set
   ----------------------------------------------------

   FUNCTION Scheduling_Period (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Natural;


   FUNCTION Scheduling_Period (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Double;


   FUNCTION Processor_Utilization_Over_Deadline (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Double;


   FUNCTION Processor_Utilization_Over_Period (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Double;






PRIVATE

   TYPE Tasks_Set IS NEW Generic_Task_Set.Set WITH NULL RECORD;






END Task_Set;








