------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Graph_Item_Pkg;                    use Graph_Item_Pkg;
with Gtk.Toolbar;                       use Gtk.Toolbar;
with Gtk.Widget;                        use Gtk.Widget;
with Gtk.Main;                          use Gtk.Main;
with Ada.Numerics;                      use Ada.Numerics;
with Ada.Numerics.Elementary_Functions; use Ada.Numerics.Elementary_Functions;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Task_Dependencies;                 use Task_Dependencies;
with Task_Set;                          use Task_Set;
with Tasks;                             use Tasks;
use Task_Set.Generic_Task_Set;
with Project;                           use Project;
with Translate;                         use Translate;

with Framework_Config; use Framework_Config;

package body Graph_Callbacks_Pkg is

   procedure Canvas_Zoomed
     (Canvas : access Interactive_Canvas_Record'Class)
   is
      function Internal
        (Canvas : access Interactive_Canvas_Record'Class;
         Item   : access Canvas_Item_Record'Class)
         return   Boolean
      is
         It : constant Item_Object_Access := Item_Object_Access (Item);
      begin
         --  Change the size the item occupies on the screen.
         --
         Set_Screen_Size
           (It,
            To_Canvas_Coordinates (Canvas, It.W),
            To_Canvas_Coordinates (Canvas, It.H));
         return True;
      end Internal;
   begin
      For_Each_Item (Canvas, Internal'Unrestricted_Access);
   end Canvas_Zoomed;

   procedure Zoom_In (Canvas : access Interactive_Canvas_Record'Class) is
   begin
      for J in  Zoom_Levels'First .. Zoom_Levels'Last - 1 loop
         if Zoom_Levels (J) = Get_Zoom (Canvas) then
            Zoom (Canvas, Zoom_Levels (J + 1), 5);
         end if;
      end loop;
   end Zoom_In;

   procedure Zoom_Out (Canvas : access Interactive_Canvas_Record'Class) is
   begin
      for J in  Zoom_Levels'First + 1 .. Zoom_Levels'Last loop
         if Zoom_Levels (J) = Get_Zoom (Canvas) then
            Zoom (Canvas, Zoom_Levels (J - 1), 5);
         end if;
      end loop;
   end Zoom_Out;

   procedure Hide_Graph_Window (G : access Graph_Window_Record'Class) is
   begin
      Hide (G);
   end Hide_Graph_Window;

   procedure Open_Graph (Button : access Gtk_Button_Record'Class) is

   begin
      Gtk_New (Graph_Window);
   end Open_Graph;

   function Close_Graph
     (Win  : access Graph_Window_Record'Class)
      return Gint
   is
   begin
      return 1;
   end Close_Graph;

   procedure Close_Win (Win : access Gtk_Window_Record'Class) is

   begin
      Gtk_Exit (0);
   end Close_Win;

   function Get_Item_Representing_Task
     (Canvas : access Interactive_Canvas_Record'Class;
      A_Task : Generic_Task_Ptr)
      return   Item_Object_Access
   is
      Item_Found : Item_Object_Access;

      function Internal
        (Canv : access Interactive_Canvas_Record'Class;
         Item : access Canvas_Item_Record'Class)
         return Boolean
      is
      begin

         if (A_Task.Name = Get_Name (Item)) then
            Item_Found := Item_Object_Access (Item);
            return False; -- Stop the For_Each_Item loop
         end if;

         return True;
      end Internal;
   begin
      For_Each_Item (Canvas, Internal'Unrestricted_Access);

      return Item_Found;
   end Get_Item_Representing_Task;

   procedure Disp1 (Win : access Graph_Window_Record'Class) is
      Borders_Space       : constant Guint := Win.G1_Borders_Space;
      Space_Between_Items : constant Guint := Win.G1_Space_Between_Items;

      Number_Of_Items_Per_Line : Guint;
      Cur_Item                 : Guint := 0;

      Da_Width : constant Guint := Get_Allocation_Width (Win);

      function Internal
        (Canvas : access Interactive_Canvas_Record'Class;
         Item   : access Canvas_Item_Record'Class)
         return   Boolean
      is
         X, Y : Guint;
      begin
         X := Borders_Space +
              (Guint (Graph_Window_Pkg.Item_Width) + Space_Between_Items) *
              (Cur_Item mod Number_Of_Items_Per_Line);

         Y := Borders_Space +
              (Guint (Graph_Window_Pkg.Item_Height) + Space_Between_Items) *
              (Cur_Item / Number_Of_Items_Per_Line);

         Move_To (Canvas, Item, Gint (X), Gint (Y));
         Cur_Item := Cur_Item + 1;
         return True;
      end Internal;

   begin

      Is_Saved := False;

      Number_Of_Items_Per_Line := 1 +
                                  (Da_Width -
                                   2 * Borders_Space -
                                   Guint (Graph_Window_Pkg.Item_Width)) /
                                  (Guint (Graph_Window_Pkg.Item_Width) +
                                   Space_Between_Items);

      For_Each_Item (Win.Buffer, Internal'Unrestricted_Access);
      Refresh_Canvas (Win.Buffer);
      Win.Cur_Graph_Type := G1;
   end Disp1;

   procedure Disp2 (Win : access Graph_Window_Record'Class) is
      Borders_Space : constant Guint := Win.G2_Borders_Space;

      Angle : Float := 0.0;
      Step  : Float;

      H_Radius : Float :=
         Float (Get_Allocation_Width (Win.Buffer) - 2 * Borders_Space) /
         2.0;
      V_Radius : Float :=
         Float (Get_Allocation_Height (Win.Buffer) - 2 * Borders_Space) /
         2.0;

      function Internal
        (Canvas : access Interactive_Canvas_Record'Class;
         Item   : access Canvas_Item_Record'Class)
         return   Boolean
      is
         X, Y : Float;
      begin
         X := Float (Borders_Space) -
              5.0 +
              H_Radius + -- X center
              H_Radius * Cos (Angle) -
              Float (Graph_Window_Pkg.Item_Width) / 2.0;

         Y := Float (Borders_Space) -
              5.0 +
              V_Radius + -- Y center
              V_Radius * Sin (Angle) -
              Float (Graph_Window_Pkg.Item_Height) / 2.0;

         Move_To (Canvas, Item, Gint (X), Gint (Y));

         Angle := Angle + Step;

         return True;
      end Internal;

   begin

      Is_Saved := False;

      if Win.G2_Force_Circle then
         H_Radius := Float'Min (H_Radius, V_Radius);
         V_Radius := Float'Min (H_Radius, V_Radius);
      end if;

      Step := 2.0 * Pi / Float (Get_Number_Of_Items (Win.Buffer));

      For_Each_Item (Win.Buffer, Internal'Unrestricted_Access);

      Refresh_Canvas (Win.Buffer);
      Win.Cur_Graph_Type := G2;
   end Disp2;

   procedure Disp3 (Win : access Graph_Window_Record'Class) is
      Borders_Space       : constant Guint := Win.G3_Borders_Space;
      Space_Between_Items : constant Guint := Win.G3_Space_Between_Items;

      An_Iterator, Src_Iterator : Iterator;
      A_Task, Src_Task          : Generic_Task_Ptr;

      Cur_Item   : Guint := 0;
      Cur_Column : Guint := 1;

      Src_Set, Dest_Set, Done_Set, Succ_Set, Pit_Set : Tasks_Set;

   begin

      Is_Saved := False;

      if not Is_Empty (Sys.Tasks) then
         -- Search after (a) source task(s)
         --
         loop
            Current_Element (Sys.Tasks, A_Task, An_Iterator);

            if (not Has_Predecessor (Sys.Dependencies, A_Task) and
                Has_Successor (Sys.Dependencies, A_Task))
            then
               Move_To
                 (Win.Buffer,
                  Get_Item_Representing_Task (Win.Buffer, A_Task),
                  Gint (Borders_Space),
                  Gint (Borders_Space +
                        Cur_Item *
                        (Guint (Graph_Window_Pkg.Item_Height) +
                         Space_Between_Items)));

               Cur_Item := Cur_Item + 1;

               Add (Done_Set, A_Task);
               Add (Src_Set, A_Task);
            end if;

            exit when Is_Last_Element (Sys.Tasks, An_Iterator);
            Next_Element (Sys.Tasks, An_Iterator);
         end loop;

         -- Now we have found the source(s), we have to loop until we haven't
         --any
         -- more successor.
         --
         -- Don't try to think about it : I tell you it can't be in the same
         --loop
         -- (or it wouldn't be comprehensible :)
         --
         loop
            Reset (Dest_Set);
            Cur_Item := 0;

            if not Is_Empty (Src_Set) then
               Reset_Iterator (Src_Set, Src_Iterator);
               loop
                  Current_Element (Src_Set, Src_Task, Src_Iterator);

                  Succ_Set :=
                     Get_Successors_List (Sys.Dependencies, Src_Task);
                  exit when Get_Number_Of_Elements (Succ_Set) = 0;

                  -- Then, for each of these successors, we have to check it
                  --hasn't been
                  -- already processed. If it hasn't been, we can lay it out
                  --depending
                  -- of it's number in the current column.
                  --
                  if not Is_Empty (Succ_Set) then
                     Reset_Iterator (Succ_Set, An_Iterator);
                     loop
                        Current_Element (Succ_Set, A_Task, An_Iterator);

                        begin
                           if Search_Task (Done_Set, A_Task.Name) =
                              null
                           then
                              null;
                           end if;

                        exception
                           when Task_Set.Task_Not_Found =>
                              if (Has_Successor
                                     (Sys.Dependencies,
                                      A_Task))
                              then
                                 Move_To
                                   (Win.Buffer,
                                    Get_Item_Representing_Task
                                       (Win.Buffer,
                                        A_Task),
                                    Gint (Borders_Space +
                                          Cur_Column *
                                          (Guint (Graph_Window_Pkg.Item_Width)
                                           +
                                           Space_Between_Items)),
                                    Gint (Borders_Space +
                                          Cur_Item *
                                          (Guint (Graph_Window_Pkg.Item_Height)
                                           +
                                           Space_Between_Items)));

                                 Cur_Item := Cur_Item + 1;

                                 Add (Dest_Set, A_Task);
                                 Add (Done_Set, A_Task);
                              else
                                 Add (Pit_Set, A_Task);
                                 Add (Done_Set, A_Task);
                              end if;
                        end;

                        exit when Is_Last_Element (Succ_Set, An_Iterator);
                        Next_Element (Succ_Set, An_Iterator);
                     end loop;
                  end if;
                  exit when Is_Last_Element (Src_Set, Src_Iterator);
                  Next_Element (Src_Set, Src_Iterator);

               end loop;
            end if;
            exit when Get_Number_Of_Elements (Dest_Set) = 0;

            Src_Set    := Dest_Set;
            Cur_Column := Cur_Column + 1;
         end loop;
         if not Is_Empty (Pit_Set) then
            Reset_Iterator (Pit_Set, An_Iterator);

            for X in  1 .. Get_Number_Of_Elements (Pit_Set) loop
               Current_Element (Pit_Set, A_Task, An_Iterator);

               Move_To
                 (Win.Buffer,
                  Get_Item_Representing_Task (Win.Buffer, A_Task),
                  Gint (Borders_Space +
                        Cur_Column *
                        (Guint (Graph_Window_Pkg.Item_Width) +
                         Space_Between_Items)),
                  Gint (Borders_Space +
                        Cur_Item *
                        (Guint (Graph_Window_Pkg.Item_Height) +
                         Space_Between_Items)));
               Cur_Item := Cur_Item + 1;
               Next_Element (Pit_Set, An_Iterator);
            end loop;
         end if;
         Refresh_Canvas (Win.Buffer);
         Win.Cur_Graph_Type := G3;
      end if;
   end Disp3;

   procedure Callbacks_Init
     (Win : Graph_Window_Access;
      Tb  : Graph_Toolbar_Access)
   is

   begin

      Graph_Window_Gint_Cb.Connect
        (Win,
         "delete_event",
         Graph_Window_Gint_Cb.To_Marshaller (Close_Graph'Access));

      Canvas_Cb.Connect
        (Win.Buffer,
         "zoomed",
         Canvas_Cb.To_Marshaller (Canvas_Zoomed'Access));

      Canvas_Cb.Object_Connect
        (Append_Item
            (Tb.Action_Toolbar,
             Tooltip_Text => " Zoom out ",
             Icon         => Gtk_Widget (Tb.Pixmap_Zoom_Out)),
         "clicked",
         Canvas_Cb.To_Marshaller (Zoom_Out'Access),
         Slot_Object => Win.Buffer);

      Canvas_Cb.Object_Connect
        (Append_Item
            (Tb.Action_Toolbar,
             Tooltip_Text => " Zoom in ",
             Icon         => Gtk_Widget (Tb.Pixmap_Zoom_In)),
         "clicked",
         Canvas_Cb.To_Marshaller (Zoom_In'Access),
         Slot_Object => Win.Buffer);

      Append_Space (Tb.Action_Toolbar);
      Append_Space (Tb.Action_Toolbar);
      Append_Space (Tb.Action_Toolbar);
      Append_Space (Tb.Action_Toolbar);

      Graph_Window_Cb.Object_Connect
        (Append_Item
            (Tb.Action_Toolbar,
             Tooltip_Text => To_String (Lb_Disp1 (Current_Language)),
             Icon         => Gtk_Widget (Tb.Pixmap_Disp1)),
         "clicked",
         Graph_Window_Cb.To_Marshaller (Disp1'Access),
         Slot_Object => Win);

      Graph_Window_Cb.Object_Connect
        (Append_Item
            (Tb.Action_Toolbar,
             Tooltip_Text => To_String (Lb_Disp2 (Current_Language)),
             Icon         => Gtk_Widget (Tb.Pixmap_Disp2)),
         "clicked",
         Graph_Window_Cb.To_Marshaller (Disp2'Access),
         Slot_Object => Win);

      Graph_Window_Cb.Object_Connect
        (Append_Item
            (Tb.Action_Toolbar,
             Tooltip_Text => To_String (Lb_Disp3 (Current_Language)),
             Icon         => Gtk_Widget (Tb.Pixmap_Disp3)),
         "clicked",
         Graph_Window_Cb.To_Marshaller (Disp3'Access),
         Slot_Object => Win);

      Append_Space (Tb.Action_Toolbar);
      Append_Space (Tb.Action_Toolbar);
      Append_Space (Tb.Action_Toolbar);
      Append_Space (Tb.Action_Toolbar);
      Graph_Window_Cb.Object_Connect
        (Append_Item
            (Tb.Action_Toolbar,
             Tooltip_Text => To_String (Lb_Exit (Current_Language)),
             Icon         => Gtk_Widget (Tb.Pixmap_Close)),
         "clicked",
         Graph_Window_Cb.To_Marshaller (Hide_Graph_Window'Access),
         Slot_Object => Win);

   end Callbacks_Init;

end Graph_Callbacks_Pkg;
