------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Framework_Config; use Framework_Config;
with Translate;        use Translate;

with Graph_Toolbar_Pkg;     use Graph_Toolbar_Pkg;
with Graph_Callbacks_Pkg;   use Graph_Callbacks_Pkg;
with Gtk.Enums;             use Gtk.Enums;
with Project;               use Project;
with Graph_Item_Pkg;        use Graph_Item_Pkg;
with Task_Set;              use Task_Set;
use Task_Set.Generic_Task_Set;
with Tasks;                 use Tasks;
with Message_Set;           use Message_Set;
use Message_Set.Generic_Message_Set;
with Messages;              use Messages;
with Buffer_Set;            use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Buffers;               use Buffers;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Task_Dependencies;     use Task_Dependencies;
use Task_Dependencies.Half_Dep_Set;
with Dependencies;          use Dependencies;

package body Graph_Window_Pkg is

   pragma Warnings (Off);

   procedure Rebuild_Canvas (Graph_Window : Graph_Window_Access) is

      function Internal_Append
        (Canva : access Interactive_Canvas_Record'Class;
         Item  : access Canvas_Item_Record'Class)
         return  Boolean
      is
      begin
         Remove (Graph_Window.Buffer, Item);

         return True;
      end Internal_Append;

   begin
      if (Graph_Window_Pkg.Canvas_Is_Created = True) then
         For_Each_Item
           (Graph_Window.Buffer,
            Internal_Append'Unrestricted_Access);

         -- include of task,message and buffer
         -- into the window
         --
         Import_Task (Graph_Window);
         Import_Message (Graph_Window);
         Import_Buffer (Graph_Window);
         Import_Dependencies (Graph_Window);

      end if;

   end Rebuild_Canvas;

   procedure Import_Task (Graph_Window : Graph_Window_Access) is

      My_Iterator : Tasks_Iterator;
      A_Task      : Generic_Task_Ptr;
      Item_Tmp    : Item_Object_Access;
      H_Position  : constant Gint := 60;
      X_Current   : Gint          := 10;
      Y_Current   : Gint          := 90;

   begin
      if (not Is_Empty (Sys.Tasks)) then
         Reset_Iterator (Sys.Tasks, My_Iterator);
         loop
            Current_Element (Sys.Tasks, A_Task, My_Iterator);
            if Name_Is_New (Graph_Window.Buffer, A_Task.Name) then
               Item_Object_New (Item_Tmp, To_Unbounded_String ("task"), True);
               Set_Name (Item_Tmp, A_Task.Name);
               Set_Visibility (Item_Tmp, False);
               Put (Graph_Window.Buffer, Item_Tmp);
               Move_To (Graph_Window.Buffer, Item_Tmp, X_Current, Y_Current);
               X_Current := X_Current + H_Position;
            end if;

            exit when Is_Last_Element (Sys.Tasks, My_Iterator) = True;
            Next_Element (Sys.Tasks, My_Iterator);
         end loop;
      end if;
   end Import_Task;

   procedure Import_Message (Graph_Window : Graph_Window_Access) is
      My_Iterator : Messages_Iterator;
      A_Message   : Message_Ptr;
      Item_Tmp    : Item_Object_Access;
      H_Position  : constant Gint := 90;
      X_Current   : Gint          := 10;
      Y_Current   : Gint          := 20;

   begin
      if (not Is_Empty (Sys.Messages)) then
         Reset_Iterator (Sys.Messages, My_Iterator);
         loop
            Current_Element (Sys.Messages, A_Message, My_Iterator);
            if Name_Is_New (Graph_Window.Buffer, A_Message.Name) then
               Item_Object_New
                 (Item_Tmp,
                  To_Unbounded_String ("message"),
                  True);
               Set_Name (Item_Tmp, A_Message.Name);
               Set_Visibility (Item_Tmp, False);
               Put (Graph_Window.Buffer, Item_Tmp);
               Move_To (Graph_Window.Buffer, Item_Tmp, X_Current, Y_Current);
               X_Current := X_Current + H_Position;
            end if;
            exit when Is_Last_Element (Sys.Messages, My_Iterator) = True;
            Next_Element (Sys.Messages, My_Iterator);
         end loop;
      end if;
   end Import_Message;

   procedure Import_Buffer (Graph_Window : Graph_Window_Access) is
      My_Iterator : Buffers_Iterator;
      A_Buffer    : Buffer_Ptr;
      Item_Tmp    : Item_Object_Access;
      H_Position  : constant Gint := 90;
      X_Current   : Gint          := 10;
      Y_Current   : Gint          := 150;
   begin
      if (not Is_Empty (Sys.Buffers)) then
         Reset_Iterator (Sys.Buffers, My_Iterator);
         loop
            Current_Element (Sys.Buffers, A_Buffer, My_Iterator);
            if Name_Is_New (Graph_Window.Buffer, A_Buffer.Name) then
               Item_Object_New
                 (Item_Tmp,
                  To_Unbounded_String ("buffer"),
                  True);
               Set_Name (Item_Tmp, A_Buffer.Name);
               Set_Visibility (Item_Tmp, False);
               Put (Graph_Window.Buffer, Item_Tmp);
               Move_To (Graph_Window.Buffer, Item_Tmp, X_Current, Y_Current);
               X_Current := X_Current + H_Position;
            end if;
            exit when Is_Last_Element (Sys.Buffers, My_Iterator) = True;
            Next_Element (Sys.Buffers, My_Iterator);
         end loop;
      end if;
   end Import_Buffer;

   procedure Import_Dependencies (Graph_Window : Graph_Window_Access) is

      My_Iterator       : Tasks_Dependencies_Iterator;
      A_Half_Dep        : Dependency_Ptr;
      Item1, Item2      : Canvas_Item;
      My_Tasks_Set      : Tasks_Set;
      My_Other_Iterator : Tasks_Iterator;
      A_Task            : Generic_Task_Ptr;

   begin

      if not Is_Empty (Sys.Dependencies.all) then
         Reset_Iterator (Sys.Dependencies.all, My_Iterator);
         loop
            Current_Element (Sys.Dependencies.all, A_Half_Dep, My_Iterator);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               Item1 :=
                  Search_Item_By_Name
                    (Graph_Window.Buffer,
                     To_Unbounded_String ("task"),
                     A_Half_Dep.Precedency_To_Task.Name);
            end if;
            if A_Half_Dep.Depend_Type = Communication_Dependency then
               Item1 :=
                  Search_Item_By_Name
                    (Graph_Window.Buffer,
                     To_Unbounded_String ("task"),
                     A_Half_Dep.Communication_Dependent_Task.Name);
            end if;
            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               Item1 :=
                  Search_Item_By_Name
                    (Graph_Window.Buffer,
                     To_Unbounded_String ("task"),
                     A_Half_Dep.Buffer_Dependent_Task.Name);
            end if;

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               Item2 :=
                  Search_Item_By_Name
                    (Graph_Window.Buffer,
                     To_Unbounded_String ("task"),
                     A_Half_Dep.Precedency_From_Task.Name);
            end if;
            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               Item2 :=
                  Search_Item_By_Name
                    (Graph_Window.Buffer,
                     To_Unbounded_String ("buffer"),
                     A_Half_Dep.Buffer_Dependency_Object.Name);
            end if;
            if A_Half_Dep.Depend_Type = Communication_Dependency then
               Item2 :=
                  Search_Item_By_Name
                    (Graph_Window.Buffer,
                     To_Unbounded_String ("message"),
                     A_Half_Dep.Communication_Dependency_Object.Name);
            end if;

            Set_Visibility (Item1, True);
            Set_Visibility (Item2, True);

            if A_Half_Dep.Depend_Type = Precedence_Dependency then
               if not Has_Link_Between
                        (Graph_Window.Buffer,
                         Item2,
                         Item1)
               then
                  Add_Link (Graph_Window.Buffer, Item2, Item1);
               end if;
            end if;

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if A_Half_Dep.Buffer_Orientation = From_Task_To_Object then
                  if not Has_Link_Between
                           (Graph_Window.Buffer,
                            Item1,
                            Item2)
                  then
                     Add_Link (Graph_Window.Buffer, Item1, Item2);
                  end if;
               else
                  if not Has_Link_Between
                           (Graph_Window.Buffer,
                            Item2,
                            Item1)
                  then
                     Add_Link (Graph_Window.Buffer, Item2, Item1);
                  end if;
               end if;
            end if;

            if A_Half_Dep.Depend_Type = Buffer_Dependency then
               if A_Half_Dep.Buffer_Orientation = From_Task_To_Object then
                  if not Has_Link_Between
                           (Graph_Window.Buffer,
                            Item1,
                            Item2)
                  then
                     Add_Link (Graph_Window.Buffer, Item1, Item2);
                  end if;
               else
                  if not Has_Link_Between
                           (Graph_Window.Buffer,
                            Item2,
                            Item1)
                  then
                     Add_Link (Graph_Window.Buffer, Item2, Item1);
                  end if;
               end if;
            end if;

            if A_Half_Dep.Depend_Type = Communication_Dependency then
               if A_Half_Dep.Communication_Orientation =
                  From_Task_To_Object
               then
                  if not Has_Link_Between
                           (Graph_Window.Buffer,
                            Item1,
                            Item2)
                  then
                     Add_Link (Graph_Window.Buffer, Item1, Item2);
                  end if;
               else
                  if not Has_Link_Between
                           (Graph_Window.Buffer,
                            Item2,
                            Item1)
                  then
                     Add_Link (Graph_Window.Buffer, Item2, Item1);
                  end if;
               end if;
            end if;

            exit when Is_Last_Element (Sys.Dependencies.all, My_Iterator) =
                      True;
            Next_Element (Sys.Dependencies.all, My_Iterator);
         end loop;
      end if;

   end Import_Dependencies;

   procedure Gtk_New (Graph_Window : in out Graph_Window_Access) is

   begin

      if Is_Hidden = False then
         Graph_Window := new Graph_Window_Record;
         Is_Hidden    := True;
         Graph_Window_Pkg.Initialize (Graph_Window);
      else
         -- include of tasks, messages and buffers
         -- into the window
         --
         Import_Task (Graph_Window);
         Import_Message (Graph_Window);
         Import_Buffer (Graph_Window);
         Import_Dependencies (Graph_Window);

         Show (Graph_Window);
      end if;

   end Gtk_New;

   procedure Initialize (Graph_Window : Graph_Window_Access) is

      Toolbar : Graph_Toolbar_Access;

   begin

      Graph_Window_Pkg.Canvas_Is_Created := True;

      -- A child window
      --
      Gtk.Window.Initialize (Graph_Window, Window_Toplevel);
      Set_Policy
        (Graph_Window,
         Allow_Shrink => False,
         Allow_Grow   => True,
         Auto_Shrink  => True);

      Set_Title (Graph_Window, To_String (Lb_Graph_View (Current_Language)));
      Set_USize (Graph_Window, 600, 400);

      Gtk_New_Vbox (Graph_Window.V_Box, Homogeneous => False, Spacing => 0);
      Add (Gtk_Window (Graph_Window), Graph_Window.V_Box);

      New_Tool_Bar (Toolbar, Graph_Window);
      Add (Graph_Window.V_Box, Toolbar.Action_Toolbar);
      Add (Graph_Window.V_Box, Toolbar.Disp_Toolbar);

      -- we want to scroll with Large graphics
      --
      Gtk_New (Graph_Window.Scrolled);
      Set_Usize (Graph_Window.Scrolled, 600, 350);
      Add (Graph_Window.V_Box, Graph_Window.Scrolled);

      Gtk_New (Graph_Window.Buffer);

      Add (Graph_Window.Scrolled, Graph_Window.Buffer);
      Show_All (Graph_Window);
      Settoolbar (Graph_Window.Buffer, Toolbar);

      Gdk_New (Graph_Window.Green_Gc, Get_Window (Graph_Window.Buffer));

      -- Store the preferences
      --
      Graph_Window.G1_Borders_Space       := Guint (200);
      Graph_Window.G1_Space_Between_Items := Guint (100);
      Graph_Window.G2_Borders_Space       := Guint (200);
      Graph_Window.G2_Force_Circle        := True;
      Graph_Window.G3_Borders_Space       := Guint (100);
      Graph_Window.G3_Space_Between_Items := Guint (150);

      Graph_Callbacks_Pkg.Callbacks_Init (Graph_Window, Toolbar);
      Show_All (Graph_Window);

      -- Include of task,message and buffer
      --
      Import_Task (Graph_Window);
      Import_Message (Graph_Window);
      Import_Buffer (Graph_Window);
      Import_Dependencies (Graph_Window);
      Is_Hidden := True;

      -- Load Item x/y positions
      --
      Set_Coordinates (Graph_Window.Buffer, Sys);

   end Initialize;

end Graph_Window_Pkg;
