------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Gdk.Drawable;          use Gdk.Drawable;
with Gtk.Arguments;         use Gtk.Arguments;
with Gtk.Drawing_Area;      use Gtk.Drawing_Area;
with Gtk.Handlers;          use Gtk.Handlers;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Processor_Set;         use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Tasks;                 use Tasks;
with Resources;             use Resources;
with Messages;              use Messages;
with Buffers;               use Buffers;
with Time_Units;            use Time_Units;
use Time_Units.A_Time_Unit;
with Unbounded_Strings;     use Unbounded_Strings;
with Events;                use Events;
with Framework_Config;      use Framework_Config;
with Translate;             use Translate;
with Select_Time_Line_Pkg;  use Select_Time_Line_Pkg;
use Select_Time_Line_Pkg.Generic_Time_Line_Object_Set;
with Objects;               use Objects;
with Editor_Config;         use Editor_Config;

package body Draw_Scheduling_Pkg is

   package Draw_Cb is new Gtk.Handlers.Return_Callback (
      Widget_Type => Gtk_Double_Buffer_Record,
      Return_Type => Boolean);

   package Allocation_Cb is new Handlers.Callback (Gtk_Double_Buffer_Record);
   package Allocation_Marshaller is new
     Allocation_Cb.Marshallers.Generic_Marshaller (
      Gtk_Allocation_Access,
      To_Allocation);

   procedure Color_Allocation
     (Draw_Scheduling : access Draw_Scheduling_Record)
   is
   begin

      if not Draw_Scheduling.Have_Colors then

         Gdk.GC.Gdk_New
           (Draw_Scheduling.White_Gc,
            Get_Window (Draw_Scheduling.Buffer));
         Gdk.GC.Set_Foreground
           (Draw_Scheduling.White_Gc,
            Gdk.Color.White (Gtk.Widget.Get_Default_Colormap));

         Gdk.GC.Gdk_New
           (Draw_Scheduling.Black_Gc,
            Get_Window (Draw_Scheduling.Buffer));
         Gdk.GC.Set_Foreground
           (Draw_Scheduling.Black_Gc,
            Gdk.Color.Black (Gtk.Widget.Get_Default_Colormap));

         Draw_Scheduling.Red_Color := Gdk.Color.Parse ("Red");
         Gdk.Color.Alloc
           (Gtk.Widget.Get_Default_Colormap,
            Draw_Scheduling.Red_Color);

         Draw_Scheduling.Blue_Color := Gdk.Color.Parse ("Blue");
         Gdk.Color.Alloc
           (Gtk.Widget.Get_Default_Colormap,
            Draw_Scheduling.Blue_Color);

         Draw_Scheduling.Have_Colors := True;

      end if;
   end Color_Allocation;

   function Configure
     (Buffer : access Gtk_Double_Buffer_Record'Class;
      Event  : Gdk.Event.Gdk_Event)
      return   Boolean
   is
   begin
      if Draw_Scheduling.Initialized then
         Redraw_Time_Line (Draw_Scheduling);
      else
         Color_Allocation (Draw_Scheduling);

         -- Check if window size has changed !
         --

         Draw_Scheduling.Current_Width  :=
           Gint (Get_Allocation_Width (Draw_Scheduling.Buffer));
         Draw_Scheduling.Current_Height :=
           Gint (Get_Allocation_Height (Draw_Scheduling.Buffer));

         Draw_Rectangle
           (Get_Pixmap (Draw_Scheduling.Buffer),
            Draw_Scheduling.White_Gc,
            Filled => True,
            X      => 0,
            Y      => 0,
            Width  => Draw_Scheduling.Current_Width,
            Height => Draw_Scheduling.Current_Height);

         Double_Buffer.Draw (Draw_Scheduling.Buffer);

      end if;

      return True;
   end Configure;

   --  This function is called once gtk has decided what size and position
   --  the widget will actually have, or everytime the widget is resized.
   --  This would be a good time for instance for resizing the component
   --  sub-widgets.
   --  Note that we have to move and resize the widget ourselves, and that
   --  for such a simple case, we could simply rely on the ancestor's
   --  size_allocation function.
   --
   procedure Size_Allocate
     (Buffer     : access Gtk_Double_Buffer_Record'Class;
      Allocation : Gtk_Allocation_Access)
   is
   begin
      if Realized_Is_Set (Buffer) then
         Draw_Scheduling.Current_Width  :=
           Gint (Get_Allocation_Width (Draw_Scheduling.Buffer));
         Draw_Scheduling.Current_Height :=
           Gint (Get_Allocation_Height (Draw_Scheduling.Buffer));
         if Draw_Scheduling.Initialized then
            Redraw_Time_Line (Draw_Scheduling);
         else
            Color_Allocation (Draw_Scheduling);

            Draw_Rectangle
              (Get_Pixmap (Draw_Scheduling.Buffer),
               Draw_Scheduling.White_Gc,
               Filled => True,
               X      => 0,
               Y      => 0,
               Width  => Draw_Scheduling.Current_Width,
               Height => Draw_Scheduling.Current_Height);

            Double_Buffer.Draw (Draw_Scheduling.Buffer);

         end if;
      end if;
      Gtk.Handlers.Emit_Stop_By_Name (Buffer, "size_allocate");
   end Size_Allocate;

   procedure Gtk_New (Draw_Scheduling : out Draw_Scheduling_Access) is

   begin
      Draw_Scheduling := new Draw_Scheduling_Record;
      Draw_Scheduling_Pkg.Initialize (Draw_Scheduling);
   end Gtk_New;

   procedure Initialize
     (Draw_Scheduling : access Draw_Scheduling_Record'Class)
   is

   begin

      Draw_Scheduling.Current_Width  := Draw_Scheduling.First_Width;
      Draw_Scheduling.Current_Height := Draw_Scheduling.First_Height;

      Gtk_New (Draw_Scheduling.Buffer);

      Gtk.Drawing_Area.Set_Usize
        (Gtk_Drawing_Area (Draw_Scheduling.Buffer),
         Draw_Scheduling.First_Width,
         Draw_Scheduling.First_Height);
      Show_All (Draw_Scheduling.Buffer);

      Load
        (Draw_Scheduling.Font,
         "-adobe-helvetica-medium-o-normal--14-140-75-75-p-78-iso8859-1");

      Draw_Cb.Connect
        (Draw_Scheduling.Buffer,
         "configure_event",
         Draw_Cb.To_Marshaller (Configure'Access));
      Allocation_Cb.Connect
        (Draw_Scheduling.Buffer,
         "size_allocate",
         Allocation_Marshaller.To_Marshaller (Size_Allocate'Access));

   end Initialize;

   procedure Clear_Time_Line
     (Draw_Scheduling : access Draw_Scheduling_Record)
   is
   begin

      Color_Allocation (Draw_Scheduling);

      Draw_Scheduling.Initialized := False;

      Draw_Rectangle
        (Get_Pixmap (Draw_Scheduling.Buffer),
         Draw_Scheduling.White_Gc,
         Filled => True,
         X      => 0,
         Y      => 0,
         Width  => Draw_Scheduling.Current_Width,
         Height => Draw_Scheduling.Current_Height);

      Double_Buffer.Draw (Draw_Scheduling.Buffer);

   end Clear_Time_Line;

   procedure Draw_Time_Line
     (Draw_Scheduling : access Draw_Scheduling_Record;
      Task1           : Tasks_Set;
      Resource1       : Resources_Set;
      Message1        : Messages_Set;
      Buffer1         : Buffers_Set;
      Period1         : Natural;
      Start_Draw1     : Natural;
      Sched1          : Scheduling_Table_Ptr)
   is

      I  : Gint := 0;
      Ok : Boolean;

   begin

      Draw_Scheduling.Tasks      := Task1;
      Draw_Scheduling.Resources  := Resource1;
      Draw_Scheduling.Messages   := Message1;
      Draw_Scheduling.Buffers    := Buffer1;
      Draw_Scheduling.Period     := Period1;
      Draw_Scheduling.Start_Draw := Start_Draw1;
      Draw_Scheduling.Sched      := Sched1;

      -- Set Left_margin
      --
      Draw_Scheduling.Left_Margin := 20;

      -- Compute Widget size
      --
      Ok := False;
      I  := Gint (Period1 - Start_Draw1) * Draw_Scheduling.Timeunit_Margin +
            Draw_Scheduling.Left_Margin +
            Draw_Scheduling.Right_Margin;
      if I > Draw_Scheduling.Current_Width then
         Draw_Scheduling.Current_Width := I;
         Ok                            := True;
      end if;

      if Is_Empty (Time_Line_To_Display) then
         I := Draw_Scheduling.High_Margin +
              Draw_Scheduling.Interligne *
              Gint (Get_Number_Of_Elements (Task1));
         if Schedule_With_Resources then
            I := I +
                 Draw_Scheduling.Interligne *
                 Gint (Get_Number_Of_Elements (Resource1));
         end if;
         if Schedule_With_Precedencies then
            I := I +
                 Draw_Scheduling.Interligne *
                 Gint (Get_Number_Of_Elements (Buffer1));
            I := I +
                 Draw_Scheduling.Interligne *
                 Gint (Get_Number_Of_Elements (Message1));
         end if;
      else
         I := Draw_Scheduling.Interligne *
              Gint (Get_Number_Of_Elements (Time_Line_To_Display));
      end if;
      if I > Draw_Scheduling.Current_Height then
         Draw_Scheduling.Current_Height := I;
         Ok                             := True;
      end if;

      if Ok then
         Gtk.Drawing_Area.Set_Usize
           (Gtk_Drawing_Area (Draw_Scheduling.Buffer),
            Draw_Scheduling.Current_Width,
            Draw_Scheduling.Current_Height);
      end if;

      Draw_Scheduling.Initialized := True;

      Redraw_Time_Line (Draw_Scheduling);

   end Draw_Time_Line;

   procedure Redraw_Time_Line
     (Draw_Scheduling : access Draw_Scheduling_Record'Class)
   is

      My_Task_Iterator     : Tasks_Iterator;
      My_Resource_Iterator : Resources_Iterator;
      My_Buffer_Iterator   : Buffers_Iterator;
      My_Message_Iterator  : Messages_Iterator;
      A_Task               : Generic_Task_Ptr;
      A_Message            : Message_Ptr;
      A_Resource           : Generic_Resource_Ptr;
      A_Buffer             : Buffer_Ptr;

      I : Gint := 0;

      Msg, Name : Unbounded_String;

      type Object_Range is new Integer range
         0 .. (Max_Tasks + Max_Resources + Max_Buffers + Max_Messages);
      Task_Name     : array (Object_Range) of Unbounded_String;
      Resource_Name : array (Object_Range) of Unbounded_String;
      Message_Name  : array (Object_Range) of Unbounded_String;
      Buffer_Name   : array (Object_Range) of Unbounded_String;

      Task_Id    : Gint;
      Message_Id : Gint;
      Buffer_Id  : Gint;

      Time_U       : Gint;
      Save_Start_U : array (Object_Range) of Gint := (others => 0);
      Save_Start_M : array (Object_Range) of Gint := (others => 0);
      Ok           : Boolean;

   begin

      ---------------------------------
      -- Clear work space
      ---------------------------------

      Draw_Rectangle
        (Get_Pixmap (Draw_Scheduling.Buffer),
         Draw_Scheduling.White_Gc,
         Filled => True,
         X      => 0,
         Y      => 0,
         Width  => Draw_Scheduling.Current_Width,
         Height => Draw_Scheduling.Current_Height);

      ---------------------------------
      -- Draw a line per task
      ---------------------------------

      I := 0;
      Reset_Iterator (Draw_Scheduling.Tasks, My_Task_Iterator);

      loop

         Current_Element (Draw_Scheduling.Tasks, A_Task, My_Task_Iterator);

         if To_Display (A_Task.Name, Task_Object_Type) then

            Msg := Lb_Task_Name (Current_Language) &
                   "=" &
                   A_Task.Name &
                   "        ";

            case A_Task.Task_Type is
               when Aperiodic_Type =>
                  Msg :=
                    Msg &
                    Lb_Capacity (Current_Language) &
                    To_Unbounded_String ("=") &
                    A_Task.Capacity'Img &
                    To_Unbounded_String ("; ") &
                    Lb_Deadline (Current_Language) &
                    To_Unbounded_String ("=") &
                    Aperiodic_Task_Ptr (A_Task).Deadline'Img &
                    To_Unbounded_String ("; ") &
                    Lb_Start_Time (Current_Language) &
                    To_Unbounded_String ("=") &
                    A_Task.Start_Time'Img &
                    To_Unbounded_String ("; ") &
                    Lb_Priority (Current_Language) &
                    To_Unbounded_String ("=") &
                    A_Task.Priority'Img &
                    To_Unbounded_String ("; Cpu=") &
                    A_Task.Cpu_Name;
               when others =>
                  Msg :=
                    Msg &
                    Lb_Period (Current_Language) &
                    To_Unbounded_String ("=") &
                    Periodic_Task_Ptr (A_Task).Period'Img &
                    To_Unbounded_String ("; ") &
                    Lb_Capacity (Current_Language) &
                    To_Unbounded_String ("=") &
                    A_Task.Capacity'Img &
                    To_Unbounded_String ("; ") &
                    Lb_Deadline (Current_Language) &
                    To_Unbounded_String ("=") &
                    Periodic_Task_Ptr (A_Task).Deadline'Img &
                    To_Unbounded_String ("; ") &
                    Lb_Start_Time (Current_Language) &
                    To_Unbounded_String ("=") &
                    A_Task.Start_Time'Img &
                    To_Unbounded_String ("; ") &
                    Lb_Priority (Current_Language) &
                    To_Unbounded_String ("=") &
                    A_Task.Priority'Img &
                    To_Unbounded_String ("; Cpu=") &
                    A_Task.Cpu_Name;
            end case;
            if A_Task.Task_Type = Parametric_Type then
               Msg := Msg &
                      To_Unbounded_String (";") &
                      Lb_Activation_Rule (Current_Language) &
                      To_Unbounded_String ("=") &
                      Parametric_Task_Ptr (A_Task).Activation_Rule;
            end if;

            Draw_Text
              (Get_Pixmap (Draw_Scheduling.Buffer),
               Draw_Scheduling.Font,
               Draw_Scheduling.Black_Gc,
               Draw_Scheduling.Left_Margin,
               Draw_Scheduling.High_Margin +
               I * Draw_Scheduling.Interligne +
               20,
               To_String (Msg));

            Draw_Line
              (Get_Pixmap (Draw_Scheduling.Buffer),
               Draw_Scheduling.Black_Gc,
               Draw_Scheduling.Left_Margin,
               Draw_Scheduling.High_Margin + I * Draw_Scheduling.Interligne,
               Draw_Scheduling.Left_Margin +
               Gint (Draw_Scheduling.Period - Draw_Scheduling.Start_Draw) *
               Draw_Scheduling.Timeunit_Margin,
               Draw_Scheduling.High_Margin + I * Draw_Scheduling.Interligne);

            -- For each line, draw unit time lines
            --
            for J in  0 .. Gint (Draw_Scheduling.Period -
                                 Draw_Scheduling.Start_Draw)
            loop
               Draw_Line
                 (Get_Pixmap (Draw_Scheduling.Buffer),
                  Draw_Scheduling.Black_Gc,
                  Draw_Scheduling.Left_Margin +
                  J * Draw_Scheduling.Timeunit_Margin,
                  Draw_Scheduling.High_Margin +
                  I * Draw_Scheduling.Interligne -
                  3,
                  Draw_Scheduling.Left_Margin +
                  J * Draw_Scheduling.Timeunit_Margin,
                  Draw_Scheduling.High_Margin +
                  I * Draw_Scheduling.Interligne +
                  3);
            end loop;

            -- Register each task in the name table
            --
            Task_Name (Object_Range (I))  := A_Task.Name;

            I := I + 1;
         end if;

         exit when Is_Last_Element (Draw_Scheduling.Tasks, My_Task_Iterator);
         Next_Element (Draw_Scheduling.Tasks, My_Task_Iterator);

      end loop;

      -------------------------------------
      -- Draw a line for each resource
      -------------------------------------

      if Schedule_With_Resources then

         if not Is_Empty (Draw_Scheduling.Resources) then

            Reset_Iterator (Draw_Scheduling.Resources, My_Resource_Iterator);

            loop

               Current_Element
                 (Draw_Scheduling.Resources,
                  A_Resource,
                  My_Resource_Iterator);

               if To_Display (A_Resource.Name, Resource_Object_Type) then

                  Msg := Lb_Resource_Name (Current_Language) &
                         "=" &
                         A_Resource.Name &
                         "        ";

                  if (A_Resource.protocol = Priority_Ceiling_Protocol) then
                     Msg := Msg & To_Unbounded_String ("Protocol = PCP");
                  else
                     if (A_Resource.protocol =
                         Priority_Inheritance_Protocol)
                     then
                        Msg := Msg & To_Unbounded_String ("Protocol = PIP");
                     else
                        if (A_Resource.protocol =
                            Immediate_Priority_Ceiling_Protocol)
                        then
                           Msg := Msg &
                                  To_Unbounded_String ("Protocol = IPCP");

                        else
                           Msg := Msg &
                                  To_Unbounded_String
                                     ("Protocol = No protocol");
                        end if;
                     end if;
                  end if;

                  Msg := Msg &
                         To_Unbounded_String ("; Cpu=") &
                         A_Task.Cpu_Name;

                  Draw_Text
                    (Get_Pixmap (Draw_Scheduling.Buffer),
                     Draw_Scheduling.Font,
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,
                     Draw_Scheduling.High_Margin +
                     I * Draw_Scheduling.Interligne +
                     20,
                     To_String (Msg));

                  Draw_Line
                    (Get_Pixmap (Draw_Scheduling.Buffer),
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,
                     Draw_Scheduling.High_Margin +
                     I * Draw_Scheduling.Interligne,
                     Draw_Scheduling.Left_Margin +
                     Gint (Draw_Scheduling.Period -
                           Draw_Scheduling.Start_Draw) *
                     Draw_Scheduling.Timeunit_Margin,
                     Draw_Scheduling.High_Margin +
                     I * Draw_Scheduling.Interligne);

                  -- For each line, draw unit time lines
                  --
                  for J in  0 .. Gint (Draw_Scheduling.Period -
                                       Draw_Scheduling.Start_Draw)
                  loop
                     Draw_Line
                       (Get_Pixmap (Draw_Scheduling.Buffer),
                        Draw_Scheduling.Black_Gc,
                        Draw_Scheduling.Left_Margin +
                        J * Draw_Scheduling.Timeunit_Margin,
                        Draw_Scheduling.High_Margin +
                        I * Draw_Scheduling.Interligne -
                        3,
                        Draw_Scheduling.Left_Margin +
                        J * Draw_Scheduling.Timeunit_Margin,
                        Draw_Scheduling.High_Margin +
                        I * Draw_Scheduling.Interligne +
                        3);
                  end loop;

                  -- Register each resource in the name table
                  --
                  Resource_Name (Object_Range (I))  := A_Resource.Name;

                  I := I + 1;
               end if;

               exit when Is_Last_Element
                            (Draw_Scheduling.Resources,
                             My_Resource_Iterator);
               Next_Element (Draw_Scheduling.Resources, My_Resource_Iterator);

            end loop;
         end if;
      end if;

      -------------------------------------
      -- Draw a line for each message and buffer
      -------------------------------------

      if Schedule_With_Precedencies then

         if not Is_Empty (Draw_Scheduling.Messages) then

            Reset_Iterator (Draw_Scheduling.Messages, My_Message_Iterator);

            loop

               Current_Element
                 (Draw_Scheduling.Messages,
                  A_Message,
                  My_Message_Iterator);

               if To_Display (A_Message.Name, Message_Object_Type) then

                  Msg := Lb_Message_Name (Current_Language) &
                         "=" &
                         A_Message.Name &
                         "        ";

                  if A_Message.Message_Type = Periodic_Type then
                     Msg := Msg &
                            To_Unbounded_String ("Pe=") &
                            Periodic_Message_Ptr (A_Message).Period'Img;
                  else
                     Msg := Msg & To_Unbounded_String (" ;Pe= 0");
                  end if;

                  Msg := Msg &
                         To_Unbounded_String (" ; Delay=") &
                         A_Message.Response_Time'Img;
                  Msg := Msg &
                         To_Unbounded_String (" ; Size = ") &
                         A_Message.Size'Img;

                  Draw_Text
                    (Get_Pixmap (Draw_Scheduling.Buffer),
                     Draw_Scheduling.Font,
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,
                     Draw_Scheduling.High_Margin +
                     I * Draw_Scheduling.Interligne +
                     20,
                     To_String (Msg));

                  Draw_Line
                    (Get_Pixmap (Draw_Scheduling.Buffer),
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,
                     Draw_Scheduling.High_Margin +
                     I * Draw_Scheduling.Interligne,
                     Draw_Scheduling.Left_Margin +
                     Gint (Draw_Scheduling.Period -
                           Draw_Scheduling.Start_Draw) *
                     Draw_Scheduling.Timeunit_Margin,
                     Draw_Scheduling.High_Margin +
                     I * Draw_Scheduling.Interligne);

                  -- For each line, draw unit time lines
                  --
                  for J in  0 .. Gint (Draw_Scheduling.Period -
                                       Draw_Scheduling.Start_Draw)
                  loop
                     Draw_Line
                       (Get_Pixmap (Draw_Scheduling.Buffer),
                        Draw_Scheduling.Black_Gc,
                        Draw_Scheduling.Left_Margin +
                        J * Draw_Scheduling.Timeunit_Margin,
                        Draw_Scheduling.High_Margin +
                        I * Draw_Scheduling.Interligne -
                        3,
                        Draw_Scheduling.Left_Margin +
                        J * Draw_Scheduling.Timeunit_Margin,
                        Draw_Scheduling.High_Margin +
                        I * Draw_Scheduling.Interligne +
                        3);
                  end loop;

                  -- Register each message in the name table
                  --
                  Message_Name (Object_Range (I))  := A_Message.Name;

                  I := I + 1;
               end if;

               exit when Is_Last_Element
                            (Draw_Scheduling.Messages,
                             My_Message_Iterator);
               Next_Element (Draw_Scheduling.Messages, My_Message_Iterator);

            end loop;
         end if;

         if not Is_Empty (Draw_Scheduling.Buffers) then

            Reset_Iterator (Draw_Scheduling.Buffers, My_Buffer_Iterator);

            loop

               Current_Element
                 (Draw_Scheduling.Buffers,
                  A_Buffer,
                  My_Buffer_Iterator);

               if To_Display (A_Buffer.Name, Buffer_Object_Type) then

                  Msg := Lb_Buffer_Name (Current_Language) &
                         "=" &
                         A_Buffer.Name &
                         "        ";

                  Msg := Msg &
                         To_Unbounded_String ("Size = ") &
                         A_Buffer.Size'Img;

                  Msg := Msg &
                         To_Unbounded_String (" ; Cpu = ") &
                         A_Buffer.Cpu_Name;

                  Draw_Text
                    (Get_Pixmap (Draw_Scheduling.Buffer),
                     Draw_Scheduling.Font,
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,
                     Draw_Scheduling.High_Margin +
                     I * Draw_Scheduling.Interligne +
                     20,
                     To_String (Msg));

                  Draw_Line
                    (Get_Pixmap (Draw_Scheduling.Buffer),
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,
                     Draw_Scheduling.High_Margin +
                     I * Draw_Scheduling.Interligne,
                     Draw_Scheduling.Left_Margin +
                     Gint (Draw_Scheduling.Period -
                           Draw_Scheduling.Start_Draw) *
                     Draw_Scheduling.Timeunit_Margin,
                     Draw_Scheduling.High_Margin +
                     I * Draw_Scheduling.Interligne);

                  -- For each line, draw unit time lines
                  --
                  for J in  0 .. Gint (Draw_Scheduling.Period -
                                       Draw_Scheduling.Start_Draw)
                  loop
                     Draw_Line
                       (Get_Pixmap (Draw_Scheduling.Buffer),
                        Draw_Scheduling.Black_Gc,
                        Draw_Scheduling.Left_Margin +
                        J * Draw_Scheduling.Timeunit_Margin,
                        Draw_Scheduling.High_Margin +
                        I * Draw_Scheduling.Interligne -
                        3,
                        Draw_Scheduling.Left_Margin +
                        J * Draw_Scheduling.Timeunit_Margin,
                        Draw_Scheduling.High_Margin +
                        I * Draw_Scheduling.Interligne +
                        3);
                  end loop;

                  -- Register each buffer in the name table
                  --
                  Buffer_Name (Object_Range (I))  := A_Buffer.Name;

                  I := I + 1;
               end if;

               exit when Is_Last_Element
                            (Draw_Scheduling.Buffers,
                             My_Buffer_Iterator);
               Next_Element (Draw_Scheduling.Buffers, My_Buffer_Iterator);

            end loop;
         end if;

      end if;

      ------------------------------------------
      -- Draw information from scheduling table
      ------------------------------------------

      for U in  0 .. Draw_Scheduling.Sched.Nb_Entries - 1 loop -- processors
         if Draw_Scheduling.Sched.Entries (U).Data.Error_Msg =
            Empty_String
         then

            -- Scan scheduling sequence, but only for period-start-draw units
            --of times
            --
            for K in  0 .. Draw_Scheduling.Sched.Entries (U).Data.Result.
                 Nb_Entries -
                           1
            loop

               if (Draw_Scheduling.Sched.Entries (U).Data.Result.Entries (K).
                    Data >=
                   Draw_Scheduling.Start_Draw) and
                  (Draw_Scheduling.Sched.Entries (U).Data.Result.Entries (K).
                    Data <=
                   Draw_Scheduling.Period)
               then

                  case 
                       Draw_Scheduling.Sched.Entries (U).Data.Result.Entries (
                       K).Item.Event_Type is

                     ----------------------------------------
                     -- Draw task running time in black
                     -- (time when the task owns the processor)
                     ----------------------------------------
                     when Running_Task =>

                        Name :=
                          Draw_Scheduling.Sched.Entries (U).Data.Result.
                          Entries (K).Item.Running_Task.Name;

                        if To_Display (Name, Task_Object_Type) then
                           Ok := False;
                           for L in  Object_Range loop
                              if (Task_Name (L) = Name) then
                                 Ok      := True;
                                 Task_Id := Gint (L);
                                 Time_U  :=
                                   Gint (
                                   Draw_Scheduling.Sched.Entries (U).Data.
                                   Result.Entries (K).Data -
                                         Draw_Scheduling.Start_Draw);
                              end if;
                           end loop;

                           if not Ok then
                              raise Task_Not_Found;
                           end if;

                           Gdk.GC.Set_Foreground
                             (Draw_Scheduling.Black_Gc,
                              Gdk.Color.Black
                                 (Gtk.Widget.Get_Default_Colormap));

                           Draw_Rectangle
                             (Get_Pixmap (Draw_Scheduling.Buffer),
                              Draw_Scheduling.Black_Gc,
                              True,
                              Draw_Scheduling.Timeunit_Margin * Time_U +
                              Draw_Scheduling.Left_Margin,
                              Task_Id * Draw_Scheduling.Interligne +
                              Draw_Scheduling.High_Margin -
                              3,
                              Draw_Scheduling.Timeunit_Margin,
                              7);
                        end if;

                     -----------------------------------------------------
                     -- Draw each task wake up time in red
                     -----------------------------------------------------
                     when Task_Activation =>

                        Name :=
                          Draw_Scheduling.Sched.Entries (U).Data.Result.
                          Entries (K).Item.Activation_Task.Name;

                        if To_Display (Name, Task_Object_Type) then
                           Ok := False;
                           for L in  Object_Range loop
                              if (Task_Name (L) = Name) then
                                 Ok      := True;
                                 Task_Id := Gint (L);
                                 Time_U  :=
                                   Gint (
                                   Draw_Scheduling.Sched.Entries (U).Data.
                                   Result.Entries (K).Data -
                                         Draw_Scheduling.Start_Draw);
                              end if;
                           end loop;

                           if not Ok then
                              raise Task_Not_Found;
                           end if;

                           Gdk.GC.Set_Foreground
                             (Draw_Scheduling.Black_Gc,
                              Draw_Scheduling.Red_Color);

                           Draw_Rectangle
                             (Get_Pixmap (Draw_Scheduling.Buffer),
                              Draw_Scheduling.Black_Gc,
                              True,
                              Draw_Scheduling.Left_Margin +
                              Draw_Scheduling.Timeunit_Margin * Time_U,
                              Draw_Scheduling.High_Margin +
                              Task_Id * Draw_Scheduling.Interligne -
                              8,
                              3,
                              16);

                        end if;

                     -----------------------------------------------------
                     -- Draw shared resource allocation in blue color
                     -----------------------------------------------------
                     when Allocate_Resource =>

                        Name :=
                          Draw_Scheduling.Sched.Entries (U).Data.Result.
                          Entries (K).Item.Allocate_Resource.Name;

                        if To_Display (Name, Resource_Object_Type) then

                           Ok := False;
                           for L in  Object_Range loop
                              if (Resource_Name (L) = Name) then
                                 Ok      := True;
                                 Task_Id := Gint (L);
                                 Time_U  :=
                                   Gint (
                                   Draw_Scheduling.Sched.Entries (U).Data.
                                   Result.Entries (K).Data -
                                         Draw_Scheduling.Start_Draw);
                              end if;
                           end loop;

                           if not Ok then
                              raise Resource_Not_Found;
                           end if;

                           Gdk.GC.Set_Foreground
                             (Draw_Scheduling.Black_Gc,
                              Draw_Scheduling.Blue_Color);

                           Draw_Rectangle
                             (Get_Pixmap (Draw_Scheduling.Buffer),
                              Draw_Scheduling.Black_Gc,
                              True,
                              Draw_Scheduling.Left_Margin +
                              Draw_Scheduling.Timeunit_Margin * Time_U,
                              Draw_Scheduling.High_Margin +
                              Task_Id * Draw_Scheduling.Interligne -
                              8,
                              3,
                              16);

                           Save_Start_U (Object_Range (Task_Id))  :=
                             Draw_Scheduling.Left_Margin +
                             Draw_Scheduling.Timeunit_Margin * Time_U;

                        end if;

                     ----------------------------------------
                     -- Draw shared resource releasing in black
                     ----------------------------------------
                     when Release_Resource =>

                        Name :=
                          Draw_Scheduling.Sched.Entries (U).Data.Result.
                          Entries (K).Item.Release_Resource.Name;

                        if To_Display (Name, Resource_Object_Type) then
                           Ok := False;
                           for L in  Object_Range loop
                              if (Resource_Name (L) = Name) then
                                 Ok      := True;
                                 Task_Id := Gint (L);
                                 Time_U  :=
                                   Gint (
                                   Draw_Scheduling.Sched.Entries (U).Data.
                                   Result.Entries (K).Data -
                                         Draw_Scheduling.Start_Draw) +
                                   1;
                              end if;
                           end loop;

                           if not Ok then
                              raise Task_Not_Found;
                           end if;

                           Gdk.GC.Set_Foreground
                             (Draw_Scheduling.Black_Gc,
                              Gdk.Color.Black
                                 (Gtk.Widget.Get_Default_Colormap));

                           Draw_Rectangle
                             (Get_Pixmap (Draw_Scheduling.Buffer),
                              Draw_Scheduling.Black_Gc,
                              True,
                              Save_Start_U (Object_Range (Task_Id)),
                              Task_Id * Draw_Scheduling.Interligne +
                              Draw_Scheduling.High_Margin -
                              3,
                              Draw_Scheduling.Left_Margin +
                              Draw_Scheduling.Timeunit_Margin * Time_U -
                              Save_Start_U (Object_Range (Task_Id)),
                              7);
                        end if;
                     ----------------------------------------------------
                     -- Event about buffers
                     -- Draw in red buffer write and in blue buffer read
                     ----------------------------------------------------

                     when Write_To_Buffer =>

                        Name :=
                          Draw_Scheduling.Sched.Entries (U).Data.Result.
                          Entries (K).Item.Write_Buffer.Name;

                        if To_Display (Name, Buffer_Object_Type) then

                           Ok := False;
                           for L in  Object_Range loop
                              if (Buffer_Name (L) = Name) then
                                 Ok        := True;
                                 Buffer_Id := Gint (L);
                                 Time_U    :=
                                   Gint (
                                   Draw_Scheduling.Sched.Entries (U).Data.
                                   Result.Entries (K).Data);
                              end if;
                           end loop;
                           if not Ok then
                              raise Buffer_Not_Found;
                           end if;

                           Gdk.GC.Set_Foreground
                             (Draw_Scheduling.Black_Gc,
                              Draw_Scheduling.Blue_Color);

                           Draw_Rectangle
                             (Get_Pixmap (Draw_Scheduling.Buffer),
                              Draw_Scheduling.Black_Gc,
                              True,
                              Draw_Scheduling.Timeunit_Margin * Time_U +
                              Draw_Scheduling.Left_Margin,
                              Buffer_Id * Draw_Scheduling.Interligne +
                              Draw_Scheduling.High_Margin -
                              3,
                              Draw_Scheduling.Timeunit_Margin,
                              7);

                        end if;

                     when Read_From_Buffer =>

                        Name :=
                          Draw_Scheduling.Sched.Entries (U).Data.Result.
                          Entries (K).Item.Read_Buffer.Name;

                        if To_Display (Name, Buffer_Object_Type) then

                           Ok := False;
                           for L in  Object_Range loop
                              if (Buffer_Name (L) = Name) then
                                 Ok        := True;
                                 Buffer_Id := Gint (L);
                                 Time_U    :=
                                   Gint (
                                   Draw_Scheduling.Sched.Entries (U).Data.
                                   Result.Entries (K).Data -
                                         Draw_Scheduling.Start_Draw);
                              end if;
                           end loop;

                           if not Ok then
                              raise Buffer_Not_Found;
                           end if;

                           Gdk.GC.Set_Foreground
                             (Draw_Scheduling.Black_Gc,
                              Draw_Scheduling.Red_Color);

                           Draw_Rectangle
                             (Get_Pixmap (Draw_Scheduling.Buffer),
                              Draw_Scheduling.Black_Gc,
                              True,
                              Draw_Scheduling.Timeunit_Margin * Time_U +
                              Draw_Scheduling.Left_Margin,
                              Buffer_Id * Draw_Scheduling.Interligne +
                              Draw_Scheduling.High_Margin -
                              3,
                              Draw_Scheduling.Timeunit_Margin,
                              7);

                        end if;

                     ----------------------------------------
                     -- Event about messages
                     -- Draw in red message communication time
                     ----------------------------------------

                     when Send_Message =>

                        Name :=
                          Draw_Scheduling.Sched.Entries (U).Data.Result.
                          Entries (K).Item.Send_Message.Name;

                        if To_Display (Name, Message_Object_Type) then

                           Ok := False;
                           for L in  Object_Range loop
                              if (Message_Name (L) = Name) then
                                 Ok         := True;
                                 Message_Id := Gint (L);
                                 Time_U     :=
                                   Gint (
                                   Draw_Scheduling.Sched.Entries (U).Data.
                                   Result.Entries (K).Data);
                              end if;
                           end loop;
                           if not Ok then
                              raise Message_Not_Found;
                           end if;

                           -- Save the time the message is sent
                           --
                           Save_Start_M (Object_Range (Message_Id))  :=
                             Time_U;
                        end if;

                     when Receive_Message =>

                        Name :=
                          Draw_Scheduling.Sched.Entries (U).Data.Result.
                          Entries (K).Item.Receive_Message.Name;

                        if To_Display (Name, Message_Object_Type) then

                           Ok := False;
                           for L in  Object_Range loop
                              if (Message_Name (L) = Name) then
                                 Ok         := True;
                                 Message_Id := Gint (L);
                                 Time_U     :=
                                   Gint (
                                   Draw_Scheduling.Sched.Entries (U).Data.
                                   Result.Entries (K).Data -
                                         Draw_Scheduling.Start_Draw);
                              end if;
                           end loop;

                           if not Ok then
                              raise Message_Not_Found;
                           end if;
                           Gdk.GC.Set_Foreground
                             (Draw_Scheduling.Black_Gc,
                              Draw_Scheduling.Red_Color);

                           Draw_Rectangle
                             (Get_Pixmap (Draw_Scheduling.Buffer),
                              Draw_Scheduling.Black_Gc,
                              True,

                              Draw_Scheduling.Left_Margin +
                              Save_Start_M (Object_Range (Message_Id)) *
                              Draw_Scheduling.Timeunit_Margin,

                              Message_Id * Draw_Scheduling.Interligne +
                              Draw_Scheduling.High_Margin -
                              3,

                              Draw_Scheduling.Timeunit_Margin *
                              (Time_U -
                               Save_Start_M (Object_Range (Message_Id))),
                              7);

                        end if;

                     -------------------------------------------
                     -- Events we have nothing to display
                     -------------------------------------------
                     when others =>
                        null;

                  end case;
               end if;

            end loop;
         end if;
      end loop;

      Gdk.GC.Set_Foreground
        (Draw_Scheduling.Black_Gc,
         Gdk.Color.Black (Gtk.Widget.Get_Default_Colormap));

      ---------------------
      -- Flush data in the window
      ---------------------
      Draw (Draw_Scheduling.Buffer);

   end Redraw_Time_Line;

end Draw_Scheduling_Pkg;
