------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Processors;            use Processors;
with Processor_Set;         use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Gtk.Radio_Menu_Item;   use Gtk.Radio_Menu_Item;
with Project;               use Project;
with Glib;                  use Glib;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Unbounded_Strings;     use Unbounded_Strings;
use Unbounded_Strings.Strings_Table_Package;
with Call_Framework;        use Call_Framework;
use Call_Framework.Framework_Request_Package;
use Call_Framework.Framework_Response_Package;
with Parameters;            use Parameters;
with Parameters.extended;   use Parameters.extended;
use Parameters.extended.Framework_Parameters_Table_Package;
with Text_Tools;            use Text_Tools;
with Task_Set;              use Task_Set;
use Task_Set.Generic_Task_Set;

package body Event_Table_Analysis_Pkg.Callbacks is

   procedure On_Cancel_Pressed (Object : access Gtk_Button_Record'Class) is
   begin
      Destroy (Event_Table_Analysis);
   end On_Cancel_Pressed;

   -- Now, call scheduling analysis for each processor
   --
   procedure Call_Processor_Analysis (Name : Unbounded_String) is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_ptr;

   begin
      Initialize (Response_List);
      Initialize (Request_List);

      if Preemption_From_Simulation then
         Initialize (A_Request);
         A_Request.Target    := Name;
         A_Request.Statement := Scheduling_Simulation_Preemption_Number;
         Add (Request_List, A_Request);
      end if;

      if Context_Switch_From_Simulation then
         Initialize (A_Request);
         A_Request.Target    := Name;
         A_Request.Statement := Scheduling_Simulation_Context_Switch_Number;
         Add (Request_List, A_Request);
      end if;

      if (Response_Time_From_Simulation (Worst_Case) or
          Response_Time_From_Simulation (Best_Case) or
          Response_Time_From_Simulation (Average_Case) or
          Response_Time_From_Simulation (Plot_Case))
      then

         Initialize (A_Request);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("worst_case");
         if Response_Time_From_Simulation (Worst_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("best_case");
         if Response_Time_From_Simulation (Best_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("average_case");
         if Response_Time_From_Simulation (Average_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Request.Target := Name;

         A_Request.Statement := Scheduling_Simulation_Response_Time;
         Add (Request_List, A_Request);
      end if;

      if (Blocking_Time_From_Simulation (Worst_Case) or
          Blocking_Time_From_Simulation (Best_Case) or
          Blocking_Time_From_Simulation (Average_Case) or
          Blocking_Time_From_Simulation (Plot_Case))
      then

         Initialize (A_Request);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("worst_case");
         if Blocking_Time_From_Simulation (Worst_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("best_case");
         if Blocking_Time_From_Simulation (Best_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("average_case");
         if Blocking_Time_From_Simulation (Average_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Request.Target := Name;

         A_Request.Statement := Scheduling_Simulation_Blocking_Time;
         Add (Request_List, A_Request);
      end if;

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title
        (Event_Table_Analysis.Text,
         Response_List.Entries (0).Title);
      for J in  0 .. Response_List.Nb_Entries - 1 loop
         Write_Text
           (Event_Table_Analysis.Text,
            Response_List.Entries (J).Text);
      end loop;

   end Call_Processor_Analysis;

   procedure On_Close_Pressed (Object : access Gtk_Button_Record'Class) is

      A_Processor           : Processor_Ptr;
      My_Processor_Iterator : Processors_Iterator;

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

      Cpu_Index : Gint;

   begin

      -- Get customization provided by the user
      --
      Context_Switch_From_Simulation     :=
         Get_Active (Event_Table_Analysis.Context_Switch);
      Preemption_From_Simulation         :=
         Get_Active (Event_Table_Analysis.Number_Preemption);
      Run_Event_Analyzer_From_Simulation :=
         Get_Active (Event_Table_Analysis.Event_Analyzer);
      Schedule_All_Processeurs           :=
         Get_Active (Event_Table_Analysis.Schedule_All_Processeurs);

      for I in  Data_From_Simulation_Table'Range loop
         Blocking_Time_From_Simulation (I) :=
            Get_Active (Event_Table_Analysis.Blocking_Time (I));
         Response_Time_From_Simulation (I) :=
            Get_Active (Event_Table_Analysis.Response_Time (I));
      end loop;

      Cpu_Index :=
         Gint'Val
           (Gint (Get_Number_Of_Elements (Sys.Processors)) -
            Gint (Selected_Button
                     (Event_Table_Analysis.Clist_Processor_Menu)));
      Cpu_Index := Cpu_Index - 1;

      Get_Element_Number
        (Sys.Processors,
         A_Processor,
         Processors_Range (Cpu_Index));
      Last_Selected_Processor := A_Processor.Name;

      -- Call the framework now
      --
      Initialize (A_Request);

      if Schedule_All_Processeurs = False then
         Call_Processor_Analysis (Last_Selected_Processor);
      else
         Reset_Iterator (Sys.Processors, My_Processor_Iterator);
         loop
            Current_Element
              (Sys.Processors,
               A_Processor,
               My_Processor_Iterator);

            Call_Processor_Analysis (A_Processor.Name);

            exit when Is_Last_Element (Sys.Processors, My_Processor_Iterator);
            Next_Element (Sys.Processors, My_Processor_Iterator);
         end loop;

      end if;

      -- Now run, event analyzers,
      --

      if Run_Event_Analyzer_From_Simulation then

         Initialize (Response_List);
         Initialize (Request_List);

         Initialize (A_Request);
         A_Request.Statement := Scheduling_Simulation_Run_Event_Handler;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Event_Table_Analysis.Text, Response_List);

      end if;

      Destroy (Event_Table_Analysis);

   end On_Close_Pressed;

end Event_Table_Analysis_Pkg.Callbacks;
