------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Glib;                               use Glib;
with Gdk.Types;                          use Gdk.Types;
with Gtk.Enums;                          use Gtk.Enums;
with Callbacks_Cheddar;                  use Callbacks_Cheddar;
with Event_Table_Analysis_Pkg.Callbacks;
use Event_Table_Analysis_Pkg.Callbacks;
with Framework_Config;                   use Framework_Config;
with Translate;                          use Translate;
with Processors;                         use Processors;
with Processor_Set;                      use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Mmi_Tools;                          use Mmi_Tools;
with Ada.Strings.Unbounded;              use Ada.Strings.Unbounded;
with Unbounded_Strings;                  use Unbounded_Strings;
use Unbounded_Strings.Strings_Table_Package;

package body Event_Table_Analysis_Pkg is

   procedure Gtk_New
     (Event_Table_Analysis : out Event_Table_Analysis_Access;
      Text                 : Gtk_Text;
      Sys                  : in System)
   is

   begin
      Event_Table_Analysis := new Event_Table_Analysis_Record;
      Event_Table_Analysis_Pkg.Initialize (Event_Table_Analysis, Text, Sys);

   end Gtk_New;

   procedure Initialize
     (Event_Table_Analysis : access Event_Table_Analysis_Record'Class;
      Text                 : Gtk_Text;
      Sys                  : in System)
   is

      Processor_Iterator : Processors_Iterator;
      Processor_Items    : Strings_Table;
      A_Processor        : Processor_Ptr;

   begin

      Event_Table_Analysis.Text := Text;

      Gtk.Window.Initialize (Event_Table_Analysis, Window_Toplevel);
      Set_Title
        (Event_Table_Analysis,
         To_String (Lb_Scheduling_Options (Current_Language)));
      Set_Policy (Event_Table_Analysis, True, True, True);
      Set_Position (Event_Table_Analysis, Win_Pos_Center_Always);
      Set_Modal (Event_Table_Analysis, True);
      Set_Extension_Events (Event_Table_Analysis, Extension_Events_All);
      Set_Flags (Event_Table_Analysis, Can_Default);
      Grab_Focus (Event_Table_Analysis);

      Gtk_New_Vbox (Event_Table_Analysis.Event_Table_Analysis_Vbox);
      Add
        (Event_Table_Analysis,
         Event_Table_Analysis.Event_Table_Analysis_Vbox);

      -- Table of options choosen by the users before simulation
      --
      Gtk_New (Event_Table_Analysis.Value_Table, 2, 2, True);
      Set_Border_Width (Event_Table_Analysis.Value_Table, 14);
      Set_Row_Spacings (Event_Table_Analysis.Value_Table, 8);
      Set_Col_Spacings (Event_Table_Analysis.Value_Table, 0);
      Pack_Start
        (Event_Table_Analysis.Event_Table_Analysis_Vbox,
         Event_Table_Analysis.Value_Table);

      ----------------------------------------------------------------
      -- List of processors on which simulations have to be run
      ----------------------------------------------------------------

      Initialize (Processor_Items);
      Reset_Iterator (Sys.Processors, Processor_Iterator);

      loop
         Current_Element (Sys.Processors, A_Processor, Processor_Iterator);
         Add (Processor_Items, A_Processor.Name);

         exit when Is_Last_Element (Sys.Processors, Processor_Iterator);

         Next_Element (Sys.Processors, Processor_Iterator);
      end loop;

      Event_Table_Analysis.Clist_Processor_Menu := Widget_SList.Null_List;
      Build_Option_Menu
        (Event_Table_Analysis.Processor_Menu,
         Event_Table_Analysis.Clist_Processor_Menu,
         Processor_Items,
         -1,
         null);

      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Processor_Menu,
         3,
         4,
         0,
         1);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String (Lb_Processor (Current_Language) & Lb_Colon));

      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         2,
         3,
         0,
         1);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String
            (Lb_Run_Simulation_On_One_Processor_Only (Current_Language) &
             Lb_Colon));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         0,
         1,
         0,
         1);

      Gtk_New (Event_Table_Analysis.Schedule_All_Processeurs, "");
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Schedule_All_Processeurs,
         1,
         2,
         0,
         1);

      Gtk_New_Hseparator (Event_Table_Analysis.Hseparator);
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Hseparator,
         0,
         5,
         1,
         2);

      ----------------------------------------------------------------
      -- Event analyzers, number of context switch or preemption
      ----------------------------------------------------------------

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String
            (Lb_Number_Of_Context_Switch (Current_Language) & Lb_Colon));

      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         0,
         1,
         2,
         3);

      Gtk_New (Event_Table_Analysis.Context_Switch, "");
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Context_Switch,
         1,
         2,
         2,
         3);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String (Lb_Number_Of_Preemption (Current_Language) & Lb_Colon));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         0,
         1,
         3,
         4);

      Gtk_New (Event_Table_Analysis.Number_Preemption, "");
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Number_Preemption,
         1,
         2,
         3,
         4);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String (Lb_Run_Event_Analyzers (Current_Language) & Lb_Colon));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         0,
         1,
         4,
         5);

      Gtk_New (Event_Table_Analysis.Event_Analyzer, "");
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Event_Analyzer,
         1,
         2,
         4,
         5);

      ----------------------------------------------------------------
      -- Response time and blocking time
      ----------------------------------------------------------------

      Gtk_New_Hseparator (Event_Table_Analysis.Hseparator);
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Hseparator,
         0,
         5,
         6,
         7);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String (Lb_Worst_Case (Current_Language)));
      Set_Alignment (Event_Table_Analysis.Label, 0.5, 0.5);
      Set_Padding (Event_Table_Analysis.Label, 0, 0);
      Set_Justify (Event_Table_Analysis.Label, Justify_Left);
      Set_Line_Wrap (Event_Table_Analysis.Label, False);
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         1,
         2,
         7,
         8);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String (Lb_Best_Case (Current_Language)));
      Set_Alignment (Event_Table_Analysis.Label, 0.5, 0.5);
      Set_Padding (Event_Table_Analysis.Label, 0, 0);
      Set_Justify (Event_Table_Analysis.Label, Justify_Left);
      Set_Line_Wrap (Event_Table_Analysis.Label, False);
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         2,
         3,
         7,
         8);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String (Lb_Average_Case (Current_Language)));
      Set_Alignment (Event_Table_Analysis.Label, 0.5, 0.5);
      Set_Padding (Event_Table_Analysis.Label, 0, 0);
      Set_Justify (Event_Table_Analysis.Label, Justify_Left);
      Set_Line_Wrap (Event_Table_Analysis.Label, False);
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         3,
         4,
         7,
         8);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String (Lb_Curve (Current_Language)));
      Set_Alignment (Event_Table_Analysis.Label, 0.5, 0.5);
      Set_Padding (Event_Table_Analysis.Label, 0, 0);
      Set_Justify (Event_Table_Analysis.Label, Justify_Left);
      Set_Line_Wrap (Event_Table_Analysis.Label, False);
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         4,
         5,
         7,
         8);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String
            (Lb_Task_Response_Time_From_Simulation (Current_Language) &
             Lb_Colon));
      Set_Alignment (Event_Table_Analysis.Label, 0.5, 0.5);
      Set_Padding (Event_Table_Analysis.Label, 0, 0);
      Set_Justify (Event_Table_Analysis.Label, Justify_Left);
      Set_Line_Wrap (Event_Table_Analysis.Label, False);
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         0,
         1,
         9,
         10);

      Gtk_New
        (Event_Table_Analysis.Label,
         To_String (Lb_Blocking_Time (Current_Language) & Lb_Colon));
      Set_Alignment (Event_Table_Analysis.Label, 0.5, 0.5);
      Set_Padding (Event_Table_Analysis.Label, 0, 0);
      Set_Justify (Event_Table_Analysis.Label, Justify_Left);
      Set_Line_Wrap (Event_Table_Analysis.Label, False);
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Label,
         0,
         1,
         10,
         11);

      -- A table of Check buttons  (worst/best/average/curve cases)
      --

      Gtk_New
        (Event_Table_Analysis.Align,
         Xalign => 0.5,
         Yalign => 0.5,
         Xscale => 0.1,
         Yscale => 0.1);
      Gtk_New (Event_Table_Analysis.Response_Time (Worst_Case), "");
      Add
        (Event_Table_Analysis.Align,
         Event_Table_Analysis.Response_Time (Worst_Case));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Align,
         1,
         2,
         9,
         10);
      Gtk_New
        (Event_Table_Analysis.Align,
         Xalign => 0.5,
         Yalign => 0.5,
         Xscale => 0.1,
         Yscale => 0.1);
      Gtk_New (Event_Table_Analysis.Response_Time (Best_Case), "");
      Add
        (Event_Table_Analysis.Align,
         Event_Table_Analysis.Response_Time (Best_Case));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Align,
         2,
         3,
         9,
         10);
      Gtk_New
        (Event_Table_Analysis.Align,
         Xalign => 0.5,
         Yalign => 0.5,
         Xscale => 0.1,
         Yscale => 0.1);
      Gtk_New (Event_Table_Analysis.Response_Time (Average_Case), "");
      Add
        (Event_Table_Analysis.Align,
         Event_Table_Analysis.Response_Time (Average_Case));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Align,
         3,
         4,
         9,
         10);
      Gtk_New
        (Event_Table_Analysis.Align,
         Xalign => 0.5,
         Yalign => 0.5,
         Xscale => 0.1,
         Yscale => 0.1);
      Gtk_New (Event_Table_Analysis.Response_Time (Plot_Case), "");
      Add
        (Event_Table_Analysis.Align,
         Event_Table_Analysis.Response_Time (Plot_Case));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Align,
         4,
         5,
         9,
         10);

      Gtk_New
        (Event_Table_Analysis.Align,
         Xalign => 0.5,
         Yalign => 0.5,
         Xscale => 0.1,
         Yscale => 0.1);
      Gtk_New (Event_Table_Analysis.Blocking_Time (Worst_Case), "");
      Add
        (Event_Table_Analysis.Align,
         Event_Table_Analysis.Blocking_Time (Worst_Case));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Align,
         1,
         2,
         10,
         11);
      Gtk_New
        (Event_Table_Analysis.Align,
         Xalign => 0.5,
         Yalign => 0.5,
         Xscale => 0.1,
         Yscale => 0.1);
      Gtk_New (Event_Table_Analysis.Blocking_Time (Best_Case), "");
      Add
        (Event_Table_Analysis.Align,
         Event_Table_Analysis.Blocking_Time (Best_Case));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Align,
         2,
         3,
         10,
         11);
      Gtk_New
        (Event_Table_Analysis.Align,
         Xalign => 0.5,
         Yalign => 0.5,
         Xscale => 0.1,
         Yscale => 0.1);
      Gtk_New (Event_Table_Analysis.Blocking_Time (Average_Case), "");
      Add
        (Event_Table_Analysis.Align,
         Event_Table_Analysis.Blocking_Time (Average_Case));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Align,
         3,
         4,
         10,
         11);
      Gtk_New
        (Event_Table_Analysis.Align,
         Xalign => 0.5,
         Yalign => 0.5,
         Xscale => 0.1,
         Yscale => 0.1);
      Gtk_New (Event_Table_Analysis.Blocking_Time (Plot_Case), "");
      Add
        (Event_Table_Analysis.Align,
         Event_Table_Analysis.Blocking_Time (Plot_Case));
      Attach
        (Event_Table_Analysis.Value_Table,
         Event_Table_Analysis.Align,
         4,
         5,
         10,
         11);

      ----------------------------------------------
      -- Initiliaze the widet of the window
      ----------------------------------------------

      Set_Active
        (Event_Table_Analysis.Context_Switch,
         Context_Switch_From_Simulation);
      Set_Active
        (Event_Table_Analysis.Number_Preemption,
         Preemption_From_Simulation);
      Set_Active
        (Event_Table_Analysis.Event_Analyzer,
         Run_Event_Analyzer_From_Simulation);
      Set_Active
        (Event_Table_Analysis.Schedule_All_Processeurs,
         Schedule_All_Processeurs);

      for I in  Data_From_Simulation'Range loop
         Set_Active
           (Event_Table_Analysis.Response_Time (I),
            Response_Time_From_Simulation (I));
         Set_Active
           (Event_Table_Analysis.Blocking_Time (I),
            Blocking_Time_From_Simulation (I));
      end loop;

      ---------------------------------------------------------
      -- The bottom part of the widget (ok and cancel buttons)
      ---------------------------------------------------------

      Gtk_New_Hseparator (Event_Table_Analysis.Hseparator);
      Pack_Start
        (Event_Table_Analysis.Event_Table_Analysis_Vbox,
         Event_Table_Analysis.Hseparator);

      Gtk_New_Hbox (Event_Table_Analysis.Button_Hbox);
      Pack_Start
        (Event_Table_Analysis.Event_Table_Analysis_Vbox,
         Event_Table_Analysis.Button_Hbox,
         False,
         True,
         5);

      Gtk_New (Event_Table_Analysis.Close_Button, "Ok");
      Add
        (Event_Table_Analysis.Button_Hbox,
         Event_Table_Analysis.Close_Button);
      Button_Callback.Connect
        (Event_Table_Analysis.Close_Button,
         "clicked",
         Button_Callback.To_Marshaller (On_Close_Pressed'Access));

      Gtk_New
        (Event_Table_Analysis.Cancel_Button,
         To_String (Lb_Cancel (Current_Language)));
      Add
        (Event_Table_Analysis.Button_Hbox,
         Event_Table_Analysis.Cancel_Button);
      Button_Callback.Connect
        (Event_Table_Analysis.Cancel_Button,
         "clicked",
         Button_Callback.To_Marshaller (On_Cancel_Pressed'Access));

   end Initialize;

end Event_Table_Analysis_Pkg;
