------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Strings.Unbounded;     use Ada.Strings.Unbounded;
with Call_Framework;            use Call_Framework;
use Call_Framework.Framework_Request_Package;
use Call_Framework.Framework_Response_Package;
with Parameters;                use Parameters;
with Parameters.extended;       use Parameters.extended;
use Parameters.extended.Framework_Parameters_Table_Package;
with Call_Scheduling_Framework; use Call_Scheduling_Framework;
with Multiprocessor_Services;   use Multiprocessor_Services;
use Multiprocessor_Services.Scheduling_Result_Per_Processor;
with Framework_Config;          use Framework_Config;
with Strings;                   use Strings;
with Project;                   use Project;
with Unbounded_Strings;         use Unbounded_Strings;
with Mmi_Tools;                 use Mmi_Tools;
with Translate;                 use Translate;
with Text_Tools;                use Text_Tools;
with Processors;                use Processors;
with Processor_Set;             use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Task_Set;                  use Task_Set;
use Task_Set.Generic_Task_Set;
with Resource_Set;              use Resource_Set;
use Resource_Set.Generic_Resource_Set;
with Message_Set;               use Message_Set;
use Message_Set.Generic_Message_Set;
with Buffer_Set;                use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Select_Time_Line_Pkg;      use Select_Time_Line_Pkg;
with Editor_Config;             use Editor_Config;
with Events;                    use Events;

package body Schedule_From_To_Pkg.Callbacks is

   Draw_From, Draw_Upto, Schedule_To : Unbounded_String;

   Last_Period : constant Natural :=
      Framework_Config.Max_Scheduling_Period - 1;

   Max_Period : Natural := 0;
   Start_Draw : Integer := 0;
   Stop_Draw  : Integer := 0;

   Ok : Boolean;

   procedure On_Cancel_Pressed (Object : access Gtk_Button_Record'Class) is
   begin
      Destroy (Schedule_From_To);
   end On_Cancel_Pressed;

   -- Now, call scheduling analysis for each processor
   --
   procedure Call_Processor_Analysis (Name : Unbounded_String) is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_ptr;

   begin
      Initialize (Response_List);
      Initialize (Request_List);

      if Preemption_From_Simulation then
         Initialize (A_Request);
         A_Request.Target    := Name;
         A_Request.Statement := Scheduling_Simulation_Preemption_Number;
         Add (Request_List, A_Request);
      end if;

      if Context_Switch_From_Simulation then
         Initialize (A_Request);
         A_Request.Target    := Name;
         A_Request.Statement := Scheduling_Simulation_Context_Switch_Number;
         Add (Request_List, A_Request);
      end if;

      if (Response_Time_From_Simulation (Worst_Case) or
          Response_Time_From_Simulation (Best_Case) or
          Response_Time_From_Simulation (Average_Case) or
          Response_Time_From_Simulation (Plot_Case))
      then

         Initialize (A_Request);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("worst_case");
         if Response_Time_From_Simulation (Worst_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("best_case");
         if Response_Time_From_Simulation (Best_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("average_case");
         if Response_Time_From_Simulation (Average_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Request.Target := Name;

         A_Request.Statement := Scheduling_Simulation_Response_Time;
         Add (Request_List, A_Request);
      end if;

      if (Blocking_Time_From_Simulation (Worst_Case) or
          Blocking_Time_From_Simulation (Best_Case) or
          Blocking_Time_From_Simulation (Average_Case) or
          Blocking_Time_From_Simulation (Plot_Case))
      then

         Initialize (A_Request);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("worst_case");
         if Blocking_Time_From_Simulation (Worst_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("best_case");
         if Blocking_Time_From_Simulation (Best_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String ("average_case");
         if Blocking_Time_From_Simulation (Average_Case) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);

         A_Request.Target := Name;

         A_Request.Statement := Scheduling_Simulation_Blocking_Time;
         Add (Request_List, A_Request);
      end if;

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title (Schedule_From_To.Text, Response_List.Entries (0).Title);
      for J in  0 .. Response_List.Nb_Entries - 1 loop
         Write_Text (Schedule_From_To.Text, Response_List.Entries (J).Text);
      end loop;

   end Call_Processor_Analysis;

   procedure Call_Uncustomized_Simulation is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_ptr;

   begin

      Initialize (A_Request);
      A_Request.Statement   := Scheduling_Simulation_Time_Line;
      A_Param               := new Parameter (Integer_Parameter);
      A_Param.Name          := To_Unbounded_String ("period");
      A_Param.Integer_value := Max_Period;
      Add (A_Request.Param, A_Param);
      A_Param               := new Parameter (Boolean_Parameter);
      A_Param.Name          := To_Unbounded_String ("schedule_with_offsets");
      A_Param.Boolean_value := Schedule_With_Offsets;
      Add (A_Request.Param, A_Param);
      A_Param               := new Parameter (Boolean_Parameter);
      A_Param.Name          :=
         To_Unbounded_String ("schedule_with_precedencies");
      A_Param.Boolean_value := Schedule_With_Precedencies;
      Add (A_Request.Param, A_Param);
      A_Param               := new Parameter (Boolean_Parameter);
      A_Param.Name          :=
         To_Unbounded_String ("schedule_with_resources");
      A_Param.Boolean_value := Schedule_With_Resources;
      Add (A_Request.Param, A_Param);
      A_Param               := new Parameter (Boolean_Parameter);
      A_Param.Name          := To_Unbounded_String ("predictable");
      A_Param.Boolean_value := Is_Global_Predictable_Seed;
      Add (A_Request.Param, A_Param);

      if Has_Global_Seed then
         A_Param               := new Parameter (Integer_Parameter);
         A_Param.Name          := To_Unbounded_String ("seed_value");
         A_Param.Integer_value := Global_Seed_Value;
         Add (A_Request.Param, A_Param);
      end if;

      for I in  Time_Unit_Event_Type'Range loop
         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String (To_Lower (I'Img));
         if Generate_Events (I) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);
      end loop;

      Add (Request_List, A_Request);

      Initialize (A_Request);
      A_Request.Target    := Empty_String;
      A_Request.Statement := Scheduling_Simulation_Basics;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      Write_Title_And_Text (Schedule_From_To.Text, Response_List);

      Draw_Time_Line
        (Schedule_From_To.Draw_Scheduling,
         Sys.Tasks,
         Sys.Resources,
         Sys.Messages,
         Sys.Buffers,
         Stop_Draw,
         Start_Draw,
         Call_Scheduling_Framework.Sched);

   end Call_Uncustomized_Simulation;

   procedure Call_Customized_Simulation is

      A_Processor           : Processor_Ptr;
      My_Processor_Iterator : Processors_Iterator;

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_ptr;

   begin

      Initialize (A_Request);

      A_Request.Statement   := Scheduling_Simulation_Time_Line;
      A_Param               := new Parameter (Integer_Parameter);
      A_Param.Name          := To_Unbounded_String ("period");
      A_Param.Integer_value := Max_Period;
      Add (A_Request.Param, A_Param);
      A_Param               := new Parameter (Boolean_Parameter);
      A_Param.Name          := To_Unbounded_String ("schedule_with_offsets");
      A_Param.Boolean_value := Schedule_With_Offsets;
      Add (A_Request.Param, A_Param);
      A_Param               := new Parameter (Boolean_Parameter);
      A_Param.Name          :=
         To_Unbounded_String ("schedule_with_precedencies");
      A_Param.Boolean_value := Schedule_With_Precedencies;
      Add (A_Request.Param, A_Param);
      A_Param               := new Parameter (Boolean_Parameter);
      A_Param.Name          :=
         To_Unbounded_String ("schedule_with_resources");
      A_Param.Boolean_value := Schedule_With_Resources;
      Add (A_Request.Param, A_Param);
      A_Param               := new Parameter (Boolean_Parameter);
      A_Param.Name          := To_Unbounded_String ("predictable");
      A_Param.Boolean_value := Is_Global_Predictable_Seed;
      Add (A_Request.Param, A_Param);

      if Has_Global_Seed then
         A_Param               := new Parameter (Integer_Parameter);
         A_Param.Name          := To_Unbounded_String ("seed_value");
         A_Param.Integer_value := Global_Seed_Value;
         Add (A_Request.Param, A_Param);
      end if;

      for I in  Time_Unit_Event_Type'Range loop
         A_Param      := new Parameter (Boolean_Parameter);
         A_Param.Name := To_Unbounded_String (To_Lower (I'Img));
         if Generate_Events (I) then
            A_Param.Boolean_value := True;
         else
            A_Param.Boolean_value := False;
         end if;
         Add (A_Request.Param, A_Param);
      end loop;

      Add (Request_List, A_Request);

      Sequential_Framework_Request (Sys, Request_List, Response_List);

      -- Put to the screen the resulting event table
      --
      if Schedule_And_Display_Event_Table then
         Put (Call_Scheduling_Framework.Sched);
      end if;

      Write_Title_And_Text (Schedule_From_To.Text, Response_List);

      Draw_Time_Line
        (Schedule_From_To.Draw_Scheduling,
         Sys.Tasks,
         Sys.Resources,
         Sys.Messages,
         Sys.Buffers,
         Stop_Draw,
         Start_Draw,
         Call_Scheduling_Framework.Sched);

      -- export event table in the targeted XML file
      --
      if Schedule_And_Export_Event_Table then
         Write_To_Xml_File
           (Call_Scheduling_Framework.Sched,
            Sys,
            Xml_Event_Table_File_Name);
      end if;

      if Schedule_All_Processeurs = False then
         Call_Processor_Analysis (Last_Selected_Processor);
      else
         Reset_Iterator (Sys.Processors, My_Processor_Iterator);
         loop
            Current_Element
              (Sys.Processors,
               A_Processor,
               My_Processor_Iterator);

            Call_Processor_Analysis (A_Processor.Name);

            exit when Is_Last_Element (Sys.Processors, My_Processor_Iterator);
            Next_Element (Sys.Processors, My_Processor_Iterator);
         end loop;

      end if;

      -- Now run, event analyzers,
      --

      if Run_Event_Analyzer_From_Simulation then

         Initialize (Response_List);
         Initialize (Request_List);

         Initialize (A_Request);
         A_Request.Statement := Scheduling_Simulation_Run_Event_Handler;
         Add (Request_List, A_Request);

         Sequential_Framework_Request (Sys, Request_List, Response_List);

         Write_Title_And_Text (Schedule_From_To.Text, Response_List);

      end if;
   end Call_Customized_Simulation;

   procedure Call_Simulation is
   begin
      -- According to from when the scheduling simulation is demanded
      -- by the user, we build a different request
      --
      if Schedule_From_To.Customized_Simulation then
         Call_Customized_Simulation;
      else
         Call_Uncustomized_Simulation;
      end if;
   end Call_Simulation;

   procedure On_Ok_Pressed (Object : access Gtk_Button_Record'Class) is

      -- Number of time line
      -- to display
      --
      Object_Number : Natural := 0;

   begin

      Draw_From   :=
         To_Unbounded_String (Get_Text (Schedule_From_To.Draw_From_Entry));
      Draw_Upto   :=
         To_Unbounded_String (Get_Text (Schedule_From_To.Draw_Upto_Entry));
      Schedule_To :=
         To_Unbounded_String (Get_Text (Schedule_From_To.Schedule_To_Entry));

      To_Natural (Draw_Upto, Stop_Draw, Ok);
      if not Ok then
         Show_Message_Box
           (Lb_Double_Quote &
            Lb_Draw_To (Current_Language) &
            Lb_Double_Quote &
            Lb_Must_Be_Numeric (Current_Language));
         return;
      end if;

      To_Integer (Draw_From, Start_Draw, Ok);
      if not Ok then
         Show_Message_Box
           (Lb_Double_Quote &
            Lb_Draw_From (Current_Language) &
            Lb_Double_Quote &
            Lb_Must_Be_Numeric (Current_Language));
         return;
      end if;

      To_Integer (Schedule_To, Max_Period, Ok);
      if not Ok then
         Show_Message_Box
           (Lb_Double_Quote &
            Lb_Schedule_To (Current_Language) &
            Lb_Double_Quote &
            Lb_Must_Be_Numeric (Current_Language));
         return;
      end if;

      if (Max_Period > Last_Period) then
         Show_Message_Box
           (Lb_Double_Quote &
            Lb_Schedule_To (Current_Language) &
            Lb_Double_Quote &
            Lb_Must_Be (Current_Language) &
            To_Unbounded_String (" <= ") &
            Last_Period'Img);
         return;
      end if;

      if (Max_Period = 0) then
         Show_Message_Box
           (Lb_Double_Quote &
            Lb_Schedule_To (Current_Language) &
            Lb_Double_Quote &
            Lb_Must_Be (Current_Language) &
            To_Unbounded_String (" > 0"));
         return;
      end if;

      if (Start_Draw < 0) then
         Show_Message_Box
           (Lb_Double_Quote &
            Lb_Draw_From (Current_Language) &
            Lb_Double_Quote &
            Lb_Must_Be (Current_Language) &
            To_Unbounded_String (" >= 0"));
         return;
      end if;

      if (Stop_Draw < 0) then
         Show_Message_Box
           (Lb_Double_Quote &
            Lb_Draw_To (Current_Language) &
            Lb_Double_Quote &
            Lb_Must_Be (Current_Language) &
            To_Unbounded_String (" >= 0"));
         return;
      end if;

      if (Start_Draw > Max_Period) then
         Show_Message_Box
           (Lb_Double_Quote &
            Lb_Draw_From (Current_Language) &
            Lb_Must_Be (Current_Language) &
            Lb_Double_Quote &
            To_Unbounded_String (" <= ") &
            Lb_Double_Quote &
            Lb_Schedule_To (Current_Language) &
            Lb_Double_Quote);
         return;
      end if;

      if (Stop_Draw - Start_Draw > 1500) then
         Show_Message_Box (Lb_Draw_Time_Line_Too_Large (Current_Language));
         return;
      end if;

      -- Check if we need to select which tasks/messages/buffers
      -- to display with time lines
      --
      Object_Number := Natural (Get_Number_Of_Elements (Sys.Tasks)) +
                       Natural (Get_Number_Of_Elements (Sys.Messages)) +
                       Natural (Get_Number_Of_Elements (Sys.Buffers));

      if Schedule_With_Resources then
         Object_Number := Object_Number +
                          Natural (Get_Number_Of_Elements (Sys.Resources));
      end if;

      if Object_Number > Editor_Config.Max_Time_Line_To_Display then
         Gtk_New (Select_Time_Line, Sys, Call_Simulation'Access);
         Show_All (Select_Time_Line);
      else
         Call_Simulation;
      end if;

      -- The window can not be destroyed ... because
      -- the callback used by the Select_Time_Line widget
      -- will call it !!!
      -- So, just hide it, and do not destroy it !!!!
      --
      Hide (Schedule_From_To);

   end On_Ok_Pressed;

end Schedule_From_To_Pkg.Callbacks;
