------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Framework_Config;               use Framework_Config;
with Translate;                      use Translate;
with Tasks;                          use Tasks;
with Task_Set;                       use Task_Set;
use Task_Set.Generic_Task_Set;
with Processors;                     use Processors;
with Processor_Set;                  use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Ada.Strings.Unbounded;          use Ada.Strings.Unbounded;
with Gtk;                            use Gtk;
with Gtk.Enums;                      use Gtk.Enums;
with Callbacks_Cheddar;              use Callbacks_Cheddar;
with Schedule_From_To_Pkg.Callbacks; use Schedule_From_To_Pkg.Callbacks;

package body Schedule_From_To_Pkg is

   procedure Gtk_New
     (Schedule_From_To      : out Schedule_From_To_Access;
      Text                  : Gtk_Text;
      Draw_Scheduling       : Draw_Scheduling_Access;
      Sys                   : in System;
      Customized_Simulation : Boolean := False)
   is

   begin
      Schedule_From_To := new Schedule_From_To_Record;
      Schedule_From_To_Pkg.Initialize
        (Schedule_From_To,
         Text,
         Draw_Scheduling,
         Sys,
         Customized_Simulation);

   end Gtk_New;

   procedure Initialize
     (Schedule_From_To      : access Schedule_From_To_Record'Class;
      Text                  : Gtk_Text;
      Draw_Scheduling       : Draw_Scheduling_Access;
      Sys                   : in System;
      Customized_Simulation : Boolean := False)
   is

      A_Task      : Generic_Task_Ptr;
      Iterator1   : Tasks_Iterator;
      A_Processor : Processor_Ptr;
      Iterator2   : Processors_Iterator;
      Max_Period  : Natural          := 0;
      Period      : Natural          := 0;
      Last_Period : constant Natural :=
         Framework_Config.Max_Scheduling_Period - 1;

   begin

      -- First of all : get the largest period of all
      -- the defined tasks
      --
      Reset_Iterator (Sys.Tasks, Iterator1);
      loop

         Current_Element (Sys.Tasks, A_Task, Iterator1);
         if (A_Task.Task_Type = Periodic_Type) or
            (A_Task.Task_Type = Poisson_Type) or
            (A_Task.Task_Type = Sporadic_Type)
         then
            Max_Period :=
               Natural'Max (Max_Period, Periodic_Task_Ptr (A_Task).Period);
         else
            Max_Period := Natural'Max (Max_Period, A_Task.Deadline);
         end if;

         exit when Is_Last_Element (Sys.Tasks, Iterator1);

         Next_Element (Sys.Tasks, Iterator1);

      end loop;

      -- Then, compute the base period (or hyper-period)
      --
      Reset_Iterator (Sys.Processors, Iterator2);

      loop

         Current_Element (Sys.Processors, A_Processor, Iterator2);

         if (Get_Number_Of_Task_From_Processor (Sys.Tasks, A_Processor.Name) /=
             0)
         then

            begin
               Period := Scheduling_Period (Sys.Tasks, A_Processor.Name);
               if Period > Max_Period then
                  Max_Period := Period;
               end if;

            exception
               when others =>
                  null;
            end;
         end if;

         exit when Is_Last_Element (Sys.Processors, Iterator2);

         Next_Element (Sys.Processors, Iterator2);

      end loop;

      if (Max_Period >= Last_Period) then
         Max_Period := Last_Period;
      end if;

      Schedule_From_To.Customized_Simulation := Customized_Simulation;
      Schedule_From_To.Text                  := Text;
      Schedule_From_To.Draw_Scheduling       := Draw_Scheduling;

      Gtk.Window.Initialize (Schedule_From_To, Window_Dialog);
      Set_Title
        (Schedule_From_To,
         To_String (Lb_Scheduling_Simulation (Current_Language)));
      Set_Policy (Schedule_From_To, False, True, False);
      Set_Position (Schedule_From_To, Win_Pos_Center_Always);
      Set_Modal (Schedule_From_To, True);

      Gtk_New_Vbox (Schedule_From_To.Vbox, False, 0);

      Add (Schedule_From_To, Schedule_From_To.Vbox);

      Gtk_New_Hbox (Schedule_From_To.Hbox, False, 0);
      Set_Border_Width (Schedule_From_To.Hbox, 16);
      Pack_Start
        (Schedule_From_To.Vbox,
         Schedule_From_To.Hbox,
         True,
         True,
         0);

      Gtk_New
        (Schedule_From_To.Schedule_From_To_Label,
         To_String (Lb_Schedule_To (Current_Language) & Lb_Colon));
      Set_Alignment (Schedule_From_To.Schedule_From_To_Label, 0.5, 0.5);
      Set_Padding (Schedule_From_To.Schedule_From_To_Label, 0, 0);
      Set_Justify (Schedule_From_To.Schedule_From_To_Label, Justify_Left);
      Set_Line_Wrap (Schedule_From_To.Schedule_From_To_Label, False);
      Pack_Start
        (Schedule_From_To.Hbox,
         Schedule_From_To.Schedule_From_To_Label,
         False,
         False,
         0);

      Gtk_New (Schedule_From_To.Schedule_To_Entry);
      Set_Editable (Schedule_From_To.Schedule_To_Entry, True);
      Set_Max_Length (Schedule_From_To.Schedule_To_Entry, 0);
      Set_Text (Schedule_From_To.Schedule_To_Entry, Max_Period'Img);
      Set_Visibility (Schedule_From_To.Schedule_To_Entry, True);
      Pack_Start
        (Schedule_From_To.Hbox,
         Schedule_From_To.Schedule_To_Entry,
         True,
         True,
         0);

      Gtk_New_Hbox (Schedule_From_To.Hbox, False, 0);
      Set_Border_Width (Schedule_From_To.Hbox, 16);
      Pack_Start
        (Schedule_From_To.Vbox,
         Schedule_From_To.Hbox,
         True,
         True,
         0);

      Gtk_New
        (Schedule_From_To.Schedule_From_To_Label,
         To_String (Lb_Draw_From (Current_Language) & Lb_Colon));
      Set_Alignment (Schedule_From_To.Schedule_From_To_Label, 0.5, 0.5);
      Set_Padding (Schedule_From_To.Schedule_From_To_Label, 0, 0);
      Set_Justify (Schedule_From_To.Schedule_From_To_Label, Justify_Left);
      Set_Line_Wrap (Schedule_From_To.Schedule_From_To_Label, False);
      Pack_Start
        (Schedule_From_To.Hbox,
         Schedule_From_To.Schedule_From_To_Label,
         False,
         False,
         0);

      Gtk_New (Schedule_From_To.Draw_From_Entry);
      Set_Editable (Schedule_From_To.Draw_From_Entry, True);
      Set_Max_Length (Schedule_From_To.Draw_From_Entry, 0);
      Set_Text (Schedule_From_To.Draw_From_Entry, "0");
      Set_Visibility (Schedule_From_To.Draw_From_Entry, True);
      Pack_Start
        (Schedule_From_To.Hbox,
         Schedule_From_To.Draw_From_Entry,
         True,
         True,
         0);

      Gtk_New_Hbox (Schedule_From_To.Hbox, True, 0);
      Set_Border_Width (Schedule_From_To.Hbox, 16);
      Pack_Start
        (Schedule_From_To.Vbox,
         Schedule_From_To.Hbox,
         True,
         True,
         0);

      Gtk_New
        (Schedule_From_To.Schedule_From_To_Label,
         To_String (Lb_Draw_To (Current_Language) & Lb_Colon));
      Set_Alignment (Schedule_From_To.Schedule_From_To_Label, 0.5, 0.5);
      Set_Padding (Schedule_From_To.Schedule_From_To_Label, 0, 0);
      Set_Justify (Schedule_From_To.Schedule_From_To_Label, Justify_Left);
      Set_Line_Wrap (Schedule_From_To.Schedule_From_To_Label, False);
      Pack_Start
        (Schedule_From_To.Hbox,
         Schedule_From_To.Schedule_From_To_Label,
         False,
         False,
         0);

      Gtk_New (Schedule_From_To.Draw_Upto_Entry);
      Set_Editable (Schedule_From_To.Draw_Upto_Entry, True);
      Set_Max_Length (Schedule_From_To.Draw_Upto_Entry, 0);
      Set_Text (Schedule_From_To.Draw_Upto_Entry, Max_Period'Img);
      Set_Visibility (Schedule_From_To.Draw_Upto_Entry, True);
      Pack_Start
        (Schedule_From_To.Hbox,
         Schedule_From_To.Draw_Upto_Entry,
         True,
         True,
         0);

      Gtk_New_Hseparator (Schedule_From_To.Separator);
      Pack_Start
        (Schedule_From_To.Vbox,
         Schedule_From_To.Separator,
         True,
         True,
         0);

      Gtk_New_Hbox (Schedule_From_To.Hbox, True, 0);
      Set_Border_Width (Schedule_From_To.Hbox, 9);
      Pack_Start
        (Schedule_From_To.Vbox,
         Schedule_From_To.Hbox,
         True,
         True,
         0);

      Gtk_New (Schedule_From_To.Ok_Button, "Ok");
      Set_Usize (Schedule_From_To.Ok_Button, 70, 25);
      Button_Callback.Connect
        (Schedule_From_To.Ok_Button,
         "clicked",
         Button_Callback.To_Marshaller (On_Ok_Pressed'Access));
      Pack_Start
        (Schedule_From_To.Hbox,
         Schedule_From_To.Ok_Button,
         False,
         False,
         0);

      Gtk_New
        (Schedule_From_To.Cancel_Button,
         To_String (Lb_Cancel (Current_Language)));
      Set_Usize (Schedule_From_To.Cancel_Button, 70, 25);
      Button_Callback.Connect
        (Schedule_From_To.Cancel_Button,
         "clicked",
         Button_Callback.To_Marshaller (On_Cancel_Pressed'Access));

      Pack_Start
        (Schedule_From_To.Hbox,
         Schedule_From_To.Cancel_Button,
         False,
         False,
         0);

   end Initialize;

end Schedule_From_To_Pkg;
