------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Glib;                  use Glib;
with Gdk;
with Gdk.Drawable;          use Gdk.Drawable;
with Gdk.Event;             use Gdk.Event;
with Gdk.Font;              use Gdk.Font;
with Gdk.Pixmap;            use Gdk.Pixmap;
with Gdk.Rectangle;         use Gdk.Rectangle;
with Gdk.Types;             use Gdk.Types;
with Gdk.Window;            use Gdk.Window;
with Gtk;                   use Gtk;
with Gtk.Box;               use Gtk.Box;
with Gtk.Enums;             use Gtk.Enums;
with Gtk.Drawing_Area;      use Gtk.Drawing_Area;
with Gtk.Main;              use Gtk.Main;
with Gtk.Handlers;          use Gtk.Handlers;
with Gtk.Style;             use Gtk.Style;
with Gtk.Window;            use Gtk.Window;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Events;                use Events;
with Tasks;                 use Tasks;
with Time_Units;            use Time_Units;
use Time_Units.A_Time_Unit;

package body Simple_Mmi is

   procedure Draw_Scheduling
     (Task1  : Tasks_Set;
      Period : Natural;
      Sched1 : Scheduling_Sequence)
   is

      Left_Margin     : constant Natural := 60;
      High_Margin     : constant Natural := 20;
      Interligne      : constant Natural := 50;
      Timeunit_Margin : constant Natural := 20;

      Pixmap : Gdk_Pixmap;
      Font   : Gdk_Font;
      Window : Gtk_Window;

      package Configured is new Gtk.Handlers.Callback (
         Widget_Type => Gtk_Drawing_Area_Record);

      procedure Draw_Scheduling
        (Widget : access Gtk_Drawing_Area_Record'Class)
      is

         My_Iterator : Tasks_Iterator;
         A_Task      : Generic_Task_Ptr;
         I           : Natural := 0;

         Msg       : Unbounded_String;
         Task_Name : array (Tasks_Range) of Unbounded_String;
         Task_Id   : Natural;
         Time_U    : Natural;
         Ok        : Boolean;

      begin

         ---------------------------------
         -- Draw a line per task
         ---------------------------------

         Reset_Iterator (Task1, My_Iterator);

         loop

            Current_Element (Task1, A_Task, My_Iterator);

            if (A_Task.Task_Type = Periodic_Type) then
               Msg :=
                 To_Unbounded_String ("P=") &
                 Periodic_Task_Ptr (A_Task).Period'Img &
                 To_Unbounded_String ("; C=") &
                 A_Task.Capacity'Img &
                 To_Unbounded_String ("; D=") &
                 Periodic_Task_Ptr (A_Task).Deadline'Img &
                 To_Unbounded_String ("; S=") &
                 A_Task.Start_Time'Img;
            else
               Msg := To_Unbounded_String ("C=") &
                      A_Task.Capacity'Img &
                      To_Unbounded_String ("; D=") &
                      Aperiodic_Task_Ptr (A_Task).Deadline'Img &
                      To_Unbounded_String ("; S=") &
                      A_Task.Start_Time'Img;
            end if;

            Draw_Text
              (Pixmap,
               Font,
               Get_Black (Get_Style (Widget)),
               Gint (Left_Margin),
               Gint (High_Margin + I * Interligne + 20),
               To_String (Msg));

            Draw_Text
              (Pixmap,
               Font,
               Get_Black (Get_Style (Widget)),
               10,
               Gint (High_Margin + I * Interligne),
               To_String (A_Task.Name));

            Draw_Line
              (Pixmap,
               Get_Black (Get_Style (Widget)),
               Gint (Left_Margin),
               Gint (High_Margin + I * Interligne),
               Gint (Left_Margin + Period * Timeunit_Margin),
               Gint (High_Margin + I * Interligne));

            -- For each line, draw unit time lines
            for J in  0 .. Period loop
               Draw_Line
                 (Pixmap,
                  Get_Black (Get_Style (Widget)),
                  Gint (Left_Margin + J * Timeunit_Margin),
                  Gint (High_Margin + I * Interligne - 3),
                  Gint (Left_Margin + J * Timeunit_Margin),
                  Gint (High_Margin + I * Interligne + 3));
            end loop;

            -- Register each task in the name table
            Task_Name (Tasks_Range (I))  := A_Task.Name;

            I := I + 1;

            exit when Is_Last_Element (Task1, My_Iterator);
            Next_Element (Task1, My_Iterator);

         end loop;

         ----------------------------------------
         -- Now, draw resulting scheduling
         ----------------------------------------

         for K in  0 .. Sched1.Nb_Entries - 1 loop

            -- find task position
            Ok := False;
            for J in  Tasks_Range loop
               if Sched1.Entries (K).Item.Event_Type = Running_Task then
                  if (Task_Name (J) =
                      Sched1.Entries (K).Item.Running_Task.Name)
                  then
                     Task_Id := Natural (J);
                     Time_U  := Sched1.Entries (K).Data;
                     Ok      := True;
                  end if;
               end if;
            end loop;
            if (not Ok) then
               raise Task_Not_Found;
            end if;

            Draw_Rectangle
              (Pixmap,
               Get_Black (Get_Style (Widget)),
               True,
               Gint (Time_U * Timeunit_Margin + Left_Margin),
               Gint (Task_Id * Interligne + High_Margin - 3),
               Gint (Timeunit_Margin),
               Gint (7));

         end loop;

         Draw (Widget);
      end Draw_Scheduling;

      procedure Button_Press_Event
        (Widget : access Gtk_Drawing_Area_Record'Class;
         Event  : in Gdk_Event)
      is
         use type Gdk.Pixmap.Gdk_Pixmap;
      begin
         if Get_Button (Event) = 1 and Pixmap /= Null_Pixmap then
            Draw_Scheduling (Widget);
         end if;
      end Button_Press_Event;

      ---------------------
      -- Configure_Event --
      ---------------------

      procedure Configure_Event
        (Drawing_Area : access Gtk_Drawing_Area_Record'Class)
      is
         Win    : Gdk_Window;
         Width  : Gint;
         Height : Gint;

      begin
         Win := Get_Window (Drawing_Area);
         Get_Size (Win, Width, Height);

         Gdk.Pixmap.Gdk_New (Pixmap, Win, Width, Height, -1);
         Draw_Rectangle
           (Pixmap,
            Get_White (Get_Style (Drawing_Area)),
            True,
            0,
            0,
            Width,
            Height);
      end Configure_Event;

      ------------------
      -- Expose_Event --
      ------------------

      procedure Expose_Event
        (Drawing_Area : access Gtk_Drawing_Area_Record'Class;
         Event        : in Gdk.Event.Gdk_Event)
      is
         Area : constant Gdk_Rectangle := Get_Area (Event);
      begin
         Draw_Pixmap
           (Get_Window (Drawing_Area),
            Get_Fg_GC (Get_Style (Drawing_Area), State_Normal),
            Pixmap,
            Area.X,
            Area.Y,
            Area.X,
            Area.Y,
            Gint (Area.Width),
            Gint (Area.Height));
      end Expose_Event;

      -------------------------
      -- Motion_Notify_Event --
      -------------------------

      procedure Motion_Notify_Event
        (Widget : access Gtk_Drawing_Area_Record'Class;
         Event  : in Gdk_Event)
      is
         X, Y   : Gint;
         State  : Gdk_Modifier_Type;
         Result : Gdk_Window;

         use type Gdk.Pixmap.Gdk_Pixmap;
      begin
         if Get_Is_Hint (Event) then
            Get_Pointer (Get_Window (Widget), X, Y, State, Result);
         else
            X     := Gint (Get_X (Event));
            Y     := Gint (Get_Y (Event));
            State := Get_State (Event);
         end if;

         if (State and Button1_Mask) /= 0 and Pixmap /= Null_Pixmap then
            Draw_Scheduling (Widget);
         end if;
      end Motion_Notify_Event;

      Vbox         : Gtk_Box;
      Drawing_Area : Gtk_Drawing_Area;

   begin
      Gtk.Main.Init;
      Gtk_New (Window, Window_Toplevel);
      Set_Title (Window, "Scheduling Result");
      Set_Border_Width (Window, Border_Width => 5);

      Gtk_New_Vbox (Vbox, Homogeneous => False, Spacing => 0);
      Add (Window, Vbox);
      Show (Vbox);

      Gtk_New (Drawing_Area);

      Size
        (Drawing_Area,
         Gint (Period * Timeunit_Margin + Left_Margin),
         Gint (High_Margin +
               Interligne * Natural (Get_Number_Of_Elements (Task1))));

      Pack_Start (In_Box => Vbox, Child => Drawing_Area);
      Show (Drawing_Area);
      Unrealize (Drawing_Area);

      Set_Events
        (Drawing_Area,
         Exposure_Mask or
         Leave_Notify_Mask or
         Button_Press_Mask or
         Pointer_Motion_Mask or
         Pointer_Motion_Hint_Mask);

      Configured.Connect
        (Widget => Drawing_Area,
         Name   => "expose_event",
         Marsh  => Configured.To_Marshaller (Expose_Event'Access));

      Configured.Connect
        (Widget => Drawing_Area,
         Name   => "motion_notify_event",
         Marsh  => Configured.To_Marshaller (Motion_Notify_Event'Access));

      Configured.Connect
        (Widget => Drawing_Area,
         Name   => "configure_event",
         Marsh  => Configured.To_Marshaller (Configure_Event'Access));

      Configured.Connect
        (Widget => Drawing_Area,
         Name   => "button_press_event",
         Marsh  => Configured.To_Marshaller (Button_Press_Event'Access));

      Load
        (Font,
         "-adobe-helvetica-medium-o-normal--14-140-75-75-p-78-iso8859-1");

      Show_All (Window);
      Gtk.Main.Main;
   end Draw_Scheduling;

end Simple_Mmi;
