------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with GNAT.Current_Exception; use GNAT.Current_Exception;
with Glib;                   use Glib;
with Translate;              use Translate;
with Framework_Config;       use Framework_Config;
with Unbounded_Strings;      use Unbounded_Strings;
use Unbounded_Strings.Strings_Table_Package;
with Project;                use Project;
with Gtk.Radio_Menu_Item;    use Gtk.Radio_Menu_Item;
with Processors;             use Processors;
with Task_Set;               use Task_Set;
use Task_Set.Generic_Task_Set;
with Buffer_Set;             use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Resource_Set;           use Resource_Set;
use Resource_Set.Generic_Resource_Set;

package body Update_Address_Spaces_Set_Pkg.Callbacks.Integrity_Checks is

   procedure Check_Integrity_Constraints
     (Update_Address_Spaces_Set : in out Update_Address_Spaces_Set_Access;
      Error_Message             : in out Unbounded_String;
      Ok                        : in out Boolean)
   is

      Address_Space_Data_String    : Unbounded_String;
      Address_Space_Text_String    : Unbounded_String;
      Address_Space_Heap_String    : Unbounded_String;
      Address_Space_Stack_String   : Unbounded_String;
      Address_Space_Quantum_String : Unbounded_String;

      Sched       : Gint;
      Cpu_Index   : Gint := 0;
      A_Processor : Processor_Ptr;
      Integer_Ok  : Boolean;

   begin

      Ok := True;

      Address_Space_Name :=
         To_Unbounded_String (Get_Text (Update_Address_Spaces_Set.Name));

      Address_Space_Heap_String  :=
         To_Unbounded_String
           (Get_Text (Update_Address_Spaces_Set.Heap_Memory_Size));
      Address_Space_Text_String  :=
         To_Unbounded_String
           (Get_Text (Update_Address_Spaces_Set.Text_Memory_Size));
      Address_Space_Data_String  :=
         To_Unbounded_String
           (Get_Text (Update_Address_Spaces_Set.Data_Memory_Size));
      Address_Space_Stack_String :=
         To_Unbounded_String
           (Get_Text (Update_Address_Spaces_Set.Stack_Memory_Size));

      Address_Space_Quantum_String :=
         To_Unbounded_String (Get_Text (Update_Address_Spaces_Set.Quantum));
      if (Address_Space_Quantum_String = "") then
         Address_Space_Quantum := 0;
      else
         To_Integer
           (Address_Space_Quantum_String,
            Address_Space_Quantum,
            Integer_Ok);
         if not Integer_Ok then
            Error_Message := "Quantum" &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      if Get_Active (Update_Address_Spaces_Set.Option) then
         Address_Space_Preemptive := Preemptive;
      else
         Address_Space_Preemptive := Not_Preemptive;
      end if;

      Address_Space_File_Name :=
         To_Unbounded_String (Get_Text (Update_Address_Spaces_Set.File_Name));

      Sched :=
         Gint'Val
           (14 -
            Gint (Selected_Button
                     (Update_Address_Spaces_Set.Clist_Scheduler_Menu)));

      if (Sched = 0) then
         Address_Space_Scheduler := No_Scheduling_Protocol;
      end if;
      if (Sched = 1) then
         Address_Space_Scheduler := Earliest_Deadline_First_Protocol;
      end if;
      if (Sched = 2) then
         Address_Space_Scheduler := Least_Laxity_First_Protocol;
      end if;
      if (Sched = 3) then
         Address_Space_Scheduler := Rate_Monotonic_Protocol;
      end if;
      if (Sched = 4) then
         Address_Space_Scheduler := Deadline_Monotonic_Protocol;
      end if;
      if (Sched = 5) then
         Address_Space_Scheduler :=
           Posix_1003_Highest_Priority_First_Protocol;
      end if;
      if (Sched = 6) then
         Address_Space_Scheduler := Time_Sharing_Based_On_Wait_Time_Protocol;
      end if;
      if (Sched = 7) then
         Address_Space_Scheduler := Round_Robin_Protocol;
      end if;
      if (Sched = 8) then
         Address_Space_Scheduler :=
           Maximum_Urgency_First_Based_On_Laxity_Protocol;
      end if;
      if (Sched = 9) then
         Address_Space_Scheduler :=
           Maximum_Urgency_First_Based_On_Deadline_Protocol;
      end if;
      if (Sched = 10) then
         Address_Space_Scheduler := D_Over_Protocol;
      end if;
      if (Sched = 11) then
         Address_Space_Scheduler := Time_Sharing_Based_On_Cpu_Usage_Protocol;
      end if;
      if (Sched = 12) then
         Address_Space_Scheduler := pipeline_User_Defined_Protocol;
      end if;
      if (Sched = 13) then
         Address_Space_Scheduler := Automata_User_Defined_Protocol;
      end if;
      if (Sched = 14) then
         Address_Space_Scheduler := Compiled_User_Defined_Protocol;
      end if;

      Cpu_Index :=
         Gint'Val
           (Gint (Get_Number_Of_Elements (Sys.Processors)) -
            Gint (Selected_Button
                     (Update_Address_Spaces_Set.Clist_Processor_Menu)));
      Cpu_Index := Cpu_Index - 1;

      Get_Element_Number
        (Sys.Processors,
         A_Processor,
         Processors_Range (Cpu_Index));
      Address_Space_Processor := A_Processor.Name;

      To_Integer (Address_Space_Text_String, Address_Space_Text, Integer_Ok);
      if not Integer_Ok then
         if Address_Space_Text_String /= "" then
            Error_Message := Lb_Text_Memory_Size (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer (Address_Space_Data_String, Address_Space_Data, Integer_Ok);
      if not Integer_Ok then
         if Address_Space_Data_String /= To_Unbounded_String ("") then
            Error_Message := Lb_Data_Memory_Size (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer (Address_Space_Heap_String, Address_Space_Heap, Integer_Ok);
      if not Integer_Ok then
         if Address_Space_Heap_String /= To_Unbounded_String ("") then
            Error_Message := Lb_Heap_Memory_Size (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer
        (Address_Space_Stack_String,
         Address_Space_Stack,
         Integer_Ok);
      if not Integer_Ok then
         if Address_Space_Stack_String /= To_Unbounded_String ("") then
            Error_Message := Lb_Stack_Memory_Size (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      Check_Address_Space
        (Sys.Address_Spaces,
         Integrity_Checks.Address_Space_Name,
         Integrity_Checks.Address_Space_Processor,
         Integrity_Checks.Address_Space_Text,
         Integrity_Checks.Address_Space_Stack,
         Integrity_Checks.Address_Space_Data,
         Integrity_Checks.Address_Space_Heap,
         Integrity_Checks.Address_Space_Preemptive,
         Integrity_Checks.Address_Space_Quantum,
         Integrity_Checks.Address_Space_File_Name,
         Integrity_Checks.Address_Space_Scheduler);

   exception
      when others =>
         Ok            := False;
         Error_Message := To_Unbounded_String (Exception_Message);

   end Check_Integrity_Constraints;

end Update_Address_Spaces_Set_Pkg.Callbacks.Integrity_Checks;
