------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with GNAT.Current_Exception; use GNAT.Current_Exception;
with Translate;              use Translate;
with Framework_Config;       use Framework_Config;
with Unbounded_Strings;      use Unbounded_Strings;
use Unbounded_Strings.Strings_Table_Package;
with Project;                use Project;
with Task_Set;               use Task_Set;
use Task_Set.Generic_Task_Set;

package body Update_Buffers_Set_Pkg.Callbacks.Integrity_Checks is

   procedure Check_Integrity_Constraints
     (Update_Buffers_Set : in out Update_Buffers_Set_Access;
      Error_Message      : in out Unbounded_String;
      Ok                 : in out Boolean)
   is

      Buffer_Qs_String   : Unbounded_String;
      Buffer_Size_String : Unbounded_String;
      Integer_Ok         : Boolean;
      Buffer_Ok          : Boolean;

   begin

      Ok := True;

      Buffer_Name :=
         To_Unbounded_String (Get_Text (Update_Buffers_Set.Name));

      Buffer_Processor     :=
         To_Unbounded_String (Get_Text (Update_Buffers_Set.Cpu_Name));
      Buffer_Address_Space :=
         To_Unbounded_String
           (Get_Text (Get_Entry (Update_Buffers_Set.Address_Space_Menu)));

      Buffer_Size_String :=
         To_Unbounded_String (Get_Text (Update_Buffers_Set.Size));

      To_Integer (Buffer_Size_String, Buffer_Size, Integer_Ok);
      if not Integer_Ok then
         Error_Message := Lb_Size (Current_Language) &
                          Lb_Must_Be_Numeric (Current_Language);
         Ok            := False;
         return;
      end if;

      Buffer_Qs_String :=
         To_Unbounded_String
           (Get_Text (Get_Entry (Update_Buffers_Set.Qs_Menu)));
      Display_String_To_Qs (Buffer_Qs_String, Buffer_Qs, Buffer_Ok);

      Check_Buffer
        (Sys.Buffers,
         Buffer_Name,
         Buffer_Size,
         Buffer_Processor,
         Buffer_Address_Space);

   exception
      when others =>
         Ok            := False;
         Error_Message := To_Unbounded_String (Exception_Message);

   end Check_Integrity_Constraints;

   procedure Check_Task_Role_Integrity_Constraints
     (Update_Buffers_Set : in out Update_Buffers_Set_Access;
      Error_Message      : in out Unbounded_String;
      Ok                 : in out Boolean)
   is

      Time_String : Unbounded_String;
      Size_String : Unbounded_String;
      Type_String : Unbounded_String;
      Integer_Ok  : Boolean;
      Role_Ok     : Boolean;

   begin

      Ok := True;

      -- Task name
      --
      Buffer_Role_Task :=
         To_Unbounded_String
           (Get_Text (Get_Entry (Update_Buffers_Set.Affect_Task)));

      -- Role Time
      --
      Time_String :=
         To_Unbounded_String (Get_Text (Update_Buffers_Set.Affect_Roletime));

      To_Integer (Time_String, Buffer_Role_Time, Integer_Ok);
      if not Integer_Ok then
         if (Time_String /= "") then
            Error_Message := Lb_Time (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      -- Role size
      --
      Size_String :=
         To_Unbounded_String (Get_Text (Update_Buffers_Set.Affect_Size));

      To_Integer (Size_String, Buffer_Role_Size, Integer_Ok);
      if not Integer_Ok then
         if (Size_String /= "") then
            Error_Message := Lb_Size (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      -- Type of task access (producer or consumer ?)
      --
      Type_String :=
         To_Unbounded_String
           (Get_Text (Get_Entry (Update_Buffers_Set.Affect_Roletype)));
      To_Role_type (Type_String, Buffer_Role_Type, Role_Ok);

      Check_Buffer_Roles
        (Sys.Tasks,
         Integrity_Checks.Buffer_Name,
         Integrity_Checks.Buffer_Role_Task,
         Integrity_Checks.Buffer_Role_Time,
         Integrity_Checks.Buffer_Role_Size);

   exception
      when others =>
         Ok            := False;
         Error_Message := To_Unbounded_String (Exception_Message);

   end Check_Task_Role_Integrity_Constraints;

end Update_Buffers_Set_Pkg.Callbacks.Integrity_Checks;
