------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with GNAT.Current_Exception; use GNAT.Current_Exception;
with Translate;              use Translate;
with Framework_Config;       use Framework_Config;
with Unbounded_Strings;      use Unbounded_Strings;
use Unbounded_Strings.Strings_Table_Package;
with Task_Set;               use Task_Set;
use Task_Set.Generic_Task_Set;
with Project;                use Project;

package body Update_Resources_Set_Pkg.Callbacks.Integrity_Checks is

   procedure Check_Integrity_Constraints
     (Update_Resources_Set : in out Update_Resources_Set_Access;
      Error_Message        : in out Unbounded_String;
      Ok                   : in out Boolean)
   is

      Resource_State_String    : Unbounded_String;
      Resource_Protocol_String : Unbounded_String;

   begin

      Ok := True;

      Resource_Name :=
         To_Unbounded_String (Get_Text (Update_Resources_Set.Name));

      Resource_State_String :=
         To_Unbounded_String (Get_Text (Update_Resources_Set.State));

      if Resource_State_String = "" then
         Resource_State := 0;
      else
         To_Integer (Resource_State_String, Resource_State, Ok);
         if not Ok then
            Error_Message := Lb_Resource_State (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            return;
         end if;
      end if;

      Resource_Protocol_String :=
         To_Unbounded_String
           (Gtk.GEntry.Get_Text (Get_Entry (Update_Resources_Set.Protocol)));

      if Resource_Protocol_String = "PCP" then
         Resource_Protocol := Priority_Ceiling_Protocol;
      else
         if Resource_Protocol_String = "PIP" then
            Resource_Protocol := Priority_Inheritance_Protocol;
         else
            if Resource_Protocol_String = "IPCP" then
               Resource_Protocol := Immediate_Priority_Ceiling_Protocol;
            else
               Resource_Protocol := No_Protocol;
            end if;
         end if;
      end if;

      Resource_Processor     :=
         To_Unbounded_String (Get_Text (Update_Resources_Set.Cpu_Name));
      Resource_Address_Space :=
         To_Unbounded_String
           (Get_Text (Get_Entry (Update_Resources_Set.Address_Space_Menu)));

      Check_Resource
        (Sys.Resources,
         Integrity_Checks.Resource_Name,
         Integrity_Checks.Resource_State,
         Integrity_Checks.Resource_Processor,
         Integrity_Checks.Resource_Address_Space,
         Integrity_Checks.Resource_Protocol);

      for I in  0 .. Update_Resources_Set.Task_Lists_Share.Nb_Entries - 1 loop

         Check_Task_Critical_Section
           (Sys.Tasks,
            Integrity_Checks.Resource_Name,
            Integrity_Checks.Resource_Processor,
            Update_Resources_Set.Task_Lists_Share.Entries (I).Item,
            Update_Resources_Set.Task_Lists_Share.Entries (I).Data.Task_Begin,
            Update_Resources_Set.Task_Lists_Share.Entries (I).Data.Task_End);

      end loop;

   exception
      when others =>
         Ok            := False;
         Error_Message := To_Unbounded_String (Exception_Message);

   end Check_Integrity_Constraints;

   procedure Check_Task_Sharing_Integrity_Constraints
     (Update_Resources_Set : in out Update_Resources_Set_Access;
      Error_Message        : in out Unbounded_String;
      Ok                   : in out Boolean)
   is

      Share_Begin_String : Unbounded_String;
      Share_End_String   : Unbounded_String;
      Integer_Ok         : Boolean;

   begin

      Ok := True;

      Resource_Task_Name :=
         To_Unbounded_String
           (Get_Text (Get_Entry (Update_Resources_Set.Resource_Affect_Task)));
      Share_Begin_String :=
         To_Unbounded_String
           (Get_Text (Update_Resources_Set.Resource_Affect_Begin));
      Share_End_String   :=
         To_Unbounded_String
           (Get_Text (Update_Resources_Set.Resource_Affect_End));

      To_Integer (Share_Begin_String, Resource_Share_Begin, Integer_Ok);
      if not Integer_Ok then
         Error_Message := Lb_Task_Begin (Current_Language) &
                          Lb_Must_Be_Numeric (Current_Language);
         Ok            := False;
         return;
      end if;

      To_Integer (Share_End_String, Resource_Share_End, Integer_Ok);
      if not Integer_Ok then
         Error_Message := Lb_Task_End (Current_Language) &
                          Lb_Must_Be_Numeric (Current_Language);
         Ok            := False;
         return;
      end if;

   exception
      when others =>
         Ok            := False;
         Error_Message := To_Unbounded_String (Exception_Message);

   end Check_Task_Sharing_Integrity_Constraints;

end Update_Resources_Set_Pkg.Callbacks.Integrity_Checks;
