------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Gtk.Radio_Menu_Item;    use Gtk.Radio_Menu_Item;
with Glib;                   use Glib;
with Translate;              use Translate;
with Framework_Config;       use Framework_Config;
with Unbounded_Strings;      use Unbounded_Strings;
use Unbounded_Strings.Strings_Table_Package;
with Project;                use Project;
with Address_Space_Set;      use Address_Space_Set;
use Address_Space_Set.Generic_Address_Space_Set;
with Buffer_Set;             use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Resource_Set;           use Resource_Set;
use Resource_Set.Generic_Resource_Set;
with Network_Set;            use Network_Set;
use Network_Set.Generic_Network_Set;
with GNAT.Current_Exception; use GNAT.Current_Exception;

package body Update_Tasks_Set_Pkg.Callbacks.Integrity_Checks is

   procedure Check_Integrity_Constraints
     (Update_Tasks_Set : in out Update_Tasks_Set_Access;
      Error_Message    : in out Unbounded_String;
      Ok               : in out Boolean)
   is

      Task_Type_Index : Gint := 0;
      Policy_Index    : Gint := 0;

      Integer_Ok, Double_Ok, Boolean_Ok : Boolean;

      Task_Priority_String          : Unbounded_String;
      Task_Capacity_String          : Unbounded_String;
      Task_Jitter_String            : Unbounded_String;
      Task_Deadline_String          : Unbounded_String;
      Task_Period_String            : Unbounded_String;
      Task_Start_Time_String        : Unbounded_String;
      Task_Blocking_Time_String     : Unbounded_String;
      Task_Seed_Value_String        : Unbounded_String;
      Task_Stack_Memory_Size_String : Unbounded_String;
      Task_Text_Memory_Size_String  : Unbounded_String;
      Task_Criticality_String       : Unbounded_String;

      Current      : Unbounded_String;
      A_Offset     : Offset_Type;
      A_Param      : Parameter_ptr;
      A_Param_Type : Gint;

   begin

      Ok := True;

      --
      -- First task page : general information
      --

      Task_Name                     :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Name));
      Task_Priority_String          :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Priority));
      Task_Capacity_String          :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Capacity));
      Task_Jitter_String            :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Jitter));
      Task_Deadline_String          :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Deadline));
      Task_Period_String            :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Period));
      Task_Criticality_String       :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Criticality));
      Task_Start_Time_String        :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Start_Time));
      Task_Blocking_Time_String     :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Blocking_Time));
      Task_Parametric_Rule          :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Activation_Rule));
      Task_Text_Memory_Size_String  :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Text_Memory_Size));
      Task_Stack_Memory_Size_String :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Stack_Memory_Size));
      Task_Seed_Value_String        :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Seed_Value));
      Task_Processor                :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Cpu_Name));
      Task_Address_Space            :=
         To_Unbounded_String
           (Get_Text (Get_Entry (Update_Tasks_Set.Address_Space_Menu)));

      -- Convert value from menu to Tasks_Type
      --
      Task_Type_Index :=
        Gint (4 - Selected_Button (Update_Tasks_Set.Clist_Task_Type_Menu));
      Task_Type       := Tasks_Type'Val (Task_Type_Index);

      -- Convert value from menu to Policies
      --
      Policy_Index :=
        Gint (2 - Selected_Button (Update_Tasks_Set.Clist_Policy_Menu));
      Task_Policy  := Policies'Val (Policy_Index);

      To_Integer (Task_Priority_String, Task_Priority, Integer_Ok);
      if not Integer_Ok then
         if Task_Priority_String /= "" then
            Error_Message := Lb_Priority (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer (Task_Capacity_String, Task_Capacity, Integer_Ok);
      if not Integer_Ok then
         if Task_Capacity_String /= "" then
            Error_Message := Lb_Capacity (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer (Task_Jitter_String, Task_Jitter, Integer_Ok);
      if not Integer_Ok then
         if Task_Jitter_String /= "" then
            Error_Message := Lb_Jitter (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer (Task_Deadline_String, Task_Deadline, Integer_Ok);
      if not Integer_Ok then
         if Task_Deadline_String /= "" then
            Error_Message := Lb_Deadline (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer (Task_Criticality_String, Task_Criticality, Integer_Ok);
      if not Integer_Ok then
         if Task_Criticality_String /= "" then
            Error_Message := Lb_Criticality (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer (Task_Period_String, Task_Period, Integer_Ok);
      if not Integer_Ok then
         if Task_Period_String /= "" then
            Error_Message := Lb_Period (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer (Task_Seed_Value_String, Task_Seed_Value, Integer_Ok);
      if not Integer_Ok then
         if Task_Seed_Value_String /= "" then
            Error_Message := Lb_Seed (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      if Get_Active (Update_Tasks_Set.Seed_Predictable) then
         Task_Predictable := True;
      else
         Task_Predictable := False;
      end if;

      To_Integer (Task_Start_Time_String, Task_Start_Time, Integer_Ok);
      if not Integer_Ok then
         if Task_Start_Time_String /= "" then
            Error_Message := Lb_Start_Time (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer (Task_Blocking_Time_String, Task_Blocking_Time, Integer_Ok);
      if not Integer_Ok then
         if Task_Blocking_Time_String /= "" then
            Error_Message := Lb_Blocking_Time (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer
        (Task_Text_Memory_Size_String,
         Task_Text_Memory_Size,
         Integer_Ok);
      if not Integer_Ok then
         if Task_Text_Memory_Size_String /= "" then
            Error_Message := Lb_Text_Memory_Size (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      To_Integer
        (Task_Stack_Memory_Size_String,
         Task_Stack_Memory_Size,
         Integer_Ok);
      if not Integer_Ok then
         if Task_Stack_Memory_Size_String /= "" then
            Error_Message := Lb_Stack_Memory_Size (Current_Language) &
                             Lb_Must_Be_Numeric (Current_Language);
            Ok            := False;
            return;
         end if;
      end if;

      --
      --- Second page of the widget : user defined parameters
      --
      Initialize (Task_Param);

      for I in  0 .. User_Defined_Parameters_Range'Last - 1 loop

         Current :=
            To_Unbounded_String
              (Get_Text (Update_Tasks_Set.Param (I).Entry_Name));

         if Current /= Empty_String then

            A_Param_Type :=
              Gint (Selected_Button
                       (Update_Tasks_Set.Param (I).Clist_Parameter_Menu));

            if A_Param_Type = 3 then
               A_Param := new Parameter (Integer_Parameter);
            else
               if A_Param_Type = 2 then
                  A_Param := new Parameter (Double_Parameter);
               else
                  if A_Param_Type = 1 then
                     A_Param := new Parameter (String_Parameter);
                  else
                     A_Param := new Parameter (Boolean_Parameter);
                  end if;
               end if;
            end if;

            A_Param.Name := Current;

            Current :=
               To_Unbounded_String
                 (Get_Text (Update_Tasks_Set.Param (I).Entry_Value));

            -- Integer case
            --
            if A_Param_Type = 3 then
               if (Current = "") then
                  A_Param.Integer_value := 0;
               else
                  To_Integer (Current, A_Param.Integer_value, Integer_Ok);
                  if not Integer_Ok then
                     Error_Message :=
                       To_Unbounded_String
                            ("Parameter " &
                             Integer'Image (Integer (I + 1)) &
                             " ") &
                       Lb_Must_Be (Current_Language) &
                       To_Unbounded_String (" integer");
                     Ok            := False;
                     return;
                  end if;
               end if;

            -- Double case
            --
            else
               if A_Param_Type = 2 then
                  if (Current = "") then
                     A_Param.double_value := 0.0;
                  else
                     To_Double (Current, A_Param.double_value, Double_Ok);
                     if not Double_Ok then
                        Error_Message :=
                          To_Unbounded_String
                               ("Parameter " &
                                Integer'Image (Integer (I + 1)) &
                                " ") &
                          Lb_Must_Be (Current_Language) &
                          To_Unbounded_String (" double");
                        Ok            := False;
                        return;
                     end if;
                  end if;

               -- String case
               --
               else
                  if A_Param_Type = 1 then
                     A_Param.string_value := Current;

                  -- Boolean case
                  --
                  else
                     if (Current = "") then
                        A_Param.Boolean_value := True;
                     else
                        To_Boolean
                          (Current,
                           A_Param.Boolean_value,
                           Boolean_Ok);
                        if not Boolean_Ok then
                           Error_Message :=
                             To_Unbounded_String
                                  ("Parameter " &
                                   Integer'Image (Integer (I + 1)) &
                                   " ") &
                             Lb_Must_Be (Current_Language) &
                             To_Unbounded_String (" boolean");
                           Ok            := False;
                           return;
                        end if;
                     end if;

                  end if;
               end if;
            end if;

            Add (Task_Param, A_Param);
         end if;

      end loop;

      --
      -- Offset information
      --
      Initialize (Task_Offset);

      for I in  0 .. Offsets_Range'Last - 1 loop

         Current :=
            To_Unbounded_String
              (Get_Text (Update_Tasks_Set.Offset (I).Entry_Val));
         if (Current = "") then
            A_Offset.Offset_Value := 0;
         else
            To_Integer (Current, A_Offset.Offset_Value, Integer_Ok);
            if not Integer_Ok then
               Error_Message := To_Unbounded_String ("Offset") &
                                Lb_Must_Be_Numeric (Current_Language) &
                                To_Unbounded_String ("Offset(" & I'Img & ")") &
                                Lb_Value (Current_Language);
               Ok            := False;
               return;
            end if;
         end if;

         Current :=
            To_Unbounded_String
              (Get_Text (Update_Tasks_Set.Offset (I).Entry_Act));
         if (Current = "") then
            A_Offset.Activation := 0;
         else
            To_Integer (Current, A_Offset.Activation, Integer_Ok);
            if not Integer_Ok then
               Error_Message := To_Unbounded_String ("Offset") &
                                Lb_Must_Be_Numeric (Current_Language) &
                                To_Unbounded_String
                                   ("Offset(" & I'Img & ") ") &
                                Lb_Activation (Current_Language);
               Ok            := False;
               return;
            end if;
         end if;

         if A_Offset.Activation /= 0 then
            Add (Task_Offset, A_Offset);
         end if;

      end loop;

      Check_Task
        (Sys.Tasks,
         Integrity_Checks.Task_Name,
         Integrity_Checks.Task_Processor,
         Integrity_Checks.Task_Address_Space,
         Integrity_Checks.Task_Type,
         Integrity_Checks.Task_Start_Time,
         Integrity_Checks.Task_Capacity,
         Integrity_Checks.Task_Period,
         Integrity_Checks.Task_Deadline,
         Integrity_Checks.Task_Jitter,
         Integrity_Checks.Task_Blocking_Time,
         Integrity_Checks.Task_Priority,
         Integrity_Checks.Task_Criticality,
         Integrity_Checks.Task_Policy,
         Integrity_Checks.Task_Offset,
         Integrity_Checks.Task_Stack_Memory_Size,
         Integrity_Checks.Task_Text_Memory_Size,
         Integrity_Checks.Task_Param,
         Integrity_Checks.Task_Parametric_Rule,
         Integrity_Checks.Task_Seed_Value,
         Integrity_Checks.Task_Predictable);

   exception
      when others =>
         Ok            := False;
         Error_Message := To_Unbounded_String (Exception_Message);

   end Check_Integrity_Constraints;

end Update_Tasks_Set_Pkg.Callbacks.Integrity_Checks;
