------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Glib;                                            use Glib;
with Mmi_Tools;                                       use Mmi_Tools;
with Translate;                                       use Translate;
with Framework_Config;                                use Framework_Config;
with Offsets;                                         use Offsets;
use Offsets.Offsets_Table_package;
with Parameters;                                      use Parameters;
use Parameters.User_Defined_Parameters_Table_Package;
with Unbounded_Strings;                               use Unbounded_Strings;
use Unbounded_Strings.Strings_Table_Package;
with Project;                                         use Project;
with Tasks;                                           use Tasks;
with Address_Spaces;                                  use Address_Spaces;
with Address_Space_Set;                               use Address_Space_Set;
use Address_Space_Set.Generic_Address_Space_Set;
with Graph_Window_Pkg;                                use Graph_Window_Pkg;
with Graph_Item_Pkg;                                  use Graph_Item_Pkg;
with Graph_Add_Pkg;                                   use Graph_Add_Pkg;
with My_Canvas;                                       use My_Canvas;
with Buffer_Set;                                      use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Double_Util;                                     use Double_Util;
with Ada.Numerics.Float_Random;                       use 
  Ada.Numerics.Float_Random;
with Ada.Numerics.Aux;                                use Ada.Numerics.Aux;
with Resource_Set;                                    use Resource_Set;
use Resource_Set.Generic_Resource_Set;
with Network_Set;                                     use Network_Set;
use Network_Set.Generic_Network_Set;
with Task_Dependencies;                               use Task_Dependencies;
with GNAT.Current_Exception;                          use 
  GNAT.Current_Exception;
with Update_Tasks_Set_Pkg.Callbacks.Integrity_Checks;

package body Update_Tasks_Set_Pkg.Callbacks is

   procedure Page_Switch
     (Notebook : access Gtk_Notebook_Record'Class;
      Page     : in Gtk.Gtk_Notebook_Page;
      Page_Num : in Gtk_Notebook)
   is
      pragma Warnings (Off, Page_Num);
      Old_Page : constant Gtk.Gtk_Notebook_Page := Get_Cur_Page (Notebook);
      Pixmap   : Gtk_Pixmap;
   begin
      Pixmap := Gtk_Pixmap (Get_Child (Gtk_Box (Get_Tab_Label (Page)), 0));
      Gtk.Pixmap.Set
        (Pixmap,
         Update_Tasks_Set.Book_Opened,
         Update_Tasks_Set.Book_Opened_Mask);
      Pixmap :=
        Gtk_Pixmap (Get_Child (Gtk_Box (Get_Menu_Label (Page)), 0));
      Gtk.Pixmap.Set
        (Pixmap,
         Update_Tasks_Set.Book_Opened,
         Update_Tasks_Set.Book_Opened_Mask);

      if Old_Page /= null then
         Pixmap :=
           Gtk_Pixmap (Get_Child (Gtk_Box (Get_Tab_Label (Old_Page)), 0));
         Gtk.Pixmap.Set
           (Pixmap,
            Update_Tasks_Set.Book_Closed,
            Update_Tasks_Set.Book_Closed_Mask);
         Pixmap :=
           Gtk_Pixmap (Get_Child (Gtk_Box (Get_Menu_Label (Old_Page)), 0));
         Gtk.Pixmap.Set
           (Pixmap,
            Update_Tasks_Set.Book_Closed,
            Update_Tasks_Set.Book_Closed_Mask);
      end if;
   end Page_Switch;

   procedure Show_Tasks
     (Update_Tasks_Set : access Update_Tasks_Set_Record'Class)
   is

      A_Task          : Generic_Task_Ptr;
      My_Iterator     : Tasks_Iterator;
      Line            : Gint;
      Task_Share      : Chars_Ptr_Array (1 .. 12);
      Capacity        : Unbounded_String;
      Start_Time      : Unbounded_String;
      Jitter          : Unbounded_String;
      Deadline        : Unbounded_String;
      Period          : Unbounded_String;
      Priority        : Unbounded_String;
      Blocking_Time   : Unbounded_String;
      Parametric_Rule : Unbounded_String;

      Policy    : Unbounded_String;
      Task_Type : Unbounded_String;

   begin

      Clear (Update_Tasks_Set.Tasks_List);

      Reset_Iterator (Sys.Tasks, My_Iterator);
      if not Is_Empty (Sys.Tasks) then
         loop

            Current_Element (Sys.Tasks, A_Task, My_Iterator);
            Capacity      :=
               To_Unbounded_String (Natural'Image (A_Task.Capacity));
            Deadline      :=
               To_Unbounded_String (Natural'Image (A_Task.Deadline));
            Start_Time    :=
               To_Unbounded_String (Natural'Image (A_Task.Start_Time));
            Priority      :=
               To_Unbounded_String
                 (Natural'Image (Natural (A_Task.Priority)));
            Blocking_Time :=
               To_Unbounded_String (Natural'Image (A_Task.Blocking_Time));

            case A_Task.Task_Type is
               when Aperiodic_Type =>
                  Jitter          := To_Unbounded_String (" 0");
                  Period          := To_Unbounded_String (" 0");
                  Parametric_Rule := To_Unbounded_String ("");
                  Task_Type       := To_Unbounded_String ("Aperiodic");
               when Periodic_Type =>
                  Jitter          :=
                     To_Unbounded_String
                       (Natural'Image (Periodic_Task_Ptr (A_Task).Jitter));
                  Period          :=
                     To_Unbounded_String
                       (Natural'Image (Periodic_Task_Ptr (A_Task).Period));
                  Parametric_Rule := To_Unbounded_String ("");
                  Task_Type       := To_Unbounded_String ("Periodic");
               when Poisson_Type =>
                  Jitter          :=
                     To_Unbounded_String
                       (Natural'Image (Poisson_Task_Ptr (A_Task).Jitter));
                  Period          :=
                     To_Unbounded_String
                       (Natural'Image (Poisson_Task_Ptr (A_Task).Period));
                  Parametric_Rule := To_Unbounded_String ("");
                  Task_Type       := To_Unbounded_String ("Poisson");
               when Sporadic_Type =>
                  Jitter          :=
                     To_Unbounded_String
                       (Natural'Image (Sporadic_Task_Ptr (A_Task).Jitter));
                  Period          :=
                     To_Unbounded_String
                       (Natural'Image (Sporadic_Task_Ptr (A_Task).Period));
                  Parametric_Rule := To_Unbounded_String ("");
                  Task_Type       := To_Unbounded_String ("Sporadic");
               when Parametric_Type =>
                  Jitter          :=
                     To_Unbounded_String
                       (Natural'Image (Poisson_Task_Ptr (A_Task).Jitter));
                  Period          :=
                     To_Unbounded_String
                       (Natural'Image (Poisson_Task_Ptr (A_Task).Period));
                  Parametric_Rule :=
                    Parametric_Task_Ptr (A_Task).Activation_Rule;
                  Task_Type       := To_Unbounded_String ("User-defined");
            end case;

            case A_Task.Policy is
               when Sched_Fifo =>
                  Policy := To_Unbounded_String ("Fifo");
               when Sched_Rr =>
                  Policy := To_Unbounded_String ("RR");
               when Sched_Others =>
                  Policy := To_Unbounded_String ("Other");
            end case;

            -- Create the line
            --
            Task_Share :=
              To_String (A_Task.Name) +
              To_String (Task_Type) +
              To_String (A_Task.Cpu_Name) +
              To_String (Policy) +
              To_String (Priority) +
              To_String (Capacity) +
              To_String (Jitter) +
              To_String (Deadline) +
              To_String (Period) +
              To_String (Start_Time) +
              To_String (Blocking_Time) +
              To_String (Parametric_Rule);

            -- Display the Line
            --
            Line := Append (Update_Tasks_Set.Tasks_List, Task_Share);

            exit when Is_Last_Element (Sys.Tasks, My_Iterator);

            Next_Element (Sys.Tasks, My_Iterator);

         end loop;
      end if;
   end Show_Tasks;

   procedure On_Click_List (List : access Gtk_Clist_Record'Class) is

      Name            : Unbounded_String;
      Capacity        : Unbounded_String;
      Jitter          : Unbounded_String;
      Deadline        : Unbounded_String;
      Period          : Unbounded_String;
      Start_Time      : Unbounded_String;
      Priority        : Unbounded_String;
      Blocking_Time   : Unbounded_String;
      Parametric_Rule : Unbounded_String;
      Row             : Gint;

      A_Task  : Generic_Task_Ptr;
      Current : Gint;

      Location_Items  : Strings_Table;
      Parameter_Items :
        array (User_Defined_Parameters_Range) of Strings_Table;

      A_Address_Space        : Address_Space_Ptr;
      Address_Space_Iterator :
        Address_Space_Set.Generic_Address_Space_Set.Iterator;

      Empty_Menu        : String_List.Glist;
      Current_Task_Type : Gint;
      Current_Policy    : Gint;

   begin

      Row  := Get_Focus_Row (Update_Tasks_Set.Tasks_List);
      Name :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Tasks_List, Row, 0));

      Update_Tasks_Set.Current_Task := Name;

      Priority        :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Tasks_List, Row, 4));
      Capacity        :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Tasks_List, Row, 5));
      Jitter          :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Tasks_List, Row, 6));
      Deadline        :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Tasks_List, Row, 7));
      Period          :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Tasks_List, Row, 8));
      Start_Time      :=
         To_Unbounded_String (Get_Text (Update_Tasks_Set.Tasks_List, Row, 9));
      Blocking_Time   :=
         To_Unbounded_String
           (Get_Text (Update_Tasks_Set.Tasks_List, Row, 10));
      Parametric_Rule :=
         To_Unbounded_String
           (Get_Text (Update_Tasks_Set.Tasks_List, Row, 11));

      Gtk.GEntry.Set_Text (Update_Tasks_Set.Name, To_String (Name));
      Gtk.GEntry.Set_Text (Update_Tasks_Set.Priority, To_String (Priority));
      Gtk.GEntry.Set_Text (Update_Tasks_Set.Capacity, To_String (Capacity));
      Gtk.GEntry.Set_Text (Update_Tasks_Set.Jitter, To_String (Jitter));
      Gtk.GEntry.Set_Text (Update_Tasks_Set.Deadline, To_String (Deadline));
      Gtk.GEntry.Set_Text (Update_Tasks_Set.Period, To_String (Period));
      Gtk.GEntry.Set_Text
        (Update_Tasks_Set.Start_Time,
         To_String (Start_Time));
      Gtk.GEntry.Set_Text
        (Update_Tasks_Set.Blocking_Time,
         To_String (Blocking_Time));
      Gtk.GEntry.Set_Text
        (Update_Tasks_Set.Activation_Rule,
         To_String (Parametric_Rule));

      A_Task := Search_Task (Sys.Tasks, Update_Tasks_Set.Current_Task);

      A_Address_Space :=
         Search_Address_Space (Sys.Address_Spaces, A_Task.Address_Space_Name);
      Set_Text
        (Update_Tasks_Set.Cpu_Name,
         To_String (A_Address_Space.Cpu_Name));

      Update_Tasks_Set.Clist_Address_Space_Menu := Empty_Menu;
      Reset_Iterator (Sys.Address_Spaces, Address_Space_Iterator);
      loop
         Current_Element
           (Sys.Address_Spaces,
            A_Address_Space,
            Address_Space_Iterator);
         if A_Address_Space.Name = A_Task.Address_Space_Name then
            String_List.Append
              (Update_Tasks_Set.Clist_Address_Space_Menu,
               To_String (A_Address_Space.Name));
         end if;
         exit when Is_Last_Element
                      (Sys.Address_Spaces,
                       Address_Space_Iterator);

         Next_Element (Sys.Address_Spaces, Address_Space_Iterator);
      end loop;
      Reset_Iterator (Sys.Address_Spaces, Address_Space_Iterator);
      loop
         Current_Element
           (Sys.Address_Spaces,
            A_Address_Space,
            Address_Space_Iterator);
         if A_Address_Space.Name /= A_Task.Address_Space_Name then
            String_List.Append
              (Update_Tasks_Set.Clist_Address_Space_Menu,
               To_String (A_Address_Space.Name));
         end if;
         exit when Is_Last_Element
                      (Sys.Address_Spaces,
                       Address_Space_Iterator);

         Next_Element (Sys.Address_Spaces, Address_Space_Iterator);
      end loop;

      Set_Popdown_Strings
        (Update_Tasks_Set.Address_Space_Menu,
         Update_Tasks_Set.Clist_Address_Space_Menu);

      Show_All (Update_Tasks_Set);

      -- Rebuild type menu
      --
      Initialize (Location_Items);
      Add (Location_Items, To_Unbounded_String ("Periodic"));
      Add (Location_Items, To_Unbounded_String ("Aperiodic"));
      Add (Location_Items, To_Unbounded_String ("Sporadic"));
      Add (Location_Items, To_Unbounded_String ("Poisson"));
      Add (Location_Items, To_Unbounded_String ("User-defined"));

      Current_Task_Type := Gint'(Tasks_Type'Pos (A_Task.Task_Type));

      Gtk.Option_Menu.Gtk_New (Update_Tasks_Set.Task_Type_Menu);
      Update_Tasks_Set.Clist_Task_Type_Menu := Widget_SList.Null_List;

      Build_Option_Menu
        (Update_Tasks_Set.Task_Type_Menu,
         Update_Tasks_Set.Clist_Task_Type_Menu,
         Location_Items,
         Current_Task_Type,
         null);
      Show_All (Update_Tasks_Set.Task_Type_Menu);
      Attach_Defaults
        (Update_Tasks_Set.Task_Table,
         Update_Tasks_Set.Task_Type_Menu,
         1,
         2,
         1,
         2);

      Show_All (Update_Tasks_Set);

      -- Rebuild policy menu
      --
      Initialize (Location_Items);
      Add (Location_Items, To_Unbounded_String ("SCHED_FIFO"));
      Add (Location_Items, To_Unbounded_String ("SCHED_RR"));
      Add (Location_Items, To_Unbounded_String ("SCHED_OTHERS"));

      case A_Task.Policy is
         when Sched_Fifo =>
            Current_Policy := 0;
         when Sched_Rr =>
            Current_Policy := 1;
         when Sched_Others =>
            Current_Policy := 2;
      end case;

      Gtk.Option_Menu.Gtk_New (Update_Tasks_Set.Policy_Menu);
      Update_Tasks_Set.Clist_Policy_Menu := Widget_SList.Null_List;

      Build_Option_Menu
        (Update_Tasks_Set.Policy_Menu,
         Update_Tasks_Set.Clist_Policy_Menu,
         Location_Items,
         Current_Policy,
         null);
      Show_All (Update_Tasks_Set.Policy_Menu);
      Attach_Defaults
        (Update_Tasks_Set.Task_Table,
         Update_Tasks_Set.Policy_Menu,
         1,
         2,
         4,
         5);
      Show_All (Update_Tasks_Set);

      -- Seed value and predictable attribute
      --
      if (A_Task.Task_Type = Poisson_Type) or
         (A_Task.Task_Type = Parametric_Type) or
         (A_Task.Task_Type = Sporadic_Type)
      then

         if (Poisson_Task_Ptr (A_Task).Predictable) then
            Set_Active (Update_Tasks_Set.Seed_Predictable, True);
            Set_Active (Update_Tasks_Set.Seed_Unpredictable, False);
         else
            Set_Active (Update_Tasks_Set.Seed_Predictable, False);
            Set_Active (Update_Tasks_Set.Seed_Unpredictable, True);
         end if;

         Gtk.GEntry.Set_Text
           (Update_Tasks_Set.Seed_Value,
            Poisson_Task_Ptr (A_Task).Seed'Img);

      else
         Gtk.GEntry.Set_Text (Update_Tasks_Set.Seed_Value, "");
      end if;

      -- Criticality level
      --
      Gtk.GEntry.Set_Text
        (Update_Tasks_Set.Criticality,
         A_Task.Criticality'Img);

      -- Task memory information
      --
      Gtk.GEntry.Set_Text
        (Update_Tasks_Set.Text_Memory_Size,
         A_Task.Text_Memory_Size'Img);
      Gtk.GEntry.Set_Text
        (Update_Tasks_Set.Stack_Memory_Size,
         A_Task.Stack_Memory_Size'Img);

      -- Offset information
      --
      for I in  0 .. Offsets_Range'Last - 1 loop
         Gtk.GEntry.Set_Text (Update_Tasks_Set.Offset (I).Entry_Act, "");
         Gtk.GEntry.Set_Text (Update_Tasks_Set.Offset (I).Entry_Val, "");
      end loop;
      for I in  0 .. A_Task.Offset.Nb_Entries - 1 loop
         Gtk.GEntry.Set_Text
           (Update_Tasks_Set.Offset (I).Entry_Act,
            A_Task.Offset.Entries (I).Activation'Img);
         Gtk.GEntry.Set_Text
           (Update_Tasks_Set.Offset (I).Entry_Val,
            A_Task.Offset.Entries (I).Offset_Value'Img);
      end loop;

      -- User defined parameters
      --
      for I in  0 .. User_Defined_Parameters_Range'Last - 1 loop
         Set_Text (Update_Tasks_Set.Param (I).Entry_Name, "");
         Set_Text (Update_Tasks_Set.Param (I).Entry_Value, "");
      end loop;

      for I in  A_Task.Param.Nb_Entries .. User_Defined_Parameters_Range'Last -
                                           1
      loop
         Initialize (Parameter_Items (I));
         Add (Parameter_Items (I), To_Unbounded_String ("Integer"));
         Add (Parameter_Items (I), To_Unbounded_String ("Double"));
         Add (Parameter_Items (I), To_Unbounded_String ("String"));
         Add (Parameter_Items (I), To_Unbounded_String ("Boolean"));

         Destroy (Update_Tasks_Set.Param (I).Parameter_Menu);
         Update_Tasks_Set.Param (I).Clist_Parameter_Menu :=
           Widget_SList.Null_List;
         Gtk.Option_Menu.Gtk_New (Update_Tasks_Set.Param (I).Parameter_Menu);

         Build_Option_Menu
           (Update_Tasks_Set.Param (I).Parameter_Menu,
            Update_Tasks_Set.Param (I).Clist_Parameter_Menu,
            Parameter_Items (I),
            -1,
            null);
         Show_All (Update_Tasks_Set.Param (I).Parameter_Menu);
         Attach
           (Update_Tasks_Set.Parameter_Table,
            Update_Tasks_Set.Param (I).Parameter_Menu,
            1,
            2,
            1 + Guint (I),
            2 + Guint (I),
            Expand or Fill,
            0,
            0,
            0);
      end loop;
      Show_All (Update_Tasks_Set);

      for I in  0 .. A_Task.Param.Nb_Entries - 1 loop
         Set_Text
           (Update_Tasks_Set.Param (I).Entry_Name,
            To_String (A_Task.Param.Entries (I).Name));

         if A_Task.Param.Entries (I).ptype = Boolean_Parameter then
            Set_Text
              (Update_Tasks_Set.Param (I).Entry_Value,
               A_Task.Param.Entries (I).Boolean_value'Img);
            Current := 3;
         else
            if A_Task.Param.Entries (I).ptype = Integer_Parameter then
               Set_Text
                 (Update_Tasks_Set.Param (I).Entry_Value,
                  A_Task.Param.Entries (I).Integer_value'Img);
               Current := 0;
            else
               if A_Task.Param.Entries (I).ptype = Double_Parameter then
                  Set_Text
                    (Update_Tasks_Set.Param (I).Entry_Value,
                     To_String
                        (Format (A_Task.Param.Entries (I).double_value)));
                  Current := 1;
               else
                  Set_Text
                    (Update_Tasks_Set.Param (I).Entry_Value,
                     To_String (A_Task.Param.Entries (I).string_value));
                  Current := 2;
               end if;
            end if;
         end if;

         Initialize (Parameter_Items (I));
         Add (Parameter_Items (I), To_Unbounded_String ("Integer"));
         Add (Parameter_Items (I), To_Unbounded_String ("Double"));
         Add (Parameter_Items (I), To_Unbounded_String ("String"));
         Add (Parameter_Items (I), To_Unbounded_String ("Boolean"));

         Destroy (Update_Tasks_Set.Param (I).Parameter_Menu);
         Update_Tasks_Set.Param (I).Clist_Parameter_Menu :=
           Widget_SList.Null_List;
         Gtk.Option_Menu.Gtk_New (Update_Tasks_Set.Param (I).Parameter_Menu);

         Build_Option_Menu
           (Update_Tasks_Set.Param (I).Parameter_Menu,
            Update_Tasks_Set.Param (I).Clist_Parameter_Menu,
            Parameter_Items (I),
            Current,
            null);
         Show_All (Update_Tasks_Set.Param (I).Parameter_Menu);
         Attach
           (Update_Tasks_Set.Parameter_Table,
            Update_Tasks_Set.Param (I).Parameter_Menu,
            1,
            2,
            1 + Guint (I),
            2 + Guint (I),
            Expand or Fill,
            0,
            0,
            0);

      end loop;
      Show_All (Update_Tasks_Set);

   end On_Click_List;

   procedure On_Close_Pressed (Object : access Gtk_Button_Record'Class) is
   begin

      Is_New   := False;
      Is_Saved := False;

      Destroy (Update_Tasks_Set);
   end On_Close_Pressed;

   procedure On_Add_Pressed (Object : access Gtk_Button_Record'Class) is

      Ok            : Boolean := True;
      Error_Message : Unbounded_String;

   begin

      Integrity_Checks.Check_Integrity_Constraints
        (Update_Tasks_Set,
         Error_Message,
         Ok);
      if not Ok then
         Show_Message_Box (Error_Message);
         return;
      end if;

      --
      -- Update task set
      --

      Add_Task
        (Sys.Tasks,
         Integrity_Checks.Task_Name,
         Integrity_Checks.Task_Processor,
         Integrity_Checks.Task_Address_Space,
         Integrity_Checks.Task_Type,
         Integrity_Checks.Task_Start_Time,
         Integrity_Checks.Task_Capacity,
         Integrity_Checks.Task_Period,
         Integrity_Checks.Task_Deadline,
         Integrity_Checks.Task_Jitter,
         Integrity_Checks.Task_Blocking_Time,
         Integrity_Checks.Task_Priority,
         Integrity_Checks.Task_Criticality,
         Integrity_Checks.Task_Policy,
         Integrity_Checks.Task_Offset,
         Integrity_Checks.Task_Stack_Memory_Size,
         Integrity_Checks.Task_Text_Memory_Size,
         Integrity_Checks.Task_Param,
         Integrity_Checks.Task_Parametric_Rule,
         Integrity_Checks.Task_Seed_Value,
         Integrity_Checks.Task_Predictable);

      Update_Tasks_Set.Current_Task := Integrity_Checks.Task_Name;

      Show_Tasks (Update_Tasks_Set);

      if Add_New_Item /= null then
         Set_Name (Items_List (Last_Item - 1), Integrity_Checks.Task_Name);
         Refresh_Canvas (Graph_Window.Buffer);
      end if;

   exception
      when Task_Set.Invalid_Parameter =>
         begin
            Show_Message_Box (Exception_Message);
            return;
         end;
      when others =>
         begin
            Show_Message_Box (Exception_Name & Exception_Message);
            return;
         end;

   end On_Add_Pressed;

   procedure On_Seed_Pressed (Object : access Gtk_Button_Record'Class) is
      Seed : Generator;
      Nmb  : Integer;
   begin

      -- Give random value for the tasks seed
      --
      Reset (Seed);

      -- Range from integer'first .. integer'last
      --
      Nmb :=
        Integer (Double (Random (Seed)) *
                 Double (Integer'Last - Integer'First) +
                 Double (Integer'First));
      Gtk.GEntry.Set_Text (Update_Tasks_Set.Seed_Value, Nmb'Img);

   end On_Seed_Pressed;

   procedure On_Delete_Pressed (Object : access Gtk_Button_Record'Class) is

      My_Iterator : Task_Set.Generic_Task_Set.Iterator;
      A_Task      : Generic_Task_Ptr;

   begin

      if Update_Tasks_Set.Current_Task = "" then
         Show_Message_Box (Lb_Select_Task_First (Current_Language));
         return;
      end if;

      if Is_Empty (Sys.Tasks) then
         Show_Message_Box (Lb_Can_Not_Modify (Current_Language));
         return;
      end if;

      Reset_Iterator (Sys.Tasks, My_Iterator);
      if not Is_Empty (Sys.Tasks) then
         loop

            Current_Element (Sys.Tasks, A_Task, My_Iterator);
            if A_Task.Name = Update_Tasks_Set.Current_Task then

               Delete_Task (Sys, A_Task);
               Rebuild_Canvas (Graph_Window);

               -- Clear main entrys
               --
               Update_Tasks_Set.Current_Task := To_Unbounded_String ("");

               Gtk.GEntry.Set_Text (Update_Tasks_Set.Name, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Priority, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Capacity, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Jitter, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Deadline, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Criticality, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Period, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Start_Time, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Blocking_Time, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Activation_Rule, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Seed_Value, "");
               Gtk.GEntry.Set_Text (Update_Tasks_Set.Cpu_Name, "");

               -- Clear User defined parameters
               --
               for I in  0 .. User_Defined_Parameters_Range'Last - 1 loop
                  Gtk.GEntry.Set_Text
                    (Update_Tasks_Set.Param (I).Entry_Name,
                     "");
                  Gtk.GEntry.Set_Text
                    (Update_Tasks_Set.Param (I).Entry_Value,
                     "");
               end loop;

               -- Clear Offset parameters
               --
               for I in  0 .. Offsets_Range'Last - 1 loop
                  Gtk.GEntry.Set_Text
                    (Update_Tasks_Set.Offset (I).Entry_Val,
                     "");
                  Gtk.GEntry.Set_Text
                    (Update_Tasks_Set.Offset (I).Entry_Act,
                     "");
               end loop;

               exit;
            end if;

            exit when Is_Last_Element (Sys.Tasks, My_Iterator);
            Next_Element (Sys.Tasks, My_Iterator);

         end loop;
         Show_Tasks (Update_Tasks_Set);
      end if;

   end On_Delete_Pressed;

   procedure On_Cancel_Pressed (Object : access Gtk_Button_Record'Class) is
   begin
      Duplicate (Update_Tasks_Set.Undo, Sys);
      Rebuild_Canvas (Graph_Window);
      Destroy (Update_Tasks_Set);
   end On_Cancel_Pressed;

   procedure On_Advanced_Pressed (Object : access Gtk_Button_Record'Class) is
   begin
      if Visible_Is_Set (Update_Tasks_Set.Vbox_List) then
         Show_Advanced (Update_Tasks_Set, False);
      else
         Show_Advanced (Update_Tasks_Set, True);
      end if;
   end On_Advanced_Pressed;

   procedure On_Modify_Pressed (Object : access Gtk_Button_Record'Class) is

      A_Task        : Generic_Task_Ptr;
      Created_Task  : Generic_Task_Ptr;
      My_Iterator   : Task_Set.Generic_Task_Set.Iterator;
      Ok            : Boolean := True;
      Error_Message : Unbounded_String;

   begin

      if Update_Tasks_Set.Current_Task = "" then
         Show_Message_Box (Lb_Select_Task_First (Current_Language));
         return;
      end if;

      if Is_Empty (Sys.Tasks) then
         Show_Message_Box (Lb_Can_Not_Modify (Current_Language));
         return;
      end if;

      Integrity_Checks.Check_Integrity_Constraints
        (Update_Tasks_Set,
         Error_Message,
         Ok);
      if not Ok then
         Show_Message_Box (Error_Message);
         return;
      end if;

      --
      -- Update task set
      --

      Reset_Iterator (Sys.Tasks, My_Iterator);
      loop

         Current_Element (Sys.Tasks, A_Task, My_Iterator);
         if A_Task.Name = Update_Tasks_Set.Current_Task then

            Delete (Sys.Tasks, A_Task);

            Add_Task
              (Sys.Tasks,
               Created_Task,
               Integrity_Checks.Task_Name,
               Integrity_Checks.Task_Processor,
               Integrity_Checks.Task_Address_Space,
               Integrity_Checks.Task_Type,
               Integrity_Checks.Task_Start_Time,
               Integrity_Checks.Task_Capacity,
               Integrity_Checks.Task_Period,
               Integrity_Checks.Task_Deadline,
               Integrity_Checks.Task_Jitter,
               Integrity_Checks.Task_Blocking_Time,
               Integrity_Checks.Task_Priority,
               Integrity_Checks.Task_Criticality,
               Integrity_Checks.Task_Policy,
               Integrity_Checks.Task_Offset,
               Integrity_Checks.Task_Stack_Memory_Size,
               Integrity_Checks.Task_Text_Memory_Size,
               Integrity_Checks.Task_Param,
               Integrity_Checks.Task_Parametric_Rule,
               Integrity_Checks.Task_Seed_Value,
               Integrity_Checks.Task_Predictable);

            Update_All_Task_Dependencies (Sys.Dependencies, Created_Task);

            Update_Tasks_Set.Current_Task := Integrity_Checks.Task_Name;
            exit;
         end if;

         exit when Is_Last_Element (Sys.Tasks, My_Iterator);
         Next_Element (Sys.Tasks, My_Iterator);

      end loop;

      Show_Tasks (Update_Tasks_Set);

   exception
      when Task_Set.Invalid_Parameter =>
         begin
            Show_Message_Box (Exception_Message);
            return;
         end;
      when others =>
         begin
            Show_Message_Box (Exception_Name & Exception_Message);
            return;
         end;

   end On_Modify_Pressed;

   procedure On_Change_Cpu (Widget : access Gtk_Editable_Record'Class) is

      Task_Address_Space : Unbounded_String;
      A_Address_Space    : Address_Space_Ptr;

   begin
      Task_Address_Space :=
         To_Unbounded_String
           (Get_Text (Get_Entry (Update_Tasks_Set.Address_Space_Menu)));
      A_Address_Space    :=
         Search_Address_Space (Sys.Address_Spaces, Task_Address_Space);
      Set_Text
        (Update_Tasks_Set.Cpu_Name,
         To_String (A_Address_Space.Cpu_Name));
   end On_Change_Cpu;

end Update_Tasks_Set_Pkg.Callbacks;
