------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Unchecked_Deallocation;

package body Lists is

   procedure Free is new Unchecked_Deallocation (Cell, Cell_Ptr);

   procedure Initialize (L : in out List) is

      Current  : Cell_Ptr := L.Head;
      Previous : Cell_Ptr;

   begin
      while Current /= null loop
         Free (Current.Info);
         Previous := Current;
         Current  := Current.Next;
         Free (Previous);
      end loop;

      L.Head := null;
      L.Tail := null;
   end Initialize;

   procedure Add (L : in out List; E : in Element) is

      New_Cell : Cell_Ptr;

   begin
      New_Cell      := new Cell;
      New_Cell.Info := E;

      New_Cell.Next := L.Head;
      L.Head        := New_Cell;

      if (L.Tail = null) then
         L.Tail := New_Cell;
      else
         New_Cell.Next.Previous := New_Cell;
      end if;

   end Add;

   procedure Put (L : in List) is

      Current : Cell_Ptr := L.Head;

   begin
      while Current /= null loop
         Put (Current.Info);
         Current := Current.Next;
      end loop;
   end Put;

   function Get_Number_Of_Elements (L : in List) return Natural is
      Current : Cell_Ptr := L.Head;
      Size    : Natural  := 0;

   begin
      while Current /= null loop
         Size    := Size + 1;
         Current := Current.Next;
      end loop;

      return Size;

   end Get_Number_Of_Elements;

   procedure Delete (L : in out List; E : in List) is

      Current : Cell_Ptr := E.Head;

   begin
      while Current /= null loop
         Delete (L, Current.Info);
         Current := Current.Next;
      end loop;
   end Delete;

   procedure Delete (L : in out List; E : in Element) is

      Current  : Cell_Ptr := L.Head;
      Is_Found : Boolean  := False;

   begin
      while Current /= null loop
         if Current.Info = E then
            Is_Found := True;

            -- Delete the last item
            --
            if (Current.Previous = null) and (Current.Next = null) then
               L.Head := null;
               L.Tail := null;
            else

               -- Delete the Head item
               --
               if (Current.Previous = null) then
                  L.Head                := Current.Next;
                  Current.Next.Previous := null;
               else
                  -- delete the Tail item
                  --
                  if (Current.Next = null) then
                     L.Tail                := Current.Previous;
                     Current.Previous.Next := null;
                  -- Middle item
                  --
                  else
                     Current.Previous.Next := Current.Next;
                     Current.Next.Previous := Current.Previous;
                  end if;
               end if;
            end if;

            Free (Current.Info);
            Free (Current);
            exit;
         end if;

         Current := Current.Next;
      end loop;

      if not Is_Found then
         raise Element_Not_Found;
      end if;

   end Delete;

   procedure Delete (L : in out List) is

      Current  : Cell_Ptr := L.Head;
      Previous : Cell_Ptr;

   begin
      while Current /= null loop
         Free (Current.Info);
         Previous := Current;
         Current  := Current.Next;
         Free (Previous);
      end loop;

   end Delete;

   procedure Duplicate (Src : in List; Dest : in out List) is

      Current : Cell_Ptr := Src.Tail;

   begin

      while Current /= null loop

         Add (Dest, Current.Info);

         Current := Current.Previous;
      end loop;

   end Duplicate;

   function Get_Head (L : in List) return Element is
   begin
      return (L.Head.Info);
   end Get_Head;

   function Get_Tail (L : in List) return Element is
   begin
      return (L.Tail.Info);
   end Get_Tail;

   function Is_Empty (L : in List) return Boolean is
   begin
      return (L.Head = null) or (L.Tail = null);
   end Is_Empty;

   procedure Previous_Element (L : in List; My_Iterator : in out Iterator) is
   begin
      My_Iterator.Current := My_Iterator.Current.Previous;
   end Previous_Element;

   procedure Next_Element (L : in List; My_Iterator : in out Iterator) is
   begin
      My_Iterator.Current := My_Iterator.Current.Next;
   end Next_Element;

   function Is_Tail_Element
     (L           : in List;
      My_Iterator : in Iterator)
      return        Boolean
   is
   begin
      return (My_Iterator.Current = L.Tail) or (L.Tail = null);
   end Is_Tail_Element;

   function Is_Head_Element
     (L           : in List;
      My_Iterator : in Iterator)
      return        Boolean
   is
   begin
      return (My_Iterator.Current = L.Head) or (L.Head = null);
   end Is_Head_Element;

   procedure Current_Element
     (L              : in List;
      Return_Element : out Element;
      My_Iterator    : in Iterator)
   is
   begin
      Return_Element := My_Iterator.Current.Info;
   end Current_Element;

   procedure Reset_Head_Iterator
     (L           : in List;
      My_Iterator : in out Iterator)
   is
   begin
      My_Iterator.Current := L.Head;
   end Reset_Head_Iterator;

   procedure Reset_Tail_Iterator
     (L           : in List;
      My_Iterator : in out Iterator)
   is
   begin
      My_Iterator.Current := L.Tail;
   end Reset_Tail_Iterator;

end Lists;
