------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2009   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Numerics.Aux;              use Ada.Numerics.Aux;
with Ada.Strings.Unbounded;         use Ada.Strings.Unbounded;
with Ada.Strings.Unbounded.Text_IO; use Ada.Strings.Unbounded.Text_IO;
with Convert_Unbounded_Strings;
with Tables;
with Lists;
with Unchecked_Deallocation;

package Unbounded_Strings is

   -- Some standard strings constants
   --
   Unbounded_Sp : constant Unbounded_String := To_Unbounded_String (" ");

   Unbounded_Lf : constant Unbounded_String :=
      To_Unbounded_String ("" & ASCII.LF);

   Unbounded_Ht : constant Unbounded_String :=
      To_Unbounded_String ("" & ASCII.HT);

   Empty_String : constant Unbounded_String := To_Unbounded_String ("");

   ----------------------------------------------
   -- String tables
   ----------------------------------------------
   -- Maximum number of strings composing a table of strings
   --
   Max_Strings : constant Natural := 100;

   procedure Initialize (St : out Unbounded_String);

   package Strings_Table_Package is new Tables (
      Unbounded_String,
      Max_Strings,
      Put);
   use Strings_Table_Package;

   subtype Strings_Range is Strings_Table_Package.Table_Range;
   subtype Strings_Table is Strings_Table_Package.Table;

   ------------------------------------------------------
   -- Convert unbounded_strings to standard numeric types
   ------------------------------------------------------

   procedure To_Natural is new Convert_Unbounded_Strings (Natural, 0);
   procedure To_Integer is new Convert_Unbounded_Strings (Integer, 0);
   procedure To_Boolean is new Convert_Unbounded_Strings (Boolean, True);

   procedure To_Double
     (From : in Unbounded_String;
      To   : out Double;
      Ok   : out Boolean);

   function To_Upper (S : Unbounded_String) return Unbounded_String;

   function To_Lower (S : Unbounded_String) return Unbounded_String;

   function To_Upper (S : String) return Unbounded_String;

   function To_Lower (S : String) return Unbounded_String;

   --------------------------------------------------------------------
   -- Pointers to unbounded_string and sub programms to manipulate it
   --------------------------------------------------------------------

   type Unbounded_String_Ptr is access Unbounded_String;

   procedure Free is new Unchecked_Deallocation (
      Unbounded_String,
      Unbounded_String_Ptr);

   procedure Put (S : Unbounded_String_Ptr);

   ------------------------------------------------------
   -- List of unbounded_string
   ------------------------------------------------------

   package Unbounded_String_List_Package is new Lists (
      Unbounded_String_Ptr,
      Put,
      Free);
   use Unbounded_String_List_Package;
   subtype Unbounded_String_List is Unbounded_String_List_Package.List;
   subtype Unbounded_String_Iterator is Unbounded_String_List_Package.Iterator;

end Unbounded_Strings;
